/*********************************************************************
   PicoTCP. Copyright (c) 2012-2017 Altran Intelligent Systems. Some rights reserved.
   See COPYING, LICENSE.GPLv2 and LICENSE.GPLv3 for usage.

   .

   Authors: Daniele Lacamera
 *********************************************************************/

#ifndef INCLUDE_PICO_ETHERNET
#define INCLUDE_PICO_ETHERNET

#include "pico_config.h"
#include "pico_frame.h"

extern struct pico_protocol pico_proto_ethernet;

#endif /* INCLUDE_PICO_ETHERNET */
