/*
**	naomi_regs.h
*/

#pragma once

enum
{
	NAOMI_ROM_OFFSETH_addr		= 0x5f7000,
	NAOMI_ROM_OFFSETL_addr		= 0x5f7004,
	NAOMI_ROM_DATA_addr			= 0x5f7008,
	NAOMI_DMA_OFFSETH_addr		= 0x5f700C,
	NAOMI_DMA_OFFSETL_addr		= 0x5f7010,
	NAOMI_DMA_COUNT_addr		= 0x5f7014,
	// Naomi 840-0001E communication board
	NAOMI_COMM2_CTRL_addr		= 0x5f7018,
	NAOMI_COMM2_OFFSET_addr		= 0x5f701C,
	NAOMI_COMM2_DATA_addr		= 0x5F7020,
	NAOMI_COMM2_STATUS0_addr	= 0x5F7024,
	NAOMI_COMM2_STATUS1_addr	= 0x5F7028,

	NAOMI_DIMM_COMMAND			= 0x5f703c,
	NAOMI_DIMM_OFFSETL			= 0x5f7040,
	NAOMI_DIMM_PARAMETERL		= 0x5f7044,
	NAOMI_DIMM_PARAMETERH		= 0x5f7048,
	NAOMI_DIMM_STATUS			= 0x5f704C,

	NAOMI_LED_addr				= 0x5F7068,
	NAOMI_BOARDID_WRITE_addr	= 0x5F7078,
	NAOMI_BOARDID_READ_addr		= 0x5F707C,

	NAOMI_MBOARD_OFFSET_addr    = 0x5F7050,
	NAOMI_MBOARD_DATA_addr      = 0x5F7054,
	NAOMI_MBOARD_STATUS_addr    = 0x5F705C,
};
