/*
	Created on: Nov 3, 2018

	Copyright 2018 flyinghead

	This file is part of reicast.

    reicast is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    reicast is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with reicast.  If not, see <https://www.gnu.org/licenses/>.
 */
#pragma once

enum
{
	AW_EPR_OFFSETL_addr				= 0x5f7000,
	AW_EPR_OFFSETH_addr				= 0x5f7004,
	AW_MPR_RECORD_INDEX_addr		= 0x5f700C,
	AW_MPR_FIRST_FILE_INDEX_addr	= 0x5f7010,
	AW_MPR_FILE_OFFSETL_addr		= 0x5f7014,
	AW_MPR_FILE_OFFSETH_addr		= 0x5f7018,
	AW_PIO_DATA_addr				= 0x5f7080,
};
