/*
    This file is part of Flycast.

    Flycast is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Flycast is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Flycast.  If not, see <https://www.gnu.org/licenses/>.
 */
#ifndef LIBRETRO_CORE_OPTIONS_INTL_H__
#define LIBRETRO_CORE_OPTIONS_INTL_H__

#if defined(_MSC_VER) && (_MSC_VER >= 1500 && _MSC_VER < 1900)
/* https://support.microsoft.com/en-us/kb/980263 */
#pragma execution_character_set("utf-8")
#pragma warning(disable:4566)
#endif

#include <libretro.h>

#include "libretro_core_option_defines.h"

/*
 ********************************
 * VERSION: 2.0
 ********************************
 *
 * - 2.0: Add support for core options v2 interface
 * - 1.3: Move translations to libretro_core_options_intl.h
 *        - libretro_core_options_intl.h includes BOM and utf-8
 *          fix for MSVC 2010-2013
 *        - Added HAVE_NO_LANGEXTRA flag to disable translations
 *          on platforms/compilers without BOM support
 * - 1.2: Use core options v1 interface when
 *        RETRO_ENVIRONMENT_GET_CORE_OPTIONS_VERSION is >= 1
 *        (previously required RETRO_ENVIRONMENT_GET_CORE_OPTIONS_VERSION == 1)
 * - 1.1: Support generation of core options v0 retro_core_option_value
 *        arrays containing options with a single value
 * - 1.0: First commit
*/

#ifdef __cplusplus
extern "C" {
#endif

/*
 ********************************
 * Core Option Definitions
 ********************************
*/
/* RETRO_LANGUAGE_AR */

#define CATEGORY_SYSTEM_LABEL_AR "نظام"
#define CATEGORY_SYSTEM_INFO_0_AR NULL
#define CATEGORY_VIDEO_LABEL_AR "فيديو"
#define CATEGORY_VIDEO_INFO_0_AR NULL
#define CATEGORY_PERFORMANCE_LABEL_AR "الأداء"
#define CATEGORY_PERFORMANCE_INFO_0_AR NULL
#define CATEGORY_HACKS_LABEL_AR NULL
#define CATEGORY_HACKS_INFO_0_AR NULL
#define CATEGORY_INPUT_LABEL_AR "الإدخال"
#define CATEGORY_INPUT_INFO_0_AR NULL
#define CATEGORY_EXPANSIONS_LABEL_AR NULL
#define CATEGORY_EXPANSIONS_INFO_0_AR NULL
#define CATEGORY_VMU_LABEL_AR NULL
#define CATEGORY_VMU_INFO_0_AR NULL
#define CORE_OPTION_NAME_REGION_LABEL_AR "المنطقة"
#define OPTION_VAL_JAPAN_AR NULL
#define OPTION_VAL_USA_AR NULL
#define OPTION_VAL_EUROPE_AR NULL
#define OPTION_VAL_DEFAULT_AR "الافتراضي"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_AR "اللغة"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_AR NULL
#define OPTION_VAL_JAPANESE_AR "Japanese - 日本語"
#define OPTION_VAL_ENGLISH_AR NULL
#define OPTION_VAL_GERMAN_AR "German - Deutsch"
#define OPTION_VAL_FRENCH_AR "French - Français"
#define OPTION_VAL_SPANISH_AR "Spanish - Español"
#define OPTION_VAL_ITALIAN_AR "Italian - Italiano"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_AR NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_AR NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_AR NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_AR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_AR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_AR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_AR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_AR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_AR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_AR NULL
#define CORE_OPTION_NAME_UPNP_LABEL_AR NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_AR NULL
#define CORE_OPTION_NAME_DCNET_LABEL_AR NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_AR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_AR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_AR NULL
#define OPTION_VAL_320X240_AR NULL
#define OPTION_VAL_640X480_AR NULL
#define OPTION_VAL_800X600_AR NULL
#define OPTION_VAL_960X720_AR NULL
#define OPTION_VAL_1024X768_AR NULL
#define OPTION_VAL_1280X960_AR NULL
#define OPTION_VAL_1440X1080_AR NULL
#define OPTION_VAL_1600X1200_AR NULL
#define OPTION_VAL_1920X1440_AR NULL
#define OPTION_VAL_2560X1920_AR NULL
#define OPTION_VAL_2880X2160_AR NULL
#define OPTION_VAL_3200X2400_AR NULL
#define OPTION_VAL_3840X2880_AR NULL
#define OPTION_VAL_4480X3360_AR NULL
#define OPTION_VAL_5120X3840_AR NULL
#define OPTION_VAL_5760X4320_AR NULL
#define OPTION_VAL_6400X4800_AR NULL
#define OPTION_VAL_7040X5280_AR NULL
#define OPTION_VAL_7680X5760_AR NULL
#define OPTION_VAL_8320X6240_AR NULL
#define OPTION_VAL_8960X6720_AR NULL
#define OPTION_VAL_9600X7200_AR NULL
#define OPTION_VAL_10240X7680_AR NULL
#define OPTION_VAL_10880X8160_AR NULL
#define OPTION_VAL_11520X8640_AR NULL
#define OPTION_VAL_12160X9120_AR NULL
#define OPTION_VAL_12800X9600_AR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_AR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_AR NULL
#define OPTION_VAL_VGA_AR NULL
#define OPTION_VAL_TV_RGB_AR NULL
#define OPTION_VAL_TV_COMPOSITE_AR NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_AR NULL
#define OPTION_VAL_NTSC_AR NULL
#define OPTION_VAL_PAL_AR NULL
#define OPTION_VAL_PAL_N_AR NULL
#define OPTION_VAL_PAL_M_AR NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_AR "اتجاه الشاشة"
#define OPTION_VAL_HORIZONTAL_AR NULL
#define OPTION_VAL_VERTICAL_AR NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_AR NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_AR NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_AR NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_AR NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_AR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_AR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_AR NULL
#define OPTION_VAL_512MB_AR NULL
#define OPTION_VAL_1GB_AR NULL
#define OPTION_VAL_2GB_AR NULL
#define OPTION_VAL_4GB_AR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_AR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_AR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_AR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_AR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_AR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_AR NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_AR NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_AR NULL
#define CORE_OPTION_NAME_FOG_LABEL_AR NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_AR NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_AR NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_AR NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_AR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_AR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_AR NULL
#define OPTION_VAL_1_AR NULL
#define OPTION_VAL_2_AR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_AR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_AR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_AR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_AR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_AR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_AR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_AR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_AR NULL
#define OPTION_VAL_2_O26_AR NULL
#define OPTION_VAL_4_AR NULL
#define OPTION_VAL_6_AR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_AR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_AR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_AR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_AR NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_AR NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_AR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_AR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_AR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_AR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_AR NULL
#define OPTION_VAL_SOME_AR "عادي"
#define OPTION_VAL_MORE_AR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_AR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_AR NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_AR NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_AR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_AR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_AR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_AR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_AR NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_AR NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_AR NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_AR NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_AR NULL
#define OPTION_VAL_100_AR NULL
#define OPTION_VAL_110_AR NULL
#define OPTION_VAL_120_AR NULL
#define OPTION_VAL_130_AR NULL
#define OPTION_VAL_140_AR NULL
#define OPTION_VAL_150_AR NULL
#define OPTION_VAL_160_AR NULL
#define OPTION_VAL_170_AR NULL
#define OPTION_VAL_180_AR NULL
#define OPTION_VAL_190_AR NULL
#define OPTION_VAL_200_AR NULL
#define OPTION_VAL_210_AR NULL
#define OPTION_VAL_220_AR NULL
#define OPTION_VAL_230_AR NULL
#define OPTION_VAL_240_AR NULL
#define OPTION_VAL_250_AR NULL
#define OPTION_VAL_260_AR NULL
#define OPTION_VAL_270_AR NULL
#define OPTION_VAL_280_AR NULL
#define OPTION_VAL_290_AR NULL
#define OPTION_VAL_300_AR NULL
#define OPTION_VAL_310_AR NULL
#define OPTION_VAL_320_AR NULL
#define OPTION_VAL_330_AR NULL
#define OPTION_VAL_340_AR NULL
#define OPTION_VAL_350_AR NULL
#define OPTION_VAL_360_AR NULL
#define OPTION_VAL_370_AR NULL
#define OPTION_VAL_380_AR NULL
#define OPTION_VAL_390_AR NULL
#define OPTION_VAL_400_AR NULL
#define OPTION_VAL_410_AR NULL
#define OPTION_VAL_420_AR NULL
#define OPTION_VAL_430_AR NULL
#define OPTION_VAL_440_AR NULL
#define OPTION_VAL_450_AR NULL
#define OPTION_VAL_460_AR NULL
#define OPTION_VAL_470_AR NULL
#define OPTION_VAL_480_AR NULL
#define OPTION_VAL_490_AR NULL
#define OPTION_VAL_500_AR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_AR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_AR NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_AR NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_AR NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_AR NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_AR NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_AR NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_AR NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_AR NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_AR NULL
#define OPTION_VAL_0_AR NULL
#define OPTION_VAL_5_AR NULL
#define OPTION_VAL_10_AR NULL
#define OPTION_VAL_15_AR NULL
#define OPTION_VAL_20_AR NULL
#define OPTION_VAL_25_AR NULL
#define OPTION_VAL_30_AR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_AR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_AR NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_AR NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_AR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_AR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_AR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_AR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_AR NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_AR NULL
#define OPTION_VAL_50_AR NULL
#define OPTION_VAL_60_AR NULL
#define OPTION_VAL_70_AR NULL
#define OPTION_VAL_80_AR NULL
#define OPTION_VAL_90_AR NULL
#define OPTION_VAL_100_O47_AR NULL
#define OPTION_VAL_110_O47_AR NULL
#define OPTION_VAL_120_O47_AR NULL
#define OPTION_VAL_130_O47_AR NULL
#define OPTION_VAL_140_O47_AR NULL
#define OPTION_VAL_150_O47_AR NULL
#define OPTION_VAL_160_O47_AR NULL
#define OPTION_VAL_170_O47_AR NULL
#define OPTION_VAL_180_O47_AR NULL
#define OPTION_VAL_190_O47_AR NULL
#define OPTION_VAL_200_O47_AR NULL
#define OPTION_VAL_210_O47_AR NULL
#define OPTION_VAL_220_O47_AR NULL
#define OPTION_VAL_230_O47_AR NULL
#define OPTION_VAL_240_O47_AR NULL
#define OPTION_VAL_250_O47_AR NULL
#define OPTION_VAL_260_O47_AR NULL
#define OPTION_VAL_270_O47_AR NULL
#define OPTION_VAL_280_O47_AR NULL
#define OPTION_VAL_290_O47_AR NULL
#define OPTION_VAL_300_O47_AR NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_AR NULL
#define OPTION_VAL_WHITE_AR NULL
#define OPTION_VAL_RED_AR "أحمر"
#define OPTION_VAL_GREEN_AR "أخضر"
#define OPTION_VAL_BLUE_AR "أزرق"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_AR NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_AR NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_AR NULL
#define OPTION_VAL_VMU_AR NULL
#define OPTION_VAL_PURUPURU_AR NULL
#define OPTION_VAL_DREAMPOTATO_AR NULL
#define OPTION_VAL_NONE_AR "لاشيء"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_AR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_AR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_AR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_AR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_AR NULL
#define OPTION_VAL_VMU_A1_AR NULL
#define OPTION_VAL_ALL_VMUS_AR NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_AR NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_AR NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_AR NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_AR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_AR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_AR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_AR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_AR NULL
#define OPTION_VAL_UPPER_LEFT_AR NULL
#define OPTION_VAL_UPPER_RIGHT_AR NULL
#define OPTION_VAL_LOWER_LEFT_AR NULL
#define OPTION_VAL_LOWER_RIGHT_AR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_AR NULL
#define OPTION_VAL_1X_AR NULL
#define OPTION_VAL_3X_AR NULL
#define OPTION_VAL_5X_AR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_AR NULL
#define OPTION_VAL_DEFAULT_ON_00_AR NULL
#define OPTION_VAL_DEFAULT_OFF_01_AR NULL
#define OPTION_VAL_BLACK_02_AR NULL
#define OPTION_VAL_LIGHT_BLUE_04_AR NULL
#define OPTION_VAL_CYAN_06_AR NULL
#define OPTION_VAL_CYAN_BLUE_07_AR NULL
#define OPTION_VAL_LIGHT_GREEN_08_AR NULL
#define OPTION_VAL_CYAN_GREEN_09_AR NULL
#define OPTION_VAL_LIGHT_CYAN_10_AR NULL
#define OPTION_VAL_PURPLE_12_AR NULL
#define OPTION_VAL_LIGHT_PURPLE_13_AR NULL
#define OPTION_VAL_YELLOW_14_AR "أصفر"
#define OPTION_VAL_GRAY_15_AR NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_AR NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_AR NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_AR NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_AR NULL
#define OPTION_VAL_LIGHT_RED_2_20_AR NULL
#define OPTION_VAL_MAGENTA_21_AR NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_AR NULL
#define OPTION_VAL_LIGHT_ORANGE_23_AR NULL
#define OPTION_VAL_ORANGE_24_AR NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_AR NULL
#define OPTION_VAL_LIGHT_YELLOW_26_AR NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_AR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_AR NULL
#define OPTION_VAL_40_AR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_AR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_AR NULL

struct retro_core_option_v2_category option_cats_ar[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_AR,
      CATEGORY_SYSTEM_INFO_0_AR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_AR,
      CATEGORY_VIDEO_INFO_0_AR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_AR,
      CATEGORY_PERFORMANCE_INFO_0_AR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_AR,
      CATEGORY_HACKS_INFO_0_AR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_AR,
      CATEGORY_INPUT_INFO_0_AR
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_AR,
      CATEGORY_EXPANSIONS_INFO_0_AR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_AR,
      CATEGORY_VMU_INFO_0_AR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ar[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_AR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_AR },
         { "USA",     OPTION_VAL_USA_AR },
         { "Europe",  OPTION_VAL_EUROPE_AR },
         { "Default", OPTION_VAL_DEFAULT_AR },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_AR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_AR },
         { "English",  OPTION_VAL_ENGLISH_AR },
         { "German",   OPTION_VAL_GERMAN_AR },
         { "French",   OPTION_VAL_FRENCH_AR },
         { "Spanish",  OPTION_VAL_SPANISH_AR },
         { "Italian",  OPTION_VAL_ITALIAN_AR },
         { "Default",  OPTION_VAL_DEFAULT_AR },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_AR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_AR },
         { "640x480",    OPTION_VAL_640X480_AR },
         { "800x600",    OPTION_VAL_800X600_AR },
         { "960x720",    OPTION_VAL_960X720_AR },
         { "1024x768",   OPTION_VAL_1024X768_AR },
         { "1280x960",   OPTION_VAL_1280X960_AR },
         { "1440x1080",  OPTION_VAL_1440X1080_AR },
         { "1600x1200",  OPTION_VAL_1600X1200_AR },
         { "1920x1440",  OPTION_VAL_1920X1440_AR },
         { "2560x1920",  OPTION_VAL_2560X1920_AR },
         { "2880x2160",  OPTION_VAL_2880X2160_AR },
         { "3200x2400",  OPTION_VAL_3200X2400_AR },
         { "3840x2880",  OPTION_VAL_3840X2880_AR },
         { "4480x3360",  OPTION_VAL_4480X3360_AR },
         { "5120x3840",  OPTION_VAL_5120X3840_AR },
         { "5760x4320",  OPTION_VAL_5760X4320_AR },
         { "6400x4800",  OPTION_VAL_6400X4800_AR },
         { "7040x5280",  OPTION_VAL_7040X5280_AR },
         { "7680x5760",  OPTION_VAL_7680X5760_AR },
         { "8320x6240",  OPTION_VAL_8320X6240_AR },
         { "8960x6720",  OPTION_VAL_8960X6720_AR },
         { "9600x7200",  OPTION_VAL_9600X7200_AR },
         { "10240x7680", OPTION_VAL_10240X7680_AR },
         { "10880x8160", OPTION_VAL_10880X8160_AR },
         { "11520x8640", OPTION_VAL_11520X8640_AR },
         { "12160x9120", OPTION_VAL_12160X9120_AR },
         { "12800x9600", OPTION_VAL_12800X9600_AR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_AR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_AR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_AR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_AR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_AR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_AR },
         { "PAL",     OPTION_VAL_PAL_AR },
         { "PAL_N",   OPTION_VAL_PAL_N_AR },
         { "PAL_M",   OPTION_VAL_PAL_M_AR },
         { "Default", OPTION_VAL_DEFAULT_AR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_AR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_AR },
         { "vertical",   OPTION_VAL_VERTICAL_AR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_AR,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_AR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_AR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_AR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_AR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_AR },
         { "1GB",   OPTION_VAL_1GB_AR },
         { "2GB",   OPTION_VAL_2GB_AR },
         { "4GB",   OPTION_VAL_4GB_AR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_AR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_AR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_AR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_AR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_AR },
         { "1",  OPTION_VAL_1_AR },
         { "2",  OPTION_VAL_2_AR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_AR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_AR },
         { "4",  OPTION_VAL_4_AR },
         { "6",  OPTION_VAL_6_AR },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_AR,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_AR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_AR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_AR,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_AR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_AR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_AR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_AR },
         { "more", OPTION_VAL_MORE_AR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_AR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_AR },
         { "110", OPTION_VAL_110_AR },
         { "120", OPTION_VAL_120_AR },
         { "130", OPTION_VAL_130_AR },
         { "140", OPTION_VAL_140_AR },
         { "150", OPTION_VAL_150_AR },
         { "160", OPTION_VAL_160_AR },
         { "170", OPTION_VAL_170_AR },
         { "180", OPTION_VAL_180_AR },
         { "190", OPTION_VAL_190_AR },
         { "200", OPTION_VAL_200_AR },
         { "210", OPTION_VAL_210_AR },
         { "220", OPTION_VAL_220_AR },
         { "230", OPTION_VAL_230_AR },
         { "240", OPTION_VAL_240_AR },
         { "250", OPTION_VAL_250_AR },
         { "260", OPTION_VAL_260_AR },
         { "270", OPTION_VAL_270_AR },
         { "280", OPTION_VAL_280_AR },
         { "290", OPTION_VAL_290_AR },
         { "300", OPTION_VAL_300_AR },
         { "310", OPTION_VAL_310_AR },
         { "320", OPTION_VAL_320_AR },
         { "330", OPTION_VAL_330_AR },
         { "340", OPTION_VAL_340_AR },
         { "350", OPTION_VAL_350_AR },
         { "360", OPTION_VAL_360_AR },
         { "370", OPTION_VAL_370_AR },
         { "380", OPTION_VAL_380_AR },
         { "390", OPTION_VAL_390_AR },
         { "400", OPTION_VAL_400_AR },
         { "410", OPTION_VAL_410_AR },
         { "420", OPTION_VAL_420_AR },
         { "430", OPTION_VAL_430_AR },
         { "440", OPTION_VAL_440_AR },
         { "450", OPTION_VAL_450_AR },
         { "460", OPTION_VAL_460_AR },
         { "470", OPTION_VAL_470_AR },
         { "480", OPTION_VAL_480_AR },
         { "490", OPTION_VAL_490_AR },
         { "500", OPTION_VAL_500_AR },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_AR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_AR },
         { "5%",  OPTION_VAL_5_AR },
         { "10%", OPTION_VAL_10_AR },
         { "15%", OPTION_VAL_15_AR },
         { "20%", OPTION_VAL_20_AR },
         { "25%", OPTION_VAL_25_AR },
         { "30%", OPTION_VAL_30_AR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_AR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_AR },
         { "5%",  OPTION_VAL_5_AR },
         { "10%", OPTION_VAL_10_AR },
         { "15%", OPTION_VAL_15_AR },
         { "20%", OPTION_VAL_20_AR },
         { "25%", OPTION_VAL_25_AR },
         { "30%", OPTION_VAL_30_AR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_AR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_AR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_AR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_AR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_AR },
         { "60%",  OPTION_VAL_60_AR },
         { "70%",  OPTION_VAL_70_AR },
         { "80%",  OPTION_VAL_80_AR },
         { "90%",  OPTION_VAL_90_AR },
         { "100%", OPTION_VAL_100_O47_AR },
         { "110%", OPTION_VAL_110_O47_AR },
         { "120%", OPTION_VAL_120_O47_AR },
         { "130%", OPTION_VAL_130_O47_AR },
         { "140%", OPTION_VAL_140_O47_AR },
         { "150%", OPTION_VAL_150_O47_AR },
         { "160%", OPTION_VAL_160_O47_AR },
         { "170%", OPTION_VAL_170_O47_AR },
         { "180%", OPTION_VAL_180_O47_AR },
         { "190%", OPTION_VAL_190_O47_AR },
         { "200%", OPTION_VAL_200_O47_AR },
         { "210%", OPTION_VAL_210_O47_AR },
         { "220%", OPTION_VAL_220_O47_AR },
         { "230%", OPTION_VAL_230_O47_AR },
         { "240%", OPTION_VAL_240_O47_AR },
         { "250%", OPTION_VAL_250_O47_AR },
         { "260%", OPTION_VAL_260_O47_AR },
         { "270%", OPTION_VAL_270_O47_AR },
         { "280%", OPTION_VAL_280_O47_AR },
         { "290%", OPTION_VAL_290_O47_AR },
         { "300%", OPTION_VAL_300_O47_AR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_AR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AR },
         { "Red",      OPTION_VAL_RED_AR },
         { "Green",    OPTION_VAL_GREEN_AR },
         { "Blue",     OPTION_VAL_BLUE_AR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_AR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AR },
         { "Red",      OPTION_VAL_RED_AR },
         { "Green",    OPTION_VAL_GREEN_AR },
         { "Blue",     OPTION_VAL_BLUE_AR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_AR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AR },
         { "Red",      OPTION_VAL_RED_AR },
         { "Green",    OPTION_VAL_GREEN_AR },
         { "Blue",     OPTION_VAL_BLUE_AR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_AR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AR },
         { "Red",      OPTION_VAL_RED_AR },
         { "Green",    OPTION_VAL_GREEN_AR },
         { "Blue",     OPTION_VAL_BLUE_AR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_AR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AR },
         { "Purupuru", OPTION_VAL_PURUPURU_AR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_AR },
         { "None",     OPTION_VAL_NONE_AR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_AR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AR },
         { "Purupuru", OPTION_VAL_PURUPURU_AR },
         { "None",     OPTION_VAL_NONE_AR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_AR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AR },
         { "Purupuru", OPTION_VAL_PURUPURU_AR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_AR },
         { "None",     OPTION_VAL_NONE_AR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_AR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AR },
         { "Purupuru", OPTION_VAL_PURUPURU_AR },
         { "None",     OPTION_VAL_NONE_AR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_AR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AR },
         { "Purupuru", OPTION_VAL_PURUPURU_AR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_AR },
         { "None",     OPTION_VAL_NONE_AR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_AR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AR },
         { "Purupuru", OPTION_VAL_PURUPURU_AR },
         { "None",     OPTION_VAL_NONE_AR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_AR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AR },
         { "Purupuru", OPTION_VAL_PURUPURU_AR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_AR },
         { "None",     OPTION_VAL_NONE_AR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_AR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AR },
         { "Purupuru", OPTION_VAL_PURUPURU_AR },
         { "None",     OPTION_VAL_NONE_AR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_AR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_AR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_AR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_AR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_AR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_AR,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_AR,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_AR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_AR,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_AR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_AR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_AR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_AR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AR },
         { "2x", OPTION_VAL_2_O26_AR },
         { "3x", OPTION_VAL_3X_AR },
         { "4x", OPTION_VAL_4_AR },
         { "5x", OPTION_VAL_5X_AR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AR },
         { "20%",  OPTION_VAL_20_AR },
         { "30%",  OPTION_VAL_30_AR },
         { "40%",  OPTION_VAL_40_AR },
         { "50%",  OPTION_VAL_50_AR },
         { "60%",  OPTION_VAL_60_AR },
         { "70%",  OPTION_VAL_70_AR },
         { "80%",  OPTION_VAL_80_AR },
         { "90%",  OPTION_VAL_90_AR },
         { "100%", OPTION_VAL_100_O47_AR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AR },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AR },
         { "2x", OPTION_VAL_2_O26_AR },
         { "3x", OPTION_VAL_3X_AR },
         { "4x", OPTION_VAL_4_AR },
         { "5x", OPTION_VAL_5X_AR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AR },
         { "20%",  OPTION_VAL_20_AR },
         { "30%",  OPTION_VAL_30_AR },
         { "40%",  OPTION_VAL_40_AR },
         { "50%",  OPTION_VAL_50_AR },
         { "60%",  OPTION_VAL_60_AR },
         { "70%",  OPTION_VAL_70_AR },
         { "80%",  OPTION_VAL_80_AR },
         { "90%",  OPTION_VAL_90_AR },
         { "100%", OPTION_VAL_100_O47_AR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AR },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AR },
         { "2x", OPTION_VAL_2_O26_AR },
         { "3x", OPTION_VAL_3X_AR },
         { "4x", OPTION_VAL_4_AR },
         { "5x", OPTION_VAL_5X_AR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AR },
         { "20%",  OPTION_VAL_20_AR },
         { "30%",  OPTION_VAL_30_AR },
         { "40%",  OPTION_VAL_40_AR },
         { "50%",  OPTION_VAL_50_AR },
         { "60%",  OPTION_VAL_60_AR },
         { "70%",  OPTION_VAL_70_AR },
         { "80%",  OPTION_VAL_80_AR },
         { "90%",  OPTION_VAL_90_AR },
         { "100%", OPTION_VAL_100_O47_AR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AR },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AR },
         { "2x", OPTION_VAL_2_O26_AR },
         { "3x", OPTION_VAL_3X_AR },
         { "4x", OPTION_VAL_4_AR },
         { "5x", OPTION_VAL_5X_AR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AR },
         { "BLUE 03",           OPTION_VAL_BLUE_AR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AR },
         { "GREEN 05",          OPTION_VAL_GREEN_AR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AR },
         { "RED 11",            OPTION_VAL_RED_AR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AR },
         { "WHITE 28",          OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_AR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AR },
         { "20%",  OPTION_VAL_20_AR },
         { "30%",  OPTION_VAL_30_AR },
         { "40%",  OPTION_VAL_40_AR },
         { "50%",  OPTION_VAL_50_AR },
         { "60%",  OPTION_VAL_60_AR },
         { "70%",  OPTION_VAL_70_AR },
         { "80%",  OPTION_VAL_80_AR },
         { "90%",  OPTION_VAL_90_AR },
         { "100%", OPTION_VAL_100_O47_AR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ar = {
   option_cats_ar,
   option_defs_ar
};

/* RETRO_LANGUAGE_AST */

#define CATEGORY_SYSTEM_LABEL_AST "Sistema"
#define CATEGORY_SYSTEM_INFO_0_AST NULL
#define CATEGORY_VIDEO_LABEL_AST "Videu"
#define CATEGORY_VIDEO_INFO_0_AST NULL
#define CATEGORY_PERFORMANCE_LABEL_AST "Rindimientu"
#define CATEGORY_PERFORMANCE_INFO_0_AST NULL
#define CATEGORY_HACKS_LABEL_AST "Hacks de la emulación"
#define CATEGORY_HACKS_INFO_0_AST NULL
#define CATEGORY_INPUT_LABEL_AST "Controles"
#define CATEGORY_INPUT_INFO_0_AST NULL
#define CATEGORY_EXPANSIONS_LABEL_AST NULL
#define CATEGORY_EXPANSIONS_INFO_0_AST NULL
#define CATEGORY_VMU_LABEL_AST NULL
#define CATEGORY_VMU_INFO_0_AST NULL
#define CORE_OPTION_NAME_REGION_LABEL_AST "Rexón"
#define OPTION_VAL_JAPAN_AST "Xapón"
#define OPTION_VAL_USA_AST "Estaos Xuníos"
#define OPTION_VAL_EUROPE_AST "Europa"
#define OPTION_VAL_DEFAULT_AST NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_AST "Llingua"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_AST NULL
#define OPTION_VAL_JAPANESE_AST "Xaponés"
#define OPTION_VAL_ENGLISH_AST "Inglés"
#define OPTION_VAL_GERMAN_AST "Alemán"
#define OPTION_VAL_FRENCH_AST "Francés"
#define OPTION_VAL_SPANISH_AST "Castellán"
#define OPTION_VAL_ITALIAN_AST "Italianu"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_AST NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_AST NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_AST "Activar el DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_AST "Activa la emulación del procesador de señal dixital d'audiu (DSP) de Dreamcast. Ameyora la precisión del soníu xeneráu mas aumenta los requirimientos de rindimientu."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_AST NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_AST NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_AST NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_AST NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_AST NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_AST NULL
#define CORE_OPTION_NAME_UPNP_LABEL_AST NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_AST NULL
#define CORE_OPTION_NAME_DCNET_LABEL_AST NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_AST NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_AST "Resolución interna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_AST "Modifica la resolución de la renderización."
#define OPTION_VAL_320X240_AST NULL
#define OPTION_VAL_640X480_AST NULL
#define OPTION_VAL_800X600_AST "800x600 (x1,25)"
#define OPTION_VAL_960X720_AST "960x720 (x1,5)"
#define OPTION_VAL_1024X768_AST "1024x768 (x1,6)"
#define OPTION_VAL_1280X960_AST NULL
#define OPTION_VAL_1440X1080_AST "1440x1080 (x2,25)"
#define OPTION_VAL_1600X1200_AST "1600x1200 (x2,5)"
#define OPTION_VAL_1920X1440_AST NULL
#define OPTION_VAL_2560X1920_AST NULL
#define OPTION_VAL_2880X2160_AST "2880x2160 (x4,5)"
#define OPTION_VAL_3200X2400_AST NULL
#define OPTION_VAL_3840X2880_AST NULL
#define OPTION_VAL_4480X3360_AST NULL
#define OPTION_VAL_5120X3840_AST NULL
#define OPTION_VAL_5760X4320_AST NULL
#define OPTION_VAL_6400X4800_AST NULL
#define OPTION_VAL_7040X5280_AST NULL
#define OPTION_VAL_7680X5760_AST NULL
#define OPTION_VAL_8320X6240_AST NULL
#define OPTION_VAL_8960X6720_AST NULL
#define OPTION_VAL_9600X7200_AST NULL
#define OPTION_VAL_10240X7680_AST NULL
#define OPTION_VAL_10880X8160_AST NULL
#define OPTION_VAL_11520X8640_AST NULL
#define OPTION_VAL_12160X9120_AST NULL
#define OPTION_VAL_12800X9600_AST NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_AST "Tipu de cable"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_AST NULL
#define OPTION_VAL_VGA_AST NULL
#define OPTION_VAL_TV_RGB_AST NULL
#define OPTION_VAL_TV_COMPOSITE_AST "TV (compuestu)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_AST NULL
#define OPTION_VAL_NTSC_AST NULL
#define OPTION_VAL_PAL_AST NULL
#define OPTION_VAL_PAL_N_AST "PAL-N (Arxentina, Paraguái ya Uruguái)"
#define OPTION_VAL_PAL_M_AST "PAL-M (Brasil)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_AST "Orientación de la pantalla"
#define OPTION_VAL_HORIZONTAL_AST NULL
#define OPTION_VAL_VERTICAL_AST NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_AST NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_AST NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_AST NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_AST NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_AST NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_AST NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_AST NULL
#define OPTION_VAL_512MB_AST NULL
#define OPTION_VAL_1GB_AST NULL
#define OPTION_VAL_2GB_AST NULL
#define OPTION_VAL_4GB_AST NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_AST NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_AST NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_AST NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_AST NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_AST NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_AST NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_AST NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_AST NULL
#define CORE_OPTION_NAME_FOG_LABEL_AST "Efeutos de nublina"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_AST "Modificador del volume"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_AST NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_AST "Filtru anisotrópicu"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_AST NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_AST NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_AST NULL
#define OPTION_VAL_1_AST NULL
#define OPTION_VAL_2_AST NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_AST NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_AST NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_AST NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_AST NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_AST NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_AST NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_AST NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_AST NULL
#define OPTION_VAL_2_O26_AST "x2"
#define OPTION_VAL_4_AST "x4"
#define OPTION_VAL_6_AST "x6"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_AST NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_AST NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_AST NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_AST NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_AST NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_AST NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_AST NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_AST NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_AST NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_AST NULL
#define OPTION_VAL_SOME_AST NULL
#define OPTION_VAL_MORE_AST NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_AST NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_AST NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_AST NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_AST NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_AST NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_AST NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_AST NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_AST "Aumenta la velocidá de carga de los GD-ROMs."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_AST NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_AST NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_AST NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_AST NULL
#define OPTION_VAL_100_AST NULL
#define OPTION_VAL_110_AST NULL
#define OPTION_VAL_120_AST NULL
#define OPTION_VAL_130_AST NULL
#define OPTION_VAL_140_AST NULL
#define OPTION_VAL_150_AST NULL
#define OPTION_VAL_160_AST NULL
#define OPTION_VAL_170_AST NULL
#define OPTION_VAL_180_AST NULL
#define OPTION_VAL_190_AST NULL
#define OPTION_VAL_200_AST NULL
#define OPTION_VAL_210_AST NULL
#define OPTION_VAL_220_AST NULL
#define OPTION_VAL_230_AST NULL
#define OPTION_VAL_240_AST NULL
#define OPTION_VAL_250_AST NULL
#define OPTION_VAL_260_AST NULL
#define OPTION_VAL_270_AST NULL
#define OPTION_VAL_280_AST NULL
#define OPTION_VAL_290_AST NULL
#define OPTION_VAL_300_AST NULL
#define OPTION_VAL_310_AST NULL
#define OPTION_VAL_320_AST NULL
#define OPTION_VAL_330_AST NULL
#define OPTION_VAL_340_AST NULL
#define OPTION_VAL_350_AST NULL
#define OPTION_VAL_360_AST NULL
#define OPTION_VAL_370_AST NULL
#define OPTION_VAL_380_AST NULL
#define OPTION_VAL_390_AST NULL
#define OPTION_VAL_400_AST NULL
#define OPTION_VAL_410_AST NULL
#define OPTION_VAL_420_AST NULL
#define OPTION_VAL_430_AST NULL
#define OPTION_VAL_440_AST NULL
#define OPTION_VAL_450_AST NULL
#define OPTION_VAL_460_AST NULL
#define OPTION_VAL_470_AST NULL
#define OPTION_VAL_480_AST NULL
#define OPTION_VAL_490_AST NULL
#define OPTION_VAL_500_AST NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_AST NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_AST NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_AST NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_AST NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_AST NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_AST NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_AST NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_AST NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_AST NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_AST NULL
#define OPTION_VAL_0_AST NULL
#define OPTION_VAL_5_AST NULL
#define OPTION_VAL_10_AST NULL
#define OPTION_VAL_15_AST NULL
#define OPTION_VAL_20_AST NULL
#define OPTION_VAL_25_AST NULL
#define OPTION_VAL_30_AST NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_AST NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_AST NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_AST NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_AST NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_AST NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_AST NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_AST NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_AST NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_AST NULL
#define OPTION_VAL_50_AST NULL
#define OPTION_VAL_60_AST NULL
#define OPTION_VAL_70_AST NULL
#define OPTION_VAL_80_AST NULL
#define OPTION_VAL_90_AST NULL
#define OPTION_VAL_100_O47_AST NULL
#define OPTION_VAL_110_O47_AST NULL
#define OPTION_VAL_120_O47_AST NULL
#define OPTION_VAL_130_O47_AST NULL
#define OPTION_VAL_140_O47_AST NULL
#define OPTION_VAL_150_O47_AST NULL
#define OPTION_VAL_160_O47_AST NULL
#define OPTION_VAL_170_O47_AST NULL
#define OPTION_VAL_180_O47_AST NULL
#define OPTION_VAL_190_O47_AST NULL
#define OPTION_VAL_200_O47_AST NULL
#define OPTION_VAL_210_O47_AST NULL
#define OPTION_VAL_220_O47_AST NULL
#define OPTION_VAL_230_O47_AST NULL
#define OPTION_VAL_240_O47_AST NULL
#define OPTION_VAL_250_O47_AST NULL
#define OPTION_VAL_260_O47_AST NULL
#define OPTION_VAL_270_O47_AST NULL
#define OPTION_VAL_280_O47_AST NULL
#define OPTION_VAL_290_O47_AST NULL
#define OPTION_VAL_300_O47_AST NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_AST NULL
#define OPTION_VAL_WHITE_AST "Blancu"
#define OPTION_VAL_RED_AST "Coloráu"
#define OPTION_VAL_GREEN_AST "Verde"
#define OPTION_VAL_BLUE_AST "Azul"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_AST NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_AST NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_AST NULL
#define OPTION_VAL_VMU_AST NULL
#define OPTION_VAL_PURUPURU_AST NULL
#define OPTION_VAL_DREAMPOTATO_AST NULL
#define OPTION_VAL_NONE_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_AST NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_AST NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_AST NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_AST NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_AST NULL
#define OPTION_VAL_VMU_A1_AST NULL
#define OPTION_VAL_ALL_VMUS_AST "Toles VMUs"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_AST "Soníos de la VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_AST "Cuando s'activa esta opción, reprodúcense los pitíos de la VMU."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_AST NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_AST NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_AST NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_AST NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_AST NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_AST NULL
#define OPTION_VAL_UPPER_LEFT_AST NULL
#define OPTION_VAL_UPPER_RIGHT_AST NULL
#define OPTION_VAL_LOWER_LEFT_AST NULL
#define OPTION_VAL_LOWER_RIGHT_AST NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_AST NULL
#define OPTION_VAL_1X_AST "x1"
#define OPTION_VAL_3X_AST "x3"
#define OPTION_VAL_5X_AST "x5"
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_AST NULL
#define OPTION_VAL_DEFAULT_ON_00_AST NULL
#define OPTION_VAL_DEFAULT_OFF_01_AST NULL
#define OPTION_VAL_BLACK_02_AST "Prietu"
#define OPTION_VAL_LIGHT_BLUE_04_AST "Azul claro"
#define OPTION_VAL_CYAN_06_AST "Cianu"
#define OPTION_VAL_CYAN_BLUE_07_AST "Azul ciano"
#define OPTION_VAL_LIGHT_GREEN_08_AST "Verde claro"
#define OPTION_VAL_CYAN_GREEN_09_AST "Verde ciano"
#define OPTION_VAL_LIGHT_CYAN_10_AST "Cianu claro"
#define OPTION_VAL_PURPLE_12_AST "Moráu"
#define OPTION_VAL_LIGHT_PURPLE_13_AST "Moráu claro"
#define OPTION_VAL_YELLOW_14_AST "Mariellu"
#define OPTION_VAL_GRAY_15_AST "Buxu"
#define OPTION_VAL_LIGHT_PURPLE_2_16_AST "Moráu claro (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_AST "Verde claro (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_AST "Verde claro (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_AST "Cianu claro (2)"
#define OPTION_VAL_LIGHT_RED_2_20_AST "Coloráu claro (2)"
#define OPTION_VAL_MAGENTA_21_AST "Maxenta"
#define OPTION_VAL_LIGHT_PURPLE_3_22_AST "Moráu claro (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_AST "Naranxa claro"
#define OPTION_VAL_ORANGE_24_AST "Naranxa"
#define OPTION_VAL_LIGHT_PURPLE_4_25_AST "Moráu claro (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_AST "Mariellu claro"
#define OPTION_VAL_LIGHT_YELLOW_2_27_AST "Mariellu claro (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_AST NULL
#define OPTION_VAL_40_AST NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_AST NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_AST NULL

struct retro_core_option_v2_category option_cats_ast[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_AST,
      CATEGORY_SYSTEM_INFO_0_AST
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_AST,
      CATEGORY_VIDEO_INFO_0_AST
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_AST,
      CATEGORY_PERFORMANCE_INFO_0_AST
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_AST,
      CATEGORY_HACKS_INFO_0_AST
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_AST,
      CATEGORY_INPUT_INFO_0_AST
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_AST,
      CATEGORY_EXPANSIONS_INFO_0_AST
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_AST,
      CATEGORY_VMU_INFO_0_AST
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ast[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_AST,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_AST },
         { "USA",     OPTION_VAL_USA_AST },
         { "Europe",  OPTION_VAL_EUROPE_AST },
         { "Default", OPTION_VAL_DEFAULT_AST },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_AST,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_AST },
         { "English",  OPTION_VAL_ENGLISH_AST },
         { "German",   OPTION_VAL_GERMAN_AST },
         { "French",   OPTION_VAL_FRENCH_AST },
         { "Spanish",  OPTION_VAL_SPANISH_AST },
         { "Italian",  OPTION_VAL_ITALIAN_AST },
         { "Default",  OPTION_VAL_DEFAULT_AST },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_AST,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_AST },
         { "640x480",    OPTION_VAL_640X480_AST },
         { "800x600",    OPTION_VAL_800X600_AST },
         { "960x720",    OPTION_VAL_960X720_AST },
         { "1024x768",   OPTION_VAL_1024X768_AST },
         { "1280x960",   OPTION_VAL_1280X960_AST },
         { "1440x1080",  OPTION_VAL_1440X1080_AST },
         { "1600x1200",  OPTION_VAL_1600X1200_AST },
         { "1920x1440",  OPTION_VAL_1920X1440_AST },
         { "2560x1920",  OPTION_VAL_2560X1920_AST },
         { "2880x2160",  OPTION_VAL_2880X2160_AST },
         { "3200x2400",  OPTION_VAL_3200X2400_AST },
         { "3840x2880",  OPTION_VAL_3840X2880_AST },
         { "4480x3360",  OPTION_VAL_4480X3360_AST },
         { "5120x3840",  OPTION_VAL_5120X3840_AST },
         { "5760x4320",  OPTION_VAL_5760X4320_AST },
         { "6400x4800",  OPTION_VAL_6400X4800_AST },
         { "7040x5280",  OPTION_VAL_7040X5280_AST },
         { "7680x5760",  OPTION_VAL_7680X5760_AST },
         { "8320x6240",  OPTION_VAL_8320X6240_AST },
         { "8960x6720",  OPTION_VAL_8960X6720_AST },
         { "9600x7200",  OPTION_VAL_9600X7200_AST },
         { "10240x7680", OPTION_VAL_10240X7680_AST },
         { "10880x8160", OPTION_VAL_10880X8160_AST },
         { "11520x8640", OPTION_VAL_11520X8640_AST },
         { "12160x9120", OPTION_VAL_12160X9120_AST },
         { "12800x9600", OPTION_VAL_12800X9600_AST },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_AST,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_AST },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_AST },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_AST },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_AST,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_AST },
         { "PAL",     OPTION_VAL_PAL_AST },
         { "PAL_N",   OPTION_VAL_PAL_N_AST },
         { "PAL_M",   OPTION_VAL_PAL_M_AST },
         { "Default", OPTION_VAL_DEFAULT_AST },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_AST,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_AST },
         { "vertical",   OPTION_VAL_VERTICAL_AST },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_AST,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_AST },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_AST },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_AST },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_AST,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_AST },
         { "1GB",   OPTION_VAL_1GB_AST },
         { "2GB",   OPTION_VAL_2GB_AST },
         { "4GB",   OPTION_VAL_4GB_AST },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_AST,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_AST,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_AST,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_AST,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_AST },
         { "1",  OPTION_VAL_1_AST },
         { "2",  OPTION_VAL_2_AST },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_AST,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_AST },
         { "4",  OPTION_VAL_4_AST },
         { "6",  OPTION_VAL_6_AST },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_AST,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_AST,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_AST,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_AST,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_AST,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_AST,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_AST,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_AST },
         { "more", OPTION_VAL_MORE_AST },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_AST,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_AST },
         { "110", OPTION_VAL_110_AST },
         { "120", OPTION_VAL_120_AST },
         { "130", OPTION_VAL_130_AST },
         { "140", OPTION_VAL_140_AST },
         { "150", OPTION_VAL_150_AST },
         { "160", OPTION_VAL_160_AST },
         { "170", OPTION_VAL_170_AST },
         { "180", OPTION_VAL_180_AST },
         { "190", OPTION_VAL_190_AST },
         { "200", OPTION_VAL_200_AST },
         { "210", OPTION_VAL_210_AST },
         { "220", OPTION_VAL_220_AST },
         { "230", OPTION_VAL_230_AST },
         { "240", OPTION_VAL_240_AST },
         { "250", OPTION_VAL_250_AST },
         { "260", OPTION_VAL_260_AST },
         { "270", OPTION_VAL_270_AST },
         { "280", OPTION_VAL_280_AST },
         { "290", OPTION_VAL_290_AST },
         { "300", OPTION_VAL_300_AST },
         { "310", OPTION_VAL_310_AST },
         { "320", OPTION_VAL_320_AST },
         { "330", OPTION_VAL_330_AST },
         { "340", OPTION_VAL_340_AST },
         { "350", OPTION_VAL_350_AST },
         { "360", OPTION_VAL_360_AST },
         { "370", OPTION_VAL_370_AST },
         { "380", OPTION_VAL_380_AST },
         { "390", OPTION_VAL_390_AST },
         { "400", OPTION_VAL_400_AST },
         { "410", OPTION_VAL_410_AST },
         { "420", OPTION_VAL_420_AST },
         { "430", OPTION_VAL_430_AST },
         { "440", OPTION_VAL_440_AST },
         { "450", OPTION_VAL_450_AST },
         { "460", OPTION_VAL_460_AST },
         { "470", OPTION_VAL_470_AST },
         { "480", OPTION_VAL_480_AST },
         { "490", OPTION_VAL_490_AST },
         { "500", OPTION_VAL_500_AST },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_AST,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_AST },
         { "5%",  OPTION_VAL_5_AST },
         { "10%", OPTION_VAL_10_AST },
         { "15%", OPTION_VAL_15_AST },
         { "20%", OPTION_VAL_20_AST },
         { "25%", OPTION_VAL_25_AST },
         { "30%", OPTION_VAL_30_AST },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_AST,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_AST },
         { "5%",  OPTION_VAL_5_AST },
         { "10%", OPTION_VAL_10_AST },
         { "15%", OPTION_VAL_15_AST },
         { "20%", OPTION_VAL_20_AST },
         { "25%", OPTION_VAL_25_AST },
         { "30%", OPTION_VAL_30_AST },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_AST,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_AST,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_AST,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_AST,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_AST },
         { "60%",  OPTION_VAL_60_AST },
         { "70%",  OPTION_VAL_70_AST },
         { "80%",  OPTION_VAL_80_AST },
         { "90%",  OPTION_VAL_90_AST },
         { "100%", OPTION_VAL_100_O47_AST },
         { "110%", OPTION_VAL_110_O47_AST },
         { "120%", OPTION_VAL_120_O47_AST },
         { "130%", OPTION_VAL_130_O47_AST },
         { "140%", OPTION_VAL_140_O47_AST },
         { "150%", OPTION_VAL_150_O47_AST },
         { "160%", OPTION_VAL_160_O47_AST },
         { "170%", OPTION_VAL_170_O47_AST },
         { "180%", OPTION_VAL_180_O47_AST },
         { "190%", OPTION_VAL_190_O47_AST },
         { "200%", OPTION_VAL_200_O47_AST },
         { "210%", OPTION_VAL_210_O47_AST },
         { "220%", OPTION_VAL_220_O47_AST },
         { "230%", OPTION_VAL_230_O47_AST },
         { "240%", OPTION_VAL_240_O47_AST },
         { "250%", OPTION_VAL_250_O47_AST },
         { "260%", OPTION_VAL_260_O47_AST },
         { "270%", OPTION_VAL_270_O47_AST },
         { "280%", OPTION_VAL_280_O47_AST },
         { "290%", OPTION_VAL_290_O47_AST },
         { "300%", OPTION_VAL_300_O47_AST },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_AST,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AST },
         { "Red",      OPTION_VAL_RED_AST },
         { "Green",    OPTION_VAL_GREEN_AST },
         { "Blue",     OPTION_VAL_BLUE_AST },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_AST,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AST },
         { "Red",      OPTION_VAL_RED_AST },
         { "Green",    OPTION_VAL_GREEN_AST },
         { "Blue",     OPTION_VAL_BLUE_AST },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_AST,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AST },
         { "Red",      OPTION_VAL_RED_AST },
         { "Green",    OPTION_VAL_GREEN_AST },
         { "Blue",     OPTION_VAL_BLUE_AST },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_AST,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_AST },
         { "Red",      OPTION_VAL_RED_AST },
         { "Green",    OPTION_VAL_GREEN_AST },
         { "Blue",     OPTION_VAL_BLUE_AST },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_AST,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AST },
         { "Purupuru", OPTION_VAL_PURUPURU_AST },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_AST },
         { "None",     OPTION_VAL_NONE_AST },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_AST,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AST },
         { "Purupuru", OPTION_VAL_PURUPURU_AST },
         { "None",     OPTION_VAL_NONE_AST },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_AST,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AST },
         { "Purupuru", OPTION_VAL_PURUPURU_AST },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_AST },
         { "None",     OPTION_VAL_NONE_AST },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_AST,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AST },
         { "Purupuru", OPTION_VAL_PURUPURU_AST },
         { "None",     OPTION_VAL_NONE_AST },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_AST,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AST },
         { "Purupuru", OPTION_VAL_PURUPURU_AST },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_AST },
         { "None",     OPTION_VAL_NONE_AST },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_AST,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AST },
         { "Purupuru", OPTION_VAL_PURUPURU_AST },
         { "None",     OPTION_VAL_NONE_AST },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_AST,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AST },
         { "Purupuru", OPTION_VAL_PURUPURU_AST },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_AST },
         { "None",     OPTION_VAL_NONE_AST },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_AST,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_AST },
         { "Purupuru", OPTION_VAL_PURUPURU_AST },
         { "None",     OPTION_VAL_NONE_AST },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_AST,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_AST,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_AST,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_AST },
         { "All VMUs", OPTION_VAL_ALL_VMUS_AST },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_AST,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_AST,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_AST,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_AST,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_AST,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_AST,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_AST,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_AST,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AST },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AST },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AST },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AST },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AST },
         { "2x", OPTION_VAL_2_O26_AST },
         { "3x", OPTION_VAL_3X_AST },
         { "4x", OPTION_VAL_4_AST },
         { "5x", OPTION_VAL_5X_AST },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AST },
         { "20%",  OPTION_VAL_20_AST },
         { "30%",  OPTION_VAL_30_AST },
         { "40%",  OPTION_VAL_40_AST },
         { "50%",  OPTION_VAL_50_AST },
         { "60%",  OPTION_VAL_60_AST },
         { "70%",  OPTION_VAL_70_AST },
         { "80%",  OPTION_VAL_80_AST },
         { "90%",  OPTION_VAL_90_AST },
         { "100%", OPTION_VAL_100_O47_AST },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AST },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AST },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AST },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AST },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AST },
         { "2x", OPTION_VAL_2_O26_AST },
         { "3x", OPTION_VAL_3X_AST },
         { "4x", OPTION_VAL_4_AST },
         { "5x", OPTION_VAL_5X_AST },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AST },
         { "20%",  OPTION_VAL_20_AST },
         { "30%",  OPTION_VAL_30_AST },
         { "40%",  OPTION_VAL_40_AST },
         { "50%",  OPTION_VAL_50_AST },
         { "60%",  OPTION_VAL_60_AST },
         { "70%",  OPTION_VAL_70_AST },
         { "80%",  OPTION_VAL_80_AST },
         { "90%",  OPTION_VAL_90_AST },
         { "100%", OPTION_VAL_100_O47_AST },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AST },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AST },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AST },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AST },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AST },
         { "2x", OPTION_VAL_2_O26_AST },
         { "3x", OPTION_VAL_3X_AST },
         { "4x", OPTION_VAL_4_AST },
         { "5x", OPTION_VAL_5X_AST },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AST },
         { "20%",  OPTION_VAL_20_AST },
         { "30%",  OPTION_VAL_30_AST },
         { "40%",  OPTION_VAL_40_AST },
         { "50%",  OPTION_VAL_50_AST },
         { "60%",  OPTION_VAL_60_AST },
         { "70%",  OPTION_VAL_70_AST },
         { "80%",  OPTION_VAL_80_AST },
         { "90%",  OPTION_VAL_90_AST },
         { "100%", OPTION_VAL_100_O47_AST },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_AST },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_AST },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_AST },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_AST },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_AST },
         { "2x", OPTION_VAL_2_O26_AST },
         { "3x", OPTION_VAL_3X_AST },
         { "4x", OPTION_VAL_4_AST },
         { "5x", OPTION_VAL_5X_AST },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_AST },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_AST },
         { "BLACK 02",          OPTION_VAL_BLACK_02_AST },
         { "BLUE 03",           OPTION_VAL_BLUE_AST },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_AST },
         { "GREEN 05",          OPTION_VAL_GREEN_AST },
         { "CYAN 06",           OPTION_VAL_CYAN_06_AST },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_AST },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_AST },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_AST },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_AST },
         { "RED 11",            OPTION_VAL_RED_AST },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_AST },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_AST },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_AST },
         { "GRAY 15",           OPTION_VAL_GRAY_15_AST },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_AST },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_AST },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_AST },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_AST },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_AST },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_AST },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_AST },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_AST },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_AST },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_AST },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_AST },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_AST },
         { "WHITE 28",          OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_AST,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_AST },
         { "20%",  OPTION_VAL_20_AST },
         { "30%",  OPTION_VAL_30_AST },
         { "40%",  OPTION_VAL_40_AST },
         { "50%",  OPTION_VAL_50_AST },
         { "60%",  OPTION_VAL_60_AST },
         { "70%",  OPTION_VAL_70_AST },
         { "80%",  OPTION_VAL_80_AST },
         { "90%",  OPTION_VAL_90_AST },
         { "100%", OPTION_VAL_100_O47_AST },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ast = {
   option_cats_ast,
   option_defs_ast
};

/* RETRO_LANGUAGE_BE */

#define CATEGORY_SYSTEM_LABEL_BE "Сістэма"
#define CATEGORY_SYSTEM_INFO_0_BE "Настройка рэгіёну, мовы, BIOS і асноўных параметраў абсталявання."
#define CATEGORY_VIDEO_LABEL_BE "Відэа"
#define CATEGORY_VIDEO_INFO_0_BE "Налады раздзялення, парадка-незалежнай празрыстасці і візуальных эфектаў."
#define CATEGORY_PERFORMANCE_LABEL_BE "Прадукцыйнасць"
#define CATEGORY_PERFORMANCE_INFO_0_BE "Настройкі шматпаточнага рэндэрынгу і пропуску кадраў."
#define CATEGORY_HACKS_LABEL_BE "Хакі эмуляцыі"
#define CATEGORY_HACKS_INFO_0_BE "Налады шырокафарматных чыт-кодаў, хуткасці загрузкі GD-ROM і замены тэкстураў."
#define CATEGORY_INPUT_LABEL_BE "Увод"
#define CATEGORY_INPUT_INFO_0_BE "Налады геймпада і светлавога пісталета."
#define CATEGORY_EXPANSIONS_LABEL_BE "Модулі пашырэння кантролераў"
#define CATEGORY_EXPANSIONS_INFO_0_BE "Выбар прылады (VMU, вібрапак), падлучанага да слотаў пашырэння кантролераў."
#define CATEGORY_VMU_LABEL_BE "Візуальныя карты памяці"
#define CATEGORY_VMU_INFO_0_BE "Настройкі файлаў захаванняў VMU для кожнай гульні і адлюстравання VMU на экране."
#define CORE_OPTION_NAME_REGION_LABEL_BE "Рэгіён"
#define OPTION_VAL_JAPAN_BE "Японія"
#define OPTION_VAL_USA_BE "ЗША"
#define OPTION_VAL_EUROPE_BE "Еўропа"
#define OPTION_VAL_DEFAULT_BE "Прадвызначана"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_BE "Мова"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_BE "Выбар мовы, якая выкарыстоўваецца ў BIOS, а таксама ў любых гульнях з падтрымкай некалькіх моваў."
#define OPTION_VAL_JAPANESE_BE "Японская"
#define OPTION_VAL_ENGLISH_BE "Англійская"
#define OPTION_VAL_GERMAN_BE "Нямецкая"
#define OPTION_VAL_FRENCH_BE "Французская"
#define OPTION_VAL_SPANISH_BE "Іспанская"
#define OPTION_VAL_ITALIAN_BE "Італьянская"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_BE "Эмуляцыя BIOS (патрабуецца перазапуск)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_BE "Прымусова выкарыстоўваць высокаўзроўневую эмуляцыю BIOS."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_BE "Уключыць DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_BE "Уключае эмуляцыю гукавога DSP (лічбавага сігнальнага працэсара) Dreamcast. Паляпшае дакладнасць згенераванага гуку, але павялічвае сістэмныя патрабаванні."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_BE "Дазваляць сэрвісныя кнопкі аркадных гульняў"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_BE "Уключае кнопку SERVICE для ўваходу ў налады аўтаматаў аркадных гульняў."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_BE "Рэжым Free Play для гульняў NAOMI"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_BE "Змяняць наладкі манетызацыі для ўключэння бясплатнай гульні."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_BE "Эмуляцыя Broadband Adapter"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_BE NULL
#define CORE_OPTION_NAME_UPNP_LABEL_BE "Уключыць UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_BE "Выкарыстоўваць UPnP для аўтаматычнай наладкі вашага Інтэрнэт-роўтара для анлайн-гульняў."
#define CORE_OPTION_NAME_DCNET_LABEL_BE NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_BE NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_BE "Унутраная раздзяляльнасць"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_BE "Змяніць раздзяляльнасць рэндэрынгу."
#define OPTION_VAL_320X240_BE "320x240 (напалову)"
#define OPTION_VAL_640X480_BE "640x480 (родная)"
#define OPTION_VAL_800X600_BE NULL
#define OPTION_VAL_960X720_BE NULL
#define OPTION_VAL_1024X768_BE NULL
#define OPTION_VAL_1280X960_BE NULL
#define OPTION_VAL_1440X1080_BE NULL
#define OPTION_VAL_1600X1200_BE NULL
#define OPTION_VAL_1920X1440_BE NULL
#define OPTION_VAL_2560X1920_BE NULL
#define OPTION_VAL_2880X2160_BE NULL
#define OPTION_VAL_3200X2400_BE NULL
#define OPTION_VAL_3840X2880_BE NULL
#define OPTION_VAL_4480X3360_BE NULL
#define OPTION_VAL_5120X3840_BE NULL
#define OPTION_VAL_5760X4320_BE NULL
#define OPTION_VAL_6400X4800_BE NULL
#define OPTION_VAL_7040X5280_BE NULL
#define OPTION_VAL_7680X5760_BE NULL
#define OPTION_VAL_8320X6240_BE NULL
#define OPTION_VAL_8960X6720_BE NULL
#define OPTION_VAL_9600X7200_BE NULL
#define OPTION_VAL_10240X7680_BE NULL
#define OPTION_VAL_10880X8160_BE NULL
#define OPTION_VAL_11520X8640_BE NULL
#define OPTION_VAL_12160X9120_BE NULL
#define OPTION_VAL_12800X9600_BE NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_BE "Тып кабеля"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_BE "Тып выходнага сігналу. 'ТБ (кампазіт)' з'яўляецца найболей сумяшчальным."
#define OPTION_VAL_VGA_BE NULL
#define OPTION_VAL_TV_RGB_BE "ТБ (RGB)"
#define OPTION_VAL_TV_COMPOSITE_BE "ТБ (кампазітны)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_BE "Стандарт вяшчання"
#define OPTION_VAL_NTSC_BE NULL
#define OPTION_VAL_PAL_BE "PAL (свет)"
#define OPTION_VAL_PAL_N_BE "PAL-N (Аргенціна, Парагвай, Уругвай)"
#define OPTION_VAL_PAL_M_BE "PAL-M (Бразілія)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_BE "Арыентацыя экрана"
#define OPTION_VAL_HORIZONTAL_BE "Гарызантальная"
#define OPTION_VAL_VERTICAL_BE "Вертыкальная"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_BE "Альфа-сартаванне"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_BE "Выбар спосабу сартавання празрыстых палігонаў."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_BE "Па пластах (хуткая, найменш дакладная)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_BE "Па трыкутніках (стандартная)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_BE "Па пікселях (дакладная, самая павольная)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_BE "Памер назапашвальнага буфера пікселяў"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_BE "Вялікія значэнні могуць спатрэбіцца для правільнага адлюстравання высокіх раздзяленняў."
#define OPTION_VAL_512MB_BE "512 МБ"
#define OPTION_VAL_1GB_BE "1 ГБ"
#define OPTION_VAL_2GB_BE "2 ГБ"
#define OPTION_VAL_4GB_BE "4 ГБ"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_BE "Максімум празрыстых пластоў"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_BE "Для складаных сцэн могуць запатрабавацца больш высокія значэнні."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_BE "Поўная эмуляцыя фрэймбуфера"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_BE "Уключае поўную эмуляцыю кадравага буфера ў VRAM. Карысна для гульняў, якія наўпрост вырабляюць чытанне ці запіс фрэймбуфера ў VRAM. Пры ўключэнні ўнутранае раздзяленне будзе прымусова ўсталявана ў 640×480 і магчыма моцнае падзенне хуткасці."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_BE "Уключыць RTT буфер (рэндэру ў тэкстуру)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_BE "Капіюе адмаляваныя тэкстуры з GPU зваротна ў VRAM. Дадзеная настройка звычайна ўключана для гульняў, якія гэтага патрабуюць. Пры ўключэнні не выкарыстоўваецца маштабаванне рэндэрынгу тэкстураў і магчыма падзенне прадукцыйнасці."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_BE "Mip-тэкстураванне"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_BE "Можа павысіць прадукцыйнасць і знізіць мігаценне пры адлюстраванні на адлегласці паменшаных копій тэкстураў."
#define CORE_OPTION_NAME_FOG_LABEL_BE "Эфекты туману"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_BE "Мадыфікатар аб'ёму"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_BE "Функцыя графічнага працэсара Dreamcast, якая выкарыстоўваецца ў гульнях для адмалёўкі ценяў аб'ектаў. Па змаўчанні павінна быць уключана - уплыў на прадукцыйнасць звычайна мінімальны ці малазначны."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_BE "Анізатропнае фільтраванне"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_BE "Паляпшае якасць тэкстураў на паверхнях пад непрамымі кутамі агляду ў адносінах да камеры. Больш высокія значэнні падвышаюць патрабаванні да GPU. Змены налады ўступаюць у сілу пры перазапуску."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_BE "Згладжванне тэкстур"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_BE "Выбар метаду згладжвання тэкстураў. Выкарыстоўваецца для прымусовага фільтравання ўсіх тэкстураў, каб атрымаць больш дакладны (або згладжаны) малюнак. Любое значэнне, адрознае ад стандартнага можа прыводзіць да памылак рэндэрынгу. Для прымянення наладкі патрабуецца перазапуск."
#define OPTION_VAL_1_BE "Прымусова метадам бліжэйшага суседа"
#define OPTION_VAL_2_BE "Прымусова лінейнае"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_BE "Затрымліваць змену кадраў"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_BE "Дапамагае пазбегнуць мігацення экрана або скажэння заставак. Не рэкамендуецца на марудных платформах."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_BE "Вызначаць змену частаты кадраў"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_BE "Апавяшчаць франтэнд аб зменах унутранай частаты кадраў (напр. з 60 да 30 кад/хв). Паляпшае размеркаванне кадраў у гульнях з сталай частатой у 30 ці 20 кад/хв, але павінна быць выключана для гульняў з плывучым значэннем частаты кадраў (напр. Ecco the Dolphin, Unreal Tournament). Недаступна пры вкл. опцыі 'Аўтапропуск кадраў'."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_BE "Фільтр постапрацоўкі PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_BE "Робіць постапрацоўку выявы для імітацыі эфектаў, спецыфічных для GPU PowerVR2 і аналагавых відэасігналаў."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_BE "Апскейлінг тэкстур (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_BE "Паляпшае 2D піксель-арт з ручной адмалёўкай. Рэкамендуецца толькі для 2D-гульняў з піксельнай графікай."
#define OPTION_VAL_2_O26_BE NULL
#define OPTION_VAL_4_BE NULL
#define OPTION_VAL_6_BE NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_BE "Максімальны памер згладжаных тэкстураў пры апскейлінгу"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_BE "Ўстаноўка максімальнага памеру тэкстураў пры маштабаванні. Для тэкстураў з памерам больш зададзенага значэння маштабаванне не прымяняецца."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_BE "Натыўная глыбіня інтэрпаляцыі"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_BE "Дапамагае пры праблемах з глыбінёй і скажэннем тэкстураў на відэакартах AMD. У шэрагу выпадкаў таксама развязвае праблемы на відэакартах Intel."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_BE "Выпраўленне прасочвання пры апскейлінгу"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_BE "Дапамагае ў выпадку прасочвання тэкстураў пры маштабаванні. Адключэнне дапамагае выправіць скажэнні пікселяў пры апскейлінгу 2D-гульняў (MVC2, CVS, KOF і інш.)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_BE "Рэндэрынг у асобным патоку"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_BE "Запускае графічны і цэнтральны працэсары ў розных патоках. Настойліва рэкамендуецца."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_BE "Аўтапропуск кадраў"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_BE "Аўтаматычна прапускаць кадры пры запаволенні эмулятара. Ужываецца толькі з уключанай опцыяй 'Рэндэрынг у асобным патоку'."
#define OPTION_VAL_SOME_BE "Звычайна"
#define OPTION_VAL_MORE_BE "Найбольшы"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_BE "Пропуск кадраў"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_BE "Устанаўлівае колькасць пропускаў кадраў паміж кадрамі якія адлюстроўваюцца."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_BE "Шырокаэкранныя чыт-коды (патрабуецца перазапуск)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_BE "Актывуе чыт-коды, якія дазваляюць адлюстроўваць некаторыя гульні ў фармаце шырокага экрана."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_BE "Шырокафарматны хак"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_BE "Адмалёўваць геаметрыю за межамі стандартных суадносін 4:3. Можа выклікаць скажэнні графікі ў пашыранай вобласці."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_BE "Хуткая загрузка GD-ROM (зніжае дакладнасць)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_BE "Паскарае загрузку GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_BE "Мод Dreamcast на 32 Мбайт АЗП"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_BE "Уключае мод для Dreamcast на 32 мегабайта АЗП. Можа ўплываць на сумяшчальнасць"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_BE "Разгон/запавольванне SH4"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_BE "Змена стандартнай частаты працэсара SH4 у 200 Мгц. Зніжэнне частаты дапамагае для слабых платформаў. Разгон можа падвышаць частату кадраў у некаторых гульнях. Выкарыстоўвайце з асцярожнасцю."
#define OPTION_VAL_100_BE "100 МГц"
#define OPTION_VAL_110_BE "110 МГц"
#define OPTION_VAL_120_BE "120 МГц"
#define OPTION_VAL_130_BE "130 МГц"
#define OPTION_VAL_140_BE "140 МГц"
#define OPTION_VAL_150_BE "150 МГц"
#define OPTION_VAL_160_BE "160 МГц"
#define OPTION_VAL_170_BE "170 МГц"
#define OPTION_VAL_180_BE "180 МГц"
#define OPTION_VAL_190_BE "190 МГц"
#define OPTION_VAL_200_BE "200 МГц"
#define OPTION_VAL_210_BE "210 МГц"
#define OPTION_VAL_220_BE "220 МГц"
#define OPTION_VAL_230_BE "230 МГц"
#define OPTION_VAL_240_BE "240 МГц"
#define OPTION_VAL_250_BE "250 МГц"
#define OPTION_VAL_260_BE "260 МГц"
#define OPTION_VAL_270_BE "270 МГц"
#define OPTION_VAL_280_BE "280 МГц"
#define OPTION_VAL_290_BE "290 МГц"
#define OPTION_VAL_300_BE "300 МГц"
#define OPTION_VAL_310_BE "310 МГц"
#define OPTION_VAL_320_BE "320 МГц"
#define OPTION_VAL_330_BE "330 МГц"
#define OPTION_VAL_340_BE "340 МГц"
#define OPTION_VAL_350_BE "350 МГц"
#define OPTION_VAL_360_BE "360 МГц"
#define OPTION_VAL_370_BE "370 МГц"
#define OPTION_VAL_380_BE "380 МГц"
#define OPTION_VAL_390_BE "390 МГц"
#define OPTION_VAL_400_BE "400 МГц"
#define OPTION_VAL_410_BE "410 МГц"
#define OPTION_VAL_420_BE "420 МГц"
#define OPTION_VAL_430_BE "430 МГц"
#define OPTION_VAL_440_BE "440 МГц"
#define OPTION_VAL_450_BE "450 МГц"
#define OPTION_VAL_460_BE "460 МГц"
#define OPTION_VAL_470_BE "470 МГц"
#define OPTION_VAL_480_BE "480 МГц"
#define OPTION_VAL_490_BE "490 МГц"
#define OPTION_VAL_500_BE "500 МГц"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_BE "Загружаць карыстальніцкія тэкстуры"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_BE "Загружаць прыстасаваныя тэкстуры з тэчкі 'system/dc/textures/<game-id>/'."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_BE NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_BE NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_BE "Здымаць дамп тэкстур"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_BE "Пры кожным выкарыстанні гульнёй новай тэкстуры яна будзе захоўвацца ў файл .png у тэчцы 'system/dc/texdump/<game-id>/'."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_BE NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_BE NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_BE "Мёртвая зона аналагавага стыку"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_BE "Настройка велічыні адхілення аналагавага джойсціка для спрацоўвання."
#define OPTION_VAL_0_BE "0 %"
#define OPTION_VAL_5_BE "5 %"
#define OPTION_VAL_10_BE "10 %"
#define OPTION_VAL_15_BE "15 %"
#define OPTION_VAL_20_BE "20 %"
#define OPTION_VAL_25_BE "25 %"
#define OPTION_VAL_30_BE "30 %"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_BE "Мёртвая зона трыгераў"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_BE "Настройка ступені націску трыгераў для спрацоўвання."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_BE "Лічбавыя трыгеры"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_BE "Пры ўключэнні трыгеры будуць апрацоўвацца як стандартныя кнопкі, цалкам націснутыя ці не націснутыя, без прамежкавых палажэнняў."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_BE "Трансляваць лічбавыя сігналы"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_BE "Трансляваць падзеі лічбавага ўводу і стан вібрацыі ў TCP-порт 8000. Сумяшчальна з параметрам MAME \"-output network\"."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_BE "Паказваць наладкі светлавога пісталета"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_BE "Уключае канфігурацыю параметраў прыцэла светлавога пісталета. Для прымянення наладкі патрабуецца пераключыць хуткае меню."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_BE "Маштабаванне прыцэлу пісталета"
#define OPTION_VAL_50_BE "50 %"
#define OPTION_VAL_60_BE "60 %"
#define OPTION_VAL_70_BE "70 %"
#define OPTION_VAL_80_BE "80 %"
#define OPTION_VAL_90_BE "90 %"
#define OPTION_VAL_100_O47_BE "100 %"
#define OPTION_VAL_110_O47_BE "110 %"
#define OPTION_VAL_120_O47_BE "120 %"
#define OPTION_VAL_130_O47_BE "130 %"
#define OPTION_VAL_140_O47_BE "140 %"
#define OPTION_VAL_150_O47_BE "150 %"
#define OPTION_VAL_160_O47_BE "160 %"
#define OPTION_VAL_170_O47_BE "170 %"
#define OPTION_VAL_180_O47_BE "180 %"
#define OPTION_VAL_190_O47_BE "190 %"
#define OPTION_VAL_200_O47_BE "200 %"
#define OPTION_VAL_210_O47_BE "210 %"
#define OPTION_VAL_220_O47_BE "220 %"
#define OPTION_VAL_230_O47_BE "230 %"
#define OPTION_VAL_240_O47_BE "240 %"
#define OPTION_VAL_250_O47_BE "250 %"
#define OPTION_VAL_260_O47_BE "260 %"
#define OPTION_VAL_270_O47_BE "270 %"
#define OPTION_VAL_280_O47_BE "280 %"
#define OPTION_VAL_290_O47_BE "290 %"
#define OPTION_VAL_300_O47_BE "300 %"
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_BE "Адлюстраванне прыцэла пісталета 1"
#define OPTION_VAL_WHITE_BE "Белы"
#define OPTION_VAL_RED_BE "Чырвоная"
#define OPTION_VAL_GREEN_BE "Зялёная"
#define OPTION_VAL_BLUE_BE "Блакітная"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_BE "Адлюстраванне прыцэла пісталета 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_BE "Адлюстраванне прыцэла пісталета 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_BE "Адлюстраванне прыцэла пісталета 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_BE "Прылада ў слоце пашырэння A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_BE "Выбар прылады, уключанай у слот пашырэння A1 (порт А слот 1)."
#define OPTION_VAL_VMU_BE NULL
#define OPTION_VAL_PURUPURU_BE "Вібрапак"
#define OPTION_VAL_DREAMPOTATO_BE NULL
#define OPTION_VAL_NONE_BE "Няма"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_BE "Прылада ў слоце пашырэння A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_BE "Выбар прылады, уключанай у слот пашырэння A2 (порт А слот 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_BE "Прылада ў слоце пашырэння B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_BE "Выбар прылады, уключанай у слот пашырэння B1 (порт B слот 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_BE "Прылада ў слоце пашырэння B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_BE "Выбар прылады, уключанай у слот пашырэння B2 (порт B слот 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_BE "Прылада ў слоце пашырэння C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_BE "Выбар прылады, уключанай у слот пашырэння C1 (порт C слот 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_BE "Прылада ў слоце пашырэння C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_BE "Выбар прылады, уключанай у слот пашырэння C2 (порт C слот 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_BE "Прылада ў слоце пашырэння D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_BE "Выбар прылады, уключанай у слот пашырэння D1 (порт D слот 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_BE "Прылада ў слоце пашырэння D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_BE "Выбар прылады, уключанай у слот пашырэння D2 (порт D слот 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_BE "Асобныя візуальныя карты памяці (VMU) для гульняў"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_BE "VMU для кожнай гульні"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_BE "Калі выключана, усе гульні выкарыстоўваюць да 8 файлаў захавання VMU (A1/A2/B1/B2/C1/C2/D1/D2), размешчаных у сістэмнай тэчцы RetroArch.\nНалада 'VMU A1' стварае на кожную запушчаную гульню ўнікальны файл VMU 'A1' у тэчцы захавання RetroArch.\nНалада 'Усе VMU' стварае да 8 унікальных файлаў VMU (A1/A2/B1/B2/C1/C2/D1/D2) на кожную запушчаную гульню."
#define OPTION_VAL_VMU_A1_BE NULL
#define OPTION_VAL_ALL_VMUS_BE "Усе VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_BE "Гукавыя сігналы візуальных карт памяці (VMU)"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_BE "Сігналы VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_BE "Уключае прайграванне сігналаў VMU."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_BE NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_BE NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_BE "Паказваць настройкі адлюстравання візуальных карт памяці (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_BE "Паказваць наладкі адлюстравання VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_BE "Уключае канфігурацыю памеру, становішча, колеру і бачнасці якая эмулюецца LCD-экранам VMU. Для прымянення наладкі патрабуецца пераключэнне хуткага меню."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_BE "Паказаць экран VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_BE "Палажэнне экрана VMU 1"
#define OPTION_VAL_UPPER_LEFT_BE "Зверху злева"
#define OPTION_VAL_UPPER_RIGHT_BE "Зверху справа"
#define OPTION_VAL_LOWER_LEFT_BE "Унізе злева"
#define OPTION_VAL_LOWER_RIGHT_BE "Унізе справа"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_BE "Памер экрана VMU 1"
#define OPTION_VAL_1X_BE NULL
#define OPTION_VAL_3X_BE NULL
#define OPTION_VAL_5X_BE NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_BE "Колер уключаных пікселяў VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_BE "Па змаўчанні ЎКЛ"
#define OPTION_VAL_DEFAULT_OFF_01_BE "Па змаўчанні ВЫКЛ"
#define OPTION_VAL_BLACK_02_BE "Чорны"
#define OPTION_VAL_LIGHT_BLUE_04_BE "Светла-блакітны"
#define OPTION_VAL_CYAN_06_BE "Марской хвалі"
#define OPTION_VAL_CYAN_BLUE_07_BE "Блакітна-сіні"
#define OPTION_VAL_LIGHT_GREEN_08_BE "Светла-зялёны"
#define OPTION_VAL_CYAN_GREEN_09_BE "Блакітна-зялёны"
#define OPTION_VAL_LIGHT_CYAN_10_BE "Светла-блакітны"
#define OPTION_VAL_PURPLE_12_BE "Фіялетавы"
#define OPTION_VAL_LIGHT_PURPLE_13_BE "Светла-фіялетавы"
#define OPTION_VAL_YELLOW_14_BE "Жоўтая"
#define OPTION_VAL_GRAY_15_BE "Шэры"
#define OPTION_VAL_LIGHT_PURPLE_2_16_BE "Светла-фіялетавы (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_BE "Светла-зялёны (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_BE "Светла-зялёны (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_BE "Светла-блакітны (2)"
#define OPTION_VAL_LIGHT_RED_2_20_BE "Светла-чырвоны (2)"
#define OPTION_VAL_MAGENTA_21_BE "Пурпурны"
#define OPTION_VAL_LIGHT_PURPLE_3_22_BE "Светла-фіялетавы (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_BE "Светла-аранжавы"
#define OPTION_VAL_ORANGE_24_BE "Аранжавы"
#define OPTION_VAL_LIGHT_PURPLE_4_25_BE "Светла-фіялетавы (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_BE "Светла-жоўты"
#define OPTION_VAL_LIGHT_YELLOW_2_27_BE "Светла-жоўты (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_BE "Колер выключаных пікселяў VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_BE "Бачнасць экрана VMU 1"
#define OPTION_VAL_40_BE "40 %"
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_BE "Паказаць экран VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_BE "Палажэнне экрана VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_BE "Памер экрана VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_BE "Колер уключаных пікселяў VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_BE "Колер выключаных пікселяў VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_BE "Бачнасць экрана VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_BE "Паказаць экран VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_BE "Палажэнне экрана VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_BE "Памер экрана VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_BE "Колер уключаных пікселяў VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_BE "Колер выключаных пікселяў VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_BE "Бачнасць экрана VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_BE "Паказаць экран VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_BE "Палажэнне экрана VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_BE "Памер экрана VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_BE "Колер уключаных пікселяў VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_BE "Колер выключаных пікселяў VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_BE "Бачнасць экрана VMU 4"

struct retro_core_option_v2_category option_cats_be[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_BE,
      CATEGORY_SYSTEM_INFO_0_BE
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_BE,
      CATEGORY_VIDEO_INFO_0_BE
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_BE,
      CATEGORY_PERFORMANCE_INFO_0_BE
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_BE,
      CATEGORY_HACKS_INFO_0_BE
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_BE,
      CATEGORY_INPUT_INFO_0_BE
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_BE,
      CATEGORY_EXPANSIONS_INFO_0_BE
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_BE,
      CATEGORY_VMU_INFO_0_BE
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_be[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_BE,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_BE },
         { "USA",     OPTION_VAL_USA_BE },
         { "Europe",  OPTION_VAL_EUROPE_BE },
         { "Default", OPTION_VAL_DEFAULT_BE },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_BE,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_BE },
         { "English",  OPTION_VAL_ENGLISH_BE },
         { "German",   OPTION_VAL_GERMAN_BE },
         { "French",   OPTION_VAL_FRENCH_BE },
         { "Spanish",  OPTION_VAL_SPANISH_BE },
         { "Italian",  OPTION_VAL_ITALIAN_BE },
         { "Default",  OPTION_VAL_DEFAULT_BE },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_BE,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_BE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_BE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_BE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_BE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_BE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_BE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_BE,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_BE },
         { "640x480",    OPTION_VAL_640X480_BE },
         { "800x600",    OPTION_VAL_800X600_BE },
         { "960x720",    OPTION_VAL_960X720_BE },
         { "1024x768",   OPTION_VAL_1024X768_BE },
         { "1280x960",   OPTION_VAL_1280X960_BE },
         { "1440x1080",  OPTION_VAL_1440X1080_BE },
         { "1600x1200",  OPTION_VAL_1600X1200_BE },
         { "1920x1440",  OPTION_VAL_1920X1440_BE },
         { "2560x1920",  OPTION_VAL_2560X1920_BE },
         { "2880x2160",  OPTION_VAL_2880X2160_BE },
         { "3200x2400",  OPTION_VAL_3200X2400_BE },
         { "3840x2880",  OPTION_VAL_3840X2880_BE },
         { "4480x3360",  OPTION_VAL_4480X3360_BE },
         { "5120x3840",  OPTION_VAL_5120X3840_BE },
         { "5760x4320",  OPTION_VAL_5760X4320_BE },
         { "6400x4800",  OPTION_VAL_6400X4800_BE },
         { "7040x5280",  OPTION_VAL_7040X5280_BE },
         { "7680x5760",  OPTION_VAL_7680X5760_BE },
         { "8320x6240",  OPTION_VAL_8320X6240_BE },
         { "8960x6720",  OPTION_VAL_8960X6720_BE },
         { "9600x7200",  OPTION_VAL_9600X7200_BE },
         { "10240x7680", OPTION_VAL_10240X7680_BE },
         { "10880x8160", OPTION_VAL_10880X8160_BE },
         { "11520x8640", OPTION_VAL_11520X8640_BE },
         { "12160x9120", OPTION_VAL_12160X9120_BE },
         { "12800x9600", OPTION_VAL_12800X9600_BE },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_BE,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_BE },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_BE },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_BE },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_BE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_BE },
         { "PAL",     OPTION_VAL_PAL_BE },
         { "PAL_N",   OPTION_VAL_PAL_N_BE },
         { "PAL_M",   OPTION_VAL_PAL_M_BE },
         { "Default", OPTION_VAL_DEFAULT_BE },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_BE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_BE },
         { "vertical",   OPTION_VAL_VERTICAL_BE },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_BE,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_BE },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_BE },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_BE },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_BE,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_BE },
         { "1GB",   OPTION_VAL_1GB_BE },
         { "2GB",   OPTION_VAL_2GB_BE },
         { "4GB",   OPTION_VAL_4GB_BE },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_BE,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_BE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_BE,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_BE,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_BE },
         { "1",  OPTION_VAL_1_BE },
         { "2",  OPTION_VAL_2_BE },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_BE,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_BE },
         { "4",  OPTION_VAL_4_BE },
         { "6",  OPTION_VAL_6_BE },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_BE,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_BE,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_BE,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_BE,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_BE,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_BE,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_BE,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_BE },
         { "more", OPTION_VAL_MORE_BE },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_BE,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_BE },
         { "110", OPTION_VAL_110_BE },
         { "120", OPTION_VAL_120_BE },
         { "130", OPTION_VAL_130_BE },
         { "140", OPTION_VAL_140_BE },
         { "150", OPTION_VAL_150_BE },
         { "160", OPTION_VAL_160_BE },
         { "170", OPTION_VAL_170_BE },
         { "180", OPTION_VAL_180_BE },
         { "190", OPTION_VAL_190_BE },
         { "200", OPTION_VAL_200_BE },
         { "210", OPTION_VAL_210_BE },
         { "220", OPTION_VAL_220_BE },
         { "230", OPTION_VAL_230_BE },
         { "240", OPTION_VAL_240_BE },
         { "250", OPTION_VAL_250_BE },
         { "260", OPTION_VAL_260_BE },
         { "270", OPTION_VAL_270_BE },
         { "280", OPTION_VAL_280_BE },
         { "290", OPTION_VAL_290_BE },
         { "300", OPTION_VAL_300_BE },
         { "310", OPTION_VAL_310_BE },
         { "320", OPTION_VAL_320_BE },
         { "330", OPTION_VAL_330_BE },
         { "340", OPTION_VAL_340_BE },
         { "350", OPTION_VAL_350_BE },
         { "360", OPTION_VAL_360_BE },
         { "370", OPTION_VAL_370_BE },
         { "380", OPTION_VAL_380_BE },
         { "390", OPTION_VAL_390_BE },
         { "400", OPTION_VAL_400_BE },
         { "410", OPTION_VAL_410_BE },
         { "420", OPTION_VAL_420_BE },
         { "430", OPTION_VAL_430_BE },
         { "440", OPTION_VAL_440_BE },
         { "450", OPTION_VAL_450_BE },
         { "460", OPTION_VAL_460_BE },
         { "470", OPTION_VAL_470_BE },
         { "480", OPTION_VAL_480_BE },
         { "490", OPTION_VAL_490_BE },
         { "500", OPTION_VAL_500_BE },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_BE,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_BE },
         { "5%",  OPTION_VAL_5_BE },
         { "10%", OPTION_VAL_10_BE },
         { "15%", OPTION_VAL_15_BE },
         { "20%", OPTION_VAL_20_BE },
         { "25%", OPTION_VAL_25_BE },
         { "30%", OPTION_VAL_30_BE },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_BE,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_BE },
         { "5%",  OPTION_VAL_5_BE },
         { "10%", OPTION_VAL_10_BE },
         { "15%", OPTION_VAL_15_BE },
         { "20%", OPTION_VAL_20_BE },
         { "25%", OPTION_VAL_25_BE },
         { "30%", OPTION_VAL_30_BE },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_BE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_BE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_BE,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_BE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_BE },
         { "60%",  OPTION_VAL_60_BE },
         { "70%",  OPTION_VAL_70_BE },
         { "80%",  OPTION_VAL_80_BE },
         { "90%",  OPTION_VAL_90_BE },
         { "100%", OPTION_VAL_100_O47_BE },
         { "110%", OPTION_VAL_110_O47_BE },
         { "120%", OPTION_VAL_120_O47_BE },
         { "130%", OPTION_VAL_130_O47_BE },
         { "140%", OPTION_VAL_140_O47_BE },
         { "150%", OPTION_VAL_150_O47_BE },
         { "160%", OPTION_VAL_160_O47_BE },
         { "170%", OPTION_VAL_170_O47_BE },
         { "180%", OPTION_VAL_180_O47_BE },
         { "190%", OPTION_VAL_190_O47_BE },
         { "200%", OPTION_VAL_200_O47_BE },
         { "210%", OPTION_VAL_210_O47_BE },
         { "220%", OPTION_VAL_220_O47_BE },
         { "230%", OPTION_VAL_230_O47_BE },
         { "240%", OPTION_VAL_240_O47_BE },
         { "250%", OPTION_VAL_250_O47_BE },
         { "260%", OPTION_VAL_260_O47_BE },
         { "270%", OPTION_VAL_270_O47_BE },
         { "280%", OPTION_VAL_280_O47_BE },
         { "290%", OPTION_VAL_290_O47_BE },
         { "300%", OPTION_VAL_300_O47_BE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_BE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_BE },
         { "Red",      OPTION_VAL_RED_BE },
         { "Green",    OPTION_VAL_GREEN_BE },
         { "Blue",     OPTION_VAL_BLUE_BE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_BE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_BE },
         { "Red",      OPTION_VAL_RED_BE },
         { "Green",    OPTION_VAL_GREEN_BE },
         { "Blue",     OPTION_VAL_BLUE_BE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_BE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_BE },
         { "Red",      OPTION_VAL_RED_BE },
         { "Green",    OPTION_VAL_GREEN_BE },
         { "Blue",     OPTION_VAL_BLUE_BE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_BE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_BE },
         { "Red",      OPTION_VAL_RED_BE },
         { "Green",    OPTION_VAL_GREEN_BE },
         { "Blue",     OPTION_VAL_BLUE_BE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_BE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BE },
         { "Purupuru", OPTION_VAL_PURUPURU_BE },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_BE },
         { "None",     OPTION_VAL_NONE_BE },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_BE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BE },
         { "Purupuru", OPTION_VAL_PURUPURU_BE },
         { "None",     OPTION_VAL_NONE_BE },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_BE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BE },
         { "Purupuru", OPTION_VAL_PURUPURU_BE },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_BE },
         { "None",     OPTION_VAL_NONE_BE },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_BE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BE },
         { "Purupuru", OPTION_VAL_PURUPURU_BE },
         { "None",     OPTION_VAL_NONE_BE },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_BE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BE },
         { "Purupuru", OPTION_VAL_PURUPURU_BE },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_BE },
         { "None",     OPTION_VAL_NONE_BE },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_BE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BE },
         { "Purupuru", OPTION_VAL_PURUPURU_BE },
         { "None",     OPTION_VAL_NONE_BE },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_BE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BE },
         { "Purupuru", OPTION_VAL_PURUPURU_BE },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_BE },
         { "None",     OPTION_VAL_NONE_BE },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_BE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BE },
         { "Purupuru", OPTION_VAL_PURUPURU_BE },
         { "None",     OPTION_VAL_NONE_BE },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_BE,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_BE,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_BE,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_BE },
         { "All VMUs", OPTION_VAL_ALL_VMUS_BE },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_BE,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_BE,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_BE,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_BE,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_BE,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_BE,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_BE,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_BE,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_BE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_BE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_BE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_BE },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_BE },
         { "2x", OPTION_VAL_2_O26_BE },
         { "3x", OPTION_VAL_3X_BE },
         { "4x", OPTION_VAL_4_BE },
         { "5x", OPTION_VAL_5X_BE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BE },
         { "BLUE 03",           OPTION_VAL_BLUE_BE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BE },
         { "GREEN 05",          OPTION_VAL_GREEN_BE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BE },
         { "RED 11",            OPTION_VAL_RED_BE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BE },
         { "WHITE 28",          OPTION_VAL_WHITE_BE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BE },
         { "BLUE 03",           OPTION_VAL_BLUE_BE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BE },
         { "GREEN 05",          OPTION_VAL_GREEN_BE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BE },
         { "RED 11",            OPTION_VAL_RED_BE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BE },
         { "WHITE 28",          OPTION_VAL_WHITE_BE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_BE },
         { "20%",  OPTION_VAL_20_BE },
         { "30%",  OPTION_VAL_30_BE },
         { "40%",  OPTION_VAL_40_BE },
         { "50%",  OPTION_VAL_50_BE },
         { "60%",  OPTION_VAL_60_BE },
         { "70%",  OPTION_VAL_70_BE },
         { "80%",  OPTION_VAL_80_BE },
         { "90%",  OPTION_VAL_90_BE },
         { "100%", OPTION_VAL_100_O47_BE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_BE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_BE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_BE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_BE },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_BE },
         { "2x", OPTION_VAL_2_O26_BE },
         { "3x", OPTION_VAL_3X_BE },
         { "4x", OPTION_VAL_4_BE },
         { "5x", OPTION_VAL_5X_BE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BE },
         { "BLUE 03",           OPTION_VAL_BLUE_BE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BE },
         { "GREEN 05",          OPTION_VAL_GREEN_BE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BE },
         { "RED 11",            OPTION_VAL_RED_BE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BE },
         { "WHITE 28",          OPTION_VAL_WHITE_BE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BE },
         { "BLUE 03",           OPTION_VAL_BLUE_BE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BE },
         { "GREEN 05",          OPTION_VAL_GREEN_BE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BE },
         { "RED 11",            OPTION_VAL_RED_BE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BE },
         { "WHITE 28",          OPTION_VAL_WHITE_BE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_BE },
         { "20%",  OPTION_VAL_20_BE },
         { "30%",  OPTION_VAL_30_BE },
         { "40%",  OPTION_VAL_40_BE },
         { "50%",  OPTION_VAL_50_BE },
         { "60%",  OPTION_VAL_60_BE },
         { "70%",  OPTION_VAL_70_BE },
         { "80%",  OPTION_VAL_80_BE },
         { "90%",  OPTION_VAL_90_BE },
         { "100%", OPTION_VAL_100_O47_BE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_BE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_BE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_BE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_BE },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_BE },
         { "2x", OPTION_VAL_2_O26_BE },
         { "3x", OPTION_VAL_3X_BE },
         { "4x", OPTION_VAL_4_BE },
         { "5x", OPTION_VAL_5X_BE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BE },
         { "BLUE 03",           OPTION_VAL_BLUE_BE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BE },
         { "GREEN 05",          OPTION_VAL_GREEN_BE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BE },
         { "RED 11",            OPTION_VAL_RED_BE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BE },
         { "WHITE 28",          OPTION_VAL_WHITE_BE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BE },
         { "BLUE 03",           OPTION_VAL_BLUE_BE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BE },
         { "GREEN 05",          OPTION_VAL_GREEN_BE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BE },
         { "RED 11",            OPTION_VAL_RED_BE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BE },
         { "WHITE 28",          OPTION_VAL_WHITE_BE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_BE },
         { "20%",  OPTION_VAL_20_BE },
         { "30%",  OPTION_VAL_30_BE },
         { "40%",  OPTION_VAL_40_BE },
         { "50%",  OPTION_VAL_50_BE },
         { "60%",  OPTION_VAL_60_BE },
         { "70%",  OPTION_VAL_70_BE },
         { "80%",  OPTION_VAL_80_BE },
         { "90%",  OPTION_VAL_90_BE },
         { "100%", OPTION_VAL_100_O47_BE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_BE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_BE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_BE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_BE },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_BE },
         { "2x", OPTION_VAL_2_O26_BE },
         { "3x", OPTION_VAL_3X_BE },
         { "4x", OPTION_VAL_4_BE },
         { "5x", OPTION_VAL_5X_BE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BE },
         { "BLUE 03",           OPTION_VAL_BLUE_BE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BE },
         { "GREEN 05",          OPTION_VAL_GREEN_BE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BE },
         { "RED 11",            OPTION_VAL_RED_BE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BE },
         { "WHITE 28",          OPTION_VAL_WHITE_BE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BE },
         { "BLUE 03",           OPTION_VAL_BLUE_BE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BE },
         { "GREEN 05",          OPTION_VAL_GREEN_BE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BE },
         { "RED 11",            OPTION_VAL_RED_BE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BE },
         { "WHITE 28",          OPTION_VAL_WHITE_BE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_BE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_BE },
         { "20%",  OPTION_VAL_20_BE },
         { "30%",  OPTION_VAL_30_BE },
         { "40%",  OPTION_VAL_40_BE },
         { "50%",  OPTION_VAL_50_BE },
         { "60%",  OPTION_VAL_60_BE },
         { "70%",  OPTION_VAL_70_BE },
         { "80%",  OPTION_VAL_80_BE },
         { "90%",  OPTION_VAL_90_BE },
         { "100%", OPTION_VAL_100_O47_BE },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_be = {
   option_cats_be,
   option_defs_be
};

/* RETRO_LANGUAGE_BG */

#define CATEGORY_SYSTEM_LABEL_BG NULL
#define CATEGORY_SYSTEM_INFO_0_BG NULL
#define CATEGORY_VIDEO_LABEL_BG "Видео"
#define CATEGORY_VIDEO_INFO_0_BG NULL
#define CATEGORY_PERFORMANCE_LABEL_BG NULL
#define CATEGORY_PERFORMANCE_INFO_0_BG NULL
#define CATEGORY_HACKS_LABEL_BG NULL
#define CATEGORY_HACKS_INFO_0_BG NULL
#define CATEGORY_INPUT_LABEL_BG "Вход"
#define CATEGORY_INPUT_INFO_0_BG NULL
#define CATEGORY_EXPANSIONS_LABEL_BG NULL
#define CATEGORY_EXPANSIONS_INFO_0_BG NULL
#define CATEGORY_VMU_LABEL_BG NULL
#define CATEGORY_VMU_INFO_0_BG NULL
#define CORE_OPTION_NAME_REGION_LABEL_BG NULL
#define OPTION_VAL_JAPAN_BG NULL
#define OPTION_VAL_USA_BG NULL
#define OPTION_VAL_EUROPE_BG NULL
#define OPTION_VAL_DEFAULT_BG NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_BG NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_BG NULL
#define OPTION_VAL_JAPANESE_BG NULL
#define OPTION_VAL_ENGLISH_BG NULL
#define OPTION_VAL_GERMAN_BG NULL
#define OPTION_VAL_FRENCH_BG NULL
#define OPTION_VAL_SPANISH_BG NULL
#define OPTION_VAL_ITALIAN_BG NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_BG NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_BG NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_BG NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_BG NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_BG NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_BG NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_BG NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_BG NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_BG NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_BG NULL
#define CORE_OPTION_NAME_UPNP_LABEL_BG NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_BG NULL
#define CORE_OPTION_NAME_DCNET_LABEL_BG NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_BG NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_BG NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_BG NULL
#define OPTION_VAL_320X240_BG NULL
#define OPTION_VAL_640X480_BG NULL
#define OPTION_VAL_800X600_BG NULL
#define OPTION_VAL_960X720_BG NULL
#define OPTION_VAL_1024X768_BG NULL
#define OPTION_VAL_1280X960_BG NULL
#define OPTION_VAL_1440X1080_BG NULL
#define OPTION_VAL_1600X1200_BG NULL
#define OPTION_VAL_1920X1440_BG NULL
#define OPTION_VAL_2560X1920_BG NULL
#define OPTION_VAL_2880X2160_BG NULL
#define OPTION_VAL_3200X2400_BG NULL
#define OPTION_VAL_3840X2880_BG NULL
#define OPTION_VAL_4480X3360_BG NULL
#define OPTION_VAL_5120X3840_BG NULL
#define OPTION_VAL_5760X4320_BG NULL
#define OPTION_VAL_6400X4800_BG NULL
#define OPTION_VAL_7040X5280_BG NULL
#define OPTION_VAL_7680X5760_BG NULL
#define OPTION_VAL_8320X6240_BG NULL
#define OPTION_VAL_8960X6720_BG NULL
#define OPTION_VAL_9600X7200_BG NULL
#define OPTION_VAL_10240X7680_BG NULL
#define OPTION_VAL_10880X8160_BG NULL
#define OPTION_VAL_11520X8640_BG NULL
#define OPTION_VAL_12160X9120_BG NULL
#define OPTION_VAL_12800X9600_BG NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_BG NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_BG NULL
#define OPTION_VAL_VGA_BG NULL
#define OPTION_VAL_TV_RGB_BG NULL
#define OPTION_VAL_TV_COMPOSITE_BG NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_BG NULL
#define OPTION_VAL_NTSC_BG NULL
#define OPTION_VAL_PAL_BG NULL
#define OPTION_VAL_PAL_N_BG NULL
#define OPTION_VAL_PAL_M_BG NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_BG NULL
#define OPTION_VAL_HORIZONTAL_BG NULL
#define OPTION_VAL_VERTICAL_BG NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_BG NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_BG NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_BG NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_BG NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_BG NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_BG NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_BG NULL
#define OPTION_VAL_512MB_BG NULL
#define OPTION_VAL_1GB_BG NULL
#define OPTION_VAL_2GB_BG NULL
#define OPTION_VAL_4GB_BG NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_BG NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_BG NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_BG NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_BG NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_BG NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_BG NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_BG NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_BG NULL
#define CORE_OPTION_NAME_FOG_LABEL_BG NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_BG NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_BG NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_BG NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_BG NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_BG NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_BG NULL
#define OPTION_VAL_1_BG NULL
#define OPTION_VAL_2_BG NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_BG NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_BG NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_BG NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_BG NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_BG NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_BG NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_BG NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_BG NULL
#define OPTION_VAL_2_O26_BG NULL
#define OPTION_VAL_4_BG NULL
#define OPTION_VAL_6_BG NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_BG NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_BG NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_BG NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_BG NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_BG NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_BG NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_BG NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_BG NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_BG NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_BG NULL
#define OPTION_VAL_SOME_BG NULL
#define OPTION_VAL_MORE_BG NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_BG NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_BG NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_BG NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_BG NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_BG NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_BG NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_BG NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_BG NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_BG NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_BG NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_BG NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_BG NULL
#define OPTION_VAL_100_BG NULL
#define OPTION_VAL_110_BG NULL
#define OPTION_VAL_120_BG NULL
#define OPTION_VAL_130_BG NULL
#define OPTION_VAL_140_BG NULL
#define OPTION_VAL_150_BG NULL
#define OPTION_VAL_160_BG NULL
#define OPTION_VAL_170_BG NULL
#define OPTION_VAL_180_BG NULL
#define OPTION_VAL_190_BG NULL
#define OPTION_VAL_200_BG NULL
#define OPTION_VAL_210_BG NULL
#define OPTION_VAL_220_BG NULL
#define OPTION_VAL_230_BG NULL
#define OPTION_VAL_240_BG NULL
#define OPTION_VAL_250_BG NULL
#define OPTION_VAL_260_BG NULL
#define OPTION_VAL_270_BG NULL
#define OPTION_VAL_280_BG NULL
#define OPTION_VAL_290_BG NULL
#define OPTION_VAL_300_BG NULL
#define OPTION_VAL_310_BG NULL
#define OPTION_VAL_320_BG NULL
#define OPTION_VAL_330_BG NULL
#define OPTION_VAL_340_BG NULL
#define OPTION_VAL_350_BG NULL
#define OPTION_VAL_360_BG NULL
#define OPTION_VAL_370_BG NULL
#define OPTION_VAL_380_BG NULL
#define OPTION_VAL_390_BG NULL
#define OPTION_VAL_400_BG NULL
#define OPTION_VAL_410_BG NULL
#define OPTION_VAL_420_BG NULL
#define OPTION_VAL_430_BG NULL
#define OPTION_VAL_440_BG NULL
#define OPTION_VAL_450_BG NULL
#define OPTION_VAL_460_BG NULL
#define OPTION_VAL_470_BG NULL
#define OPTION_VAL_480_BG NULL
#define OPTION_VAL_490_BG NULL
#define OPTION_VAL_500_BG NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_BG NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_BG NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_BG NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_BG NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_BG NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_BG NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_BG NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_BG NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_BG NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_BG NULL
#define OPTION_VAL_0_BG NULL
#define OPTION_VAL_5_BG NULL
#define OPTION_VAL_10_BG NULL
#define OPTION_VAL_15_BG NULL
#define OPTION_VAL_20_BG NULL
#define OPTION_VAL_25_BG NULL
#define OPTION_VAL_30_BG NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_BG NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_BG NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_BG NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_BG NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_BG NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_BG NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_BG NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_BG NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_BG NULL
#define OPTION_VAL_50_BG NULL
#define OPTION_VAL_60_BG NULL
#define OPTION_VAL_70_BG NULL
#define OPTION_VAL_80_BG NULL
#define OPTION_VAL_90_BG NULL
#define OPTION_VAL_100_O47_BG NULL
#define OPTION_VAL_110_O47_BG NULL
#define OPTION_VAL_120_O47_BG NULL
#define OPTION_VAL_130_O47_BG NULL
#define OPTION_VAL_140_O47_BG NULL
#define OPTION_VAL_150_O47_BG NULL
#define OPTION_VAL_160_O47_BG NULL
#define OPTION_VAL_170_O47_BG NULL
#define OPTION_VAL_180_O47_BG NULL
#define OPTION_VAL_190_O47_BG NULL
#define OPTION_VAL_200_O47_BG NULL
#define OPTION_VAL_210_O47_BG NULL
#define OPTION_VAL_220_O47_BG NULL
#define OPTION_VAL_230_O47_BG NULL
#define OPTION_VAL_240_O47_BG NULL
#define OPTION_VAL_250_O47_BG NULL
#define OPTION_VAL_260_O47_BG NULL
#define OPTION_VAL_270_O47_BG NULL
#define OPTION_VAL_280_O47_BG NULL
#define OPTION_VAL_290_O47_BG NULL
#define OPTION_VAL_300_O47_BG NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_BG NULL
#define OPTION_VAL_WHITE_BG NULL
#define OPTION_VAL_RED_BG NULL
#define OPTION_VAL_GREEN_BG NULL
#define OPTION_VAL_BLUE_BG NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_BG NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_BG NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_BG NULL
#define OPTION_VAL_VMU_BG NULL
#define OPTION_VAL_PURUPURU_BG NULL
#define OPTION_VAL_DREAMPOTATO_BG NULL
#define OPTION_VAL_NONE_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_BG NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_BG NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_BG NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_BG NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_BG NULL
#define OPTION_VAL_VMU_A1_BG NULL
#define OPTION_VAL_ALL_VMUS_BG NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_BG NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_BG NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_BG NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_BG NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_BG NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_BG NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_BG NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_BG NULL
#define OPTION_VAL_UPPER_LEFT_BG NULL
#define OPTION_VAL_UPPER_RIGHT_BG NULL
#define OPTION_VAL_LOWER_LEFT_BG NULL
#define OPTION_VAL_LOWER_RIGHT_BG NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_BG NULL
#define OPTION_VAL_1X_BG NULL
#define OPTION_VAL_3X_BG NULL
#define OPTION_VAL_5X_BG NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_BG NULL
#define OPTION_VAL_DEFAULT_ON_00_BG NULL
#define OPTION_VAL_DEFAULT_OFF_01_BG NULL
#define OPTION_VAL_BLACK_02_BG NULL
#define OPTION_VAL_LIGHT_BLUE_04_BG NULL
#define OPTION_VAL_CYAN_06_BG NULL
#define OPTION_VAL_CYAN_BLUE_07_BG NULL
#define OPTION_VAL_LIGHT_GREEN_08_BG NULL
#define OPTION_VAL_CYAN_GREEN_09_BG NULL
#define OPTION_VAL_LIGHT_CYAN_10_BG NULL
#define OPTION_VAL_PURPLE_12_BG NULL
#define OPTION_VAL_LIGHT_PURPLE_13_BG NULL
#define OPTION_VAL_YELLOW_14_BG NULL
#define OPTION_VAL_GRAY_15_BG NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_BG NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_BG NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_BG NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_BG NULL
#define OPTION_VAL_LIGHT_RED_2_20_BG NULL
#define OPTION_VAL_MAGENTA_21_BG NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_BG NULL
#define OPTION_VAL_LIGHT_ORANGE_23_BG NULL
#define OPTION_VAL_ORANGE_24_BG NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_BG NULL
#define OPTION_VAL_LIGHT_YELLOW_26_BG NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_BG NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_BG NULL
#define OPTION_VAL_40_BG NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_BG NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_BG NULL

struct retro_core_option_v2_category option_cats_bg[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_BG,
      CATEGORY_SYSTEM_INFO_0_BG
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_BG,
      CATEGORY_VIDEO_INFO_0_BG
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_BG,
      CATEGORY_PERFORMANCE_INFO_0_BG
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_BG,
      CATEGORY_HACKS_INFO_0_BG
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_BG,
      CATEGORY_INPUT_INFO_0_BG
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_BG,
      CATEGORY_EXPANSIONS_INFO_0_BG
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_BG,
      CATEGORY_VMU_INFO_0_BG
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_bg[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_BG,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_BG },
         { "USA",     OPTION_VAL_USA_BG },
         { "Europe",  OPTION_VAL_EUROPE_BG },
         { "Default", OPTION_VAL_DEFAULT_BG },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_BG,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_BG },
         { "English",  OPTION_VAL_ENGLISH_BG },
         { "German",   OPTION_VAL_GERMAN_BG },
         { "French",   OPTION_VAL_FRENCH_BG },
         { "Spanish",  OPTION_VAL_SPANISH_BG },
         { "Italian",  OPTION_VAL_ITALIAN_BG },
         { "Default",  OPTION_VAL_DEFAULT_BG },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_BG,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_BG,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_BG,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_BG,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_BG,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_BG,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_BG,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_BG,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_BG },
         { "640x480",    OPTION_VAL_640X480_BG },
         { "800x600",    OPTION_VAL_800X600_BG },
         { "960x720",    OPTION_VAL_960X720_BG },
         { "1024x768",   OPTION_VAL_1024X768_BG },
         { "1280x960",   OPTION_VAL_1280X960_BG },
         { "1440x1080",  OPTION_VAL_1440X1080_BG },
         { "1600x1200",  OPTION_VAL_1600X1200_BG },
         { "1920x1440",  OPTION_VAL_1920X1440_BG },
         { "2560x1920",  OPTION_VAL_2560X1920_BG },
         { "2880x2160",  OPTION_VAL_2880X2160_BG },
         { "3200x2400",  OPTION_VAL_3200X2400_BG },
         { "3840x2880",  OPTION_VAL_3840X2880_BG },
         { "4480x3360",  OPTION_VAL_4480X3360_BG },
         { "5120x3840",  OPTION_VAL_5120X3840_BG },
         { "5760x4320",  OPTION_VAL_5760X4320_BG },
         { "6400x4800",  OPTION_VAL_6400X4800_BG },
         { "7040x5280",  OPTION_VAL_7040X5280_BG },
         { "7680x5760",  OPTION_VAL_7680X5760_BG },
         { "8320x6240",  OPTION_VAL_8320X6240_BG },
         { "8960x6720",  OPTION_VAL_8960X6720_BG },
         { "9600x7200",  OPTION_VAL_9600X7200_BG },
         { "10240x7680", OPTION_VAL_10240X7680_BG },
         { "10880x8160", OPTION_VAL_10880X8160_BG },
         { "11520x8640", OPTION_VAL_11520X8640_BG },
         { "12160x9120", OPTION_VAL_12160X9120_BG },
         { "12800x9600", OPTION_VAL_12800X9600_BG },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_BG,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_BG },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_BG },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_BG },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_BG,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_BG },
         { "PAL",     OPTION_VAL_PAL_BG },
         { "PAL_N",   OPTION_VAL_PAL_N_BG },
         { "PAL_M",   OPTION_VAL_PAL_M_BG },
         { "Default", OPTION_VAL_DEFAULT_BG },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_BG,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_BG },
         { "vertical",   OPTION_VAL_VERTICAL_BG },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_BG,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_BG },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_BG },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_BG },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_BG,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_BG },
         { "1GB",   OPTION_VAL_1GB_BG },
         { "2GB",   OPTION_VAL_2GB_BG },
         { "4GB",   OPTION_VAL_4GB_BG },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_BG,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_BG,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_BG,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_BG,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_BG },
         { "1",  OPTION_VAL_1_BG },
         { "2",  OPTION_VAL_2_BG },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_BG,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_BG },
         { "4",  OPTION_VAL_4_BG },
         { "6",  OPTION_VAL_6_BG },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_BG,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_BG,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_BG,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_BG,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_BG,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_BG,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_BG,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_BG },
         { "more", OPTION_VAL_MORE_BG },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_BG,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_BG },
         { "110", OPTION_VAL_110_BG },
         { "120", OPTION_VAL_120_BG },
         { "130", OPTION_VAL_130_BG },
         { "140", OPTION_VAL_140_BG },
         { "150", OPTION_VAL_150_BG },
         { "160", OPTION_VAL_160_BG },
         { "170", OPTION_VAL_170_BG },
         { "180", OPTION_VAL_180_BG },
         { "190", OPTION_VAL_190_BG },
         { "200", OPTION_VAL_200_BG },
         { "210", OPTION_VAL_210_BG },
         { "220", OPTION_VAL_220_BG },
         { "230", OPTION_VAL_230_BG },
         { "240", OPTION_VAL_240_BG },
         { "250", OPTION_VAL_250_BG },
         { "260", OPTION_VAL_260_BG },
         { "270", OPTION_VAL_270_BG },
         { "280", OPTION_VAL_280_BG },
         { "290", OPTION_VAL_290_BG },
         { "300", OPTION_VAL_300_BG },
         { "310", OPTION_VAL_310_BG },
         { "320", OPTION_VAL_320_BG },
         { "330", OPTION_VAL_330_BG },
         { "340", OPTION_VAL_340_BG },
         { "350", OPTION_VAL_350_BG },
         { "360", OPTION_VAL_360_BG },
         { "370", OPTION_VAL_370_BG },
         { "380", OPTION_VAL_380_BG },
         { "390", OPTION_VAL_390_BG },
         { "400", OPTION_VAL_400_BG },
         { "410", OPTION_VAL_410_BG },
         { "420", OPTION_VAL_420_BG },
         { "430", OPTION_VAL_430_BG },
         { "440", OPTION_VAL_440_BG },
         { "450", OPTION_VAL_450_BG },
         { "460", OPTION_VAL_460_BG },
         { "470", OPTION_VAL_470_BG },
         { "480", OPTION_VAL_480_BG },
         { "490", OPTION_VAL_490_BG },
         { "500", OPTION_VAL_500_BG },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_BG,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_BG },
         { "5%",  OPTION_VAL_5_BG },
         { "10%", OPTION_VAL_10_BG },
         { "15%", OPTION_VAL_15_BG },
         { "20%", OPTION_VAL_20_BG },
         { "25%", OPTION_VAL_25_BG },
         { "30%", OPTION_VAL_30_BG },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_BG,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_BG },
         { "5%",  OPTION_VAL_5_BG },
         { "10%", OPTION_VAL_10_BG },
         { "15%", OPTION_VAL_15_BG },
         { "20%", OPTION_VAL_20_BG },
         { "25%", OPTION_VAL_25_BG },
         { "30%", OPTION_VAL_30_BG },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_BG,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_BG,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_BG,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_BG,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_BG },
         { "60%",  OPTION_VAL_60_BG },
         { "70%",  OPTION_VAL_70_BG },
         { "80%",  OPTION_VAL_80_BG },
         { "90%",  OPTION_VAL_90_BG },
         { "100%", OPTION_VAL_100_O47_BG },
         { "110%", OPTION_VAL_110_O47_BG },
         { "120%", OPTION_VAL_120_O47_BG },
         { "130%", OPTION_VAL_130_O47_BG },
         { "140%", OPTION_VAL_140_O47_BG },
         { "150%", OPTION_VAL_150_O47_BG },
         { "160%", OPTION_VAL_160_O47_BG },
         { "170%", OPTION_VAL_170_O47_BG },
         { "180%", OPTION_VAL_180_O47_BG },
         { "190%", OPTION_VAL_190_O47_BG },
         { "200%", OPTION_VAL_200_O47_BG },
         { "210%", OPTION_VAL_210_O47_BG },
         { "220%", OPTION_VAL_220_O47_BG },
         { "230%", OPTION_VAL_230_O47_BG },
         { "240%", OPTION_VAL_240_O47_BG },
         { "250%", OPTION_VAL_250_O47_BG },
         { "260%", OPTION_VAL_260_O47_BG },
         { "270%", OPTION_VAL_270_O47_BG },
         { "280%", OPTION_VAL_280_O47_BG },
         { "290%", OPTION_VAL_290_O47_BG },
         { "300%", OPTION_VAL_300_O47_BG },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_BG,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_BG },
         { "Red",      OPTION_VAL_RED_BG },
         { "Green",    OPTION_VAL_GREEN_BG },
         { "Blue",     OPTION_VAL_BLUE_BG },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_BG,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_BG },
         { "Red",      OPTION_VAL_RED_BG },
         { "Green",    OPTION_VAL_GREEN_BG },
         { "Blue",     OPTION_VAL_BLUE_BG },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_BG,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_BG },
         { "Red",      OPTION_VAL_RED_BG },
         { "Green",    OPTION_VAL_GREEN_BG },
         { "Blue",     OPTION_VAL_BLUE_BG },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_BG,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_BG },
         { "Red",      OPTION_VAL_RED_BG },
         { "Green",    OPTION_VAL_GREEN_BG },
         { "Blue",     OPTION_VAL_BLUE_BG },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_BG,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BG },
         { "Purupuru", OPTION_VAL_PURUPURU_BG },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_BG },
         { "None",     OPTION_VAL_NONE_BG },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_BG,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BG },
         { "Purupuru", OPTION_VAL_PURUPURU_BG },
         { "None",     OPTION_VAL_NONE_BG },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_BG,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BG },
         { "Purupuru", OPTION_VAL_PURUPURU_BG },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_BG },
         { "None",     OPTION_VAL_NONE_BG },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_BG,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BG },
         { "Purupuru", OPTION_VAL_PURUPURU_BG },
         { "None",     OPTION_VAL_NONE_BG },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_BG,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BG },
         { "Purupuru", OPTION_VAL_PURUPURU_BG },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_BG },
         { "None",     OPTION_VAL_NONE_BG },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_BG,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BG },
         { "Purupuru", OPTION_VAL_PURUPURU_BG },
         { "None",     OPTION_VAL_NONE_BG },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_BG,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BG },
         { "Purupuru", OPTION_VAL_PURUPURU_BG },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_BG },
         { "None",     OPTION_VAL_NONE_BG },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_BG,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_BG },
         { "Purupuru", OPTION_VAL_PURUPURU_BG },
         { "None",     OPTION_VAL_NONE_BG },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_BG,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_BG,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_BG,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_BG },
         { "All VMUs", OPTION_VAL_ALL_VMUS_BG },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_BG,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_BG,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_BG,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_BG,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_BG,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_BG,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_BG,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_BG,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_BG },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_BG },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_BG },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_BG },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_BG },
         { "2x", OPTION_VAL_2_O26_BG },
         { "3x", OPTION_VAL_3X_BG },
         { "4x", OPTION_VAL_4_BG },
         { "5x", OPTION_VAL_5X_BG },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BG },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BG },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BG },
         { "BLUE 03",           OPTION_VAL_BLUE_BG },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BG },
         { "GREEN 05",          OPTION_VAL_GREEN_BG },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BG },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BG },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BG },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BG },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BG },
         { "RED 11",            OPTION_VAL_RED_BG },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BG },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BG },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BG },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BG },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BG },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BG },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BG },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BG },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BG },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BG },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BG },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BG },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BG },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BG },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BG },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BG },
         { "WHITE 28",          OPTION_VAL_WHITE_BG },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BG },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BG },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BG },
         { "BLUE 03",           OPTION_VAL_BLUE_BG },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BG },
         { "GREEN 05",          OPTION_VAL_GREEN_BG },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BG },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BG },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BG },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BG },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BG },
         { "RED 11",            OPTION_VAL_RED_BG },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BG },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BG },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BG },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BG },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BG },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BG },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BG },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BG },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BG },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BG },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BG },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BG },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BG },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BG },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BG },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BG },
         { "WHITE 28",          OPTION_VAL_WHITE_BG },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_BG },
         { "20%",  OPTION_VAL_20_BG },
         { "30%",  OPTION_VAL_30_BG },
         { "40%",  OPTION_VAL_40_BG },
         { "50%",  OPTION_VAL_50_BG },
         { "60%",  OPTION_VAL_60_BG },
         { "70%",  OPTION_VAL_70_BG },
         { "80%",  OPTION_VAL_80_BG },
         { "90%",  OPTION_VAL_90_BG },
         { "100%", OPTION_VAL_100_O47_BG },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_BG },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_BG },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_BG },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_BG },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_BG },
         { "2x", OPTION_VAL_2_O26_BG },
         { "3x", OPTION_VAL_3X_BG },
         { "4x", OPTION_VAL_4_BG },
         { "5x", OPTION_VAL_5X_BG },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BG },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BG },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BG },
         { "BLUE 03",           OPTION_VAL_BLUE_BG },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BG },
         { "GREEN 05",          OPTION_VAL_GREEN_BG },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BG },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BG },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BG },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BG },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BG },
         { "RED 11",            OPTION_VAL_RED_BG },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BG },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BG },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BG },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BG },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BG },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BG },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BG },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BG },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BG },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BG },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BG },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BG },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BG },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BG },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BG },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BG },
         { "WHITE 28",          OPTION_VAL_WHITE_BG },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BG },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BG },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BG },
         { "BLUE 03",           OPTION_VAL_BLUE_BG },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BG },
         { "GREEN 05",          OPTION_VAL_GREEN_BG },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BG },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BG },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BG },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BG },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BG },
         { "RED 11",            OPTION_VAL_RED_BG },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BG },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BG },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BG },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BG },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BG },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BG },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BG },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BG },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BG },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BG },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BG },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BG },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BG },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BG },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BG },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BG },
         { "WHITE 28",          OPTION_VAL_WHITE_BG },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_BG },
         { "20%",  OPTION_VAL_20_BG },
         { "30%",  OPTION_VAL_30_BG },
         { "40%",  OPTION_VAL_40_BG },
         { "50%",  OPTION_VAL_50_BG },
         { "60%",  OPTION_VAL_60_BG },
         { "70%",  OPTION_VAL_70_BG },
         { "80%",  OPTION_VAL_80_BG },
         { "90%",  OPTION_VAL_90_BG },
         { "100%", OPTION_VAL_100_O47_BG },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_BG },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_BG },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_BG },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_BG },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_BG },
         { "2x", OPTION_VAL_2_O26_BG },
         { "3x", OPTION_VAL_3X_BG },
         { "4x", OPTION_VAL_4_BG },
         { "5x", OPTION_VAL_5X_BG },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BG },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BG },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BG },
         { "BLUE 03",           OPTION_VAL_BLUE_BG },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BG },
         { "GREEN 05",          OPTION_VAL_GREEN_BG },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BG },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BG },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BG },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BG },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BG },
         { "RED 11",            OPTION_VAL_RED_BG },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BG },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BG },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BG },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BG },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BG },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BG },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BG },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BG },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BG },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BG },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BG },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BG },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BG },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BG },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BG },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BG },
         { "WHITE 28",          OPTION_VAL_WHITE_BG },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BG },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BG },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BG },
         { "BLUE 03",           OPTION_VAL_BLUE_BG },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BG },
         { "GREEN 05",          OPTION_VAL_GREEN_BG },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BG },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BG },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BG },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BG },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BG },
         { "RED 11",            OPTION_VAL_RED_BG },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BG },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BG },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BG },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BG },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BG },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BG },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BG },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BG },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BG },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BG },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BG },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BG },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BG },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BG },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BG },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BG },
         { "WHITE 28",          OPTION_VAL_WHITE_BG },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_BG },
         { "20%",  OPTION_VAL_20_BG },
         { "30%",  OPTION_VAL_30_BG },
         { "40%",  OPTION_VAL_40_BG },
         { "50%",  OPTION_VAL_50_BG },
         { "60%",  OPTION_VAL_60_BG },
         { "70%",  OPTION_VAL_70_BG },
         { "80%",  OPTION_VAL_80_BG },
         { "90%",  OPTION_VAL_90_BG },
         { "100%", OPTION_VAL_100_O47_BG },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_BG },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_BG },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_BG },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_BG },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_BG },
         { "2x", OPTION_VAL_2_O26_BG },
         { "3x", OPTION_VAL_3X_BG },
         { "4x", OPTION_VAL_4_BG },
         { "5x", OPTION_VAL_5X_BG },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BG },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BG },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BG },
         { "BLUE 03",           OPTION_VAL_BLUE_BG },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BG },
         { "GREEN 05",          OPTION_VAL_GREEN_BG },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BG },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BG },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BG },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BG },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BG },
         { "RED 11",            OPTION_VAL_RED_BG },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BG },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BG },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BG },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BG },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BG },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BG },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BG },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BG },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BG },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BG },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BG },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BG },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BG },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BG },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BG },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BG },
         { "WHITE 28",          OPTION_VAL_WHITE_BG },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_BG },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_BG },
         { "BLACK 02",          OPTION_VAL_BLACK_02_BG },
         { "BLUE 03",           OPTION_VAL_BLUE_BG },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_BG },
         { "GREEN 05",          OPTION_VAL_GREEN_BG },
         { "CYAN 06",           OPTION_VAL_CYAN_06_BG },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_BG },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_BG },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_BG },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_BG },
         { "RED 11",            OPTION_VAL_RED_BG },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_BG },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_BG },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_BG },
         { "GRAY 15",           OPTION_VAL_GRAY_15_BG },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_BG },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_BG },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_BG },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_BG },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_BG },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_BG },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_BG },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_BG },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_BG },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_BG },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_BG },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_BG },
         { "WHITE 28",          OPTION_VAL_WHITE_BG },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_BG,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_BG },
         { "20%",  OPTION_VAL_20_BG },
         { "30%",  OPTION_VAL_30_BG },
         { "40%",  OPTION_VAL_40_BG },
         { "50%",  OPTION_VAL_50_BG },
         { "60%",  OPTION_VAL_60_BG },
         { "70%",  OPTION_VAL_70_BG },
         { "80%",  OPTION_VAL_80_BG },
         { "90%",  OPTION_VAL_90_BG },
         { "100%", OPTION_VAL_100_O47_BG },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_bg = {
   option_cats_bg,
   option_defs_bg
};

/* RETRO_LANGUAGE_CA */

#define CATEGORY_SYSTEM_LABEL_CA "Sistema"
#define CATEGORY_SYSTEM_INFO_0_CA "Canvia la configuració de regió, idioma, BIOS i del maquinari base."
#define CATEGORY_VIDEO_LABEL_CA "Vídeo"
#define CATEGORY_VIDEO_INFO_0_CA "Configura resolució, les transparències independents de l'ordre i els efectes visuals."
#define CATEGORY_PERFORMANCE_LABEL_CA "Rendiment"
#define CATEGORY_PERFORMANCE_INFO_0_CA "Configura el renderitzat multifil i l'omissió de fotogrames."
#define CATEGORY_HACKS_LABEL_CA "Correccions de l'emulació"
#define CATEGORY_HACKS_INFO_0_CA "Canvia la configuració de les correccions de pantalla panoràmica, la velocitat de càrrega de GD-ROM i la substitució de textures."
#define CATEGORY_INPUT_LABEL_CA "Entrada"
#define CATEGORY_INPUT_INFO_0_CA "Canvia la configuració dels controladors i les pistoles de llum."
#define CATEGORY_EXPANSIONS_LABEL_CA "Ports d'expansió dels controladors"
#define CATEGORY_EXPANSIONS_INFO_0_CA "Selecciona el dispositiu (VMU, dispositiu vibració) connectat a cada port d'expansió."
#define CATEGORY_VMU_LABEL_CA "Unitat de la memòria visual"
#define CATEGORY_VMU_INFO_0_CA "Configura fitxers per desar partides diferents per cada joc i la visibilitat en pantalla per cada joc."
#define CORE_OPTION_NAME_REGION_LABEL_CA "Regió"
#define OPTION_VAL_JAPAN_CA "Japó"
#define OPTION_VAL_USA_CA NULL
#define OPTION_VAL_EUROPE_CA "Europa"
#define OPTION_VAL_DEFAULT_CA "Per defecte"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_CA "Llengua"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_CA "Canvia la llengua que farà servir la BIOS i els jocs que contenen diferents idiomes."
#define OPTION_VAL_JAPANESE_CA "Japonès"
#define OPTION_VAL_ENGLISH_CA "Anglès"
#define OPTION_VAL_GERMAN_CA "Alemany"
#define OPTION_VAL_FRENCH_CA "Francès"
#define OPTION_VAL_SPANISH_CA "Español"
#define OPTION_VAL_ITALIAN_CA "Italià"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_CA "HLE BIOS (Es requereix reinici)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_CA "Obliga l'ús d'una BIOS amb emulació d'alt nivell."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_CA "Habilitar DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CA "Habilita l'emulació del processador d'àudio DSP de Dreamcast. Millora la precisió del so generat, però augmenta els requisits de maquinari."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CA "Permet els botons de servei d'arcade"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CA "Habilita el botó SERVICE per jocs arcade, per accedir a la configuració de la màquina."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CA "Estableix els jocs de NAOMI com a gratuïts"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CA "Modifica la configuració de monedes del joc a partides gratuïtes."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_CA "Emulació de l'adaptador Broadband"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CA "Emula l'adaptador de banda ampla per cable Ethernet en compte del mòdem (es requereix reinici)"
#define CORE_OPTION_NAME_UPNP_LABEL_CA "Habilitat UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_CA "Fes servir UPnP per configurar automàticament el router per jocs en línia."
#define CORE_OPTION_NAME_DCNET_LABEL_CA "Fes servir DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_CA "Fes servir el servei en el núvol de DCNet per l'accés a internet de la Dreamcast."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CA "Resolució Interna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CA "Modifica la resolució interna."
#define OPTION_VAL_320X240_CA "320 × 240 (50 %)"
#define OPTION_VAL_640X480_CA "640x480b (Natiu)"
#define OPTION_VAL_800X600_CA "800x600 (x1,25)"
#define OPTION_VAL_960X720_CA "960 × 720 (150 %)"
#define OPTION_VAL_1024X768_CA "1024x768 (x1,6)"
#define OPTION_VAL_1280X960_CA NULL
#define OPTION_VAL_1440X1080_CA "1440x1080 (x2,25)"
#define OPTION_VAL_1600X1200_CA "1600x1200 (x2,5)"
#define OPTION_VAL_1920X1440_CA "1920 × 1440 (300 %)"
#define OPTION_VAL_2560X1920_CA "2560 × 1920 (400 %)"
#define OPTION_VAL_2880X2160_CA "2880x2160 (x4,5)"
#define OPTION_VAL_3200X2400_CA "3200 × 2400 (500 %)"
#define OPTION_VAL_3840X2880_CA "3840 × 2880 (600 %)"
#define OPTION_VAL_4480X3360_CA NULL
#define OPTION_VAL_5120X3840_CA NULL
#define OPTION_VAL_5760X4320_CA NULL
#define OPTION_VAL_6400X4800_CA NULL
#define OPTION_VAL_7040X5280_CA NULL
#define OPTION_VAL_7680X5760_CA NULL
#define OPTION_VAL_8320X6240_CA NULL
#define OPTION_VAL_8960X6720_CA NULL
#define OPTION_VAL_9600X7200_CA NULL
#define OPTION_VAL_10240X7680_CA NULL
#define OPTION_VAL_10880X8160_CA NULL
#define OPTION_VAL_11520X8640_CA NULL
#define OPTION_VAL_12160X9120_CA NULL
#define OPTION_VAL_12800X9600_CA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_CA "Tipus de cable"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CA "Tipus del senyal de sortida. 'TV (Composició)' és el tipus de senyal més suportat."
#define OPTION_VAL_VGA_CA NULL
#define OPTION_VAL_TV_RGB_CA NULL
#define OPTION_VAL_TV_COMPOSITE_CA "TV (Composició)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_CA "Estàndard de transmissió"
#define OPTION_VAL_NTSC_CA NULL
#define OPTION_VAL_PAL_CA "Pal (Mundial)"
#define OPTION_VAL_PAL_N_CA NULL
#define OPTION_VAL_PAL_M_CA "PAL-M (Brasil)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CA "Orientació de la pantalla"
#define OPTION_VAL_HORIZONTAL_CA "Horitzontal"
#define OPTION_VAL_VERTICAL_CA NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CA "Ordenar per transparències"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_CA "Selecciona l'ordre de dibuix dels polígons transparents."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CA "Per bandes (ràpid però menys acurat)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_CA NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_CA "Per píxel (acurat però lent)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CA "Mida de la memòria intermèdia dels píxels d'acumulació"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CA "És possible que sigui necessari un valor alt perquè la imatge es mostri correctament amb resolucions superiors a la nativa."
#define OPTION_VAL_512MB_CA "512 MB"
#define OPTION_VAL_1GB_CA "1 GB"
#define OPTION_VAL_2GB_CA "2 GB"
#define OPTION_VAL_4GB_CA "4 GB"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_CA "Nombre màxim de capes transparents"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CA "Les escenes més complexes poden necessitar un valor més alt."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CA "Emulació completa de la memòria intermèdia"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CA "Activa l'emulació completa de la memòria intermèdia de fotogrames a la VRAM. Ideal per aquells jocs que llegeixen o escriuen directament a la memòria intermèdia de la VRAM. Si s'activa aquesta opció, s'ajustarà la resolució interna a 640 x 480 i el rendiment es pot veure afectat."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CA "Habilita la memòria intermèdia RTT (Renderitzat a textures)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CA "Copia les textures renderitzades a la targeta gràfica a la memòria VRAM. Aquesta opció cal deixar-la activada per aquells jocs que ho necessitin. En activar aquesta opció, es desactiva l'escalat de textures i el rendiment es pot veure afectat."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_CA NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_CA "En activar aquesta opció, les textures mostraran una versió més reduïda de mida si es troben més allunyades de la pantalla. Pot millorar el rendiment i reduir els efectes de distorsió."
#define CORE_OPTION_NAME_FOG_LABEL_CA "Efectes de boira"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CA "Modificador de volum"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CA "Una característica de la targeta gràfica de Dreamcast que solen fer servir alguns jocs per dibuixar les ombres dels objectes. Aquesta opció cal activar-se en general, no afecta el rendiment."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CA "Filtre anisotròpic"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CA "Millora la qualitat de les textures si es troben en un angle oblic respecte a la càmera. Un valor alt necessitarà més recursos de la targeta gràfica. Cal reiniciar perquè els canvis tinguin efecte."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CA "Filtratge de textures"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CA "Indica el mode de filtratge de textures que es farà servir. Es pot forçar un mode de filtratge concret en totes les textures amb l'objectiu d'aconseguir un aspecte per definit. Qualsevol valor que no sigui el valor per defecte pot provocar errors de renderitzat. Cal reiniciar per aplicar els canvis."
#define OPTION_VAL_1_CA "Força píxel més proper"
#define OPTION_VAL_2_CA "Forçar lineal"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CA "Retardar el canvi de fotogrames"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CA "Útil per evitar parpelleig constant errors gràfics en els vídeos. No es recomana activar-ho en plataformes amb pocs recursos."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CA "Detecta els canvis en la taxa de fotogrames"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CA "Informa al front-end si canvia la freqüència d'actualització interna. Millora la cadència de fotogrames en aquells jocs que s'executin a 30 o 20 fotogrames per segon fixes, però aquesta opció cal desactivar-se en aquells jocs que tinguin freqüències de fotogrames desbloquejades. NOTA: Aquesta opció no està disponible si s'activa Omissió automàtica de fotogrames."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CA "Filtre de postprocessat PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CA "Postprocessa la imatge renderitzada per emular els efectes dedicats de la GPU PowerVR2 i els senyals de vídeo analògic."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_CA "Augment de mida de les textures (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CA "Millora els gràfics de tipus píxel art dibuixats a mà. Activa-ho només per jocs 2D pixelats."
#define OPTION_VAL_2_O26_CA NULL
#define OPTION_VAL_4_CA NULL
#define OPTION_VAL_6_CA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CA "Mida màxima de l'escalat de textures"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_CA "Selecciona un valor màxim per a l'escalat de textures. Si la mida de la textura és superior al valor seleccionat, no s'escalarà."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CA "Interpolació nativa de profunditat"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CA "Esmorteeix la corrupció de textures i els problemes gràfics de profunditat en fer servir una GPU de AMB. En alguns casos també pot ajudar en una GPU de Intel."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_CA "Corregeix els defectes de les vores en escalar"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_CA "Dona suport si les textures mostren vores sagnants en escalar-les. Desactivar aquesta opció pot ser útil si hi ha distorsió de píxels en escalar jocs 2D (MVC2, CVS, KOF, etç...)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CA "Renderitzat multifil"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CA "Executa la GPU i la CPU en diferents fils. Molt recomanat."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CA "Omissió automàtica de fotogrames"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CA "Omet fotogrames automàticament si l'emulador funciona amb lentitud. Nota: Aquesta opció només funciona si l'opció 'Renderitzat multifil' està activada."
#define OPTION_VAL_SOME_CA NULL
#define OPTION_VAL_MORE_CA "Màxim"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CA "Omissió de fotogrames"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CA "Selecciona el nombre de fotogrames a ometre per cada fotograma en pantalla que es mostra."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CA "Correccions per pantalles panoràmiques (Es requereix reinici)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CA "Activa els trucs que permet a certs jocs mostrar-se correctament en pantalles panoràmiques."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CA "Correcció de pantalles panoràmiques"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CA "Dibuixa geometries més enllà de la relació d'aspecte 4:3. Pot produir errors gràfics en les zones visibles."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CA "Lectura ràpida de GD-ROM (poc acurat)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CA "Augmenta la velocitat de càrrega de GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_CA NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_CA "Activa la modificació de 32 MB de RAM per Dreamcast. Pot afectar la compatibilitat"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_CA NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_CA "Canvia la velocitat del rellotge del processador principal SH4 dels 200 MHz per defecte. Baixar la velocitat pot ajudar a les plataformes més lentes. Augmentar la velocitat pot augmentar la velocitat en alguns jocs. Cal utilitzar amb precaució."
#define OPTION_VAL_100_CA NULL
#define OPTION_VAL_110_CA NULL
#define OPTION_VAL_120_CA NULL
#define OPTION_VAL_130_CA NULL
#define OPTION_VAL_140_CA NULL
#define OPTION_VAL_150_CA NULL
#define OPTION_VAL_160_CA NULL
#define OPTION_VAL_170_CA NULL
#define OPTION_VAL_180_CA NULL
#define OPTION_VAL_190_CA NULL
#define OPTION_VAL_200_CA NULL
#define OPTION_VAL_210_CA NULL
#define OPTION_VAL_220_CA NULL
#define OPTION_VAL_230_CA NULL
#define OPTION_VAL_240_CA NULL
#define OPTION_VAL_250_CA NULL
#define OPTION_VAL_260_CA NULL
#define OPTION_VAL_270_CA NULL
#define OPTION_VAL_280_CA NULL
#define OPTION_VAL_290_CA NULL
#define OPTION_VAL_300_CA NULL
#define OPTION_VAL_310_CA NULL
#define OPTION_VAL_320_CA NULL
#define OPTION_VAL_330_CA NULL
#define OPTION_VAL_340_CA NULL
#define OPTION_VAL_350_CA NULL
#define OPTION_VAL_360_CA NULL
#define OPTION_VAL_370_CA NULL
#define OPTION_VAL_380_CA NULL
#define OPTION_VAL_390_CA NULL
#define OPTION_VAL_400_CA NULL
#define OPTION_VAL_410_CA NULL
#define OPTION_VAL_420_CA NULL
#define OPTION_VAL_430_CA NULL
#define OPTION_VAL_440_CA NULL
#define OPTION_VAL_450_CA NULL
#define OPTION_VAL_460_CA NULL
#define OPTION_VAL_470_CA NULL
#define OPTION_VAL_480_CA NULL
#define OPTION_VAL_490_CA NULL
#define OPTION_VAL_500_CA NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CA "Carregar textures personalitzades"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_CA "Carrega textures personalitzades que es trobin a la carpeta 'system/dc/textures/<game-id>/'."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_CA "Carregar textures personalitzades"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_CA "Carrega les textures personalitzades a l'inici. Pot millorar el rendiment però augment l'ús de memòria."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CA "Abocar textures"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_CA "Cada vegada que un joc faci servir una nova textura, aquesta es desarà en un fitxer .png a la carpeta 'system/dc/texdump/<game-id>/'."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_CA "Substitueix les textures"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_CA "Sempre aboca les textures que son reemplaçades per les textures personalitzades."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CA "Zona morta palanca analògica"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_CA "Selecciona la distància a la qual es reconeix la pulsació del gatell."
#define OPTION_VAL_0_CA NULL
#define OPTION_VAL_5_CA NULL
#define OPTION_VAL_10_CA NULL
#define OPTION_VAL_15_CA NULL
#define OPTION_VAL_20_CA NULL
#define OPTION_VAL_25_CA NULL
#define OPTION_VAL_30_CA NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CA "Zona morta dels gatells"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_CA "Selecciona la distància a la qual es reconeix la pulsació del gatell."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CA "Gatells digitals"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_CA "En activar aquesta opció, els gatells funcionaran com botons normals, el que significa que el senyal serà polsat o no polsat, sense valor intermedi."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CA "Transmet les sortides digitals"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CA "Transmet els senyals digitals i l'estat de vibració del port TCP 8000. És compatible amb l'opció de MAME '-output network'."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CA "Mostra la configuració de la pistola de llum"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CA "Permet configurar les diferents mires de les pistoles de llum. NOTA: És imprescindible sortir i tornar a entrar al menú ràpid perquè el canvi funcioni."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_CA "Escala del punt de mira de la pistola"
#define OPTION_VAL_50_CA NULL
#define OPTION_VAL_60_CA NULL
#define OPTION_VAL_70_CA NULL
#define OPTION_VAL_80_CA NULL
#define OPTION_VAL_90_CA NULL
#define OPTION_VAL_100_O47_CA NULL
#define OPTION_VAL_110_O47_CA NULL
#define OPTION_VAL_120_O47_CA NULL
#define OPTION_VAL_130_O47_CA NULL
#define OPTION_VAL_140_O47_CA NULL
#define OPTION_VAL_150_O47_CA NULL
#define OPTION_VAL_160_O47_CA NULL
#define OPTION_VAL_170_O47_CA NULL
#define OPTION_VAL_180_O47_CA NULL
#define OPTION_VAL_190_O47_CA NULL
#define OPTION_VAL_200_O47_CA NULL
#define OPTION_VAL_210_O47_CA NULL
#define OPTION_VAL_220_O47_CA NULL
#define OPTION_VAL_230_O47_CA NULL
#define OPTION_VAL_240_O47_CA NULL
#define OPTION_VAL_250_O47_CA NULL
#define OPTION_VAL_260_O47_CA NULL
#define OPTION_VAL_270_O47_CA NULL
#define OPTION_VAL_280_O47_CA NULL
#define OPTION_VAL_290_O47_CA NULL
#define OPTION_VAL_300_O47_CA NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CA "Mostra el punt de mira de la pistola 1"
#define OPTION_VAL_WHITE_CA "Blanc"
#define OPTION_VAL_RED_CA "Vermell"
#define OPTION_VAL_GREEN_CA "Verd"
#define OPTION_VAL_BLUE_CA "Blau"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CA "Mostra el punt de mira de la pistola 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CA "Mostra el punt de mira de la pistola 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CA "Mostra el punt de mira de la pistola 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_CA "Dispositu en el port d'expansió A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_CA "Selecciona el dispositiu connectat a la ranura d'expansió A1 (port A ranura 1)."
#define OPTION_VAL_VMU_CA NULL
#define OPTION_VAL_PURUPURU_CA "Pack de vibració"
#define OPTION_VAL_DREAMPOTATO_CA NULL
#define OPTION_VAL_NONE_CA "Cap"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_CA "Dispositiu en el port d'expansió A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_CA "Selecciona el dispositiu connectat a la ranura d'expansió A2 (port A ranura 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_CA "Dispositiu en el port d'expansió B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_CA "Selecciona el dispositiu connectat a la ranura d'expansió B1 (port B ranura 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_CA "Dispostiu en el port d'expanció B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_CA "Selecciona el dispositiu connectat a la ranura d'expansió B2 (port B ranura 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_CA "Dispositiu en el port d'expansió C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_CA "Selecciona el dispositiu connectat a la ranura d'expansió C1 (port C ranura 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_CA "Dispositiu en el port d'expansió C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_CA "Selecciona el dispositiu connectat a la ranura d'expansió C2 (port C ranura 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_CA "Dispositiu en el port d'expansió D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_CA "Selecciona el dispositiu connectat a la ranura d'expansió D1 (port D ranura 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_CA "Dispositiu en el port d'expanció D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_CA "Selecciona el dispositiu connectat a la ranura d'expansió D2 (port D ranura 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CA "Separar Visual Memory (VMU) per cada joc"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CA "VMUs individuals per cada joc"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CA "En desactivar aquesta opció, tots els jocs compartiran els 8 fitxers de desats de VMU (A1/A2/B1/B2/C1/C2/D1/D2) que es troben a la carpeta del sistema de RetroArch. L'opció \"VMU A1\" crearà un únic fitxer de la VMU A1 a la carpeta de desats de RetroArch per cada joc que s'executi. L'opció \"Totes les VMU\" crearà 8 fitxers VMU únics per cada joc que s'executi."
#define OPTION_VAL_VMU_A1_CA NULL
#define OPTION_VAL_ALL_VMUS_CA "Totes les VMUs"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CA "Sons de les Visual Memory (VMU)"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_CA "Sons VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_CA "Quan està actiu, es reprodueixen els sons de la VMU."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_CA "Emmagatzematge VMU DreamPotato"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_CA "Activa l'emmagatzematge de dades VMU a l'emulador DreamPotato. Quan estigui actiu, no es fa servir l'emmagatzematge local."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CA "Mostra la configuració de visualització de les Visual Memory (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CA "Mostra la configuració de pantalla de VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CA "Permet configurar la visibilitat, mida i posició de les pantalles LCD emulades. NOTA: cal sortir i tornar a entrar al menú ràpid perquè el canvi funcioni."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CA "Pantalla de la VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CA "Posició de la pantalla de la VMU 1"
#define OPTION_VAL_UPPER_LEFT_CA "Superior esquerra"
#define OPTION_VAL_UPPER_RIGHT_CA "Superior dret"
#define OPTION_VAL_LOWER_LEFT_CA "Inferiro esquerra"
#define OPTION_VAL_LOWER_RIGHT_CA "Inferior dret"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CA "Mida de la pantalla de la VMU 1"
#define OPTION_VAL_1X_CA NULL
#define OPTION_VAL_3X_CA NULL
#define OPTION_VAL_5X_CA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CA "Color dels píxels actius de la VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_CA "Activat per defecte"
#define OPTION_VAL_DEFAULT_OFF_01_CA "Desactivat per defecte"
#define OPTION_VAL_BLACK_02_CA "Negre"
#define OPTION_VAL_LIGHT_BLUE_04_CA "Blau clar"
#define OPTION_VAL_CYAN_06_CA "Cian"
#define OPTION_VAL_CYAN_BLUE_07_CA "Blau cian"
#define OPTION_VAL_LIGHT_GREEN_08_CA "Verd lluminós"
#define OPTION_VAL_CYAN_GREEN_09_CA "Verd cian"
#define OPTION_VAL_LIGHT_CYAN_10_CA "Cian clar"
#define OPTION_VAL_PURPLE_12_CA "Lila"
#define OPTION_VAL_LIGHT_PURPLE_13_CA "Lila clar"
#define OPTION_VAL_YELLOW_14_CA "Groc"
#define OPTION_VAL_GRAY_15_CA "Gris"
#define OPTION_VAL_LIGHT_PURPLE_2_16_CA "Violeta clar (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_CA "Verd clar (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_CA "Verd clar (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_CA "Cian clar (2)"
#define OPTION_VAL_LIGHT_RED_2_20_CA "Vermell clar (2)"
#define OPTION_VAL_MAGENTA_21_CA NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_CA "Violeta clar (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_CA "Taronja clar"
#define OPTION_VAL_ORANGE_24_CA "Taronja"
#define OPTION_VAL_LIGHT_PURPLE_4_25_CA "Violeta clar (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_CA "Groc clar"
#define OPTION_VAL_LIGHT_YELLOW_2_27_CA "Groc clar (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CA "Color dels píxels inactius de la VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CA "Opacitat de la pantalla de la VMU 1"
#define OPTION_VAL_40_CA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CA "Pantalla de la VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CA "Posició de la pantalla de la VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CA "Mida de la pantalla de la VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CA "Color dels píxels actius de la VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CA "Color dels píxels inactius de la VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CA "Opacitat de la pantalla de la VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CA "Pantalla de la VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CA "Posició de la pantalla de la VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CA "Mida de la pantalla de la VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CA "Color dels píxels actius de la VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CA "Color dels píxels inactius de la VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CA "Opacitat de la pantalla de la VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CA "Pantalla de la VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CA "Posició de la pantalla de la VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CA "Mida de la pantalla de la VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CA "Color dels píxels actius de la VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CA "Color dels píxels inactius de la VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CA "Opacitat de la pantalla de la VMU 4"

struct retro_core_option_v2_category option_cats_ca[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CA,
      CATEGORY_SYSTEM_INFO_0_CA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CA,
      CATEGORY_VIDEO_INFO_0_CA
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_CA,
      CATEGORY_PERFORMANCE_INFO_0_CA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CA,
      CATEGORY_HACKS_INFO_0_CA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CA,
      CATEGORY_INPUT_INFO_0_CA
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_CA,
      CATEGORY_EXPANSIONS_INFO_0_CA
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_CA,
      CATEGORY_VMU_INFO_0_CA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ca[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_CA,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_CA },
         { "USA",     OPTION_VAL_USA_CA },
         { "Europe",  OPTION_VAL_EUROPE_CA },
         { "Default", OPTION_VAL_DEFAULT_CA },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_CA,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_CA },
         { "English",  OPTION_VAL_ENGLISH_CA },
         { "German",   OPTION_VAL_GERMAN_CA },
         { "French",   OPTION_VAL_FRENCH_CA },
         { "Spanish",  OPTION_VAL_SPANISH_CA },
         { "Italian",  OPTION_VAL_ITALIAN_CA },
         { "Default",  OPTION_VAL_DEFAULT_CA },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CA,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_CA },
         { "640x480",    OPTION_VAL_640X480_CA },
         { "800x600",    OPTION_VAL_800X600_CA },
         { "960x720",    OPTION_VAL_960X720_CA },
         { "1024x768",   OPTION_VAL_1024X768_CA },
         { "1280x960",   OPTION_VAL_1280X960_CA },
         { "1440x1080",  OPTION_VAL_1440X1080_CA },
         { "1600x1200",  OPTION_VAL_1600X1200_CA },
         { "1920x1440",  OPTION_VAL_1920X1440_CA },
         { "2560x1920",  OPTION_VAL_2560X1920_CA },
         { "2880x2160",  OPTION_VAL_2880X2160_CA },
         { "3200x2400",  OPTION_VAL_3200X2400_CA },
         { "3840x2880",  OPTION_VAL_3840X2880_CA },
         { "4480x3360",  OPTION_VAL_4480X3360_CA },
         { "5120x3840",  OPTION_VAL_5120X3840_CA },
         { "5760x4320",  OPTION_VAL_5760X4320_CA },
         { "6400x4800",  OPTION_VAL_6400X4800_CA },
         { "7040x5280",  OPTION_VAL_7040X5280_CA },
         { "7680x5760",  OPTION_VAL_7680X5760_CA },
         { "8320x6240",  OPTION_VAL_8320X6240_CA },
         { "8960x6720",  OPTION_VAL_8960X6720_CA },
         { "9600x7200",  OPTION_VAL_9600X7200_CA },
         { "10240x7680", OPTION_VAL_10240X7680_CA },
         { "10880x8160", OPTION_VAL_10880X8160_CA },
         { "11520x8640", OPTION_VAL_11520X8640_CA },
         { "12160x9120", OPTION_VAL_12160X9120_CA },
         { "12800x9600", OPTION_VAL_12800X9600_CA },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CA,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_CA },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_CA },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_CA },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_CA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_CA },
         { "PAL",     OPTION_VAL_PAL_CA },
         { "PAL_N",   OPTION_VAL_PAL_N_CA },
         { "PAL_M",   OPTION_VAL_PAL_M_CA },
         { "Default", OPTION_VAL_DEFAULT_CA },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_CA },
         { "vertical",   OPTION_VAL_VERTICAL_CA },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_CA,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CA },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_CA },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_CA },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CA,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_CA },
         { "1GB",   OPTION_VAL_1GB_CA },
         { "2GB",   OPTION_VAL_2GB_CA },
         { "4GB",   OPTION_VAL_4GB_CA },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CA,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_CA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CA,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CA,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_CA },
         { "1",  OPTION_VAL_1_CA },
         { "2",  OPTION_VAL_2_CA },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CA,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_CA },
         { "4",  OPTION_VAL_4_CA },
         { "6",  OPTION_VAL_6_CA },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_CA,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CA,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_CA,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_CA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_CA },
         { "more", OPTION_VAL_MORE_CA },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CA,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_CA },
         { "110", OPTION_VAL_110_CA },
         { "120", OPTION_VAL_120_CA },
         { "130", OPTION_VAL_130_CA },
         { "140", OPTION_VAL_140_CA },
         { "150", OPTION_VAL_150_CA },
         { "160", OPTION_VAL_160_CA },
         { "170", OPTION_VAL_170_CA },
         { "180", OPTION_VAL_180_CA },
         { "190", OPTION_VAL_190_CA },
         { "200", OPTION_VAL_200_CA },
         { "210", OPTION_VAL_210_CA },
         { "220", OPTION_VAL_220_CA },
         { "230", OPTION_VAL_230_CA },
         { "240", OPTION_VAL_240_CA },
         { "250", OPTION_VAL_250_CA },
         { "260", OPTION_VAL_260_CA },
         { "270", OPTION_VAL_270_CA },
         { "280", OPTION_VAL_280_CA },
         { "290", OPTION_VAL_290_CA },
         { "300", OPTION_VAL_300_CA },
         { "310", OPTION_VAL_310_CA },
         { "320", OPTION_VAL_320_CA },
         { "330", OPTION_VAL_330_CA },
         { "340", OPTION_VAL_340_CA },
         { "350", OPTION_VAL_350_CA },
         { "360", OPTION_VAL_360_CA },
         { "370", OPTION_VAL_370_CA },
         { "380", OPTION_VAL_380_CA },
         { "390", OPTION_VAL_390_CA },
         { "400", OPTION_VAL_400_CA },
         { "410", OPTION_VAL_410_CA },
         { "420", OPTION_VAL_420_CA },
         { "430", OPTION_VAL_430_CA },
         { "440", OPTION_VAL_440_CA },
         { "450", OPTION_VAL_450_CA },
         { "460", OPTION_VAL_460_CA },
         { "470", OPTION_VAL_470_CA },
         { "480", OPTION_VAL_480_CA },
         { "490", OPTION_VAL_490_CA },
         { "500", OPTION_VAL_500_CA },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_CA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CA },
         { "5%",  OPTION_VAL_5_CA },
         { "10%", OPTION_VAL_10_CA },
         { "15%", OPTION_VAL_15_CA },
         { "20%", OPTION_VAL_20_CA },
         { "25%", OPTION_VAL_25_CA },
         { "30%", OPTION_VAL_30_CA },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_CA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CA },
         { "5%",  OPTION_VAL_5_CA },
         { "10%", OPTION_VAL_10_CA },
         { "15%", OPTION_VAL_15_CA },
         { "20%", OPTION_VAL_20_CA },
         { "25%", OPTION_VAL_25_CA },
         { "30%", OPTION_VAL_30_CA },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_CA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CA,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_CA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_CA },
         { "60%",  OPTION_VAL_60_CA },
         { "70%",  OPTION_VAL_70_CA },
         { "80%",  OPTION_VAL_80_CA },
         { "90%",  OPTION_VAL_90_CA },
         { "100%", OPTION_VAL_100_O47_CA },
         { "110%", OPTION_VAL_110_O47_CA },
         { "120%", OPTION_VAL_120_O47_CA },
         { "130%", OPTION_VAL_130_O47_CA },
         { "140%", OPTION_VAL_140_O47_CA },
         { "150%", OPTION_VAL_150_O47_CA },
         { "160%", OPTION_VAL_160_O47_CA },
         { "170%", OPTION_VAL_170_O47_CA },
         { "180%", OPTION_VAL_180_O47_CA },
         { "190%", OPTION_VAL_190_O47_CA },
         { "200%", OPTION_VAL_200_O47_CA },
         { "210%", OPTION_VAL_210_O47_CA },
         { "220%", OPTION_VAL_220_O47_CA },
         { "230%", OPTION_VAL_230_O47_CA },
         { "240%", OPTION_VAL_240_O47_CA },
         { "250%", OPTION_VAL_250_O47_CA },
         { "260%", OPTION_VAL_260_O47_CA },
         { "270%", OPTION_VAL_270_O47_CA },
         { "280%", OPTION_VAL_280_O47_CA },
         { "290%", OPTION_VAL_290_O47_CA },
         { "300%", OPTION_VAL_300_O47_CA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CA },
         { "Red",      OPTION_VAL_RED_CA },
         { "Green",    OPTION_VAL_GREEN_CA },
         { "Blue",     OPTION_VAL_BLUE_CA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CA },
         { "Red",      OPTION_VAL_RED_CA },
         { "Green",    OPTION_VAL_GREEN_CA },
         { "Blue",     OPTION_VAL_BLUE_CA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CA },
         { "Red",      OPTION_VAL_RED_CA },
         { "Green",    OPTION_VAL_GREEN_CA },
         { "Blue",     OPTION_VAL_BLUE_CA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CA },
         { "Red",      OPTION_VAL_RED_CA },
         { "Green",    OPTION_VAL_GREEN_CA },
         { "Blue",     OPTION_VAL_BLUE_CA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_CA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CA },
         { "Purupuru", OPTION_VAL_PURUPURU_CA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CA },
         { "None",     OPTION_VAL_NONE_CA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_CA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CA },
         { "Purupuru", OPTION_VAL_PURUPURU_CA },
         { "None",     OPTION_VAL_NONE_CA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_CA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CA },
         { "Purupuru", OPTION_VAL_PURUPURU_CA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CA },
         { "None",     OPTION_VAL_NONE_CA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_CA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CA },
         { "Purupuru", OPTION_VAL_PURUPURU_CA },
         { "None",     OPTION_VAL_NONE_CA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_CA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CA },
         { "Purupuru", OPTION_VAL_PURUPURU_CA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CA },
         { "None",     OPTION_VAL_NONE_CA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_CA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CA },
         { "Purupuru", OPTION_VAL_PURUPURU_CA },
         { "None",     OPTION_VAL_NONE_CA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_CA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CA },
         { "Purupuru", OPTION_VAL_PURUPURU_CA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CA },
         { "None",     OPTION_VAL_NONE_CA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_CA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CA },
         { "Purupuru", OPTION_VAL_PURUPURU_CA },
         { "None",     OPTION_VAL_NONE_CA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_CA },
         { "All VMUs", OPTION_VAL_ALL_VMUS_CA },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CA,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_CA,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_CA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_CA,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_CA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CA,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CA },
         { "2x", OPTION_VAL_2_O26_CA },
         { "3x", OPTION_VAL_3X_CA },
         { "4x", OPTION_VAL_4_CA },
         { "5x", OPTION_VAL_5X_CA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CA },
         { "20%",  OPTION_VAL_20_CA },
         { "30%",  OPTION_VAL_30_CA },
         { "40%",  OPTION_VAL_40_CA },
         { "50%",  OPTION_VAL_50_CA },
         { "60%",  OPTION_VAL_60_CA },
         { "70%",  OPTION_VAL_70_CA },
         { "80%",  OPTION_VAL_80_CA },
         { "90%",  OPTION_VAL_90_CA },
         { "100%", OPTION_VAL_100_O47_CA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CA },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CA },
         { "2x", OPTION_VAL_2_O26_CA },
         { "3x", OPTION_VAL_3X_CA },
         { "4x", OPTION_VAL_4_CA },
         { "5x", OPTION_VAL_5X_CA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CA },
         { "20%",  OPTION_VAL_20_CA },
         { "30%",  OPTION_VAL_30_CA },
         { "40%",  OPTION_VAL_40_CA },
         { "50%",  OPTION_VAL_50_CA },
         { "60%",  OPTION_VAL_60_CA },
         { "70%",  OPTION_VAL_70_CA },
         { "80%",  OPTION_VAL_80_CA },
         { "90%",  OPTION_VAL_90_CA },
         { "100%", OPTION_VAL_100_O47_CA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CA },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CA },
         { "2x", OPTION_VAL_2_O26_CA },
         { "3x", OPTION_VAL_3X_CA },
         { "4x", OPTION_VAL_4_CA },
         { "5x", OPTION_VAL_5X_CA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CA },
         { "20%",  OPTION_VAL_20_CA },
         { "30%",  OPTION_VAL_30_CA },
         { "40%",  OPTION_VAL_40_CA },
         { "50%",  OPTION_VAL_50_CA },
         { "60%",  OPTION_VAL_60_CA },
         { "70%",  OPTION_VAL_70_CA },
         { "80%",  OPTION_VAL_80_CA },
         { "90%",  OPTION_VAL_90_CA },
         { "100%", OPTION_VAL_100_O47_CA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CA },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CA },
         { "2x", OPTION_VAL_2_O26_CA },
         { "3x", OPTION_VAL_3X_CA },
         { "4x", OPTION_VAL_4_CA },
         { "5x", OPTION_VAL_5X_CA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CA },
         { "BLUE 03",           OPTION_VAL_BLUE_CA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CA },
         { "GREEN 05",          OPTION_VAL_GREEN_CA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CA },
         { "RED 11",            OPTION_VAL_RED_CA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CA },
         { "WHITE 28",          OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CA },
         { "20%",  OPTION_VAL_20_CA },
         { "30%",  OPTION_VAL_30_CA },
         { "40%",  OPTION_VAL_40_CA },
         { "50%",  OPTION_VAL_50_CA },
         { "60%",  OPTION_VAL_60_CA },
         { "70%",  OPTION_VAL_70_CA },
         { "80%",  OPTION_VAL_80_CA },
         { "90%",  OPTION_VAL_90_CA },
         { "100%", OPTION_VAL_100_O47_CA },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ca = {
   option_cats_ca,
   option_defs_ca
};

/* RETRO_LANGUAGE_CHS */

#define CATEGORY_SYSTEM_LABEL_CHS "系统"
#define CATEGORY_SYSTEM_INFO_0_CHS "配置区域、语言、BIOS 和基础硬件设置。"
#define CATEGORY_VIDEO_LABEL_CHS "视频"
#define CATEGORY_VIDEO_INFO_0_CHS "配置分辨率、与顺序无关的透明度和视觉效果设置。"
#define CATEGORY_PERFORMANCE_LABEL_CHS "性能"
#define CATEGORY_PERFORMANCE_INFO_0_CHS "配置多线程渲染和跳帧设置。"
#define CATEGORY_HACKS_LABEL_CHS "模拟作弊"
#define CATEGORY_HACKS_INFO_0_CHS "配置宽屏覆盖、GD-ROM 加载速度和纹理替换设置。"
#define CATEGORY_INPUT_LABEL_CHS "输入"
#define CATEGORY_INPUT_INFO_0_CHS "配置手柄和光枪设置。"
#define CATEGORY_EXPANSIONS_LABEL_CHS "控制器扩展插槽"
#define CATEGORY_EXPANSIONS_INFO_0_CHS "选择插入每个控制器扩展插槽中的设备（VMU、震动设备）。"
#define CATEGORY_VMU_LABEL_CHS "视觉内存单元（VMU）"
#define CATEGORY_VMU_INFO_0_CHS "配置每款游戏的 VMU 存档文件和屏幕上的 VMU 可见性设置。"
#define CORE_OPTION_NAME_REGION_LABEL_CHS "区域"
#define OPTION_VAL_JAPAN_CHS "日本"
#define OPTION_VAL_USA_CHS "美国"
#define OPTION_VAL_EUROPE_CHS "欧洲"
#define OPTION_VAL_DEFAULT_CHS "默认"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_CHS "语言"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_CHS "更改 BIOS 以及包含多种语言的游戏所使用的语言。"
#define OPTION_VAL_JAPANESE_CHS "日语"
#define OPTION_VAL_ENGLISH_CHS "英语"
#define OPTION_VAL_GERMAN_CHS "德语"
#define OPTION_VAL_FRENCH_CHS "法语"
#define OPTION_VAL_SPANISH_CHS "西班牙语"
#define OPTION_VAL_ITALIAN_CHS "意大利语"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_CHS "HLE BIOS（需重启）"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_CHS "强制使用高级模拟 BIOS。"
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_CHS "启用 DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CHS "启用 Dreamcast 音频 DSP（数字信号处理器）的模拟。可提升生成声音的准确性，但会增加性能需求。"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CHS "允许街机服务按钮"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CHS "为街机游戏启用 SERVICE 按钮，以进入机柜设置。"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CHS "将 NAOMI 游戏设为免费游玩"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CHS "修改游戏的投币设置，使其变为免费游玩模式。"
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_CHS "宽带适配器模拟"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CHS "模拟以太网宽带适配器而非调制解调器。（需重启）"
#define CORE_OPTION_NAME_UPNP_LABEL_CHS "启用 UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_CHS "使用 UPnP 自动配置您的互联网路由器，以便进行在线游戏。"
#define CORE_OPTION_NAME_DCNET_LABEL_CHS "使用 DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_CHS "使用 DCNet 云服务访问 Dreamcast 网络。"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CHS "内部分辨率"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CHS "修改渲染分辨率。"
#define OPTION_VAL_320X240_CHS "320x240（一半）"
#define OPTION_VAL_640X480_CHS "640x480（原生）"
#define OPTION_VAL_800X600_CHS "800x600（x1.25）"
#define OPTION_VAL_960X720_CHS "960x720（x1.5）"
#define OPTION_VAL_1024X768_CHS "1024x768（x1.6）"
#define OPTION_VAL_1280X960_CHS "1280x960（x2）"
#define OPTION_VAL_1440X1080_CHS "1440x1080（x2.25）"
#define OPTION_VAL_1600X1200_CHS "1600x1200（x2.5）"
#define OPTION_VAL_1920X1440_CHS "1920x1440（x3）"
#define OPTION_VAL_2560X1920_CHS "2560x1920（x4）"
#define OPTION_VAL_2880X2160_CHS "2880x2160（x4.5）"
#define OPTION_VAL_3200X2400_CHS "3200x2400（x5）"
#define OPTION_VAL_3840X2880_CHS "3840x2880（x6）"
#define OPTION_VAL_4480X3360_CHS "4480x3360（x7）"
#define OPTION_VAL_5120X3840_CHS "5120x3840（x8）"
#define OPTION_VAL_5760X4320_CHS "5760x4320（x9）"
#define OPTION_VAL_6400X4800_CHS "6400x4800（x10）"
#define OPTION_VAL_7040X5280_CHS "7040x5280（x11）"
#define OPTION_VAL_7680X5760_CHS "7680x5760（x12）"
#define OPTION_VAL_8320X6240_CHS "8320x6240（x13）"
#define OPTION_VAL_8960X6720_CHS "8960x6720（x14）"
#define OPTION_VAL_9600X7200_CHS "9600x7200（x15）"
#define OPTION_VAL_10240X7680_CHS "10240x7680（x16）"
#define OPTION_VAL_10880X8160_CHS "10880x8160（x17）"
#define OPTION_VAL_11520X8640_CHS "11520x8640（x18）"
#define OPTION_VAL_12160X9120_CHS "12160x9120（x19）"
#define OPTION_VAL_12800X9600_CHS "12800x9600（x20）"
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_CHS "线缆类型"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CHS "输出信号类型。“电脑复合（Composite）” 是最广泛支持的。"
#define OPTION_VAL_VGA_CHS NULL
#define OPTION_VAL_TV_RGB_CHS "电视 (RGB)"
#define OPTION_VAL_TV_COMPOSITE_CHS "电视（复合）"
#define CORE_OPTION_NAME_BROADCAST_LABEL_CHS "广播制式"
#define OPTION_VAL_NTSC_CHS NULL
#define OPTION_VAL_PAL_CHS "PAL（全球）"
#define OPTION_VAL_PAL_N_CHS "PAL-N（阿根廷，巴拉圭，乌拉圭）"
#define OPTION_VAL_PAL_M_CHS "PAL-M（巴西）"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CHS "屏幕方向"
#define OPTION_VAL_HORIZONTAL_CHS "水平"
#define OPTION_VAL_VERTICAL_CHS "垂直"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CHS "阿尔法排序"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_CHS "选择如何对透明多边形进行排序。"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CHS "按条带（快速，最不精确）"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_CHS "按三角形（正常）"
#define OPTION_VAL_PER_PIXEL_ACCURATE_CHS "按像素（精确，但最慢）"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CHS "累积像素缓冲区大小"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CHS "更高的值可能在更高分辨率下才能正确输出。"
#define OPTION_VAL_512MB_CHS "512M"
#define OPTION_VAL_1GB_CHS NULL
#define OPTION_VAL_2GB_CHS NULL
#define OPTION_VAL_4GB_CHS NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_CHS "最大透明图层"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CHS "复杂场景可能需要更高的值。"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CHS "完整帧缓冲区模拟"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CHS "在 VRAM 中启用完整的帧缓冲区模拟。这对于直接读取或写入 VRAM 帧缓冲区的游戏很有用。启用后，内部分辨率将被强制设为 640x480，且性能可能会受到严重影响。"
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CHS "启用 RTT（渲染到纹理）缓冲区"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CHS "将渲染的纹理从 GPU 复制回 VRAM。此选项通常为需要它的游戏启用。启用后，纹理渲染上采样将被禁用，且性能可能会受到影响。"
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_CHS NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_CHS "启用后，纹理在远离视角时会使用更小的版本，这可以提升性能并减少闪烁。"
#define CORE_OPTION_NAME_FOG_LABEL_CHS "雾效"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CHS "体积修正器"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CHS "一项 Dreamcast GPU 特性，通常用于游戏绘制物体阴影。通常应保持启用状态——其对性能的影响通常微乎其微或可忽略不计。"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CHS "各向异性过滤"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CHS "提升相对于相机呈倾斜角度的表面纹理质量。更高的值对 GPU 要求更高。更改此设置需重启后才生效。"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CHS "纹理过滤"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CHS "要使用的纹理过滤模式。可用于强制所有纹理使用特定的过滤模式，以获得比默认模式更清晰（或更平滑）的外观。除“默认”外的其他值可能导致各种渲染问题。更改此设置需重启后才生效。"
#define OPTION_VAL_1_CHS "强制最近邻"
#define OPTION_VAL_2_CHS "强制线性"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CHS "延迟帧交换"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CHS "有助于避免闪屏或画面故障。不推荐在性能较慢的平台上使用。"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CHS "检测帧率变化"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CHS "当内部帧率发生变化时（例如从 60 fps 变为 30 fps），通知前端。可改善以锁定 30 fps 或 20 fps 运行的游戏的帧速率稳定性，但应在帧率未锁定（不稳定）的游戏（如《海豚历险记》、《虚幻竞技场》）中禁用。注意：当“自动跳帧”启用时，此项不可用。"
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CHS "PowerVR2 后处理滤镜"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CHS "对渲染图像进行后处理，以模拟 PowerVR2 GPU 和模拟视频信号特有的效果。"
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_CHS "纹理放大（xBRZ）"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CHS "增强手绘 2D 像素艺术图形。仅应与 2D 像素化游戏配合使用。"
#define OPTION_VAL_2_O26_CHS NULL
#define OPTION_VAL_4_CHS NULL
#define OPTION_VAL_6_CHS NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CHS "纹理放大最大过滤尺寸"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_CHS "为要放大的纹理选择一个最大尺寸值。如果纹理尺寸高于所选值，则不会进行放大。"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CHS "原生深度插值"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CHS "有助于解决 AMD GPU 上的纹理损坏和深度问题。在某些情况下也可能对 Intel GPU 有帮助。"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_CHS "修复放大边缘渗色"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_CHS "在放大时有助于解决纹理渗色问题。如果在 2D 游戏（如 MVC2、CVS、KOF 等）放大时像素出现扭曲，禁用它可能会有所帮助"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CHS "多线程渲染"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CHS "在不同线程上运行 GPU 和 CPU。强烈推荐。"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CHS "自动跳帧"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CHS "当模拟器运行缓慢时自动跳过帧。注意：此设置仅在“多线程渲染”启用时才生效。"
#define OPTION_VAL_SOME_CHS "正常"
#define OPTION_VAL_MORE_CHS "最大"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CHS "帧跳过"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CHS "设置每显示一帧之间要跳过的帧数。"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CHS "宽屏作弊（需重启）"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CHS "启用允许某些游戏以宽屏格式显示的作弊码。"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CHS "宽屏破解"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CHS "在正常的 4:3 宽高比之外绘制几何图形。可能在暴露的区域产生图形故障。"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CHS "GD-ROM 快速加载（不准确）"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CHS "加快 GD-ROM 的加载速度。"
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_CHS "Dreamcast 32M 内存模块"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_CHS "为 Dreamcast 启用 32MB RAM 模组。可能影响兼容性"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_CHS "SH4 CPU 降频/超频"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_CHS "将 SH4 主 CPU 时钟从默认的 200 MHz 更改。降频可能有助于性能较慢的平台。超频可能会提高某些游戏的帧率，请谨慎使用。"
#define OPTION_VAL_100_CHS NULL
#define OPTION_VAL_110_CHS NULL
#define OPTION_VAL_120_CHS NULL
#define OPTION_VAL_130_CHS NULL
#define OPTION_VAL_140_CHS NULL
#define OPTION_VAL_150_CHS NULL
#define OPTION_VAL_160_CHS NULL
#define OPTION_VAL_170_CHS NULL
#define OPTION_VAL_180_CHS NULL
#define OPTION_VAL_190_CHS NULL
#define OPTION_VAL_200_CHS NULL
#define OPTION_VAL_210_CHS NULL
#define OPTION_VAL_220_CHS NULL
#define OPTION_VAL_230_CHS NULL
#define OPTION_VAL_240_CHS NULL
#define OPTION_VAL_250_CHS NULL
#define OPTION_VAL_260_CHS NULL
#define OPTION_VAL_270_CHS NULL
#define OPTION_VAL_280_CHS NULL
#define OPTION_VAL_290_CHS NULL
#define OPTION_VAL_300_CHS NULL
#define OPTION_VAL_310_CHS NULL
#define OPTION_VAL_320_CHS NULL
#define OPTION_VAL_330_CHS NULL
#define OPTION_VAL_340_CHS NULL
#define OPTION_VAL_350_CHS NULL
#define OPTION_VAL_360_CHS NULL
#define OPTION_VAL_370_CHS NULL
#define OPTION_VAL_380_CHS NULL
#define OPTION_VAL_390_CHS NULL
#define OPTION_VAL_400_CHS NULL
#define OPTION_VAL_410_CHS NULL
#define OPTION_VAL_420_CHS NULL
#define OPTION_VAL_430_CHS NULL
#define OPTION_VAL_440_CHS NULL
#define OPTION_VAL_450_CHS NULL
#define OPTION_VAL_460_CHS NULL
#define OPTION_VAL_470_CHS NULL
#define OPTION_VAL_480_CHS NULL
#define OPTION_VAL_490_CHS NULL
#define OPTION_VAL_500_CHS NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CHS "加载自定义纹理"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_CHS "加载位于 'system/dc/textures/<游戏ID>/' 文件夹中的自定义纹理。"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_CHS NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_CHS NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CHS "转储纹理"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_CHS "每当游戏使用一个新纹理时，它将被保存为一个 .png 文件，存放于 'system/dc/texdump/<游戏ID>/' 文件夹中。"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_CHS NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_CHS NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CHS "模拟摇杆死区"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_CHS "选择在摇杆开始被处理前需要推动多远的距离。"
#define OPTION_VAL_0_CHS NULL
#define OPTION_VAL_5_CHS NULL
#define OPTION_VAL_10_CHS NULL
#define OPTION_VAL_15_CHS NULL
#define OPTION_VAL_20_CHS NULL
#define OPTION_VAL_25_CHS NULL
#define OPTION_VAL_30_CHS NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CHS "扳机死区"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_CHS "选择在开始处理之前，你需要将扳机按下的程度。"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CHS "数字扳机"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_CHS "启用后，扳机会像普通按钮一样工作，即它们将被处理为完全按下或完全未按下，没有中间状态。"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CHS "广播数字输出"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CHS "在 TCP 端口 8000 上广播数字输出和力反馈状态。与 MAME 的 “-output network” 选项兼容。"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CHS "显示光枪设置"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CHS "启用光枪准星显示选项的配置。注意：快速菜单可能需要切换才能使此设置生效。"
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_CHS "光枪准星大小缩放"
#define OPTION_VAL_50_CHS NULL
#define OPTION_VAL_60_CHS NULL
#define OPTION_VAL_70_CHS NULL
#define OPTION_VAL_80_CHS NULL
#define OPTION_VAL_90_CHS NULL
#define OPTION_VAL_100_O47_CHS NULL
#define OPTION_VAL_110_O47_CHS NULL
#define OPTION_VAL_120_O47_CHS NULL
#define OPTION_VAL_130_O47_CHS NULL
#define OPTION_VAL_140_O47_CHS NULL
#define OPTION_VAL_150_O47_CHS NULL
#define OPTION_VAL_160_O47_CHS NULL
#define OPTION_VAL_170_O47_CHS NULL
#define OPTION_VAL_180_O47_CHS NULL
#define OPTION_VAL_190_O47_CHS NULL
#define OPTION_VAL_200_O47_CHS NULL
#define OPTION_VAL_210_O47_CHS NULL
#define OPTION_VAL_220_O47_CHS NULL
#define OPTION_VAL_230_O47_CHS NULL
#define OPTION_VAL_240_O47_CHS NULL
#define OPTION_VAL_250_O47_CHS NULL
#define OPTION_VAL_260_O47_CHS NULL
#define OPTION_VAL_270_O47_CHS NULL
#define OPTION_VAL_280_O47_CHS NULL
#define OPTION_VAL_290_O47_CHS NULL
#define OPTION_VAL_300_O47_CHS NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CHS "光枪 1 准星显示"
#define OPTION_VAL_WHITE_CHS "白色"
#define OPTION_VAL_RED_CHS "红色"
#define OPTION_VAL_GREEN_CHS "绿色"
#define OPTION_VAL_BLUE_CHS "蓝色"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CHS "光枪 2 准星显示"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CHS "光枪 3 准星显示"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CHS "光枪 4 准星显示"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_CHS "A1 扩展插槽设备"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_CHS "选择插入 A1 扩展插槽（端口 A 插槽 1）中的设备。"
#define OPTION_VAL_VMU_CHS NULL
#define OPTION_VAL_PURUPURU_CHS "震动包"
#define OPTION_VAL_DREAMPOTATO_CHS NULL
#define OPTION_VAL_NONE_CHS "无"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_CHS "A2 扩展插槽设备"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_CHS "选择插入 A2 扩展插槽（端口 A 插槽 2）中的设备。"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_CHS "B1 扩展插槽设备"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_CHS "选择插入 B1 扩展插槽（端口 B 插槽 1）中的设备。"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_CHS "B2 扩展插槽设备"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_CHS "选择插入 B2 扩展插槽（端口 B 插槽 2）中的设备。"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_CHS "C1 扩展插槽设备"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_CHS "选择插入 C1 扩展插槽（端口 C 插槽 1）中的设备。"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_CHS "C2 扩展插槽设备"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_CHS "选择插入 C2 扩展插槽（端口 C 插槽 2）中的设备。"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_CHS "D1 扩展插槽设备"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_CHS "选择插入 D1 扩展插槽（端口 D 插槽 1）中的设备。"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_CHS "D2 扩展插槽设备"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_CHS "选择插入 D2 扩展插槽（端口 D 插槽 2）中的设备。"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CHS "按游戏区分的视觉内存单元/系统（VMU）"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CHS "按游戏区分的 VMU"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CHS "当禁用时，所有游戏共享最多 8 个 VMU 存档文件（A1/A2/B1/B2/C1/C2/D1/D2），这些文件位于 RetroArch 的系统文件夹中。\n“VMU A1” 设置会在 RetroArch 的存档文件夹中为每个启动的游戏创建一个唯一的 VMU 'A1' 文件。\n“所有 VMU” 设置会为每个启动的游戏创建最多 8 个唯一的 VMU 文件（A1/A2/B1/B2/C1/C2/D1/D2）。"
#define OPTION_VAL_VMU_A1_CHS NULL
#define OPTION_VAL_ALL_VMUS_CHS "所有 VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CHS "视觉内存单元/系统（VMU）声音"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_CHS "VMU 声音"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_CHS "启用时，会播放 VMU 的“哔”声。"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_CHS NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_CHS NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CHS "显示视觉内存单元/系统（VMU）屏幕设置"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CHS "显示 VMU 屏幕设置"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CHS "启用对模拟 VMU LCD 屏幕可见性、大小、位置和颜色的配置。注意：快速菜单可能需要切换才能使此设置生效。"
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CHS "VMU 屏幕 1 显示"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CHS "VMU 屏幕 1 位置"
#define OPTION_VAL_UPPER_LEFT_CHS "左上角"
#define OPTION_VAL_UPPER_RIGHT_CHS "右上角"
#define OPTION_VAL_LOWER_LEFT_CHS "左下角"
#define OPTION_VAL_LOWER_RIGHT_CHS "右下角"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CHS "VMU 屏幕 1 大小"
#define OPTION_VAL_1X_CHS "1倍"
#define OPTION_VAL_3X_CHS NULL
#define OPTION_VAL_5X_CHS "5倍"
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CHS "VMU 屏幕 1 像素点亮颜色"
#define OPTION_VAL_DEFAULT_ON_00_CHS "默认开启"
#define OPTION_VAL_DEFAULT_OFF_01_CHS "默认关闭"
#define OPTION_VAL_BLACK_02_CHS "黑色"
#define OPTION_VAL_LIGHT_BLUE_04_CHS "浅蓝色"
#define OPTION_VAL_CYAN_06_CHS "青色"
#define OPTION_VAL_CYAN_BLUE_07_CHS "青蓝色"
#define OPTION_VAL_LIGHT_GREEN_08_CHS "浅绿色"
#define OPTION_VAL_CYAN_GREEN_09_CHS "青绿色"
#define OPTION_VAL_LIGHT_CYAN_10_CHS "浅青色"
#define OPTION_VAL_PURPLE_12_CHS "紫色"
#define OPTION_VAL_LIGHT_PURPLE_13_CHS "紫色"
#define OPTION_VAL_YELLOW_14_CHS "黄色"
#define OPTION_VAL_GRAY_15_CHS "灰色"
#define OPTION_VAL_LIGHT_PURPLE_2_16_CHS "浅紫色（2）"
#define OPTION_VAL_LIGHT_GREEN_2_17_CHS "浅绿色（2）"
#define OPTION_VAL_LIGHT_GREEN_3_18_CHS "浅绿色（3）"
#define OPTION_VAL_LIGHT_CYAN_2_19_CHS "浅青色（2）"
#define OPTION_VAL_LIGHT_RED_2_20_CHS "浅红色（2）"
#define OPTION_VAL_MAGENTA_21_CHS "品红色"
#define OPTION_VAL_LIGHT_PURPLE_3_22_CHS "浅紫色（3）"
#define OPTION_VAL_LIGHT_ORANGE_23_CHS "浅橙色"
#define OPTION_VAL_ORANGE_24_CHS "橙色"
#define OPTION_VAL_LIGHT_PURPLE_4_25_CHS "浅紫色（4）"
#define OPTION_VAL_LIGHT_YELLOW_26_CHS "浅黄色"
#define OPTION_VAL_LIGHT_YELLOW_2_27_CHS "浅黄色（2）"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CHS "VMU 屏幕 1 像素关闭颜色"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CHS "VMU 屏幕 1 不透明度"
#define OPTION_VAL_40_CHS NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CHS "VMU 屏幕 2 显示"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CHS "VMU 屏幕 2 位置"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CHS "VMU 屏幕 2 尺寸"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CHS "VMU 屏幕 2 像素开启颜色"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CHS "VMU 屏幕 2 像素关闭颜色"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CHS "VMU 屏幕 2 不透明度"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CHS "VMU 屏幕 3 显示"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CHS "VMU 屏幕 3 位置"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CHS "VMU 屏幕 3 尺寸"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CHS "VMU 屏幕 3 像素开启颜色"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CHS "VMU 屏幕 3 像素关闭颜色"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CHS "VMU 屏幕 3 不透明度"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CHS "VMU 屏幕 4 显示"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CHS "VMU 屏幕 4 位置"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CHS "VMU 屏幕 4 尺寸"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CHS "VMU 屏幕 4 像素开启颜色"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CHS "VMU 屏幕 4 像素关闭颜色"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CHS "VMU 屏幕 4 不透明度"

struct retro_core_option_v2_category option_cats_chs[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CHS,
      CATEGORY_SYSTEM_INFO_0_CHS
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CHS,
      CATEGORY_VIDEO_INFO_0_CHS
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_CHS,
      CATEGORY_PERFORMANCE_INFO_0_CHS
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CHS,
      CATEGORY_HACKS_INFO_0_CHS
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CHS,
      CATEGORY_INPUT_INFO_0_CHS
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_CHS,
      CATEGORY_EXPANSIONS_INFO_0_CHS
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_CHS,
      CATEGORY_VMU_INFO_0_CHS
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_chs[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_CHS,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_CHS },
         { "USA",     OPTION_VAL_USA_CHS },
         { "Europe",  OPTION_VAL_EUROPE_CHS },
         { "Default", OPTION_VAL_DEFAULT_CHS },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_CHS,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_CHS },
         { "English",  OPTION_VAL_ENGLISH_CHS },
         { "German",   OPTION_VAL_GERMAN_CHS },
         { "French",   OPTION_VAL_FRENCH_CHS },
         { "Spanish",  OPTION_VAL_SPANISH_CHS },
         { "Italian",  OPTION_VAL_ITALIAN_CHS },
         { "Default",  OPTION_VAL_DEFAULT_CHS },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CHS,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_CHS },
         { "640x480",    OPTION_VAL_640X480_CHS },
         { "800x600",    OPTION_VAL_800X600_CHS },
         { "960x720",    OPTION_VAL_960X720_CHS },
         { "1024x768",   OPTION_VAL_1024X768_CHS },
         { "1280x960",   OPTION_VAL_1280X960_CHS },
         { "1440x1080",  OPTION_VAL_1440X1080_CHS },
         { "1600x1200",  OPTION_VAL_1600X1200_CHS },
         { "1920x1440",  OPTION_VAL_1920X1440_CHS },
         { "2560x1920",  OPTION_VAL_2560X1920_CHS },
         { "2880x2160",  OPTION_VAL_2880X2160_CHS },
         { "3200x2400",  OPTION_VAL_3200X2400_CHS },
         { "3840x2880",  OPTION_VAL_3840X2880_CHS },
         { "4480x3360",  OPTION_VAL_4480X3360_CHS },
         { "5120x3840",  OPTION_VAL_5120X3840_CHS },
         { "5760x4320",  OPTION_VAL_5760X4320_CHS },
         { "6400x4800",  OPTION_VAL_6400X4800_CHS },
         { "7040x5280",  OPTION_VAL_7040X5280_CHS },
         { "7680x5760",  OPTION_VAL_7680X5760_CHS },
         { "8320x6240",  OPTION_VAL_8320X6240_CHS },
         { "8960x6720",  OPTION_VAL_8960X6720_CHS },
         { "9600x7200",  OPTION_VAL_9600X7200_CHS },
         { "10240x7680", OPTION_VAL_10240X7680_CHS },
         { "10880x8160", OPTION_VAL_10880X8160_CHS },
         { "11520x8640", OPTION_VAL_11520X8640_CHS },
         { "12160x9120", OPTION_VAL_12160X9120_CHS },
         { "12800x9600", OPTION_VAL_12800X9600_CHS },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_CHS },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_CHS },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_CHS },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_CHS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_CHS },
         { "PAL",     OPTION_VAL_PAL_CHS },
         { "PAL_N",   OPTION_VAL_PAL_N_CHS },
         { "PAL_M",   OPTION_VAL_PAL_M_CHS },
         { "Default", OPTION_VAL_DEFAULT_CHS },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CHS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_CHS },
         { "vertical",   OPTION_VAL_VERTICAL_CHS },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CHS },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_CHS },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_CHS },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_CHS },
         { "1GB",   OPTION_VAL_1GB_CHS },
         { "2GB",   OPTION_VAL_2GB_CHS },
         { "4GB",   OPTION_VAL_4GB_CHS },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CHS,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_CHS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_CHS },
         { "1",  OPTION_VAL_1_CHS },
         { "2",  OPTION_VAL_2_CHS },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_CHS },
         { "4",  OPTION_VAL_4_CHS },
         { "6",  OPTION_VAL_6_CHS },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CHS,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CHS,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_CHS,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_CHS,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CHS,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CHS,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_CHS },
         { "more", OPTION_VAL_MORE_CHS },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CHS,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_CHS },
         { "110", OPTION_VAL_110_CHS },
         { "120", OPTION_VAL_120_CHS },
         { "130", OPTION_VAL_130_CHS },
         { "140", OPTION_VAL_140_CHS },
         { "150", OPTION_VAL_150_CHS },
         { "160", OPTION_VAL_160_CHS },
         { "170", OPTION_VAL_170_CHS },
         { "180", OPTION_VAL_180_CHS },
         { "190", OPTION_VAL_190_CHS },
         { "200", OPTION_VAL_200_CHS },
         { "210", OPTION_VAL_210_CHS },
         { "220", OPTION_VAL_220_CHS },
         { "230", OPTION_VAL_230_CHS },
         { "240", OPTION_VAL_240_CHS },
         { "250", OPTION_VAL_250_CHS },
         { "260", OPTION_VAL_260_CHS },
         { "270", OPTION_VAL_270_CHS },
         { "280", OPTION_VAL_280_CHS },
         { "290", OPTION_VAL_290_CHS },
         { "300", OPTION_VAL_300_CHS },
         { "310", OPTION_VAL_310_CHS },
         { "320", OPTION_VAL_320_CHS },
         { "330", OPTION_VAL_330_CHS },
         { "340", OPTION_VAL_340_CHS },
         { "350", OPTION_VAL_350_CHS },
         { "360", OPTION_VAL_360_CHS },
         { "370", OPTION_VAL_370_CHS },
         { "380", OPTION_VAL_380_CHS },
         { "390", OPTION_VAL_390_CHS },
         { "400", OPTION_VAL_400_CHS },
         { "410", OPTION_VAL_410_CHS },
         { "420", OPTION_VAL_420_CHS },
         { "430", OPTION_VAL_430_CHS },
         { "440", OPTION_VAL_440_CHS },
         { "450", OPTION_VAL_450_CHS },
         { "460", OPTION_VAL_460_CHS },
         { "470", OPTION_VAL_470_CHS },
         { "480", OPTION_VAL_480_CHS },
         { "490", OPTION_VAL_490_CHS },
         { "500", OPTION_VAL_500_CHS },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_CHS,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CHS },
         { "5%",  OPTION_VAL_5_CHS },
         { "10%", OPTION_VAL_10_CHS },
         { "15%", OPTION_VAL_15_CHS },
         { "20%", OPTION_VAL_20_CHS },
         { "25%", OPTION_VAL_25_CHS },
         { "30%", OPTION_VAL_30_CHS },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_CHS,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CHS },
         { "5%",  OPTION_VAL_5_CHS },
         { "10%", OPTION_VAL_10_CHS },
         { "15%", OPTION_VAL_15_CHS },
         { "20%", OPTION_VAL_20_CHS },
         { "25%", OPTION_VAL_25_CHS },
         { "30%", OPTION_VAL_30_CHS },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_CHS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CHS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CHS,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_CHS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_CHS },
         { "60%",  OPTION_VAL_60_CHS },
         { "70%",  OPTION_VAL_70_CHS },
         { "80%",  OPTION_VAL_80_CHS },
         { "90%",  OPTION_VAL_90_CHS },
         { "100%", OPTION_VAL_100_O47_CHS },
         { "110%", OPTION_VAL_110_O47_CHS },
         { "120%", OPTION_VAL_120_O47_CHS },
         { "130%", OPTION_VAL_130_O47_CHS },
         { "140%", OPTION_VAL_140_O47_CHS },
         { "150%", OPTION_VAL_150_O47_CHS },
         { "160%", OPTION_VAL_160_O47_CHS },
         { "170%", OPTION_VAL_170_O47_CHS },
         { "180%", OPTION_VAL_180_O47_CHS },
         { "190%", OPTION_VAL_190_O47_CHS },
         { "200%", OPTION_VAL_200_O47_CHS },
         { "210%", OPTION_VAL_210_O47_CHS },
         { "220%", OPTION_VAL_220_O47_CHS },
         { "230%", OPTION_VAL_230_O47_CHS },
         { "240%", OPTION_VAL_240_O47_CHS },
         { "250%", OPTION_VAL_250_O47_CHS },
         { "260%", OPTION_VAL_260_O47_CHS },
         { "270%", OPTION_VAL_270_O47_CHS },
         { "280%", OPTION_VAL_280_O47_CHS },
         { "290%", OPTION_VAL_290_O47_CHS },
         { "300%", OPTION_VAL_300_O47_CHS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHS },
         { "Red",      OPTION_VAL_RED_CHS },
         { "Green",    OPTION_VAL_GREEN_CHS },
         { "Blue",     OPTION_VAL_BLUE_CHS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHS },
         { "Red",      OPTION_VAL_RED_CHS },
         { "Green",    OPTION_VAL_GREEN_CHS },
         { "Blue",     OPTION_VAL_BLUE_CHS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHS },
         { "Red",      OPTION_VAL_RED_CHS },
         { "Green",    OPTION_VAL_GREEN_CHS },
         { "Blue",     OPTION_VAL_BLUE_CHS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHS },
         { "Red",      OPTION_VAL_RED_CHS },
         { "Green",    OPTION_VAL_GREEN_CHS },
         { "Blue",     OPTION_VAL_BLUE_CHS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_CHS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHS },
         { "Purupuru", OPTION_VAL_PURUPURU_CHS },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CHS },
         { "None",     OPTION_VAL_NONE_CHS },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_CHS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHS },
         { "Purupuru", OPTION_VAL_PURUPURU_CHS },
         { "None",     OPTION_VAL_NONE_CHS },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_CHS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHS },
         { "Purupuru", OPTION_VAL_PURUPURU_CHS },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CHS },
         { "None",     OPTION_VAL_NONE_CHS },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_CHS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHS },
         { "Purupuru", OPTION_VAL_PURUPURU_CHS },
         { "None",     OPTION_VAL_NONE_CHS },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_CHS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHS },
         { "Purupuru", OPTION_VAL_PURUPURU_CHS },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CHS },
         { "None",     OPTION_VAL_NONE_CHS },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_CHS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHS },
         { "Purupuru", OPTION_VAL_PURUPURU_CHS },
         { "None",     OPTION_VAL_NONE_CHS },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_CHS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHS },
         { "Purupuru", OPTION_VAL_PURUPURU_CHS },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CHS },
         { "None",     OPTION_VAL_NONE_CHS },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_CHS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHS },
         { "Purupuru", OPTION_VAL_PURUPURU_CHS },
         { "None",     OPTION_VAL_NONE_CHS },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CHS,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CHS,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CHS,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_CHS },
         { "All VMUs", OPTION_VAL_ALL_VMUS_CHS },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CHS,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_CHS,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_CHS,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_CHS,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_CHS,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CHS,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CHS,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CHS,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHS },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHS },
         { "2x", OPTION_VAL_2_O26_CHS },
         { "3x", OPTION_VAL_3X_CHS },
         { "4x", OPTION_VAL_4_CHS },
         { "5x", OPTION_VAL_5X_CHS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHS },
         { "20%",  OPTION_VAL_20_CHS },
         { "30%",  OPTION_VAL_30_CHS },
         { "40%",  OPTION_VAL_40_CHS },
         { "50%",  OPTION_VAL_50_CHS },
         { "60%",  OPTION_VAL_60_CHS },
         { "70%",  OPTION_VAL_70_CHS },
         { "80%",  OPTION_VAL_80_CHS },
         { "90%",  OPTION_VAL_90_CHS },
         { "100%", OPTION_VAL_100_O47_CHS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHS },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHS },
         { "2x", OPTION_VAL_2_O26_CHS },
         { "3x", OPTION_VAL_3X_CHS },
         { "4x", OPTION_VAL_4_CHS },
         { "5x", OPTION_VAL_5X_CHS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHS },
         { "20%",  OPTION_VAL_20_CHS },
         { "30%",  OPTION_VAL_30_CHS },
         { "40%",  OPTION_VAL_40_CHS },
         { "50%",  OPTION_VAL_50_CHS },
         { "60%",  OPTION_VAL_60_CHS },
         { "70%",  OPTION_VAL_70_CHS },
         { "80%",  OPTION_VAL_80_CHS },
         { "90%",  OPTION_VAL_90_CHS },
         { "100%", OPTION_VAL_100_O47_CHS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHS },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHS },
         { "2x", OPTION_VAL_2_O26_CHS },
         { "3x", OPTION_VAL_3X_CHS },
         { "4x", OPTION_VAL_4_CHS },
         { "5x", OPTION_VAL_5X_CHS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHS },
         { "20%",  OPTION_VAL_20_CHS },
         { "30%",  OPTION_VAL_30_CHS },
         { "40%",  OPTION_VAL_40_CHS },
         { "50%",  OPTION_VAL_50_CHS },
         { "60%",  OPTION_VAL_60_CHS },
         { "70%",  OPTION_VAL_70_CHS },
         { "80%",  OPTION_VAL_80_CHS },
         { "90%",  OPTION_VAL_90_CHS },
         { "100%", OPTION_VAL_100_O47_CHS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHS },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHS },
         { "2x", OPTION_VAL_2_O26_CHS },
         { "3x", OPTION_VAL_3X_CHS },
         { "4x", OPTION_VAL_4_CHS },
         { "5x", OPTION_VAL_5X_CHS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHS },
         { "BLUE 03",           OPTION_VAL_BLUE_CHS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHS },
         { "GREEN 05",          OPTION_VAL_GREEN_CHS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHS },
         { "RED 11",            OPTION_VAL_RED_CHS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHS },
         { "WHITE 28",          OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CHS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHS },
         { "20%",  OPTION_VAL_20_CHS },
         { "30%",  OPTION_VAL_30_CHS },
         { "40%",  OPTION_VAL_40_CHS },
         { "50%",  OPTION_VAL_50_CHS },
         { "60%",  OPTION_VAL_60_CHS },
         { "70%",  OPTION_VAL_70_CHS },
         { "80%",  OPTION_VAL_80_CHS },
         { "90%",  OPTION_VAL_90_CHS },
         { "100%", OPTION_VAL_100_O47_CHS },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_chs = {
   option_cats_chs,
   option_defs_chs
};

/* RETRO_LANGUAGE_CHT */

#define CATEGORY_SYSTEM_LABEL_CHT "系統"
#define CATEGORY_SYSTEM_INFO_0_CHT NULL
#define CATEGORY_VIDEO_LABEL_CHT "顯示"
#define CATEGORY_VIDEO_INFO_0_CHT "變更解析度、渲染透明度和視覺效果的設定。"
#define CATEGORY_PERFORMANCE_LABEL_CHT "效能"
#define CATEGORY_PERFORMANCE_INFO_0_CHT NULL
#define CATEGORY_HACKS_LABEL_CHT "進階"
#define CATEGORY_HACKS_INFO_0_CHT NULL
#define CATEGORY_INPUT_LABEL_CHT "輸入"
#define CATEGORY_INPUT_INFO_0_CHT NULL
#define CATEGORY_EXPANSIONS_LABEL_CHT NULL
#define CATEGORY_EXPANSIONS_INFO_0_CHT NULL
#define CATEGORY_VMU_LABEL_CHT NULL
#define CATEGORY_VMU_INFO_0_CHT NULL
#define CORE_OPTION_NAME_REGION_LABEL_CHT "發售地"
#define OPTION_VAL_JAPAN_CHT NULL
#define OPTION_VAL_USA_CHT NULL
#define OPTION_VAL_EUROPE_CHT NULL
#define OPTION_VAL_DEFAULT_CHT "預設"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_CHT "語言"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_CHT NULL
#define OPTION_VAL_JAPANESE_CHT "日語"
#define OPTION_VAL_ENGLISH_CHT "英語"
#define OPTION_VAL_GERMAN_CHT "德語"
#define OPTION_VAL_FRENCH_CHT "法語"
#define OPTION_VAL_SPANISH_CHT "西班牙語"
#define OPTION_VAL_ITALIAN_CHT "意大利語"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_CHT NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_CHT NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CHT NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CHT NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CHT NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CHT NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_CHT NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CHT NULL
#define CORE_OPTION_NAME_UPNP_LABEL_CHT NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DCNET_LABEL_CHT NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_CHT NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CHT "內部解析度"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CHT NULL
#define OPTION_VAL_320X240_CHT NULL
#define OPTION_VAL_640X480_CHT NULL
#define OPTION_VAL_800X600_CHT NULL
#define OPTION_VAL_960X720_CHT NULL
#define OPTION_VAL_1024X768_CHT NULL
#define OPTION_VAL_1280X960_CHT NULL
#define OPTION_VAL_1440X1080_CHT NULL
#define OPTION_VAL_1600X1200_CHT NULL
#define OPTION_VAL_1920X1440_CHT NULL
#define OPTION_VAL_2560X1920_CHT NULL
#define OPTION_VAL_2880X2160_CHT NULL
#define OPTION_VAL_3200X2400_CHT NULL
#define OPTION_VAL_3840X2880_CHT NULL
#define OPTION_VAL_4480X3360_CHT NULL
#define OPTION_VAL_5120X3840_CHT NULL
#define OPTION_VAL_5760X4320_CHT NULL
#define OPTION_VAL_6400X4800_CHT NULL
#define OPTION_VAL_7040X5280_CHT NULL
#define OPTION_VAL_7680X5760_CHT NULL
#define OPTION_VAL_8320X6240_CHT NULL
#define OPTION_VAL_8960X6720_CHT NULL
#define OPTION_VAL_9600X7200_CHT NULL
#define OPTION_VAL_10240X7680_CHT NULL
#define OPTION_VAL_10880X8160_CHT NULL
#define OPTION_VAL_11520X8640_CHT NULL
#define OPTION_VAL_12160X9120_CHT NULL
#define OPTION_VAL_12800X9600_CHT NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_CHT NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CHT NULL
#define OPTION_VAL_VGA_CHT NULL
#define OPTION_VAL_TV_RGB_CHT NULL
#define OPTION_VAL_TV_COMPOSITE_CHT NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_CHT NULL
#define OPTION_VAL_NTSC_CHT NULL
#define OPTION_VAL_PAL_CHT NULL
#define OPTION_VAL_PAL_N_CHT NULL
#define OPTION_VAL_PAL_M_CHT NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CHT "螢幕顯示方向"
#define OPTION_VAL_HORIZONTAL_CHT NULL
#define OPTION_VAL_VERTICAL_CHT NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CHT NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_CHT NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CHT NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_CHT NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_CHT NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CHT NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CHT NULL
#define OPTION_VAL_512MB_CHT NULL
#define OPTION_VAL_1GB_CHT NULL
#define OPTION_VAL_2GB_CHT NULL
#define OPTION_VAL_4GB_CHT NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_CHT NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CHT NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CHT NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CHT NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CHT NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_CHT NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_FOG_LABEL_CHT NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CHT NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CHT NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CHT NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CHT "材質過濾"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CHT NULL
#define OPTION_VAL_1_CHT NULL
#define OPTION_VAL_2_CHT NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CHT NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CHT NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CHT NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CHT NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_CHT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CHT NULL
#define OPTION_VAL_2_O26_CHT "2倍"
#define OPTION_VAL_4_CHT "4倍"
#define OPTION_VAL_6_CHT "6倍"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CHT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_CHT NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CHT NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CHT NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_CHT NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_CHT NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CHT NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CHT NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CHT NULL
#define OPTION_VAL_SOME_CHT "正常"
#define OPTION_VAL_MORE_CHT NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CHT NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CHT NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CHT NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CHT NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CHT NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_CHT NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_CHT NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_CHT NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_CHT NULL
#define OPTION_VAL_100_CHT NULL
#define OPTION_VAL_110_CHT NULL
#define OPTION_VAL_120_CHT NULL
#define OPTION_VAL_130_CHT NULL
#define OPTION_VAL_140_CHT NULL
#define OPTION_VAL_150_CHT NULL
#define OPTION_VAL_160_CHT NULL
#define OPTION_VAL_170_CHT NULL
#define OPTION_VAL_180_CHT NULL
#define OPTION_VAL_190_CHT NULL
#define OPTION_VAL_200_CHT NULL
#define OPTION_VAL_210_CHT NULL
#define OPTION_VAL_220_CHT NULL
#define OPTION_VAL_230_CHT NULL
#define OPTION_VAL_240_CHT NULL
#define OPTION_VAL_250_CHT NULL
#define OPTION_VAL_260_CHT NULL
#define OPTION_VAL_270_CHT NULL
#define OPTION_VAL_280_CHT NULL
#define OPTION_VAL_290_CHT NULL
#define OPTION_VAL_300_CHT NULL
#define OPTION_VAL_310_CHT NULL
#define OPTION_VAL_320_CHT NULL
#define OPTION_VAL_330_CHT NULL
#define OPTION_VAL_340_CHT NULL
#define OPTION_VAL_350_CHT NULL
#define OPTION_VAL_360_CHT NULL
#define OPTION_VAL_370_CHT NULL
#define OPTION_VAL_380_CHT NULL
#define OPTION_VAL_390_CHT NULL
#define OPTION_VAL_400_CHT NULL
#define OPTION_VAL_410_CHT NULL
#define OPTION_VAL_420_CHT NULL
#define OPTION_VAL_430_CHT NULL
#define OPTION_VAL_440_CHT NULL
#define OPTION_VAL_450_CHT NULL
#define OPTION_VAL_460_CHT NULL
#define OPTION_VAL_470_CHT NULL
#define OPTION_VAL_480_CHT NULL
#define OPTION_VAL_490_CHT NULL
#define OPTION_VAL_500_CHT NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CHT NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_CHT NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_CHT NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CHT NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_CHT NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_CHT NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CHT NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_CHT NULL
#define OPTION_VAL_0_CHT NULL
#define OPTION_VAL_5_CHT NULL
#define OPTION_VAL_10_CHT NULL
#define OPTION_VAL_15_CHT NULL
#define OPTION_VAL_20_CHT NULL
#define OPTION_VAL_25_CHT NULL
#define OPTION_VAL_30_CHT NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CHT NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CHT NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CHT NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CHT NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CHT NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_CHT NULL
#define OPTION_VAL_50_CHT NULL
#define OPTION_VAL_60_CHT NULL
#define OPTION_VAL_70_CHT NULL
#define OPTION_VAL_80_CHT NULL
#define OPTION_VAL_90_CHT NULL
#define OPTION_VAL_100_O47_CHT NULL
#define OPTION_VAL_110_O47_CHT NULL
#define OPTION_VAL_120_O47_CHT NULL
#define OPTION_VAL_130_O47_CHT NULL
#define OPTION_VAL_140_O47_CHT NULL
#define OPTION_VAL_150_O47_CHT NULL
#define OPTION_VAL_160_O47_CHT NULL
#define OPTION_VAL_170_O47_CHT NULL
#define OPTION_VAL_180_O47_CHT NULL
#define OPTION_VAL_190_O47_CHT NULL
#define OPTION_VAL_200_O47_CHT NULL
#define OPTION_VAL_210_O47_CHT NULL
#define OPTION_VAL_220_O47_CHT NULL
#define OPTION_VAL_230_O47_CHT NULL
#define OPTION_VAL_240_O47_CHT NULL
#define OPTION_VAL_250_O47_CHT NULL
#define OPTION_VAL_260_O47_CHT NULL
#define OPTION_VAL_270_O47_CHT NULL
#define OPTION_VAL_280_O47_CHT NULL
#define OPTION_VAL_290_O47_CHT NULL
#define OPTION_VAL_300_O47_CHT NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CHT NULL
#define OPTION_VAL_WHITE_CHT "白色"
#define OPTION_VAL_RED_CHT "紅色"
#define OPTION_VAL_GREEN_CHT "綠色"
#define OPTION_VAL_BLUE_CHT "藍色"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CHT NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CHT NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_CHT NULL
#define OPTION_VAL_VMU_CHT NULL
#define OPTION_VAL_PURUPURU_CHT NULL
#define OPTION_VAL_DREAMPOTATO_CHT NULL
#define OPTION_VAL_NONE_CHT "無"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_CHT NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_CHT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CHT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CHT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CHT NULL
#define OPTION_VAL_VMU_A1_CHT NULL
#define OPTION_VAL_ALL_VMUS_CHT NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_CHT NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_CHT NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_CHT NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_CHT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CHT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CHT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CHT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CHT NULL
#define OPTION_VAL_UPPER_LEFT_CHT NULL
#define OPTION_VAL_UPPER_RIGHT_CHT NULL
#define OPTION_VAL_LOWER_LEFT_CHT NULL
#define OPTION_VAL_LOWER_RIGHT_CHT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CHT NULL
#define OPTION_VAL_1X_CHT "1倍"
#define OPTION_VAL_3X_CHT "3倍"
#define OPTION_VAL_5X_CHT "5倍"
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CHT NULL
#define OPTION_VAL_DEFAULT_ON_00_CHT NULL
#define OPTION_VAL_DEFAULT_OFF_01_CHT NULL
#define OPTION_VAL_BLACK_02_CHT "黑色"
#define OPTION_VAL_LIGHT_BLUE_04_CHT "淺藍色"
#define OPTION_VAL_CYAN_06_CHT "青色"
#define OPTION_VAL_CYAN_BLUE_07_CHT "青藍色"
#define OPTION_VAL_LIGHT_GREEN_08_CHT "淺綠色"
#define OPTION_VAL_CYAN_GREEN_09_CHT "青綠色"
#define OPTION_VAL_LIGHT_CYAN_10_CHT "淺青色"
#define OPTION_VAL_PURPLE_12_CHT "紫色"
#define OPTION_VAL_LIGHT_PURPLE_13_CHT "淺紫色"
#define OPTION_VAL_YELLOW_14_CHT "黃色"
#define OPTION_VAL_GRAY_15_CHT "灰色"
#define OPTION_VAL_LIGHT_PURPLE_2_16_CHT "淺紫色 (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_CHT "淺綠色 (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_CHT "淺綠色 (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_CHT "淺青色 (2)"
#define OPTION_VAL_LIGHT_RED_2_20_CHT "淡紅色 (2)"
#define OPTION_VAL_MAGENTA_21_CHT "洋紅色"
#define OPTION_VAL_LIGHT_PURPLE_3_22_CHT "淺紫色 (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_CHT "淺橘色"
#define OPTION_VAL_ORANGE_24_CHT "橙色"
#define OPTION_VAL_LIGHT_PURPLE_4_25_CHT "淺紫色 (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_CHT "淡黃色"
#define OPTION_VAL_LIGHT_YELLOW_2_27_CHT "淡黃色 (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CHT NULL
#define OPTION_VAL_40_CHT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CHT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CHT NULL

struct retro_core_option_v2_category option_cats_cht[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CHT,
      CATEGORY_SYSTEM_INFO_0_CHT
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CHT,
      CATEGORY_VIDEO_INFO_0_CHT
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_CHT,
      CATEGORY_PERFORMANCE_INFO_0_CHT
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CHT,
      CATEGORY_HACKS_INFO_0_CHT
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CHT,
      CATEGORY_INPUT_INFO_0_CHT
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_CHT,
      CATEGORY_EXPANSIONS_INFO_0_CHT
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_CHT,
      CATEGORY_VMU_INFO_0_CHT
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_cht[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_CHT,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_CHT },
         { "USA",     OPTION_VAL_USA_CHT },
         { "Europe",  OPTION_VAL_EUROPE_CHT },
         { "Default", OPTION_VAL_DEFAULT_CHT },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_CHT,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_CHT },
         { "English",  OPTION_VAL_ENGLISH_CHT },
         { "German",   OPTION_VAL_GERMAN_CHT },
         { "French",   OPTION_VAL_FRENCH_CHT },
         { "Spanish",  OPTION_VAL_SPANISH_CHT },
         { "Italian",  OPTION_VAL_ITALIAN_CHT },
         { "Default",  OPTION_VAL_DEFAULT_CHT },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CHT,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_CHT },
         { "640x480",    OPTION_VAL_640X480_CHT },
         { "800x600",    OPTION_VAL_800X600_CHT },
         { "960x720",    OPTION_VAL_960X720_CHT },
         { "1024x768",   OPTION_VAL_1024X768_CHT },
         { "1280x960",   OPTION_VAL_1280X960_CHT },
         { "1440x1080",  OPTION_VAL_1440X1080_CHT },
         { "1600x1200",  OPTION_VAL_1600X1200_CHT },
         { "1920x1440",  OPTION_VAL_1920X1440_CHT },
         { "2560x1920",  OPTION_VAL_2560X1920_CHT },
         { "2880x2160",  OPTION_VAL_2880X2160_CHT },
         { "3200x2400",  OPTION_VAL_3200X2400_CHT },
         { "3840x2880",  OPTION_VAL_3840X2880_CHT },
         { "4480x3360",  OPTION_VAL_4480X3360_CHT },
         { "5120x3840",  OPTION_VAL_5120X3840_CHT },
         { "5760x4320",  OPTION_VAL_5760X4320_CHT },
         { "6400x4800",  OPTION_VAL_6400X4800_CHT },
         { "7040x5280",  OPTION_VAL_7040X5280_CHT },
         { "7680x5760",  OPTION_VAL_7680X5760_CHT },
         { "8320x6240",  OPTION_VAL_8320X6240_CHT },
         { "8960x6720",  OPTION_VAL_8960X6720_CHT },
         { "9600x7200",  OPTION_VAL_9600X7200_CHT },
         { "10240x7680", OPTION_VAL_10240X7680_CHT },
         { "10880x8160", OPTION_VAL_10880X8160_CHT },
         { "11520x8640", OPTION_VAL_11520X8640_CHT },
         { "12160x9120", OPTION_VAL_12160X9120_CHT },
         { "12800x9600", OPTION_VAL_12800X9600_CHT },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_CHT },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_CHT },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_CHT },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_CHT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_CHT },
         { "PAL",     OPTION_VAL_PAL_CHT },
         { "PAL_N",   OPTION_VAL_PAL_N_CHT },
         { "PAL_M",   OPTION_VAL_PAL_M_CHT },
         { "Default", OPTION_VAL_DEFAULT_CHT },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CHT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_CHT },
         { "vertical",   OPTION_VAL_VERTICAL_CHT },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CHT },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_CHT },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_CHT },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_CHT },
         { "1GB",   OPTION_VAL_1GB_CHT },
         { "2GB",   OPTION_VAL_2GB_CHT },
         { "4GB",   OPTION_VAL_4GB_CHT },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CHT,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_CHT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_CHT },
         { "1",  OPTION_VAL_1_CHT },
         { "2",  OPTION_VAL_2_CHT },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_CHT },
         { "4",  OPTION_VAL_4_CHT },
         { "6",  OPTION_VAL_6_CHT },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CHT,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CHT,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_CHT,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_CHT,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CHT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CHT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_CHT },
         { "more", OPTION_VAL_MORE_CHT },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CHT,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_CHT },
         { "110", OPTION_VAL_110_CHT },
         { "120", OPTION_VAL_120_CHT },
         { "130", OPTION_VAL_130_CHT },
         { "140", OPTION_VAL_140_CHT },
         { "150", OPTION_VAL_150_CHT },
         { "160", OPTION_VAL_160_CHT },
         { "170", OPTION_VAL_170_CHT },
         { "180", OPTION_VAL_180_CHT },
         { "190", OPTION_VAL_190_CHT },
         { "200", OPTION_VAL_200_CHT },
         { "210", OPTION_VAL_210_CHT },
         { "220", OPTION_VAL_220_CHT },
         { "230", OPTION_VAL_230_CHT },
         { "240", OPTION_VAL_240_CHT },
         { "250", OPTION_VAL_250_CHT },
         { "260", OPTION_VAL_260_CHT },
         { "270", OPTION_VAL_270_CHT },
         { "280", OPTION_VAL_280_CHT },
         { "290", OPTION_VAL_290_CHT },
         { "300", OPTION_VAL_300_CHT },
         { "310", OPTION_VAL_310_CHT },
         { "320", OPTION_VAL_320_CHT },
         { "330", OPTION_VAL_330_CHT },
         { "340", OPTION_VAL_340_CHT },
         { "350", OPTION_VAL_350_CHT },
         { "360", OPTION_VAL_360_CHT },
         { "370", OPTION_VAL_370_CHT },
         { "380", OPTION_VAL_380_CHT },
         { "390", OPTION_VAL_390_CHT },
         { "400", OPTION_VAL_400_CHT },
         { "410", OPTION_VAL_410_CHT },
         { "420", OPTION_VAL_420_CHT },
         { "430", OPTION_VAL_430_CHT },
         { "440", OPTION_VAL_440_CHT },
         { "450", OPTION_VAL_450_CHT },
         { "460", OPTION_VAL_460_CHT },
         { "470", OPTION_VAL_470_CHT },
         { "480", OPTION_VAL_480_CHT },
         { "490", OPTION_VAL_490_CHT },
         { "500", OPTION_VAL_500_CHT },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_CHT,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CHT },
         { "5%",  OPTION_VAL_5_CHT },
         { "10%", OPTION_VAL_10_CHT },
         { "15%", OPTION_VAL_15_CHT },
         { "20%", OPTION_VAL_20_CHT },
         { "25%", OPTION_VAL_25_CHT },
         { "30%", OPTION_VAL_30_CHT },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_CHT,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CHT },
         { "5%",  OPTION_VAL_5_CHT },
         { "10%", OPTION_VAL_10_CHT },
         { "15%", OPTION_VAL_15_CHT },
         { "20%", OPTION_VAL_20_CHT },
         { "25%", OPTION_VAL_25_CHT },
         { "30%", OPTION_VAL_30_CHT },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_CHT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CHT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CHT,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_CHT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_CHT },
         { "60%",  OPTION_VAL_60_CHT },
         { "70%",  OPTION_VAL_70_CHT },
         { "80%",  OPTION_VAL_80_CHT },
         { "90%",  OPTION_VAL_90_CHT },
         { "100%", OPTION_VAL_100_O47_CHT },
         { "110%", OPTION_VAL_110_O47_CHT },
         { "120%", OPTION_VAL_120_O47_CHT },
         { "130%", OPTION_VAL_130_O47_CHT },
         { "140%", OPTION_VAL_140_O47_CHT },
         { "150%", OPTION_VAL_150_O47_CHT },
         { "160%", OPTION_VAL_160_O47_CHT },
         { "170%", OPTION_VAL_170_O47_CHT },
         { "180%", OPTION_VAL_180_O47_CHT },
         { "190%", OPTION_VAL_190_O47_CHT },
         { "200%", OPTION_VAL_200_O47_CHT },
         { "210%", OPTION_VAL_210_O47_CHT },
         { "220%", OPTION_VAL_220_O47_CHT },
         { "230%", OPTION_VAL_230_O47_CHT },
         { "240%", OPTION_VAL_240_O47_CHT },
         { "250%", OPTION_VAL_250_O47_CHT },
         { "260%", OPTION_VAL_260_O47_CHT },
         { "270%", OPTION_VAL_270_O47_CHT },
         { "280%", OPTION_VAL_280_O47_CHT },
         { "290%", OPTION_VAL_290_O47_CHT },
         { "300%", OPTION_VAL_300_O47_CHT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHT },
         { "Red",      OPTION_VAL_RED_CHT },
         { "Green",    OPTION_VAL_GREEN_CHT },
         { "Blue",     OPTION_VAL_BLUE_CHT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHT },
         { "Red",      OPTION_VAL_RED_CHT },
         { "Green",    OPTION_VAL_GREEN_CHT },
         { "Blue",     OPTION_VAL_BLUE_CHT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHT },
         { "Red",      OPTION_VAL_RED_CHT },
         { "Green",    OPTION_VAL_GREEN_CHT },
         { "Blue",     OPTION_VAL_BLUE_CHT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CHT },
         { "Red",      OPTION_VAL_RED_CHT },
         { "Green",    OPTION_VAL_GREEN_CHT },
         { "Blue",     OPTION_VAL_BLUE_CHT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_CHT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHT },
         { "Purupuru", OPTION_VAL_PURUPURU_CHT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CHT },
         { "None",     OPTION_VAL_NONE_CHT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_CHT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHT },
         { "Purupuru", OPTION_VAL_PURUPURU_CHT },
         { "None",     OPTION_VAL_NONE_CHT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_CHT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHT },
         { "Purupuru", OPTION_VAL_PURUPURU_CHT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CHT },
         { "None",     OPTION_VAL_NONE_CHT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_CHT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHT },
         { "Purupuru", OPTION_VAL_PURUPURU_CHT },
         { "None",     OPTION_VAL_NONE_CHT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_CHT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHT },
         { "Purupuru", OPTION_VAL_PURUPURU_CHT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CHT },
         { "None",     OPTION_VAL_NONE_CHT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_CHT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHT },
         { "Purupuru", OPTION_VAL_PURUPURU_CHT },
         { "None",     OPTION_VAL_NONE_CHT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_CHT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHT },
         { "Purupuru", OPTION_VAL_PURUPURU_CHT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CHT },
         { "None",     OPTION_VAL_NONE_CHT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_CHT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CHT },
         { "Purupuru", OPTION_VAL_PURUPURU_CHT },
         { "None",     OPTION_VAL_NONE_CHT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CHT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CHT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CHT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_CHT },
         { "All VMUs", OPTION_VAL_ALL_VMUS_CHT },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CHT,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_CHT,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_CHT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_CHT,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_CHT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CHT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CHT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CHT,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHT },
         { "2x", OPTION_VAL_2_O26_CHT },
         { "3x", OPTION_VAL_3X_CHT },
         { "4x", OPTION_VAL_4_CHT },
         { "5x", OPTION_VAL_5X_CHT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHT },
         { "20%",  OPTION_VAL_20_CHT },
         { "30%",  OPTION_VAL_30_CHT },
         { "40%",  OPTION_VAL_40_CHT },
         { "50%",  OPTION_VAL_50_CHT },
         { "60%",  OPTION_VAL_60_CHT },
         { "70%",  OPTION_VAL_70_CHT },
         { "80%",  OPTION_VAL_80_CHT },
         { "90%",  OPTION_VAL_90_CHT },
         { "100%", OPTION_VAL_100_O47_CHT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHT },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHT },
         { "2x", OPTION_VAL_2_O26_CHT },
         { "3x", OPTION_VAL_3X_CHT },
         { "4x", OPTION_VAL_4_CHT },
         { "5x", OPTION_VAL_5X_CHT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHT },
         { "20%",  OPTION_VAL_20_CHT },
         { "30%",  OPTION_VAL_30_CHT },
         { "40%",  OPTION_VAL_40_CHT },
         { "50%",  OPTION_VAL_50_CHT },
         { "60%",  OPTION_VAL_60_CHT },
         { "70%",  OPTION_VAL_70_CHT },
         { "80%",  OPTION_VAL_80_CHT },
         { "90%",  OPTION_VAL_90_CHT },
         { "100%", OPTION_VAL_100_O47_CHT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHT },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHT },
         { "2x", OPTION_VAL_2_O26_CHT },
         { "3x", OPTION_VAL_3X_CHT },
         { "4x", OPTION_VAL_4_CHT },
         { "5x", OPTION_VAL_5X_CHT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHT },
         { "20%",  OPTION_VAL_20_CHT },
         { "30%",  OPTION_VAL_30_CHT },
         { "40%",  OPTION_VAL_40_CHT },
         { "50%",  OPTION_VAL_50_CHT },
         { "60%",  OPTION_VAL_60_CHT },
         { "70%",  OPTION_VAL_70_CHT },
         { "80%",  OPTION_VAL_80_CHT },
         { "90%",  OPTION_VAL_90_CHT },
         { "100%", OPTION_VAL_100_O47_CHT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CHT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CHT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CHT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CHT },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CHT },
         { "2x", OPTION_VAL_2_O26_CHT },
         { "3x", OPTION_VAL_3X_CHT },
         { "4x", OPTION_VAL_4_CHT },
         { "5x", OPTION_VAL_5X_CHT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CHT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CHT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CHT },
         { "BLUE 03",           OPTION_VAL_BLUE_CHT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CHT },
         { "GREEN 05",          OPTION_VAL_GREEN_CHT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CHT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CHT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CHT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CHT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CHT },
         { "RED 11",            OPTION_VAL_RED_CHT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CHT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CHT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CHT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CHT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CHT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CHT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CHT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CHT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CHT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CHT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CHT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CHT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CHT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CHT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CHT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CHT },
         { "WHITE 28",          OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CHT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CHT },
         { "20%",  OPTION_VAL_20_CHT },
         { "30%",  OPTION_VAL_30_CHT },
         { "40%",  OPTION_VAL_40_CHT },
         { "50%",  OPTION_VAL_50_CHT },
         { "60%",  OPTION_VAL_60_CHT },
         { "70%",  OPTION_VAL_70_CHT },
         { "80%",  OPTION_VAL_80_CHT },
         { "90%",  OPTION_VAL_90_CHT },
         { "100%", OPTION_VAL_100_O47_CHT },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_cht = {
   option_cats_cht,
   option_defs_cht
};

/* RETRO_LANGUAGE_CS */

#define CATEGORY_SYSTEM_LABEL_CS "Systém"
#define CATEGORY_SYSTEM_INFO_0_CS "Konfigurace regionu, jazyka, systému BIOS a základního nastavení hardwaru."
#define CATEGORY_VIDEO_LABEL_CS NULL
#define CATEGORY_VIDEO_INFO_0_CS "Konfigurace rozlišení, nastavení průhlednosti a vizuálních efektů nezávislých na pořadí."
#define CATEGORY_PERFORMANCE_LABEL_CS "Výkon"
#define CATEGORY_PERFORMANCE_INFO_0_CS "Konfigurace nastavení vykreslování ve vláknech a přeskakování snímků."
#define CATEGORY_HACKS_LABEL_CS "Emulace hacků"
#define CATEGORY_HACKS_INFO_0_CS "Konfigurace přepisu širokoúhlého zobrazení, rychlosti načítání GD-ROM a nastavení výměny textur."
#define CATEGORY_INPUT_LABEL_CS "Vstup"
#define CATEGORY_INPUT_INFO_0_CS "Konfigurace nastavení gamepadu a světelné pistole."
#define CATEGORY_EXPANSIONS_LABEL_CS "Rozšiřující sloty ovladače"
#define CATEGORY_EXPANSIONS_INFO_0_CS "Vyberte zařízení (VMU, rumble device) zapojené do každého rozšiřujícího slotu ovladače."
#define CATEGORY_VMU_LABEL_CS "Jednotka Vizuální Paměti"
#define CATEGORY_VMU_INFO_0_CS "Konfigurace ukládacích souborů VMU pro jednotlivé hry a nastavení viditelnosti VMU na obrazovce."
#define CORE_OPTION_NAME_REGION_LABEL_CS NULL
#define OPTION_VAL_JAPAN_CS "Japonsko"
#define OPTION_VAL_USA_CS NULL
#define OPTION_VAL_EUROPE_CS "Evropa"
#define OPTION_VAL_DEFAULT_CS "Výchozí"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_CS "Jazyk"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_CS "Změní jazyk používaný systémem BIOS a všemi hrami, které obsahují více jazyků."
#define OPTION_VAL_JAPANESE_CS "Japonština"
#define OPTION_VAL_ENGLISH_CS "Angličtina"
#define OPTION_VAL_GERMAN_CS "Němčina"
#define OPTION_VAL_FRENCH_CS "Francouzština"
#define OPTION_VAL_SPANISH_CS "Španělština"
#define OPTION_VAL_ITALIAN_CS "Italský"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_CS "HLE BIOS (nutný restart)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_CS "Vynucení použití vysokoúrovňové emulace systému BIOS."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_CS "Povolit DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CS "Umožňuje emulaci zvukového DSP (digitálního signálového procesoru) konzole Dreamcast. Zlepšuje přesnost generovaného zvuku, ale zvyšuje nároky na výkon."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CS "Povolení tlačítek služby Arcade"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CS "Povoluje tlačítko SERVICE pro arkádové hry, které slouží k zadání nastavení kabinetu."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CS "Nastavení her NAOMI na volné hraní"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CS "Upravte nastavení mincí ve hře na volnou hru."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_CS "Emulace širokopásmového adaptéru"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CS "Emulujte širokopásmový adaptér Ethernet namísto modemu. (Je nutný restart.)"
#define CORE_OPTION_NAME_UPNP_LABEL_CS "Povolit UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_CS "Pomocí UPnP můžete automaticky nakonfigurovat svůj internetový směrovač pro online hry."
#define CORE_OPTION_NAME_DCNET_LABEL_CS "Použít DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_CS "Pro přístup k internetu na Dreamcastu použijte cloudovou službu DCNet."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CS "Vnitřní rozlišení"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CS "Úprava rozlišení vykreslování."
#define OPTION_VAL_320X240_CS "320x240 (Poloviční)"
#define OPTION_VAL_640X480_CS "640x480 (nativní)"
#define OPTION_VAL_800X600_CS NULL
#define OPTION_VAL_960X720_CS "960 x 720 (x1.5)"
#define OPTION_VAL_1024X768_CS NULL
#define OPTION_VAL_1280X960_CS NULL
#define OPTION_VAL_1440X1080_CS "1440 x 1080 (x2.25)"
#define OPTION_VAL_1600X1200_CS NULL
#define OPTION_VAL_1920X1440_CS NULL
#define OPTION_VAL_2560X1920_CS "2560 x 1920 (x4)"
#define OPTION_VAL_2880X2160_CS NULL
#define OPTION_VAL_3200X2400_CS NULL
#define OPTION_VAL_3840X2880_CS NULL
#define OPTION_VAL_4480X3360_CS NULL
#define OPTION_VAL_5120X3840_CS NULL
#define OPTION_VAL_5760X4320_CS NULL
#define OPTION_VAL_6400X4800_CS NULL
#define OPTION_VAL_7040X5280_CS NULL
#define OPTION_VAL_7680X5760_CS NULL
#define OPTION_VAL_8320X6240_CS NULL
#define OPTION_VAL_8960X6720_CS NULL
#define OPTION_VAL_9600X7200_CS NULL
#define OPTION_VAL_10240X7680_CS NULL
#define OPTION_VAL_10880X8160_CS NULL
#define OPTION_VAL_11520X8640_CS NULL
#define OPTION_VAL_12160X9120_CS NULL
#define OPTION_VAL_12800X9600_CS NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_CS "Typ kabelu"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CS "Typ výstupního signálu. Nejvíce podporovaný je 'TV (Kompozitní)'."
#define OPTION_VAL_VGA_CS NULL
#define OPTION_VAL_TV_RGB_CS NULL
#define OPTION_VAL_TV_COMPOSITE_CS "TV (Kompozitní)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_CS "Standardní Vysílání"
#define OPTION_VAL_NTSC_CS NULL
#define OPTION_VAL_PAL_CS "PAL (Svět)"
#define OPTION_VAL_PAL_N_CS NULL
#define OPTION_VAL_PAL_M_CS "PAL-M (Brazílie)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CS "Orientace obrazovky"
#define OPTION_VAL_HORIZONTAL_CS "Horizontální"
#define OPTION_VAL_VERTICAL_CS "Vertikální"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CS "Třídění Alpha"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_CS "Vyberte způsob řazení průhledných polygonů."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CS "Per-Strip (rychlé, nejméně přesné)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_CS "Per-Triangle (normální)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_CS "Per-Pixel (přesné, ale nejpomalejší)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CS "Velikost akumulační vyrovnávací paměti pixelů"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CS "Pro správný výstup ve vyšších rozlišeních mohou být nutné vyšší hodnoty."
#define OPTION_VAL_512MB_CS NULL
#define OPTION_VAL_1GB_CS NULL
#define OPTION_VAL_2GB_CS NULL
#define OPTION_VAL_4GB_CS NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_CS "Maximální počet průhledných vrstev"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CS "U složitých scén mohou být nutné vyšší hodnoty."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CS "Úplná emulace framebufferu"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CS "Povolení úplné emulace framebufferu v paměti VRAM. To je užitečné pro hry, které přímo čtou nebo zapisují framebuffer do VRAM. Pokud je tato funkce povolena, interní rozlišení je vynuceno na 640x480 a může dojít k výraznému snížení výkonu."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CS "Povolení Vyrovnávací Paměti RTT (Render To Texture)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CS "Kopírování vykreslených textur z GPU zpět do VRAM. Tato možnost je obvykle povolena u her, které ji vyžadují. Pokud je povolena, je vypnuto zvyšování rozlišení vykreslovaných textur a může dojít ke snížení výkonu."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_CS "Mipmapování"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_CS "Pokud je tato funkce povolena, textury použijí menší verzi sebe sama, když se objeví ve větší vzdálenosti, což může zvýšit výkon a snížit třpyt."
#define CORE_OPTION_NAME_FOG_LABEL_CS "Efekty Mlhy"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CS "Modifikátor Hlasitosti"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CS "Funkce grafického procesoru Dreamcast, která se obvykle používá ve hrách ke kreslení stínů objektů. Tato funkce by měla být normálně povolena - dopad na výkon je obvykle minimální až zanedbatelný."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CS "Anizotropní Filtrování"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CS "Vylepšete kvalitu textur na površích, které jsou vůči kameře pod šikmým úhlem pohledu. Vyšší hodnoty jsou náročnější na grafický procesor. Změny tohoto nastavení se uplatní až po restartu."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CS "Filtrování textur"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CS "Režim filtrování textur, který se má použít. Tento parametr lze použít k vynucení určitého režimu filtrování textur na všech texturách, abyste dosáhli ostřejšího (nebo hladšího) vzhledu než ve výchozím nastavení. Jiné hodnoty než výchozí mohou způsobit různé problémy s vykreslováním. Změny tohoto nastavení se uplatní až po restartu."
#define OPTION_VAL_1_CS "Vynucený Nearest-Neighbor"
#define OPTION_VAL_2_CS "Vynucený lineární"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CS "Výměna Zpožděných Snímků"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CS "Je užitečné vyhnout se blikajícím obrazovkám nebo poruchovým videím. Nedoporučuje se na pomalých platformách."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CS "Detekce změn snímkové frekvence"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CS "Upozornit frontend při změně interní snímkové frekvence (např. z 60 fps na 30 fps). Zlepšuje snímkovou frekvenci ve hrách, které běží s uzamčenou snímkovou frekvencí 30 fps nebo 20 fps, ale měl by být vypnut pro hry s odemčenou (nestabilní) snímkovou frekvencí (např. Ecco the Dolphin, Unreal Tournament). Poznámka: Není k dispozici, pokud je povoleno 'Automatické přeskakování snímků'."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CS "Postprocesní Filtr PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CS "Následné zpracování vykresleného obrazu pro simulaci efektů specifických pro grafický procesor PowerVR2 a analogové video signály."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_CS "Vylepšení Textur (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CS "Vylepšete ručně kreslenou 2D pixelovou grafiku. Mělo by se používat pouze ve 2D pixelových hrách."
#define OPTION_VAL_2_O26_CS NULL
#define OPTION_VAL_4_CS NULL
#define OPTION_VAL_6_CS NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CS "Zvýšení Rozlišení Textur Max. Filtrovaná Velikost"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_CS "Vyberte maximální hodnotu velikosti textury, která má být zvětšena, pokud je velikost textury větší než vybraná hodnota, nebude zvětšena."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CS "Nativní interpolace hloubky"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CS "Pomáhá řešit problémy s poškozením textur a hloubkou na grafických procesorech AMD. V některých případech může pomoci i GPU Intel."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_CS "Fix Upscalingu krvacející okraj"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_CS "Pomáhá v případě krvácení textur při upscalingu. Jeho vypnutí může pomoci, pokud se pixely při upscalingu ve 2D hrách (MVC2, CVS, KOF atd.) deformují."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CS "Vykreslování Vláken"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CS "Spouští GPU a CPU v různých vláknech. Důrazně doporučujeme."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CS "Automatické Přeskočení Snímků"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CS "Automatické přeskakování snímků při pomalém běhu emulátoru. Poznámka: Toto nastavení platí pouze tehdy, když je povoleno 'Vláknové Vykreslování'."
#define OPTION_VAL_SOME_CS "Normální"
#define OPTION_VAL_MORE_CS "Maximální"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CS "Přeskočení Snímků"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CS "Nastaví počet snímků, které se mají přeskočit mezi jednotlivými zobrazenými snímky."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CS "Širokoúhlé Cheaty (Nutný Restart)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CS "Aktivuje cheaty, které umožňují zobrazení některých her v širokoúhlém formátu."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CS "Hack Širokoúhleho Obrazu"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CS "Kreslení geometrie mimo běžný poměr stran 4:3. Může způsobit grafické závady v odhalených oblastech."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CS "Rychlé Načítání GD-ROM (nepřesné)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CS "Zrychluje načítání GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_CS NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_CS "Povoluje 32MB RAM Mod pro Dreamcast. Může ovlivnit kompatibilitu"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_CS "Podtaktování/přetaktování procesoru SH4"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_CS "Změňte takt hlavního procesoru SH4 z výchozích 200 MHz. Podtaktování může pomoci pomalým platformám. Přetaktování může u některých her zvýšit snímkovou frekvenci. Používejte s opatrností."
#define OPTION_VAL_100_CS NULL
#define OPTION_VAL_110_CS NULL
#define OPTION_VAL_120_CS NULL
#define OPTION_VAL_130_CS NULL
#define OPTION_VAL_140_CS NULL
#define OPTION_VAL_150_CS NULL
#define OPTION_VAL_160_CS NULL
#define OPTION_VAL_170_CS NULL
#define OPTION_VAL_180_CS NULL
#define OPTION_VAL_190_CS NULL
#define OPTION_VAL_200_CS NULL
#define OPTION_VAL_210_CS NULL
#define OPTION_VAL_220_CS NULL
#define OPTION_VAL_230_CS NULL
#define OPTION_VAL_240_CS NULL
#define OPTION_VAL_250_CS NULL
#define OPTION_VAL_260_CS NULL
#define OPTION_VAL_270_CS NULL
#define OPTION_VAL_280_CS NULL
#define OPTION_VAL_290_CS NULL
#define OPTION_VAL_300_CS NULL
#define OPTION_VAL_310_CS NULL
#define OPTION_VAL_320_CS NULL
#define OPTION_VAL_330_CS NULL
#define OPTION_VAL_340_CS NULL
#define OPTION_VAL_350_CS NULL
#define OPTION_VAL_360_CS NULL
#define OPTION_VAL_370_CS NULL
#define OPTION_VAL_380_CS NULL
#define OPTION_VAL_390_CS NULL
#define OPTION_VAL_400_CS NULL
#define OPTION_VAL_410_CS NULL
#define OPTION_VAL_420_CS NULL
#define OPTION_VAL_430_CS NULL
#define OPTION_VAL_440_CS NULL
#define OPTION_VAL_450_CS NULL
#define OPTION_VAL_460_CS NULL
#define OPTION_VAL_470_CS NULL
#define OPTION_VAL_480_CS NULL
#define OPTION_VAL_490_CS NULL
#define OPTION_VAL_500_CS NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CS "Načíst Vlastní Textury"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_CS "Načtení vlastních textur umístěných ve složce 'system/dc/textures/<game-id>/'."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_CS "Přednačíst vlastní textury"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_CS "Načíst vlastní textury při spuštění hry. Může zlepšit výkon, ale zvyšuje využití paměti."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CS "Skladiště Textur"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_CS "Pokaždé, když hra použije novou texturu, uloží ji jako soubor .png do složky 'system/dc/texdump/<game-id>/'."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_CS "Vymazat nahrazené textury"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_CS "Vždy vymažte textury, které již byly nahrazeny vlastními texturami."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CS "Analogova Páčka Mrtvé Zóny"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_CS "Zvolte, jak daleko musíte zatlačit na analogovou páčku, než se začne zpracovávat."
#define OPTION_VAL_0_CS NULL
#define OPTION_VAL_5_CS NULL
#define OPTION_VAL_10_CS NULL
#define OPTION_VAL_15_CS NULL
#define OPTION_VAL_20_CS NULL
#define OPTION_VAL_25_CS NULL
#define OPTION_VAL_30_CS NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CS "Spouštěč Mrtvé Zóny"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_CS "Zvolte, jak dlouho musíte stisknout spoušť, než se začne zpracovávat."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CS "Digitální Spouštěče"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_CS "Pokud je tato funkce povolena, budou se spouštěče chovat jako běžná tlačítka, což znamená, že budou zpracovávány buď jako plně stisknuté, nebo jako nestisknuté, bez mezistupně."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CS "Digitální výstupy pro vysílání"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CS "Vysílání digitálních výstupů a stavu vynucene zpětné vazby na portu TCP 8000. Kompatibilní s volbou MAME \"-výstupní síť\"."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CS "Zobrazit Nastavení Světelné Pistole"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CS "Povolení konfigurace možností zobrazení zaměřovacího kříže světelné pistole. POZNÁMKA: Aby se toto nastavení projevilo, může být nutné přepnout rychlou nabídku."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_CS "Měřítko velikosti zaměřovacího kříže zbraně"
#define OPTION_VAL_50_CS NULL
#define OPTION_VAL_60_CS NULL
#define OPTION_VAL_70_CS NULL
#define OPTION_VAL_80_CS NULL
#define OPTION_VAL_90_CS NULL
#define OPTION_VAL_100_O47_CS NULL
#define OPTION_VAL_110_O47_CS NULL
#define OPTION_VAL_120_O47_CS NULL
#define OPTION_VAL_130_O47_CS NULL
#define OPTION_VAL_140_O47_CS NULL
#define OPTION_VAL_150_O47_CS NULL
#define OPTION_VAL_160_O47_CS NULL
#define OPTION_VAL_170_O47_CS NULL
#define OPTION_VAL_180_O47_CS NULL
#define OPTION_VAL_190_O47_CS NULL
#define OPTION_VAL_200_O47_CS NULL
#define OPTION_VAL_210_O47_CS NULL
#define OPTION_VAL_220_O47_CS NULL
#define OPTION_VAL_230_O47_CS NULL
#define OPTION_VAL_240_O47_CS NULL
#define OPTION_VAL_250_O47_CS NULL
#define OPTION_VAL_260_O47_CS NULL
#define OPTION_VAL_270_O47_CS NULL
#define OPTION_VAL_280_O47_CS NULL
#define OPTION_VAL_290_O47_CS NULL
#define OPTION_VAL_300_O47_CS NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CS "Zobrazení Kříže Zbraně 1"
#define OPTION_VAL_WHITE_CS "Bílá"
#define OPTION_VAL_RED_CS "Červený"
#define OPTION_VAL_GREEN_CS "Zelený"
#define OPTION_VAL_BLUE_CS "Modrý"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CS "Zobrazení Kříže Zbraně 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CS "Zobrazení Kříže Zbraně 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CS "Zobrazení Kříže Zbraně 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_CS "Zařízení v rozšiřujícím slotu A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_CS "Vyberte zařízení zapojené do rozšiřujícího slotu A1 (port A slot 1)."
#define OPTION_VAL_VMU_CS NULL
#define OPTION_VAL_PURUPURU_CS "Balíček vibrací"
#define OPTION_VAL_DREAMPOTATO_CS NULL
#define OPTION_VAL_NONE_CS "Žádné"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_CS "Zařízení v rozšiřujícím slotu A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_CS "Vyberte zařízení zapojené do rozšiřujícího slotu A2 (port A slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_CS "Zařízení v rozšiřujícím slotu B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_CS "Vyberte zařízení zapojené do rozšiřujícího slotu B1 (port B slot 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_CS "Zařízení v rozšiřujícím slotu B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_CS "Vyberte zařízení zapojené do rozšiřujícího slotu B2 (port B slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_CS "Zařízení v rozšiřujícím slotu C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_CS "Vyberte zařízení zapojené do rozšiřujícího slotu C1 (port C slot 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_CS "Zařízení v rozšiřujícím slotu C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_CS "Vyberte zařízení zapojené do rozšiřujícího slotu C2 (port C slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_CS "Zařízení v rozšiřujícím slotu D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_CS "Vyberte zařízení zapojené do rozšiřujícího slotu D1 (port D slot 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_CS "Zařízení v rozšiřujícím slotu D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_CS "Vyberte zařízení zapojené do rozšiřujícího slotu D2 (port D slot 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CS "Vizuální Paměťové Jednotky/Systémy pro Jednotlivé Hry (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CS "VMUs Na-Hru"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CS "Pokud je vypnuto, všechny hry sdílejí až 8 souborů VMU (A1/A2/B1/B2/C1/C2/D1/D2) umístěných v systémové složce aplikace RetroArch.\nNastavení „VMU A1“ vytvoří pro každou spuštěnou hru jedinečný soubor VMU „A1“ ve složce uložení aplikace RetroArch.\nNastavení „Všechny VMU“ vytvoří pro každou spuštěnou hru až 8 jedinečných souborů VMU (A1/A2/B1/B2/C1/C2/D1/D2)."
#define OPTION_VAL_VMU_A1_CS NULL
#define OPTION_VAL_ALL_VMUS_CS NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CS "Vizuální paměťové jednotky/systémy (VMU) Zvuky"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_CS "Zvuky VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_CS "Pokud je tato funkce povolena, přehrávají se zvukové signály VMU."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_CS "Úložiště DreamPotato VMU"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_CS "Povolí ukládání dat VMU na připojený emulátor DreamPotato. Pokud je tato funkce povolena, místní soubor VMU se nepoužívá."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CS "Zobrazit Nastavení Displeje Vizuální Paměťové Jednotky/Systému (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CS "Zobrazit Nastavení VMU Displeje"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CS "Umožňuje konfiguraci viditelnosti, velikosti, polohy a barvy obrazovky LCD emulovaného VMU. POZNÁMKA: Může být nutné přepnout Rychlé Menu, aby se toto nastavení projevilo."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CS "Displej Obrazovky VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CS "Pozice Obrazovky VMU 1"
#define OPTION_VAL_UPPER_LEFT_CS "Vlevo Nahoře"
#define OPTION_VAL_UPPER_RIGHT_CS "Vpravo Nahoře"
#define OPTION_VAL_LOWER_LEFT_CS "Vlevo Dole"
#define OPTION_VAL_LOWER_RIGHT_CS "Vpravo Dole"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CS "Velikost Obrazovky VMU 1"
#define OPTION_VAL_1X_CS NULL
#define OPTION_VAL_3X_CS NULL
#define OPTION_VAL_5X_CS NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CS "Zapnout Barvu Pixelu Obrazovky VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_CS "Výchozí ON"
#define OPTION_VAL_DEFAULT_OFF_01_CS "Výchozí OFF"
#define OPTION_VAL_BLACK_02_CS "Černá"
#define OPTION_VAL_LIGHT_BLUE_04_CS "Světle Modrá"
#define OPTION_VAL_CYAN_06_CS "Azurová"
#define OPTION_VAL_CYAN_BLUE_07_CS "Azurově Modrá"
#define OPTION_VAL_LIGHT_GREEN_08_CS "Světle Zelená"
#define OPTION_VAL_CYAN_GREEN_09_CS "Azurově Zelená"
#define OPTION_VAL_LIGHT_CYAN_10_CS "Světle Azurová"
#define OPTION_VAL_PURPLE_12_CS "Fialová"
#define OPTION_VAL_LIGHT_PURPLE_13_CS "Světle Fialová"
#define OPTION_VAL_YELLOW_14_CS "Žlutá"
#define OPTION_VAL_GRAY_15_CS "Šedá"
#define OPTION_VAL_LIGHT_PURPLE_2_16_CS "Světle Fialová (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_CS "Světle Zelená (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_CS "Světle Zelená (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_CS "Světle Azurová (2)"
#define OPTION_VAL_LIGHT_RED_2_20_CS "Světle Červená (2)"
#define OPTION_VAL_MAGENTA_21_CS "Purpurová"
#define OPTION_VAL_LIGHT_PURPLE_3_22_CS "Světle Fialová (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_CS "Světle Oranžová"
#define OPTION_VAL_ORANGE_24_CS "Oranžová"
#define OPTION_VAL_LIGHT_PURPLE_4_25_CS "Světle Fialová (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_CS "Světle Žlutá"
#define OPTION_VAL_LIGHT_YELLOW_2_27_CS "Světle Žlutá (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CS "Vypnout Barvu Pixelu Obrazovky VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CS "Průhlednost Obrazovky VMU 1"
#define OPTION_VAL_40_CS NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CS "Displej Obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CS "Pozice Obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CS "Velikost Obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CS "Zapnout Barvu Pixelu Obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CS "Vypnout Barvu Pixelu Obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CS "Průhlednost Obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CS "Displej Obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CS "Pozice Obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CS "Velikost Obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CS "Zapnout Barvu Pixelu Obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CS "Vypnout Barvu Pixelu Obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CS "Průhlednost Obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CS "Displej Obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CS "Pozice Obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CS "Velikost Obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CS "Zapnout Barvu Pixelu Obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CS "Vypnout Barvu Pixelu Obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CS "Průhlednost Obrazovky VMU 4"

struct retro_core_option_v2_category option_cats_cs[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CS,
      CATEGORY_SYSTEM_INFO_0_CS
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CS,
      CATEGORY_VIDEO_INFO_0_CS
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_CS,
      CATEGORY_PERFORMANCE_INFO_0_CS
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CS,
      CATEGORY_HACKS_INFO_0_CS
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CS,
      CATEGORY_INPUT_INFO_0_CS
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_CS,
      CATEGORY_EXPANSIONS_INFO_0_CS
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_CS,
      CATEGORY_VMU_INFO_0_CS
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_cs[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_CS,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_CS },
         { "USA",     OPTION_VAL_USA_CS },
         { "Europe",  OPTION_VAL_EUROPE_CS },
         { "Default", OPTION_VAL_DEFAULT_CS },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_CS,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_CS },
         { "English",  OPTION_VAL_ENGLISH_CS },
         { "German",   OPTION_VAL_GERMAN_CS },
         { "French",   OPTION_VAL_FRENCH_CS },
         { "Spanish",  OPTION_VAL_SPANISH_CS },
         { "Italian",  OPTION_VAL_ITALIAN_CS },
         { "Default",  OPTION_VAL_DEFAULT_CS },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CS,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_CS },
         { "640x480",    OPTION_VAL_640X480_CS },
         { "800x600",    OPTION_VAL_800X600_CS },
         { "960x720",    OPTION_VAL_960X720_CS },
         { "1024x768",   OPTION_VAL_1024X768_CS },
         { "1280x960",   OPTION_VAL_1280X960_CS },
         { "1440x1080",  OPTION_VAL_1440X1080_CS },
         { "1600x1200",  OPTION_VAL_1600X1200_CS },
         { "1920x1440",  OPTION_VAL_1920X1440_CS },
         { "2560x1920",  OPTION_VAL_2560X1920_CS },
         { "2880x2160",  OPTION_VAL_2880X2160_CS },
         { "3200x2400",  OPTION_VAL_3200X2400_CS },
         { "3840x2880",  OPTION_VAL_3840X2880_CS },
         { "4480x3360",  OPTION_VAL_4480X3360_CS },
         { "5120x3840",  OPTION_VAL_5120X3840_CS },
         { "5760x4320",  OPTION_VAL_5760X4320_CS },
         { "6400x4800",  OPTION_VAL_6400X4800_CS },
         { "7040x5280",  OPTION_VAL_7040X5280_CS },
         { "7680x5760",  OPTION_VAL_7680X5760_CS },
         { "8320x6240",  OPTION_VAL_8320X6240_CS },
         { "8960x6720",  OPTION_VAL_8960X6720_CS },
         { "9600x7200",  OPTION_VAL_9600X7200_CS },
         { "10240x7680", OPTION_VAL_10240X7680_CS },
         { "10880x8160", OPTION_VAL_10880X8160_CS },
         { "11520x8640", OPTION_VAL_11520X8640_CS },
         { "12160x9120", OPTION_VAL_12160X9120_CS },
         { "12800x9600", OPTION_VAL_12800X9600_CS },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CS,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_CS },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_CS },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_CS },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_CS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_CS },
         { "PAL",     OPTION_VAL_PAL_CS },
         { "PAL_N",   OPTION_VAL_PAL_N_CS },
         { "PAL_M",   OPTION_VAL_PAL_M_CS },
         { "Default", OPTION_VAL_DEFAULT_CS },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_CS },
         { "vertical",   OPTION_VAL_VERTICAL_CS },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_CS,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CS },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_CS },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_CS },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CS,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_CS },
         { "1GB",   OPTION_VAL_1GB_CS },
         { "2GB",   OPTION_VAL_2GB_CS },
         { "4GB",   OPTION_VAL_4GB_CS },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CS,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_CS,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CS,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CS,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_CS },
         { "1",  OPTION_VAL_1_CS },
         { "2",  OPTION_VAL_2_CS },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CS,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_CS },
         { "4",  OPTION_VAL_4_CS },
         { "6",  OPTION_VAL_6_CS },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_CS,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CS,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CS,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_CS,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_CS,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CS,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CS,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_CS },
         { "more", OPTION_VAL_MORE_CS },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CS,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_CS },
         { "110", OPTION_VAL_110_CS },
         { "120", OPTION_VAL_120_CS },
         { "130", OPTION_VAL_130_CS },
         { "140", OPTION_VAL_140_CS },
         { "150", OPTION_VAL_150_CS },
         { "160", OPTION_VAL_160_CS },
         { "170", OPTION_VAL_170_CS },
         { "180", OPTION_VAL_180_CS },
         { "190", OPTION_VAL_190_CS },
         { "200", OPTION_VAL_200_CS },
         { "210", OPTION_VAL_210_CS },
         { "220", OPTION_VAL_220_CS },
         { "230", OPTION_VAL_230_CS },
         { "240", OPTION_VAL_240_CS },
         { "250", OPTION_VAL_250_CS },
         { "260", OPTION_VAL_260_CS },
         { "270", OPTION_VAL_270_CS },
         { "280", OPTION_VAL_280_CS },
         { "290", OPTION_VAL_290_CS },
         { "300", OPTION_VAL_300_CS },
         { "310", OPTION_VAL_310_CS },
         { "320", OPTION_VAL_320_CS },
         { "330", OPTION_VAL_330_CS },
         { "340", OPTION_VAL_340_CS },
         { "350", OPTION_VAL_350_CS },
         { "360", OPTION_VAL_360_CS },
         { "370", OPTION_VAL_370_CS },
         { "380", OPTION_VAL_380_CS },
         { "390", OPTION_VAL_390_CS },
         { "400", OPTION_VAL_400_CS },
         { "410", OPTION_VAL_410_CS },
         { "420", OPTION_VAL_420_CS },
         { "430", OPTION_VAL_430_CS },
         { "440", OPTION_VAL_440_CS },
         { "450", OPTION_VAL_450_CS },
         { "460", OPTION_VAL_460_CS },
         { "470", OPTION_VAL_470_CS },
         { "480", OPTION_VAL_480_CS },
         { "490", OPTION_VAL_490_CS },
         { "500", OPTION_VAL_500_CS },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_CS,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CS },
         { "5%",  OPTION_VAL_5_CS },
         { "10%", OPTION_VAL_10_CS },
         { "15%", OPTION_VAL_15_CS },
         { "20%", OPTION_VAL_20_CS },
         { "25%", OPTION_VAL_25_CS },
         { "30%", OPTION_VAL_30_CS },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_CS,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CS },
         { "5%",  OPTION_VAL_5_CS },
         { "10%", OPTION_VAL_10_CS },
         { "15%", OPTION_VAL_15_CS },
         { "20%", OPTION_VAL_20_CS },
         { "25%", OPTION_VAL_25_CS },
         { "30%", OPTION_VAL_30_CS },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_CS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CS,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_CS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_CS },
         { "60%",  OPTION_VAL_60_CS },
         { "70%",  OPTION_VAL_70_CS },
         { "80%",  OPTION_VAL_80_CS },
         { "90%",  OPTION_VAL_90_CS },
         { "100%", OPTION_VAL_100_O47_CS },
         { "110%", OPTION_VAL_110_O47_CS },
         { "120%", OPTION_VAL_120_O47_CS },
         { "130%", OPTION_VAL_130_O47_CS },
         { "140%", OPTION_VAL_140_O47_CS },
         { "150%", OPTION_VAL_150_O47_CS },
         { "160%", OPTION_VAL_160_O47_CS },
         { "170%", OPTION_VAL_170_O47_CS },
         { "180%", OPTION_VAL_180_O47_CS },
         { "190%", OPTION_VAL_190_O47_CS },
         { "200%", OPTION_VAL_200_O47_CS },
         { "210%", OPTION_VAL_210_O47_CS },
         { "220%", OPTION_VAL_220_O47_CS },
         { "230%", OPTION_VAL_230_O47_CS },
         { "240%", OPTION_VAL_240_O47_CS },
         { "250%", OPTION_VAL_250_O47_CS },
         { "260%", OPTION_VAL_260_O47_CS },
         { "270%", OPTION_VAL_270_O47_CS },
         { "280%", OPTION_VAL_280_O47_CS },
         { "290%", OPTION_VAL_290_O47_CS },
         { "300%", OPTION_VAL_300_O47_CS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CS },
         { "Red",      OPTION_VAL_RED_CS },
         { "Green",    OPTION_VAL_GREEN_CS },
         { "Blue",     OPTION_VAL_BLUE_CS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CS },
         { "Red",      OPTION_VAL_RED_CS },
         { "Green",    OPTION_VAL_GREEN_CS },
         { "Blue",     OPTION_VAL_BLUE_CS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CS },
         { "Red",      OPTION_VAL_RED_CS },
         { "Green",    OPTION_VAL_GREEN_CS },
         { "Blue",     OPTION_VAL_BLUE_CS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CS,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CS },
         { "Red",      OPTION_VAL_RED_CS },
         { "Green",    OPTION_VAL_GREEN_CS },
         { "Blue",     OPTION_VAL_BLUE_CS },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_CS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CS },
         { "Purupuru", OPTION_VAL_PURUPURU_CS },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CS },
         { "None",     OPTION_VAL_NONE_CS },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_CS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CS },
         { "Purupuru", OPTION_VAL_PURUPURU_CS },
         { "None",     OPTION_VAL_NONE_CS },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_CS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CS },
         { "Purupuru", OPTION_VAL_PURUPURU_CS },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CS },
         { "None",     OPTION_VAL_NONE_CS },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_CS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CS },
         { "Purupuru", OPTION_VAL_PURUPURU_CS },
         { "None",     OPTION_VAL_NONE_CS },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_CS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CS },
         { "Purupuru", OPTION_VAL_PURUPURU_CS },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CS },
         { "None",     OPTION_VAL_NONE_CS },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_CS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CS },
         { "Purupuru", OPTION_VAL_PURUPURU_CS },
         { "None",     OPTION_VAL_NONE_CS },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_CS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CS },
         { "Purupuru", OPTION_VAL_PURUPURU_CS },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CS },
         { "None",     OPTION_VAL_NONE_CS },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_CS,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CS },
         { "Purupuru", OPTION_VAL_PURUPURU_CS },
         { "None",     OPTION_VAL_NONE_CS },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CS,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CS,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CS,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_CS },
         { "All VMUs", OPTION_VAL_ALL_VMUS_CS },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CS,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_CS,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_CS,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_CS,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_CS,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CS,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CS,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CS,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CS },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CS },
         { "2x", OPTION_VAL_2_O26_CS },
         { "3x", OPTION_VAL_3X_CS },
         { "4x", OPTION_VAL_4_CS },
         { "5x", OPTION_VAL_5X_CS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CS },
         { "20%",  OPTION_VAL_20_CS },
         { "30%",  OPTION_VAL_30_CS },
         { "40%",  OPTION_VAL_40_CS },
         { "50%",  OPTION_VAL_50_CS },
         { "60%",  OPTION_VAL_60_CS },
         { "70%",  OPTION_VAL_70_CS },
         { "80%",  OPTION_VAL_80_CS },
         { "90%",  OPTION_VAL_90_CS },
         { "100%", OPTION_VAL_100_O47_CS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CS },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CS },
         { "2x", OPTION_VAL_2_O26_CS },
         { "3x", OPTION_VAL_3X_CS },
         { "4x", OPTION_VAL_4_CS },
         { "5x", OPTION_VAL_5X_CS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CS },
         { "20%",  OPTION_VAL_20_CS },
         { "30%",  OPTION_VAL_30_CS },
         { "40%",  OPTION_VAL_40_CS },
         { "50%",  OPTION_VAL_50_CS },
         { "60%",  OPTION_VAL_60_CS },
         { "70%",  OPTION_VAL_70_CS },
         { "80%",  OPTION_VAL_80_CS },
         { "90%",  OPTION_VAL_90_CS },
         { "100%", OPTION_VAL_100_O47_CS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CS },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CS },
         { "2x", OPTION_VAL_2_O26_CS },
         { "3x", OPTION_VAL_3X_CS },
         { "4x", OPTION_VAL_4_CS },
         { "5x", OPTION_VAL_5X_CS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CS },
         { "20%",  OPTION_VAL_20_CS },
         { "30%",  OPTION_VAL_30_CS },
         { "40%",  OPTION_VAL_40_CS },
         { "50%",  OPTION_VAL_50_CS },
         { "60%",  OPTION_VAL_60_CS },
         { "70%",  OPTION_VAL_70_CS },
         { "80%",  OPTION_VAL_80_CS },
         { "90%",  OPTION_VAL_90_CS },
         { "100%", OPTION_VAL_100_O47_CS },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CS },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CS },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CS },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CS },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CS },
         { "2x", OPTION_VAL_2_O26_CS },
         { "3x", OPTION_VAL_3X_CS },
         { "4x", OPTION_VAL_4_CS },
         { "5x", OPTION_VAL_5X_CS },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CS },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CS },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CS },
         { "BLUE 03",           OPTION_VAL_BLUE_CS },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CS },
         { "GREEN 05",          OPTION_VAL_GREEN_CS },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CS },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CS },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CS },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CS },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CS },
         { "RED 11",            OPTION_VAL_RED_CS },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CS },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CS },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CS },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CS },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CS },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CS },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CS },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CS },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CS },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CS },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CS },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CS },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CS },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CS },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CS },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CS },
         { "WHITE 28",          OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CS,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CS },
         { "20%",  OPTION_VAL_20_CS },
         { "30%",  OPTION_VAL_30_CS },
         { "40%",  OPTION_VAL_40_CS },
         { "50%",  OPTION_VAL_50_CS },
         { "60%",  OPTION_VAL_60_CS },
         { "70%",  OPTION_VAL_70_CS },
         { "80%",  OPTION_VAL_80_CS },
         { "90%",  OPTION_VAL_90_CS },
         { "100%", OPTION_VAL_100_O47_CS },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_cs = {
   option_cats_cs,
   option_defs_cs
};

/* RETRO_LANGUAGE_CY */

#define CATEGORY_SYSTEM_LABEL_CY NULL
#define CATEGORY_SYSTEM_INFO_0_CY NULL
#define CATEGORY_VIDEO_LABEL_CY NULL
#define CATEGORY_VIDEO_INFO_0_CY NULL
#define CATEGORY_PERFORMANCE_LABEL_CY NULL
#define CATEGORY_PERFORMANCE_INFO_0_CY NULL
#define CATEGORY_HACKS_LABEL_CY NULL
#define CATEGORY_HACKS_INFO_0_CY NULL
#define CATEGORY_INPUT_LABEL_CY "Mewnbwn"
#define CATEGORY_INPUT_INFO_0_CY NULL
#define CATEGORY_EXPANSIONS_LABEL_CY NULL
#define CATEGORY_EXPANSIONS_INFO_0_CY NULL
#define CATEGORY_VMU_LABEL_CY NULL
#define CATEGORY_VMU_INFO_0_CY NULL
#define CORE_OPTION_NAME_REGION_LABEL_CY NULL
#define OPTION_VAL_JAPAN_CY NULL
#define OPTION_VAL_USA_CY NULL
#define OPTION_VAL_EUROPE_CY NULL
#define OPTION_VAL_DEFAULT_CY NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_CY NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_CY NULL
#define OPTION_VAL_JAPANESE_CY NULL
#define OPTION_VAL_ENGLISH_CY NULL
#define OPTION_VAL_GERMAN_CY NULL
#define OPTION_VAL_FRENCH_CY NULL
#define OPTION_VAL_SPANISH_CY NULL
#define OPTION_VAL_ITALIAN_CY NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_CY NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_CY NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_CY NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CY NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CY NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CY NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CY NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CY NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_CY NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CY NULL
#define CORE_OPTION_NAME_UPNP_LABEL_CY NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_CY NULL
#define CORE_OPTION_NAME_DCNET_LABEL_CY NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_CY NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CY NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CY NULL
#define OPTION_VAL_320X240_CY NULL
#define OPTION_VAL_640X480_CY NULL
#define OPTION_VAL_800X600_CY NULL
#define OPTION_VAL_960X720_CY NULL
#define OPTION_VAL_1024X768_CY NULL
#define OPTION_VAL_1280X960_CY NULL
#define OPTION_VAL_1440X1080_CY NULL
#define OPTION_VAL_1600X1200_CY NULL
#define OPTION_VAL_1920X1440_CY NULL
#define OPTION_VAL_2560X1920_CY NULL
#define OPTION_VAL_2880X2160_CY NULL
#define OPTION_VAL_3200X2400_CY NULL
#define OPTION_VAL_3840X2880_CY NULL
#define OPTION_VAL_4480X3360_CY NULL
#define OPTION_VAL_5120X3840_CY NULL
#define OPTION_VAL_5760X4320_CY NULL
#define OPTION_VAL_6400X4800_CY NULL
#define OPTION_VAL_7040X5280_CY NULL
#define OPTION_VAL_7680X5760_CY NULL
#define OPTION_VAL_8320X6240_CY NULL
#define OPTION_VAL_8960X6720_CY NULL
#define OPTION_VAL_9600X7200_CY NULL
#define OPTION_VAL_10240X7680_CY NULL
#define OPTION_VAL_10880X8160_CY NULL
#define OPTION_VAL_11520X8640_CY NULL
#define OPTION_VAL_12160X9120_CY NULL
#define OPTION_VAL_12800X9600_CY NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_CY NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CY NULL
#define OPTION_VAL_VGA_CY NULL
#define OPTION_VAL_TV_RGB_CY NULL
#define OPTION_VAL_TV_COMPOSITE_CY NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_CY NULL
#define OPTION_VAL_NTSC_CY NULL
#define OPTION_VAL_PAL_CY NULL
#define OPTION_VAL_PAL_N_CY NULL
#define OPTION_VAL_PAL_M_CY NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CY NULL
#define OPTION_VAL_HORIZONTAL_CY NULL
#define OPTION_VAL_VERTICAL_CY NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CY NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_CY NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CY NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_CY NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_CY NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CY NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CY NULL
#define OPTION_VAL_512MB_CY NULL
#define OPTION_VAL_1GB_CY NULL
#define OPTION_VAL_2GB_CY NULL
#define OPTION_VAL_4GB_CY NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_CY NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CY NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CY NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CY NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CY NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CY NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_CY NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_CY NULL
#define CORE_OPTION_NAME_FOG_LABEL_CY NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CY NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CY NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CY NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CY NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CY NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CY NULL
#define OPTION_VAL_1_CY NULL
#define OPTION_VAL_2_CY NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CY NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CY NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CY NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CY NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CY NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CY NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_CY NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CY NULL
#define OPTION_VAL_2_O26_CY NULL
#define OPTION_VAL_4_CY NULL
#define OPTION_VAL_6_CY NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CY NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_CY NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CY NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CY NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_CY NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_CY NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CY NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CY NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CY NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CY NULL
#define OPTION_VAL_SOME_CY NULL
#define OPTION_VAL_MORE_CY NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CY NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CY NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CY NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CY NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CY NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CY NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CY NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CY NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_CY NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_CY NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_CY NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_CY NULL
#define OPTION_VAL_100_CY NULL
#define OPTION_VAL_110_CY NULL
#define OPTION_VAL_120_CY NULL
#define OPTION_VAL_130_CY NULL
#define OPTION_VAL_140_CY NULL
#define OPTION_VAL_150_CY NULL
#define OPTION_VAL_160_CY NULL
#define OPTION_VAL_170_CY NULL
#define OPTION_VAL_180_CY NULL
#define OPTION_VAL_190_CY NULL
#define OPTION_VAL_200_CY NULL
#define OPTION_VAL_210_CY NULL
#define OPTION_VAL_220_CY NULL
#define OPTION_VAL_230_CY NULL
#define OPTION_VAL_240_CY NULL
#define OPTION_VAL_250_CY NULL
#define OPTION_VAL_260_CY NULL
#define OPTION_VAL_270_CY NULL
#define OPTION_VAL_280_CY NULL
#define OPTION_VAL_290_CY NULL
#define OPTION_VAL_300_CY NULL
#define OPTION_VAL_310_CY NULL
#define OPTION_VAL_320_CY NULL
#define OPTION_VAL_330_CY NULL
#define OPTION_VAL_340_CY NULL
#define OPTION_VAL_350_CY NULL
#define OPTION_VAL_360_CY NULL
#define OPTION_VAL_370_CY NULL
#define OPTION_VAL_380_CY NULL
#define OPTION_VAL_390_CY NULL
#define OPTION_VAL_400_CY NULL
#define OPTION_VAL_410_CY NULL
#define OPTION_VAL_420_CY NULL
#define OPTION_VAL_430_CY NULL
#define OPTION_VAL_440_CY NULL
#define OPTION_VAL_450_CY NULL
#define OPTION_VAL_460_CY NULL
#define OPTION_VAL_470_CY NULL
#define OPTION_VAL_480_CY NULL
#define OPTION_VAL_490_CY NULL
#define OPTION_VAL_500_CY NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CY NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_CY NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_CY NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_CY NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CY NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_CY NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_CY NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_CY NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CY NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_CY NULL
#define OPTION_VAL_0_CY NULL
#define OPTION_VAL_5_CY NULL
#define OPTION_VAL_10_CY NULL
#define OPTION_VAL_15_CY NULL
#define OPTION_VAL_20_CY NULL
#define OPTION_VAL_25_CY NULL
#define OPTION_VAL_30_CY NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CY NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_CY NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CY NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_CY NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CY NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CY NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CY NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CY NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_CY NULL
#define OPTION_VAL_50_CY NULL
#define OPTION_VAL_60_CY NULL
#define OPTION_VAL_70_CY NULL
#define OPTION_VAL_80_CY NULL
#define OPTION_VAL_90_CY NULL
#define OPTION_VAL_100_O47_CY NULL
#define OPTION_VAL_110_O47_CY NULL
#define OPTION_VAL_120_O47_CY NULL
#define OPTION_VAL_130_O47_CY NULL
#define OPTION_VAL_140_O47_CY NULL
#define OPTION_VAL_150_O47_CY NULL
#define OPTION_VAL_160_O47_CY NULL
#define OPTION_VAL_170_O47_CY NULL
#define OPTION_VAL_180_O47_CY NULL
#define OPTION_VAL_190_O47_CY NULL
#define OPTION_VAL_200_O47_CY NULL
#define OPTION_VAL_210_O47_CY NULL
#define OPTION_VAL_220_O47_CY NULL
#define OPTION_VAL_230_O47_CY NULL
#define OPTION_VAL_240_O47_CY NULL
#define OPTION_VAL_250_O47_CY NULL
#define OPTION_VAL_260_O47_CY NULL
#define OPTION_VAL_270_O47_CY NULL
#define OPTION_VAL_280_O47_CY NULL
#define OPTION_VAL_290_O47_CY NULL
#define OPTION_VAL_300_O47_CY NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CY NULL
#define OPTION_VAL_WHITE_CY NULL
#define OPTION_VAL_RED_CY "Coch"
#define OPTION_VAL_GREEN_CY NULL
#define OPTION_VAL_BLUE_CY NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CY NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CY NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_CY NULL
#define OPTION_VAL_VMU_CY NULL
#define OPTION_VAL_PURUPURU_CY NULL
#define OPTION_VAL_DREAMPOTATO_CY NULL
#define OPTION_VAL_NONE_CY "Dim"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_CY NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_CY NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CY NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CY NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CY NULL
#define OPTION_VAL_VMU_A1_CY NULL
#define OPTION_VAL_ALL_VMUS_CY NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_CY NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_CY NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_CY NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_CY NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CY NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CY NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CY NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CY NULL
#define OPTION_VAL_UPPER_LEFT_CY NULL
#define OPTION_VAL_UPPER_RIGHT_CY NULL
#define OPTION_VAL_LOWER_LEFT_CY NULL
#define OPTION_VAL_LOWER_RIGHT_CY NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CY NULL
#define OPTION_VAL_1X_CY NULL
#define OPTION_VAL_3X_CY NULL
#define OPTION_VAL_5X_CY NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CY NULL
#define OPTION_VAL_DEFAULT_ON_00_CY NULL
#define OPTION_VAL_DEFAULT_OFF_01_CY NULL
#define OPTION_VAL_BLACK_02_CY NULL
#define OPTION_VAL_LIGHT_BLUE_04_CY NULL
#define OPTION_VAL_CYAN_06_CY NULL
#define OPTION_VAL_CYAN_BLUE_07_CY NULL
#define OPTION_VAL_LIGHT_GREEN_08_CY NULL
#define OPTION_VAL_CYAN_GREEN_09_CY NULL
#define OPTION_VAL_LIGHT_CYAN_10_CY NULL
#define OPTION_VAL_PURPLE_12_CY NULL
#define OPTION_VAL_LIGHT_PURPLE_13_CY NULL
#define OPTION_VAL_YELLOW_14_CY NULL
#define OPTION_VAL_GRAY_15_CY NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_CY NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_CY NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_CY NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_CY NULL
#define OPTION_VAL_LIGHT_RED_2_20_CY NULL
#define OPTION_VAL_MAGENTA_21_CY NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_CY NULL
#define OPTION_VAL_LIGHT_ORANGE_23_CY NULL
#define OPTION_VAL_ORANGE_24_CY NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_CY NULL
#define OPTION_VAL_LIGHT_YELLOW_26_CY NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_CY NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CY NULL
#define OPTION_VAL_40_CY NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CY NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CY NULL

struct retro_core_option_v2_category option_cats_cy[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CY,
      CATEGORY_SYSTEM_INFO_0_CY
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CY,
      CATEGORY_VIDEO_INFO_0_CY
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_CY,
      CATEGORY_PERFORMANCE_INFO_0_CY
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CY,
      CATEGORY_HACKS_INFO_0_CY
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CY,
      CATEGORY_INPUT_INFO_0_CY
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_CY,
      CATEGORY_EXPANSIONS_INFO_0_CY
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_CY,
      CATEGORY_VMU_INFO_0_CY
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_cy[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_CY,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_CY },
         { "USA",     OPTION_VAL_USA_CY },
         { "Europe",  OPTION_VAL_EUROPE_CY },
         { "Default", OPTION_VAL_DEFAULT_CY },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_CY,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_CY },
         { "English",  OPTION_VAL_ENGLISH_CY },
         { "German",   OPTION_VAL_GERMAN_CY },
         { "French",   OPTION_VAL_FRENCH_CY },
         { "Spanish",  OPTION_VAL_SPANISH_CY },
         { "Italian",  OPTION_VAL_ITALIAN_CY },
         { "Default",  OPTION_VAL_DEFAULT_CY },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_CY,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_CY },
         { "640x480",    OPTION_VAL_640X480_CY },
         { "800x600",    OPTION_VAL_800X600_CY },
         { "960x720",    OPTION_VAL_960X720_CY },
         { "1024x768",   OPTION_VAL_1024X768_CY },
         { "1280x960",   OPTION_VAL_1280X960_CY },
         { "1440x1080",  OPTION_VAL_1440X1080_CY },
         { "1600x1200",  OPTION_VAL_1600X1200_CY },
         { "1920x1440",  OPTION_VAL_1920X1440_CY },
         { "2560x1920",  OPTION_VAL_2560X1920_CY },
         { "2880x2160",  OPTION_VAL_2880X2160_CY },
         { "3200x2400",  OPTION_VAL_3200X2400_CY },
         { "3840x2880",  OPTION_VAL_3840X2880_CY },
         { "4480x3360",  OPTION_VAL_4480X3360_CY },
         { "5120x3840",  OPTION_VAL_5120X3840_CY },
         { "5760x4320",  OPTION_VAL_5760X4320_CY },
         { "6400x4800",  OPTION_VAL_6400X4800_CY },
         { "7040x5280",  OPTION_VAL_7040X5280_CY },
         { "7680x5760",  OPTION_VAL_7680X5760_CY },
         { "8320x6240",  OPTION_VAL_8320X6240_CY },
         { "8960x6720",  OPTION_VAL_8960X6720_CY },
         { "9600x7200",  OPTION_VAL_9600X7200_CY },
         { "10240x7680", OPTION_VAL_10240X7680_CY },
         { "10880x8160", OPTION_VAL_10880X8160_CY },
         { "11520x8640", OPTION_VAL_11520X8640_CY },
         { "12160x9120", OPTION_VAL_12160X9120_CY },
         { "12800x9600", OPTION_VAL_12800X9600_CY },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_CY,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_CY },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_CY },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_CY },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_CY,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_CY },
         { "PAL",     OPTION_VAL_PAL_CY },
         { "PAL_N",   OPTION_VAL_PAL_N_CY },
         { "PAL_M",   OPTION_VAL_PAL_M_CY },
         { "Default", OPTION_VAL_DEFAULT_CY },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_CY,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_CY },
         { "vertical",   OPTION_VAL_VERTICAL_CY },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_CY,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_CY },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_CY },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_CY },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_CY,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_CY },
         { "1GB",   OPTION_VAL_1GB_CY },
         { "2GB",   OPTION_VAL_2GB_CY },
         { "4GB",   OPTION_VAL_4GB_CY },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_CY,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_CY,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_CY,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_CY,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_CY },
         { "1",  OPTION_VAL_1_CY },
         { "2",  OPTION_VAL_2_CY },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_CY,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_CY },
         { "4",  OPTION_VAL_4_CY },
         { "6",  OPTION_VAL_6_CY },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_CY,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_CY,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_CY,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_CY,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_CY,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_CY,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_CY,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_CY },
         { "more", OPTION_VAL_MORE_CY },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_CY,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_CY },
         { "110", OPTION_VAL_110_CY },
         { "120", OPTION_VAL_120_CY },
         { "130", OPTION_VAL_130_CY },
         { "140", OPTION_VAL_140_CY },
         { "150", OPTION_VAL_150_CY },
         { "160", OPTION_VAL_160_CY },
         { "170", OPTION_VAL_170_CY },
         { "180", OPTION_VAL_180_CY },
         { "190", OPTION_VAL_190_CY },
         { "200", OPTION_VAL_200_CY },
         { "210", OPTION_VAL_210_CY },
         { "220", OPTION_VAL_220_CY },
         { "230", OPTION_VAL_230_CY },
         { "240", OPTION_VAL_240_CY },
         { "250", OPTION_VAL_250_CY },
         { "260", OPTION_VAL_260_CY },
         { "270", OPTION_VAL_270_CY },
         { "280", OPTION_VAL_280_CY },
         { "290", OPTION_VAL_290_CY },
         { "300", OPTION_VAL_300_CY },
         { "310", OPTION_VAL_310_CY },
         { "320", OPTION_VAL_320_CY },
         { "330", OPTION_VAL_330_CY },
         { "340", OPTION_VAL_340_CY },
         { "350", OPTION_VAL_350_CY },
         { "360", OPTION_VAL_360_CY },
         { "370", OPTION_VAL_370_CY },
         { "380", OPTION_VAL_380_CY },
         { "390", OPTION_VAL_390_CY },
         { "400", OPTION_VAL_400_CY },
         { "410", OPTION_VAL_410_CY },
         { "420", OPTION_VAL_420_CY },
         { "430", OPTION_VAL_430_CY },
         { "440", OPTION_VAL_440_CY },
         { "450", OPTION_VAL_450_CY },
         { "460", OPTION_VAL_460_CY },
         { "470", OPTION_VAL_470_CY },
         { "480", OPTION_VAL_480_CY },
         { "490", OPTION_VAL_490_CY },
         { "500", OPTION_VAL_500_CY },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_CY,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CY },
         { "5%",  OPTION_VAL_5_CY },
         { "10%", OPTION_VAL_10_CY },
         { "15%", OPTION_VAL_15_CY },
         { "20%", OPTION_VAL_20_CY },
         { "25%", OPTION_VAL_25_CY },
         { "30%", OPTION_VAL_30_CY },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_CY,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_CY },
         { "5%",  OPTION_VAL_5_CY },
         { "10%", OPTION_VAL_10_CY },
         { "15%", OPTION_VAL_15_CY },
         { "20%", OPTION_VAL_20_CY },
         { "25%", OPTION_VAL_25_CY },
         { "30%", OPTION_VAL_30_CY },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_CY,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_CY,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_CY,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_CY,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_CY },
         { "60%",  OPTION_VAL_60_CY },
         { "70%",  OPTION_VAL_70_CY },
         { "80%",  OPTION_VAL_80_CY },
         { "90%",  OPTION_VAL_90_CY },
         { "100%", OPTION_VAL_100_O47_CY },
         { "110%", OPTION_VAL_110_O47_CY },
         { "120%", OPTION_VAL_120_O47_CY },
         { "130%", OPTION_VAL_130_O47_CY },
         { "140%", OPTION_VAL_140_O47_CY },
         { "150%", OPTION_VAL_150_O47_CY },
         { "160%", OPTION_VAL_160_O47_CY },
         { "170%", OPTION_VAL_170_O47_CY },
         { "180%", OPTION_VAL_180_O47_CY },
         { "190%", OPTION_VAL_190_O47_CY },
         { "200%", OPTION_VAL_200_O47_CY },
         { "210%", OPTION_VAL_210_O47_CY },
         { "220%", OPTION_VAL_220_O47_CY },
         { "230%", OPTION_VAL_230_O47_CY },
         { "240%", OPTION_VAL_240_O47_CY },
         { "250%", OPTION_VAL_250_O47_CY },
         { "260%", OPTION_VAL_260_O47_CY },
         { "270%", OPTION_VAL_270_O47_CY },
         { "280%", OPTION_VAL_280_O47_CY },
         { "290%", OPTION_VAL_290_O47_CY },
         { "300%", OPTION_VAL_300_O47_CY },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_CY,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CY },
         { "Red",      OPTION_VAL_RED_CY },
         { "Green",    OPTION_VAL_GREEN_CY },
         { "Blue",     OPTION_VAL_BLUE_CY },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_CY,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CY },
         { "Red",      OPTION_VAL_RED_CY },
         { "Green",    OPTION_VAL_GREEN_CY },
         { "Blue",     OPTION_VAL_BLUE_CY },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_CY,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CY },
         { "Red",      OPTION_VAL_RED_CY },
         { "Green",    OPTION_VAL_GREEN_CY },
         { "Blue",     OPTION_VAL_BLUE_CY },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_CY,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_CY },
         { "Red",      OPTION_VAL_RED_CY },
         { "Green",    OPTION_VAL_GREEN_CY },
         { "Blue",     OPTION_VAL_BLUE_CY },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_CY,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CY },
         { "Purupuru", OPTION_VAL_PURUPURU_CY },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CY },
         { "None",     OPTION_VAL_NONE_CY },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_CY,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CY },
         { "Purupuru", OPTION_VAL_PURUPURU_CY },
         { "None",     OPTION_VAL_NONE_CY },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_CY,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CY },
         { "Purupuru", OPTION_VAL_PURUPURU_CY },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CY },
         { "None",     OPTION_VAL_NONE_CY },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_CY,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CY },
         { "Purupuru", OPTION_VAL_PURUPURU_CY },
         { "None",     OPTION_VAL_NONE_CY },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_CY,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CY },
         { "Purupuru", OPTION_VAL_PURUPURU_CY },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CY },
         { "None",     OPTION_VAL_NONE_CY },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_CY,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CY },
         { "Purupuru", OPTION_VAL_PURUPURU_CY },
         { "None",     OPTION_VAL_NONE_CY },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_CY,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CY },
         { "Purupuru", OPTION_VAL_PURUPURU_CY },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_CY },
         { "None",     OPTION_VAL_NONE_CY },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_CY,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_CY },
         { "Purupuru", OPTION_VAL_PURUPURU_CY },
         { "None",     OPTION_VAL_NONE_CY },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CY,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_CY,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_CY,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_CY },
         { "All VMUs", OPTION_VAL_ALL_VMUS_CY },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CY,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_CY,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_CY,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_CY,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_CY,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CY,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_CY,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_CY,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CY },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CY },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CY },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CY },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CY },
         { "2x", OPTION_VAL_2_O26_CY },
         { "3x", OPTION_VAL_3X_CY },
         { "4x", OPTION_VAL_4_CY },
         { "5x", OPTION_VAL_5X_CY },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CY },
         { "20%",  OPTION_VAL_20_CY },
         { "30%",  OPTION_VAL_30_CY },
         { "40%",  OPTION_VAL_40_CY },
         { "50%",  OPTION_VAL_50_CY },
         { "60%",  OPTION_VAL_60_CY },
         { "70%",  OPTION_VAL_70_CY },
         { "80%",  OPTION_VAL_80_CY },
         { "90%",  OPTION_VAL_90_CY },
         { "100%", OPTION_VAL_100_O47_CY },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CY },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CY },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CY },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CY },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CY },
         { "2x", OPTION_VAL_2_O26_CY },
         { "3x", OPTION_VAL_3X_CY },
         { "4x", OPTION_VAL_4_CY },
         { "5x", OPTION_VAL_5X_CY },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CY },
         { "20%",  OPTION_VAL_20_CY },
         { "30%",  OPTION_VAL_30_CY },
         { "40%",  OPTION_VAL_40_CY },
         { "50%",  OPTION_VAL_50_CY },
         { "60%",  OPTION_VAL_60_CY },
         { "70%",  OPTION_VAL_70_CY },
         { "80%",  OPTION_VAL_80_CY },
         { "90%",  OPTION_VAL_90_CY },
         { "100%", OPTION_VAL_100_O47_CY },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CY },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CY },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CY },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CY },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CY },
         { "2x", OPTION_VAL_2_O26_CY },
         { "3x", OPTION_VAL_3X_CY },
         { "4x", OPTION_VAL_4_CY },
         { "5x", OPTION_VAL_5X_CY },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CY },
         { "20%",  OPTION_VAL_20_CY },
         { "30%",  OPTION_VAL_30_CY },
         { "40%",  OPTION_VAL_40_CY },
         { "50%",  OPTION_VAL_50_CY },
         { "60%",  OPTION_VAL_60_CY },
         { "70%",  OPTION_VAL_70_CY },
         { "80%",  OPTION_VAL_80_CY },
         { "90%",  OPTION_VAL_90_CY },
         { "100%", OPTION_VAL_100_O47_CY },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_CY },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_CY },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_CY },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_CY },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_CY },
         { "2x", OPTION_VAL_2_O26_CY },
         { "3x", OPTION_VAL_3X_CY },
         { "4x", OPTION_VAL_4_CY },
         { "5x", OPTION_VAL_5X_CY },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_CY },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_CY },
         { "BLACK 02",          OPTION_VAL_BLACK_02_CY },
         { "BLUE 03",           OPTION_VAL_BLUE_CY },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_CY },
         { "GREEN 05",          OPTION_VAL_GREEN_CY },
         { "CYAN 06",           OPTION_VAL_CYAN_06_CY },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_CY },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_CY },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_CY },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_CY },
         { "RED 11",            OPTION_VAL_RED_CY },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_CY },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_CY },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_CY },
         { "GRAY 15",           OPTION_VAL_GRAY_15_CY },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_CY },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_CY },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_CY },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_CY },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_CY },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_CY },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_CY },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_CY },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_CY },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_CY },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_CY },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_CY },
         { "WHITE 28",          OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_CY,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_CY },
         { "20%",  OPTION_VAL_20_CY },
         { "30%",  OPTION_VAL_30_CY },
         { "40%",  OPTION_VAL_40_CY },
         { "50%",  OPTION_VAL_50_CY },
         { "60%",  OPTION_VAL_60_CY },
         { "70%",  OPTION_VAL_70_CY },
         { "80%",  OPTION_VAL_80_CY },
         { "90%",  OPTION_VAL_90_CY },
         { "100%", OPTION_VAL_100_O47_CY },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_cy = {
   option_cats_cy,
   option_defs_cy
};

/* RETRO_LANGUAGE_DA */

#define CATEGORY_SYSTEM_LABEL_DA NULL
#define CATEGORY_SYSTEM_INFO_0_DA NULL
#define CATEGORY_VIDEO_LABEL_DA NULL
#define CATEGORY_VIDEO_INFO_0_DA NULL
#define CATEGORY_PERFORMANCE_LABEL_DA NULL
#define CATEGORY_PERFORMANCE_INFO_0_DA NULL
#define CATEGORY_HACKS_LABEL_DA NULL
#define CATEGORY_HACKS_INFO_0_DA NULL
#define CATEGORY_INPUT_LABEL_DA NULL
#define CATEGORY_INPUT_INFO_0_DA NULL
#define CATEGORY_EXPANSIONS_LABEL_DA NULL
#define CATEGORY_EXPANSIONS_INFO_0_DA NULL
#define CATEGORY_VMU_LABEL_DA NULL
#define CATEGORY_VMU_INFO_0_DA NULL
#define CORE_OPTION_NAME_REGION_LABEL_DA NULL
#define OPTION_VAL_JAPAN_DA NULL
#define OPTION_VAL_USA_DA NULL
#define OPTION_VAL_EUROPE_DA NULL
#define OPTION_VAL_DEFAULT_DA NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_DA NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_DA NULL
#define OPTION_VAL_JAPANESE_DA NULL
#define OPTION_VAL_ENGLISH_DA NULL
#define OPTION_VAL_GERMAN_DA NULL
#define OPTION_VAL_FRENCH_DA NULL
#define OPTION_VAL_SPANISH_DA NULL
#define OPTION_VAL_ITALIAN_DA NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_DA NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_DA NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_DA NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_DA NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_DA NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_DA NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_DA NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_DA NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_DA NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_DA NULL
#define CORE_OPTION_NAME_UPNP_LABEL_DA NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_DA NULL
#define CORE_OPTION_NAME_DCNET_LABEL_DA NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_DA NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_DA NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_DA NULL
#define OPTION_VAL_320X240_DA NULL
#define OPTION_VAL_640X480_DA NULL
#define OPTION_VAL_800X600_DA NULL
#define OPTION_VAL_960X720_DA NULL
#define OPTION_VAL_1024X768_DA NULL
#define OPTION_VAL_1280X960_DA NULL
#define OPTION_VAL_1440X1080_DA NULL
#define OPTION_VAL_1600X1200_DA NULL
#define OPTION_VAL_1920X1440_DA NULL
#define OPTION_VAL_2560X1920_DA NULL
#define OPTION_VAL_2880X2160_DA NULL
#define OPTION_VAL_3200X2400_DA NULL
#define OPTION_VAL_3840X2880_DA NULL
#define OPTION_VAL_4480X3360_DA NULL
#define OPTION_VAL_5120X3840_DA NULL
#define OPTION_VAL_5760X4320_DA NULL
#define OPTION_VAL_6400X4800_DA NULL
#define OPTION_VAL_7040X5280_DA NULL
#define OPTION_VAL_7680X5760_DA NULL
#define OPTION_VAL_8320X6240_DA NULL
#define OPTION_VAL_8960X6720_DA NULL
#define OPTION_VAL_9600X7200_DA NULL
#define OPTION_VAL_10240X7680_DA NULL
#define OPTION_VAL_10880X8160_DA NULL
#define OPTION_VAL_11520X8640_DA NULL
#define OPTION_VAL_12160X9120_DA NULL
#define OPTION_VAL_12800X9600_DA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_DA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_DA NULL
#define OPTION_VAL_VGA_DA NULL
#define OPTION_VAL_TV_RGB_DA NULL
#define OPTION_VAL_TV_COMPOSITE_DA NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_DA NULL
#define OPTION_VAL_NTSC_DA NULL
#define OPTION_VAL_PAL_DA NULL
#define OPTION_VAL_PAL_N_DA NULL
#define OPTION_VAL_PAL_M_DA NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_DA NULL
#define OPTION_VAL_HORIZONTAL_DA NULL
#define OPTION_VAL_VERTICAL_DA NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_DA NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_DA NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_DA NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_DA NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_DA NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_DA NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_DA NULL
#define OPTION_VAL_512MB_DA NULL
#define OPTION_VAL_1GB_DA NULL
#define OPTION_VAL_2GB_DA NULL
#define OPTION_VAL_4GB_DA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_DA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_DA NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_DA NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_DA NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_DA NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_DA NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_DA NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_DA NULL
#define CORE_OPTION_NAME_FOG_LABEL_DA NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_DA NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_DA NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_DA NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_DA NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_DA NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_DA NULL
#define OPTION_VAL_1_DA NULL
#define OPTION_VAL_2_DA NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_DA NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_DA NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_DA NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_DA NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_DA NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_DA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_DA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_DA NULL
#define OPTION_VAL_2_O26_DA NULL
#define OPTION_VAL_4_DA NULL
#define OPTION_VAL_6_DA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_DA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_DA NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_DA NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_DA NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_DA NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_DA NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_DA NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_DA NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_DA NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_DA NULL
#define OPTION_VAL_SOME_DA NULL
#define OPTION_VAL_MORE_DA NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_DA NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_DA NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_DA NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_DA NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_DA NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_DA NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_DA NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_DA NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_DA NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_DA NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_DA NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_DA NULL
#define OPTION_VAL_100_DA NULL
#define OPTION_VAL_110_DA NULL
#define OPTION_VAL_120_DA NULL
#define OPTION_VAL_130_DA NULL
#define OPTION_VAL_140_DA NULL
#define OPTION_VAL_150_DA NULL
#define OPTION_VAL_160_DA NULL
#define OPTION_VAL_170_DA NULL
#define OPTION_VAL_180_DA NULL
#define OPTION_VAL_190_DA NULL
#define OPTION_VAL_200_DA NULL
#define OPTION_VAL_210_DA NULL
#define OPTION_VAL_220_DA NULL
#define OPTION_VAL_230_DA NULL
#define OPTION_VAL_240_DA NULL
#define OPTION_VAL_250_DA NULL
#define OPTION_VAL_260_DA NULL
#define OPTION_VAL_270_DA NULL
#define OPTION_VAL_280_DA NULL
#define OPTION_VAL_290_DA NULL
#define OPTION_VAL_300_DA NULL
#define OPTION_VAL_310_DA NULL
#define OPTION_VAL_320_DA NULL
#define OPTION_VAL_330_DA NULL
#define OPTION_VAL_340_DA NULL
#define OPTION_VAL_350_DA NULL
#define OPTION_VAL_360_DA NULL
#define OPTION_VAL_370_DA NULL
#define OPTION_VAL_380_DA NULL
#define OPTION_VAL_390_DA NULL
#define OPTION_VAL_400_DA NULL
#define OPTION_VAL_410_DA NULL
#define OPTION_VAL_420_DA NULL
#define OPTION_VAL_430_DA NULL
#define OPTION_VAL_440_DA NULL
#define OPTION_VAL_450_DA NULL
#define OPTION_VAL_460_DA NULL
#define OPTION_VAL_470_DA NULL
#define OPTION_VAL_480_DA NULL
#define OPTION_VAL_490_DA NULL
#define OPTION_VAL_500_DA NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_DA NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_DA NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_DA NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_DA NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_DA NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_DA NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_DA NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_DA NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_DA NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_DA NULL
#define OPTION_VAL_0_DA NULL
#define OPTION_VAL_5_DA NULL
#define OPTION_VAL_10_DA NULL
#define OPTION_VAL_15_DA NULL
#define OPTION_VAL_20_DA NULL
#define OPTION_VAL_25_DA NULL
#define OPTION_VAL_30_DA NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_DA NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_DA NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_DA NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_DA NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_DA NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_DA NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_DA NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_DA NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_DA NULL
#define OPTION_VAL_50_DA NULL
#define OPTION_VAL_60_DA NULL
#define OPTION_VAL_70_DA NULL
#define OPTION_VAL_80_DA NULL
#define OPTION_VAL_90_DA NULL
#define OPTION_VAL_100_O47_DA NULL
#define OPTION_VAL_110_O47_DA NULL
#define OPTION_VAL_120_O47_DA NULL
#define OPTION_VAL_130_O47_DA NULL
#define OPTION_VAL_140_O47_DA NULL
#define OPTION_VAL_150_O47_DA NULL
#define OPTION_VAL_160_O47_DA NULL
#define OPTION_VAL_170_O47_DA NULL
#define OPTION_VAL_180_O47_DA NULL
#define OPTION_VAL_190_O47_DA NULL
#define OPTION_VAL_200_O47_DA NULL
#define OPTION_VAL_210_O47_DA NULL
#define OPTION_VAL_220_O47_DA NULL
#define OPTION_VAL_230_O47_DA NULL
#define OPTION_VAL_240_O47_DA NULL
#define OPTION_VAL_250_O47_DA NULL
#define OPTION_VAL_260_O47_DA NULL
#define OPTION_VAL_270_O47_DA NULL
#define OPTION_VAL_280_O47_DA NULL
#define OPTION_VAL_290_O47_DA NULL
#define OPTION_VAL_300_O47_DA NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_DA NULL
#define OPTION_VAL_WHITE_DA NULL
#define OPTION_VAL_RED_DA NULL
#define OPTION_VAL_GREEN_DA NULL
#define OPTION_VAL_BLUE_DA NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_DA NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_DA NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_DA NULL
#define OPTION_VAL_VMU_DA NULL
#define OPTION_VAL_PURUPURU_DA NULL
#define OPTION_VAL_DREAMPOTATO_DA NULL
#define OPTION_VAL_NONE_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_DA NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_DA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_DA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_DA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_DA NULL
#define OPTION_VAL_VMU_A1_DA NULL
#define OPTION_VAL_ALL_VMUS_DA NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_DA NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_DA NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_DA NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_DA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_DA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_DA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_DA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_DA NULL
#define OPTION_VAL_UPPER_LEFT_DA NULL
#define OPTION_VAL_UPPER_RIGHT_DA NULL
#define OPTION_VAL_LOWER_LEFT_DA NULL
#define OPTION_VAL_LOWER_RIGHT_DA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_DA NULL
#define OPTION_VAL_1X_DA NULL
#define OPTION_VAL_3X_DA NULL
#define OPTION_VAL_5X_DA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_DA NULL
#define OPTION_VAL_DEFAULT_ON_00_DA NULL
#define OPTION_VAL_DEFAULT_OFF_01_DA NULL
#define OPTION_VAL_BLACK_02_DA NULL
#define OPTION_VAL_LIGHT_BLUE_04_DA NULL
#define OPTION_VAL_CYAN_06_DA NULL
#define OPTION_VAL_CYAN_BLUE_07_DA NULL
#define OPTION_VAL_LIGHT_GREEN_08_DA NULL
#define OPTION_VAL_CYAN_GREEN_09_DA NULL
#define OPTION_VAL_LIGHT_CYAN_10_DA NULL
#define OPTION_VAL_PURPLE_12_DA NULL
#define OPTION_VAL_LIGHT_PURPLE_13_DA NULL
#define OPTION_VAL_YELLOW_14_DA NULL
#define OPTION_VAL_GRAY_15_DA NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_DA NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_DA NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_DA NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_DA NULL
#define OPTION_VAL_LIGHT_RED_2_20_DA NULL
#define OPTION_VAL_MAGENTA_21_DA NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_DA NULL
#define OPTION_VAL_LIGHT_ORANGE_23_DA NULL
#define OPTION_VAL_ORANGE_24_DA NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_DA NULL
#define OPTION_VAL_LIGHT_YELLOW_26_DA NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_DA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_DA NULL
#define OPTION_VAL_40_DA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_DA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_DA NULL

struct retro_core_option_v2_category option_cats_da[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_DA,
      CATEGORY_SYSTEM_INFO_0_DA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_DA,
      CATEGORY_VIDEO_INFO_0_DA
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_DA,
      CATEGORY_PERFORMANCE_INFO_0_DA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_DA,
      CATEGORY_HACKS_INFO_0_DA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_DA,
      CATEGORY_INPUT_INFO_0_DA
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_DA,
      CATEGORY_EXPANSIONS_INFO_0_DA
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_DA,
      CATEGORY_VMU_INFO_0_DA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_da[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_DA,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_DA },
         { "USA",     OPTION_VAL_USA_DA },
         { "Europe",  OPTION_VAL_EUROPE_DA },
         { "Default", OPTION_VAL_DEFAULT_DA },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_DA,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_DA },
         { "English",  OPTION_VAL_ENGLISH_DA },
         { "German",   OPTION_VAL_GERMAN_DA },
         { "French",   OPTION_VAL_FRENCH_DA },
         { "Spanish",  OPTION_VAL_SPANISH_DA },
         { "Italian",  OPTION_VAL_ITALIAN_DA },
         { "Default",  OPTION_VAL_DEFAULT_DA },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_DA,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_DA },
         { "640x480",    OPTION_VAL_640X480_DA },
         { "800x600",    OPTION_VAL_800X600_DA },
         { "960x720",    OPTION_VAL_960X720_DA },
         { "1024x768",   OPTION_VAL_1024X768_DA },
         { "1280x960",   OPTION_VAL_1280X960_DA },
         { "1440x1080",  OPTION_VAL_1440X1080_DA },
         { "1600x1200",  OPTION_VAL_1600X1200_DA },
         { "1920x1440",  OPTION_VAL_1920X1440_DA },
         { "2560x1920",  OPTION_VAL_2560X1920_DA },
         { "2880x2160",  OPTION_VAL_2880X2160_DA },
         { "3200x2400",  OPTION_VAL_3200X2400_DA },
         { "3840x2880",  OPTION_VAL_3840X2880_DA },
         { "4480x3360",  OPTION_VAL_4480X3360_DA },
         { "5120x3840",  OPTION_VAL_5120X3840_DA },
         { "5760x4320",  OPTION_VAL_5760X4320_DA },
         { "6400x4800",  OPTION_VAL_6400X4800_DA },
         { "7040x5280",  OPTION_VAL_7040X5280_DA },
         { "7680x5760",  OPTION_VAL_7680X5760_DA },
         { "8320x6240",  OPTION_VAL_8320X6240_DA },
         { "8960x6720",  OPTION_VAL_8960X6720_DA },
         { "9600x7200",  OPTION_VAL_9600X7200_DA },
         { "10240x7680", OPTION_VAL_10240X7680_DA },
         { "10880x8160", OPTION_VAL_10880X8160_DA },
         { "11520x8640", OPTION_VAL_11520X8640_DA },
         { "12160x9120", OPTION_VAL_12160X9120_DA },
         { "12800x9600", OPTION_VAL_12800X9600_DA },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_DA,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_DA },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_DA },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_DA },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_DA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_DA },
         { "PAL",     OPTION_VAL_PAL_DA },
         { "PAL_N",   OPTION_VAL_PAL_N_DA },
         { "PAL_M",   OPTION_VAL_PAL_M_DA },
         { "Default", OPTION_VAL_DEFAULT_DA },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_DA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_DA },
         { "vertical",   OPTION_VAL_VERTICAL_DA },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_DA,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_DA },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_DA },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_DA },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_DA,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_DA },
         { "1GB",   OPTION_VAL_1GB_DA },
         { "2GB",   OPTION_VAL_2GB_DA },
         { "4GB",   OPTION_VAL_4GB_DA },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_DA,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_DA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_DA,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_DA,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_DA },
         { "1",  OPTION_VAL_1_DA },
         { "2",  OPTION_VAL_2_DA },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_DA,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_DA },
         { "4",  OPTION_VAL_4_DA },
         { "6",  OPTION_VAL_6_DA },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_DA,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_DA,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_DA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_DA,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_DA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_DA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_DA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_DA },
         { "more", OPTION_VAL_MORE_DA },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_DA,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_DA },
         { "110", OPTION_VAL_110_DA },
         { "120", OPTION_VAL_120_DA },
         { "130", OPTION_VAL_130_DA },
         { "140", OPTION_VAL_140_DA },
         { "150", OPTION_VAL_150_DA },
         { "160", OPTION_VAL_160_DA },
         { "170", OPTION_VAL_170_DA },
         { "180", OPTION_VAL_180_DA },
         { "190", OPTION_VAL_190_DA },
         { "200", OPTION_VAL_200_DA },
         { "210", OPTION_VAL_210_DA },
         { "220", OPTION_VAL_220_DA },
         { "230", OPTION_VAL_230_DA },
         { "240", OPTION_VAL_240_DA },
         { "250", OPTION_VAL_250_DA },
         { "260", OPTION_VAL_260_DA },
         { "270", OPTION_VAL_270_DA },
         { "280", OPTION_VAL_280_DA },
         { "290", OPTION_VAL_290_DA },
         { "300", OPTION_VAL_300_DA },
         { "310", OPTION_VAL_310_DA },
         { "320", OPTION_VAL_320_DA },
         { "330", OPTION_VAL_330_DA },
         { "340", OPTION_VAL_340_DA },
         { "350", OPTION_VAL_350_DA },
         { "360", OPTION_VAL_360_DA },
         { "370", OPTION_VAL_370_DA },
         { "380", OPTION_VAL_380_DA },
         { "390", OPTION_VAL_390_DA },
         { "400", OPTION_VAL_400_DA },
         { "410", OPTION_VAL_410_DA },
         { "420", OPTION_VAL_420_DA },
         { "430", OPTION_VAL_430_DA },
         { "440", OPTION_VAL_440_DA },
         { "450", OPTION_VAL_450_DA },
         { "460", OPTION_VAL_460_DA },
         { "470", OPTION_VAL_470_DA },
         { "480", OPTION_VAL_480_DA },
         { "490", OPTION_VAL_490_DA },
         { "500", OPTION_VAL_500_DA },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_DA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_DA },
         { "5%",  OPTION_VAL_5_DA },
         { "10%", OPTION_VAL_10_DA },
         { "15%", OPTION_VAL_15_DA },
         { "20%", OPTION_VAL_20_DA },
         { "25%", OPTION_VAL_25_DA },
         { "30%", OPTION_VAL_30_DA },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_DA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_DA },
         { "5%",  OPTION_VAL_5_DA },
         { "10%", OPTION_VAL_10_DA },
         { "15%", OPTION_VAL_15_DA },
         { "20%", OPTION_VAL_20_DA },
         { "25%", OPTION_VAL_25_DA },
         { "30%", OPTION_VAL_30_DA },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_DA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_DA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_DA,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_DA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_DA },
         { "60%",  OPTION_VAL_60_DA },
         { "70%",  OPTION_VAL_70_DA },
         { "80%",  OPTION_VAL_80_DA },
         { "90%",  OPTION_VAL_90_DA },
         { "100%", OPTION_VAL_100_O47_DA },
         { "110%", OPTION_VAL_110_O47_DA },
         { "120%", OPTION_VAL_120_O47_DA },
         { "130%", OPTION_VAL_130_O47_DA },
         { "140%", OPTION_VAL_140_O47_DA },
         { "150%", OPTION_VAL_150_O47_DA },
         { "160%", OPTION_VAL_160_O47_DA },
         { "170%", OPTION_VAL_170_O47_DA },
         { "180%", OPTION_VAL_180_O47_DA },
         { "190%", OPTION_VAL_190_O47_DA },
         { "200%", OPTION_VAL_200_O47_DA },
         { "210%", OPTION_VAL_210_O47_DA },
         { "220%", OPTION_VAL_220_O47_DA },
         { "230%", OPTION_VAL_230_O47_DA },
         { "240%", OPTION_VAL_240_O47_DA },
         { "250%", OPTION_VAL_250_O47_DA },
         { "260%", OPTION_VAL_260_O47_DA },
         { "270%", OPTION_VAL_270_O47_DA },
         { "280%", OPTION_VAL_280_O47_DA },
         { "290%", OPTION_VAL_290_O47_DA },
         { "300%", OPTION_VAL_300_O47_DA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_DA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DA },
         { "Red",      OPTION_VAL_RED_DA },
         { "Green",    OPTION_VAL_GREEN_DA },
         { "Blue",     OPTION_VAL_BLUE_DA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_DA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DA },
         { "Red",      OPTION_VAL_RED_DA },
         { "Green",    OPTION_VAL_GREEN_DA },
         { "Blue",     OPTION_VAL_BLUE_DA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_DA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DA },
         { "Red",      OPTION_VAL_RED_DA },
         { "Green",    OPTION_VAL_GREEN_DA },
         { "Blue",     OPTION_VAL_BLUE_DA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_DA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DA },
         { "Red",      OPTION_VAL_RED_DA },
         { "Green",    OPTION_VAL_GREEN_DA },
         { "Blue",     OPTION_VAL_BLUE_DA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_DA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DA },
         { "Purupuru", OPTION_VAL_PURUPURU_DA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_DA },
         { "None",     OPTION_VAL_NONE_DA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_DA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DA },
         { "Purupuru", OPTION_VAL_PURUPURU_DA },
         { "None",     OPTION_VAL_NONE_DA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_DA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DA },
         { "Purupuru", OPTION_VAL_PURUPURU_DA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_DA },
         { "None",     OPTION_VAL_NONE_DA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_DA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DA },
         { "Purupuru", OPTION_VAL_PURUPURU_DA },
         { "None",     OPTION_VAL_NONE_DA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_DA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DA },
         { "Purupuru", OPTION_VAL_PURUPURU_DA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_DA },
         { "None",     OPTION_VAL_NONE_DA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_DA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DA },
         { "Purupuru", OPTION_VAL_PURUPURU_DA },
         { "None",     OPTION_VAL_NONE_DA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_DA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DA },
         { "Purupuru", OPTION_VAL_PURUPURU_DA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_DA },
         { "None",     OPTION_VAL_NONE_DA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_DA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DA },
         { "Purupuru", OPTION_VAL_PURUPURU_DA },
         { "None",     OPTION_VAL_NONE_DA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_DA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_DA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_DA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_DA },
         { "All VMUs", OPTION_VAL_ALL_VMUS_DA },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_DA,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_DA,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_DA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_DA,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_DA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_DA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_DA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_DA,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DA },
         { "2x", OPTION_VAL_2_O26_DA },
         { "3x", OPTION_VAL_3X_DA },
         { "4x", OPTION_VAL_4_DA },
         { "5x", OPTION_VAL_5X_DA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DA },
         { "20%",  OPTION_VAL_20_DA },
         { "30%",  OPTION_VAL_30_DA },
         { "40%",  OPTION_VAL_40_DA },
         { "50%",  OPTION_VAL_50_DA },
         { "60%",  OPTION_VAL_60_DA },
         { "70%",  OPTION_VAL_70_DA },
         { "80%",  OPTION_VAL_80_DA },
         { "90%",  OPTION_VAL_90_DA },
         { "100%", OPTION_VAL_100_O47_DA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DA },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DA },
         { "2x", OPTION_VAL_2_O26_DA },
         { "3x", OPTION_VAL_3X_DA },
         { "4x", OPTION_VAL_4_DA },
         { "5x", OPTION_VAL_5X_DA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DA },
         { "20%",  OPTION_VAL_20_DA },
         { "30%",  OPTION_VAL_30_DA },
         { "40%",  OPTION_VAL_40_DA },
         { "50%",  OPTION_VAL_50_DA },
         { "60%",  OPTION_VAL_60_DA },
         { "70%",  OPTION_VAL_70_DA },
         { "80%",  OPTION_VAL_80_DA },
         { "90%",  OPTION_VAL_90_DA },
         { "100%", OPTION_VAL_100_O47_DA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DA },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DA },
         { "2x", OPTION_VAL_2_O26_DA },
         { "3x", OPTION_VAL_3X_DA },
         { "4x", OPTION_VAL_4_DA },
         { "5x", OPTION_VAL_5X_DA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DA },
         { "20%",  OPTION_VAL_20_DA },
         { "30%",  OPTION_VAL_30_DA },
         { "40%",  OPTION_VAL_40_DA },
         { "50%",  OPTION_VAL_50_DA },
         { "60%",  OPTION_VAL_60_DA },
         { "70%",  OPTION_VAL_70_DA },
         { "80%",  OPTION_VAL_80_DA },
         { "90%",  OPTION_VAL_90_DA },
         { "100%", OPTION_VAL_100_O47_DA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DA },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DA },
         { "2x", OPTION_VAL_2_O26_DA },
         { "3x", OPTION_VAL_3X_DA },
         { "4x", OPTION_VAL_4_DA },
         { "5x", OPTION_VAL_5X_DA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DA },
         { "BLUE 03",           OPTION_VAL_BLUE_DA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DA },
         { "GREEN 05",          OPTION_VAL_GREEN_DA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DA },
         { "RED 11",            OPTION_VAL_RED_DA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DA },
         { "WHITE 28",          OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_DA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DA },
         { "20%",  OPTION_VAL_20_DA },
         { "30%",  OPTION_VAL_30_DA },
         { "40%",  OPTION_VAL_40_DA },
         { "50%",  OPTION_VAL_50_DA },
         { "60%",  OPTION_VAL_60_DA },
         { "70%",  OPTION_VAL_70_DA },
         { "80%",  OPTION_VAL_80_DA },
         { "90%",  OPTION_VAL_90_DA },
         { "100%", OPTION_VAL_100_O47_DA },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_da = {
   option_cats_da,
   option_defs_da
};

/* RETRO_LANGUAGE_DE */

#define CATEGORY_SYSTEM_LABEL_DE NULL
#define CATEGORY_SYSTEM_INFO_0_DE "Region, Sprache, BIOS und Basis-Hardware-Einstellungen konfigurieren."
#define CATEGORY_VIDEO_LABEL_DE NULL
#define CATEGORY_VIDEO_INFO_0_DE "Einstellungen für Auflösung, reihenfolgeunabhängige Transparenz und visuelle Effekte konfigurieren."
#define CATEGORY_PERFORMANCE_LABEL_DE "Leistung"
#define CATEGORY_PERFORMANCE_INFO_0_DE "Einstellungen für mehrgängiges Rendern und Bilder überspringen ändern."
#define CATEGORY_HACKS_LABEL_DE "Emulations-Hacks"
#define CATEGORY_HACKS_INFO_0_DE "Breitbildüberschreibungen, GD-ROM-Ladegeschwindigkeit und Texturaustauscheinstellungen konfigurieren."
#define CATEGORY_INPUT_LABEL_DE "Eingabe"
#define CATEGORY_INPUT_INFO_0_DE "Gamepad- und Lightgun-Einstellungen konfigurieren."
#define CATEGORY_EXPANSIONS_LABEL_DE "Controllererweiterungsslots"
#define CATEGORY_EXPANSIONS_INFO_0_DE "Das Gerät (VMU, Rumpelgerät) auswählen, das am jeweiligen Controllererweiterungsslot angeschlossen ist."
#define CATEGORY_VMU_LABEL_DE NULL
#define CATEGORY_VMU_INFO_0_DE "Konfiguriert die VMU-Speicherdateien pro Spiel und die VMU-Sichtbarkeitseinstellungen auf dem Bildschirm."
#define CORE_OPTION_NAME_REGION_LABEL_DE NULL
#define OPTION_VAL_JAPAN_DE NULL
#define OPTION_VAL_USA_DE NULL
#define OPTION_VAL_EUROPE_DE "Europa"
#define OPTION_VAL_DEFAULT_DE "Standard"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_DE "Sprache"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_DE "Ändert die Sprache, die vom BIOS und von Spielen verwendet wird, die mehrere Sprachen enthalten."
#define OPTION_VAL_JAPANESE_DE "Japanisch"
#define OPTION_VAL_ENGLISH_DE "Englisch"
#define OPTION_VAL_GERMAN_DE "Deutsch"
#define OPTION_VAL_FRENCH_DE "Französisch"
#define OPTION_VAL_SPANISH_DE "Spanisch"
#define OPTION_VAL_ITALIAN_DE "Italienisch"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_DE "HLE BIOS (Neustart erforderlich)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_DE "Erzwingt die Verwendung von High-Level-Emulation-BIOS."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_DE "DSP aktivieren"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_DE "Aktiviert Emulation von DSP des Dreamcasts (digitaler Signalprozessor). Verbessert die Genauigkeit des erzeugten Klanges, erhöht aber die Leistungsanforderungen."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_DE "Arcade-Servicetasten erlauben"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_DE "Aktiviert den SERVICE-Knopf für Arcadespiele, um die Cabineteinstellungen aufzurufen."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_DE "NAOMI-Spiele auf kostenloses Spiel setzen"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_DE "Ändert die Münzeinstellungen des Spiels zu kostenlos."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_DE "Breitbandadapter-Emulation"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_DE "Den Ethernet-Breitbandadapter anstelle des Modems emulieren. (Neustart erforderlich)"
#define CORE_OPTION_NAME_UPNP_LABEL_DE "UPnP aktivieren"
#define CORE_OPTION_NAME_UPNP_INFO_0_DE "Verwendet UPnP, um den Internetrouter automatisch für Onlinespiele zu konfigurieren."
#define CORE_OPTION_NAME_DCNET_LABEL_DE "DCNet verwenden"
#define CORE_OPTION_NAME_DCNET_INFO_0_DE "Den DCNet-Cloud-Service für den Dreamcast-Internetzugang nutzen."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_DE "Interne Auflösung"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_DE "Rendering-Auflösung ändern."
#define OPTION_VAL_320X240_DE "320×240 (Halb)"
#define OPTION_VAL_640X480_DE "640×480 (Nativ)"
#define OPTION_VAL_800X600_DE "800×600 (×1,25)"
#define OPTION_VAL_960X720_DE "960×720 (×1,5)"
#define OPTION_VAL_1024X768_DE "1024×768 (×1,6)"
#define OPTION_VAL_1280X960_DE "1280×960 (×2)"
#define OPTION_VAL_1440X1080_DE "1440×1080 (×2,25)"
#define OPTION_VAL_1600X1200_DE "1600×1200 (×2,5)"
#define OPTION_VAL_1920X1440_DE "1920×1440 (×3)"
#define OPTION_VAL_2560X1920_DE "2560×1920 (×4)"
#define OPTION_VAL_2880X2160_DE "2880×2160 (×4,5)"
#define OPTION_VAL_3200X2400_DE "3200×2400 (×5)"
#define OPTION_VAL_3840X2880_DE "3840×2880 (×6)"
#define OPTION_VAL_4480X3360_DE "4480×3360 (×7)"
#define OPTION_VAL_5120X3840_DE "5120×3840 (×8)"
#define OPTION_VAL_5760X4320_DE "5760×4320 (×9)"
#define OPTION_VAL_6400X4800_DE "6400×4800 (×10)"
#define OPTION_VAL_7040X5280_DE "7040×5280 (×11)"
#define OPTION_VAL_7680X5760_DE "7680×5760 (×12)"
#define OPTION_VAL_8320X6240_DE "8320×6240 (×13)"
#define OPTION_VAL_8960X6720_DE "8960×6720 (×14)"
#define OPTION_VAL_9600X7200_DE "9600×7200 (×15)"
#define OPTION_VAL_10240X7680_DE "10240×7680 (×6)"
#define OPTION_VAL_10880X8160_DE "10880×8160 (×17)"
#define OPTION_VAL_11520X8640_DE "11520×8640 (×18)"
#define OPTION_VAL_12160X9120_DE "12160×9120 (×19)"
#define OPTION_VAL_12800X9600_DE "12800×9600 (×20)"
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_DE "Kabeltyp"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_DE "Das Ausgabesignalkabel: „TV (Composite)“ ist das am weitesten unterstützte."
#define OPTION_VAL_VGA_DE NULL
#define OPTION_VAL_TV_RGB_DE NULL
#define OPTION_VAL_TV_COMPOSITE_DE NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_DE "Ausgabe-Regions-Standard"
#define OPTION_VAL_NTSC_DE NULL
#define OPTION_VAL_PAL_DE "PAL (Welt)"
#define OPTION_VAL_PAL_N_DE "PAL-N (Argentinien, Paraguay, Uruguay)"
#define OPTION_VAL_PAL_M_DE "PAL-M (Brasilien)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_DE "Bildschirmausrichtung"
#define OPTION_VAL_HORIZONTAL_DE NULL
#define OPTION_VAL_VERTICAL_DE "Vertikal"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_DE "Alpha-Sortierung"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_DE "Wählen, wie die transparenten Polygone sortiert werden."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_DE "Pro-Streifen (schnell, am wenigsten genau)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_DE "Pro-Dreieck (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_DE "Pro-Pixel (genau, aber langsamer)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_DE "Anhäufung Pixel Puffergröße"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_DE "Für eine korrekte Ausgabe höherer Auflösungen sind möglicherweise höhere Werte erforderlich."
#define OPTION_VAL_512MB_DE "512 MB"
#define OPTION_VAL_1GB_DE "1 GB"
#define OPTION_VAL_2GB_DE "2 GB"
#define OPTION_VAL_4GB_DE "4 GB"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_DE "Maximale transparente Ebenen"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_DE "Höhere Werte könnten für komplexe Szenen erforderlich sein."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_DE "Vollständige Bildspeicheremulation"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_DE "Vollständige Bildspeicheremulation in VRAM aktivieren. Dies ist für Spiele nützlich, die den Bildspeicher direkt in VRAM lesen oder schreiben. Wenn aktiviert, wird die interne Auflösung auf 640×480 erzwungen und die Leistung kann stark beeinträchtigt werden."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_DE "RTT-Puffer (Render-To-Texture) erlauben"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_DE "Kopiert gerenderte Texturen zurück von der GPU in VRAM. Diese Option ist normalerweise für Spiele aktiviert, die dies erfordern. Wenn aktiviert, ist das Hochskalieren von Texturen deaktiviert und die Leistung kann beeinträchtigt werden."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_DE NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_DE "Wenn aktivierte Texturen eine kleinere Version von sich selbst verwenden, sofern sie weiter entfernt erscheinen, kann dies die Leistung erhöhen und das Schimmern verringern."
#define CORE_OPTION_NAME_FOG_LABEL_DE "Nebeleffekte"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_DE "Volumen-Modifizierer"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_DE "Eine Dreamcast-GPU-Funktion, die typischerweise von Spielen benutzt wird, um Objektschatten zu zeichnen. Normalerweise sollte dies aktiviert werden – der Leistungsverlust ist in der Regel minimal bis vernachlässigbar."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_DE "Anisotropes Filtern"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_DE "Verbessert die Qualität der Texturen auf Oberflächen, die in schrägen Blickwinkeln gegenüber der Kamera liegen. Höhere Werte sind anspruchsvoller für die GPU. Änderungen an dieser Einstellung gelten erst nach einem Neustart."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_DE "Texturfilterung"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_DE "Der zu verwendende Texturfiltermodus. Dies kann verwendet werden, um einen bestimmten Texturfiltermodus für alle Texturen zu erzwingen, um ein klareres (oder glatteres) Erscheinungsbild als Standard zu erhalten. Andere Werte als Standardwerte können zu verschiedenen Rendering-Problemen führen. Änderungen an dieser Einstellung werden nur nach einem Neustart angewendet."
#define OPTION_VAL_1_DE "Nächste-Nachbarn erzwingen"
#define OPTION_VAL_2_DE "Linear erzwingen"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_DE "Bildaustausch verzögern"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_DE "Nützlich, um flackernden Bildschirm oder fehlerhafte Videos zu vermeiden. Nicht empfohlen auf langsamen Plattformen."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_DE "Änderungen der Bildrate erkennen"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_DE "Benachrichtigt das Frontend, wenn sich die interne Bildrate ändert (z. B. von 60 fps auf 30 fps). Verbessert das Bildtempo in Spielen, die mit gesperrten 30 bps oder 20 bps laufen, sollte aber für Spiele mit entsperrten (instabilen) Bildraten deaktiviert sein (z. B. Ecco the Dolphin, Unreal Tournament). Hinweis: Nicht verfügbar, wenn „Automatisches Bilder überspringen“ aktiviert ist."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_DE "PowerVR2-Nachbearbeitungsfilter"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_DE "Nachbearbeitung des gerenderten Bildes zur Simulation von Effekten speziell für PowerVR2-GPU und analoge Videosignale."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_DE "Texturen hochskalieren (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_DE "Verbessert handgezeichnete 2D-Pixelgrafiken. Sollte nur mit 2D-Pixelspielen verwendet werden."
#define OPTION_VAL_2_O26_DE NULL
#define OPTION_VAL_4_DE "4×"
#define OPTION_VAL_6_DE "6×"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_DE "Texturhochskalierung Max. gefilterte Größe"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_DE "Einen Maximalwert für die Größe einer hochzuskalierenden Textur auswählen. Wenn die Texturgröße größer als der ausgewählte Wert ist, wird sie nicht hochskaliert."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_DE "Native Tiefeninterpolation"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_DE "Hilft bei Texturfehlern und tieferen Problemen mit AMD GPUs. Kann auch Intel GPUs in einigen Fällen helfen."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_DE "Ausfransende Kanten beim Hochskalieren beheben"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_DE "Hilft im Fall von Texture Bleeding beim Hochskalieren. Ein Deaktivieren kann helfen, wenn in 2D-Spielen Pixel beim Hochskalieren verformt werden (MVC2, CVS, KOF usw.)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_DE NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_DE "Führt die GPU und CPU in verschiedenen Prozessen aus. Sehr empfohlen."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_DE "Automatisches Bilder überspringen"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_DE "Überspringt Bilder automatisch, wenn der Emulator langsam läuft. Hinweis: Diese Einstellung gilt nur, wenn 'Threaded Rendering' aktiviert ist."
#define OPTION_VAL_SOME_DE NULL
#define OPTION_VAL_MORE_DE NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_DE "Bilder überspringen"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_DE "Legt die Anzahl der Bilder fest, die zwischen jedem angezeigten Bild übersprungen werden sollen."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_DE "Breitbild-Cheats (Neustart erforderlich)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_DE "Aktiviert Cheats, mit denen bestimmte Spiele im Breitbildformat angezeigt werden können."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_DE "Breitbildmodus-Hack"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_DE "Zeichnet Geometrie außerhalb des normalen 4:3-Seitenverhältnisses. Kann grafische Fehler in den aufgedeckten Bereichen erzeugen."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_DE "GD-ROM schnelles Laden (ungenau)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_DE "Beschleunigt das Laden von GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_DE "Dreamcast 32 MB RAM Mod"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_DE "Aktiviert 32 MB RAM Mod für Dreamcast.Kann die Kompatibilität beeinflussen"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_DE "SH4 CPU unter-/übertakten"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_DE "Ändert den Haupt-CPU-Takt des SH4 statt der standardmäßigen 200 MHz. Untertakten kann bei langsamen Plattformen helfen. Übertakten kann die Bildrate für einige Spiele erhöhen. Mit Vorsicht verwenden."
#define OPTION_VAL_100_DE NULL
#define OPTION_VAL_110_DE NULL
#define OPTION_VAL_120_DE NULL
#define OPTION_VAL_130_DE NULL
#define OPTION_VAL_140_DE NULL
#define OPTION_VAL_150_DE NULL
#define OPTION_VAL_160_DE NULL
#define OPTION_VAL_170_DE NULL
#define OPTION_VAL_180_DE NULL
#define OPTION_VAL_190_DE NULL
#define OPTION_VAL_200_DE NULL
#define OPTION_VAL_210_DE NULL
#define OPTION_VAL_220_DE NULL
#define OPTION_VAL_230_DE NULL
#define OPTION_VAL_240_DE NULL
#define OPTION_VAL_250_DE NULL
#define OPTION_VAL_260_DE NULL
#define OPTION_VAL_270_DE NULL
#define OPTION_VAL_280_DE NULL
#define OPTION_VAL_290_DE NULL
#define OPTION_VAL_300_DE NULL
#define OPTION_VAL_310_DE NULL
#define OPTION_VAL_320_DE NULL
#define OPTION_VAL_330_DE NULL
#define OPTION_VAL_340_DE NULL
#define OPTION_VAL_350_DE NULL
#define OPTION_VAL_360_DE NULL
#define OPTION_VAL_370_DE NULL
#define OPTION_VAL_380_DE NULL
#define OPTION_VAL_390_DE NULL
#define OPTION_VAL_400_DE NULL
#define OPTION_VAL_410_DE NULL
#define OPTION_VAL_420_DE NULL
#define OPTION_VAL_430_DE NULL
#define OPTION_VAL_440_DE NULL
#define OPTION_VAL_450_DE NULL
#define OPTION_VAL_460_DE NULL
#define OPTION_VAL_470_DE NULL
#define OPTION_VAL_480_DE NULL
#define OPTION_VAL_490_DE NULL
#define OPTION_VAL_500_DE NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_DE "Eigene Texturen laden"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_DE "Benutzerdefinierte Texturen aus dem Ordner „system/dc/textures/<game-id>/“ laden."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_DE "Eigene Texturen vorabladen"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_DE "Eigene Texturen beim Start von Spielen vorabladen. Dies kann die Leistung verbessern, erhöht jedoch den Speicherverbrauch."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_DE "Texturen dumpen"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_DE "Jedes Mal, wenn eine neue Textur vom Spiel verwendet wird, wird sie als .png-Datei im Ordner „system/dc/texdump/<game-id>/“ gespeichert."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_DE "Ersetzte Texturen aus Speicher entfernen"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_DE "Bereits durch eigene, ersetzte Texturen immer aus dem Speicher entfernen."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_DE "Analog-Stick Totzone"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_DE "Auswählen, wie weit der analogen Stick gedrückt werden muss, bevor er verarbeitet wird."
#define OPTION_VAL_0_DE "0 %"
#define OPTION_VAL_5_DE "5 %"
#define OPTION_VAL_10_DE "10 %"
#define OPTION_VAL_15_DE "15 %"
#define OPTION_VAL_20_DE "20 %"
#define OPTION_VAL_25_DE "25 %"
#define OPTION_VAL_30_DE "30 %"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_DE "Trigger Totzone"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_DE "Auswählen, wie stark der Abzug gedrückt werden muss, bevor er verarbeitet wird."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_DE "Digitale Trigger"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_DE "Wenn aktiviert, reagiert ein Abzug wie eine normale Taste, d. h. entweder als vollständig gedrückt oder gar nicht gedrückt, ohne dass es ein Dazwischen gibt."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_DE "Digitale Ausgänge übertragen"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_DE "Digitale Ausgänge und Force-Feedback-Status auf TCP-Port 8000 übertragen. Kompatibel mit der MAME-Option „-output network“."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_DE "Lightgun-Einstellungen anzeigen"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_DE "Aktiviert die Konfiguration der Lightgun-Fadenkreuz-Anzeigeoptionen. HINWEIS: Schnellmenü muss eingeschaltet sein, damit diese Einstellung wirksam wird."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_DE "Skalierung der Fadenkreuzgröße"
#define OPTION_VAL_50_DE "50 %"
#define OPTION_VAL_60_DE "60 %"
#define OPTION_VAL_70_DE "70 %"
#define OPTION_VAL_80_DE "80 %"
#define OPTION_VAL_90_DE "90 %"
#define OPTION_VAL_100_O47_DE "100 %"
#define OPTION_VAL_110_O47_DE "110 %"
#define OPTION_VAL_120_O47_DE "120 %"
#define OPTION_VAL_130_O47_DE "130 %"
#define OPTION_VAL_140_O47_DE "140 %"
#define OPTION_VAL_150_O47_DE "150 %"
#define OPTION_VAL_160_O47_DE "160 %"
#define OPTION_VAL_170_O47_DE "170 %"
#define OPTION_VAL_180_O47_DE "180 %"
#define OPTION_VAL_190_O47_DE "190 %"
#define OPTION_VAL_200_O47_DE "200 %"
#define OPTION_VAL_210_O47_DE "210 %"
#define OPTION_VAL_220_O47_DE "220 %"
#define OPTION_VAL_230_O47_DE "230 %"
#define OPTION_VAL_240_O47_DE "240 %"
#define OPTION_VAL_250_O47_DE "250 %"
#define OPTION_VAL_260_O47_DE "260 %"
#define OPTION_VAL_270_O47_DE "270 %"
#define OPTION_VAL_280_O47_DE "280 %"
#define OPTION_VAL_290_O47_DE "290 %"
#define OPTION_VAL_300_O47_DE "300 %"
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_DE "Anzeige Pistolen-Fadenkreuz 1"
#define OPTION_VAL_WHITE_DE "Weiß"
#define OPTION_VAL_RED_DE "Rot"
#define OPTION_VAL_GREEN_DE "Grün"
#define OPTION_VAL_BLUE_DE "Blau"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_DE "Anzeige Pistolen-Fadenkreuz 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_DE "Anzeige Pistolen-Fadenkreuz 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_DE "Anzeige Pistolen-Fadenkreuz 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_DE "Gerät im Erweiterungsslot A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_DE "Das Gerät auswählen, das im Erweiterungsslot A1 angeschlossen ist (Port A Slot 1)."
#define OPTION_VAL_VMU_DE NULL
#define OPTION_VAL_PURUPURU_DE "Vibrationspaket"
#define OPTION_VAL_DREAMPOTATO_DE NULL
#define OPTION_VAL_NONE_DE "Nein"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_DE "Gerät im Erweiterungsslot A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_DE "Das Gerät auswählen, das im Erweiterungsslot A2 angeschlossen ist (Port A Slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_DE "Gerät im Erweiterungsslot B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_DE "Das Gerät auswählen, das im Erweiterungsslot B1 angeschlossen ist (Port B Slot 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_DE "Gerät im Erweiterungsslot B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_DE "Das Gerät auswählen, das im Erweiterungsslot B2 angeschlossen ist (Port B Slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_DE "Gerät im Erweiterungsslot C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_DE "Das Gerät auswählen, das im Erweiterungsslot C1 angeschlossen ist (Port C Slot 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_DE "Gerät im Erweiterungsslot C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_DE "Das Gerät auswählen, das im Erweiterungsslot C2 angeschlossen ist (Port C Slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_DE "Gerät im Erweiterungsslot D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_DE "Das Gerät auswählen, das im Erweiterungsslot D1 angeschlossen ist (Port D Slot 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_DE "Gerät im Erweiterungsslot D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_DE "Das Gerät auswählen, das im Erweiterungsslot D2 angeschlossen ist (Port D Slot 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_DE "Visual Memory Units/Systeme (VMU) pro Spiel"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_DE "VMUs pro Spiel"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_DE "Wenn deaktiviert, teilen alle Spiele bis zu 8 VMU-Speicherdateien (A1/A2/B1/B2/C1/C2/D1/D2) im Systemverzeichnis von RetroArch.\nDie Einstellung „VMU A1“ erzeugt für jedes Spiel, das gestartet wird, eine einzigartige VMU-„A1“-Datei im Speicherverzeichnis von RetroArch.\nDie Einstellung „Alle VMUs“ erzeugt bis zu 8 einzigartige VMU-Dateien (A1/A2/B1/B2/C1/C2/D1/D2) für jedes Spiel, das gestartet wird."
#define OPTION_VAL_VMU_A1_DE NULL
#define OPTION_VAL_ALL_VMUS_DE "Alle VMUs"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_DE NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_DE NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_DE "Wenn aktiviert, werden VMU-Signaltöne abgespielt."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_DE "DreamPotato VMU-Speicher"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_DE "Die Speicherung von VMU-Daten auf dem verbundenen DreamPotato-Emulator aktivieren. Wenn diese Option aktiviert ist, wird die lokale VMU-Speicherdatei nicht verwendet."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_DE "Visual Memory Unit/System (VMU) Anzeigeeinstellungen anzeigen"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_DE "VMU-Bildschirmeinstellungen anzeigen"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_DE "Konfiguration von emulierten VMU LCD-Bildschirmen über Sichtbarkeit, Größe, Position und Farbe aktivieren. HINWEIS: Das Schnellmenü muss aktiviert sein, damit diese Einstellung wirksam wird."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_DE "Anzeige VMU-Bildschirm 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_DE "Position VMU-Bildschirm 1"
#define OPTION_VAL_UPPER_LEFT_DE "Oben links"
#define OPTION_VAL_UPPER_RIGHT_DE "Oben rechts"
#define OPTION_VAL_LOWER_LEFT_DE "Unten links"
#define OPTION_VAL_LOWER_RIGHT_DE "Unten rechts"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_DE "Größe VMU-Bildschirm 1"
#define OPTION_VAL_1X_DE "1×"
#define OPTION_VAL_3X_DE "3×"
#define OPTION_VAL_5X_DE "5×"
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_DE "Pixel-an-Farbe VMU-Bildschirm 1"
#define OPTION_VAL_DEFAULT_ON_00_DE "Standard EIN"
#define OPTION_VAL_DEFAULT_OFF_01_DE "Standard AUS"
#define OPTION_VAL_BLACK_02_DE "Schwarz"
#define OPTION_VAL_LIGHT_BLUE_04_DE "Hellblau"
#define OPTION_VAL_CYAN_06_DE NULL
#define OPTION_VAL_CYAN_BLUE_07_DE "Zyanblau"
#define OPTION_VAL_LIGHT_GREEN_08_DE "Hellgrün"
#define OPTION_VAL_CYAN_GREEN_09_DE "Zyangrün"
#define OPTION_VAL_LIGHT_CYAN_10_DE "Hellzyan"
#define OPTION_VAL_PURPLE_12_DE "Lila"
#define OPTION_VAL_LIGHT_PURPLE_13_DE "Hellviolett"
#define OPTION_VAL_YELLOW_14_DE "Gelb"
#define OPTION_VAL_GRAY_15_DE "Grau"
#define OPTION_VAL_LIGHT_PURPLE_2_16_DE "Hellviolett (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_DE "Hellgrün (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_DE "Hellgrün (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_DE "Hellzyan (2)"
#define OPTION_VAL_LIGHT_RED_2_20_DE "Hellrot (2)"
#define OPTION_VAL_MAGENTA_21_DE NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_DE "Hellviolett (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_DE "Hellorange"
#define OPTION_VAL_ORANGE_24_DE NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_DE "Hellviolett (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_DE "Hellgelb"
#define OPTION_VAL_LIGHT_YELLOW_2_27_DE "Hellgelb (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_DE "Pixel-aus-Farbe VMU-Bildschirm 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_DE "Deckkraft VMU-Bildschirm 1"
#define OPTION_VAL_40_DE "40 %"
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_DE "Anzeige VMU-Bildschirm 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_DE "Position VMU-Bildschirm 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_DE "Größe VMU-Bildschirm 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_DE "Pixel-an-Farbe Bildschirm 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_DE "Pixel-aus-Farbe VMU-Bildschirm 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_DE "Deckkraft VMU-Bildschirm 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_DE "Anzeige VMU-Bildschirm 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_DE "Position VMU-Bildschirm 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_DE "Größe VMU-Bildschirm 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_DE "Pixel-an-Farbe Bildschirm 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_DE "Pixel-aus-Farbe VMU-Bildschirm 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_DE "Deckkraft VMU-Bildschirm 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_DE "Anzeige VMU-Bildschirm 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_DE "VMU-Bildschirm-4-Position"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_DE "Größe VMU-Bildschirm 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_DE "Pixel-an-Farbe Bildschirm 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_DE "Pixel-aus-Farbe VMU-Bildschirm 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_DE "Deckkraft VMU-Bildschirm 4"

struct retro_core_option_v2_category option_cats_de[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_DE,
      CATEGORY_SYSTEM_INFO_0_DE
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_DE,
      CATEGORY_VIDEO_INFO_0_DE
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_DE,
      CATEGORY_PERFORMANCE_INFO_0_DE
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_DE,
      CATEGORY_HACKS_INFO_0_DE
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_DE,
      CATEGORY_INPUT_INFO_0_DE
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_DE,
      CATEGORY_EXPANSIONS_INFO_0_DE
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_DE,
      CATEGORY_VMU_INFO_0_DE
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_de[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_DE,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_DE },
         { "USA",     OPTION_VAL_USA_DE },
         { "Europe",  OPTION_VAL_EUROPE_DE },
         { "Default", OPTION_VAL_DEFAULT_DE },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_DE,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_DE },
         { "English",  OPTION_VAL_ENGLISH_DE },
         { "German",   OPTION_VAL_GERMAN_DE },
         { "French",   OPTION_VAL_FRENCH_DE },
         { "Spanish",  OPTION_VAL_SPANISH_DE },
         { "Italian",  OPTION_VAL_ITALIAN_DE },
         { "Default",  OPTION_VAL_DEFAULT_DE },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_DE,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_DE },
         { "640x480",    OPTION_VAL_640X480_DE },
         { "800x600",    OPTION_VAL_800X600_DE },
         { "960x720",    OPTION_VAL_960X720_DE },
         { "1024x768",   OPTION_VAL_1024X768_DE },
         { "1280x960",   OPTION_VAL_1280X960_DE },
         { "1440x1080",  OPTION_VAL_1440X1080_DE },
         { "1600x1200",  OPTION_VAL_1600X1200_DE },
         { "1920x1440",  OPTION_VAL_1920X1440_DE },
         { "2560x1920",  OPTION_VAL_2560X1920_DE },
         { "2880x2160",  OPTION_VAL_2880X2160_DE },
         { "3200x2400",  OPTION_VAL_3200X2400_DE },
         { "3840x2880",  OPTION_VAL_3840X2880_DE },
         { "4480x3360",  OPTION_VAL_4480X3360_DE },
         { "5120x3840",  OPTION_VAL_5120X3840_DE },
         { "5760x4320",  OPTION_VAL_5760X4320_DE },
         { "6400x4800",  OPTION_VAL_6400X4800_DE },
         { "7040x5280",  OPTION_VAL_7040X5280_DE },
         { "7680x5760",  OPTION_VAL_7680X5760_DE },
         { "8320x6240",  OPTION_VAL_8320X6240_DE },
         { "8960x6720",  OPTION_VAL_8960X6720_DE },
         { "9600x7200",  OPTION_VAL_9600X7200_DE },
         { "10240x7680", OPTION_VAL_10240X7680_DE },
         { "10880x8160", OPTION_VAL_10880X8160_DE },
         { "11520x8640", OPTION_VAL_11520X8640_DE },
         { "12160x9120", OPTION_VAL_12160X9120_DE },
         { "12800x9600", OPTION_VAL_12800X9600_DE },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_DE,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_DE },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_DE },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_DE },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_DE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_DE },
         { "PAL",     OPTION_VAL_PAL_DE },
         { "PAL_N",   OPTION_VAL_PAL_N_DE },
         { "PAL_M",   OPTION_VAL_PAL_M_DE },
         { "Default", OPTION_VAL_DEFAULT_DE },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_DE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_DE },
         { "vertical",   OPTION_VAL_VERTICAL_DE },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_DE,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_DE },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_DE },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_DE },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_DE,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_DE },
         { "1GB",   OPTION_VAL_1GB_DE },
         { "2GB",   OPTION_VAL_2GB_DE },
         { "4GB",   OPTION_VAL_4GB_DE },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_DE,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_DE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_DE,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_DE,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_DE },
         { "1",  OPTION_VAL_1_DE },
         { "2",  OPTION_VAL_2_DE },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_DE,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_DE },
         { "4",  OPTION_VAL_4_DE },
         { "6",  OPTION_VAL_6_DE },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_DE,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_DE,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_DE,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_DE,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_DE,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_DE,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_DE,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_DE },
         { "more", OPTION_VAL_MORE_DE },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_DE,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_DE },
         { "110", OPTION_VAL_110_DE },
         { "120", OPTION_VAL_120_DE },
         { "130", OPTION_VAL_130_DE },
         { "140", OPTION_VAL_140_DE },
         { "150", OPTION_VAL_150_DE },
         { "160", OPTION_VAL_160_DE },
         { "170", OPTION_VAL_170_DE },
         { "180", OPTION_VAL_180_DE },
         { "190", OPTION_VAL_190_DE },
         { "200", OPTION_VAL_200_DE },
         { "210", OPTION_VAL_210_DE },
         { "220", OPTION_VAL_220_DE },
         { "230", OPTION_VAL_230_DE },
         { "240", OPTION_VAL_240_DE },
         { "250", OPTION_VAL_250_DE },
         { "260", OPTION_VAL_260_DE },
         { "270", OPTION_VAL_270_DE },
         { "280", OPTION_VAL_280_DE },
         { "290", OPTION_VAL_290_DE },
         { "300", OPTION_VAL_300_DE },
         { "310", OPTION_VAL_310_DE },
         { "320", OPTION_VAL_320_DE },
         { "330", OPTION_VAL_330_DE },
         { "340", OPTION_VAL_340_DE },
         { "350", OPTION_VAL_350_DE },
         { "360", OPTION_VAL_360_DE },
         { "370", OPTION_VAL_370_DE },
         { "380", OPTION_VAL_380_DE },
         { "390", OPTION_VAL_390_DE },
         { "400", OPTION_VAL_400_DE },
         { "410", OPTION_VAL_410_DE },
         { "420", OPTION_VAL_420_DE },
         { "430", OPTION_VAL_430_DE },
         { "440", OPTION_VAL_440_DE },
         { "450", OPTION_VAL_450_DE },
         { "460", OPTION_VAL_460_DE },
         { "470", OPTION_VAL_470_DE },
         { "480", OPTION_VAL_480_DE },
         { "490", OPTION_VAL_490_DE },
         { "500", OPTION_VAL_500_DE },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_DE,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_DE },
         { "5%",  OPTION_VAL_5_DE },
         { "10%", OPTION_VAL_10_DE },
         { "15%", OPTION_VAL_15_DE },
         { "20%", OPTION_VAL_20_DE },
         { "25%", OPTION_VAL_25_DE },
         { "30%", OPTION_VAL_30_DE },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_DE,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_DE },
         { "5%",  OPTION_VAL_5_DE },
         { "10%", OPTION_VAL_10_DE },
         { "15%", OPTION_VAL_15_DE },
         { "20%", OPTION_VAL_20_DE },
         { "25%", OPTION_VAL_25_DE },
         { "30%", OPTION_VAL_30_DE },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_DE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_DE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_DE,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_DE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_DE },
         { "60%",  OPTION_VAL_60_DE },
         { "70%",  OPTION_VAL_70_DE },
         { "80%",  OPTION_VAL_80_DE },
         { "90%",  OPTION_VAL_90_DE },
         { "100%", OPTION_VAL_100_O47_DE },
         { "110%", OPTION_VAL_110_O47_DE },
         { "120%", OPTION_VAL_120_O47_DE },
         { "130%", OPTION_VAL_130_O47_DE },
         { "140%", OPTION_VAL_140_O47_DE },
         { "150%", OPTION_VAL_150_O47_DE },
         { "160%", OPTION_VAL_160_O47_DE },
         { "170%", OPTION_VAL_170_O47_DE },
         { "180%", OPTION_VAL_180_O47_DE },
         { "190%", OPTION_VAL_190_O47_DE },
         { "200%", OPTION_VAL_200_O47_DE },
         { "210%", OPTION_VAL_210_O47_DE },
         { "220%", OPTION_VAL_220_O47_DE },
         { "230%", OPTION_VAL_230_O47_DE },
         { "240%", OPTION_VAL_240_O47_DE },
         { "250%", OPTION_VAL_250_O47_DE },
         { "260%", OPTION_VAL_260_O47_DE },
         { "270%", OPTION_VAL_270_O47_DE },
         { "280%", OPTION_VAL_280_O47_DE },
         { "290%", OPTION_VAL_290_O47_DE },
         { "300%", OPTION_VAL_300_O47_DE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_DE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DE },
         { "Red",      OPTION_VAL_RED_DE },
         { "Green",    OPTION_VAL_GREEN_DE },
         { "Blue",     OPTION_VAL_BLUE_DE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_DE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DE },
         { "Red",      OPTION_VAL_RED_DE },
         { "Green",    OPTION_VAL_GREEN_DE },
         { "Blue",     OPTION_VAL_BLUE_DE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_DE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DE },
         { "Red",      OPTION_VAL_RED_DE },
         { "Green",    OPTION_VAL_GREEN_DE },
         { "Blue",     OPTION_VAL_BLUE_DE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_DE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_DE },
         { "Red",      OPTION_VAL_RED_DE },
         { "Green",    OPTION_VAL_GREEN_DE },
         { "Blue",     OPTION_VAL_BLUE_DE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_DE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DE },
         { "Purupuru", OPTION_VAL_PURUPURU_DE },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_DE },
         { "None",     OPTION_VAL_NONE_DE },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_DE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DE },
         { "Purupuru", OPTION_VAL_PURUPURU_DE },
         { "None",     OPTION_VAL_NONE_DE },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_DE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DE },
         { "Purupuru", OPTION_VAL_PURUPURU_DE },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_DE },
         { "None",     OPTION_VAL_NONE_DE },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_DE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DE },
         { "Purupuru", OPTION_VAL_PURUPURU_DE },
         { "None",     OPTION_VAL_NONE_DE },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_DE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DE },
         { "Purupuru", OPTION_VAL_PURUPURU_DE },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_DE },
         { "None",     OPTION_VAL_NONE_DE },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_DE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DE },
         { "Purupuru", OPTION_VAL_PURUPURU_DE },
         { "None",     OPTION_VAL_NONE_DE },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_DE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DE },
         { "Purupuru", OPTION_VAL_PURUPURU_DE },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_DE },
         { "None",     OPTION_VAL_NONE_DE },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_DE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_DE },
         { "Purupuru", OPTION_VAL_PURUPURU_DE },
         { "None",     OPTION_VAL_NONE_DE },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_DE,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_DE,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_DE,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_DE },
         { "All VMUs", OPTION_VAL_ALL_VMUS_DE },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_DE,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_DE,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_DE,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_DE,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_DE,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_DE,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_DE,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_DE,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DE },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DE },
         { "2x", OPTION_VAL_2_O26_DE },
         { "3x", OPTION_VAL_3X_DE },
         { "4x", OPTION_VAL_4_DE },
         { "5x", OPTION_VAL_5X_DE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DE },
         { "20%",  OPTION_VAL_20_DE },
         { "30%",  OPTION_VAL_30_DE },
         { "40%",  OPTION_VAL_40_DE },
         { "50%",  OPTION_VAL_50_DE },
         { "60%",  OPTION_VAL_60_DE },
         { "70%",  OPTION_VAL_70_DE },
         { "80%",  OPTION_VAL_80_DE },
         { "90%",  OPTION_VAL_90_DE },
         { "100%", OPTION_VAL_100_O47_DE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DE },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DE },
         { "2x", OPTION_VAL_2_O26_DE },
         { "3x", OPTION_VAL_3X_DE },
         { "4x", OPTION_VAL_4_DE },
         { "5x", OPTION_VAL_5X_DE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DE },
         { "20%",  OPTION_VAL_20_DE },
         { "30%",  OPTION_VAL_30_DE },
         { "40%",  OPTION_VAL_40_DE },
         { "50%",  OPTION_VAL_50_DE },
         { "60%",  OPTION_VAL_60_DE },
         { "70%",  OPTION_VAL_70_DE },
         { "80%",  OPTION_VAL_80_DE },
         { "90%",  OPTION_VAL_90_DE },
         { "100%", OPTION_VAL_100_O47_DE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DE },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DE },
         { "2x", OPTION_VAL_2_O26_DE },
         { "3x", OPTION_VAL_3X_DE },
         { "4x", OPTION_VAL_4_DE },
         { "5x", OPTION_VAL_5X_DE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DE },
         { "20%",  OPTION_VAL_20_DE },
         { "30%",  OPTION_VAL_30_DE },
         { "40%",  OPTION_VAL_40_DE },
         { "50%",  OPTION_VAL_50_DE },
         { "60%",  OPTION_VAL_60_DE },
         { "70%",  OPTION_VAL_70_DE },
         { "80%",  OPTION_VAL_80_DE },
         { "90%",  OPTION_VAL_90_DE },
         { "100%", OPTION_VAL_100_O47_DE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_DE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_DE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_DE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_DE },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_DE },
         { "2x", OPTION_VAL_2_O26_DE },
         { "3x", OPTION_VAL_3X_DE },
         { "4x", OPTION_VAL_4_DE },
         { "5x", OPTION_VAL_5X_DE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_DE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_DE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_DE },
         { "BLUE 03",           OPTION_VAL_BLUE_DE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_DE },
         { "GREEN 05",          OPTION_VAL_GREEN_DE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_DE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_DE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_DE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_DE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_DE },
         { "RED 11",            OPTION_VAL_RED_DE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_DE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_DE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_DE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_DE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_DE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_DE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_DE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_DE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_DE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_DE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_DE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_DE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_DE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_DE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_DE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_DE },
         { "WHITE 28",          OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_DE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_DE },
         { "20%",  OPTION_VAL_20_DE },
         { "30%",  OPTION_VAL_30_DE },
         { "40%",  OPTION_VAL_40_DE },
         { "50%",  OPTION_VAL_50_DE },
         { "60%",  OPTION_VAL_60_DE },
         { "70%",  OPTION_VAL_70_DE },
         { "80%",  OPTION_VAL_80_DE },
         { "90%",  OPTION_VAL_90_DE },
         { "100%", OPTION_VAL_100_O47_DE },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_de = {
   option_cats_de,
   option_defs_de
};

/* RETRO_LANGUAGE_EL */

#define CATEGORY_SYSTEM_LABEL_EL "Σύστημα"
#define CATEGORY_SYSTEM_INFO_0_EL NULL
#define CATEGORY_VIDEO_LABEL_EL "Οδηγός Βίντεο"
#define CATEGORY_VIDEO_INFO_0_EL NULL
#define CATEGORY_PERFORMANCE_LABEL_EL "Επίδοση"
#define CATEGORY_PERFORMANCE_INFO_0_EL NULL
#define CATEGORY_HACKS_LABEL_EL NULL
#define CATEGORY_HACKS_INFO_0_EL NULL
#define CATEGORY_INPUT_LABEL_EL "Οδηγός Εισαγωγής"
#define CATEGORY_INPUT_INFO_0_EL NULL
#define CATEGORY_EXPANSIONS_LABEL_EL NULL
#define CATEGORY_EXPANSIONS_INFO_0_EL NULL
#define CATEGORY_VMU_LABEL_EL NULL
#define CATEGORY_VMU_INFO_0_EL NULL
#define CORE_OPTION_NAME_REGION_LABEL_EL "Περιοχή"
#define OPTION_VAL_JAPAN_EL NULL
#define OPTION_VAL_USA_EL NULL
#define OPTION_VAL_EUROPE_EL NULL
#define OPTION_VAL_DEFAULT_EL "Προκαθορισμένο"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_EL "Γλώσσα"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_EL NULL
#define OPTION_VAL_JAPANESE_EL "Ιαπωνικά"
#define OPTION_VAL_ENGLISH_EL "Αγγλικά"
#define OPTION_VAL_GERMAN_EL "Γερμανός"
#define OPTION_VAL_FRENCH_EL "Γαλλική γλώσσα"
#define OPTION_VAL_SPANISH_EL "Ισπανικά"
#define OPTION_VAL_ITALIAN_EL "Ιταλικά"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_EL NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_EL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_EL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_EL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_EL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_EL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_EL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_EL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_EL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_EL NULL
#define CORE_OPTION_NAME_UPNP_LABEL_EL NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_EL NULL
#define CORE_OPTION_NAME_DCNET_LABEL_EL NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_EL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_EL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_EL NULL
#define OPTION_VAL_320X240_EL NULL
#define OPTION_VAL_640X480_EL NULL
#define OPTION_VAL_800X600_EL NULL
#define OPTION_VAL_960X720_EL NULL
#define OPTION_VAL_1024X768_EL NULL
#define OPTION_VAL_1280X960_EL NULL
#define OPTION_VAL_1440X1080_EL NULL
#define OPTION_VAL_1600X1200_EL NULL
#define OPTION_VAL_1920X1440_EL NULL
#define OPTION_VAL_2560X1920_EL NULL
#define OPTION_VAL_2880X2160_EL NULL
#define OPTION_VAL_3200X2400_EL NULL
#define OPTION_VAL_3840X2880_EL NULL
#define OPTION_VAL_4480X3360_EL NULL
#define OPTION_VAL_5120X3840_EL NULL
#define OPTION_VAL_5760X4320_EL NULL
#define OPTION_VAL_6400X4800_EL NULL
#define OPTION_VAL_7040X5280_EL NULL
#define OPTION_VAL_7680X5760_EL NULL
#define OPTION_VAL_8320X6240_EL NULL
#define OPTION_VAL_8960X6720_EL NULL
#define OPTION_VAL_9600X7200_EL NULL
#define OPTION_VAL_10240X7680_EL NULL
#define OPTION_VAL_10880X8160_EL NULL
#define OPTION_VAL_11520X8640_EL NULL
#define OPTION_VAL_12160X9120_EL NULL
#define OPTION_VAL_12800X9600_EL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_EL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_EL NULL
#define OPTION_VAL_VGA_EL NULL
#define OPTION_VAL_TV_RGB_EL NULL
#define OPTION_VAL_TV_COMPOSITE_EL NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_EL NULL
#define OPTION_VAL_NTSC_EL NULL
#define OPTION_VAL_PAL_EL NULL
#define OPTION_VAL_PAL_N_EL NULL
#define OPTION_VAL_PAL_M_EL NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_EL "Περιστροφή οθόνης"
#define OPTION_VAL_HORIZONTAL_EL NULL
#define OPTION_VAL_VERTICAL_EL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_EL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_EL NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_EL NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_EL NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_EL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_EL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_EL NULL
#define OPTION_VAL_512MB_EL NULL
#define OPTION_VAL_1GB_EL NULL
#define OPTION_VAL_2GB_EL NULL
#define OPTION_VAL_4GB_EL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_EL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_EL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_EL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_EL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_EL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_EL NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_EL NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_EL NULL
#define CORE_OPTION_NAME_FOG_LABEL_EL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_EL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_EL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_EL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_EL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_EL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_EL NULL
#define OPTION_VAL_1_EL NULL
#define OPTION_VAL_2_EL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_EL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_EL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_EL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_EL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_EL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_EL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_EL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_EL NULL
#define OPTION_VAL_2_O26_EL NULL
#define OPTION_VAL_4_EL NULL
#define OPTION_VAL_6_EL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_EL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_EL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_EL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_EL NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_EL NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_EL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_EL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_EL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_EL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_EL NULL
#define OPTION_VAL_SOME_EL "Φυσιολογικά"
#define OPTION_VAL_MORE_EL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_EL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_EL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_EL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_EL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_EL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_EL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_EL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_EL NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_EL NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_EL NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_EL NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_EL NULL
#define OPTION_VAL_100_EL NULL
#define OPTION_VAL_110_EL NULL
#define OPTION_VAL_120_EL NULL
#define OPTION_VAL_130_EL NULL
#define OPTION_VAL_140_EL NULL
#define OPTION_VAL_150_EL NULL
#define OPTION_VAL_160_EL NULL
#define OPTION_VAL_170_EL NULL
#define OPTION_VAL_180_EL NULL
#define OPTION_VAL_190_EL NULL
#define OPTION_VAL_200_EL NULL
#define OPTION_VAL_210_EL NULL
#define OPTION_VAL_220_EL NULL
#define OPTION_VAL_230_EL NULL
#define OPTION_VAL_240_EL NULL
#define OPTION_VAL_250_EL NULL
#define OPTION_VAL_260_EL NULL
#define OPTION_VAL_270_EL NULL
#define OPTION_VAL_280_EL NULL
#define OPTION_VAL_290_EL NULL
#define OPTION_VAL_300_EL NULL
#define OPTION_VAL_310_EL NULL
#define OPTION_VAL_320_EL NULL
#define OPTION_VAL_330_EL NULL
#define OPTION_VAL_340_EL NULL
#define OPTION_VAL_350_EL NULL
#define OPTION_VAL_360_EL NULL
#define OPTION_VAL_370_EL NULL
#define OPTION_VAL_380_EL NULL
#define OPTION_VAL_390_EL NULL
#define OPTION_VAL_400_EL NULL
#define OPTION_VAL_410_EL NULL
#define OPTION_VAL_420_EL NULL
#define OPTION_VAL_430_EL NULL
#define OPTION_VAL_440_EL NULL
#define OPTION_VAL_450_EL NULL
#define OPTION_VAL_460_EL NULL
#define OPTION_VAL_470_EL NULL
#define OPTION_VAL_480_EL NULL
#define OPTION_VAL_490_EL NULL
#define OPTION_VAL_500_EL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_EL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_EL NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_EL NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_EL NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_EL NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_EL NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_EL NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_EL NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_EL NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_EL NULL
#define OPTION_VAL_0_EL NULL
#define OPTION_VAL_5_EL NULL
#define OPTION_VAL_10_EL NULL
#define OPTION_VAL_15_EL NULL
#define OPTION_VAL_20_EL NULL
#define OPTION_VAL_25_EL NULL
#define OPTION_VAL_30_EL NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_EL NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_EL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_EL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_EL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_EL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_EL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_EL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_EL NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_EL NULL
#define OPTION_VAL_50_EL NULL
#define OPTION_VAL_60_EL NULL
#define OPTION_VAL_70_EL NULL
#define OPTION_VAL_80_EL "το 80%"
#define OPTION_VAL_90_EL NULL
#define OPTION_VAL_100_O47_EL NULL
#define OPTION_VAL_110_O47_EL NULL
#define OPTION_VAL_120_O47_EL NULL
#define OPTION_VAL_130_O47_EL NULL
#define OPTION_VAL_140_O47_EL NULL
#define OPTION_VAL_150_O47_EL NULL
#define OPTION_VAL_160_O47_EL NULL
#define OPTION_VAL_170_O47_EL NULL
#define OPTION_VAL_180_O47_EL NULL
#define OPTION_VAL_190_O47_EL NULL
#define OPTION_VAL_200_O47_EL NULL
#define OPTION_VAL_210_O47_EL NULL
#define OPTION_VAL_220_O47_EL NULL
#define OPTION_VAL_230_O47_EL NULL
#define OPTION_VAL_240_O47_EL NULL
#define OPTION_VAL_250_O47_EL NULL
#define OPTION_VAL_260_O47_EL NULL
#define OPTION_VAL_270_O47_EL NULL
#define OPTION_VAL_280_O47_EL NULL
#define OPTION_VAL_290_O47_EL NULL
#define OPTION_VAL_300_O47_EL NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_EL NULL
#define OPTION_VAL_WHITE_EL NULL
#define OPTION_VAL_RED_EL "Κόκκινο"
#define OPTION_VAL_GREEN_EL "Πράσινο"
#define OPTION_VAL_BLUE_EL "Μπλε"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_EL NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_EL NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_EL NULL
#define OPTION_VAL_VMU_EL NULL
#define OPTION_VAL_PURUPURU_EL NULL
#define OPTION_VAL_DREAMPOTATO_EL NULL
#define OPTION_VAL_NONE_EL "Κανείς"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_EL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_EL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_EL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_EL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_EL NULL
#define OPTION_VAL_VMU_A1_EL NULL
#define OPTION_VAL_ALL_VMUS_EL NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_EL NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_EL NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_EL NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_EL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_EL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_EL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_EL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_EL NULL
#define OPTION_VAL_UPPER_LEFT_EL NULL
#define OPTION_VAL_UPPER_RIGHT_EL NULL
#define OPTION_VAL_LOWER_LEFT_EL NULL
#define OPTION_VAL_LOWER_RIGHT_EL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_EL NULL
#define OPTION_VAL_1X_EL NULL
#define OPTION_VAL_3X_EL NULL
#define OPTION_VAL_5X_EL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_EL NULL
#define OPTION_VAL_DEFAULT_ON_00_EL NULL
#define OPTION_VAL_DEFAULT_OFF_01_EL NULL
#define OPTION_VAL_BLACK_02_EL NULL
#define OPTION_VAL_LIGHT_BLUE_04_EL NULL
#define OPTION_VAL_CYAN_06_EL NULL
#define OPTION_VAL_CYAN_BLUE_07_EL NULL
#define OPTION_VAL_LIGHT_GREEN_08_EL NULL
#define OPTION_VAL_CYAN_GREEN_09_EL NULL
#define OPTION_VAL_LIGHT_CYAN_10_EL NULL
#define OPTION_VAL_PURPLE_12_EL NULL
#define OPTION_VAL_LIGHT_PURPLE_13_EL NULL
#define OPTION_VAL_YELLOW_14_EL "Κίτρινο"
#define OPTION_VAL_GRAY_15_EL NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_EL NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_EL NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_EL NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_EL NULL
#define OPTION_VAL_LIGHT_RED_2_20_EL NULL
#define OPTION_VAL_MAGENTA_21_EL NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_EL NULL
#define OPTION_VAL_LIGHT_ORANGE_23_EL NULL
#define OPTION_VAL_ORANGE_24_EL NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_EL NULL
#define OPTION_VAL_LIGHT_YELLOW_26_EL NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_EL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_EL NULL
#define OPTION_VAL_40_EL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_EL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_EL NULL

struct retro_core_option_v2_category option_cats_el[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_EL,
      CATEGORY_SYSTEM_INFO_0_EL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_EL,
      CATEGORY_VIDEO_INFO_0_EL
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_EL,
      CATEGORY_PERFORMANCE_INFO_0_EL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_EL,
      CATEGORY_HACKS_INFO_0_EL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_EL,
      CATEGORY_INPUT_INFO_0_EL
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_EL,
      CATEGORY_EXPANSIONS_INFO_0_EL
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_EL,
      CATEGORY_VMU_INFO_0_EL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_el[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_EL,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_EL },
         { "USA",     OPTION_VAL_USA_EL },
         { "Europe",  OPTION_VAL_EUROPE_EL },
         { "Default", OPTION_VAL_DEFAULT_EL },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_EL,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_EL },
         { "English",  OPTION_VAL_ENGLISH_EL },
         { "German",   OPTION_VAL_GERMAN_EL },
         { "French",   OPTION_VAL_FRENCH_EL },
         { "Spanish",  OPTION_VAL_SPANISH_EL },
         { "Italian",  OPTION_VAL_ITALIAN_EL },
         { "Default",  OPTION_VAL_DEFAULT_EL },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_EL,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_EL },
         { "640x480",    OPTION_VAL_640X480_EL },
         { "800x600",    OPTION_VAL_800X600_EL },
         { "960x720",    OPTION_VAL_960X720_EL },
         { "1024x768",   OPTION_VAL_1024X768_EL },
         { "1280x960",   OPTION_VAL_1280X960_EL },
         { "1440x1080",  OPTION_VAL_1440X1080_EL },
         { "1600x1200",  OPTION_VAL_1600X1200_EL },
         { "1920x1440",  OPTION_VAL_1920X1440_EL },
         { "2560x1920",  OPTION_VAL_2560X1920_EL },
         { "2880x2160",  OPTION_VAL_2880X2160_EL },
         { "3200x2400",  OPTION_VAL_3200X2400_EL },
         { "3840x2880",  OPTION_VAL_3840X2880_EL },
         { "4480x3360",  OPTION_VAL_4480X3360_EL },
         { "5120x3840",  OPTION_VAL_5120X3840_EL },
         { "5760x4320",  OPTION_VAL_5760X4320_EL },
         { "6400x4800",  OPTION_VAL_6400X4800_EL },
         { "7040x5280",  OPTION_VAL_7040X5280_EL },
         { "7680x5760",  OPTION_VAL_7680X5760_EL },
         { "8320x6240",  OPTION_VAL_8320X6240_EL },
         { "8960x6720",  OPTION_VAL_8960X6720_EL },
         { "9600x7200",  OPTION_VAL_9600X7200_EL },
         { "10240x7680", OPTION_VAL_10240X7680_EL },
         { "10880x8160", OPTION_VAL_10880X8160_EL },
         { "11520x8640", OPTION_VAL_11520X8640_EL },
         { "12160x9120", OPTION_VAL_12160X9120_EL },
         { "12800x9600", OPTION_VAL_12800X9600_EL },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_EL,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_EL },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_EL },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_EL },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_EL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_EL },
         { "PAL",     OPTION_VAL_PAL_EL },
         { "PAL_N",   OPTION_VAL_PAL_N_EL },
         { "PAL_M",   OPTION_VAL_PAL_M_EL },
         { "Default", OPTION_VAL_DEFAULT_EL },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_EL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_EL },
         { "vertical",   OPTION_VAL_VERTICAL_EL },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_EL,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_EL },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_EL },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_EL },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_EL,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_EL },
         { "1GB",   OPTION_VAL_1GB_EL },
         { "2GB",   OPTION_VAL_2GB_EL },
         { "4GB",   OPTION_VAL_4GB_EL },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_EL,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_EL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_EL,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_EL,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_EL },
         { "1",  OPTION_VAL_1_EL },
         { "2",  OPTION_VAL_2_EL },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_EL,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_EL },
         { "4",  OPTION_VAL_4_EL },
         { "6",  OPTION_VAL_6_EL },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_EL,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_EL,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_EL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_EL,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_EL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_EL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_EL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_EL },
         { "more", OPTION_VAL_MORE_EL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_EL,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_EL },
         { "110", OPTION_VAL_110_EL },
         { "120", OPTION_VAL_120_EL },
         { "130", OPTION_VAL_130_EL },
         { "140", OPTION_VAL_140_EL },
         { "150", OPTION_VAL_150_EL },
         { "160", OPTION_VAL_160_EL },
         { "170", OPTION_VAL_170_EL },
         { "180", OPTION_VAL_180_EL },
         { "190", OPTION_VAL_190_EL },
         { "200", OPTION_VAL_200_EL },
         { "210", OPTION_VAL_210_EL },
         { "220", OPTION_VAL_220_EL },
         { "230", OPTION_VAL_230_EL },
         { "240", OPTION_VAL_240_EL },
         { "250", OPTION_VAL_250_EL },
         { "260", OPTION_VAL_260_EL },
         { "270", OPTION_VAL_270_EL },
         { "280", OPTION_VAL_280_EL },
         { "290", OPTION_VAL_290_EL },
         { "300", OPTION_VAL_300_EL },
         { "310", OPTION_VAL_310_EL },
         { "320", OPTION_VAL_320_EL },
         { "330", OPTION_VAL_330_EL },
         { "340", OPTION_VAL_340_EL },
         { "350", OPTION_VAL_350_EL },
         { "360", OPTION_VAL_360_EL },
         { "370", OPTION_VAL_370_EL },
         { "380", OPTION_VAL_380_EL },
         { "390", OPTION_VAL_390_EL },
         { "400", OPTION_VAL_400_EL },
         { "410", OPTION_VAL_410_EL },
         { "420", OPTION_VAL_420_EL },
         { "430", OPTION_VAL_430_EL },
         { "440", OPTION_VAL_440_EL },
         { "450", OPTION_VAL_450_EL },
         { "460", OPTION_VAL_460_EL },
         { "470", OPTION_VAL_470_EL },
         { "480", OPTION_VAL_480_EL },
         { "490", OPTION_VAL_490_EL },
         { "500", OPTION_VAL_500_EL },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_EL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_EL },
         { "5%",  OPTION_VAL_5_EL },
         { "10%", OPTION_VAL_10_EL },
         { "15%", OPTION_VAL_15_EL },
         { "20%", OPTION_VAL_20_EL },
         { "25%", OPTION_VAL_25_EL },
         { "30%", OPTION_VAL_30_EL },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_EL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_EL },
         { "5%",  OPTION_VAL_5_EL },
         { "10%", OPTION_VAL_10_EL },
         { "15%", OPTION_VAL_15_EL },
         { "20%", OPTION_VAL_20_EL },
         { "25%", OPTION_VAL_25_EL },
         { "30%", OPTION_VAL_30_EL },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_EL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_EL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_EL,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_EL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_EL },
         { "60%",  OPTION_VAL_60_EL },
         { "70%",  OPTION_VAL_70_EL },
         { "80%",  OPTION_VAL_80_EL },
         { "90%",  OPTION_VAL_90_EL },
         { "100%", OPTION_VAL_100_O47_EL },
         { "110%", OPTION_VAL_110_O47_EL },
         { "120%", OPTION_VAL_120_O47_EL },
         { "130%", OPTION_VAL_130_O47_EL },
         { "140%", OPTION_VAL_140_O47_EL },
         { "150%", OPTION_VAL_150_O47_EL },
         { "160%", OPTION_VAL_160_O47_EL },
         { "170%", OPTION_VAL_170_O47_EL },
         { "180%", OPTION_VAL_180_O47_EL },
         { "190%", OPTION_VAL_190_O47_EL },
         { "200%", OPTION_VAL_200_O47_EL },
         { "210%", OPTION_VAL_210_O47_EL },
         { "220%", OPTION_VAL_220_O47_EL },
         { "230%", OPTION_VAL_230_O47_EL },
         { "240%", OPTION_VAL_240_O47_EL },
         { "250%", OPTION_VAL_250_O47_EL },
         { "260%", OPTION_VAL_260_O47_EL },
         { "270%", OPTION_VAL_270_O47_EL },
         { "280%", OPTION_VAL_280_O47_EL },
         { "290%", OPTION_VAL_290_O47_EL },
         { "300%", OPTION_VAL_300_O47_EL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_EL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EL },
         { "Red",      OPTION_VAL_RED_EL },
         { "Green",    OPTION_VAL_GREEN_EL },
         { "Blue",     OPTION_VAL_BLUE_EL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_EL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EL },
         { "Red",      OPTION_VAL_RED_EL },
         { "Green",    OPTION_VAL_GREEN_EL },
         { "Blue",     OPTION_VAL_BLUE_EL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_EL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EL },
         { "Red",      OPTION_VAL_RED_EL },
         { "Green",    OPTION_VAL_GREEN_EL },
         { "Blue",     OPTION_VAL_BLUE_EL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_EL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EL },
         { "Red",      OPTION_VAL_RED_EL },
         { "Green",    OPTION_VAL_GREEN_EL },
         { "Blue",     OPTION_VAL_BLUE_EL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_EL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EL },
         { "Purupuru", OPTION_VAL_PURUPURU_EL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_EL },
         { "None",     OPTION_VAL_NONE_EL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_EL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EL },
         { "Purupuru", OPTION_VAL_PURUPURU_EL },
         { "None",     OPTION_VAL_NONE_EL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_EL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EL },
         { "Purupuru", OPTION_VAL_PURUPURU_EL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_EL },
         { "None",     OPTION_VAL_NONE_EL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_EL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EL },
         { "Purupuru", OPTION_VAL_PURUPURU_EL },
         { "None",     OPTION_VAL_NONE_EL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_EL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EL },
         { "Purupuru", OPTION_VAL_PURUPURU_EL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_EL },
         { "None",     OPTION_VAL_NONE_EL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_EL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EL },
         { "Purupuru", OPTION_VAL_PURUPURU_EL },
         { "None",     OPTION_VAL_NONE_EL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_EL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EL },
         { "Purupuru", OPTION_VAL_PURUPURU_EL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_EL },
         { "None",     OPTION_VAL_NONE_EL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_EL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EL },
         { "Purupuru", OPTION_VAL_PURUPURU_EL },
         { "None",     OPTION_VAL_NONE_EL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_EL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_EL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_EL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_EL },
         { "All VMUs", OPTION_VAL_ALL_VMUS_EL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_EL,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_EL,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_EL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_EL,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_EL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_EL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_EL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_EL,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EL },
         { "2x", OPTION_VAL_2_O26_EL },
         { "3x", OPTION_VAL_3X_EL },
         { "4x", OPTION_VAL_4_EL },
         { "5x", OPTION_VAL_5X_EL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EL },
         { "20%",  OPTION_VAL_20_EL },
         { "30%",  OPTION_VAL_30_EL },
         { "40%",  OPTION_VAL_40_EL },
         { "50%",  OPTION_VAL_50_EL },
         { "60%",  OPTION_VAL_60_EL },
         { "70%",  OPTION_VAL_70_EL },
         { "80%",  OPTION_VAL_80_EL },
         { "90%",  OPTION_VAL_90_EL },
         { "100%", OPTION_VAL_100_O47_EL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EL },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EL },
         { "2x", OPTION_VAL_2_O26_EL },
         { "3x", OPTION_VAL_3X_EL },
         { "4x", OPTION_VAL_4_EL },
         { "5x", OPTION_VAL_5X_EL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EL },
         { "20%",  OPTION_VAL_20_EL },
         { "30%",  OPTION_VAL_30_EL },
         { "40%",  OPTION_VAL_40_EL },
         { "50%",  OPTION_VAL_50_EL },
         { "60%",  OPTION_VAL_60_EL },
         { "70%",  OPTION_VAL_70_EL },
         { "80%",  OPTION_VAL_80_EL },
         { "90%",  OPTION_VAL_90_EL },
         { "100%", OPTION_VAL_100_O47_EL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EL },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EL },
         { "2x", OPTION_VAL_2_O26_EL },
         { "3x", OPTION_VAL_3X_EL },
         { "4x", OPTION_VAL_4_EL },
         { "5x", OPTION_VAL_5X_EL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EL },
         { "20%",  OPTION_VAL_20_EL },
         { "30%",  OPTION_VAL_30_EL },
         { "40%",  OPTION_VAL_40_EL },
         { "50%",  OPTION_VAL_50_EL },
         { "60%",  OPTION_VAL_60_EL },
         { "70%",  OPTION_VAL_70_EL },
         { "80%",  OPTION_VAL_80_EL },
         { "90%",  OPTION_VAL_90_EL },
         { "100%", OPTION_VAL_100_O47_EL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EL },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EL },
         { "2x", OPTION_VAL_2_O26_EL },
         { "3x", OPTION_VAL_3X_EL },
         { "4x", OPTION_VAL_4_EL },
         { "5x", OPTION_VAL_5X_EL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EL },
         { "BLUE 03",           OPTION_VAL_BLUE_EL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EL },
         { "GREEN 05",          OPTION_VAL_GREEN_EL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EL },
         { "RED 11",            OPTION_VAL_RED_EL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EL },
         { "WHITE 28",          OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_EL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EL },
         { "20%",  OPTION_VAL_20_EL },
         { "30%",  OPTION_VAL_30_EL },
         { "40%",  OPTION_VAL_40_EL },
         { "50%",  OPTION_VAL_50_EL },
         { "60%",  OPTION_VAL_60_EL },
         { "70%",  OPTION_VAL_70_EL },
         { "80%",  OPTION_VAL_80_EL },
         { "90%",  OPTION_VAL_90_EL },
         { "100%", OPTION_VAL_100_O47_EL },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_el = {
   option_cats_el,
   option_defs_el
};

/* RETRO_LANGUAGE_EN */

#define CATEGORY_SYSTEM_LABEL_EN NULL
#define CATEGORY_SYSTEM_INFO_0_EN NULL
#define CATEGORY_VIDEO_LABEL_EN NULL
#define CATEGORY_VIDEO_INFO_0_EN NULL
#define CATEGORY_PERFORMANCE_LABEL_EN NULL
#define CATEGORY_PERFORMANCE_INFO_0_EN NULL
#define CATEGORY_HACKS_LABEL_EN NULL
#define CATEGORY_HACKS_INFO_0_EN NULL
#define CATEGORY_INPUT_LABEL_EN NULL
#define CATEGORY_INPUT_INFO_0_EN NULL
#define CATEGORY_EXPANSIONS_LABEL_EN NULL
#define CATEGORY_EXPANSIONS_INFO_0_EN NULL
#define CATEGORY_VMU_LABEL_EN NULL
#define CATEGORY_VMU_INFO_0_EN NULL
#define CORE_OPTION_NAME_REGION_LABEL_EN NULL
#define OPTION_VAL_JAPAN_EN NULL
#define OPTION_VAL_USA_EN NULL
#define OPTION_VAL_EUROPE_EN NULL
#define OPTION_VAL_DEFAULT_EN NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_EN NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_EN NULL
#define OPTION_VAL_JAPANESE_EN NULL
#define OPTION_VAL_ENGLISH_EN NULL
#define OPTION_VAL_GERMAN_EN NULL
#define OPTION_VAL_FRENCH_EN NULL
#define OPTION_VAL_SPANISH_EN NULL
#define OPTION_VAL_ITALIAN_EN NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_EN NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_EN NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_EN NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_EN NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_EN NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_EN NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_EN NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_EN NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_EN NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_EN NULL
#define CORE_OPTION_NAME_UPNP_LABEL_EN NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_EN NULL
#define CORE_OPTION_NAME_DCNET_LABEL_EN NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_EN NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_EN NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_EN NULL
#define OPTION_VAL_320X240_EN NULL
#define OPTION_VAL_640X480_EN NULL
#define OPTION_VAL_800X600_EN NULL
#define OPTION_VAL_960X720_EN NULL
#define OPTION_VAL_1024X768_EN NULL
#define OPTION_VAL_1280X960_EN NULL
#define OPTION_VAL_1440X1080_EN NULL
#define OPTION_VAL_1600X1200_EN NULL
#define OPTION_VAL_1920X1440_EN NULL
#define OPTION_VAL_2560X1920_EN NULL
#define OPTION_VAL_2880X2160_EN NULL
#define OPTION_VAL_3200X2400_EN NULL
#define OPTION_VAL_3840X2880_EN NULL
#define OPTION_VAL_4480X3360_EN NULL
#define OPTION_VAL_5120X3840_EN NULL
#define OPTION_VAL_5760X4320_EN NULL
#define OPTION_VAL_6400X4800_EN NULL
#define OPTION_VAL_7040X5280_EN NULL
#define OPTION_VAL_7680X5760_EN NULL
#define OPTION_VAL_8320X6240_EN NULL
#define OPTION_VAL_8960X6720_EN NULL
#define OPTION_VAL_9600X7200_EN NULL
#define OPTION_VAL_10240X7680_EN NULL
#define OPTION_VAL_10880X8160_EN NULL
#define OPTION_VAL_11520X8640_EN NULL
#define OPTION_VAL_12160X9120_EN NULL
#define OPTION_VAL_12800X9600_EN NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_EN NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_EN NULL
#define OPTION_VAL_VGA_EN NULL
#define OPTION_VAL_TV_RGB_EN NULL
#define OPTION_VAL_TV_COMPOSITE_EN NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_EN NULL
#define OPTION_VAL_NTSC_EN NULL
#define OPTION_VAL_PAL_EN NULL
#define OPTION_VAL_PAL_N_EN NULL
#define OPTION_VAL_PAL_M_EN NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_EN NULL
#define OPTION_VAL_HORIZONTAL_EN NULL
#define OPTION_VAL_VERTICAL_EN NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_EN NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_EN NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_EN NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_EN NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_EN NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_EN NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_EN NULL
#define OPTION_VAL_512MB_EN NULL
#define OPTION_VAL_1GB_EN NULL
#define OPTION_VAL_2GB_EN NULL
#define OPTION_VAL_4GB_EN NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_EN NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_EN NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_EN "Full frame-buffer emulation"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_EN "Enable full frame-buffer emulation in VRAM. This is useful for games that directly read or write the frame-buffer in VRAM. When enabled, Internal Resolution is forced to 640x480 and performance may be severely impacted."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_EN NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_EN NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_EN NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_EN NULL
#define CORE_OPTION_NAME_FOG_LABEL_EN NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_EN NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_EN NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_EN NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_EN NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_EN NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_EN NULL
#define OPTION_VAL_1_EN "Force Nearest-Neighbour"
#define OPTION_VAL_2_EN NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_EN NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_EN NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_EN NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_EN "Notify front-end when internal frame rate changes (e.g. from 60 fps to 30 fps). Improves frame pacing in games that run at a locked 30 fps or 20 fps, but should be disabled for games with unlocked (unstable) frame rates (e.g. Ecco the Dolphin, Unreal Tournament). Note: Unavailable when 'Auto Skip Frame' is enabled."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_EN NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_EN "Post-process the rendered image to simulate effects specific to the PowerVR2 GPU and analogue video signals."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_EN NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_EN NULL
#define OPTION_VAL_2_O26_EN NULL
#define OPTION_VAL_4_EN NULL
#define OPTION_VAL_6_EN NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_EN "Texture Upscaling Maximum Filtered Size"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_EN NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_EN NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_EN NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_EN NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_EN NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_EN NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_EN NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_EN NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_EN NULL
#define OPTION_VAL_SOME_EN NULL
#define OPTION_VAL_MORE_EN NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_EN NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_EN NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_EN NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_EN NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_EN NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_EN NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_EN NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_EN NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_EN NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_EN NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_EN NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_EN NULL
#define OPTION_VAL_100_EN NULL
#define OPTION_VAL_110_EN NULL
#define OPTION_VAL_120_EN NULL
#define OPTION_VAL_130_EN NULL
#define OPTION_VAL_140_EN NULL
#define OPTION_VAL_150_EN NULL
#define OPTION_VAL_160_EN NULL
#define OPTION_VAL_170_EN NULL
#define OPTION_VAL_180_EN NULL
#define OPTION_VAL_190_EN NULL
#define OPTION_VAL_200_EN NULL
#define OPTION_VAL_210_EN NULL
#define OPTION_VAL_220_EN NULL
#define OPTION_VAL_230_EN NULL
#define OPTION_VAL_240_EN NULL
#define OPTION_VAL_250_EN NULL
#define OPTION_VAL_260_EN NULL
#define OPTION_VAL_270_EN NULL
#define OPTION_VAL_280_EN NULL
#define OPTION_VAL_290_EN NULL
#define OPTION_VAL_300_EN NULL
#define OPTION_VAL_310_EN NULL
#define OPTION_VAL_320_EN NULL
#define OPTION_VAL_330_EN NULL
#define OPTION_VAL_340_EN NULL
#define OPTION_VAL_350_EN NULL
#define OPTION_VAL_360_EN NULL
#define OPTION_VAL_370_EN NULL
#define OPTION_VAL_380_EN NULL
#define OPTION_VAL_390_EN NULL
#define OPTION_VAL_400_EN NULL
#define OPTION_VAL_410_EN NULL
#define OPTION_VAL_420_EN NULL
#define OPTION_VAL_430_EN NULL
#define OPTION_VAL_440_EN NULL
#define OPTION_VAL_450_EN NULL
#define OPTION_VAL_460_EN NULL
#define OPTION_VAL_470_EN NULL
#define OPTION_VAL_480_EN NULL
#define OPTION_VAL_490_EN NULL
#define OPTION_VAL_500_EN NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_EN NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_EN NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_EN NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_EN NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_EN NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_EN NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_EN NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_EN NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_EN "Analogue Stick Dead-zone"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_EN "Select how far you have to push the analogue stick before it starts being processed."
#define OPTION_VAL_0_EN NULL
#define OPTION_VAL_5_EN NULL
#define OPTION_VAL_10_EN NULL
#define OPTION_VAL_15_EN NULL
#define OPTION_VAL_20_EN NULL
#define OPTION_VAL_25_EN NULL
#define OPTION_VAL_30_EN NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_EN "Trigger Dead-zone"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_EN NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_EN NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_EN NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_EN NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_EN NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_EN NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_EN "Enable configuration of light gun cross-hair display options. NOTE: Quick Menu may need to be toggled for this setting to take effect."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_EN NULL
#define OPTION_VAL_50_EN NULL
#define OPTION_VAL_60_EN NULL
#define OPTION_VAL_70_EN NULL
#define OPTION_VAL_80_EN NULL
#define OPTION_VAL_90_EN NULL
#define OPTION_VAL_100_O47_EN NULL
#define OPTION_VAL_110_O47_EN NULL
#define OPTION_VAL_120_O47_EN NULL
#define OPTION_VAL_130_O47_EN NULL
#define OPTION_VAL_140_O47_EN NULL
#define OPTION_VAL_150_O47_EN NULL
#define OPTION_VAL_160_O47_EN NULL
#define OPTION_VAL_170_O47_EN NULL
#define OPTION_VAL_180_O47_EN NULL
#define OPTION_VAL_190_O47_EN NULL
#define OPTION_VAL_200_O47_EN NULL
#define OPTION_VAL_210_O47_EN NULL
#define OPTION_VAL_220_O47_EN NULL
#define OPTION_VAL_230_O47_EN NULL
#define OPTION_VAL_240_O47_EN NULL
#define OPTION_VAL_250_O47_EN NULL
#define OPTION_VAL_260_O47_EN NULL
#define OPTION_VAL_270_O47_EN NULL
#define OPTION_VAL_280_O47_EN NULL
#define OPTION_VAL_290_O47_EN NULL
#define OPTION_VAL_300_O47_EN NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_EN NULL
#define OPTION_VAL_WHITE_EN NULL
#define OPTION_VAL_RED_EN NULL
#define OPTION_VAL_GREEN_EN NULL
#define OPTION_VAL_BLUE_EN NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_EN NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_EN NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_EN NULL
#define OPTION_VAL_VMU_EN NULL
#define OPTION_VAL_PURUPURU_EN NULL
#define OPTION_VAL_DREAMPOTATO_EN NULL
#define OPTION_VAL_NONE_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_EN NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_EN NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_EN NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_EN NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_EN NULL
#define OPTION_VAL_VMU_A1_EN NULL
#define OPTION_VAL_ALL_VMUS_EN NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_EN NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_EN NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_EN NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_EN NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_EN NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_EN NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_EN "Enable configuration of emulated VMU LCD screen visibility, size, position and colour. NOTE: Quick Menu may need to be toggled for this setting to take effect."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_EN NULL
#define OPTION_VAL_UPPER_LEFT_EN NULL
#define OPTION_VAL_UPPER_RIGHT_EN NULL
#define OPTION_VAL_LOWER_LEFT_EN NULL
#define OPTION_VAL_LOWER_RIGHT_EN NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_EN NULL
#define OPTION_VAL_1X_EN NULL
#define OPTION_VAL_3X_EN NULL
#define OPTION_VAL_5X_EN NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_EN NULL
#define OPTION_VAL_DEFAULT_ON_00_EN NULL
#define OPTION_VAL_DEFAULT_OFF_01_EN NULL
#define OPTION_VAL_BLACK_02_EN NULL
#define OPTION_VAL_LIGHT_BLUE_04_EN NULL
#define OPTION_VAL_CYAN_06_EN NULL
#define OPTION_VAL_CYAN_BLUE_07_EN NULL
#define OPTION_VAL_LIGHT_GREEN_08_EN NULL
#define OPTION_VAL_CYAN_GREEN_09_EN NULL
#define OPTION_VAL_LIGHT_CYAN_10_EN NULL
#define OPTION_VAL_PURPLE_12_EN NULL
#define OPTION_VAL_LIGHT_PURPLE_13_EN NULL
#define OPTION_VAL_YELLOW_14_EN NULL
#define OPTION_VAL_GRAY_15_EN "Grey"
#define OPTION_VAL_LIGHT_PURPLE_2_16_EN NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_EN NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_EN NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_EN NULL
#define OPTION_VAL_LIGHT_RED_2_20_EN NULL
#define OPTION_VAL_MAGENTA_21_EN NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_EN NULL
#define OPTION_VAL_LIGHT_ORANGE_23_EN NULL
#define OPTION_VAL_ORANGE_24_EN NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_EN NULL
#define OPTION_VAL_LIGHT_YELLOW_26_EN NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_EN NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_EN NULL
#define OPTION_VAL_40_EN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_EN NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_EN NULL

struct retro_core_option_v2_category option_cats_en[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_EN,
      CATEGORY_SYSTEM_INFO_0_EN
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_EN,
      CATEGORY_VIDEO_INFO_0_EN
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_EN,
      CATEGORY_PERFORMANCE_INFO_0_EN
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_EN,
      CATEGORY_HACKS_INFO_0_EN
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_EN,
      CATEGORY_INPUT_INFO_0_EN
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_EN,
      CATEGORY_EXPANSIONS_INFO_0_EN
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_EN,
      CATEGORY_VMU_INFO_0_EN
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_en[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_EN,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_EN },
         { "USA",     OPTION_VAL_USA_EN },
         { "Europe",  OPTION_VAL_EUROPE_EN },
         { "Default", OPTION_VAL_DEFAULT_EN },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_EN,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_EN },
         { "English",  OPTION_VAL_ENGLISH_EN },
         { "German",   OPTION_VAL_GERMAN_EN },
         { "French",   OPTION_VAL_FRENCH_EN },
         { "Spanish",  OPTION_VAL_SPANISH_EN },
         { "Italian",  OPTION_VAL_ITALIAN_EN },
         { "Default",  OPTION_VAL_DEFAULT_EN },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_EN,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_EN },
         { "640x480",    OPTION_VAL_640X480_EN },
         { "800x600",    OPTION_VAL_800X600_EN },
         { "960x720",    OPTION_VAL_960X720_EN },
         { "1024x768",   OPTION_VAL_1024X768_EN },
         { "1280x960",   OPTION_VAL_1280X960_EN },
         { "1440x1080",  OPTION_VAL_1440X1080_EN },
         { "1600x1200",  OPTION_VAL_1600X1200_EN },
         { "1920x1440",  OPTION_VAL_1920X1440_EN },
         { "2560x1920",  OPTION_VAL_2560X1920_EN },
         { "2880x2160",  OPTION_VAL_2880X2160_EN },
         { "3200x2400",  OPTION_VAL_3200X2400_EN },
         { "3840x2880",  OPTION_VAL_3840X2880_EN },
         { "4480x3360",  OPTION_VAL_4480X3360_EN },
         { "5120x3840",  OPTION_VAL_5120X3840_EN },
         { "5760x4320",  OPTION_VAL_5760X4320_EN },
         { "6400x4800",  OPTION_VAL_6400X4800_EN },
         { "7040x5280",  OPTION_VAL_7040X5280_EN },
         { "7680x5760",  OPTION_VAL_7680X5760_EN },
         { "8320x6240",  OPTION_VAL_8320X6240_EN },
         { "8960x6720",  OPTION_VAL_8960X6720_EN },
         { "9600x7200",  OPTION_VAL_9600X7200_EN },
         { "10240x7680", OPTION_VAL_10240X7680_EN },
         { "10880x8160", OPTION_VAL_10880X8160_EN },
         { "11520x8640", OPTION_VAL_11520X8640_EN },
         { "12160x9120", OPTION_VAL_12160X9120_EN },
         { "12800x9600", OPTION_VAL_12800X9600_EN },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_EN,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_EN },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_EN },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_EN },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_EN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_EN },
         { "PAL",     OPTION_VAL_PAL_EN },
         { "PAL_N",   OPTION_VAL_PAL_N_EN },
         { "PAL_M",   OPTION_VAL_PAL_M_EN },
         { "Default", OPTION_VAL_DEFAULT_EN },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_EN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_EN },
         { "vertical",   OPTION_VAL_VERTICAL_EN },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_EN,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_EN },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_EN },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_EN },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_EN,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_EN },
         { "1GB",   OPTION_VAL_1GB_EN },
         { "2GB",   OPTION_VAL_2GB_EN },
         { "4GB",   OPTION_VAL_4GB_EN },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_EN,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_EN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_EN,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_EN,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_EN },
         { "1",  OPTION_VAL_1_EN },
         { "2",  OPTION_VAL_2_EN },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_EN,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_EN },
         { "4",  OPTION_VAL_4_EN },
         { "6",  OPTION_VAL_6_EN },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_EN,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_EN,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_EN,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_EN,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_EN,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_EN,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_EN,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_EN },
         { "more", OPTION_VAL_MORE_EN },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_EN,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_EN },
         { "110", OPTION_VAL_110_EN },
         { "120", OPTION_VAL_120_EN },
         { "130", OPTION_VAL_130_EN },
         { "140", OPTION_VAL_140_EN },
         { "150", OPTION_VAL_150_EN },
         { "160", OPTION_VAL_160_EN },
         { "170", OPTION_VAL_170_EN },
         { "180", OPTION_VAL_180_EN },
         { "190", OPTION_VAL_190_EN },
         { "200", OPTION_VAL_200_EN },
         { "210", OPTION_VAL_210_EN },
         { "220", OPTION_VAL_220_EN },
         { "230", OPTION_VAL_230_EN },
         { "240", OPTION_VAL_240_EN },
         { "250", OPTION_VAL_250_EN },
         { "260", OPTION_VAL_260_EN },
         { "270", OPTION_VAL_270_EN },
         { "280", OPTION_VAL_280_EN },
         { "290", OPTION_VAL_290_EN },
         { "300", OPTION_VAL_300_EN },
         { "310", OPTION_VAL_310_EN },
         { "320", OPTION_VAL_320_EN },
         { "330", OPTION_VAL_330_EN },
         { "340", OPTION_VAL_340_EN },
         { "350", OPTION_VAL_350_EN },
         { "360", OPTION_VAL_360_EN },
         { "370", OPTION_VAL_370_EN },
         { "380", OPTION_VAL_380_EN },
         { "390", OPTION_VAL_390_EN },
         { "400", OPTION_VAL_400_EN },
         { "410", OPTION_VAL_410_EN },
         { "420", OPTION_VAL_420_EN },
         { "430", OPTION_VAL_430_EN },
         { "440", OPTION_VAL_440_EN },
         { "450", OPTION_VAL_450_EN },
         { "460", OPTION_VAL_460_EN },
         { "470", OPTION_VAL_470_EN },
         { "480", OPTION_VAL_480_EN },
         { "490", OPTION_VAL_490_EN },
         { "500", OPTION_VAL_500_EN },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_EN,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_EN },
         { "5%",  OPTION_VAL_5_EN },
         { "10%", OPTION_VAL_10_EN },
         { "15%", OPTION_VAL_15_EN },
         { "20%", OPTION_VAL_20_EN },
         { "25%", OPTION_VAL_25_EN },
         { "30%", OPTION_VAL_30_EN },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_EN,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_EN },
         { "5%",  OPTION_VAL_5_EN },
         { "10%", OPTION_VAL_10_EN },
         { "15%", OPTION_VAL_15_EN },
         { "20%", OPTION_VAL_20_EN },
         { "25%", OPTION_VAL_25_EN },
         { "30%", OPTION_VAL_30_EN },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_EN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_EN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_EN,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_EN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_EN },
         { "60%",  OPTION_VAL_60_EN },
         { "70%",  OPTION_VAL_70_EN },
         { "80%",  OPTION_VAL_80_EN },
         { "90%",  OPTION_VAL_90_EN },
         { "100%", OPTION_VAL_100_O47_EN },
         { "110%", OPTION_VAL_110_O47_EN },
         { "120%", OPTION_VAL_120_O47_EN },
         { "130%", OPTION_VAL_130_O47_EN },
         { "140%", OPTION_VAL_140_O47_EN },
         { "150%", OPTION_VAL_150_O47_EN },
         { "160%", OPTION_VAL_160_O47_EN },
         { "170%", OPTION_VAL_170_O47_EN },
         { "180%", OPTION_VAL_180_O47_EN },
         { "190%", OPTION_VAL_190_O47_EN },
         { "200%", OPTION_VAL_200_O47_EN },
         { "210%", OPTION_VAL_210_O47_EN },
         { "220%", OPTION_VAL_220_O47_EN },
         { "230%", OPTION_VAL_230_O47_EN },
         { "240%", OPTION_VAL_240_O47_EN },
         { "250%", OPTION_VAL_250_O47_EN },
         { "260%", OPTION_VAL_260_O47_EN },
         { "270%", OPTION_VAL_270_O47_EN },
         { "280%", OPTION_VAL_280_O47_EN },
         { "290%", OPTION_VAL_290_O47_EN },
         { "300%", OPTION_VAL_300_O47_EN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_EN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EN },
         { "Red",      OPTION_VAL_RED_EN },
         { "Green",    OPTION_VAL_GREEN_EN },
         { "Blue",     OPTION_VAL_BLUE_EN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_EN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EN },
         { "Red",      OPTION_VAL_RED_EN },
         { "Green",    OPTION_VAL_GREEN_EN },
         { "Blue",     OPTION_VAL_BLUE_EN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_EN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EN },
         { "Red",      OPTION_VAL_RED_EN },
         { "Green",    OPTION_VAL_GREEN_EN },
         { "Blue",     OPTION_VAL_BLUE_EN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_EN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EN },
         { "Red",      OPTION_VAL_RED_EN },
         { "Green",    OPTION_VAL_GREEN_EN },
         { "Blue",     OPTION_VAL_BLUE_EN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_EN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EN },
         { "Purupuru", OPTION_VAL_PURUPURU_EN },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_EN },
         { "None",     OPTION_VAL_NONE_EN },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_EN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EN },
         { "Purupuru", OPTION_VAL_PURUPURU_EN },
         { "None",     OPTION_VAL_NONE_EN },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_EN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EN },
         { "Purupuru", OPTION_VAL_PURUPURU_EN },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_EN },
         { "None",     OPTION_VAL_NONE_EN },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_EN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EN },
         { "Purupuru", OPTION_VAL_PURUPURU_EN },
         { "None",     OPTION_VAL_NONE_EN },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_EN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EN },
         { "Purupuru", OPTION_VAL_PURUPURU_EN },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_EN },
         { "None",     OPTION_VAL_NONE_EN },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_EN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EN },
         { "Purupuru", OPTION_VAL_PURUPURU_EN },
         { "None",     OPTION_VAL_NONE_EN },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_EN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EN },
         { "Purupuru", OPTION_VAL_PURUPURU_EN },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_EN },
         { "None",     OPTION_VAL_NONE_EN },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_EN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EN },
         { "Purupuru", OPTION_VAL_PURUPURU_EN },
         { "None",     OPTION_VAL_NONE_EN },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_EN,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_EN,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_EN,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_EN },
         { "All VMUs", OPTION_VAL_ALL_VMUS_EN },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_EN,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_EN,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_EN,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_EN,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_EN,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_EN,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_EN,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_EN,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EN },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EN },
         { "2x", OPTION_VAL_2_O26_EN },
         { "3x", OPTION_VAL_3X_EN },
         { "4x", OPTION_VAL_4_EN },
         { "5x", OPTION_VAL_5X_EN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EN },
         { "20%",  OPTION_VAL_20_EN },
         { "30%",  OPTION_VAL_30_EN },
         { "40%",  OPTION_VAL_40_EN },
         { "50%",  OPTION_VAL_50_EN },
         { "60%",  OPTION_VAL_60_EN },
         { "70%",  OPTION_VAL_70_EN },
         { "80%",  OPTION_VAL_80_EN },
         { "90%",  OPTION_VAL_90_EN },
         { "100%", OPTION_VAL_100_O47_EN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EN },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EN },
         { "2x", OPTION_VAL_2_O26_EN },
         { "3x", OPTION_VAL_3X_EN },
         { "4x", OPTION_VAL_4_EN },
         { "5x", OPTION_VAL_5X_EN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EN },
         { "20%",  OPTION_VAL_20_EN },
         { "30%",  OPTION_VAL_30_EN },
         { "40%",  OPTION_VAL_40_EN },
         { "50%",  OPTION_VAL_50_EN },
         { "60%",  OPTION_VAL_60_EN },
         { "70%",  OPTION_VAL_70_EN },
         { "80%",  OPTION_VAL_80_EN },
         { "90%",  OPTION_VAL_90_EN },
         { "100%", OPTION_VAL_100_O47_EN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EN },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EN },
         { "2x", OPTION_VAL_2_O26_EN },
         { "3x", OPTION_VAL_3X_EN },
         { "4x", OPTION_VAL_4_EN },
         { "5x", OPTION_VAL_5X_EN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EN },
         { "20%",  OPTION_VAL_20_EN },
         { "30%",  OPTION_VAL_30_EN },
         { "40%",  OPTION_VAL_40_EN },
         { "50%",  OPTION_VAL_50_EN },
         { "60%",  OPTION_VAL_60_EN },
         { "70%",  OPTION_VAL_70_EN },
         { "80%",  OPTION_VAL_80_EN },
         { "90%",  OPTION_VAL_90_EN },
         { "100%", OPTION_VAL_100_O47_EN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EN },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EN },
         { "2x", OPTION_VAL_2_O26_EN },
         { "3x", OPTION_VAL_3X_EN },
         { "4x", OPTION_VAL_4_EN },
         { "5x", OPTION_VAL_5X_EN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EN },
         { "BLUE 03",           OPTION_VAL_BLUE_EN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EN },
         { "GREEN 05",          OPTION_VAL_GREEN_EN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EN },
         { "RED 11",            OPTION_VAL_RED_EN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EN },
         { "WHITE 28",          OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_EN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EN },
         { "20%",  OPTION_VAL_20_EN },
         { "30%",  OPTION_VAL_30_EN },
         { "40%",  OPTION_VAL_40_EN },
         { "50%",  OPTION_VAL_50_EN },
         { "60%",  OPTION_VAL_60_EN },
         { "70%",  OPTION_VAL_70_EN },
         { "80%",  OPTION_VAL_80_EN },
         { "90%",  OPTION_VAL_90_EN },
         { "100%", OPTION_VAL_100_O47_EN },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_en = {
   option_cats_en,
   option_defs_en
};

/* RETRO_LANGUAGE_EO */

#define CATEGORY_SYSTEM_LABEL_EO NULL
#define CATEGORY_SYSTEM_INFO_0_EO NULL
#define CATEGORY_VIDEO_LABEL_EO "Video Driver"
#define CATEGORY_VIDEO_INFO_0_EO NULL
#define CATEGORY_PERFORMANCE_LABEL_EO NULL
#define CATEGORY_PERFORMANCE_INFO_0_EO NULL
#define CATEGORY_HACKS_LABEL_EO NULL
#define CATEGORY_HACKS_INFO_0_EO NULL
#define CATEGORY_INPUT_LABEL_EO "Input Driver"
#define CATEGORY_INPUT_INFO_0_EO NULL
#define CATEGORY_EXPANSIONS_LABEL_EO NULL
#define CATEGORY_EXPANSIONS_INFO_0_EO NULL
#define CATEGORY_VMU_LABEL_EO NULL
#define CATEGORY_VMU_INFO_0_EO NULL
#define CORE_OPTION_NAME_REGION_LABEL_EO "Regiono"
#define OPTION_VAL_JAPAN_EO NULL
#define OPTION_VAL_USA_EO NULL
#define OPTION_VAL_EUROPE_EO NULL
#define OPTION_VAL_DEFAULT_EO "Defaŭlta"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_EO "Lingvo"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_EO NULL
#define OPTION_VAL_JAPANESE_EO NULL
#define OPTION_VAL_ENGLISH_EO NULL
#define OPTION_VAL_GERMAN_EO NULL
#define OPTION_VAL_FRENCH_EO NULL
#define OPTION_VAL_SPANISH_EO NULL
#define OPTION_VAL_ITALIAN_EO NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_EO NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_EO NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_EO NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_EO NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_EO NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_EO NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_EO NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_EO NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_EO NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_EO NULL
#define CORE_OPTION_NAME_UPNP_LABEL_EO NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_EO NULL
#define CORE_OPTION_NAME_DCNET_LABEL_EO NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_EO NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_EO "Interna Rezolucio"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_EO NULL
#define OPTION_VAL_320X240_EO NULL
#define OPTION_VAL_640X480_EO NULL
#define OPTION_VAL_800X600_EO NULL
#define OPTION_VAL_960X720_EO NULL
#define OPTION_VAL_1024X768_EO NULL
#define OPTION_VAL_1280X960_EO NULL
#define OPTION_VAL_1440X1080_EO NULL
#define OPTION_VAL_1600X1200_EO NULL
#define OPTION_VAL_1920X1440_EO NULL
#define OPTION_VAL_2560X1920_EO NULL
#define OPTION_VAL_2880X2160_EO NULL
#define OPTION_VAL_3200X2400_EO NULL
#define OPTION_VAL_3840X2880_EO NULL
#define OPTION_VAL_4480X3360_EO NULL
#define OPTION_VAL_5120X3840_EO NULL
#define OPTION_VAL_5760X4320_EO NULL
#define OPTION_VAL_6400X4800_EO NULL
#define OPTION_VAL_7040X5280_EO NULL
#define OPTION_VAL_7680X5760_EO NULL
#define OPTION_VAL_8320X6240_EO NULL
#define OPTION_VAL_8960X6720_EO NULL
#define OPTION_VAL_9600X7200_EO NULL
#define OPTION_VAL_10240X7680_EO NULL
#define OPTION_VAL_10880X8160_EO NULL
#define OPTION_VAL_11520X8640_EO NULL
#define OPTION_VAL_12160X9120_EO NULL
#define OPTION_VAL_12800X9600_EO NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_EO NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_EO NULL
#define OPTION_VAL_VGA_EO NULL
#define OPTION_VAL_TV_RGB_EO NULL
#define OPTION_VAL_TV_COMPOSITE_EO NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_EO NULL
#define OPTION_VAL_NTSC_EO NULL
#define OPTION_VAL_PAL_EO NULL
#define OPTION_VAL_PAL_N_EO NULL
#define OPTION_VAL_PAL_M_EO NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_EO "Ekrana orientigo"
#define OPTION_VAL_HORIZONTAL_EO NULL
#define OPTION_VAL_VERTICAL_EO NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_EO NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_EO NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_EO NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_EO NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_EO NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_EO NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_EO NULL
#define OPTION_VAL_512MB_EO NULL
#define OPTION_VAL_1GB_EO NULL
#define OPTION_VAL_2GB_EO NULL
#define OPTION_VAL_4GB_EO NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_EO NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_EO NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_EO NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_EO NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_EO NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_EO NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_EO NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_EO NULL
#define CORE_OPTION_NAME_FOG_LABEL_EO NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_EO NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_EO NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_EO NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_EO NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_EO NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_EO NULL
#define OPTION_VAL_1_EO NULL
#define OPTION_VAL_2_EO NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_EO NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_EO NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_EO NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_EO NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_EO NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_EO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_EO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_EO NULL
#define OPTION_VAL_2_O26_EO NULL
#define OPTION_VAL_4_EO NULL
#define OPTION_VAL_6_EO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_EO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_EO NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_EO NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_EO NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_EO NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_EO NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_EO NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_EO NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_EO NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_EO NULL
#define OPTION_VAL_SOME_EO "Normala"
#define OPTION_VAL_MORE_EO NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_EO NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_EO NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_EO NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_EO NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_EO NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_EO NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_EO NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_EO NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_EO NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_EO NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_EO NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_EO NULL
#define OPTION_VAL_100_EO NULL
#define OPTION_VAL_110_EO NULL
#define OPTION_VAL_120_EO NULL
#define OPTION_VAL_130_EO NULL
#define OPTION_VAL_140_EO NULL
#define OPTION_VAL_150_EO NULL
#define OPTION_VAL_160_EO NULL
#define OPTION_VAL_170_EO NULL
#define OPTION_VAL_180_EO NULL
#define OPTION_VAL_190_EO NULL
#define OPTION_VAL_200_EO NULL
#define OPTION_VAL_210_EO NULL
#define OPTION_VAL_220_EO NULL
#define OPTION_VAL_230_EO NULL
#define OPTION_VAL_240_EO NULL
#define OPTION_VAL_250_EO NULL
#define OPTION_VAL_260_EO NULL
#define OPTION_VAL_270_EO NULL
#define OPTION_VAL_280_EO NULL
#define OPTION_VAL_290_EO NULL
#define OPTION_VAL_300_EO NULL
#define OPTION_VAL_310_EO NULL
#define OPTION_VAL_320_EO NULL
#define OPTION_VAL_330_EO NULL
#define OPTION_VAL_340_EO NULL
#define OPTION_VAL_350_EO NULL
#define OPTION_VAL_360_EO NULL
#define OPTION_VAL_370_EO NULL
#define OPTION_VAL_380_EO NULL
#define OPTION_VAL_390_EO NULL
#define OPTION_VAL_400_EO NULL
#define OPTION_VAL_410_EO NULL
#define OPTION_VAL_420_EO NULL
#define OPTION_VAL_430_EO NULL
#define OPTION_VAL_440_EO NULL
#define OPTION_VAL_450_EO NULL
#define OPTION_VAL_460_EO NULL
#define OPTION_VAL_470_EO NULL
#define OPTION_VAL_480_EO NULL
#define OPTION_VAL_490_EO NULL
#define OPTION_VAL_500_EO NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_EO NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_EO NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_EO NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_EO NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_EO NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_EO NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_EO NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_EO NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_EO NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_EO NULL
#define OPTION_VAL_0_EO NULL
#define OPTION_VAL_5_EO NULL
#define OPTION_VAL_10_EO NULL
#define OPTION_VAL_15_EO NULL
#define OPTION_VAL_20_EO NULL
#define OPTION_VAL_25_EO NULL
#define OPTION_VAL_30_EO NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_EO NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_EO NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_EO NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_EO NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_EO NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_EO NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_EO NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_EO NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_EO NULL
#define OPTION_VAL_50_EO NULL
#define OPTION_VAL_60_EO NULL
#define OPTION_VAL_70_EO NULL
#define OPTION_VAL_80_EO NULL
#define OPTION_VAL_90_EO NULL
#define OPTION_VAL_100_O47_EO NULL
#define OPTION_VAL_110_O47_EO NULL
#define OPTION_VAL_120_O47_EO NULL
#define OPTION_VAL_130_O47_EO NULL
#define OPTION_VAL_140_O47_EO NULL
#define OPTION_VAL_150_O47_EO NULL
#define OPTION_VAL_160_O47_EO NULL
#define OPTION_VAL_170_O47_EO NULL
#define OPTION_VAL_180_O47_EO NULL
#define OPTION_VAL_190_O47_EO NULL
#define OPTION_VAL_200_O47_EO NULL
#define OPTION_VAL_210_O47_EO NULL
#define OPTION_VAL_220_O47_EO NULL
#define OPTION_VAL_230_O47_EO NULL
#define OPTION_VAL_240_O47_EO NULL
#define OPTION_VAL_250_O47_EO NULL
#define OPTION_VAL_260_O47_EO NULL
#define OPTION_VAL_270_O47_EO NULL
#define OPTION_VAL_280_O47_EO NULL
#define OPTION_VAL_290_O47_EO NULL
#define OPTION_VAL_300_O47_EO NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_EO NULL
#define OPTION_VAL_WHITE_EO NULL
#define OPTION_VAL_RED_EO NULL
#define OPTION_VAL_GREEN_EO NULL
#define OPTION_VAL_BLUE_EO NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_EO NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_EO NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_EO NULL
#define OPTION_VAL_VMU_EO NULL
#define OPTION_VAL_PURUPURU_EO NULL
#define OPTION_VAL_DREAMPOTATO_EO NULL
#define OPTION_VAL_NONE_EO "Neniom"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_EO NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_EO NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_EO NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_EO NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_EO NULL
#define OPTION_VAL_VMU_A1_EO NULL
#define OPTION_VAL_ALL_VMUS_EO NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_EO NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_EO NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_EO NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_EO NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_EO NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_EO NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_EO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_EO NULL
#define OPTION_VAL_UPPER_LEFT_EO NULL
#define OPTION_VAL_UPPER_RIGHT_EO NULL
#define OPTION_VAL_LOWER_LEFT_EO NULL
#define OPTION_VAL_LOWER_RIGHT_EO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_EO NULL
#define OPTION_VAL_1X_EO NULL
#define OPTION_VAL_3X_EO NULL
#define OPTION_VAL_5X_EO NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_EO NULL
#define OPTION_VAL_DEFAULT_ON_00_EO NULL
#define OPTION_VAL_DEFAULT_OFF_01_EO NULL
#define OPTION_VAL_BLACK_02_EO NULL
#define OPTION_VAL_LIGHT_BLUE_04_EO NULL
#define OPTION_VAL_CYAN_06_EO NULL
#define OPTION_VAL_CYAN_BLUE_07_EO NULL
#define OPTION_VAL_LIGHT_GREEN_08_EO NULL
#define OPTION_VAL_CYAN_GREEN_09_EO NULL
#define OPTION_VAL_LIGHT_CYAN_10_EO NULL
#define OPTION_VAL_PURPLE_12_EO NULL
#define OPTION_VAL_LIGHT_PURPLE_13_EO NULL
#define OPTION_VAL_YELLOW_14_EO NULL
#define OPTION_VAL_GRAY_15_EO NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_EO NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_EO NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_EO NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_EO NULL
#define OPTION_VAL_LIGHT_RED_2_20_EO NULL
#define OPTION_VAL_MAGENTA_21_EO NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_EO NULL
#define OPTION_VAL_LIGHT_ORANGE_23_EO NULL
#define OPTION_VAL_ORANGE_24_EO NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_EO NULL
#define OPTION_VAL_LIGHT_YELLOW_26_EO NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_EO NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_EO NULL
#define OPTION_VAL_40_EO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_EO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_EO NULL

struct retro_core_option_v2_category option_cats_eo[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_EO,
      CATEGORY_SYSTEM_INFO_0_EO
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_EO,
      CATEGORY_VIDEO_INFO_0_EO
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_EO,
      CATEGORY_PERFORMANCE_INFO_0_EO
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_EO,
      CATEGORY_HACKS_INFO_0_EO
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_EO,
      CATEGORY_INPUT_INFO_0_EO
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_EO,
      CATEGORY_EXPANSIONS_INFO_0_EO
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_EO,
      CATEGORY_VMU_INFO_0_EO
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_eo[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_EO,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_EO },
         { "USA",     OPTION_VAL_USA_EO },
         { "Europe",  OPTION_VAL_EUROPE_EO },
         { "Default", OPTION_VAL_DEFAULT_EO },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_EO,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_EO },
         { "English",  OPTION_VAL_ENGLISH_EO },
         { "German",   OPTION_VAL_GERMAN_EO },
         { "French",   OPTION_VAL_FRENCH_EO },
         { "Spanish",  OPTION_VAL_SPANISH_EO },
         { "Italian",  OPTION_VAL_ITALIAN_EO },
         { "Default",  OPTION_VAL_DEFAULT_EO },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_EO,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_EO },
         { "640x480",    OPTION_VAL_640X480_EO },
         { "800x600",    OPTION_VAL_800X600_EO },
         { "960x720",    OPTION_VAL_960X720_EO },
         { "1024x768",   OPTION_VAL_1024X768_EO },
         { "1280x960",   OPTION_VAL_1280X960_EO },
         { "1440x1080",  OPTION_VAL_1440X1080_EO },
         { "1600x1200",  OPTION_VAL_1600X1200_EO },
         { "1920x1440",  OPTION_VAL_1920X1440_EO },
         { "2560x1920",  OPTION_VAL_2560X1920_EO },
         { "2880x2160",  OPTION_VAL_2880X2160_EO },
         { "3200x2400",  OPTION_VAL_3200X2400_EO },
         { "3840x2880",  OPTION_VAL_3840X2880_EO },
         { "4480x3360",  OPTION_VAL_4480X3360_EO },
         { "5120x3840",  OPTION_VAL_5120X3840_EO },
         { "5760x4320",  OPTION_VAL_5760X4320_EO },
         { "6400x4800",  OPTION_VAL_6400X4800_EO },
         { "7040x5280",  OPTION_VAL_7040X5280_EO },
         { "7680x5760",  OPTION_VAL_7680X5760_EO },
         { "8320x6240",  OPTION_VAL_8320X6240_EO },
         { "8960x6720",  OPTION_VAL_8960X6720_EO },
         { "9600x7200",  OPTION_VAL_9600X7200_EO },
         { "10240x7680", OPTION_VAL_10240X7680_EO },
         { "10880x8160", OPTION_VAL_10880X8160_EO },
         { "11520x8640", OPTION_VAL_11520X8640_EO },
         { "12160x9120", OPTION_VAL_12160X9120_EO },
         { "12800x9600", OPTION_VAL_12800X9600_EO },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_EO,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_EO },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_EO },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_EO },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_EO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_EO },
         { "PAL",     OPTION_VAL_PAL_EO },
         { "PAL_N",   OPTION_VAL_PAL_N_EO },
         { "PAL_M",   OPTION_VAL_PAL_M_EO },
         { "Default", OPTION_VAL_DEFAULT_EO },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_EO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_EO },
         { "vertical",   OPTION_VAL_VERTICAL_EO },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_EO,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_EO },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_EO },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_EO },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_EO,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_EO },
         { "1GB",   OPTION_VAL_1GB_EO },
         { "2GB",   OPTION_VAL_2GB_EO },
         { "4GB",   OPTION_VAL_4GB_EO },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_EO,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_EO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_EO,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_EO,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_EO },
         { "1",  OPTION_VAL_1_EO },
         { "2",  OPTION_VAL_2_EO },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_EO,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_EO },
         { "4",  OPTION_VAL_4_EO },
         { "6",  OPTION_VAL_6_EO },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_EO,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_EO,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_EO,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_EO,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_EO,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_EO,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_EO,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_EO },
         { "more", OPTION_VAL_MORE_EO },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_EO,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_EO },
         { "110", OPTION_VAL_110_EO },
         { "120", OPTION_VAL_120_EO },
         { "130", OPTION_VAL_130_EO },
         { "140", OPTION_VAL_140_EO },
         { "150", OPTION_VAL_150_EO },
         { "160", OPTION_VAL_160_EO },
         { "170", OPTION_VAL_170_EO },
         { "180", OPTION_VAL_180_EO },
         { "190", OPTION_VAL_190_EO },
         { "200", OPTION_VAL_200_EO },
         { "210", OPTION_VAL_210_EO },
         { "220", OPTION_VAL_220_EO },
         { "230", OPTION_VAL_230_EO },
         { "240", OPTION_VAL_240_EO },
         { "250", OPTION_VAL_250_EO },
         { "260", OPTION_VAL_260_EO },
         { "270", OPTION_VAL_270_EO },
         { "280", OPTION_VAL_280_EO },
         { "290", OPTION_VAL_290_EO },
         { "300", OPTION_VAL_300_EO },
         { "310", OPTION_VAL_310_EO },
         { "320", OPTION_VAL_320_EO },
         { "330", OPTION_VAL_330_EO },
         { "340", OPTION_VAL_340_EO },
         { "350", OPTION_VAL_350_EO },
         { "360", OPTION_VAL_360_EO },
         { "370", OPTION_VAL_370_EO },
         { "380", OPTION_VAL_380_EO },
         { "390", OPTION_VAL_390_EO },
         { "400", OPTION_VAL_400_EO },
         { "410", OPTION_VAL_410_EO },
         { "420", OPTION_VAL_420_EO },
         { "430", OPTION_VAL_430_EO },
         { "440", OPTION_VAL_440_EO },
         { "450", OPTION_VAL_450_EO },
         { "460", OPTION_VAL_460_EO },
         { "470", OPTION_VAL_470_EO },
         { "480", OPTION_VAL_480_EO },
         { "490", OPTION_VAL_490_EO },
         { "500", OPTION_VAL_500_EO },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_EO,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_EO },
         { "5%",  OPTION_VAL_5_EO },
         { "10%", OPTION_VAL_10_EO },
         { "15%", OPTION_VAL_15_EO },
         { "20%", OPTION_VAL_20_EO },
         { "25%", OPTION_VAL_25_EO },
         { "30%", OPTION_VAL_30_EO },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_EO,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_EO },
         { "5%",  OPTION_VAL_5_EO },
         { "10%", OPTION_VAL_10_EO },
         { "15%", OPTION_VAL_15_EO },
         { "20%", OPTION_VAL_20_EO },
         { "25%", OPTION_VAL_25_EO },
         { "30%", OPTION_VAL_30_EO },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_EO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_EO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_EO,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_EO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_EO },
         { "60%",  OPTION_VAL_60_EO },
         { "70%",  OPTION_VAL_70_EO },
         { "80%",  OPTION_VAL_80_EO },
         { "90%",  OPTION_VAL_90_EO },
         { "100%", OPTION_VAL_100_O47_EO },
         { "110%", OPTION_VAL_110_O47_EO },
         { "120%", OPTION_VAL_120_O47_EO },
         { "130%", OPTION_VAL_130_O47_EO },
         { "140%", OPTION_VAL_140_O47_EO },
         { "150%", OPTION_VAL_150_O47_EO },
         { "160%", OPTION_VAL_160_O47_EO },
         { "170%", OPTION_VAL_170_O47_EO },
         { "180%", OPTION_VAL_180_O47_EO },
         { "190%", OPTION_VAL_190_O47_EO },
         { "200%", OPTION_VAL_200_O47_EO },
         { "210%", OPTION_VAL_210_O47_EO },
         { "220%", OPTION_VAL_220_O47_EO },
         { "230%", OPTION_VAL_230_O47_EO },
         { "240%", OPTION_VAL_240_O47_EO },
         { "250%", OPTION_VAL_250_O47_EO },
         { "260%", OPTION_VAL_260_O47_EO },
         { "270%", OPTION_VAL_270_O47_EO },
         { "280%", OPTION_VAL_280_O47_EO },
         { "290%", OPTION_VAL_290_O47_EO },
         { "300%", OPTION_VAL_300_O47_EO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_EO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EO },
         { "Red",      OPTION_VAL_RED_EO },
         { "Green",    OPTION_VAL_GREEN_EO },
         { "Blue",     OPTION_VAL_BLUE_EO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_EO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EO },
         { "Red",      OPTION_VAL_RED_EO },
         { "Green",    OPTION_VAL_GREEN_EO },
         { "Blue",     OPTION_VAL_BLUE_EO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_EO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EO },
         { "Red",      OPTION_VAL_RED_EO },
         { "Green",    OPTION_VAL_GREEN_EO },
         { "Blue",     OPTION_VAL_BLUE_EO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_EO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_EO },
         { "Red",      OPTION_VAL_RED_EO },
         { "Green",    OPTION_VAL_GREEN_EO },
         { "Blue",     OPTION_VAL_BLUE_EO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_EO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EO },
         { "Purupuru", OPTION_VAL_PURUPURU_EO },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_EO },
         { "None",     OPTION_VAL_NONE_EO },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_EO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EO },
         { "Purupuru", OPTION_VAL_PURUPURU_EO },
         { "None",     OPTION_VAL_NONE_EO },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_EO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EO },
         { "Purupuru", OPTION_VAL_PURUPURU_EO },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_EO },
         { "None",     OPTION_VAL_NONE_EO },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_EO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EO },
         { "Purupuru", OPTION_VAL_PURUPURU_EO },
         { "None",     OPTION_VAL_NONE_EO },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_EO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EO },
         { "Purupuru", OPTION_VAL_PURUPURU_EO },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_EO },
         { "None",     OPTION_VAL_NONE_EO },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_EO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EO },
         { "Purupuru", OPTION_VAL_PURUPURU_EO },
         { "None",     OPTION_VAL_NONE_EO },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_EO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EO },
         { "Purupuru", OPTION_VAL_PURUPURU_EO },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_EO },
         { "None",     OPTION_VAL_NONE_EO },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_EO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_EO },
         { "Purupuru", OPTION_VAL_PURUPURU_EO },
         { "None",     OPTION_VAL_NONE_EO },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_EO,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_EO,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_EO,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_EO },
         { "All VMUs", OPTION_VAL_ALL_VMUS_EO },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_EO,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_EO,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_EO,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_EO,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_EO,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_EO,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_EO,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_EO,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EO },
         { "2x", OPTION_VAL_2_O26_EO },
         { "3x", OPTION_VAL_3X_EO },
         { "4x", OPTION_VAL_4_EO },
         { "5x", OPTION_VAL_5X_EO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EO },
         { "20%",  OPTION_VAL_20_EO },
         { "30%",  OPTION_VAL_30_EO },
         { "40%",  OPTION_VAL_40_EO },
         { "50%",  OPTION_VAL_50_EO },
         { "60%",  OPTION_VAL_60_EO },
         { "70%",  OPTION_VAL_70_EO },
         { "80%",  OPTION_VAL_80_EO },
         { "90%",  OPTION_VAL_90_EO },
         { "100%", OPTION_VAL_100_O47_EO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EO },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EO },
         { "2x", OPTION_VAL_2_O26_EO },
         { "3x", OPTION_VAL_3X_EO },
         { "4x", OPTION_VAL_4_EO },
         { "5x", OPTION_VAL_5X_EO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EO },
         { "20%",  OPTION_VAL_20_EO },
         { "30%",  OPTION_VAL_30_EO },
         { "40%",  OPTION_VAL_40_EO },
         { "50%",  OPTION_VAL_50_EO },
         { "60%",  OPTION_VAL_60_EO },
         { "70%",  OPTION_VAL_70_EO },
         { "80%",  OPTION_VAL_80_EO },
         { "90%",  OPTION_VAL_90_EO },
         { "100%", OPTION_VAL_100_O47_EO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EO },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EO },
         { "2x", OPTION_VAL_2_O26_EO },
         { "3x", OPTION_VAL_3X_EO },
         { "4x", OPTION_VAL_4_EO },
         { "5x", OPTION_VAL_5X_EO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EO },
         { "20%",  OPTION_VAL_20_EO },
         { "30%",  OPTION_VAL_30_EO },
         { "40%",  OPTION_VAL_40_EO },
         { "50%",  OPTION_VAL_50_EO },
         { "60%",  OPTION_VAL_60_EO },
         { "70%",  OPTION_VAL_70_EO },
         { "80%",  OPTION_VAL_80_EO },
         { "90%",  OPTION_VAL_90_EO },
         { "100%", OPTION_VAL_100_O47_EO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_EO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_EO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_EO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_EO },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_EO },
         { "2x", OPTION_VAL_2_O26_EO },
         { "3x", OPTION_VAL_3X_EO },
         { "4x", OPTION_VAL_4_EO },
         { "5x", OPTION_VAL_5X_EO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_EO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_EO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_EO },
         { "BLUE 03",           OPTION_VAL_BLUE_EO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_EO },
         { "GREEN 05",          OPTION_VAL_GREEN_EO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_EO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_EO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_EO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_EO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_EO },
         { "RED 11",            OPTION_VAL_RED_EO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_EO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_EO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_EO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_EO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_EO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_EO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_EO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_EO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_EO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_EO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_EO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_EO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_EO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_EO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_EO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_EO },
         { "WHITE 28",          OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_EO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_EO },
         { "20%",  OPTION_VAL_20_EO },
         { "30%",  OPTION_VAL_30_EO },
         { "40%",  OPTION_VAL_40_EO },
         { "50%",  OPTION_VAL_50_EO },
         { "60%",  OPTION_VAL_60_EO },
         { "70%",  OPTION_VAL_70_EO },
         { "80%",  OPTION_VAL_80_EO },
         { "90%",  OPTION_VAL_90_EO },
         { "100%", OPTION_VAL_100_O47_EO },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_eo = {
   option_cats_eo,
   option_defs_eo
};

/* RETRO_LANGUAGE_ES */

#define CATEGORY_SYSTEM_LABEL_ES "Sistema"
#define CATEGORY_SYSTEM_INFO_0_ES "Cambia los ajustes de región, idioma, BIOS y del hardware base."
#define CATEGORY_VIDEO_LABEL_ES "Vídeo"
#define CATEGORY_VIDEO_INFO_0_ES "Cambia los ajustes de resolución, las transparencias independientes del orden y los efectos visuales."
#define CATEGORY_PERFORMANCE_LABEL_ES "Rendimiento"
#define CATEGORY_PERFORMANCE_INFO_0_ES "Configura los ajustes del renderizado multihilo y la omisión de fotogramas."
#define CATEGORY_HACKS_LABEL_ES "Arreglos de emulación"
#define CATEGORY_HACKS_INFO_0_ES "Cambia los ajustes de forzado de imagen panorámica, la velocidad de carga de GD-ROM y la sustitución de texturas."
#define CATEGORY_INPUT_LABEL_ES "Entrada"
#define CATEGORY_INPUT_INFO_0_ES "Cambia los ajustes de los mandos y las pistolas de luz."
#define CATEGORY_EXPANSIONS_LABEL_ES "Puertos de expansión de los mandos"
#define CATEGORY_EXPANSIONS_INFO_0_ES "Selecciona el dispositivo (Visual Memory/VMU/vibración) que irá conectado en cada puerto de expansión."
#define CATEGORY_VMU_LABEL_ES "Visual Memory"
#define CATEGORY_VMU_INFO_0_ES "Cambia los ajustes de los archivos de VMU individuales para cada juego y la visibilidad en pantalla de las mismas."
#define CORE_OPTION_NAME_REGION_LABEL_ES "Región"
#define OPTION_VAL_JAPAN_ES "Japón"
#define OPTION_VAL_USA_ES "Estados Unidos"
#define OPTION_VAL_EUROPE_ES "Europa"
#define OPTION_VAL_DEFAULT_ES "Valor predeterminado"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_ES "Idioma"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_ES "Cambia el idioma que utilizará la BIOS y cualquier juego que contenga varios idiomas."
#define OPTION_VAL_JAPANESE_ES "Japonés"
#define OPTION_VAL_ENGLISH_ES "Inglés"
#define OPTION_VAL_GERMAN_ES "Alemán"
#define OPTION_VAL_FRENCH_ES "Francés"
#define OPTION_VAL_SPANISH_ES "Español"
#define OPTION_VAL_ITALIAN_ES "Italiano"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_ES "BIOS HLE (es necesario reiniciar)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_ES "Fuerza el uso de una BIOS creada con emulación de alto nivel."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_ES "Activar DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_ES "Habilita la emulación del DSP (procesador de señales digitales) de audio de Dreamcast. Mejora la precisión del sonido generado a costa de aumentar los requisitos."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_ES "Activar botones de servicio de arcade"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_ES "Activa el botón SERVICE para los juegos para recreativas/arcades, con el que acceder a la configuración del mueble."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_ES "Partidas gratuitas en NAOMI"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_ES "Modifica la configuración de monedas del juego a partidas gratuitas."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_ES "Emulación del Broadband Adapter (BBA)"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_ES "Emula el adaptador de banda ancha por Ethernet en lugar del módem (es necesario reiniciar)."
#define CORE_OPTION_NAME_UPNP_LABEL_ES "Activar UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_ES "Utiliza UPnP para configurar automáticamente tu router/enrutador de Internet para poder conectarte a juegos online."
#define CORE_OPTION_NAME_DCNET_LABEL_ES "Utilizar DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_ES "Utiliza el servicio en la nube DCNet para conectar la Dreamcast a Internet."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_ES "Resolución interna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_ES "Modifica la resolución de renderizado."
#define OPTION_VAL_320X240_ES "320 × 240 (50 %)"
#define OPTION_VAL_640X480_ES "640 × 480 (nativa)"
#define OPTION_VAL_800X600_ES "800 × 600 (125 %)"
#define OPTION_VAL_960X720_ES "960 × 720 (150 %)"
#define OPTION_VAL_1024X768_ES "1024 × 768 (160 %)"
#define OPTION_VAL_1280X960_ES "1280 × 960 (200 %)"
#define OPTION_VAL_1440X1080_ES "1440 × 1080 (225 %)"
#define OPTION_VAL_1600X1200_ES "1600 × 1200 (250 %)"
#define OPTION_VAL_1920X1440_ES "1920 × 1440 (300 %)"
#define OPTION_VAL_2560X1920_ES "2560 × 1920 (400 %)"
#define OPTION_VAL_2880X2160_ES "2880 × 2160 (450 %)"
#define OPTION_VAL_3200X2400_ES "3200 × 2400 (500 %)"
#define OPTION_VAL_3840X2880_ES "3840 × 2880 (600 %)"
#define OPTION_VAL_4480X3360_ES "4480 × 3360 (700 %)"
#define OPTION_VAL_5120X3840_ES "5120 × 3840 (800 %)"
#define OPTION_VAL_5760X4320_ES "5760 × 4320 (900 %)"
#define OPTION_VAL_6400X4800_ES "6400 × 4800 (1000 %)"
#define OPTION_VAL_7040X5280_ES "7040 × 5280 (1100 %)"
#define OPTION_VAL_7680X5760_ES "7680 × 5760 (1200 %)"
#define OPTION_VAL_8320X6240_ES "8320 × 6240 (1300 %)"
#define OPTION_VAL_8960X6720_ES "8960 × 6720 (1400 %)"
#define OPTION_VAL_9600X7200_ES "9600 × 7200 (1500 %)"
#define OPTION_VAL_10240X7680_ES "10240 × 7680 (1600 %)"
#define OPTION_VAL_10880X8160_ES "10880 × 8160 (1700 %)"
#define OPTION_VAL_11520X8640_ES "11520 × 8640 (1800 %)"
#define OPTION_VAL_12160X9120_ES "12160 × 9120 (1900 %)"
#define OPTION_VAL_12800X9600_ES "12800 × 9600 (2000 %)"
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_ES "Tipo de cable"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_ES "El tipo de señal de salida de la imagen. «TV (Compuesto)» es el formato más compatible de todos."
#define OPTION_VAL_VGA_ES NULL
#define OPTION_VAL_TV_RGB_ES NULL
#define OPTION_VAL_TV_COMPOSITE_ES "TV (Compuesto)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_ES "Estándar de transmisión"
#define OPTION_VAL_NTSC_ES NULL
#define OPTION_VAL_PAL_ES "PAL (Resto del mundo)"
#define OPTION_VAL_PAL_N_ES NULL
#define OPTION_VAL_PAL_M_ES "PAL-M (Brasil)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_ES "Orientación de pantalla"
#define OPTION_VAL_HORIZONTAL_ES NULL
#define OPTION_VAL_VERTICAL_ES NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_ES "Ordenación de transparencias"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_ES "Selecciona el orden de dibujado de los polígonos transparentes."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_ES "Por bandas (rápida, menos precisa)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_ES "Por triángulos (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_ES "Por píxeles (precisa, pero lenta)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_ES "Tamaño del búfer de acumulación de píxeles"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_ES "Es posible que sea necesario un valor alto para que la imagen se muestre bien a resoluciones superiores a la nativa."
#define OPTION_VAL_512MB_ES "512 MB"
#define OPTION_VAL_1GB_ES "1 GB"
#define OPTION_VAL_2GB_ES "2 GB"
#define OPTION_VAL_4GB_ES "4 GB"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_ES "N.º máximo de capas transparentes"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_ES "Las escenas más complejas podrían necesitar un valor más alto."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_ES "Emulación completa del framebuffer"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_ES "Activa la emulación completa del framebuffer dentro de la VRAM. Ideal para aquellos juegos que lean o escriban directamente al framebuffer en la VRAM. Al activar esta opción, se forzará una resolución interna de 640 × 480 y el rendimiento podría verse gravemente afectado."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_ES "Activar el búfer RTT (renderizado a texturas)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_ES "Copia las texturas renderizadas en la GPU de vuelta a la VRAM. Esta opción suele dejarse activada para aquellos juegos que la requieran. Al activar esta opción, se desactivará el escalado de texturas y el rendimiento podría verse afectado."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_ES "«Mipmapping»"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_ES "Al activar esta opción, las texturas mostrarán una versión con un tamaño más reducido cuanto más lejos estén de la pantalla. Puede mejorar el rendimiento y reducir los efectos de distorsión."
#define CORE_OPTION_NAME_FOG_LABEL_ES "Efectos de niebla"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_ES "Modificador de volúmenes"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_ES "Una característica de la GPU de Dreamcast que suelen utilizar los juegos para dibujar las sombras de objetos. Esta opción debe activarse por norma general; apenas afecta al rendimiento."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_ES "Filtrado anisotrópico"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_ES "Mejora la calidad de las texturas en aquellas superficies que se encuentren en un ángulo oblicuo respecto a la cámara. Un valor muy alto necesitará de mayores recursos de la GPU. Es necesario reiniciar para aplicar los cambios en este ajuste."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_ES "Filtrado de texturas"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_ES "Indica el modo de filtrado de texturas que se utilizará. Puede utilizarse para forzar un modo de filtrado concreto en todas las texturas con el fin de conseguir un aspecto más definido (o suavizado) que el valor predeterminado. Cualquier valor que no sea el predeterminado puede provocar problemas de renderizado. Es necesario reiniciar para aplicar los cambios en este ajuste."
#define OPTION_VAL_1_ES "Forzar vecino más cercano"
#define OPTION_VAL_2_ES "Forzar lineal"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_ES "Retrasar cambio de fotogramas"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_ES "Ideal para evitar parpadeos constantes o vídeos con fallos gráficos. No se recomienda su uso en plataformas lentas."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_ES "Detectar cambios en la frecuencia de actualización"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_ES "Informa en el front-end cuando cambie la frecuencia de actualización interna (p. ej.: de 60 a 30 FPS). Mejora la cadencia de fotogramas en aquellos juegos que se ejecuten a 30 o 20 FPS fijos, pero esta opción debería desactivarse en aquellos juegos con frecuencias de fotogramas desbloqueadas (es decir, inestables, p. ej.: Ecco the Dolphin, Unreal Tournament...). NOTA: esta opción no estará disponible si se activa Omisión automática de fotogramas."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_ES "Filtro de posprocesamiento del PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_ES "Posprocesa la imagen renderizada para simular los efectos dedicados de la GPU PowerVR2 y las señales de vídeo analógico."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_ES "Reescalado de texturas (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_ES "Mejora los gráficos 2D de corte «pixel art» dibujado a mano. Utilizar solo con juegos en 2D pixelado."
#define OPTION_VAL_2_O26_ES NULL
#define OPTION_VAL_4_ES NULL
#define OPTION_VAL_6_ES NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_ES "Tamaño máximo del escalado de texturas"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_ES "Selecciona un tamaño máximo para el escalado de texturas. Si el tamaño de una textura es superior al valor seleccionado, no se reescalará."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_ES "Interpolación nativa de profundidad"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_ES "Mitiga la corrupción de texturas y los problemas de profundidad en GPU de AMD. En algunos casos podría ayudar también en GPU de Intel."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_ES "Corregir sangrado de bordes al escalar"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_ES "Ayuda cuando las texturas muestren bordes sangrados al escalarlas. Desactivar esta opción puede ser útil si hay distorsión de píxeles al escalar juegos 2D (MVC2, CVS, KOF, etcétera)."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_ES "Renderizado multihilo"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_ES "Ejecuta la GPU y la CPU en hilos distintos. Muy recomendado."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_ES "Omisión automática de fotogramas"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_ES "Omite fotogramas de forma automática cuando el emulador se esté ejecutando lentamente. Nota: este ajuste solo surtirá efecto cuando la opción «Renderizado multihilo» esté activada."
#define OPTION_VAL_SOME_ES NULL
#define OPTION_VAL_MORE_ES "Valor máximo"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_ES "Omisión de fotogramas"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_ES "Establece la cantidad de fotogramas que se omitirán tras cada fotograma que se muestre."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_ES "Trucos para pantallas panorámicas (es necesario reiniciar)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_ES "Activa los trucos que permiten a ciertos juegos mostrarse en formato panorámico."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_ES "Arreglo para pantallas panorámicas"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_ES "Dibuja la geometría más allá de la relación de aspecto normal de 4:3. Puede producir fallos gráficos en las zonas reveladas."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_ES "Lectura rápida de GD-ROM (inexacta)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_ES "Acelera la carga de los GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_ES "Mod de 32 MB de RAM para Dreamcast"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_ES "Activa el mod de 32 MB de RAM para Dreamcast. Podría afectar a la compatibilidad."
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_ES "Velocidad de la CPU SH4"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_ES "Cambia la velocidad de reloj de la CPU SH4 principal de los 200 MHz predeterminados. Una bajada de velocidad podría ayudar a plataformas lentas. Una subida podría acelerar la velocidad de fotogramas en ciertos juegos. Utilizar con precaución."
#define OPTION_VAL_100_ES "100 MHz"
#define OPTION_VAL_110_ES "110 MHz"
#define OPTION_VAL_120_ES "120 MHz"
#define OPTION_VAL_130_ES "130 MHz"
#define OPTION_VAL_140_ES "140 MHz"
#define OPTION_VAL_150_ES "150 MHz"
#define OPTION_VAL_160_ES "160 MHz"
#define OPTION_VAL_170_ES "170 MHz"
#define OPTION_VAL_180_ES "180 MHz"
#define OPTION_VAL_190_ES "190 MHz"
#define OPTION_VAL_200_ES "200 MHz"
#define OPTION_VAL_210_ES "210 MHz"
#define OPTION_VAL_220_ES "220 MHz"
#define OPTION_VAL_230_ES "230 MHz"
#define OPTION_VAL_240_ES "240 MHz"
#define OPTION_VAL_250_ES "250 MHz"
#define OPTION_VAL_260_ES "260 MHz"
#define OPTION_VAL_270_ES "270 MHz"
#define OPTION_VAL_280_ES "280 MHz"
#define OPTION_VAL_290_ES "290 MHz"
#define OPTION_VAL_300_ES "300 MHz"
#define OPTION_VAL_310_ES "310 MHz"
#define OPTION_VAL_320_ES "320 MHz"
#define OPTION_VAL_330_ES "330 MHz"
#define OPTION_VAL_340_ES "340 MHz"
#define OPTION_VAL_350_ES "350 MHz"
#define OPTION_VAL_360_ES "360 MHz"
#define OPTION_VAL_370_ES "370 MHz"
#define OPTION_VAL_380_ES "380 MHz"
#define OPTION_VAL_390_ES "390 MHz"
#define OPTION_VAL_400_ES "400 MHz"
#define OPTION_VAL_410_ES "410 MHz"
#define OPTION_VAL_420_ES "420 MHz"
#define OPTION_VAL_430_ES "430 MHz"
#define OPTION_VAL_440_ES "440 MHz"
#define OPTION_VAL_450_ES "450 MHz"
#define OPTION_VAL_460_ES "460 MHz"
#define OPTION_VAL_470_ES "470 MHz"
#define OPTION_VAL_480_ES "480 MHz"
#define OPTION_VAL_490_ES "490 MHz"
#define OPTION_VAL_500_ES "500 MHz"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_ES "Cargar texturas personalizadas"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_ES "Carga las texturas personalizadas que se encuentren en la carpeta «system/dc/textures/<id-del-juego>/»."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_ES "Precargar texturas personalizadas"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_ES "Carga las texturas personalizadas al ejecutar un juego. Podría mejorar el rendimiento, pero aumentará el consumo de memoria."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_ES "Volcar texturas"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_ES "Cada vez que aparezca una textura nueva en el juego, se guardará en formato .PNG dentro de la carpeta «system/dc/texdump/<id-del-juego>/»."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_ES "Volcar texturas ya reemplazadas"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_ES "Vuelca siempre las texturas que ya hayan sido reemplazadas por otras personalizadas."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_ES "Zona muerta del stick analógico"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_ES "Selecciona la distancia a partir de la cual se empezará a reconocer el movimiento del stick analógico."
#define OPTION_VAL_0_ES "0 %"
#define OPTION_VAL_5_ES "5 %"
#define OPTION_VAL_10_ES "10 %"
#define OPTION_VAL_15_ES "15 %"
#define OPTION_VAL_20_ES "20 %"
#define OPTION_VAL_25_ES "25 %"
#define OPTION_VAL_30_ES "30 %"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_ES "Zona muerta de los gatillos"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_ES "Selecciona la distancia a partir de la cual se empezará a reconocer la pulsación del gatillo."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_ES "Gatillos digitales"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_ES "Al activar esta opción, los gatillos funcionarán como botones normales, lo que significa que se procesarán como o pulsados por completo o sin pulsar, sin término medio."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_ES "Transmitir salidas digitales"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_ES "Transmite las salidas digitales y el estado de la vibración al puerto TCP 8000. Compatible con la opción de MAME «-output network»."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_ES "Mostrar ajustes de las pistolas de luz"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_ES "Permite configurar los puntos de mira de las pistolas de luz. NOTA: es necesario salir y volver a entrar del menú rápido para que el cambio surta efecto."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_ES "Escala del punto de mira de la pistola"
#define OPTION_VAL_50_ES "50 %"
#define OPTION_VAL_60_ES "60 %"
#define OPTION_VAL_70_ES "70 %"
#define OPTION_VAL_80_ES "80 %"
#define OPTION_VAL_90_ES "90 %"
#define OPTION_VAL_100_O47_ES "100 %"
#define OPTION_VAL_110_O47_ES "110 %"
#define OPTION_VAL_120_O47_ES "120 %"
#define OPTION_VAL_130_O47_ES "130 %"
#define OPTION_VAL_140_O47_ES "140 %"
#define OPTION_VAL_150_O47_ES "150 %"
#define OPTION_VAL_160_O47_ES "160 %"
#define OPTION_VAL_170_O47_ES "170 %"
#define OPTION_VAL_180_O47_ES "180 %"
#define OPTION_VAL_190_O47_ES "190 %"
#define OPTION_VAL_200_O47_ES "200 %"
#define OPTION_VAL_210_O47_ES "210 %"
#define OPTION_VAL_220_O47_ES "220 %"
#define OPTION_VAL_230_O47_ES "230 %"
#define OPTION_VAL_240_O47_ES "240 %"
#define OPTION_VAL_250_O47_ES "250 %"
#define OPTION_VAL_260_O47_ES "260 %"
#define OPTION_VAL_270_O47_ES "270 %"
#define OPTION_VAL_280_O47_ES "280 %"
#define OPTION_VAL_290_O47_ES "290 %"
#define OPTION_VAL_300_O47_ES "300 %"
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_ES "Mostrar punto de mira del arma 1"
#define OPTION_VAL_WHITE_ES "Blanco"
#define OPTION_VAL_RED_ES "Rojo"
#define OPTION_VAL_GREEN_ES "Verde"
#define OPTION_VAL_BLUE_ES "Azul"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_ES "Mostrar punto de mira del arma 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_ES "Mostrar punto de mira del arma 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_ES "Mostrar punto de mira del arma 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_ES "Dispositivo del puerto de expansión A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_ES "Selecciona el dispositivo conectado al puerto de expansión A1 (puerto de mando A, ranura 1)."
#define OPTION_VAL_VMU_ES "Visual Memory (VMU)"
#define OPTION_VAL_PURUPURU_ES "Vibration Pack (vibración)"
#define OPTION_VAL_DREAMPOTATO_ES NULL
#define OPTION_VAL_NONE_ES "No compartir"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_ES "Dispositivo del puerto de expansión A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_ES "Selecciona el dispositivo conectado al puerto de expansión A2 (puerto de mando A, ranura 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_ES "Dispositivo del puerto de expansión B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_ES "Selecciona el dispositivo conectado al puerto de expansión B1 (puerto de mando B, ranura 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_ES "Dispositivo del puerto de expansión B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_ES "Selecciona el dispositivo conectado al puerto de expansión B2 (puerto de mando B, ranura 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_ES "Dispositivo del puerto de expansión C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_ES "Selecciona el dispositivo conectado al puerto de expansión C1 (puerto de mando C, ranura 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_ES "Dispositivo del puerto de expansión C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_ES "Selecciona el dispositivo conectado al puerto de expansión C2 (puerto de mando C, ranura 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_ES "Dispositivo del puerto de expansión D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_ES "Selecciona el dispositivo conectado al puerto de expansión D1 (puerto de mando D, ranura 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_ES "Dispositivo del puerto de expansión D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_ES "Selecciona el dispositivo conectado al puerto de expansión D2 (puerto de mando D, ranura 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_ES "Separar Visual Memory (VMU) por juego"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_ES "VMUs individuales para cada juego"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_ES "Al desactivar esta opción, todos los juegos compartirán ocho archivos de guardado de VMU (A1/A2/B1/B2/C1/C2/D1/D2) que se encontrarán en la carpeta del sistema de RetroArch.\nEl ajuste «VMU A1» creará un archivo único de la VMU A1 en la carpeta de guardados de RetroArch para cada juego que se ejecute.\n«Todas las VMU» creará ocho archivos VMU únicos (A1/A2/B1/B2/C1/C2/D1/D2) para cada juego que se ejecute."
#define OPTION_VAL_VMU_A1_ES NULL
#define OPTION_VAL_ALL_VMUS_ES "Todas las VMUs"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_ES "Sonidos de las Visual Memory (VMU)"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_ES "Sonidos de Visual Memory/VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_ES "Al activar esta opción, se reproducirán los pitidos de las Visual Memory/VMU."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_ES "Almacenar VMU en DreamPotato"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_ES "Activa el almacenamiento de datos de la Visual Memory/VMU al emulador DreamPotato conectado. Al activar esta opción, no se utilizará el archivo de guardado local como Visual Memory/VMU."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_ES "Mostrar ajustes de visualización de las Visual Memory (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_ES "Mostrar ajustes de visualización de las VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_ES "Permite configurar la visibilidad, tamaño, posición y color de las pantallas LCD emuladas de las VMU. NOTA: es necesario salir y volver a entrar del menú rápido para que el cambio surta efecto."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_ES "Pantalla de la VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_ES "Posición de la pantalla de la VMU 1"
#define OPTION_VAL_UPPER_LEFT_ES "Esquina superior izquierda"
#define OPTION_VAL_UPPER_RIGHT_ES "Esquina superior derecha"
#define OPTION_VAL_LOWER_LEFT_ES "Esquina inferior izquierda"
#define OPTION_VAL_LOWER_RIGHT_ES "Esquina interior derecha"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_ES "Tamaño de la pantalla de la VMU 1"
#define OPTION_VAL_1X_ES NULL
#define OPTION_VAL_3X_ES NULL
#define OPTION_VAL_5X_ES NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_ES "Color de píxeles encendidos de la VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_ES "Predeterminado (encendido)"
#define OPTION_VAL_DEFAULT_OFF_01_ES "Predeterminado (apagado)"
#define OPTION_VAL_BLACK_02_ES "Negro"
#define OPTION_VAL_LIGHT_BLUE_04_ES "Azul claro"
#define OPTION_VAL_CYAN_06_ES "Cian"
#define OPTION_VAL_CYAN_BLUE_07_ES "Azul celeste"
#define OPTION_VAL_LIGHT_GREEN_08_ES "Verde claro"
#define OPTION_VAL_CYAN_GREEN_09_ES "Verde cian"
#define OPTION_VAL_LIGHT_CYAN_10_ES "Cian claro"
#define OPTION_VAL_PURPLE_12_ES "Violeta"
#define OPTION_VAL_LIGHT_PURPLE_13_ES "Violeta claro"
#define OPTION_VAL_YELLOW_14_ES "Amarillo"
#define OPTION_VAL_GRAY_15_ES "Gris"
#define OPTION_VAL_LIGHT_PURPLE_2_16_ES "Violeta claro (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_ES "Verde claro (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_ES "Verde claro (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_ES "Cian claro (2)"
#define OPTION_VAL_LIGHT_RED_2_20_ES "Rojo claro (2)"
#define OPTION_VAL_MAGENTA_21_ES NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_ES "Violeta claro (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_ES "Naranja claro"
#define OPTION_VAL_ORANGE_24_ES "Naranja"
#define OPTION_VAL_LIGHT_PURPLE_4_25_ES "Violeta claro (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_ES "Amarillo claro"
#define OPTION_VAL_LIGHT_YELLOW_2_27_ES "Amarillo claro (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_ES "Color de píxeles apagados de la VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_ES "Opacidad de la pantalla de la VMU 1"
#define OPTION_VAL_40_ES "40 %"
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_ES "Pantalla de la VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_ES "Posición de la pantalla de la VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_ES "Tamaño de la pantalla de la VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_ES "Color de píxeles encendidos de la VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_ES "Color de píxeles apagados de la VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_ES "Opacidad de la pantalla de la VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_ES "Pantalla de la VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_ES "Posición de la pantalla de la VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_ES "Tamaño de la pantalla de la VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_ES "Color de píxeles encendidos de la VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_ES "Color de píxeles apagados de la VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_ES "Opacidad de la pantalla de la VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_ES "Pantalla de la VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_ES "Posición de la pantalla de la VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_ES "Tamaño de la pantalla de la VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_ES "Color de píxeles encendidos de la VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_ES "Color de píxeles apagados de la VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_ES "Opacidad de la pantalla de la VMU 4"

struct retro_core_option_v2_category option_cats_es[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_ES,
      CATEGORY_SYSTEM_INFO_0_ES
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_ES,
      CATEGORY_VIDEO_INFO_0_ES
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_ES,
      CATEGORY_PERFORMANCE_INFO_0_ES
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_ES,
      CATEGORY_HACKS_INFO_0_ES
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_ES,
      CATEGORY_INPUT_INFO_0_ES
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_ES,
      CATEGORY_EXPANSIONS_INFO_0_ES
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_ES,
      CATEGORY_VMU_INFO_0_ES
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_es[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_ES,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_ES },
         { "USA",     OPTION_VAL_USA_ES },
         { "Europe",  OPTION_VAL_EUROPE_ES },
         { "Default", OPTION_VAL_DEFAULT_ES },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_ES,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_ES },
         { "English",  OPTION_VAL_ENGLISH_ES },
         { "German",   OPTION_VAL_GERMAN_ES },
         { "French",   OPTION_VAL_FRENCH_ES },
         { "Spanish",  OPTION_VAL_SPANISH_ES },
         { "Italian",  OPTION_VAL_ITALIAN_ES },
         { "Default",  OPTION_VAL_DEFAULT_ES },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_ES,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_ES },
         { "640x480",    OPTION_VAL_640X480_ES },
         { "800x600",    OPTION_VAL_800X600_ES },
         { "960x720",    OPTION_VAL_960X720_ES },
         { "1024x768",   OPTION_VAL_1024X768_ES },
         { "1280x960",   OPTION_VAL_1280X960_ES },
         { "1440x1080",  OPTION_VAL_1440X1080_ES },
         { "1600x1200",  OPTION_VAL_1600X1200_ES },
         { "1920x1440",  OPTION_VAL_1920X1440_ES },
         { "2560x1920",  OPTION_VAL_2560X1920_ES },
         { "2880x2160",  OPTION_VAL_2880X2160_ES },
         { "3200x2400",  OPTION_VAL_3200X2400_ES },
         { "3840x2880",  OPTION_VAL_3840X2880_ES },
         { "4480x3360",  OPTION_VAL_4480X3360_ES },
         { "5120x3840",  OPTION_VAL_5120X3840_ES },
         { "5760x4320",  OPTION_VAL_5760X4320_ES },
         { "6400x4800",  OPTION_VAL_6400X4800_ES },
         { "7040x5280",  OPTION_VAL_7040X5280_ES },
         { "7680x5760",  OPTION_VAL_7680X5760_ES },
         { "8320x6240",  OPTION_VAL_8320X6240_ES },
         { "8960x6720",  OPTION_VAL_8960X6720_ES },
         { "9600x7200",  OPTION_VAL_9600X7200_ES },
         { "10240x7680", OPTION_VAL_10240X7680_ES },
         { "10880x8160", OPTION_VAL_10880X8160_ES },
         { "11520x8640", OPTION_VAL_11520X8640_ES },
         { "12160x9120", OPTION_VAL_12160X9120_ES },
         { "12800x9600", OPTION_VAL_12800X9600_ES },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_ES,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_ES },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_ES },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_ES },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_ES,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_ES },
         { "PAL",     OPTION_VAL_PAL_ES },
         { "PAL_N",   OPTION_VAL_PAL_N_ES },
         { "PAL_M",   OPTION_VAL_PAL_M_ES },
         { "Default", OPTION_VAL_DEFAULT_ES },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_ES,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_ES },
         { "vertical",   OPTION_VAL_VERTICAL_ES },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_ES,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_ES },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_ES },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_ES },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_ES,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_ES },
         { "1GB",   OPTION_VAL_1GB_ES },
         { "2GB",   OPTION_VAL_2GB_ES },
         { "4GB",   OPTION_VAL_4GB_ES },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_ES,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_ES,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_ES,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_ES,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_ES },
         { "1",  OPTION_VAL_1_ES },
         { "2",  OPTION_VAL_2_ES },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_ES,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_ES },
         { "4",  OPTION_VAL_4_ES },
         { "6",  OPTION_VAL_6_ES },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_ES,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_ES,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_ES,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_ES,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_ES,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_ES,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_ES,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_ES },
         { "more", OPTION_VAL_MORE_ES },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_ES,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_ES },
         { "110", OPTION_VAL_110_ES },
         { "120", OPTION_VAL_120_ES },
         { "130", OPTION_VAL_130_ES },
         { "140", OPTION_VAL_140_ES },
         { "150", OPTION_VAL_150_ES },
         { "160", OPTION_VAL_160_ES },
         { "170", OPTION_VAL_170_ES },
         { "180", OPTION_VAL_180_ES },
         { "190", OPTION_VAL_190_ES },
         { "200", OPTION_VAL_200_ES },
         { "210", OPTION_VAL_210_ES },
         { "220", OPTION_VAL_220_ES },
         { "230", OPTION_VAL_230_ES },
         { "240", OPTION_VAL_240_ES },
         { "250", OPTION_VAL_250_ES },
         { "260", OPTION_VAL_260_ES },
         { "270", OPTION_VAL_270_ES },
         { "280", OPTION_VAL_280_ES },
         { "290", OPTION_VAL_290_ES },
         { "300", OPTION_VAL_300_ES },
         { "310", OPTION_VAL_310_ES },
         { "320", OPTION_VAL_320_ES },
         { "330", OPTION_VAL_330_ES },
         { "340", OPTION_VAL_340_ES },
         { "350", OPTION_VAL_350_ES },
         { "360", OPTION_VAL_360_ES },
         { "370", OPTION_VAL_370_ES },
         { "380", OPTION_VAL_380_ES },
         { "390", OPTION_VAL_390_ES },
         { "400", OPTION_VAL_400_ES },
         { "410", OPTION_VAL_410_ES },
         { "420", OPTION_VAL_420_ES },
         { "430", OPTION_VAL_430_ES },
         { "440", OPTION_VAL_440_ES },
         { "450", OPTION_VAL_450_ES },
         { "460", OPTION_VAL_460_ES },
         { "470", OPTION_VAL_470_ES },
         { "480", OPTION_VAL_480_ES },
         { "490", OPTION_VAL_490_ES },
         { "500", OPTION_VAL_500_ES },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_ES,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_ES },
         { "5%",  OPTION_VAL_5_ES },
         { "10%", OPTION_VAL_10_ES },
         { "15%", OPTION_VAL_15_ES },
         { "20%", OPTION_VAL_20_ES },
         { "25%", OPTION_VAL_25_ES },
         { "30%", OPTION_VAL_30_ES },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_ES,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_ES },
         { "5%",  OPTION_VAL_5_ES },
         { "10%", OPTION_VAL_10_ES },
         { "15%", OPTION_VAL_15_ES },
         { "20%", OPTION_VAL_20_ES },
         { "25%", OPTION_VAL_25_ES },
         { "30%", OPTION_VAL_30_ES },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_ES,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_ES,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_ES,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_ES,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_ES },
         { "60%",  OPTION_VAL_60_ES },
         { "70%",  OPTION_VAL_70_ES },
         { "80%",  OPTION_VAL_80_ES },
         { "90%",  OPTION_VAL_90_ES },
         { "100%", OPTION_VAL_100_O47_ES },
         { "110%", OPTION_VAL_110_O47_ES },
         { "120%", OPTION_VAL_120_O47_ES },
         { "130%", OPTION_VAL_130_O47_ES },
         { "140%", OPTION_VAL_140_O47_ES },
         { "150%", OPTION_VAL_150_O47_ES },
         { "160%", OPTION_VAL_160_O47_ES },
         { "170%", OPTION_VAL_170_O47_ES },
         { "180%", OPTION_VAL_180_O47_ES },
         { "190%", OPTION_VAL_190_O47_ES },
         { "200%", OPTION_VAL_200_O47_ES },
         { "210%", OPTION_VAL_210_O47_ES },
         { "220%", OPTION_VAL_220_O47_ES },
         { "230%", OPTION_VAL_230_O47_ES },
         { "240%", OPTION_VAL_240_O47_ES },
         { "250%", OPTION_VAL_250_O47_ES },
         { "260%", OPTION_VAL_260_O47_ES },
         { "270%", OPTION_VAL_270_O47_ES },
         { "280%", OPTION_VAL_280_O47_ES },
         { "290%", OPTION_VAL_290_O47_ES },
         { "300%", OPTION_VAL_300_O47_ES },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_ES,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ES },
         { "Red",      OPTION_VAL_RED_ES },
         { "Green",    OPTION_VAL_GREEN_ES },
         { "Blue",     OPTION_VAL_BLUE_ES },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_ES,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ES },
         { "Red",      OPTION_VAL_RED_ES },
         { "Green",    OPTION_VAL_GREEN_ES },
         { "Blue",     OPTION_VAL_BLUE_ES },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_ES,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ES },
         { "Red",      OPTION_VAL_RED_ES },
         { "Green",    OPTION_VAL_GREEN_ES },
         { "Blue",     OPTION_VAL_BLUE_ES },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_ES,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ES },
         { "Red",      OPTION_VAL_RED_ES },
         { "Green",    OPTION_VAL_GREEN_ES },
         { "Blue",     OPTION_VAL_BLUE_ES },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_ES,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ES },
         { "Purupuru", OPTION_VAL_PURUPURU_ES },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_ES },
         { "None",     OPTION_VAL_NONE_ES },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_ES,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ES },
         { "Purupuru", OPTION_VAL_PURUPURU_ES },
         { "None",     OPTION_VAL_NONE_ES },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_ES,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ES },
         { "Purupuru", OPTION_VAL_PURUPURU_ES },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_ES },
         { "None",     OPTION_VAL_NONE_ES },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_ES,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ES },
         { "Purupuru", OPTION_VAL_PURUPURU_ES },
         { "None",     OPTION_VAL_NONE_ES },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_ES,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ES },
         { "Purupuru", OPTION_VAL_PURUPURU_ES },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_ES },
         { "None",     OPTION_VAL_NONE_ES },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_ES,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ES },
         { "Purupuru", OPTION_VAL_PURUPURU_ES },
         { "None",     OPTION_VAL_NONE_ES },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_ES,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ES },
         { "Purupuru", OPTION_VAL_PURUPURU_ES },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_ES },
         { "None",     OPTION_VAL_NONE_ES },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_ES,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ES },
         { "Purupuru", OPTION_VAL_PURUPURU_ES },
         { "None",     OPTION_VAL_NONE_ES },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_ES,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_ES,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_ES,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_ES },
         { "All VMUs", OPTION_VAL_ALL_VMUS_ES },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_ES,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_ES,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_ES,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_ES,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_ES,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_ES,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_ES,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_ES,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ES },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ES },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ES },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ES },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ES },
         { "2x", OPTION_VAL_2_O26_ES },
         { "3x", OPTION_VAL_3X_ES },
         { "4x", OPTION_VAL_4_ES },
         { "5x", OPTION_VAL_5X_ES },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ES },
         { "20%",  OPTION_VAL_20_ES },
         { "30%",  OPTION_VAL_30_ES },
         { "40%",  OPTION_VAL_40_ES },
         { "50%",  OPTION_VAL_50_ES },
         { "60%",  OPTION_VAL_60_ES },
         { "70%",  OPTION_VAL_70_ES },
         { "80%",  OPTION_VAL_80_ES },
         { "90%",  OPTION_VAL_90_ES },
         { "100%", OPTION_VAL_100_O47_ES },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ES },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ES },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ES },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ES },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ES },
         { "2x", OPTION_VAL_2_O26_ES },
         { "3x", OPTION_VAL_3X_ES },
         { "4x", OPTION_VAL_4_ES },
         { "5x", OPTION_VAL_5X_ES },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ES },
         { "20%",  OPTION_VAL_20_ES },
         { "30%",  OPTION_VAL_30_ES },
         { "40%",  OPTION_VAL_40_ES },
         { "50%",  OPTION_VAL_50_ES },
         { "60%",  OPTION_VAL_60_ES },
         { "70%",  OPTION_VAL_70_ES },
         { "80%",  OPTION_VAL_80_ES },
         { "90%",  OPTION_VAL_90_ES },
         { "100%", OPTION_VAL_100_O47_ES },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ES },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ES },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ES },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ES },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ES },
         { "2x", OPTION_VAL_2_O26_ES },
         { "3x", OPTION_VAL_3X_ES },
         { "4x", OPTION_VAL_4_ES },
         { "5x", OPTION_VAL_5X_ES },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ES },
         { "20%",  OPTION_VAL_20_ES },
         { "30%",  OPTION_VAL_30_ES },
         { "40%",  OPTION_VAL_40_ES },
         { "50%",  OPTION_VAL_50_ES },
         { "60%",  OPTION_VAL_60_ES },
         { "70%",  OPTION_VAL_70_ES },
         { "80%",  OPTION_VAL_80_ES },
         { "90%",  OPTION_VAL_90_ES },
         { "100%", OPTION_VAL_100_O47_ES },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ES },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ES },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ES },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ES },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ES },
         { "2x", OPTION_VAL_2_O26_ES },
         { "3x", OPTION_VAL_3X_ES },
         { "4x", OPTION_VAL_4_ES },
         { "5x", OPTION_VAL_5X_ES },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ES },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ES },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ES },
         { "BLUE 03",           OPTION_VAL_BLUE_ES },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ES },
         { "GREEN 05",          OPTION_VAL_GREEN_ES },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ES },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ES },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ES },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ES },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ES },
         { "RED 11",            OPTION_VAL_RED_ES },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ES },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ES },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ES },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ES },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ES },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ES },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ES },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ES },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ES },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ES },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ES },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ES },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ES },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ES },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ES },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ES },
         { "WHITE 28",          OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_ES,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ES },
         { "20%",  OPTION_VAL_20_ES },
         { "30%",  OPTION_VAL_30_ES },
         { "40%",  OPTION_VAL_40_ES },
         { "50%",  OPTION_VAL_50_ES },
         { "60%",  OPTION_VAL_60_ES },
         { "70%",  OPTION_VAL_70_ES },
         { "80%",  OPTION_VAL_80_ES },
         { "90%",  OPTION_VAL_90_ES },
         { "100%", OPTION_VAL_100_O47_ES },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_es = {
   option_cats_es,
   option_defs_es
};

/* RETRO_LANGUAGE_FA */

#define CATEGORY_SYSTEM_LABEL_FA "سیستم"
#define CATEGORY_SYSTEM_INFO_0_FA NULL
#define CATEGORY_VIDEO_LABEL_FA "ویدیو"
#define CATEGORY_VIDEO_INFO_0_FA NULL
#define CATEGORY_PERFORMANCE_LABEL_FA "عملکرد"
#define CATEGORY_PERFORMANCE_INFO_0_FA NULL
#define CATEGORY_HACKS_LABEL_FA NULL
#define CATEGORY_HACKS_INFO_0_FA NULL
#define CATEGORY_INPUT_LABEL_FA "داده"
#define CATEGORY_INPUT_INFO_0_FA NULL
#define CATEGORY_EXPANSIONS_LABEL_FA NULL
#define CATEGORY_EXPANSIONS_INFO_0_FA NULL
#define CATEGORY_VMU_LABEL_FA NULL
#define CATEGORY_VMU_INFO_0_FA NULL
#define CORE_OPTION_NAME_REGION_LABEL_FA "منطقه"
#define OPTION_VAL_JAPAN_FA "ژاپن"
#define OPTION_VAL_USA_FA "ایالات متحده آمریکا"
#define OPTION_VAL_EUROPE_FA "اروپا"
#define OPTION_VAL_DEFAULT_FA "پیش‌فرض"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_FA "زبان"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_FA NULL
#define OPTION_VAL_JAPANESE_FA "ژاپنی"
#define OPTION_VAL_ENGLISH_FA "انگلیسی"
#define OPTION_VAL_GERMAN_FA "آلمانی"
#define OPTION_VAL_FRENCH_FA "فرانسوی"
#define OPTION_VAL_SPANISH_FA "اسپانیایی"
#define OPTION_VAL_ITALIAN_FA "ایتالیایی"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_FA NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_FA NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_FA NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_FA NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_FA NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_FA NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_FA NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_FA NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_FA NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_FA NULL
#define CORE_OPTION_NAME_UPNP_LABEL_FA NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_FA NULL
#define CORE_OPTION_NAME_DCNET_LABEL_FA NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_FA NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_FA NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_FA NULL
#define OPTION_VAL_320X240_FA NULL
#define OPTION_VAL_640X480_FA NULL
#define OPTION_VAL_800X600_FA NULL
#define OPTION_VAL_960X720_FA NULL
#define OPTION_VAL_1024X768_FA NULL
#define OPTION_VAL_1280X960_FA NULL
#define OPTION_VAL_1440X1080_FA NULL
#define OPTION_VAL_1600X1200_FA NULL
#define OPTION_VAL_1920X1440_FA NULL
#define OPTION_VAL_2560X1920_FA NULL
#define OPTION_VAL_2880X2160_FA NULL
#define OPTION_VAL_3200X2400_FA NULL
#define OPTION_VAL_3840X2880_FA NULL
#define OPTION_VAL_4480X3360_FA NULL
#define OPTION_VAL_5120X3840_FA NULL
#define OPTION_VAL_5760X4320_FA NULL
#define OPTION_VAL_6400X4800_FA NULL
#define OPTION_VAL_7040X5280_FA NULL
#define OPTION_VAL_7680X5760_FA NULL
#define OPTION_VAL_8320X6240_FA NULL
#define OPTION_VAL_8960X6720_FA NULL
#define OPTION_VAL_9600X7200_FA NULL
#define OPTION_VAL_10240X7680_FA NULL
#define OPTION_VAL_10880X8160_FA NULL
#define OPTION_VAL_11520X8640_FA NULL
#define OPTION_VAL_12160X9120_FA NULL
#define OPTION_VAL_12800X9600_FA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_FA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_FA NULL
#define OPTION_VAL_VGA_FA NULL
#define OPTION_VAL_TV_RGB_FA NULL
#define OPTION_VAL_TV_COMPOSITE_FA NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_FA NULL
#define OPTION_VAL_NTSC_FA NULL
#define OPTION_VAL_PAL_FA NULL
#define OPTION_VAL_PAL_N_FA NULL
#define OPTION_VAL_PAL_M_FA NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_FA "جهت صفحه نمایش"
#define OPTION_VAL_HORIZONTAL_FA NULL
#define OPTION_VAL_VERTICAL_FA NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_FA NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_FA NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_FA NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_FA NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_FA NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_FA NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_FA NULL
#define OPTION_VAL_512MB_FA NULL
#define OPTION_VAL_1GB_FA NULL
#define OPTION_VAL_2GB_FA NULL
#define OPTION_VAL_4GB_FA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_FA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_FA NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_FA NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_FA NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_FA NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_FA NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_FA NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_FA NULL
#define CORE_OPTION_NAME_FOG_LABEL_FA NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_FA NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_FA NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_FA NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_FA NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_FA NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_FA NULL
#define OPTION_VAL_1_FA NULL
#define OPTION_VAL_2_FA NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_FA NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_FA NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_FA NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_FA NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_FA NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_FA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_FA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_FA NULL
#define OPTION_VAL_2_O26_FA NULL
#define OPTION_VAL_4_FA NULL
#define OPTION_VAL_6_FA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_FA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_FA NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_FA NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_FA NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_FA NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_FA NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_FA NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_FA NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_FA NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_FA NULL
#define OPTION_VAL_SOME_FA NULL
#define OPTION_VAL_MORE_FA NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_FA NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_FA NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_FA NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_FA NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_FA NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_FA NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_FA NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_FA NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_FA NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_FA NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_FA NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_FA NULL
#define OPTION_VAL_100_FA NULL
#define OPTION_VAL_110_FA NULL
#define OPTION_VAL_120_FA NULL
#define OPTION_VAL_130_FA NULL
#define OPTION_VAL_140_FA NULL
#define OPTION_VAL_150_FA NULL
#define OPTION_VAL_160_FA NULL
#define OPTION_VAL_170_FA NULL
#define OPTION_VAL_180_FA NULL
#define OPTION_VAL_190_FA NULL
#define OPTION_VAL_200_FA NULL
#define OPTION_VAL_210_FA NULL
#define OPTION_VAL_220_FA NULL
#define OPTION_VAL_230_FA NULL
#define OPTION_VAL_240_FA NULL
#define OPTION_VAL_250_FA NULL
#define OPTION_VAL_260_FA NULL
#define OPTION_VAL_270_FA NULL
#define OPTION_VAL_280_FA NULL
#define OPTION_VAL_290_FA NULL
#define OPTION_VAL_300_FA NULL
#define OPTION_VAL_310_FA NULL
#define OPTION_VAL_320_FA NULL
#define OPTION_VAL_330_FA NULL
#define OPTION_VAL_340_FA NULL
#define OPTION_VAL_350_FA NULL
#define OPTION_VAL_360_FA NULL
#define OPTION_VAL_370_FA NULL
#define OPTION_VAL_380_FA NULL
#define OPTION_VAL_390_FA NULL
#define OPTION_VAL_400_FA NULL
#define OPTION_VAL_410_FA NULL
#define OPTION_VAL_420_FA NULL
#define OPTION_VAL_430_FA NULL
#define OPTION_VAL_440_FA NULL
#define OPTION_VAL_450_FA NULL
#define OPTION_VAL_460_FA NULL
#define OPTION_VAL_470_FA NULL
#define OPTION_VAL_480_FA NULL
#define OPTION_VAL_490_FA NULL
#define OPTION_VAL_500_FA NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_FA NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_FA NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_FA NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_FA NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_FA NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_FA NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_FA NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_FA NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_FA NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_FA NULL
#define OPTION_VAL_0_FA NULL
#define OPTION_VAL_5_FA NULL
#define OPTION_VAL_10_FA NULL
#define OPTION_VAL_15_FA NULL
#define OPTION_VAL_20_FA NULL
#define OPTION_VAL_25_FA NULL
#define OPTION_VAL_30_FA NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_FA NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_FA NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_FA NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_FA NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_FA NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_FA NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_FA NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_FA NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_FA NULL
#define OPTION_VAL_50_FA NULL
#define OPTION_VAL_60_FA NULL
#define OPTION_VAL_70_FA NULL
#define OPTION_VAL_80_FA NULL
#define OPTION_VAL_90_FA NULL
#define OPTION_VAL_100_O47_FA NULL
#define OPTION_VAL_110_O47_FA NULL
#define OPTION_VAL_120_O47_FA NULL
#define OPTION_VAL_130_O47_FA NULL
#define OPTION_VAL_140_O47_FA NULL
#define OPTION_VAL_150_O47_FA NULL
#define OPTION_VAL_160_O47_FA NULL
#define OPTION_VAL_170_O47_FA NULL
#define OPTION_VAL_180_O47_FA NULL
#define OPTION_VAL_190_O47_FA NULL
#define OPTION_VAL_200_O47_FA NULL
#define OPTION_VAL_210_O47_FA NULL
#define OPTION_VAL_220_O47_FA NULL
#define OPTION_VAL_230_O47_FA NULL
#define OPTION_VAL_240_O47_FA NULL
#define OPTION_VAL_250_O47_FA NULL
#define OPTION_VAL_260_O47_FA NULL
#define OPTION_VAL_270_O47_FA NULL
#define OPTION_VAL_280_O47_FA NULL
#define OPTION_VAL_290_O47_FA NULL
#define OPTION_VAL_300_O47_FA NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_FA NULL
#define OPTION_VAL_WHITE_FA NULL
#define OPTION_VAL_RED_FA NULL
#define OPTION_VAL_GREEN_FA NULL
#define OPTION_VAL_BLUE_FA NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_FA NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_FA NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_FA NULL
#define OPTION_VAL_VMU_FA NULL
#define OPTION_VAL_PURUPURU_FA NULL
#define OPTION_VAL_DREAMPOTATO_FA NULL
#define OPTION_VAL_NONE_FA "هیچکدام"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_FA NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_FA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_FA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_FA NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_FA NULL
#define OPTION_VAL_VMU_A1_FA NULL
#define OPTION_VAL_ALL_VMUS_FA NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_FA NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_FA NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_FA NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_FA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_FA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_FA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_FA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_FA NULL
#define OPTION_VAL_UPPER_LEFT_FA NULL
#define OPTION_VAL_UPPER_RIGHT_FA NULL
#define OPTION_VAL_LOWER_LEFT_FA NULL
#define OPTION_VAL_LOWER_RIGHT_FA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_FA NULL
#define OPTION_VAL_1X_FA NULL
#define OPTION_VAL_3X_FA NULL
#define OPTION_VAL_5X_FA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_FA NULL
#define OPTION_VAL_DEFAULT_ON_00_FA NULL
#define OPTION_VAL_DEFAULT_OFF_01_FA NULL
#define OPTION_VAL_BLACK_02_FA NULL
#define OPTION_VAL_LIGHT_BLUE_04_FA NULL
#define OPTION_VAL_CYAN_06_FA NULL
#define OPTION_VAL_CYAN_BLUE_07_FA NULL
#define OPTION_VAL_LIGHT_GREEN_08_FA NULL
#define OPTION_VAL_CYAN_GREEN_09_FA NULL
#define OPTION_VAL_LIGHT_CYAN_10_FA NULL
#define OPTION_VAL_PURPLE_12_FA NULL
#define OPTION_VAL_LIGHT_PURPLE_13_FA NULL
#define OPTION_VAL_YELLOW_14_FA NULL
#define OPTION_VAL_GRAY_15_FA NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_FA NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_FA NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_FA NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_FA NULL
#define OPTION_VAL_LIGHT_RED_2_20_FA NULL
#define OPTION_VAL_MAGENTA_21_FA NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_FA NULL
#define OPTION_VAL_LIGHT_ORANGE_23_FA NULL
#define OPTION_VAL_ORANGE_24_FA NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_FA NULL
#define OPTION_VAL_LIGHT_YELLOW_26_FA NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_FA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_FA NULL
#define OPTION_VAL_40_FA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_FA NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_FA NULL

struct retro_core_option_v2_category option_cats_fa[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_FA,
      CATEGORY_SYSTEM_INFO_0_FA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_FA,
      CATEGORY_VIDEO_INFO_0_FA
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_FA,
      CATEGORY_PERFORMANCE_INFO_0_FA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_FA,
      CATEGORY_HACKS_INFO_0_FA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_FA,
      CATEGORY_INPUT_INFO_0_FA
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_FA,
      CATEGORY_EXPANSIONS_INFO_0_FA
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_FA,
      CATEGORY_VMU_INFO_0_FA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_fa[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_FA,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_FA },
         { "USA",     OPTION_VAL_USA_FA },
         { "Europe",  OPTION_VAL_EUROPE_FA },
         { "Default", OPTION_VAL_DEFAULT_FA },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_FA,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_FA },
         { "English",  OPTION_VAL_ENGLISH_FA },
         { "German",   OPTION_VAL_GERMAN_FA },
         { "French",   OPTION_VAL_FRENCH_FA },
         { "Spanish",  OPTION_VAL_SPANISH_FA },
         { "Italian",  OPTION_VAL_ITALIAN_FA },
         { "Default",  OPTION_VAL_DEFAULT_FA },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_FA,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_FA },
         { "640x480",    OPTION_VAL_640X480_FA },
         { "800x600",    OPTION_VAL_800X600_FA },
         { "960x720",    OPTION_VAL_960X720_FA },
         { "1024x768",   OPTION_VAL_1024X768_FA },
         { "1280x960",   OPTION_VAL_1280X960_FA },
         { "1440x1080",  OPTION_VAL_1440X1080_FA },
         { "1600x1200",  OPTION_VAL_1600X1200_FA },
         { "1920x1440",  OPTION_VAL_1920X1440_FA },
         { "2560x1920",  OPTION_VAL_2560X1920_FA },
         { "2880x2160",  OPTION_VAL_2880X2160_FA },
         { "3200x2400",  OPTION_VAL_3200X2400_FA },
         { "3840x2880",  OPTION_VAL_3840X2880_FA },
         { "4480x3360",  OPTION_VAL_4480X3360_FA },
         { "5120x3840",  OPTION_VAL_5120X3840_FA },
         { "5760x4320",  OPTION_VAL_5760X4320_FA },
         { "6400x4800",  OPTION_VAL_6400X4800_FA },
         { "7040x5280",  OPTION_VAL_7040X5280_FA },
         { "7680x5760",  OPTION_VAL_7680X5760_FA },
         { "8320x6240",  OPTION_VAL_8320X6240_FA },
         { "8960x6720",  OPTION_VAL_8960X6720_FA },
         { "9600x7200",  OPTION_VAL_9600X7200_FA },
         { "10240x7680", OPTION_VAL_10240X7680_FA },
         { "10880x8160", OPTION_VAL_10880X8160_FA },
         { "11520x8640", OPTION_VAL_11520X8640_FA },
         { "12160x9120", OPTION_VAL_12160X9120_FA },
         { "12800x9600", OPTION_VAL_12800X9600_FA },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_FA,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_FA },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_FA },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_FA },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_FA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_FA },
         { "PAL",     OPTION_VAL_PAL_FA },
         { "PAL_N",   OPTION_VAL_PAL_N_FA },
         { "PAL_M",   OPTION_VAL_PAL_M_FA },
         { "Default", OPTION_VAL_DEFAULT_FA },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_FA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_FA },
         { "vertical",   OPTION_VAL_VERTICAL_FA },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_FA,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_FA },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_FA },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_FA },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_FA,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_FA },
         { "1GB",   OPTION_VAL_1GB_FA },
         { "2GB",   OPTION_VAL_2GB_FA },
         { "4GB",   OPTION_VAL_4GB_FA },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_FA,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_FA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_FA,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_FA,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_FA },
         { "1",  OPTION_VAL_1_FA },
         { "2",  OPTION_VAL_2_FA },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_FA,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_FA },
         { "4",  OPTION_VAL_4_FA },
         { "6",  OPTION_VAL_6_FA },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_FA,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_FA,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_FA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_FA,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_FA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_FA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_FA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_FA },
         { "more", OPTION_VAL_MORE_FA },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_FA,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_FA },
         { "110", OPTION_VAL_110_FA },
         { "120", OPTION_VAL_120_FA },
         { "130", OPTION_VAL_130_FA },
         { "140", OPTION_VAL_140_FA },
         { "150", OPTION_VAL_150_FA },
         { "160", OPTION_VAL_160_FA },
         { "170", OPTION_VAL_170_FA },
         { "180", OPTION_VAL_180_FA },
         { "190", OPTION_VAL_190_FA },
         { "200", OPTION_VAL_200_FA },
         { "210", OPTION_VAL_210_FA },
         { "220", OPTION_VAL_220_FA },
         { "230", OPTION_VAL_230_FA },
         { "240", OPTION_VAL_240_FA },
         { "250", OPTION_VAL_250_FA },
         { "260", OPTION_VAL_260_FA },
         { "270", OPTION_VAL_270_FA },
         { "280", OPTION_VAL_280_FA },
         { "290", OPTION_VAL_290_FA },
         { "300", OPTION_VAL_300_FA },
         { "310", OPTION_VAL_310_FA },
         { "320", OPTION_VAL_320_FA },
         { "330", OPTION_VAL_330_FA },
         { "340", OPTION_VAL_340_FA },
         { "350", OPTION_VAL_350_FA },
         { "360", OPTION_VAL_360_FA },
         { "370", OPTION_VAL_370_FA },
         { "380", OPTION_VAL_380_FA },
         { "390", OPTION_VAL_390_FA },
         { "400", OPTION_VAL_400_FA },
         { "410", OPTION_VAL_410_FA },
         { "420", OPTION_VAL_420_FA },
         { "430", OPTION_VAL_430_FA },
         { "440", OPTION_VAL_440_FA },
         { "450", OPTION_VAL_450_FA },
         { "460", OPTION_VAL_460_FA },
         { "470", OPTION_VAL_470_FA },
         { "480", OPTION_VAL_480_FA },
         { "490", OPTION_VAL_490_FA },
         { "500", OPTION_VAL_500_FA },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_FA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_FA },
         { "5%",  OPTION_VAL_5_FA },
         { "10%", OPTION_VAL_10_FA },
         { "15%", OPTION_VAL_15_FA },
         { "20%", OPTION_VAL_20_FA },
         { "25%", OPTION_VAL_25_FA },
         { "30%", OPTION_VAL_30_FA },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_FA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_FA },
         { "5%",  OPTION_VAL_5_FA },
         { "10%", OPTION_VAL_10_FA },
         { "15%", OPTION_VAL_15_FA },
         { "20%", OPTION_VAL_20_FA },
         { "25%", OPTION_VAL_25_FA },
         { "30%", OPTION_VAL_30_FA },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_FA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_FA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_FA,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_FA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_FA },
         { "60%",  OPTION_VAL_60_FA },
         { "70%",  OPTION_VAL_70_FA },
         { "80%",  OPTION_VAL_80_FA },
         { "90%",  OPTION_VAL_90_FA },
         { "100%", OPTION_VAL_100_O47_FA },
         { "110%", OPTION_VAL_110_O47_FA },
         { "120%", OPTION_VAL_120_O47_FA },
         { "130%", OPTION_VAL_130_O47_FA },
         { "140%", OPTION_VAL_140_O47_FA },
         { "150%", OPTION_VAL_150_O47_FA },
         { "160%", OPTION_VAL_160_O47_FA },
         { "170%", OPTION_VAL_170_O47_FA },
         { "180%", OPTION_VAL_180_O47_FA },
         { "190%", OPTION_VAL_190_O47_FA },
         { "200%", OPTION_VAL_200_O47_FA },
         { "210%", OPTION_VAL_210_O47_FA },
         { "220%", OPTION_VAL_220_O47_FA },
         { "230%", OPTION_VAL_230_O47_FA },
         { "240%", OPTION_VAL_240_O47_FA },
         { "250%", OPTION_VAL_250_O47_FA },
         { "260%", OPTION_VAL_260_O47_FA },
         { "270%", OPTION_VAL_270_O47_FA },
         { "280%", OPTION_VAL_280_O47_FA },
         { "290%", OPTION_VAL_290_O47_FA },
         { "300%", OPTION_VAL_300_O47_FA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_FA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FA },
         { "Red",      OPTION_VAL_RED_FA },
         { "Green",    OPTION_VAL_GREEN_FA },
         { "Blue",     OPTION_VAL_BLUE_FA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_FA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FA },
         { "Red",      OPTION_VAL_RED_FA },
         { "Green",    OPTION_VAL_GREEN_FA },
         { "Blue",     OPTION_VAL_BLUE_FA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_FA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FA },
         { "Red",      OPTION_VAL_RED_FA },
         { "Green",    OPTION_VAL_GREEN_FA },
         { "Blue",     OPTION_VAL_BLUE_FA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_FA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FA },
         { "Red",      OPTION_VAL_RED_FA },
         { "Green",    OPTION_VAL_GREEN_FA },
         { "Blue",     OPTION_VAL_BLUE_FA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_FA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FA },
         { "Purupuru", OPTION_VAL_PURUPURU_FA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_FA },
         { "None",     OPTION_VAL_NONE_FA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_FA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FA },
         { "Purupuru", OPTION_VAL_PURUPURU_FA },
         { "None",     OPTION_VAL_NONE_FA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_FA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FA },
         { "Purupuru", OPTION_VAL_PURUPURU_FA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_FA },
         { "None",     OPTION_VAL_NONE_FA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_FA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FA },
         { "Purupuru", OPTION_VAL_PURUPURU_FA },
         { "None",     OPTION_VAL_NONE_FA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_FA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FA },
         { "Purupuru", OPTION_VAL_PURUPURU_FA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_FA },
         { "None",     OPTION_VAL_NONE_FA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_FA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FA },
         { "Purupuru", OPTION_VAL_PURUPURU_FA },
         { "None",     OPTION_VAL_NONE_FA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_FA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FA },
         { "Purupuru", OPTION_VAL_PURUPURU_FA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_FA },
         { "None",     OPTION_VAL_NONE_FA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_FA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FA },
         { "Purupuru", OPTION_VAL_PURUPURU_FA },
         { "None",     OPTION_VAL_NONE_FA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_FA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_FA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_FA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_FA },
         { "All VMUs", OPTION_VAL_ALL_VMUS_FA },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_FA,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_FA,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_FA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_FA,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_FA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_FA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_FA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_FA,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FA },
         { "2x", OPTION_VAL_2_O26_FA },
         { "3x", OPTION_VAL_3X_FA },
         { "4x", OPTION_VAL_4_FA },
         { "5x", OPTION_VAL_5X_FA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FA },
         { "20%",  OPTION_VAL_20_FA },
         { "30%",  OPTION_VAL_30_FA },
         { "40%",  OPTION_VAL_40_FA },
         { "50%",  OPTION_VAL_50_FA },
         { "60%",  OPTION_VAL_60_FA },
         { "70%",  OPTION_VAL_70_FA },
         { "80%",  OPTION_VAL_80_FA },
         { "90%",  OPTION_VAL_90_FA },
         { "100%", OPTION_VAL_100_O47_FA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FA },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FA },
         { "2x", OPTION_VAL_2_O26_FA },
         { "3x", OPTION_VAL_3X_FA },
         { "4x", OPTION_VAL_4_FA },
         { "5x", OPTION_VAL_5X_FA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FA },
         { "20%",  OPTION_VAL_20_FA },
         { "30%",  OPTION_VAL_30_FA },
         { "40%",  OPTION_VAL_40_FA },
         { "50%",  OPTION_VAL_50_FA },
         { "60%",  OPTION_VAL_60_FA },
         { "70%",  OPTION_VAL_70_FA },
         { "80%",  OPTION_VAL_80_FA },
         { "90%",  OPTION_VAL_90_FA },
         { "100%", OPTION_VAL_100_O47_FA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FA },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FA },
         { "2x", OPTION_VAL_2_O26_FA },
         { "3x", OPTION_VAL_3X_FA },
         { "4x", OPTION_VAL_4_FA },
         { "5x", OPTION_VAL_5X_FA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FA },
         { "20%",  OPTION_VAL_20_FA },
         { "30%",  OPTION_VAL_30_FA },
         { "40%",  OPTION_VAL_40_FA },
         { "50%",  OPTION_VAL_50_FA },
         { "60%",  OPTION_VAL_60_FA },
         { "70%",  OPTION_VAL_70_FA },
         { "80%",  OPTION_VAL_80_FA },
         { "90%",  OPTION_VAL_90_FA },
         { "100%", OPTION_VAL_100_O47_FA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FA },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FA },
         { "2x", OPTION_VAL_2_O26_FA },
         { "3x", OPTION_VAL_3X_FA },
         { "4x", OPTION_VAL_4_FA },
         { "5x", OPTION_VAL_5X_FA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FA },
         { "BLUE 03",           OPTION_VAL_BLUE_FA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FA },
         { "GREEN 05",          OPTION_VAL_GREEN_FA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FA },
         { "RED 11",            OPTION_VAL_RED_FA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FA },
         { "WHITE 28",          OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_FA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FA },
         { "20%",  OPTION_VAL_20_FA },
         { "30%",  OPTION_VAL_30_FA },
         { "40%",  OPTION_VAL_40_FA },
         { "50%",  OPTION_VAL_50_FA },
         { "60%",  OPTION_VAL_60_FA },
         { "70%",  OPTION_VAL_70_FA },
         { "80%",  OPTION_VAL_80_FA },
         { "90%",  OPTION_VAL_90_FA },
         { "100%", OPTION_VAL_100_O47_FA },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_fa = {
   option_cats_fa,
   option_defs_fa
};

/* RETRO_LANGUAGE_FI */

#define CATEGORY_SYSTEM_LABEL_FI "Järjestelmä"
#define CATEGORY_SYSTEM_INFO_0_FI NULL
#define CATEGORY_VIDEO_LABEL_FI NULL
#define CATEGORY_VIDEO_INFO_0_FI NULL
#define CATEGORY_PERFORMANCE_LABEL_FI "Suorituskyky"
#define CATEGORY_PERFORMANCE_INFO_0_FI NULL
#define CATEGORY_HACKS_LABEL_FI "Emulointikikat"
#define CATEGORY_HACKS_INFO_0_FI "Määritä laajakuvaohituksen, GD-ROM-latausnopeuden ja tekstuurikorvauksen asetukset."
#define CATEGORY_INPUT_LABEL_FI "Syöte"
#define CATEGORY_INPUT_INFO_0_FI NULL
#define CATEGORY_EXPANSIONS_LABEL_FI NULL
#define CATEGORY_EXPANSIONS_INFO_0_FI NULL
#define CATEGORY_VMU_LABEL_FI NULL
#define CATEGORY_VMU_INFO_0_FI NULL
#define CORE_OPTION_NAME_REGION_LABEL_FI "Alue"
#define OPTION_VAL_JAPAN_FI NULL
#define OPTION_VAL_USA_FI NULL
#define OPTION_VAL_EUROPE_FI "Eurooppa"
#define OPTION_VAL_DEFAULT_FI "Oletus"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_FI "Kieli"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_FI NULL
#define OPTION_VAL_JAPANESE_FI NULL
#define OPTION_VAL_ENGLISH_FI NULL
#define OPTION_VAL_GERMAN_FI NULL
#define OPTION_VAL_FRENCH_FI NULL
#define OPTION_VAL_SPANISH_FI NULL
#define OPTION_VAL_ITALIAN_FI NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_FI NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_FI NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_FI "Käytä DSP:tä"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_FI NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_FI NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_FI NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_FI NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_FI NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_FI "Laajakaista-adapterin emulaatio"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_FI NULL
#define CORE_OPTION_NAME_UPNP_LABEL_FI "Käytä UPnP:tä"
#define CORE_OPTION_NAME_UPNP_INFO_0_FI "Käytä UPnP:tä Internet-reitittimen määrittämiseksi verkkopelejä varten."
#define CORE_OPTION_NAME_DCNET_LABEL_FI "Käytä DCNetiä"
#define CORE_OPTION_NAME_DCNET_INFO_0_FI "Käytä DCNet-pilvipalvelua Dreamcastin Internet-käyttöön."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_FI "Sisäinen resoluutio"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_FI "Muokkaa renderöintitarkkuutta."
#define OPTION_VAL_320X240_FI "320x240 (Puolikas)"
#define OPTION_VAL_640X480_FI "640x480 (Natiivi)"
#define OPTION_VAL_800X600_FI NULL
#define OPTION_VAL_960X720_FI NULL
#define OPTION_VAL_1024X768_FI NULL
#define OPTION_VAL_1280X960_FI NULL
#define OPTION_VAL_1440X1080_FI NULL
#define OPTION_VAL_1600X1200_FI NULL
#define OPTION_VAL_1920X1440_FI NULL
#define OPTION_VAL_2560X1920_FI NULL
#define OPTION_VAL_2880X2160_FI NULL
#define OPTION_VAL_3200X2400_FI NULL
#define OPTION_VAL_3840X2880_FI NULL
#define OPTION_VAL_4480X3360_FI NULL
#define OPTION_VAL_5120X3840_FI NULL
#define OPTION_VAL_5760X4320_FI NULL
#define OPTION_VAL_6400X4800_FI NULL
#define OPTION_VAL_7040X5280_FI NULL
#define OPTION_VAL_7680X5760_FI NULL
#define OPTION_VAL_8320X6240_FI NULL
#define OPTION_VAL_8960X6720_FI NULL
#define OPTION_VAL_9600X7200_FI NULL
#define OPTION_VAL_10240X7680_FI NULL
#define OPTION_VAL_10880X8160_FI NULL
#define OPTION_VAL_11520X8640_FI NULL
#define OPTION_VAL_12160X9120_FI NULL
#define OPTION_VAL_12800X9600_FI NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_FI "Kaapelin tyyppi"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_FI NULL
#define OPTION_VAL_VGA_FI NULL
#define OPTION_VAL_TV_RGB_FI NULL
#define OPTION_VAL_TV_COMPOSITE_FI NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_FI NULL
#define OPTION_VAL_NTSC_FI NULL
#define OPTION_VAL_PAL_FI NULL
#define OPTION_VAL_PAL_N_FI "PAL-N (Argentiina, Paraguay, Uruguay)"
#define OPTION_VAL_PAL_M_FI "PAL-M (Brasilia)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_FI "Näytön suunta"
#define OPTION_VAL_HORIZONTAL_FI NULL
#define OPTION_VAL_VERTICAL_FI NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_FI NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_FI NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_FI NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_FI NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_FI NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_FI NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_FI NULL
#define OPTION_VAL_512MB_FI "512 Mt"
#define OPTION_VAL_1GB_FI "1 Gt"
#define OPTION_VAL_2GB_FI "2 Gt"
#define OPTION_VAL_4GB_FI "4 Gt"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_FI NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_FI NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_FI NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_FI NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_FI NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_FI NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_FI NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_FI NULL
#define CORE_OPTION_NAME_FOG_LABEL_FI "Sumutehosteet"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_FI NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_FI NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_FI "Anisotrooppinen suodatus"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_FI NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_FI "Tekstuurisuodatus"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_FI NULL
#define OPTION_VAL_1_FI NULL
#define OPTION_VAL_2_FI NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_FI NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_FI NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_FI NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_FI NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_FI "PowerVR2-jälkikäsittelysuodatin"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_FI NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_FI "Tekstuurin yliskaalaus (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_FI NULL
#define OPTION_VAL_2_O26_FI NULL
#define OPTION_VAL_4_FI NULL
#define OPTION_VAL_6_FI NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_FI NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_FI NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_FI NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_FI NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_FI NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_FI NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_FI "Säikeistetty renderöinti"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_FI NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_FI NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_FI NULL
#define OPTION_VAL_SOME_FI "Normaali"
#define OPTION_VAL_MORE_FI NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_FI NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_FI NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_FI NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_FI NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_FI NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_FI NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_FI NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_FI NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_FI NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_FI NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_FI "SH4-suorittimen ali-/ylikellotus"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_FI NULL
#define OPTION_VAL_100_FI NULL
#define OPTION_VAL_110_FI NULL
#define OPTION_VAL_120_FI NULL
#define OPTION_VAL_130_FI NULL
#define OPTION_VAL_140_FI NULL
#define OPTION_VAL_150_FI NULL
#define OPTION_VAL_160_FI NULL
#define OPTION_VAL_170_FI NULL
#define OPTION_VAL_180_FI NULL
#define OPTION_VAL_190_FI NULL
#define OPTION_VAL_200_FI NULL
#define OPTION_VAL_210_FI NULL
#define OPTION_VAL_220_FI NULL
#define OPTION_VAL_230_FI NULL
#define OPTION_VAL_240_FI NULL
#define OPTION_VAL_250_FI NULL
#define OPTION_VAL_260_FI NULL
#define OPTION_VAL_270_FI NULL
#define OPTION_VAL_280_FI NULL
#define OPTION_VAL_290_FI NULL
#define OPTION_VAL_300_FI NULL
#define OPTION_VAL_310_FI NULL
#define OPTION_VAL_320_FI NULL
#define OPTION_VAL_330_FI NULL
#define OPTION_VAL_340_FI NULL
#define OPTION_VAL_350_FI NULL
#define OPTION_VAL_360_FI NULL
#define OPTION_VAL_370_FI NULL
#define OPTION_VAL_380_FI NULL
#define OPTION_VAL_390_FI NULL
#define OPTION_VAL_400_FI NULL
#define OPTION_VAL_410_FI NULL
#define OPTION_VAL_420_FI NULL
#define OPTION_VAL_430_FI NULL
#define OPTION_VAL_440_FI NULL
#define OPTION_VAL_450_FI NULL
#define OPTION_VAL_460_FI NULL
#define OPTION_VAL_470_FI NULL
#define OPTION_VAL_480_FI NULL
#define OPTION_VAL_490_FI NULL
#define OPTION_VAL_500_FI NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_FI NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_FI NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_FI NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_FI NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_FI NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_FI NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_FI NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_FI NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_FI NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_FI NULL
#define OPTION_VAL_0_FI "0 %"
#define OPTION_VAL_5_FI "5 %"
#define OPTION_VAL_10_FI "10 %"
#define OPTION_VAL_15_FI "15 %"
#define OPTION_VAL_20_FI "20 %"
#define OPTION_VAL_25_FI "25 %"
#define OPTION_VAL_30_FI "30 %"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_FI NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_FI NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_FI NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_FI NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_FI NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_FI NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_FI "Näytä valoaseen asetukset"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_FI NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_FI "Aseen tähtäimen koko"
#define OPTION_VAL_50_FI "50 %"
#define OPTION_VAL_60_FI "60 %"
#define OPTION_VAL_70_FI "70 %"
#define OPTION_VAL_80_FI "80 %"
#define OPTION_VAL_90_FI "90 %"
#define OPTION_VAL_100_O47_FI "100 %"
#define OPTION_VAL_110_O47_FI "110 %"
#define OPTION_VAL_120_O47_FI "120 %"
#define OPTION_VAL_130_O47_FI "130 %"
#define OPTION_VAL_140_O47_FI "140 %"
#define OPTION_VAL_150_O47_FI "150 %"
#define OPTION_VAL_160_O47_FI "160 %"
#define OPTION_VAL_170_O47_FI "170 %"
#define OPTION_VAL_180_O47_FI "180 %"
#define OPTION_VAL_190_O47_FI "190 %"
#define OPTION_VAL_200_O47_FI "200 %"
#define OPTION_VAL_210_O47_FI "210 %"
#define OPTION_VAL_220_O47_FI "220 %"
#define OPTION_VAL_230_O47_FI "230 %"
#define OPTION_VAL_240_O47_FI "240 %"
#define OPTION_VAL_250_O47_FI "250 %"
#define OPTION_VAL_260_O47_FI "260 %"
#define OPTION_VAL_270_O47_FI "270 %"
#define OPTION_VAL_280_O47_FI "280 %"
#define OPTION_VAL_290_O47_FI "290 %"
#define OPTION_VAL_300_O47_FI "300 %"
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_FI NULL
#define OPTION_VAL_WHITE_FI "Valkoinen"
#define OPTION_VAL_RED_FI "Punainen"
#define OPTION_VAL_GREEN_FI "Vihreä"
#define OPTION_VAL_BLUE_FI "Sininen"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_FI NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_FI NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_FI NULL
#define OPTION_VAL_VMU_FI NULL
#define OPTION_VAL_PURUPURU_FI NULL
#define OPTION_VAL_DREAMPOTATO_FI NULL
#define OPTION_VAL_NONE_FI "Ei mitään"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_FI NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_FI NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_FI NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_FI NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_FI NULL
#define OPTION_VAL_VMU_A1_FI NULL
#define OPTION_VAL_ALL_VMUS_FI NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_FI "VMU-äänet"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_FI "Kun käytössä, VMU-piippaukset toistetaan."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_FI NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_FI NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_FI NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_FI NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_FI NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_FI NULL
#define OPTION_VAL_UPPER_LEFT_FI "Ylävasen"
#define OPTION_VAL_UPPER_RIGHT_FI "Yläoikea"
#define OPTION_VAL_LOWER_LEFT_FI "Alavasen"
#define OPTION_VAL_LOWER_RIGHT_FI "Alaoikea"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_FI NULL
#define OPTION_VAL_1X_FI NULL
#define OPTION_VAL_3X_FI NULL
#define OPTION_VAL_5X_FI NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_FI NULL
#define OPTION_VAL_DEFAULT_ON_00_FI NULL
#define OPTION_VAL_DEFAULT_OFF_01_FI NULL
#define OPTION_VAL_BLACK_02_FI "Musta"
#define OPTION_VAL_LIGHT_BLUE_04_FI "Vaaleansininen"
#define OPTION_VAL_CYAN_06_FI "Syaani"
#define OPTION_VAL_CYAN_BLUE_07_FI "Turkoosi sininen"
#define OPTION_VAL_LIGHT_GREEN_08_FI "Vaaleanvihreä"
#define OPTION_VAL_CYAN_GREEN_09_FI "Syaani vihreä"
#define OPTION_VAL_LIGHT_CYAN_10_FI NULL
#define OPTION_VAL_PURPLE_12_FI "Violetti"
#define OPTION_VAL_LIGHT_PURPLE_13_FI NULL
#define OPTION_VAL_YELLOW_14_FI "Keltainen"
#define OPTION_VAL_GRAY_15_FI "Harmaa"
#define OPTION_VAL_LIGHT_PURPLE_2_16_FI NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_FI NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_FI NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_FI NULL
#define OPTION_VAL_LIGHT_RED_2_20_FI NULL
#define OPTION_VAL_MAGENTA_21_FI NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_FI NULL
#define OPTION_VAL_LIGHT_ORANGE_23_FI NULL
#define OPTION_VAL_ORANGE_24_FI "Oranssi"
#define OPTION_VAL_LIGHT_PURPLE_4_25_FI NULL
#define OPTION_VAL_LIGHT_YELLOW_26_FI NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_FI NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_FI NULL
#define OPTION_VAL_40_FI "40 %"
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_FI NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_FI NULL

struct retro_core_option_v2_category option_cats_fi[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_FI,
      CATEGORY_SYSTEM_INFO_0_FI
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_FI,
      CATEGORY_VIDEO_INFO_0_FI
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_FI,
      CATEGORY_PERFORMANCE_INFO_0_FI
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_FI,
      CATEGORY_HACKS_INFO_0_FI
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_FI,
      CATEGORY_INPUT_INFO_0_FI
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_FI,
      CATEGORY_EXPANSIONS_INFO_0_FI
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_FI,
      CATEGORY_VMU_INFO_0_FI
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_fi[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_FI,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_FI },
         { "USA",     OPTION_VAL_USA_FI },
         { "Europe",  OPTION_VAL_EUROPE_FI },
         { "Default", OPTION_VAL_DEFAULT_FI },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_FI,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_FI },
         { "English",  OPTION_VAL_ENGLISH_FI },
         { "German",   OPTION_VAL_GERMAN_FI },
         { "French",   OPTION_VAL_FRENCH_FI },
         { "Spanish",  OPTION_VAL_SPANISH_FI },
         { "Italian",  OPTION_VAL_ITALIAN_FI },
         { "Default",  OPTION_VAL_DEFAULT_FI },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_FI,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_FI },
         { "640x480",    OPTION_VAL_640X480_FI },
         { "800x600",    OPTION_VAL_800X600_FI },
         { "960x720",    OPTION_VAL_960X720_FI },
         { "1024x768",   OPTION_VAL_1024X768_FI },
         { "1280x960",   OPTION_VAL_1280X960_FI },
         { "1440x1080",  OPTION_VAL_1440X1080_FI },
         { "1600x1200",  OPTION_VAL_1600X1200_FI },
         { "1920x1440",  OPTION_VAL_1920X1440_FI },
         { "2560x1920",  OPTION_VAL_2560X1920_FI },
         { "2880x2160",  OPTION_VAL_2880X2160_FI },
         { "3200x2400",  OPTION_VAL_3200X2400_FI },
         { "3840x2880",  OPTION_VAL_3840X2880_FI },
         { "4480x3360",  OPTION_VAL_4480X3360_FI },
         { "5120x3840",  OPTION_VAL_5120X3840_FI },
         { "5760x4320",  OPTION_VAL_5760X4320_FI },
         { "6400x4800",  OPTION_VAL_6400X4800_FI },
         { "7040x5280",  OPTION_VAL_7040X5280_FI },
         { "7680x5760",  OPTION_VAL_7680X5760_FI },
         { "8320x6240",  OPTION_VAL_8320X6240_FI },
         { "8960x6720",  OPTION_VAL_8960X6720_FI },
         { "9600x7200",  OPTION_VAL_9600X7200_FI },
         { "10240x7680", OPTION_VAL_10240X7680_FI },
         { "10880x8160", OPTION_VAL_10880X8160_FI },
         { "11520x8640", OPTION_VAL_11520X8640_FI },
         { "12160x9120", OPTION_VAL_12160X9120_FI },
         { "12800x9600", OPTION_VAL_12800X9600_FI },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_FI,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_FI },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_FI },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_FI },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_FI,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_FI },
         { "PAL",     OPTION_VAL_PAL_FI },
         { "PAL_N",   OPTION_VAL_PAL_N_FI },
         { "PAL_M",   OPTION_VAL_PAL_M_FI },
         { "Default", OPTION_VAL_DEFAULT_FI },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_FI,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_FI },
         { "vertical",   OPTION_VAL_VERTICAL_FI },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_FI,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_FI },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_FI },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_FI },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_FI,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_FI },
         { "1GB",   OPTION_VAL_1GB_FI },
         { "2GB",   OPTION_VAL_2GB_FI },
         { "4GB",   OPTION_VAL_4GB_FI },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_FI,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_FI,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_FI,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_FI,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_FI },
         { "1",  OPTION_VAL_1_FI },
         { "2",  OPTION_VAL_2_FI },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_FI,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_FI },
         { "4",  OPTION_VAL_4_FI },
         { "6",  OPTION_VAL_6_FI },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_FI,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_FI,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_FI,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_FI,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_FI,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_FI,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_FI,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_FI },
         { "more", OPTION_VAL_MORE_FI },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_FI,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_FI },
         { "110", OPTION_VAL_110_FI },
         { "120", OPTION_VAL_120_FI },
         { "130", OPTION_VAL_130_FI },
         { "140", OPTION_VAL_140_FI },
         { "150", OPTION_VAL_150_FI },
         { "160", OPTION_VAL_160_FI },
         { "170", OPTION_VAL_170_FI },
         { "180", OPTION_VAL_180_FI },
         { "190", OPTION_VAL_190_FI },
         { "200", OPTION_VAL_200_FI },
         { "210", OPTION_VAL_210_FI },
         { "220", OPTION_VAL_220_FI },
         { "230", OPTION_VAL_230_FI },
         { "240", OPTION_VAL_240_FI },
         { "250", OPTION_VAL_250_FI },
         { "260", OPTION_VAL_260_FI },
         { "270", OPTION_VAL_270_FI },
         { "280", OPTION_VAL_280_FI },
         { "290", OPTION_VAL_290_FI },
         { "300", OPTION_VAL_300_FI },
         { "310", OPTION_VAL_310_FI },
         { "320", OPTION_VAL_320_FI },
         { "330", OPTION_VAL_330_FI },
         { "340", OPTION_VAL_340_FI },
         { "350", OPTION_VAL_350_FI },
         { "360", OPTION_VAL_360_FI },
         { "370", OPTION_VAL_370_FI },
         { "380", OPTION_VAL_380_FI },
         { "390", OPTION_VAL_390_FI },
         { "400", OPTION_VAL_400_FI },
         { "410", OPTION_VAL_410_FI },
         { "420", OPTION_VAL_420_FI },
         { "430", OPTION_VAL_430_FI },
         { "440", OPTION_VAL_440_FI },
         { "450", OPTION_VAL_450_FI },
         { "460", OPTION_VAL_460_FI },
         { "470", OPTION_VAL_470_FI },
         { "480", OPTION_VAL_480_FI },
         { "490", OPTION_VAL_490_FI },
         { "500", OPTION_VAL_500_FI },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_FI,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_FI },
         { "5%",  OPTION_VAL_5_FI },
         { "10%", OPTION_VAL_10_FI },
         { "15%", OPTION_VAL_15_FI },
         { "20%", OPTION_VAL_20_FI },
         { "25%", OPTION_VAL_25_FI },
         { "30%", OPTION_VAL_30_FI },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_FI,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_FI },
         { "5%",  OPTION_VAL_5_FI },
         { "10%", OPTION_VAL_10_FI },
         { "15%", OPTION_VAL_15_FI },
         { "20%", OPTION_VAL_20_FI },
         { "25%", OPTION_VAL_25_FI },
         { "30%", OPTION_VAL_30_FI },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_FI,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_FI,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_FI,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_FI,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_FI },
         { "60%",  OPTION_VAL_60_FI },
         { "70%",  OPTION_VAL_70_FI },
         { "80%",  OPTION_VAL_80_FI },
         { "90%",  OPTION_VAL_90_FI },
         { "100%", OPTION_VAL_100_O47_FI },
         { "110%", OPTION_VAL_110_O47_FI },
         { "120%", OPTION_VAL_120_O47_FI },
         { "130%", OPTION_VAL_130_O47_FI },
         { "140%", OPTION_VAL_140_O47_FI },
         { "150%", OPTION_VAL_150_O47_FI },
         { "160%", OPTION_VAL_160_O47_FI },
         { "170%", OPTION_VAL_170_O47_FI },
         { "180%", OPTION_VAL_180_O47_FI },
         { "190%", OPTION_VAL_190_O47_FI },
         { "200%", OPTION_VAL_200_O47_FI },
         { "210%", OPTION_VAL_210_O47_FI },
         { "220%", OPTION_VAL_220_O47_FI },
         { "230%", OPTION_VAL_230_O47_FI },
         { "240%", OPTION_VAL_240_O47_FI },
         { "250%", OPTION_VAL_250_O47_FI },
         { "260%", OPTION_VAL_260_O47_FI },
         { "270%", OPTION_VAL_270_O47_FI },
         { "280%", OPTION_VAL_280_O47_FI },
         { "290%", OPTION_VAL_290_O47_FI },
         { "300%", OPTION_VAL_300_O47_FI },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_FI,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FI },
         { "Red",      OPTION_VAL_RED_FI },
         { "Green",    OPTION_VAL_GREEN_FI },
         { "Blue",     OPTION_VAL_BLUE_FI },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_FI,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FI },
         { "Red",      OPTION_VAL_RED_FI },
         { "Green",    OPTION_VAL_GREEN_FI },
         { "Blue",     OPTION_VAL_BLUE_FI },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_FI,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FI },
         { "Red",      OPTION_VAL_RED_FI },
         { "Green",    OPTION_VAL_GREEN_FI },
         { "Blue",     OPTION_VAL_BLUE_FI },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_FI,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FI },
         { "Red",      OPTION_VAL_RED_FI },
         { "Green",    OPTION_VAL_GREEN_FI },
         { "Blue",     OPTION_VAL_BLUE_FI },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_FI,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FI },
         { "Purupuru", OPTION_VAL_PURUPURU_FI },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_FI },
         { "None",     OPTION_VAL_NONE_FI },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_FI,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FI },
         { "Purupuru", OPTION_VAL_PURUPURU_FI },
         { "None",     OPTION_VAL_NONE_FI },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_FI,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FI },
         { "Purupuru", OPTION_VAL_PURUPURU_FI },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_FI },
         { "None",     OPTION_VAL_NONE_FI },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_FI,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FI },
         { "Purupuru", OPTION_VAL_PURUPURU_FI },
         { "None",     OPTION_VAL_NONE_FI },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_FI,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FI },
         { "Purupuru", OPTION_VAL_PURUPURU_FI },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_FI },
         { "None",     OPTION_VAL_NONE_FI },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_FI,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FI },
         { "Purupuru", OPTION_VAL_PURUPURU_FI },
         { "None",     OPTION_VAL_NONE_FI },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_FI,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FI },
         { "Purupuru", OPTION_VAL_PURUPURU_FI },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_FI },
         { "None",     OPTION_VAL_NONE_FI },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_FI,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FI },
         { "Purupuru", OPTION_VAL_PURUPURU_FI },
         { "None",     OPTION_VAL_NONE_FI },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_FI,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_FI,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_FI,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_FI },
         { "All VMUs", OPTION_VAL_ALL_VMUS_FI },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_FI,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_FI,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_FI,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_FI,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_FI,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_FI,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_FI,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_FI,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FI },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FI },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FI },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FI },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FI },
         { "2x", OPTION_VAL_2_O26_FI },
         { "3x", OPTION_VAL_3X_FI },
         { "4x", OPTION_VAL_4_FI },
         { "5x", OPTION_VAL_5X_FI },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FI },
         { "20%",  OPTION_VAL_20_FI },
         { "30%",  OPTION_VAL_30_FI },
         { "40%",  OPTION_VAL_40_FI },
         { "50%",  OPTION_VAL_50_FI },
         { "60%",  OPTION_VAL_60_FI },
         { "70%",  OPTION_VAL_70_FI },
         { "80%",  OPTION_VAL_80_FI },
         { "90%",  OPTION_VAL_90_FI },
         { "100%", OPTION_VAL_100_O47_FI },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FI },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FI },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FI },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FI },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FI },
         { "2x", OPTION_VAL_2_O26_FI },
         { "3x", OPTION_VAL_3X_FI },
         { "4x", OPTION_VAL_4_FI },
         { "5x", OPTION_VAL_5X_FI },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FI },
         { "20%",  OPTION_VAL_20_FI },
         { "30%",  OPTION_VAL_30_FI },
         { "40%",  OPTION_VAL_40_FI },
         { "50%",  OPTION_VAL_50_FI },
         { "60%",  OPTION_VAL_60_FI },
         { "70%",  OPTION_VAL_70_FI },
         { "80%",  OPTION_VAL_80_FI },
         { "90%",  OPTION_VAL_90_FI },
         { "100%", OPTION_VAL_100_O47_FI },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FI },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FI },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FI },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FI },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FI },
         { "2x", OPTION_VAL_2_O26_FI },
         { "3x", OPTION_VAL_3X_FI },
         { "4x", OPTION_VAL_4_FI },
         { "5x", OPTION_VAL_5X_FI },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FI },
         { "20%",  OPTION_VAL_20_FI },
         { "30%",  OPTION_VAL_30_FI },
         { "40%",  OPTION_VAL_40_FI },
         { "50%",  OPTION_VAL_50_FI },
         { "60%",  OPTION_VAL_60_FI },
         { "70%",  OPTION_VAL_70_FI },
         { "80%",  OPTION_VAL_80_FI },
         { "90%",  OPTION_VAL_90_FI },
         { "100%", OPTION_VAL_100_O47_FI },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FI },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FI },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FI },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FI },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FI },
         { "2x", OPTION_VAL_2_O26_FI },
         { "3x", OPTION_VAL_3X_FI },
         { "4x", OPTION_VAL_4_FI },
         { "5x", OPTION_VAL_5X_FI },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FI },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FI },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FI },
         { "BLUE 03",           OPTION_VAL_BLUE_FI },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FI },
         { "GREEN 05",          OPTION_VAL_GREEN_FI },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FI },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FI },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FI },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FI },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FI },
         { "RED 11",            OPTION_VAL_RED_FI },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FI },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FI },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FI },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FI },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FI },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FI },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FI },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FI },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FI },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FI },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FI },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FI },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FI },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FI },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FI },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FI },
         { "WHITE 28",          OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_FI,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FI },
         { "20%",  OPTION_VAL_20_FI },
         { "30%",  OPTION_VAL_30_FI },
         { "40%",  OPTION_VAL_40_FI },
         { "50%",  OPTION_VAL_50_FI },
         { "60%",  OPTION_VAL_60_FI },
         { "70%",  OPTION_VAL_70_FI },
         { "80%",  OPTION_VAL_80_FI },
         { "90%",  OPTION_VAL_90_FI },
         { "100%", OPTION_VAL_100_O47_FI },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_fi = {
   option_cats_fi,
   option_defs_fi
};

/* RETRO_LANGUAGE_FR */

#define CATEGORY_SYSTEM_LABEL_FR "Système"
#define CATEGORY_SYSTEM_INFO_0_FR "Configurer les réglages de région, de langue, du BIOS et du matériel de base."
#define CATEGORY_VIDEO_LABEL_FR "Vidéo"
#define CATEGORY_VIDEO_INFO_0_FR "Configurez les réglages de résolution, de transparence indépendante de l'ordre et des effets visuels."
#define CATEGORY_PERFORMANCE_LABEL_FR "Performances"
#define CATEGORY_PERFORMANCE_INFO_0_FR "Configurer les réglages de rendu et de saut d'images."
#define CATEGORY_HACKS_LABEL_FR "Hacks d'émulation"
#define CATEGORY_HACKS_INFO_0_FR "Configurez les réglages de remplacements pour écrans larges, de vitesse de chargement des GD-ROM et de remplacement de textures."
#define CATEGORY_INPUT_LABEL_FR "Entrées"
#define CATEGORY_INPUT_INFO_0_FR "Configurer les réglages des manettes et pistolets."
#define CATEGORY_EXPANSIONS_LABEL_FR "Ports d'extension de la manette"
#define CATEGORY_EXPANSIONS_INFO_0_FR "Sélectionnez le périphérique (VMU, dispositif de vibration) branché dans chaque emplacement d'extension de la manette."
#define CATEGORY_VMU_LABEL_FR "Unité de mémoire visuelle (VMU)"
#define CATEGORY_VMU_INFO_0_FR "Configurer les réglages de fichiers de sauvegarde VMU par jeu et de visibilité des VMU à l'écran."
#define CORE_OPTION_NAME_REGION_LABEL_FR "Région"
#define OPTION_VAL_JAPAN_FR "Japon"
#define OPTION_VAL_USA_FR "États-Unis"
#define OPTION_VAL_EUROPE_FR NULL
#define OPTION_VAL_DEFAULT_FR "Par défaut"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_FR "Langue "
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_FR "Modifie la langue utilisée par le BIOS et par tous les jeux qui contiennent plusieurs langues."
#define OPTION_VAL_JAPANESE_FR "Japonais"
#define OPTION_VAL_ENGLISH_FR "Anglais"
#define OPTION_VAL_GERMAN_FR "Allemand"
#define OPTION_VAL_FRENCH_FR "Français"
#define OPTION_VAL_SPANISH_FR "Espagnol"
#define OPTION_VAL_ITALIAN_FR "Italien"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_FR "BIOS HLE (Redémarrage requis)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_FR "Forcer l'utilisation d'un BIOS en émulation de haut niveau."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_FR "Activer le DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_FR "Active l'émulation du DSP audio de la Dreamcast (processeur de signal numérique). Améliore la précision du son généré, mais augmente les exigences de performances."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_FR "Autoriser les boutons de service d'arcade"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_FR "Active le bouton SERVICE pour les jeux d'arcade, pour entrer dans les paramètres de la borne."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_FR "Définir les jeux NAOMI en Free Play"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_FR "Modifier les réglages de pièces du jeu en mode gratuit."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_FR "Émulation de l'adaptateur haut débit"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_FR "Émuler l'adaptateur haut débit Ethernet au lieu du modem. (Redémarrage requis)"
#define CORE_OPTION_NAME_UPNP_LABEL_FR "Activer l'UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_FR "Utilisez l'UPnP pour configurer automatiquement votre routeur Internet pour les jeux en ligne."
#define CORE_OPTION_NAME_DCNET_LABEL_FR "Utiliser DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_FR "Utilisez le service cloud DCNet pour l'accès à internet de la Dreamcast."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_FR "Résolution interne"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_FR "Modifier la résolution de rendu."
#define OPTION_VAL_320X240_FR "320x240 (Moitié)"
#define OPTION_VAL_640X480_FR "640x480 (Natif)"
#define OPTION_VAL_800X600_FR "800x600 (1,25x)"
#define OPTION_VAL_960X720_FR "960x720 (1,5x)"
#define OPTION_VAL_1024X768_FR "1024x768 (1,6x)"
#define OPTION_VAL_1280X960_FR "1280x960 (2x)"
#define OPTION_VAL_1440X1080_FR "1440x1080 (2,25x)"
#define OPTION_VAL_1600X1200_FR "1600x1200 (2,5x)"
#define OPTION_VAL_1920X1440_FR "1920x1440 (3x)"
#define OPTION_VAL_2560X1920_FR "2560x1920 (4x)"
#define OPTION_VAL_2880X2160_FR "2880x2160 (4,5x)"
#define OPTION_VAL_3200X2400_FR "3200x2400 (5x)"
#define OPTION_VAL_3840X2880_FR "3840x2880 (6x)"
#define OPTION_VAL_4480X3360_FR "4480x3360 (7x)"
#define OPTION_VAL_5120X3840_FR "5120x3840 (8x)"
#define OPTION_VAL_5760X4320_FR "5760x4320 (9x)"
#define OPTION_VAL_6400X4800_FR "6400x4800 (10x)"
#define OPTION_VAL_7040X5280_FR "7040x5280 (11x)"
#define OPTION_VAL_7680X5760_FR "7680x5760 (12x)"
#define OPTION_VAL_8320X6240_FR "8320x6240 (13x)"
#define OPTION_VAL_8960X6720_FR "8960x6720 (14x)"
#define OPTION_VAL_9600X7200_FR "9600x7200 (15x)"
#define OPTION_VAL_10240X7680_FR "10240x7680 (16x)"
#define OPTION_VAL_10880X8160_FR "10880x8160 (17x)"
#define OPTION_VAL_11520X8640_FR "11520x8640 (18x)"
#define OPTION_VAL_12160X9120_FR "12160x9120 (19x)"
#define OPTION_VAL_12800X9600_FR "12800x9600 (20x)"
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_FR "Type de câble"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_FR "Le type de signal de sortie. 'TV (Composite)' est le plus largement pris en charge."
#define OPTION_VAL_VGA_FR NULL
#define OPTION_VAL_TV_RGB_FR "TV (RVB)"
#define OPTION_VAL_TV_COMPOSITE_FR NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_FR "Standard de diffusion"
#define OPTION_VAL_NTSC_FR NULL
#define OPTION_VAL_PAL_FR "PAL (Monde)"
#define OPTION_VAL_PAL_N_FR "PAL-N (Argentine, Paraguay, Uruguay)"
#define OPTION_VAL_PAL_M_FR "PAL-M (Brésil)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_FR "Orientation de l'écran"
#define OPTION_VAL_HORIZONTAL_FR "Horizontale"
#define OPTION_VAL_VERTICAL_FR "Verticale"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_FR "Tri alpha"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_FR "Sélectionner comment les polygones transparents sont triés."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_FR "Par bande (rapide, moins précis)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_FR "Par triangle (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_FR "Par pixel (précis, mais le plus lent)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_FR "Taille du tampon d'accumulation des pixels"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_FR "Des valeurs plus élevées peuvent être requises pour que les résolutions plus élevées s'affichent correctement."
#define OPTION_VAL_512MB_FR "512 Mo"
#define OPTION_VAL_1GB_FR "1 Go"
#define OPTION_VAL_2GB_FR "2 Go"
#define OPTION_VAL_4GB_FR "4 Go"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_FR "Calques transparents maximum"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_FR "Des valeurs plus élevées peuvent être requises pour les scènes complexes."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_FR "Émulation complète du framebuffer"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_FR "Active l'émulation complète du framebuffer dans la VRAM. Ceci est utile pour les jeux qui lisent ou écrivent directement le framebuffer dans la VRAM. Lorsque cette option est activée, la résolution interne est forcée à 640x480 et les performances peuvent être sévèrement affectées."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_FR "Activer la mémoire tampon du RTT (rendu vers texture)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_FR "Copier les textures rendues depuis le processeur graphique vers la VRAM. Cette option est normalement activée pour les jeux qui en ont besoin. Lorsque cette option est activée, la mise à l'échelle du rendu des textures est désactivée et les performances peuvent être impactées."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_FR NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_FR "Lorsque cette option est activée, les textures utilisent une version plus petite d'elles-mêmes lorsqu'elles apparaissent plus loin, cela peut augmenter les performances et réduire le scintillement."
#define CORE_OPTION_NAME_FOG_LABEL_FR "Effets de brouillard"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_FR "Modificateur du volume"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_FR "Une fonctionnalité du processeur graphique de la Dreamcast qui est généralement utilisée par les jeux pour le rendu des ombres d'objets. Cette option devrait normalement être activée — l'impact sur les performances va généralement de minimal à négligeable."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_FR "Filtrage anisotrope"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_FR "Améliore la qualité des textures sur les surfaces qui sont à des angles de vue obliques par rapport à la caméra. Des valeurs plus élevées sont plus exigeantes sur le processeur graphique. Les modifications de ce réglage ne sont appliqués qu'après redémarrage."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_FR "Filtrage des textures"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_FR "Le mode de filtrage des textures à utiliser. Cela peut être utilisé pour forcer un certain mode de filtrage de texture sur toutes les textures pour obtenir une apparence plus nette (ou plus lisse) que par défaut. Les valeurs autres que celles par défaut peuvent causer divers problèmes de rendu. Les modifications de ce réglage ne sont appliquées qu'après redémarrage."
#define OPTION_VAL_1_FR "Forcer au plus proche"
#define OPTION_VAL_2_FR "Forcer linéaire"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_FR "Retarder l'échange d'images"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_FR "Utile pour éviter les écrans clignotants ou les vidéos buggées. Non recommandé sur les plates-formes lentes."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_FR "Détecter les changements de fréquence d'images"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_FR "Notifier l'interface frontale lorsque la fréquence d’images interne change (par exemple, de 60 i/s à 30 i/s). Améliore la cadence des images dans les jeux qui fonctionnent à 30 i/s ou 20 i/s, mais devrait être désactivé pour les jeux avec des fréquences d'images (non stables) déverrouillées (Ecco the Dolphin, Unreal Tournament par exemple). Remarque : Indisponible lorsque 'Saut d'images automatique' est activé."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_FR "Filtre de post-traitement PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_FR "Post-traitement de l'image rendue pour simuler les effets spécifiques au processeur graphique PowerVR2 et aux signaux vidéo analogiques."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_FR "Agrandissement des textures (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_FR "Améliorer les graphismes en pixels 2D dessinés à la main. Devrait être utilisé uniquement avec les jeux en 2D pixelisés."
#define OPTION_VAL_2_O26_FR "x2"
#define OPTION_VAL_4_FR "x4"
#define OPTION_VAL_6_FR "x6"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_FR "Facteur maximal du filtre d'agrandissement de textures"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_FR "Sélectionner une valeur de taille maximale pour les textures à agrandir, si la taille de la texture est supérieure à la valeur sélectionnée, elle ne sera pas redimensionnée."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_FR "Interpolation de profondeur native"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_FR "Aide pour la corruption de textures et les problèmes de profondeur sur les processeurs graphiques AMD. Peut aussi aider les processeurs graphiques Intel dans certains cas."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_FR "Corriger les bords débordant à l'agrandissement"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_FR "Aide avec le cas de débordement de textures lors de l'agrandissement. Désactiver cette option peut aider si les pixels se déforment lors de l'agrandissement dans les jeux 2D (MVC2, CVS, KOF, etc.)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_FR "Rendu sur plusieurs fils d'exécution"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_FR "Fais tourner le processeur graphique et le processeur sur différents fils d'exécution. Fortement recommandé."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_FR "Saut d'images automatique"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_FR "Sauter automatiquement des images lorsque l'émulateur s'exécute lentement. Remarque : Cette option ne s'applique que lorsque 'Rendu sur plusieurs fils d'exécution' est activé."
#define OPTION_VAL_SOME_FR NULL
#define OPTION_VAL_MORE_FR "Maximal"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_FR "Saut d'images"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_FR "Définit le nombre d'images à sauter entre chaque image affichée."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_FR "Cheats d'écran large (Redémarrage requis)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_FR "Active les cheats qui permettent à certains jeux de s'afficher au format écran large."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_FR "Hack d'écran large"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_FR "Afficher un rendu de la géométrie en dehors du rapport d'aspect normal de 4:3. Peut produire des bugs graphiques dans les zones révélées."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_FR "Chargement rapide des GD-ROM (inexact)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_FR "Accélère le chargement des GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_FR "Mod RAM Dreamcast 32 Mo"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_FR "Active le mod RAM 32 Mo pour Dreamcast. Peut affecter la compatibilité"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_FR "Under/overclock processeur du SH4"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_FR "Modifie la fréquence d'horloge du processeur principal du SH4 cadencé par défaut à 200 MHz. La diminuer peut aider les plates-formes lentes. L'augmenter peut améliorer la fréquence d'images pour certains jeux. À utiliser avec prudence."
#define OPTION_VAL_100_FR NULL
#define OPTION_VAL_110_FR NULL
#define OPTION_VAL_120_FR NULL
#define OPTION_VAL_130_FR NULL
#define OPTION_VAL_140_FR NULL
#define OPTION_VAL_150_FR NULL
#define OPTION_VAL_160_FR NULL
#define OPTION_VAL_170_FR NULL
#define OPTION_VAL_180_FR NULL
#define OPTION_VAL_190_FR NULL
#define OPTION_VAL_200_FR NULL
#define OPTION_VAL_210_FR NULL
#define OPTION_VAL_220_FR NULL
#define OPTION_VAL_230_FR NULL
#define OPTION_VAL_240_FR NULL
#define OPTION_VAL_250_FR NULL
#define OPTION_VAL_260_FR NULL
#define OPTION_VAL_270_FR NULL
#define OPTION_VAL_280_FR NULL
#define OPTION_VAL_290_FR NULL
#define OPTION_VAL_300_FR NULL
#define OPTION_VAL_310_FR NULL
#define OPTION_VAL_320_FR NULL
#define OPTION_VAL_330_FR NULL
#define OPTION_VAL_340_FR NULL
#define OPTION_VAL_350_FR NULL
#define OPTION_VAL_360_FR NULL
#define OPTION_VAL_370_FR NULL
#define OPTION_VAL_380_FR NULL
#define OPTION_VAL_390_FR NULL
#define OPTION_VAL_400_FR NULL
#define OPTION_VAL_410_FR NULL
#define OPTION_VAL_420_FR NULL
#define OPTION_VAL_430_FR NULL
#define OPTION_VAL_440_FR NULL
#define OPTION_VAL_450_FR NULL
#define OPTION_VAL_460_FR NULL
#define OPTION_VAL_470_FR NULL
#define OPTION_VAL_480_FR NULL
#define OPTION_VAL_490_FR NULL
#define OPTION_VAL_500_FR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_FR "Charger les textures personnalisées"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_FR "Charger les textures personnalisées situées dans le dossier 'system/dc/textures/<game-id>/'."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_FR "Précharger les textures personnalisées"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_FR "Précharge les textures personnalisées au lancement du jeu. Peut améliorer les performances mais augmente l'utilisation de la mémoire."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_FR "Importer les textures"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_FR "Chaque fois qu'une nouvelle texture est utilisée par le jeu, elle sera sauvegardée en tant que fichier .png dans le dossier 'system/dc/texdump/<game-id>/'."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_FR "Importer les textures remplacées"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_FR "Toujours importer les textures qui sont déjà remplacées par des textures personnalisées."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_FR "Deadzone du stick analogique"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_FR "Sélectionner la distance à laquelle vous devez pousser le stick analogique avant que ne commence son traitement."
#define OPTION_VAL_0_FR NULL
#define OPTION_VAL_5_FR NULL
#define OPTION_VAL_10_FR NULL
#define OPTION_VAL_15_FR NULL
#define OPTION_VAL_20_FR NULL
#define OPTION_VAL_25_FR NULL
#define OPTION_VAL_30_FR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_FR "Deadzone des gâchettes"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_FR "Sélectionner combien vous devez appuyer sur la gâchette avant que ne commence son traitement."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_FR "Gâchettes numériques"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_FR "Lorsque cette option est activée, les gâchettes agiront comme des touches normales, ce qui signifie qu'elles seront traités comme étant entièrement pressées ou pas du tout, sans aucun intermédiaire."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_FR "Diffuser les sorties numériques"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_FR "Diffuser les sorties numériques et l'état du retour de force sur le port TCP 8000. Compatible avec l'option de MAME \"-output network\" ."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_FR "Afficher les réglages du pistolet"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_FR "Activer la configuration des options d'affichage du viseur des pistolets. REMARQUE : le menu rapide devrait être réouvert pour que ce paramètre prenne effet."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_FR "Mise à l'échelle du viseur de pistolet"
#define OPTION_VAL_50_FR NULL
#define OPTION_VAL_60_FR NULL
#define OPTION_VAL_70_FR NULL
#define OPTION_VAL_80_FR NULL
#define OPTION_VAL_90_FR NULL
#define OPTION_VAL_100_O47_FR NULL
#define OPTION_VAL_110_O47_FR NULL
#define OPTION_VAL_120_O47_FR NULL
#define OPTION_VAL_130_O47_FR NULL
#define OPTION_VAL_140_O47_FR NULL
#define OPTION_VAL_150_O47_FR NULL
#define OPTION_VAL_160_O47_FR NULL
#define OPTION_VAL_170_O47_FR NULL
#define OPTION_VAL_180_O47_FR NULL
#define OPTION_VAL_190_O47_FR NULL
#define OPTION_VAL_200_O47_FR NULL
#define OPTION_VAL_210_O47_FR NULL
#define OPTION_VAL_220_O47_FR NULL
#define OPTION_VAL_230_O47_FR NULL
#define OPTION_VAL_240_O47_FR NULL
#define OPTION_VAL_250_O47_FR NULL
#define OPTION_VAL_260_O47_FR NULL
#define OPTION_VAL_270_O47_FR NULL
#define OPTION_VAL_280_O47_FR NULL
#define OPTION_VAL_290_O47_FR NULL
#define OPTION_VAL_300_O47_FR NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_FR "Afficher le viseur du pistolet 1"
#define OPTION_VAL_WHITE_FR "Blanc"
#define OPTION_VAL_RED_FR "Rouge"
#define OPTION_VAL_GREEN_FR "Vert"
#define OPTION_VAL_BLUE_FR "Bleu"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_FR "Afficher le viseur du pistolet 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_FR "Afficher le viseur du pistolet 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_FR "Afficher le viseur du pistolet 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_FR "Périphérique dans le port d'extension A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_FR "Sélectionner le périphérique branché dans le port d'extension A1 (port A emplacement 1)."
#define OPTION_VAL_VMU_FR NULL
#define OPTION_VAL_PURUPURU_FR NULL
#define OPTION_VAL_DREAMPOTATO_FR NULL
#define OPTION_VAL_NONE_FR "Ne pas partager"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_FR "Périphérique dans le port d'extension A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_FR "Sélectionner le périphérique branché dans le port d'extension A2 (port A emplacement 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_FR "Périphérique dans le port d'extension B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_FR "Sélectionner le périphérique branché dans le port d'extension B1 (port B emplacement 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_FR "Périphérique dans le port d'extension B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_FR "Sélectionner le périphérique branché dans le port d'extension B2 (port B emplacement 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_FR "Périphérique dans le port d'extension C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_FR "Sélectionner le périphérique branché dans le port d'extension C1 (port C emplacement 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_FR "Périphérique dans le port d'extension C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_FR "Sélectionner le périphérique branché dans le port d'extension C2 (port C emplacement 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_FR "Périphérique dans le port d'extension D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_FR "Sélectionner le périphérique branché dans le port d'extension D1 (port D emplacement 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_FR "Périphérique dans le port d'extension D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_FR "Sélectionner le périphérique branché dans le port d'extension D2 (port D emplacement 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_FR "Unités/systèmes de mémoire visuelle par jeu (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_FR "VMU par jeu"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_FR "Lorsque cette option est désactivée, tous les jeux partagent jusqu'à 8 fichiers de sauvegarde VMU (A1/A2/B1/B2/C1/D1/D1/D2) situés dans le dossier system de RetroArch.\nLe réglage 'VMU A1' crée un fichier VMU 'A1' unique dans le dossier de sauvegarde de RetroArch pour chaque jeu qui est lancé.\nLe réglage 'Toutes les VMU' crée jusqu'à 8 fichiers VMU uniques (A1/A2/B1/B2/C1/C2/D1/D2) pour chaque jeu qui est lancé."
#define OPTION_VAL_VMU_A1_FR NULL
#define OPTION_VAL_ALL_VMUS_FR "Tous les VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_FR "Sons des unités/systèmes de mémoire visuelle (VMU)"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_FR "Sons des VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_FR "Si cette option est activée, les bips des VMU sont joués."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_FR "Stockage VMU DreamPotato"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_FR "Activer le stockage des données VMU sur l'émulateur DreamPotato connecté. Lorsque cette option est activée, le fichier de sauvegarde VMU local n'est pas utilisé."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_FR "Afficher les réglage d'affichage de l'unité/système de mémoire visuelle (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_FR "Afficher les réglages d'affichage du VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_FR "Activer la configuration de la visibilité, de la taille, de la position et de la couleur de l'écran du VMU émulé. REMARQUE : Le menu rapide devrait être réouvert pour que ce réglage prenne effet."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_FR "Affichage de l'écran du VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_FR "Position de l'écran du VMU 1"
#define OPTION_VAL_UPPER_LEFT_FR "En haut à gauche"
#define OPTION_VAL_UPPER_RIGHT_FR "En haut à droite"
#define OPTION_VAL_LOWER_LEFT_FR "En bas à gauche"
#define OPTION_VAL_LOWER_RIGHT_FR "En bas à droite"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_FR "Taille de l'écran du VMU 1"
#define OPTION_VAL_1X_FR "x1"
#define OPTION_VAL_3X_FR "x3"
#define OPTION_VAL_5X_FR "x5"
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_FR "Couleur des pixels allumés de l'écran du VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_FR "Par défaut activé"
#define OPTION_VAL_DEFAULT_OFF_01_FR "Par défaut désactivé"
#define OPTION_VAL_BLACK_02_FR "Noire"
#define OPTION_VAL_LIGHT_BLUE_04_FR "Bleu clair"
#define OPTION_VAL_CYAN_06_FR NULL
#define OPTION_VAL_CYAN_BLUE_07_FR "Bleu cyan"
#define OPTION_VAL_LIGHT_GREEN_08_FR "Vert clair"
#define OPTION_VAL_CYAN_GREEN_09_FR "Vert cyan"
#define OPTION_VAL_LIGHT_CYAN_10_FR "Cyan clair"
#define OPTION_VAL_PURPLE_12_FR "Violette"
#define OPTION_VAL_LIGHT_PURPLE_13_FR "Violet clair"
#define OPTION_VAL_YELLOW_14_FR "Jaune"
#define OPTION_VAL_GRAY_15_FR "Grise"
#define OPTION_VAL_LIGHT_PURPLE_2_16_FR "Violet clair (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_FR "Vert clair (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_FR "Vert clair (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_FR "Cyan clair (2)"
#define OPTION_VAL_LIGHT_RED_2_20_FR "Rouge clair (2)"
#define OPTION_VAL_MAGENTA_21_FR NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_FR "Violet clair (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_FR "Orange clair"
#define OPTION_VAL_ORANGE_24_FR NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_FR "Violet clair (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_FR "Jaune clair"
#define OPTION_VAL_LIGHT_YELLOW_2_27_FR "Jaune clair (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_FR "Couleur des pixels éteints de l'écran du VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_FR "Opacité de l'écran du VMU 1"
#define OPTION_VAL_40_FR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_FR "Affichage de l'écran du VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_FR "Position de l'écran du VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_FR "Taille de l'écran du VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_FR "Couleur des pixels allumés de l'écran du VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_FR "Couleur des pixels éteints de l'écran du VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_FR "Opacité de l'écran du VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_FR "Affichage de l'écran du VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_FR "Position de l'écran du VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_FR "Taille de l'écran du VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_FR "Couleur des pixels allumés de l'écran du VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_FR "Couleur des pixels éteints de l'écran du VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_FR "Opacité de l'écran du VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_FR "Affichage de l'écran du VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_FR "Position de l'écran du VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_FR "Taille de l'écran du VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_FR "Couleur des pixels allumés de l'écran du VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_FR "Couleur des pixels éteints de l'écran du VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_FR "Opacité de l'écran du VMU 4"

struct retro_core_option_v2_category option_cats_fr[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_FR,
      CATEGORY_SYSTEM_INFO_0_FR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_FR,
      CATEGORY_VIDEO_INFO_0_FR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_FR,
      CATEGORY_PERFORMANCE_INFO_0_FR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_FR,
      CATEGORY_HACKS_INFO_0_FR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_FR,
      CATEGORY_INPUT_INFO_0_FR
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_FR,
      CATEGORY_EXPANSIONS_INFO_0_FR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_FR,
      CATEGORY_VMU_INFO_0_FR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_fr[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_FR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_FR },
         { "USA",     OPTION_VAL_USA_FR },
         { "Europe",  OPTION_VAL_EUROPE_FR },
         { "Default", OPTION_VAL_DEFAULT_FR },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_FR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_FR },
         { "English",  OPTION_VAL_ENGLISH_FR },
         { "German",   OPTION_VAL_GERMAN_FR },
         { "French",   OPTION_VAL_FRENCH_FR },
         { "Spanish",  OPTION_VAL_SPANISH_FR },
         { "Italian",  OPTION_VAL_ITALIAN_FR },
         { "Default",  OPTION_VAL_DEFAULT_FR },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_FR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_FR },
         { "640x480",    OPTION_VAL_640X480_FR },
         { "800x600",    OPTION_VAL_800X600_FR },
         { "960x720",    OPTION_VAL_960X720_FR },
         { "1024x768",   OPTION_VAL_1024X768_FR },
         { "1280x960",   OPTION_VAL_1280X960_FR },
         { "1440x1080",  OPTION_VAL_1440X1080_FR },
         { "1600x1200",  OPTION_VAL_1600X1200_FR },
         { "1920x1440",  OPTION_VAL_1920X1440_FR },
         { "2560x1920",  OPTION_VAL_2560X1920_FR },
         { "2880x2160",  OPTION_VAL_2880X2160_FR },
         { "3200x2400",  OPTION_VAL_3200X2400_FR },
         { "3840x2880",  OPTION_VAL_3840X2880_FR },
         { "4480x3360",  OPTION_VAL_4480X3360_FR },
         { "5120x3840",  OPTION_VAL_5120X3840_FR },
         { "5760x4320",  OPTION_VAL_5760X4320_FR },
         { "6400x4800",  OPTION_VAL_6400X4800_FR },
         { "7040x5280",  OPTION_VAL_7040X5280_FR },
         { "7680x5760",  OPTION_VAL_7680X5760_FR },
         { "8320x6240",  OPTION_VAL_8320X6240_FR },
         { "8960x6720",  OPTION_VAL_8960X6720_FR },
         { "9600x7200",  OPTION_VAL_9600X7200_FR },
         { "10240x7680", OPTION_VAL_10240X7680_FR },
         { "10880x8160", OPTION_VAL_10880X8160_FR },
         { "11520x8640", OPTION_VAL_11520X8640_FR },
         { "12160x9120", OPTION_VAL_12160X9120_FR },
         { "12800x9600", OPTION_VAL_12800X9600_FR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_FR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_FR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_FR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_FR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_FR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_FR },
         { "PAL",     OPTION_VAL_PAL_FR },
         { "PAL_N",   OPTION_VAL_PAL_N_FR },
         { "PAL_M",   OPTION_VAL_PAL_M_FR },
         { "Default", OPTION_VAL_DEFAULT_FR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_FR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_FR },
         { "vertical",   OPTION_VAL_VERTICAL_FR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_FR,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_FR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_FR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_FR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_FR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_FR },
         { "1GB",   OPTION_VAL_1GB_FR },
         { "2GB",   OPTION_VAL_2GB_FR },
         { "4GB",   OPTION_VAL_4GB_FR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_FR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_FR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_FR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_FR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_FR },
         { "1",  OPTION_VAL_1_FR },
         { "2",  OPTION_VAL_2_FR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_FR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_FR },
         { "4",  OPTION_VAL_4_FR },
         { "6",  OPTION_VAL_6_FR },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_FR,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_FR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_FR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_FR,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_FR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_FR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_FR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_FR },
         { "more", OPTION_VAL_MORE_FR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_FR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_FR },
         { "110", OPTION_VAL_110_FR },
         { "120", OPTION_VAL_120_FR },
         { "130", OPTION_VAL_130_FR },
         { "140", OPTION_VAL_140_FR },
         { "150", OPTION_VAL_150_FR },
         { "160", OPTION_VAL_160_FR },
         { "170", OPTION_VAL_170_FR },
         { "180", OPTION_VAL_180_FR },
         { "190", OPTION_VAL_190_FR },
         { "200", OPTION_VAL_200_FR },
         { "210", OPTION_VAL_210_FR },
         { "220", OPTION_VAL_220_FR },
         { "230", OPTION_VAL_230_FR },
         { "240", OPTION_VAL_240_FR },
         { "250", OPTION_VAL_250_FR },
         { "260", OPTION_VAL_260_FR },
         { "270", OPTION_VAL_270_FR },
         { "280", OPTION_VAL_280_FR },
         { "290", OPTION_VAL_290_FR },
         { "300", OPTION_VAL_300_FR },
         { "310", OPTION_VAL_310_FR },
         { "320", OPTION_VAL_320_FR },
         { "330", OPTION_VAL_330_FR },
         { "340", OPTION_VAL_340_FR },
         { "350", OPTION_VAL_350_FR },
         { "360", OPTION_VAL_360_FR },
         { "370", OPTION_VAL_370_FR },
         { "380", OPTION_VAL_380_FR },
         { "390", OPTION_VAL_390_FR },
         { "400", OPTION_VAL_400_FR },
         { "410", OPTION_VAL_410_FR },
         { "420", OPTION_VAL_420_FR },
         { "430", OPTION_VAL_430_FR },
         { "440", OPTION_VAL_440_FR },
         { "450", OPTION_VAL_450_FR },
         { "460", OPTION_VAL_460_FR },
         { "470", OPTION_VAL_470_FR },
         { "480", OPTION_VAL_480_FR },
         { "490", OPTION_VAL_490_FR },
         { "500", OPTION_VAL_500_FR },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_FR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_FR },
         { "5%",  OPTION_VAL_5_FR },
         { "10%", OPTION_VAL_10_FR },
         { "15%", OPTION_VAL_15_FR },
         { "20%", OPTION_VAL_20_FR },
         { "25%", OPTION_VAL_25_FR },
         { "30%", OPTION_VAL_30_FR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_FR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_FR },
         { "5%",  OPTION_VAL_5_FR },
         { "10%", OPTION_VAL_10_FR },
         { "15%", OPTION_VAL_15_FR },
         { "20%", OPTION_VAL_20_FR },
         { "25%", OPTION_VAL_25_FR },
         { "30%", OPTION_VAL_30_FR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_FR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_FR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_FR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_FR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_FR },
         { "60%",  OPTION_VAL_60_FR },
         { "70%",  OPTION_VAL_70_FR },
         { "80%",  OPTION_VAL_80_FR },
         { "90%",  OPTION_VAL_90_FR },
         { "100%", OPTION_VAL_100_O47_FR },
         { "110%", OPTION_VAL_110_O47_FR },
         { "120%", OPTION_VAL_120_O47_FR },
         { "130%", OPTION_VAL_130_O47_FR },
         { "140%", OPTION_VAL_140_O47_FR },
         { "150%", OPTION_VAL_150_O47_FR },
         { "160%", OPTION_VAL_160_O47_FR },
         { "170%", OPTION_VAL_170_O47_FR },
         { "180%", OPTION_VAL_180_O47_FR },
         { "190%", OPTION_VAL_190_O47_FR },
         { "200%", OPTION_VAL_200_O47_FR },
         { "210%", OPTION_VAL_210_O47_FR },
         { "220%", OPTION_VAL_220_O47_FR },
         { "230%", OPTION_VAL_230_O47_FR },
         { "240%", OPTION_VAL_240_O47_FR },
         { "250%", OPTION_VAL_250_O47_FR },
         { "260%", OPTION_VAL_260_O47_FR },
         { "270%", OPTION_VAL_270_O47_FR },
         { "280%", OPTION_VAL_280_O47_FR },
         { "290%", OPTION_VAL_290_O47_FR },
         { "300%", OPTION_VAL_300_O47_FR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_FR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FR },
         { "Red",      OPTION_VAL_RED_FR },
         { "Green",    OPTION_VAL_GREEN_FR },
         { "Blue",     OPTION_VAL_BLUE_FR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_FR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FR },
         { "Red",      OPTION_VAL_RED_FR },
         { "Green",    OPTION_VAL_GREEN_FR },
         { "Blue",     OPTION_VAL_BLUE_FR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_FR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FR },
         { "Red",      OPTION_VAL_RED_FR },
         { "Green",    OPTION_VAL_GREEN_FR },
         { "Blue",     OPTION_VAL_BLUE_FR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_FR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_FR },
         { "Red",      OPTION_VAL_RED_FR },
         { "Green",    OPTION_VAL_GREEN_FR },
         { "Blue",     OPTION_VAL_BLUE_FR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_FR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FR },
         { "Purupuru", OPTION_VAL_PURUPURU_FR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_FR },
         { "None",     OPTION_VAL_NONE_FR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_FR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FR },
         { "Purupuru", OPTION_VAL_PURUPURU_FR },
         { "None",     OPTION_VAL_NONE_FR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_FR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FR },
         { "Purupuru", OPTION_VAL_PURUPURU_FR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_FR },
         { "None",     OPTION_VAL_NONE_FR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_FR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FR },
         { "Purupuru", OPTION_VAL_PURUPURU_FR },
         { "None",     OPTION_VAL_NONE_FR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_FR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FR },
         { "Purupuru", OPTION_VAL_PURUPURU_FR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_FR },
         { "None",     OPTION_VAL_NONE_FR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_FR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FR },
         { "Purupuru", OPTION_VAL_PURUPURU_FR },
         { "None",     OPTION_VAL_NONE_FR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_FR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FR },
         { "Purupuru", OPTION_VAL_PURUPURU_FR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_FR },
         { "None",     OPTION_VAL_NONE_FR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_FR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_FR },
         { "Purupuru", OPTION_VAL_PURUPURU_FR },
         { "None",     OPTION_VAL_NONE_FR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_FR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_FR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_FR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_FR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_FR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_FR,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_FR,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_FR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_FR,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_FR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_FR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_FR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_FR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FR },
         { "2x", OPTION_VAL_2_O26_FR },
         { "3x", OPTION_VAL_3X_FR },
         { "4x", OPTION_VAL_4_FR },
         { "5x", OPTION_VAL_5X_FR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FR },
         { "20%",  OPTION_VAL_20_FR },
         { "30%",  OPTION_VAL_30_FR },
         { "40%",  OPTION_VAL_40_FR },
         { "50%",  OPTION_VAL_50_FR },
         { "60%",  OPTION_VAL_60_FR },
         { "70%",  OPTION_VAL_70_FR },
         { "80%",  OPTION_VAL_80_FR },
         { "90%",  OPTION_VAL_90_FR },
         { "100%", OPTION_VAL_100_O47_FR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FR },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FR },
         { "2x", OPTION_VAL_2_O26_FR },
         { "3x", OPTION_VAL_3X_FR },
         { "4x", OPTION_VAL_4_FR },
         { "5x", OPTION_VAL_5X_FR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FR },
         { "20%",  OPTION_VAL_20_FR },
         { "30%",  OPTION_VAL_30_FR },
         { "40%",  OPTION_VAL_40_FR },
         { "50%",  OPTION_VAL_50_FR },
         { "60%",  OPTION_VAL_60_FR },
         { "70%",  OPTION_VAL_70_FR },
         { "80%",  OPTION_VAL_80_FR },
         { "90%",  OPTION_VAL_90_FR },
         { "100%", OPTION_VAL_100_O47_FR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FR },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FR },
         { "2x", OPTION_VAL_2_O26_FR },
         { "3x", OPTION_VAL_3X_FR },
         { "4x", OPTION_VAL_4_FR },
         { "5x", OPTION_VAL_5X_FR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FR },
         { "20%",  OPTION_VAL_20_FR },
         { "30%",  OPTION_VAL_30_FR },
         { "40%",  OPTION_VAL_40_FR },
         { "50%",  OPTION_VAL_50_FR },
         { "60%",  OPTION_VAL_60_FR },
         { "70%",  OPTION_VAL_70_FR },
         { "80%",  OPTION_VAL_80_FR },
         { "90%",  OPTION_VAL_90_FR },
         { "100%", OPTION_VAL_100_O47_FR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_FR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_FR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_FR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_FR },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_FR },
         { "2x", OPTION_VAL_2_O26_FR },
         { "3x", OPTION_VAL_3X_FR },
         { "4x", OPTION_VAL_4_FR },
         { "5x", OPTION_VAL_5X_FR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_FR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_FR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_FR },
         { "BLUE 03",           OPTION_VAL_BLUE_FR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_FR },
         { "GREEN 05",          OPTION_VAL_GREEN_FR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_FR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_FR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_FR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_FR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_FR },
         { "RED 11",            OPTION_VAL_RED_FR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_FR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_FR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_FR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_FR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_FR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_FR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_FR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_FR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_FR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_FR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_FR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_FR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_FR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_FR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_FR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_FR },
         { "WHITE 28",          OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_FR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_FR },
         { "20%",  OPTION_VAL_20_FR },
         { "30%",  OPTION_VAL_30_FR },
         { "40%",  OPTION_VAL_40_FR },
         { "50%",  OPTION_VAL_50_FR },
         { "60%",  OPTION_VAL_60_FR },
         { "70%",  OPTION_VAL_70_FR },
         { "80%",  OPTION_VAL_80_FR },
         { "90%",  OPTION_VAL_90_FR },
         { "100%", OPTION_VAL_100_O47_FR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_fr = {
   option_cats_fr,
   option_defs_fr
};

/* RETRO_LANGUAGE_GA */

#define CATEGORY_SYSTEM_LABEL_GA "Córas"
#define CATEGORY_SYSTEM_INFO_0_GA "Cumraigh réigiún, teanga, BIOS agus socruithe crua-earraí bunúsacha."
#define CATEGORY_VIDEO_LABEL_GA "Físeán"
#define CATEGORY_VIDEO_INFO_0_GA "Cumraigh socruithe taifeach, trédhearcachta neamhspleách ar ord agus éifeachtaí amhairc."
#define CATEGORY_PERFORMANCE_LABEL_GA "Feidhmíocht"
#define CATEGORY_PERFORMANCE_INFO_0_GA "Cumraigh socruithe rindreála snáithithe agus scipeála frámaí."
#define CATEGORY_HACKS_LABEL_GA "Cleasanna Aithrise"
#define CATEGORY_HACKS_INFO_0_GA "Cumraigh sáruithe scáileáin leathan, luas lódála GD-ROM agus socruithe athsholáthair uigeachta."
#define CATEGORY_INPUT_LABEL_GA "Ionchur"
#define CATEGORY_INPUT_INFO_0_GA "Cumraigh socruithe an gamepad agus an ghunna solais."
#define CATEGORY_EXPANSIONS_LABEL_GA "Sliotáin Leathnaithe Rialaitheora"
#define CATEGORY_EXPANSIONS_INFO_0_GA "Roghnaigh an gléas (VMU, gléas creathadh) atá plugáilte isteach i ngach sliotán leathnúcháin rialtóra."
#define CATEGORY_VMU_LABEL_GA "Aonad Cuimhne Amhairc"
#define CATEGORY_VMU_INFO_0_GA "Cumraigh comhaid sábhála VMU in aghaidh an chluiche agus socruithe infheictheachta VMU ar an scáileán."
#define CORE_OPTION_NAME_REGION_LABEL_GA "Réigiún"
#define OPTION_VAL_JAPAN_GA "An tSeapáin"
#define OPTION_VAL_USA_GA "Stáit Aontaithe Mheiriceá"
#define OPTION_VAL_EUROPE_GA "An Eoraip"
#define OPTION_VAL_DEFAULT_GA "Réamhshocrú"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_GA "Teanga"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_GA "Athraíonn sé an teanga a úsáideann an BIOS agus aon chluiche ina bhfuil ilteangacha."
#define OPTION_VAL_JAPANESE_GA "Seapáinis"
#define OPTION_VAL_ENGLISH_GA "Béarla"
#define OPTION_VAL_GERMAN_GA "Gearmáinis"
#define OPTION_VAL_FRENCH_GA "Fraincis"
#define OPTION_VAL_SPANISH_GA "Spáinnis"
#define OPTION_VAL_ITALIAN_GA "Iodáilis"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_GA "BIOS HLE (Atosú Riachtanach)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_GA "Éigean úsáid BIOS aithrise ardleibhéil."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_GA "Cumasaigh DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_GA "Cumasaigh aithris ar DSP fuaime (próiseálaí comhartha digiteach) an Dreamcast. Feabhsaíonn sé cruinneas an fhuaime ghinte, ach méadaíonn sé na ceanglais feidhmíochta."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_GA "Ceadaigh Cnaipí Seirbhíse Arcade"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_GA "Cumasaíonn sé an cnaipe SERVICE le haghaidh cluichí stuara, chun socruithe caibinéid a iontráil."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_GA "Socraigh Cluichí NAOMI go Súgradh Saor in Aisce"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_GA "Athraigh socruithe mona an chluiche le haghaidh súgartha saor in aisce."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_GA "Aithris Oiriúntóra Leathanbhanda"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_GA "Déan aithris ar an oiriúntóir leathanbhanda Ethernet in ionad an mhóideim. (Atosú Riachtanach)"
#define CORE_OPTION_NAME_UPNP_LABEL_GA "Cumasaigh UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_GA "Bain úsáid as UPnP chun do ródaire Idirlín a chumrú go huathoibríoch le haghaidh cluichí ar líne."
#define CORE_OPTION_NAME_DCNET_LABEL_GA "Úsáid DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_GA "Bain úsáid as seirbhís scamall DCNet le haghaidh rochtain Idirlín Dreamcast."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_GA "Rún Inmheánach"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_GA "Modhnaigh réiteach rindreála."
#define OPTION_VAL_320X240_GA "320x240 (Leath)"
#define OPTION_VAL_640X480_GA "640x480 (Dúchasach)"
#define OPTION_VAL_800X600_GA NULL
#define OPTION_VAL_960X720_GA NULL
#define OPTION_VAL_1024X768_GA NULL
#define OPTION_VAL_1280X960_GA NULL
#define OPTION_VAL_1440X1080_GA NULL
#define OPTION_VAL_1600X1200_GA NULL
#define OPTION_VAL_1920X1440_GA NULL
#define OPTION_VAL_2560X1920_GA NULL
#define OPTION_VAL_2880X2160_GA NULL
#define OPTION_VAL_3200X2400_GA NULL
#define OPTION_VAL_3840X2880_GA NULL
#define OPTION_VAL_4480X3360_GA NULL
#define OPTION_VAL_5120X3840_GA NULL
#define OPTION_VAL_5760X4320_GA NULL
#define OPTION_VAL_6400X4800_GA NULL
#define OPTION_VAL_7040X5280_GA NULL
#define OPTION_VAL_7680X5760_GA NULL
#define OPTION_VAL_8320X6240_GA NULL
#define OPTION_VAL_8960X6720_GA NULL
#define OPTION_VAL_9600X7200_GA NULL
#define OPTION_VAL_10240X7680_GA NULL
#define OPTION_VAL_10880X8160_GA NULL
#define OPTION_VAL_11520X8640_GA NULL
#define OPTION_VAL_12160X9120_GA NULL
#define OPTION_VAL_12800X9600_GA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_GA "Cineál Cábla"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_GA "An cineál comhartha aschuir. Is é 'Teilifís (Ilchodach)' an ceann is mó a dtacaítear leis."
#define OPTION_VAL_VGA_GA NULL
#define OPTION_VAL_TV_RGB_GA NULL
#define OPTION_VAL_TV_COMPOSITE_GA "Teilifís (Ilchodach)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_GA "Caighdeán Craolta"
#define OPTION_VAL_NTSC_GA NULL
#define OPTION_VAL_PAL_GA "PAL (Domhanda)"
#define OPTION_VAL_PAL_N_GA "PAL-N (An Airgintín, Paragua, Uragua)"
#define OPTION_VAL_PAL_M_GA "PAL-M (An Bhrasaíl)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_GA "Treoshuíomh an Scáileáin"
#define OPTION_VAL_HORIZONTAL_GA "Cothrománach"
#define OPTION_VAL_VERTICAL_GA "Ingearach"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_GA "Sórtáil Alfa"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_GA "Roghnaigh conas a shórtálfar na polagáin thrédhearcacha."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_GA "In aghaidh an Stiall (tapa, is lú cruinn)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_GA "In aghaidh an Triantáin (gnáth)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_GA "In aghaidh an Phicteilín (cruinn, ach is moille)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_GA "Méid Maoláin Picteil Carntha"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_GA "B’fhéidir go mbeadh luachanna níos airde ag teastáil chun taifeach níos airde a aschur i gceart."
#define OPTION_VAL_512MB_GA NULL
#define OPTION_VAL_1GB_GA NULL
#define OPTION_VAL_2GB_GA NULL
#define OPTION_VAL_4GB_GA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_GA "Uasmhéid Sraitheanna Trédhearcacha"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_GA "D’fhéadfadh luachanna níos airde a bheith ag teastáil le haghaidh radhairc chasta."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_GA "Aithris iomlán maoláin fráma"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_GA "Cumasaigh aithris iomlán maolán fráma i VRAM. Tá sé seo úsáideach do chluichí a léann nó a scríobhann an maolán fráma i VRAM go díreach. Nuair a bhíonn sé cumasaithe, cuirtear iallach ar an Taifeach Inmheánach go 640x480 agus d’fhéadfadh tionchar mór a bheith aige ar fheidhmíocht."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_GA "Cumasaigh Maolán RTT (Rindreáil go hUigeacht)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_GA "Cóipeáil uigeachtaí rindreáilte ar ais ón GPU go VRAM. De ghnáth, bíonn an rogha seo cumasaithe le haghaidh cluichí a bhfuil gá leo. Nuair a bhíonn sé cumasaithe, díchumasaítear uas-scálú rindreála uigeachtaí agus d’fhéadfadh tionchar a bheith aige ar fheidhmíocht."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_GA "Mipmapáil"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_GA "Nuair a bheidh sé cumasaithe, úsáidfidh uigeachtaí leagan níos lú díobh féin nuair a bheidh siad níos faide ar shiúl, rud a fhéadann feidhmíocht a mhéadú agus lonrú a laghdú."
#define CORE_OPTION_NAME_FOG_LABEL_GA "Éifeachtaí Ceo"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_GA "Modhnóir Imleabhair"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_GA "Gné GPU Dreamcast a úsáideann cluichí de ghnáth chun scáthanna réad a tharraingt. Ba chóir é seo a chumasú de ghnáth - is gnách go mbíonn an tionchar ar fheidhmíocht beag go neamhbhríoch."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_GA "Scagadh Ainiseatrópach"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_GA "Feabhsaigh cáilíocht na n-uigeachtaí ar dhromchlaí atá ag uillinneacha féachana claonta i gcomparáid leis an gceamara. Bíonn luachanna níos airde níos déine ar an GPU. Ní bheidh feidhm ag athruithe ar an socrú seo ach amháin tar éis atosú."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_GA "Scagadh Uigeachta"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_GA "An modh scagtha uigeachta le húsáid. Is féidir é seo a úsáid chun modh scagtha uigeachta áirithe a chur i bhfeidhm ar gach uigeacht chun cuma níos géire (nó níos míne) a fháil ná an Réamhshocrú. D’fhéadfadh luachanna seachas an Réamhshocrú fadhbanna éagsúla rindreála a chruthú. Ní bheidh feidhm ag athruithe ar an socrú seo ach amháin tar éis atosú."
#define OPTION_VAL_1_GA "Fórsáil an Chomharsa is Gaire"
#define OPTION_VAL_2_GA "Fórsa Líneach"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_GA "Moill ar Mhalartú Fráma"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_GA "Úsáideach chun scáileáin splancacha nó físeáin glitcheacha a sheachaint. Ní mholtar é ar ardáin mhalla."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_GA "Athruithe ar Ráta Fráma a Bhraith"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_GA "Cuir an comhéadan ar an eolas nuair a athraíonn an ráta fráma inmheánach (e.g. ó 60 fps go 30 fps). Feabhsaíonn sé luas fráma i gcluichí a ritheann ag 30 fps nó 20 fps faoi ghlas, ach ba chóir é a dhíchumasú i gcluichí le rátaí fráma neamhghlasáilte (éagobhsaí) (e.g. Ecco the Dolphin, Unreal Tournament). Tabhair faoi deara: Níl sé ar fáil nuair a bhíonn 'Auto Skip Frame' cumasaithe."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_GA "Scagaire Iarphróiseála PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_GA "Déan an íomhá rindreáilte a iarphróiseáil chun éifeachtaí a insamhladh atá sainiúil don GPU PowerVR2 agus do chomharthaí físe analógacha."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_GA "Uasghrádú Uigeachta (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_GA "Feabhsaigh grafaicí ealaíne picteilíní 2T lámhdhéanta. Níor cheart é a úsáid ach le cluichí picteilíní 2T."
#define OPTION_VAL_2_O26_GA NULL
#define OPTION_VAL_4_GA NULL
#define OPTION_VAL_6_GA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_GA "Uasmhéid Scagtha Uasmhéid Uigeachta"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_GA "Roghnaigh uasluach méide le haghaidh uigeachta atá le huaslagadh, mura bhfuil méid na huigeachta níos airde ná an luach roghnaithe ní dhéanfar í a uaslagadh."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_GA "Idirshuíomh Doimhneachta Dúchasach"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_GA "Cabhraíonn sé le truailliú uigeachta agus fadhbanna doimhneachta ar GPUanna AMD. Is féidir leis cabhrú le GPUanna Intel i gcásanna áirithe freisin."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_GA "Deisigh Imeall Fuilteach Uasghrádaithe"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_GA "Cuidíonn sé le cás fuilithe uigeachta agus uas-scálú á dhéanamh. Is féidir le díchumasú cabhrú má tá picteilíní ag saobhadh agus uas-scálú á dhéanamh i gcluichí 2T (MVC2, CVS, KOF, srl.)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_GA "Rindreáil Snáithithe"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_GA "Ritheann sé an GPU agus an LAP ar shnáitheanna difriúla. Moltar go mór."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_GA "Fráma Uathoibríoch Léim"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_GA "Léim go huathoibríoch frámaí nuair a bhíonn an t-aithriseoir ag rith go mall. Tabhair faoi deara: Ní bhaineann an socrú seo ach amháin nuair a bhíonn 'Rindreáil Snáithithe' cumasaithe."
#define OPTION_VAL_SOME_GA "Gnáth"
#define OPTION_VAL_MORE_GA "Uasmhéid"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_GA "Scipeáil Fráma"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_GA "Socraíonn sé líon na bhfrámaí le léim idir gach fráma a thaispeántar."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_GA "Aicearra Leathan-Scáileáin (Atosú Riachtanach)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_GA "Gníomhaíonn sé aicearraí a ligeann do chluichí áirithe a thaispeáint i bhformáid scáileáin leathan."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_GA "Haic Leathan-Scáileáin"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_GA "Tarraing geoiméadracht lasmuigh den ghnáthchóimheas gné 4:3. D’fhéadfadh sé go mbeadh fabhtanna grafacha sna ceantair nochta."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_GA "Luchtú Tapa GD-ROM (míchruinn)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_GA "Luasghéadaíonn sé luchtú GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_GA "Modúl RAM Dreamcast 32MB"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_GA "Cumasaíonn sé Mod RAM 32MB do Dreamcast. D’fhéadfadh sé difear a dhéanamh do chomhoiriúnacht"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_GA "Róchlogáil/fo-chlogáil LAP SH4"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_GA "Athraigh clog LAP príomhúil SH4 ón 200 MHz réamhshocraithe. D’fhéadfadh tearc-chlogáil cabhrú le hardáin a mhoilliú. D’fhéadfadh ró-chlogáil an ráta fráma a mhéadú do roinnt cluichí. Bain úsáid as go cúramach."
#define OPTION_VAL_100_GA NULL
#define OPTION_VAL_110_GA NULL
#define OPTION_VAL_120_GA NULL
#define OPTION_VAL_130_GA NULL
#define OPTION_VAL_140_GA NULL
#define OPTION_VAL_150_GA NULL
#define OPTION_VAL_160_GA NULL
#define OPTION_VAL_170_GA NULL
#define OPTION_VAL_180_GA NULL
#define OPTION_VAL_190_GA NULL
#define OPTION_VAL_200_GA NULL
#define OPTION_VAL_210_GA NULL
#define OPTION_VAL_220_GA NULL
#define OPTION_VAL_230_GA NULL
#define OPTION_VAL_240_GA NULL
#define OPTION_VAL_250_GA NULL
#define OPTION_VAL_260_GA NULL
#define OPTION_VAL_270_GA NULL
#define OPTION_VAL_280_GA NULL
#define OPTION_VAL_290_GA NULL
#define OPTION_VAL_300_GA NULL
#define OPTION_VAL_310_GA NULL
#define OPTION_VAL_320_GA NULL
#define OPTION_VAL_330_GA NULL
#define OPTION_VAL_340_GA NULL
#define OPTION_VAL_350_GA NULL
#define OPTION_VAL_360_GA NULL
#define OPTION_VAL_370_GA NULL
#define OPTION_VAL_380_GA NULL
#define OPTION_VAL_390_GA NULL
#define OPTION_VAL_400_GA NULL
#define OPTION_VAL_410_GA NULL
#define OPTION_VAL_420_GA NULL
#define OPTION_VAL_430_GA NULL
#define OPTION_VAL_440_GA NULL
#define OPTION_VAL_450_GA NULL
#define OPTION_VAL_460_GA NULL
#define OPTION_VAL_470_GA NULL
#define OPTION_VAL_480_GA NULL
#define OPTION_VAL_490_GA NULL
#define OPTION_VAL_500_GA NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_GA "Luchtaigh Uigeachtaí Saincheaptha"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_GA "Luchtaigh uigeachtaí saincheaptha atá suite sa bhfillteán 'system/dc/textures/<game-id>/'."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_GA "Réamhluchtaigh Uigeachtaí Saincheaptha"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_GA "Réamhlódáil uigeachtaí saincheaptha ag tús an chluiche. Féadfaidh sé feidhmíocht a fheabhsú ach méadaíonn sé úsáid cuimhne."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_GA "Uigeachtaí Dumpála"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_GA "Gach uair a úsáideann an cluiche uigeacht nua, sábhálfar í mar chomhad .png sa bhfillteán 'system/dc/texdump/<game-id>/'."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_GA "Uigeachtaí Athsholáthraithe Dumpála"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_GA "Dumpáil uigeachtaí atá curtha ina n-áit cheana féin le huigeachtaí saincheaptha i gcónaí."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_GA "Crios Marbh Bata Analógach"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_GA "Roghnaigh cé chomh fada is gá duit an bata analógach a bhrú sula dtosaíonn sé á phróiseáil."
#define OPTION_VAL_0_GA NULL
#define OPTION_VAL_5_GA NULL
#define OPTION_VAL_10_GA NULL
#define OPTION_VAL_15_GA NULL
#define OPTION_VAL_20_GA NULL
#define OPTION_VAL_25_GA NULL
#define OPTION_VAL_30_GA NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_GA "Spriocdhíon Marbh"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_GA "Roghnaigh cé mhéad a chaithfidh tú an truicear a bhrú sula dtosaíonn sé á phróiseáil."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_GA "Spreagthóirí Digiteacha"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_GA "Nuair a bheidh siad cumasaithe, feidhmeoidh na spreagthóirí mar chnaipí rialta, rud a chiallaíonn go ndéanfar iad a phróiseáil mar chnaipí brúite go hiomlán nó gan brú ar chor ar bith, gan aon rud eatarthu."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_GA "Aschuir Dhigiteacha Craolta"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_GA "Aschuir dhigiteacha craolta agus staid aiseolais fórsa ar phort TCP 8000. Ag luí leis an rogha MAME \"-output network\"."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_GA "Taispeáin Socruithe Gunna Solais"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_GA "Cumasaigh cumraíocht roghanna taispeána crosaire gunna solais. TABHAIR FAOI DEARA: B’fhéidir go mbeidh ort an Roghchlár Tapa a athrú chun an socrú seo a chur i bhfeidhm."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_GA "Scálú Méid Croslíne Gunna"
#define OPTION_VAL_50_GA NULL
#define OPTION_VAL_60_GA NULL
#define OPTION_VAL_70_GA NULL
#define OPTION_VAL_80_GA NULL
#define OPTION_VAL_90_GA NULL
#define OPTION_VAL_100_O47_GA NULL
#define OPTION_VAL_110_O47_GA NULL
#define OPTION_VAL_120_O47_GA NULL
#define OPTION_VAL_130_O47_GA NULL
#define OPTION_VAL_140_O47_GA NULL
#define OPTION_VAL_150_O47_GA NULL
#define OPTION_VAL_160_O47_GA NULL
#define OPTION_VAL_170_O47_GA NULL
#define OPTION_VAL_180_O47_GA NULL
#define OPTION_VAL_190_O47_GA NULL
#define OPTION_VAL_200_O47_GA NULL
#define OPTION_VAL_210_O47_GA NULL
#define OPTION_VAL_220_O47_GA NULL
#define OPTION_VAL_230_O47_GA NULL
#define OPTION_VAL_240_O47_GA NULL
#define OPTION_VAL_250_O47_GA NULL
#define OPTION_VAL_260_O47_GA NULL
#define OPTION_VAL_270_O47_GA NULL
#define OPTION_VAL_280_O47_GA NULL
#define OPTION_VAL_290_O47_GA NULL
#define OPTION_VAL_300_O47_GA NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_GA "Taispeáin Croslíne don Ghunna 1"
#define OPTION_VAL_WHITE_GA "Bán"
#define OPTION_VAL_RED_GA "Dearg"
#define OPTION_VAL_GREEN_GA "Glas"
#define OPTION_VAL_BLUE_GA "Gorm"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_GA "Taispeáin Croslíne don Ghunna 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_GA "Taispeáin Croslíne don Ghunna 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_GA "Taispeáin Croslíne don Ghunna 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_GA "Gléas i Sliotán Leathnaithe A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_GA "Roghnaigh an gléas atá plugáilte isteach sa sliotán leathnúcháin A1 (port A sliotán 1)."
#define OPTION_VAL_VMU_GA NULL
#define OPTION_VAL_PURUPURU_GA "Pacáiste Creathadh"
#define OPTION_VAL_DREAMPOTATO_GA NULL
#define OPTION_VAL_NONE_GA "Dada"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_GA "Gléas i Sliotán Leathnaithe A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_GA "Roghnaigh an gléas atá plugáilte isteach sa sliotán leathnúcháin A2 (port A sliotán 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_GA "Gléas i Sliotán Leathnaithe B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_GA "Roghnaigh an gléas atá plugáilte isteach sa sliotán leathnúcháin B1 (port B sliotán 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_GA "Gléas i Sliotán Leathnaithe B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_GA "Roghnaigh an gléas atá plugáilte isteach sa sliotán leathnúcháin B2 (port B sliotán 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_GA "Gléas i Sliotán Leathnaithe C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_GA "Roghnaigh an gléas atá plugáilte isteach sa sliotán leathnúcháin C1 (port C sliotán 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_GA "Gléas i Sliotán Leathnaithe C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_GA "Roghnaigh an gléas atá plugáilte isteach sa sliotán leathnúcháin C2 (port C sliotán 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_GA "Gléas i Sliotán Leathnaithe D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_GA "Roghnaigh an gléas atá plugáilte isteach sa sliotán leathnúcháin D1 (port D sliotán 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_GA "Gléas i Sliotán Leathnaithe D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_GA "Roghnaigh an gléas atá plugáilte isteach sa sliotán leathnúcháin D2 (port D sliotán 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_GA "Aonaid/Córais Cuimhne Amhairc In Aghaidh an Chluiche (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_GA "VMUanna In aghaidh an Chluiche"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_GA "Nuair a bhíonn sé díchumasaithe, roinneann gach cluiche suas le 8 gcomhad sábhála VMU (A1/A2/B1/B2/C1/C2/D1/D2) atá suite i bhfillteán córais RetroArch.\nCruthaíonn an socrú 'VMU A1' comhad VMU uathúil 'A1' i bhfillteán sábhála RetroArch do gach cluiche a sheoltar.\nCruthaíonn an socrú 'Gach VMU' suas le 8 gcomhad VMU uathúla (A1/A2/B1/B2/C1/C2/D1/D2) do gach cluiche a sheoltar."
#define OPTION_VAL_VMU_A1_GA NULL
#define OPTION_VAL_ALL_VMUS_GA "Gach VMUanna"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_GA "Fuaimeanna Aonaid/Córais Cuimhne Amhairc (VMU)"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_GA "Fuaimeanna VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_GA "Nuair a bhíonn sé cumasaithe, seinntear bípeanna VMU."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_GA "Stóráil DreamPotato VMU"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_GA "Cumasaigh stóráil sonraí VMU ar an aithriseoir DreamPotato ceangailte. Nuair a bhíonn sé cumasaithe, ní úsáidtear an comhad sábhála VMU áitiúil."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_GA "Taispeáin Socruithe Taispeána Aonaid/Córais Cuimhne Amhairc (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_GA "Taispeáin Socruithe Taispeána VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_GA "Cumasaigh cumraíocht infheictheachta, méide, suíomh agus datha scáileáin LCD VMU aithrisithe. TABHAIR FAOI DEARA: B’fhéidir go mbeidh ort an Roghchlár Tapa a athrú chun an socrú seo a chur i bhfeidhm."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_GA "Taispeántas Scáileán 1 VMU"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_GA "Suíomh Scáileáin VMU 1"
#define OPTION_VAL_UPPER_LEFT_GA "Uachtarach ar Chlé"
#define OPTION_VAL_UPPER_RIGHT_GA "Uachtarach ar Dheis"
#define OPTION_VAL_LOWER_LEFT_GA "Íochtarach ar Chlé"
#define OPTION_VAL_LOWER_RIGHT_GA "Íochtarach ar Dheis"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_GA "Méid Scáileáin VMU 1"
#define OPTION_VAL_1X_GA NULL
#define OPTION_VAL_3X_GA NULL
#define OPTION_VAL_5X_GA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_GA "Scáileán VMU 1 Picteilín Ar Dath"
#define OPTION_VAL_DEFAULT_ON_00_GA "Réamhshocrú AR"
#define OPTION_VAL_DEFAULT_OFF_01_GA "Réamhshocrú AS"
#define OPTION_VAL_BLACK_02_GA "Dubh"
#define OPTION_VAL_LIGHT_BLUE_04_GA "Gorm Éadrom"
#define OPTION_VAL_CYAN_06_GA "Cian"
#define OPTION_VAL_CYAN_BLUE_07_GA "Gorm Cian"
#define OPTION_VAL_LIGHT_GREEN_08_GA "Glas Éadrom"
#define OPTION_VAL_CYAN_GREEN_09_GA "Glas Cian"
#define OPTION_VAL_LIGHT_CYAN_10_GA "Cian Éadrom"
#define OPTION_VAL_PURPLE_12_GA "Corcra"
#define OPTION_VAL_LIGHT_PURPLE_13_GA "Corcra Éadrom"
#define OPTION_VAL_YELLOW_14_GA "Buí"
#define OPTION_VAL_GRAY_15_GA "Liath"
#define OPTION_VAL_LIGHT_PURPLE_2_16_GA "Corcra Éadrom (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_GA "Glas Éadrom (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_GA "Glas Éadrom (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_GA "Cian Éadrom (2)"
#define OPTION_VAL_LIGHT_RED_2_20_GA "Dearg Éadrom (2)"
#define OPTION_VAL_MAGENTA_21_GA "Maigeanta"
#define OPTION_VAL_LIGHT_PURPLE_3_22_GA "Corcra Éadrom (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_GA "Oráiste Éadrom"
#define OPTION_VAL_ORANGE_24_GA "Oráiste"
#define OPTION_VAL_LIGHT_PURPLE_4_25_GA "Corcra Éadrom (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_GA "Buí Éadrom"
#define OPTION_VAL_LIGHT_YELLOW_2_27_GA "Buí Éadrom (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_GA "Scáileán VMU 1 Picteilín As Dath"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_GA "Teimhneacht Scáileáin 1 VMU"
#define OPTION_VAL_40_GA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_GA "Taispeántas Scáileán 2 VMU"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_GA "Suíomh Scáileáin VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_GA "Méid Scáileáin VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_GA "Scáileán VMU 2 Picteilín Ar Dath"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_GA "Scáileán VMU 2 Picteilín As Dath"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_GA "Teimhneacht Scáileáin VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_GA "Taispeántas Scáileán VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_GA "Scáileán VMU 3 Seasamh"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_GA "Méid Scáileáin VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_GA "Scáileán VMU 3 Picteilín Ar Dath"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_GA "Scáileán VMU 3 Picteilín As Dath"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_GA "Teimhneacht Scáileáin VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_GA "Taispeántas Scáileán VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_GA "Scáileán VMU 4 Seasamh"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_GA "Méid Scáileáin VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_GA "Scáileán VMU 4 Picteilín Ar Dath"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_GA "Scáileán VMU 4 Picteilín As Dath"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_GA "Teimhneacht Scáileáin VMU 4"

struct retro_core_option_v2_category option_cats_ga[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_GA,
      CATEGORY_SYSTEM_INFO_0_GA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_GA,
      CATEGORY_VIDEO_INFO_0_GA
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_GA,
      CATEGORY_PERFORMANCE_INFO_0_GA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_GA,
      CATEGORY_HACKS_INFO_0_GA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_GA,
      CATEGORY_INPUT_INFO_0_GA
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_GA,
      CATEGORY_EXPANSIONS_INFO_0_GA
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_GA,
      CATEGORY_VMU_INFO_0_GA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ga[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_GA,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_GA },
         { "USA",     OPTION_VAL_USA_GA },
         { "Europe",  OPTION_VAL_EUROPE_GA },
         { "Default", OPTION_VAL_DEFAULT_GA },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_GA,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_GA },
         { "English",  OPTION_VAL_ENGLISH_GA },
         { "German",   OPTION_VAL_GERMAN_GA },
         { "French",   OPTION_VAL_FRENCH_GA },
         { "Spanish",  OPTION_VAL_SPANISH_GA },
         { "Italian",  OPTION_VAL_ITALIAN_GA },
         { "Default",  OPTION_VAL_DEFAULT_GA },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_GA,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_GA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_GA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_GA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_GA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_GA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_GA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_GA,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_GA },
         { "640x480",    OPTION_VAL_640X480_GA },
         { "800x600",    OPTION_VAL_800X600_GA },
         { "960x720",    OPTION_VAL_960X720_GA },
         { "1024x768",   OPTION_VAL_1024X768_GA },
         { "1280x960",   OPTION_VAL_1280X960_GA },
         { "1440x1080",  OPTION_VAL_1440X1080_GA },
         { "1600x1200",  OPTION_VAL_1600X1200_GA },
         { "1920x1440",  OPTION_VAL_1920X1440_GA },
         { "2560x1920",  OPTION_VAL_2560X1920_GA },
         { "2880x2160",  OPTION_VAL_2880X2160_GA },
         { "3200x2400",  OPTION_VAL_3200X2400_GA },
         { "3840x2880",  OPTION_VAL_3840X2880_GA },
         { "4480x3360",  OPTION_VAL_4480X3360_GA },
         { "5120x3840",  OPTION_VAL_5120X3840_GA },
         { "5760x4320",  OPTION_VAL_5760X4320_GA },
         { "6400x4800",  OPTION_VAL_6400X4800_GA },
         { "7040x5280",  OPTION_VAL_7040X5280_GA },
         { "7680x5760",  OPTION_VAL_7680X5760_GA },
         { "8320x6240",  OPTION_VAL_8320X6240_GA },
         { "8960x6720",  OPTION_VAL_8960X6720_GA },
         { "9600x7200",  OPTION_VAL_9600X7200_GA },
         { "10240x7680", OPTION_VAL_10240X7680_GA },
         { "10880x8160", OPTION_VAL_10880X8160_GA },
         { "11520x8640", OPTION_VAL_11520X8640_GA },
         { "12160x9120", OPTION_VAL_12160X9120_GA },
         { "12800x9600", OPTION_VAL_12800X9600_GA },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_GA,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_GA },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_GA },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_GA },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_GA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_GA },
         { "PAL",     OPTION_VAL_PAL_GA },
         { "PAL_N",   OPTION_VAL_PAL_N_GA },
         { "PAL_M",   OPTION_VAL_PAL_M_GA },
         { "Default", OPTION_VAL_DEFAULT_GA },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_GA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_GA },
         { "vertical",   OPTION_VAL_VERTICAL_GA },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_GA,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_GA },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_GA },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_GA },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_GA,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_GA },
         { "1GB",   OPTION_VAL_1GB_GA },
         { "2GB",   OPTION_VAL_2GB_GA },
         { "4GB",   OPTION_VAL_4GB_GA },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_GA,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_GA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_GA,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_GA,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_GA },
         { "1",  OPTION_VAL_1_GA },
         { "2",  OPTION_VAL_2_GA },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_GA,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_GA },
         { "4",  OPTION_VAL_4_GA },
         { "6",  OPTION_VAL_6_GA },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_GA,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_GA,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_GA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_GA,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_GA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_GA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_GA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_GA },
         { "more", OPTION_VAL_MORE_GA },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_GA,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_GA },
         { "110", OPTION_VAL_110_GA },
         { "120", OPTION_VAL_120_GA },
         { "130", OPTION_VAL_130_GA },
         { "140", OPTION_VAL_140_GA },
         { "150", OPTION_VAL_150_GA },
         { "160", OPTION_VAL_160_GA },
         { "170", OPTION_VAL_170_GA },
         { "180", OPTION_VAL_180_GA },
         { "190", OPTION_VAL_190_GA },
         { "200", OPTION_VAL_200_GA },
         { "210", OPTION_VAL_210_GA },
         { "220", OPTION_VAL_220_GA },
         { "230", OPTION_VAL_230_GA },
         { "240", OPTION_VAL_240_GA },
         { "250", OPTION_VAL_250_GA },
         { "260", OPTION_VAL_260_GA },
         { "270", OPTION_VAL_270_GA },
         { "280", OPTION_VAL_280_GA },
         { "290", OPTION_VAL_290_GA },
         { "300", OPTION_VAL_300_GA },
         { "310", OPTION_VAL_310_GA },
         { "320", OPTION_VAL_320_GA },
         { "330", OPTION_VAL_330_GA },
         { "340", OPTION_VAL_340_GA },
         { "350", OPTION_VAL_350_GA },
         { "360", OPTION_VAL_360_GA },
         { "370", OPTION_VAL_370_GA },
         { "380", OPTION_VAL_380_GA },
         { "390", OPTION_VAL_390_GA },
         { "400", OPTION_VAL_400_GA },
         { "410", OPTION_VAL_410_GA },
         { "420", OPTION_VAL_420_GA },
         { "430", OPTION_VAL_430_GA },
         { "440", OPTION_VAL_440_GA },
         { "450", OPTION_VAL_450_GA },
         { "460", OPTION_VAL_460_GA },
         { "470", OPTION_VAL_470_GA },
         { "480", OPTION_VAL_480_GA },
         { "490", OPTION_VAL_490_GA },
         { "500", OPTION_VAL_500_GA },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_GA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_GA },
         { "5%",  OPTION_VAL_5_GA },
         { "10%", OPTION_VAL_10_GA },
         { "15%", OPTION_VAL_15_GA },
         { "20%", OPTION_VAL_20_GA },
         { "25%", OPTION_VAL_25_GA },
         { "30%", OPTION_VAL_30_GA },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_GA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_GA },
         { "5%",  OPTION_VAL_5_GA },
         { "10%", OPTION_VAL_10_GA },
         { "15%", OPTION_VAL_15_GA },
         { "20%", OPTION_VAL_20_GA },
         { "25%", OPTION_VAL_25_GA },
         { "30%", OPTION_VAL_30_GA },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_GA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_GA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_GA,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_GA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_GA },
         { "60%",  OPTION_VAL_60_GA },
         { "70%",  OPTION_VAL_70_GA },
         { "80%",  OPTION_VAL_80_GA },
         { "90%",  OPTION_VAL_90_GA },
         { "100%", OPTION_VAL_100_O47_GA },
         { "110%", OPTION_VAL_110_O47_GA },
         { "120%", OPTION_VAL_120_O47_GA },
         { "130%", OPTION_VAL_130_O47_GA },
         { "140%", OPTION_VAL_140_O47_GA },
         { "150%", OPTION_VAL_150_O47_GA },
         { "160%", OPTION_VAL_160_O47_GA },
         { "170%", OPTION_VAL_170_O47_GA },
         { "180%", OPTION_VAL_180_O47_GA },
         { "190%", OPTION_VAL_190_O47_GA },
         { "200%", OPTION_VAL_200_O47_GA },
         { "210%", OPTION_VAL_210_O47_GA },
         { "220%", OPTION_VAL_220_O47_GA },
         { "230%", OPTION_VAL_230_O47_GA },
         { "240%", OPTION_VAL_240_O47_GA },
         { "250%", OPTION_VAL_250_O47_GA },
         { "260%", OPTION_VAL_260_O47_GA },
         { "270%", OPTION_VAL_270_O47_GA },
         { "280%", OPTION_VAL_280_O47_GA },
         { "290%", OPTION_VAL_290_O47_GA },
         { "300%", OPTION_VAL_300_O47_GA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_GA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_GA },
         { "Red",      OPTION_VAL_RED_GA },
         { "Green",    OPTION_VAL_GREEN_GA },
         { "Blue",     OPTION_VAL_BLUE_GA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_GA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_GA },
         { "Red",      OPTION_VAL_RED_GA },
         { "Green",    OPTION_VAL_GREEN_GA },
         { "Blue",     OPTION_VAL_BLUE_GA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_GA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_GA },
         { "Red",      OPTION_VAL_RED_GA },
         { "Green",    OPTION_VAL_GREEN_GA },
         { "Blue",     OPTION_VAL_BLUE_GA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_GA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_GA },
         { "Red",      OPTION_VAL_RED_GA },
         { "Green",    OPTION_VAL_GREEN_GA },
         { "Blue",     OPTION_VAL_BLUE_GA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_GA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GA },
         { "Purupuru", OPTION_VAL_PURUPURU_GA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_GA },
         { "None",     OPTION_VAL_NONE_GA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_GA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GA },
         { "Purupuru", OPTION_VAL_PURUPURU_GA },
         { "None",     OPTION_VAL_NONE_GA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_GA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GA },
         { "Purupuru", OPTION_VAL_PURUPURU_GA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_GA },
         { "None",     OPTION_VAL_NONE_GA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_GA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GA },
         { "Purupuru", OPTION_VAL_PURUPURU_GA },
         { "None",     OPTION_VAL_NONE_GA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_GA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GA },
         { "Purupuru", OPTION_VAL_PURUPURU_GA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_GA },
         { "None",     OPTION_VAL_NONE_GA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_GA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GA },
         { "Purupuru", OPTION_VAL_PURUPURU_GA },
         { "None",     OPTION_VAL_NONE_GA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_GA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GA },
         { "Purupuru", OPTION_VAL_PURUPURU_GA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_GA },
         { "None",     OPTION_VAL_NONE_GA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_GA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GA },
         { "Purupuru", OPTION_VAL_PURUPURU_GA },
         { "None",     OPTION_VAL_NONE_GA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_GA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_GA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_GA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_GA },
         { "All VMUs", OPTION_VAL_ALL_VMUS_GA },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_GA,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_GA,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_GA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_GA,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_GA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_GA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_GA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_GA,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_GA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_GA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_GA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_GA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_GA },
         { "2x", OPTION_VAL_2_O26_GA },
         { "3x", OPTION_VAL_3X_GA },
         { "4x", OPTION_VAL_4_GA },
         { "5x", OPTION_VAL_5X_GA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GA },
         { "BLUE 03",           OPTION_VAL_BLUE_GA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GA },
         { "GREEN 05",          OPTION_VAL_GREEN_GA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GA },
         { "RED 11",            OPTION_VAL_RED_GA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GA },
         { "WHITE 28",          OPTION_VAL_WHITE_GA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GA },
         { "BLUE 03",           OPTION_VAL_BLUE_GA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GA },
         { "GREEN 05",          OPTION_VAL_GREEN_GA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GA },
         { "RED 11",            OPTION_VAL_RED_GA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GA },
         { "WHITE 28",          OPTION_VAL_WHITE_GA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_GA },
         { "20%",  OPTION_VAL_20_GA },
         { "30%",  OPTION_VAL_30_GA },
         { "40%",  OPTION_VAL_40_GA },
         { "50%",  OPTION_VAL_50_GA },
         { "60%",  OPTION_VAL_60_GA },
         { "70%",  OPTION_VAL_70_GA },
         { "80%",  OPTION_VAL_80_GA },
         { "90%",  OPTION_VAL_90_GA },
         { "100%", OPTION_VAL_100_O47_GA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_GA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_GA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_GA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_GA },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_GA },
         { "2x", OPTION_VAL_2_O26_GA },
         { "3x", OPTION_VAL_3X_GA },
         { "4x", OPTION_VAL_4_GA },
         { "5x", OPTION_VAL_5X_GA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GA },
         { "BLUE 03",           OPTION_VAL_BLUE_GA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GA },
         { "GREEN 05",          OPTION_VAL_GREEN_GA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GA },
         { "RED 11",            OPTION_VAL_RED_GA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GA },
         { "WHITE 28",          OPTION_VAL_WHITE_GA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GA },
         { "BLUE 03",           OPTION_VAL_BLUE_GA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GA },
         { "GREEN 05",          OPTION_VAL_GREEN_GA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GA },
         { "RED 11",            OPTION_VAL_RED_GA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GA },
         { "WHITE 28",          OPTION_VAL_WHITE_GA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_GA },
         { "20%",  OPTION_VAL_20_GA },
         { "30%",  OPTION_VAL_30_GA },
         { "40%",  OPTION_VAL_40_GA },
         { "50%",  OPTION_VAL_50_GA },
         { "60%",  OPTION_VAL_60_GA },
         { "70%",  OPTION_VAL_70_GA },
         { "80%",  OPTION_VAL_80_GA },
         { "90%",  OPTION_VAL_90_GA },
         { "100%", OPTION_VAL_100_O47_GA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_GA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_GA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_GA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_GA },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_GA },
         { "2x", OPTION_VAL_2_O26_GA },
         { "3x", OPTION_VAL_3X_GA },
         { "4x", OPTION_VAL_4_GA },
         { "5x", OPTION_VAL_5X_GA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GA },
         { "BLUE 03",           OPTION_VAL_BLUE_GA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GA },
         { "GREEN 05",          OPTION_VAL_GREEN_GA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GA },
         { "RED 11",            OPTION_VAL_RED_GA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GA },
         { "WHITE 28",          OPTION_VAL_WHITE_GA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GA },
         { "BLUE 03",           OPTION_VAL_BLUE_GA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GA },
         { "GREEN 05",          OPTION_VAL_GREEN_GA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GA },
         { "RED 11",            OPTION_VAL_RED_GA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GA },
         { "WHITE 28",          OPTION_VAL_WHITE_GA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_GA },
         { "20%",  OPTION_VAL_20_GA },
         { "30%",  OPTION_VAL_30_GA },
         { "40%",  OPTION_VAL_40_GA },
         { "50%",  OPTION_VAL_50_GA },
         { "60%",  OPTION_VAL_60_GA },
         { "70%",  OPTION_VAL_70_GA },
         { "80%",  OPTION_VAL_80_GA },
         { "90%",  OPTION_VAL_90_GA },
         { "100%", OPTION_VAL_100_O47_GA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_GA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_GA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_GA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_GA },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_GA },
         { "2x", OPTION_VAL_2_O26_GA },
         { "3x", OPTION_VAL_3X_GA },
         { "4x", OPTION_VAL_4_GA },
         { "5x", OPTION_VAL_5X_GA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GA },
         { "BLUE 03",           OPTION_VAL_BLUE_GA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GA },
         { "GREEN 05",          OPTION_VAL_GREEN_GA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GA },
         { "RED 11",            OPTION_VAL_RED_GA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GA },
         { "WHITE 28",          OPTION_VAL_WHITE_GA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GA },
         { "BLUE 03",           OPTION_VAL_BLUE_GA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GA },
         { "GREEN 05",          OPTION_VAL_GREEN_GA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GA },
         { "RED 11",            OPTION_VAL_RED_GA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GA },
         { "WHITE 28",          OPTION_VAL_WHITE_GA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_GA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_GA },
         { "20%",  OPTION_VAL_20_GA },
         { "30%",  OPTION_VAL_30_GA },
         { "40%",  OPTION_VAL_40_GA },
         { "50%",  OPTION_VAL_50_GA },
         { "60%",  OPTION_VAL_60_GA },
         { "70%",  OPTION_VAL_70_GA },
         { "80%",  OPTION_VAL_80_GA },
         { "90%",  OPTION_VAL_90_GA },
         { "100%", OPTION_VAL_100_O47_GA },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ga = {
   option_cats_ga,
   option_defs_ga
};

/* RETRO_LANGUAGE_GL */

#define CATEGORY_SYSTEM_LABEL_GL "Sistema"
#define CATEGORY_SYSTEM_INFO_0_GL "Configure a rexión, o idioma, a BIOS e a configuración do hardware base."
#define CATEGORY_VIDEO_LABEL_GL "Vídeo"
#define CATEGORY_VIDEO_INFO_0_GL "Configure a resolución, a transparencia independente da orde e os axustes de efectos visuais."
#define CATEGORY_PERFORMANCE_LABEL_GL "Rendemento"
#define CATEGORY_PERFORMANCE_INFO_0_GL "Configure as opcións de renderizado en fíos e omitir marcos."
#define CATEGORY_HACKS_LABEL_GL "Hacks de emulación"
#define CATEGORY_HACKS_INFO_0_GL "Configura substitucións de pantalla panorámica, velocidade de carga de GD-ROM e configuración de substitución de texturas."
#define CATEGORY_INPUT_LABEL_GL "Entrada"
#define CATEGORY_INPUT_INFO_0_GL "Configura a configuración do gamepad e da pistola lixeira."
#define CATEGORY_EXPANSIONS_LABEL_GL "Rañuras de expansión do controlador"
#define CATEGORY_EXPANSIONS_INFO_0_GL "Seleccione o dispositivo (VMU, dispositivo rumble) conectado a cada rañura de expansión do controlador."
#define CATEGORY_VMU_LABEL_GL "Unidade de Memoria Visual"
#define CATEGORY_VMU_INFO_0_GL "Configura os ficheiros de gardar VMU por xogo e os axustes de visibilidade da VMU en pantalla."
#define CORE_OPTION_NAME_REGION_LABEL_GL "Rexión"
#define OPTION_VAL_JAPAN_GL "Xapón"
#define OPTION_VAL_USA_GL NULL
#define OPTION_VAL_EUROPE_GL "Europa"
#define OPTION_VAL_DEFAULT_GL "Por defecto"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_GL "Idioma"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_GL "Cambia o idioma que usa a BIOS e os xogos que conteñan varios idiomas."
#define OPTION_VAL_JAPANESE_GL "Xaponés"
#define OPTION_VAL_ENGLISH_GL "Inglés"
#define OPTION_VAL_GERMAN_GL "Alemán"
#define OPTION_VAL_FRENCH_GL "Francés"
#define OPTION_VAL_SPANISH_GL "Español"
#define OPTION_VAL_ITALIAN_GL "Italiano"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_GL "BIOS HLE (reinicio necesario)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_GL "Forza o uso de emulación BIOS de alto nivel."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_GL "Activar DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_GL "Activa a emulación do DSP de audio (procesador de sinal dixital) do Dreamcast. Mellora a precisión do son xerado, pero aumenta os requisitos de rendemento."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_GL "Permitir botóns de servizo de Arcade"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_GL "Activa o botón SERVICE para xogos de arcade para entrar na configuración do armario."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_GL "Configura NAOMI Games Para Xogar de Balde"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_GL "Modifica a configuración de moedas do xogo para xogar libremente."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_GL "Emulación de adaptador de banda ancha"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_GL "Emular o adaptador Ethernet de banda ancha en lugar do módem. (Requírese reinicio)"
#define CORE_OPTION_NAME_UPNP_LABEL_GL "Activar UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_GL "Use UPnP para configurar automaticamente o seu enrutador de Internet para xogos en liña."
#define CORE_OPTION_NAME_DCNET_LABEL_GL "Usar DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_GL "Usar o servizo na nube DCNet para acceso a Internet de Dreamcast."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_GL "Resolución interna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_GL "Modificar a resolución do renderizado."
#define OPTION_VAL_320X240_GL "320 x 240 (a metade)"
#define OPTION_VAL_640X480_GL "640 x 480 (nativo)"
#define OPTION_VAL_800X600_GL NULL
#define OPTION_VAL_960X720_GL NULL
#define OPTION_VAL_1024X768_GL NULL
#define OPTION_VAL_1280X960_GL NULL
#define OPTION_VAL_1440X1080_GL NULL
#define OPTION_VAL_1600X1200_GL NULL
#define OPTION_VAL_1920X1440_GL NULL
#define OPTION_VAL_2560X1920_GL NULL
#define OPTION_VAL_2880X2160_GL NULL
#define OPTION_VAL_3200X2400_GL NULL
#define OPTION_VAL_3840X2880_GL NULL
#define OPTION_VAL_4480X3360_GL NULL
#define OPTION_VAL_5120X3840_GL NULL
#define OPTION_VAL_5760X4320_GL NULL
#define OPTION_VAL_6400X4800_GL NULL
#define OPTION_VAL_7040X5280_GL NULL
#define OPTION_VAL_7680X5760_GL NULL
#define OPTION_VAL_8320X6240_GL NULL
#define OPTION_VAL_8960X6720_GL NULL
#define OPTION_VAL_9600X7200_GL NULL
#define OPTION_VAL_10240X7680_GL NULL
#define OPTION_VAL_10880X8160_GL NULL
#define OPTION_VAL_11520X8640_GL NULL
#define OPTION_VAL_12160X9120_GL NULL
#define OPTION_VAL_12800X9600_GL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_GL "Tipo de cable"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_GL "O tipo de sinal de saída. 'TV (Composite)' é o máis soportado."
#define OPTION_VAL_VGA_GL NULL
#define OPTION_VAL_TV_RGB_GL NULL
#define OPTION_VAL_TV_COMPOSITE_GL "TV (composto)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_GL "Estándar de emisión"
#define OPTION_VAL_NTSC_GL NULL
#define OPTION_VAL_PAL_GL "PAL (Mundo)"
#define OPTION_VAL_PAL_N_GL "PAL-N (Arxentina, Paraguai, Uruguai)"
#define OPTION_VAL_PAL_M_GL "PAL-M (Brasil)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_GL "Orientación da Pantalla"
#define OPTION_VAL_HORIZONTAL_GL NULL
#define OPTION_VAL_VERTICAL_GL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_GL "Clasificación Alpha"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_GL "Seleccione como se ordenan os polígonos transparentes."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_GL "Por franxa (rápido, menos preciso)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_GL "Por triángulo (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_GL "Por píxel (preciso, pero máis lento)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_GL "Tamaño do búfer de píxeles de acumulación"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_GL "É posible que se precisen valores máis altos para que as resolucións máis altas saian correctamente."
#define OPTION_VAL_512MB_GL NULL
#define OPTION_VAL_1GB_GL NULL
#define OPTION_VAL_2GB_GL NULL
#define OPTION_VAL_4GB_GL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_GL "Máximo de capas transparentes"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_GL "Poden ser necesarios valores máis altos para escenas complexas."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_GL "Emulación completa de framebuffer"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_GL "Activa a emulación completa de framebuffer na VRAM. Isto é útil para xogos que len ou escriben directamente o framebuffer na VRAM. Cando está activada, a resolución interna é obrigada a 640 x 480 e o rendemento pode verse gravemente afectado."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_GL "Activa o buffer RTT (Render To Texture)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_GL "Copia texturas renderizadas de volta da GPU á VRAM. Esta opción normalmente está habilitada para xogos que o requiran. Cando se activa, a ampliación da renderización de texturas está desactivada e o rendemento pode verse afectado."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_GL "Mapeo mip"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_GL "Cando as texturas activadas usarán unha versión máis pequena de si mesmas cando aparezan máis lonxe, pode aumentar o rendemento e reducir o brillo."
#define CORE_OPTION_NAME_FOG_LABEL_GL "Efectos néboa"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_GL "Modificador de volume"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_GL "Unha función GPU de Dreamcast que normalmente usan os xogos para debuxar sombras de obxectos. Normalmente debería estar activado: o impacto no rendemento adoita ser mínimo ou insignificante."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_GL "Filtrado anisotrópico"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_GL "Mellora a calidade das texturas en superficies que están en ángulos de visión oblicuos con respecto á cámara. Os valores máis altos son máis esixentes na GPU. Os cambios nesta configuración só se aplican despois de reiniciar."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_GL "Filtrado de Texturas"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_GL "O modo de filtrado de texturas a usar. Isto pódese usar para forzar un determinado modo de filtrado de texturas en todas as texturas para obter unha aparencia máis nítida (ou máis suave) que a predeterminada. Os valores distintos do predeterminado poden causar varios problemas de renderización. Os cambios nesta configuración só se aplican despois de reiniciar."
#define OPTION_VAL_1_GL "Forza: Veciño máis próximo"
#define OPTION_VAL_2_GL "Forza lineal"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_GL "Intercambio de fotogramas con retardo"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_GL "Útil para evitar o parpadeo nas pantallas ou os vídeos con fallos. Non recomendado en plataformas lentas."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_GL "Detectar cambios na taxa de Fotogramas"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_GL "Notificar a interface cando cambie a velocidade de fotogramas interna (por exemplo, de 60 fps a 30 fps). Mellora o ritmo de cadros nos xogos que se executan a 30 ou 20 fps bloqueados, pero que debería desactivarse para xogos con velocidades de cadros desbloqueadas (inestables) (por exemplo, Ecco the Dolphin, Unreal Tournament). Nota: Non está dispoñible cando se activa \"Saltar marco automático\"."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_GL "Filtrado de postprocesado PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_GL "Postproceso da imaxe renderizada para simular efectos específicos da PowerVR2 GPU e os sinais analóxicos de vídeo."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_GL "Aumento das texturas (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_GL "Millora dos gráficos 2D pixel art debuxados a man. Só se debería usar con xogos pixelados en 2D."
#define OPTION_VAL_2_O26_GL NULL
#define OPTION_VAL_4_GL NULL
#define OPTION_VAL_6_GL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_GL "Aumento da escala das texturas. Tamaño filtrado"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_GL "Seleccione un valor de tamaño máximo para unha textura a escalar, se o tamaño da textura é superior ao valor seleccionado, non se ampliará."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_GL "Interpolación de profundidade nativa"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_GL "Axuda coa corrupción da textura e os problemas de profundidade nas GPU AMD. Tamén pode axudar ás GPU Intel nalgúns casos."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_GL "Corrixir Bleeding Edge de luxo"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_GL "Axuda co caso de sangramento da textura ao aumentar a escala. Desactivalo pode axudar se os píxeles se deforman ao aumentar a escala en xogos en 2D (MVC2, CVS, KOF, etc.)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_GL "Renderizado de fíos"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_GL "Fai funcionar a GPU e a CPU en fíos diferentes. Altamente recomendado."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_GL "Salto de fotograma automático"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_GL "Automáticamente salta fotogramas cando o emulador está correndo lento. Nota: Esta opción só se aplica cando 'Renderizado por fíos' está activado."
#define OPTION_VAL_SOME_GL NULL
#define OPTION_VAL_MORE_GL "Máximo"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_GL "Salto de fotogramas"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_GL "Especifica o número de fotogramas para saltar entre cada fotograma que é mostrado."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_GL "Trucos de pantalla ancha (necesario reiniciar)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_GL "Activa trucos que permiten que determinados xogos se mostren en formato de pantalla panorámica."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_GL "Corte de pantalla ancha"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_GL "Debuxa xeometría fóra da relación de aspecto normal de 4:3. Pode producir fallas gráficas nas áreas reveladas."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_GL "Carga rápida de GD-ROM (inexacta)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_GL "Acelera a carga de GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_GL NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_GL "Activa 32 MB de RAM Mod para Dreamcast. Pode afectar á compatibilidade"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_GL "CPU SH4 under/overclock"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_GL "Cambia o reloxo da CPU principal do SH4 dos 200 MHz predeterminados. O underclocking pode axudar a ralentizar as plataformas. O overclocking pode aumentar a velocidade de fotogramas nalgúns xogos. Use con precaución."
#define OPTION_VAL_100_GL NULL
#define OPTION_VAL_110_GL NULL
#define OPTION_VAL_120_GL NULL
#define OPTION_VAL_130_GL NULL
#define OPTION_VAL_140_GL NULL
#define OPTION_VAL_150_GL NULL
#define OPTION_VAL_160_GL NULL
#define OPTION_VAL_170_GL NULL
#define OPTION_VAL_180_GL NULL
#define OPTION_VAL_190_GL NULL
#define OPTION_VAL_200_GL NULL
#define OPTION_VAL_210_GL NULL
#define OPTION_VAL_220_GL NULL
#define OPTION_VAL_230_GL NULL
#define OPTION_VAL_240_GL NULL
#define OPTION_VAL_250_GL NULL
#define OPTION_VAL_260_GL NULL
#define OPTION_VAL_270_GL NULL
#define OPTION_VAL_280_GL NULL
#define OPTION_VAL_290_GL NULL
#define OPTION_VAL_300_GL NULL
#define OPTION_VAL_310_GL NULL
#define OPTION_VAL_320_GL NULL
#define OPTION_VAL_330_GL NULL
#define OPTION_VAL_340_GL NULL
#define OPTION_VAL_350_GL NULL
#define OPTION_VAL_360_GL NULL
#define OPTION_VAL_370_GL NULL
#define OPTION_VAL_380_GL NULL
#define OPTION_VAL_390_GL NULL
#define OPTION_VAL_400_GL NULL
#define OPTION_VAL_410_GL NULL
#define OPTION_VAL_420_GL NULL
#define OPTION_VAL_430_GL NULL
#define OPTION_VAL_440_GL NULL
#define OPTION_VAL_450_GL NULL
#define OPTION_VAL_460_GL NULL
#define OPTION_VAL_470_GL NULL
#define OPTION_VAL_480_GL NULL
#define OPTION_VAL_490_GL NULL
#define OPTION_VAL_500_GL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_GL "Carga as texturas personalizadas"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_GL "Carga texturas personalizadas situadas no cartafol \"system/dc/textures/<game-id>/\"."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_GL "Precargar texturas personalizadas"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_GL "Precarga as texturas personalizadas ao iniciar o xogo. Pode mellorar o rendemento, pero aumenta o uso de memoria."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_GL "Texturas de volcado"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_GL "Cada vez que o xogo use unha textura nova, gardarase como ficheiro .png no cartafol 'system/dc/texdump/<game-id>/'."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_GL "Exportar texturas substituídas"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_GL "Exporta sempre as texturas que xa foron substituídas por texturas personalizadas."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_GL "Zona morta panca analóxica"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_GL "Selecciona ata onde tes que empuxar o stick analóxico antes de que comece a procesarse."
#define OPTION_VAL_0_GL "0 %"
#define OPTION_VAL_5_GL "5 %"
#define OPTION_VAL_10_GL NULL
#define OPTION_VAL_15_GL NULL
#define OPTION_VAL_20_GL NULL
#define OPTION_VAL_25_GL NULL
#define OPTION_VAL_30_GL NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_GL "Activar zona morta"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_GL "Seleccione canto ten que premer o gatillo antes de comezar a procesarse."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_GL "Disparadores dixitais"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_GL "Cando estean activados, os disparadores actuarán como botóns normais, o que significa que se procesarán como totalmente presionados ou sen presionar nada, sen intermediarios."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_GL "Saídas dixitais de emisión"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_GL "Saídas dixitais de transmisión e estado de retroalimentación forzada no porto TCP 8000. Compatible coa opción MAME \"-output network\"."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_GL "Mostrar a configuración da arma lixeira"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_GL "Activa a configuración das opcións de visualización do punto de mira de pistola lixeira. NOTA: É posible que teña que cambiar o menú rápido para que esta configuración teña efecto."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_GL "Escala do tamaño da mira de pistola"
#define OPTION_VAL_50_GL NULL
#define OPTION_VAL_60_GL NULL
#define OPTION_VAL_70_GL NULL
#define OPTION_VAL_80_GL NULL
#define OPTION_VAL_90_GL NULL
#define OPTION_VAL_100_O47_GL NULL
#define OPTION_VAL_110_O47_GL NULL
#define OPTION_VAL_120_O47_GL NULL
#define OPTION_VAL_130_O47_GL NULL
#define OPTION_VAL_140_O47_GL NULL
#define OPTION_VAL_150_O47_GL NULL
#define OPTION_VAL_160_O47_GL NULL
#define OPTION_VAL_170_O47_GL NULL
#define OPTION_VAL_180_O47_GL NULL
#define OPTION_VAL_190_O47_GL NULL
#define OPTION_VAL_200_O47_GL NULL
#define OPTION_VAL_210_O47_GL NULL
#define OPTION_VAL_220_O47_GL NULL
#define OPTION_VAL_230_O47_GL NULL
#define OPTION_VAL_240_O47_GL NULL
#define OPTION_VAL_250_O47_GL NULL
#define OPTION_VAL_260_O47_GL NULL
#define OPTION_VAL_270_O47_GL NULL
#define OPTION_VAL_280_O47_GL NULL
#define OPTION_VAL_290_O47_GL NULL
#define OPTION_VAL_300_O47_GL NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_GL "Pantalla Gun Crosshair 1"
#define OPTION_VAL_WHITE_GL "Branco"
#define OPTION_VAL_RED_GL "Vermello"
#define OPTION_VAL_GREEN_GL "Verde"
#define OPTION_VAL_BLUE_GL "Azul"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_GL "Pantalla Gun Crosshair 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_GL "Pantalla Gun Crosshair 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_GL "Pantalla Gun Crosshair 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_GL "Dispositivo na rañura de expansión A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_GL "Seleccione o dispositivo conectado a rañura de expansión A1 (porto A rañura 1)."
#define OPTION_VAL_VMU_GL NULL
#define OPTION_VAL_PURUPURU_GL "Paquete de vibracións"
#define OPTION_VAL_DREAMPOTATO_GL NULL
#define OPTION_VAL_NONE_GL "Ningún"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_GL "Dispositivo na rañura de expansión A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_GL "Seleccione o dispositivo conectado a rañura de expansión A2 (porto A rañura 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_GL "Dispositivo na rañura de expansión B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_GL "Seleccione o dispositivo conectado a rañura de expansión B1 (porto B rañura 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_GL "Dispositivo na rañura de expansión B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_GL "Seleccione o dispositivo conectado á rañura de expansión B2 (porto B rañura 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_GL "Dispositivo na rañura de expansión C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_GL "Seleccione o dispositivo conectado a rañura de expansión C1 (porto C rañura 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_GL "Dispositivo na rañura de expansión C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_GL "Seleccione o dispositivo conectado a rañura de expansión C2 (porto C rañura 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_GL "Dispositivo na rañura de expansión D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_GL "Seleccione o dispositivo conectado a rañura de expansión D1 (porto D rañura 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_GL "Dispositivo na rañura de expansión D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_GL "Seleccione o dispositivo conectado a rañura de expansión D2 (porto D rañura 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_GL "Unidades/sistemas de memoria visual (VMU) por xogo"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_GL "VMUs por xogo"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_GL "Cando está desactivado, todos os xogos comparten até 8 arquivos de gardado VMU (A1/A2/B1/B2/C1/C2/D1/D2) situados na carpeta do sistema de RetroArch.\nA configuración 'VMU A1' crea un arquivo VMU 'A1' único na carpeta de gardado de RetroArch para cada xogo que se lanza.\nA configuración 'Todos os VMUs' crea até 8 arquivos VMU únicos (A1/A2/B1/B2/C1/C2/D1/D2) para cada xogo que se lanza."
#define OPTION_VAL_VMU_A1_GL NULL
#define OPTION_VAL_ALL_VMUS_GL "Todas as VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_GL "Sons de sistemas/unidades de memoria visual (VMU)"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_GL "Sons VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_GL "Cando está activado, reprodúcense os pitidos VMU."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_GL "Almacenamento de VMU de DreamPotato"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_GL "Activa o almacenamento dos datos da VMU no emulador DreamPotato conectado. Cando está activado, non se utiliza o ficheiro local de gardado da VMU."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_GL "Amosa as opcións de pantalla da Unidade/Sistema de Memoria visual (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_GL "Amosa as opcións de pantalla da VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_GL "Activa a configuración da visibilidade, tamaño, posición e cor da pantalla LCD VMU emulada. NOTA: É posible que teña que cambiar o menú rápido para que esta configuración teña efecto."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_GL "Pantalla 1 da VMU"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_GL "Posición da pantalla 1 VMU"
#define OPTION_VAL_UPPER_LEFT_GL "Superior esquerda"
#define OPTION_VAL_UPPER_RIGHT_GL "Superior dereita"
#define OPTION_VAL_LOWER_LEFT_GL "Inferior esquerda"
#define OPTION_VAL_LOWER_RIGHT_GL "Inferior dereita"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_GL "Tamaño da pantalla 1 VMU"
#define OPTION_VAL_1X_GL NULL
#define OPTION_VAL_3X_GL NULL
#define OPTION_VAL_5X_GL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_GL "Activar pixel en cor na pantalla 1 VMU"
#define OPTION_VAL_DEFAULT_ON_00_GL "Por defecto activado"
#define OPTION_VAL_DEFAULT_OFF_01_GL "Por defecto desactivado"
#define OPTION_VAL_BLACK_02_GL "Negro"
#define OPTION_VAL_LIGHT_BLUE_04_GL "Azul claro"
#define OPTION_VAL_CYAN_06_GL "Azul ceo"
#define OPTION_VAL_CYAN_BLUE_07_GL "Cián"
#define OPTION_VAL_LIGHT_GREEN_08_GL "Verde claro"
#define OPTION_VAL_CYAN_GREEN_09_GL "Cián Verde"
#define OPTION_VAL_LIGHT_CYAN_10_GL "Cián claro"
#define OPTION_VAL_PURPLE_12_GL "Malva"
#define OPTION_VAL_LIGHT_PURPLE_13_GL "Malva claro"
#define OPTION_VAL_YELLOW_14_GL "Amarelo"
#define OPTION_VAL_GRAY_15_GL "Gris"
#define OPTION_VAL_LIGHT_PURPLE_2_16_GL "Malva claro (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_GL "Verde claro (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_GL "Verde claro (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_GL "Cián claro (2)"
#define OPTION_VAL_LIGHT_RED_2_20_GL "Vermello claro (2)"
#define OPTION_VAL_MAGENTA_21_GL "Maxenta"
#define OPTION_VAL_LIGHT_PURPLE_3_22_GL "Malva claro (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_GL "Laranxa claro"
#define OPTION_VAL_ORANGE_24_GL "Laranxa"
#define OPTION_VAL_LIGHT_PURPLE_4_25_GL "Malva claro (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_GL "Amarelo claro"
#define OPTION_VAL_LIGHT_YELLOW_2_27_GL "Amarelo claro (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_GL "Desactivar píxel en cor na pantalla 1 VMU"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_GL "Opacidade da pantalla 1 VMU"
#define OPTION_VAL_40_GL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_GL "Pantalla 2 da VMU"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_GL "Posición da pantalla 2 VMU"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_GL "Tamaño da pantalla 2 VMU"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_GL "Activar pixel en cor na pantalla 2 VMU"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_GL "Desactivar píxel en cor na pantalla 2 VMU"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_GL "Opacidade da pantalla 2 VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_GL "Pantalla 3 da VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_GL "Posición da pantalla 3 VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_GL "Tamaño da pantalla 3 VMU"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_GL "Activar pixel en cor na pantalla 3 VMU"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_GL "Desactivar píxel en cor na pantalla 3 VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_GL "Opacidade da pantalla 3 VMU"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_GL "Pantalla 4 da VMU"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_GL "Posición da pantalla 4 VMU"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_GL "Tamaño da pantalla 4 VMU"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_GL "Activar pixel en cor na pantalla 4 VMU"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_GL "Desactivar píxel en cor na pantalla 4 VMU"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_GL "Opacidade da pantalla 4 VMU"

struct retro_core_option_v2_category option_cats_gl[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_GL,
      CATEGORY_SYSTEM_INFO_0_GL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_GL,
      CATEGORY_VIDEO_INFO_0_GL
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_GL,
      CATEGORY_PERFORMANCE_INFO_0_GL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_GL,
      CATEGORY_HACKS_INFO_0_GL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_GL,
      CATEGORY_INPUT_INFO_0_GL
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_GL,
      CATEGORY_EXPANSIONS_INFO_0_GL
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_GL,
      CATEGORY_VMU_INFO_0_GL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_gl[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_GL,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_GL },
         { "USA",     OPTION_VAL_USA_GL },
         { "Europe",  OPTION_VAL_EUROPE_GL },
         { "Default", OPTION_VAL_DEFAULT_GL },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_GL,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_GL },
         { "English",  OPTION_VAL_ENGLISH_GL },
         { "German",   OPTION_VAL_GERMAN_GL },
         { "French",   OPTION_VAL_FRENCH_GL },
         { "Spanish",  OPTION_VAL_SPANISH_GL },
         { "Italian",  OPTION_VAL_ITALIAN_GL },
         { "Default",  OPTION_VAL_DEFAULT_GL },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_GL,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_GL },
         { "640x480",    OPTION_VAL_640X480_GL },
         { "800x600",    OPTION_VAL_800X600_GL },
         { "960x720",    OPTION_VAL_960X720_GL },
         { "1024x768",   OPTION_VAL_1024X768_GL },
         { "1280x960",   OPTION_VAL_1280X960_GL },
         { "1440x1080",  OPTION_VAL_1440X1080_GL },
         { "1600x1200",  OPTION_VAL_1600X1200_GL },
         { "1920x1440",  OPTION_VAL_1920X1440_GL },
         { "2560x1920",  OPTION_VAL_2560X1920_GL },
         { "2880x2160",  OPTION_VAL_2880X2160_GL },
         { "3200x2400",  OPTION_VAL_3200X2400_GL },
         { "3840x2880",  OPTION_VAL_3840X2880_GL },
         { "4480x3360",  OPTION_VAL_4480X3360_GL },
         { "5120x3840",  OPTION_VAL_5120X3840_GL },
         { "5760x4320",  OPTION_VAL_5760X4320_GL },
         { "6400x4800",  OPTION_VAL_6400X4800_GL },
         { "7040x5280",  OPTION_VAL_7040X5280_GL },
         { "7680x5760",  OPTION_VAL_7680X5760_GL },
         { "8320x6240",  OPTION_VAL_8320X6240_GL },
         { "8960x6720",  OPTION_VAL_8960X6720_GL },
         { "9600x7200",  OPTION_VAL_9600X7200_GL },
         { "10240x7680", OPTION_VAL_10240X7680_GL },
         { "10880x8160", OPTION_VAL_10880X8160_GL },
         { "11520x8640", OPTION_VAL_11520X8640_GL },
         { "12160x9120", OPTION_VAL_12160X9120_GL },
         { "12800x9600", OPTION_VAL_12800X9600_GL },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_GL,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_GL },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_GL },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_GL },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_GL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_GL },
         { "PAL",     OPTION_VAL_PAL_GL },
         { "PAL_N",   OPTION_VAL_PAL_N_GL },
         { "PAL_M",   OPTION_VAL_PAL_M_GL },
         { "Default", OPTION_VAL_DEFAULT_GL },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_GL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_GL },
         { "vertical",   OPTION_VAL_VERTICAL_GL },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_GL,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_GL },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_GL },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_GL },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_GL,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_GL },
         { "1GB",   OPTION_VAL_1GB_GL },
         { "2GB",   OPTION_VAL_2GB_GL },
         { "4GB",   OPTION_VAL_4GB_GL },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_GL,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_GL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_GL,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_GL,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_GL },
         { "1",  OPTION_VAL_1_GL },
         { "2",  OPTION_VAL_2_GL },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_GL,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_GL },
         { "4",  OPTION_VAL_4_GL },
         { "6",  OPTION_VAL_6_GL },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_GL,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_GL,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_GL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_GL,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_GL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_GL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_GL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_GL },
         { "more", OPTION_VAL_MORE_GL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_GL,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_GL },
         { "110", OPTION_VAL_110_GL },
         { "120", OPTION_VAL_120_GL },
         { "130", OPTION_VAL_130_GL },
         { "140", OPTION_VAL_140_GL },
         { "150", OPTION_VAL_150_GL },
         { "160", OPTION_VAL_160_GL },
         { "170", OPTION_VAL_170_GL },
         { "180", OPTION_VAL_180_GL },
         { "190", OPTION_VAL_190_GL },
         { "200", OPTION_VAL_200_GL },
         { "210", OPTION_VAL_210_GL },
         { "220", OPTION_VAL_220_GL },
         { "230", OPTION_VAL_230_GL },
         { "240", OPTION_VAL_240_GL },
         { "250", OPTION_VAL_250_GL },
         { "260", OPTION_VAL_260_GL },
         { "270", OPTION_VAL_270_GL },
         { "280", OPTION_VAL_280_GL },
         { "290", OPTION_VAL_290_GL },
         { "300", OPTION_VAL_300_GL },
         { "310", OPTION_VAL_310_GL },
         { "320", OPTION_VAL_320_GL },
         { "330", OPTION_VAL_330_GL },
         { "340", OPTION_VAL_340_GL },
         { "350", OPTION_VAL_350_GL },
         { "360", OPTION_VAL_360_GL },
         { "370", OPTION_VAL_370_GL },
         { "380", OPTION_VAL_380_GL },
         { "390", OPTION_VAL_390_GL },
         { "400", OPTION_VAL_400_GL },
         { "410", OPTION_VAL_410_GL },
         { "420", OPTION_VAL_420_GL },
         { "430", OPTION_VAL_430_GL },
         { "440", OPTION_VAL_440_GL },
         { "450", OPTION_VAL_450_GL },
         { "460", OPTION_VAL_460_GL },
         { "470", OPTION_VAL_470_GL },
         { "480", OPTION_VAL_480_GL },
         { "490", OPTION_VAL_490_GL },
         { "500", OPTION_VAL_500_GL },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_GL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_GL },
         { "5%",  OPTION_VAL_5_GL },
         { "10%", OPTION_VAL_10_GL },
         { "15%", OPTION_VAL_15_GL },
         { "20%", OPTION_VAL_20_GL },
         { "25%", OPTION_VAL_25_GL },
         { "30%", OPTION_VAL_30_GL },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_GL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_GL },
         { "5%",  OPTION_VAL_5_GL },
         { "10%", OPTION_VAL_10_GL },
         { "15%", OPTION_VAL_15_GL },
         { "20%", OPTION_VAL_20_GL },
         { "25%", OPTION_VAL_25_GL },
         { "30%", OPTION_VAL_30_GL },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_GL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_GL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_GL,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_GL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_GL },
         { "60%",  OPTION_VAL_60_GL },
         { "70%",  OPTION_VAL_70_GL },
         { "80%",  OPTION_VAL_80_GL },
         { "90%",  OPTION_VAL_90_GL },
         { "100%", OPTION_VAL_100_O47_GL },
         { "110%", OPTION_VAL_110_O47_GL },
         { "120%", OPTION_VAL_120_O47_GL },
         { "130%", OPTION_VAL_130_O47_GL },
         { "140%", OPTION_VAL_140_O47_GL },
         { "150%", OPTION_VAL_150_O47_GL },
         { "160%", OPTION_VAL_160_O47_GL },
         { "170%", OPTION_VAL_170_O47_GL },
         { "180%", OPTION_VAL_180_O47_GL },
         { "190%", OPTION_VAL_190_O47_GL },
         { "200%", OPTION_VAL_200_O47_GL },
         { "210%", OPTION_VAL_210_O47_GL },
         { "220%", OPTION_VAL_220_O47_GL },
         { "230%", OPTION_VAL_230_O47_GL },
         { "240%", OPTION_VAL_240_O47_GL },
         { "250%", OPTION_VAL_250_O47_GL },
         { "260%", OPTION_VAL_260_O47_GL },
         { "270%", OPTION_VAL_270_O47_GL },
         { "280%", OPTION_VAL_280_O47_GL },
         { "290%", OPTION_VAL_290_O47_GL },
         { "300%", OPTION_VAL_300_O47_GL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_GL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_GL },
         { "Red",      OPTION_VAL_RED_GL },
         { "Green",    OPTION_VAL_GREEN_GL },
         { "Blue",     OPTION_VAL_BLUE_GL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_GL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_GL },
         { "Red",      OPTION_VAL_RED_GL },
         { "Green",    OPTION_VAL_GREEN_GL },
         { "Blue",     OPTION_VAL_BLUE_GL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_GL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_GL },
         { "Red",      OPTION_VAL_RED_GL },
         { "Green",    OPTION_VAL_GREEN_GL },
         { "Blue",     OPTION_VAL_BLUE_GL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_GL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_GL },
         { "Red",      OPTION_VAL_RED_GL },
         { "Green",    OPTION_VAL_GREEN_GL },
         { "Blue",     OPTION_VAL_BLUE_GL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_GL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GL },
         { "Purupuru", OPTION_VAL_PURUPURU_GL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_GL },
         { "None",     OPTION_VAL_NONE_GL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_GL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GL },
         { "Purupuru", OPTION_VAL_PURUPURU_GL },
         { "None",     OPTION_VAL_NONE_GL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_GL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GL },
         { "Purupuru", OPTION_VAL_PURUPURU_GL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_GL },
         { "None",     OPTION_VAL_NONE_GL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_GL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GL },
         { "Purupuru", OPTION_VAL_PURUPURU_GL },
         { "None",     OPTION_VAL_NONE_GL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_GL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GL },
         { "Purupuru", OPTION_VAL_PURUPURU_GL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_GL },
         { "None",     OPTION_VAL_NONE_GL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_GL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GL },
         { "Purupuru", OPTION_VAL_PURUPURU_GL },
         { "None",     OPTION_VAL_NONE_GL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_GL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GL },
         { "Purupuru", OPTION_VAL_PURUPURU_GL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_GL },
         { "None",     OPTION_VAL_NONE_GL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_GL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_GL },
         { "Purupuru", OPTION_VAL_PURUPURU_GL },
         { "None",     OPTION_VAL_NONE_GL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_GL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_GL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_GL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_GL },
         { "All VMUs", OPTION_VAL_ALL_VMUS_GL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_GL,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_GL,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_GL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_GL,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_GL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_GL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_GL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_GL,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_GL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_GL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_GL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_GL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_GL },
         { "2x", OPTION_VAL_2_O26_GL },
         { "3x", OPTION_VAL_3X_GL },
         { "4x", OPTION_VAL_4_GL },
         { "5x", OPTION_VAL_5X_GL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_GL },
         { "20%",  OPTION_VAL_20_GL },
         { "30%",  OPTION_VAL_30_GL },
         { "40%",  OPTION_VAL_40_GL },
         { "50%",  OPTION_VAL_50_GL },
         { "60%",  OPTION_VAL_60_GL },
         { "70%",  OPTION_VAL_70_GL },
         { "80%",  OPTION_VAL_80_GL },
         { "90%",  OPTION_VAL_90_GL },
         { "100%", OPTION_VAL_100_O47_GL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_GL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_GL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_GL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_GL },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_GL },
         { "2x", OPTION_VAL_2_O26_GL },
         { "3x", OPTION_VAL_3X_GL },
         { "4x", OPTION_VAL_4_GL },
         { "5x", OPTION_VAL_5X_GL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_GL },
         { "20%",  OPTION_VAL_20_GL },
         { "30%",  OPTION_VAL_30_GL },
         { "40%",  OPTION_VAL_40_GL },
         { "50%",  OPTION_VAL_50_GL },
         { "60%",  OPTION_VAL_60_GL },
         { "70%",  OPTION_VAL_70_GL },
         { "80%",  OPTION_VAL_80_GL },
         { "90%",  OPTION_VAL_90_GL },
         { "100%", OPTION_VAL_100_O47_GL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_GL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_GL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_GL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_GL },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_GL },
         { "2x", OPTION_VAL_2_O26_GL },
         { "3x", OPTION_VAL_3X_GL },
         { "4x", OPTION_VAL_4_GL },
         { "5x", OPTION_VAL_5X_GL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_GL },
         { "20%",  OPTION_VAL_20_GL },
         { "30%",  OPTION_VAL_30_GL },
         { "40%",  OPTION_VAL_40_GL },
         { "50%",  OPTION_VAL_50_GL },
         { "60%",  OPTION_VAL_60_GL },
         { "70%",  OPTION_VAL_70_GL },
         { "80%",  OPTION_VAL_80_GL },
         { "90%",  OPTION_VAL_90_GL },
         { "100%", OPTION_VAL_100_O47_GL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_GL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_GL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_GL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_GL },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_GL },
         { "2x", OPTION_VAL_2_O26_GL },
         { "3x", OPTION_VAL_3X_GL },
         { "4x", OPTION_VAL_4_GL },
         { "5x", OPTION_VAL_5X_GL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_GL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_GL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_GL },
         { "BLUE 03",           OPTION_VAL_BLUE_GL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_GL },
         { "GREEN 05",          OPTION_VAL_GREEN_GL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_GL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_GL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_GL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_GL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_GL },
         { "RED 11",            OPTION_VAL_RED_GL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_GL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_GL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_GL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_GL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_GL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_GL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_GL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_GL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_GL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_GL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_GL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_GL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_GL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_GL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_GL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_GL },
         { "WHITE 28",          OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_GL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_GL },
         { "20%",  OPTION_VAL_20_GL },
         { "30%",  OPTION_VAL_30_GL },
         { "40%",  OPTION_VAL_40_GL },
         { "50%",  OPTION_VAL_50_GL },
         { "60%",  OPTION_VAL_60_GL },
         { "70%",  OPTION_VAL_70_GL },
         { "80%",  OPTION_VAL_80_GL },
         { "90%",  OPTION_VAL_90_GL },
         { "100%", OPTION_VAL_100_O47_GL },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_gl = {
   option_cats_gl,
   option_defs_gl
};

/* RETRO_LANGUAGE_HE */

#define CATEGORY_SYSTEM_LABEL_HE NULL
#define CATEGORY_SYSTEM_INFO_0_HE NULL
#define CATEGORY_VIDEO_LABEL_HE "וידאו"
#define CATEGORY_VIDEO_INFO_0_HE NULL
#define CATEGORY_PERFORMANCE_LABEL_HE NULL
#define CATEGORY_PERFORMANCE_INFO_0_HE NULL
#define CATEGORY_HACKS_LABEL_HE NULL
#define CATEGORY_HACKS_INFO_0_HE NULL
#define CATEGORY_INPUT_LABEL_HE "קלט"
#define CATEGORY_INPUT_INFO_0_HE NULL
#define CATEGORY_EXPANSIONS_LABEL_HE NULL
#define CATEGORY_EXPANSIONS_INFO_0_HE NULL
#define CATEGORY_VMU_LABEL_HE NULL
#define CATEGORY_VMU_INFO_0_HE NULL
#define CORE_OPTION_NAME_REGION_LABEL_HE NULL
#define OPTION_VAL_JAPAN_HE NULL
#define OPTION_VAL_USA_HE NULL
#define OPTION_VAL_EUROPE_HE NULL
#define OPTION_VAL_DEFAULT_HE NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_HE NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_HE NULL
#define OPTION_VAL_JAPANESE_HE NULL
#define OPTION_VAL_ENGLISH_HE NULL
#define OPTION_VAL_GERMAN_HE NULL
#define OPTION_VAL_FRENCH_HE NULL
#define OPTION_VAL_SPANISH_HE NULL
#define OPTION_VAL_ITALIAN_HE NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_HE NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_HE NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_HE NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_HE NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_HE NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_HE NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_HE NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_HE NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_HE NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_HE NULL
#define CORE_OPTION_NAME_UPNP_LABEL_HE NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_HE NULL
#define CORE_OPTION_NAME_DCNET_LABEL_HE NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_HE NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_HE NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_HE NULL
#define OPTION_VAL_320X240_HE NULL
#define OPTION_VAL_640X480_HE NULL
#define OPTION_VAL_800X600_HE NULL
#define OPTION_VAL_960X720_HE NULL
#define OPTION_VAL_1024X768_HE NULL
#define OPTION_VAL_1280X960_HE NULL
#define OPTION_VAL_1440X1080_HE NULL
#define OPTION_VAL_1600X1200_HE NULL
#define OPTION_VAL_1920X1440_HE NULL
#define OPTION_VAL_2560X1920_HE NULL
#define OPTION_VAL_2880X2160_HE NULL
#define OPTION_VAL_3200X2400_HE NULL
#define OPTION_VAL_3840X2880_HE NULL
#define OPTION_VAL_4480X3360_HE NULL
#define OPTION_VAL_5120X3840_HE NULL
#define OPTION_VAL_5760X4320_HE NULL
#define OPTION_VAL_6400X4800_HE NULL
#define OPTION_VAL_7040X5280_HE NULL
#define OPTION_VAL_7680X5760_HE NULL
#define OPTION_VAL_8320X6240_HE NULL
#define OPTION_VAL_8960X6720_HE NULL
#define OPTION_VAL_9600X7200_HE NULL
#define OPTION_VAL_10240X7680_HE NULL
#define OPTION_VAL_10880X8160_HE NULL
#define OPTION_VAL_11520X8640_HE NULL
#define OPTION_VAL_12160X9120_HE NULL
#define OPTION_VAL_12800X9600_HE NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_HE NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_HE NULL
#define OPTION_VAL_VGA_HE NULL
#define OPTION_VAL_TV_RGB_HE NULL
#define OPTION_VAL_TV_COMPOSITE_HE NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_HE NULL
#define OPTION_VAL_NTSC_HE NULL
#define OPTION_VAL_PAL_HE NULL
#define OPTION_VAL_PAL_N_HE NULL
#define OPTION_VAL_PAL_M_HE NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_HE NULL
#define OPTION_VAL_HORIZONTAL_HE NULL
#define OPTION_VAL_VERTICAL_HE NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_HE NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_HE NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_HE NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_HE NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_HE NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_HE NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_HE NULL
#define OPTION_VAL_512MB_HE NULL
#define OPTION_VAL_1GB_HE NULL
#define OPTION_VAL_2GB_HE NULL
#define OPTION_VAL_4GB_HE NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_HE NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_HE NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_HE NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_HE NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_HE NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_HE NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_HE NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_HE NULL
#define CORE_OPTION_NAME_FOG_LABEL_HE NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_HE NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_HE NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_HE NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_HE NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_HE NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_HE NULL
#define OPTION_VAL_1_HE NULL
#define OPTION_VAL_2_HE NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_HE NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_HE NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_HE NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_HE NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_HE NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_HE NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_HE NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_HE NULL
#define OPTION_VAL_2_O26_HE NULL
#define OPTION_VAL_4_HE NULL
#define OPTION_VAL_6_HE NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_HE NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_HE NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_HE NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_HE NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_HE NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_HE NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_HE NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_HE NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_HE NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_HE NULL
#define OPTION_VAL_SOME_HE NULL
#define OPTION_VAL_MORE_HE NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_HE NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_HE NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_HE NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_HE NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_HE NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_HE NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_HE NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_HE NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_HE NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_HE NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_HE NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_HE NULL
#define OPTION_VAL_100_HE NULL
#define OPTION_VAL_110_HE NULL
#define OPTION_VAL_120_HE NULL
#define OPTION_VAL_130_HE NULL
#define OPTION_VAL_140_HE NULL
#define OPTION_VAL_150_HE NULL
#define OPTION_VAL_160_HE NULL
#define OPTION_VAL_170_HE NULL
#define OPTION_VAL_180_HE NULL
#define OPTION_VAL_190_HE NULL
#define OPTION_VAL_200_HE NULL
#define OPTION_VAL_210_HE NULL
#define OPTION_VAL_220_HE NULL
#define OPTION_VAL_230_HE NULL
#define OPTION_VAL_240_HE NULL
#define OPTION_VAL_250_HE NULL
#define OPTION_VAL_260_HE NULL
#define OPTION_VAL_270_HE NULL
#define OPTION_VAL_280_HE NULL
#define OPTION_VAL_290_HE NULL
#define OPTION_VAL_300_HE NULL
#define OPTION_VAL_310_HE NULL
#define OPTION_VAL_320_HE NULL
#define OPTION_VAL_330_HE NULL
#define OPTION_VAL_340_HE NULL
#define OPTION_VAL_350_HE NULL
#define OPTION_VAL_360_HE NULL
#define OPTION_VAL_370_HE NULL
#define OPTION_VAL_380_HE NULL
#define OPTION_VAL_390_HE NULL
#define OPTION_VAL_400_HE NULL
#define OPTION_VAL_410_HE NULL
#define OPTION_VAL_420_HE NULL
#define OPTION_VAL_430_HE NULL
#define OPTION_VAL_440_HE NULL
#define OPTION_VAL_450_HE NULL
#define OPTION_VAL_460_HE NULL
#define OPTION_VAL_470_HE NULL
#define OPTION_VAL_480_HE NULL
#define OPTION_VAL_490_HE NULL
#define OPTION_VAL_500_HE NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_HE NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_HE NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_HE NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_HE NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_HE NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_HE NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_HE NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_HE NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_HE NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_HE NULL
#define OPTION_VAL_0_HE NULL
#define OPTION_VAL_5_HE NULL
#define OPTION_VAL_10_HE NULL
#define OPTION_VAL_15_HE NULL
#define OPTION_VAL_20_HE NULL
#define OPTION_VAL_25_HE NULL
#define OPTION_VAL_30_HE NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_HE NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_HE NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_HE NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_HE NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_HE NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_HE NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_HE NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_HE NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_HE NULL
#define OPTION_VAL_50_HE NULL
#define OPTION_VAL_60_HE NULL
#define OPTION_VAL_70_HE NULL
#define OPTION_VAL_80_HE NULL
#define OPTION_VAL_90_HE NULL
#define OPTION_VAL_100_O47_HE NULL
#define OPTION_VAL_110_O47_HE NULL
#define OPTION_VAL_120_O47_HE NULL
#define OPTION_VAL_130_O47_HE NULL
#define OPTION_VAL_140_O47_HE NULL
#define OPTION_VAL_150_O47_HE NULL
#define OPTION_VAL_160_O47_HE NULL
#define OPTION_VAL_170_O47_HE NULL
#define OPTION_VAL_180_O47_HE NULL
#define OPTION_VAL_190_O47_HE NULL
#define OPTION_VAL_200_O47_HE NULL
#define OPTION_VAL_210_O47_HE NULL
#define OPTION_VAL_220_O47_HE NULL
#define OPTION_VAL_230_O47_HE NULL
#define OPTION_VAL_240_O47_HE NULL
#define OPTION_VAL_250_O47_HE NULL
#define OPTION_VAL_260_O47_HE NULL
#define OPTION_VAL_270_O47_HE NULL
#define OPTION_VAL_280_O47_HE NULL
#define OPTION_VAL_290_O47_HE NULL
#define OPTION_VAL_300_O47_HE NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_HE NULL
#define OPTION_VAL_WHITE_HE NULL
#define OPTION_VAL_RED_HE NULL
#define OPTION_VAL_GREEN_HE NULL
#define OPTION_VAL_BLUE_HE NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_HE NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_HE NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_HE NULL
#define OPTION_VAL_VMU_HE NULL
#define OPTION_VAL_PURUPURU_HE NULL
#define OPTION_VAL_DREAMPOTATO_HE NULL
#define OPTION_VAL_NONE_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_HE NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_HE NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_HE NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_HE NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_HE NULL
#define OPTION_VAL_VMU_A1_HE NULL
#define OPTION_VAL_ALL_VMUS_HE NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_HE NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_HE NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_HE NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_HE NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_HE NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_HE NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_HE NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_HE NULL
#define OPTION_VAL_UPPER_LEFT_HE NULL
#define OPTION_VAL_UPPER_RIGHT_HE NULL
#define OPTION_VAL_LOWER_LEFT_HE NULL
#define OPTION_VAL_LOWER_RIGHT_HE NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_HE NULL
#define OPTION_VAL_1X_HE NULL
#define OPTION_VAL_3X_HE NULL
#define OPTION_VAL_5X_HE NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_HE NULL
#define OPTION_VAL_DEFAULT_ON_00_HE NULL
#define OPTION_VAL_DEFAULT_OFF_01_HE NULL
#define OPTION_VAL_BLACK_02_HE NULL
#define OPTION_VAL_LIGHT_BLUE_04_HE NULL
#define OPTION_VAL_CYAN_06_HE NULL
#define OPTION_VAL_CYAN_BLUE_07_HE NULL
#define OPTION_VAL_LIGHT_GREEN_08_HE NULL
#define OPTION_VAL_CYAN_GREEN_09_HE NULL
#define OPTION_VAL_LIGHT_CYAN_10_HE NULL
#define OPTION_VAL_PURPLE_12_HE NULL
#define OPTION_VAL_LIGHT_PURPLE_13_HE NULL
#define OPTION_VAL_YELLOW_14_HE NULL
#define OPTION_VAL_GRAY_15_HE NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_HE NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_HE NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_HE NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_HE NULL
#define OPTION_VAL_LIGHT_RED_2_20_HE NULL
#define OPTION_VAL_MAGENTA_21_HE NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_HE NULL
#define OPTION_VAL_LIGHT_ORANGE_23_HE NULL
#define OPTION_VAL_ORANGE_24_HE NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_HE NULL
#define OPTION_VAL_LIGHT_YELLOW_26_HE NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_HE NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_HE NULL
#define OPTION_VAL_40_HE NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_HE NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_HE NULL

struct retro_core_option_v2_category option_cats_he[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_HE,
      CATEGORY_SYSTEM_INFO_0_HE
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_HE,
      CATEGORY_VIDEO_INFO_0_HE
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_HE,
      CATEGORY_PERFORMANCE_INFO_0_HE
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_HE,
      CATEGORY_HACKS_INFO_0_HE
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_HE,
      CATEGORY_INPUT_INFO_0_HE
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_HE,
      CATEGORY_EXPANSIONS_INFO_0_HE
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_HE,
      CATEGORY_VMU_INFO_0_HE
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_he[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_HE,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_HE },
         { "USA",     OPTION_VAL_USA_HE },
         { "Europe",  OPTION_VAL_EUROPE_HE },
         { "Default", OPTION_VAL_DEFAULT_HE },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_HE,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_HE },
         { "English",  OPTION_VAL_ENGLISH_HE },
         { "German",   OPTION_VAL_GERMAN_HE },
         { "French",   OPTION_VAL_FRENCH_HE },
         { "Spanish",  OPTION_VAL_SPANISH_HE },
         { "Italian",  OPTION_VAL_ITALIAN_HE },
         { "Default",  OPTION_VAL_DEFAULT_HE },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_HE,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_HE },
         { "640x480",    OPTION_VAL_640X480_HE },
         { "800x600",    OPTION_VAL_800X600_HE },
         { "960x720",    OPTION_VAL_960X720_HE },
         { "1024x768",   OPTION_VAL_1024X768_HE },
         { "1280x960",   OPTION_VAL_1280X960_HE },
         { "1440x1080",  OPTION_VAL_1440X1080_HE },
         { "1600x1200",  OPTION_VAL_1600X1200_HE },
         { "1920x1440",  OPTION_VAL_1920X1440_HE },
         { "2560x1920",  OPTION_VAL_2560X1920_HE },
         { "2880x2160",  OPTION_VAL_2880X2160_HE },
         { "3200x2400",  OPTION_VAL_3200X2400_HE },
         { "3840x2880",  OPTION_VAL_3840X2880_HE },
         { "4480x3360",  OPTION_VAL_4480X3360_HE },
         { "5120x3840",  OPTION_VAL_5120X3840_HE },
         { "5760x4320",  OPTION_VAL_5760X4320_HE },
         { "6400x4800",  OPTION_VAL_6400X4800_HE },
         { "7040x5280",  OPTION_VAL_7040X5280_HE },
         { "7680x5760",  OPTION_VAL_7680X5760_HE },
         { "8320x6240",  OPTION_VAL_8320X6240_HE },
         { "8960x6720",  OPTION_VAL_8960X6720_HE },
         { "9600x7200",  OPTION_VAL_9600X7200_HE },
         { "10240x7680", OPTION_VAL_10240X7680_HE },
         { "10880x8160", OPTION_VAL_10880X8160_HE },
         { "11520x8640", OPTION_VAL_11520X8640_HE },
         { "12160x9120", OPTION_VAL_12160X9120_HE },
         { "12800x9600", OPTION_VAL_12800X9600_HE },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_HE,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_HE },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_HE },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_HE },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_HE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_HE },
         { "PAL",     OPTION_VAL_PAL_HE },
         { "PAL_N",   OPTION_VAL_PAL_N_HE },
         { "PAL_M",   OPTION_VAL_PAL_M_HE },
         { "Default", OPTION_VAL_DEFAULT_HE },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_HE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_HE },
         { "vertical",   OPTION_VAL_VERTICAL_HE },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_HE,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_HE },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_HE },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_HE },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_HE,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_HE },
         { "1GB",   OPTION_VAL_1GB_HE },
         { "2GB",   OPTION_VAL_2GB_HE },
         { "4GB",   OPTION_VAL_4GB_HE },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_HE,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_HE,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_HE,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_HE,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_HE },
         { "1",  OPTION_VAL_1_HE },
         { "2",  OPTION_VAL_2_HE },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_HE,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_HE },
         { "4",  OPTION_VAL_4_HE },
         { "6",  OPTION_VAL_6_HE },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_HE,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_HE,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_HE,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_HE,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_HE,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_HE,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_HE,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_HE },
         { "more", OPTION_VAL_MORE_HE },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_HE,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_HE },
         { "110", OPTION_VAL_110_HE },
         { "120", OPTION_VAL_120_HE },
         { "130", OPTION_VAL_130_HE },
         { "140", OPTION_VAL_140_HE },
         { "150", OPTION_VAL_150_HE },
         { "160", OPTION_VAL_160_HE },
         { "170", OPTION_VAL_170_HE },
         { "180", OPTION_VAL_180_HE },
         { "190", OPTION_VAL_190_HE },
         { "200", OPTION_VAL_200_HE },
         { "210", OPTION_VAL_210_HE },
         { "220", OPTION_VAL_220_HE },
         { "230", OPTION_VAL_230_HE },
         { "240", OPTION_VAL_240_HE },
         { "250", OPTION_VAL_250_HE },
         { "260", OPTION_VAL_260_HE },
         { "270", OPTION_VAL_270_HE },
         { "280", OPTION_VAL_280_HE },
         { "290", OPTION_VAL_290_HE },
         { "300", OPTION_VAL_300_HE },
         { "310", OPTION_VAL_310_HE },
         { "320", OPTION_VAL_320_HE },
         { "330", OPTION_VAL_330_HE },
         { "340", OPTION_VAL_340_HE },
         { "350", OPTION_VAL_350_HE },
         { "360", OPTION_VAL_360_HE },
         { "370", OPTION_VAL_370_HE },
         { "380", OPTION_VAL_380_HE },
         { "390", OPTION_VAL_390_HE },
         { "400", OPTION_VAL_400_HE },
         { "410", OPTION_VAL_410_HE },
         { "420", OPTION_VAL_420_HE },
         { "430", OPTION_VAL_430_HE },
         { "440", OPTION_VAL_440_HE },
         { "450", OPTION_VAL_450_HE },
         { "460", OPTION_VAL_460_HE },
         { "470", OPTION_VAL_470_HE },
         { "480", OPTION_VAL_480_HE },
         { "490", OPTION_VAL_490_HE },
         { "500", OPTION_VAL_500_HE },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_HE,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_HE },
         { "5%",  OPTION_VAL_5_HE },
         { "10%", OPTION_VAL_10_HE },
         { "15%", OPTION_VAL_15_HE },
         { "20%", OPTION_VAL_20_HE },
         { "25%", OPTION_VAL_25_HE },
         { "30%", OPTION_VAL_30_HE },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_HE,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_HE },
         { "5%",  OPTION_VAL_5_HE },
         { "10%", OPTION_VAL_10_HE },
         { "15%", OPTION_VAL_15_HE },
         { "20%", OPTION_VAL_20_HE },
         { "25%", OPTION_VAL_25_HE },
         { "30%", OPTION_VAL_30_HE },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_HE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_HE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_HE,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_HE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_HE },
         { "60%",  OPTION_VAL_60_HE },
         { "70%",  OPTION_VAL_70_HE },
         { "80%",  OPTION_VAL_80_HE },
         { "90%",  OPTION_VAL_90_HE },
         { "100%", OPTION_VAL_100_O47_HE },
         { "110%", OPTION_VAL_110_O47_HE },
         { "120%", OPTION_VAL_120_O47_HE },
         { "130%", OPTION_VAL_130_O47_HE },
         { "140%", OPTION_VAL_140_O47_HE },
         { "150%", OPTION_VAL_150_O47_HE },
         { "160%", OPTION_VAL_160_O47_HE },
         { "170%", OPTION_VAL_170_O47_HE },
         { "180%", OPTION_VAL_180_O47_HE },
         { "190%", OPTION_VAL_190_O47_HE },
         { "200%", OPTION_VAL_200_O47_HE },
         { "210%", OPTION_VAL_210_O47_HE },
         { "220%", OPTION_VAL_220_O47_HE },
         { "230%", OPTION_VAL_230_O47_HE },
         { "240%", OPTION_VAL_240_O47_HE },
         { "250%", OPTION_VAL_250_O47_HE },
         { "260%", OPTION_VAL_260_O47_HE },
         { "270%", OPTION_VAL_270_O47_HE },
         { "280%", OPTION_VAL_280_O47_HE },
         { "290%", OPTION_VAL_290_O47_HE },
         { "300%", OPTION_VAL_300_O47_HE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_HE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HE },
         { "Red",      OPTION_VAL_RED_HE },
         { "Green",    OPTION_VAL_GREEN_HE },
         { "Blue",     OPTION_VAL_BLUE_HE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_HE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HE },
         { "Red",      OPTION_VAL_RED_HE },
         { "Green",    OPTION_VAL_GREEN_HE },
         { "Blue",     OPTION_VAL_BLUE_HE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_HE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HE },
         { "Red",      OPTION_VAL_RED_HE },
         { "Green",    OPTION_VAL_GREEN_HE },
         { "Blue",     OPTION_VAL_BLUE_HE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_HE,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HE },
         { "Red",      OPTION_VAL_RED_HE },
         { "Green",    OPTION_VAL_GREEN_HE },
         { "Blue",     OPTION_VAL_BLUE_HE },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_HE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HE },
         { "Purupuru", OPTION_VAL_PURUPURU_HE },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_HE },
         { "None",     OPTION_VAL_NONE_HE },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_HE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HE },
         { "Purupuru", OPTION_VAL_PURUPURU_HE },
         { "None",     OPTION_VAL_NONE_HE },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_HE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HE },
         { "Purupuru", OPTION_VAL_PURUPURU_HE },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_HE },
         { "None",     OPTION_VAL_NONE_HE },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_HE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HE },
         { "Purupuru", OPTION_VAL_PURUPURU_HE },
         { "None",     OPTION_VAL_NONE_HE },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_HE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HE },
         { "Purupuru", OPTION_VAL_PURUPURU_HE },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_HE },
         { "None",     OPTION_VAL_NONE_HE },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_HE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HE },
         { "Purupuru", OPTION_VAL_PURUPURU_HE },
         { "None",     OPTION_VAL_NONE_HE },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_HE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HE },
         { "Purupuru", OPTION_VAL_PURUPURU_HE },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_HE },
         { "None",     OPTION_VAL_NONE_HE },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_HE,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HE },
         { "Purupuru", OPTION_VAL_PURUPURU_HE },
         { "None",     OPTION_VAL_NONE_HE },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_HE,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_HE,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_HE,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_HE },
         { "All VMUs", OPTION_VAL_ALL_VMUS_HE },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_HE,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_HE,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_HE,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_HE,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_HE,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_HE,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_HE,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_HE,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HE },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HE },
         { "2x", OPTION_VAL_2_O26_HE },
         { "3x", OPTION_VAL_3X_HE },
         { "4x", OPTION_VAL_4_HE },
         { "5x", OPTION_VAL_5X_HE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HE },
         { "20%",  OPTION_VAL_20_HE },
         { "30%",  OPTION_VAL_30_HE },
         { "40%",  OPTION_VAL_40_HE },
         { "50%",  OPTION_VAL_50_HE },
         { "60%",  OPTION_VAL_60_HE },
         { "70%",  OPTION_VAL_70_HE },
         { "80%",  OPTION_VAL_80_HE },
         { "90%",  OPTION_VAL_90_HE },
         { "100%", OPTION_VAL_100_O47_HE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HE },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HE },
         { "2x", OPTION_VAL_2_O26_HE },
         { "3x", OPTION_VAL_3X_HE },
         { "4x", OPTION_VAL_4_HE },
         { "5x", OPTION_VAL_5X_HE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HE },
         { "20%",  OPTION_VAL_20_HE },
         { "30%",  OPTION_VAL_30_HE },
         { "40%",  OPTION_VAL_40_HE },
         { "50%",  OPTION_VAL_50_HE },
         { "60%",  OPTION_VAL_60_HE },
         { "70%",  OPTION_VAL_70_HE },
         { "80%",  OPTION_VAL_80_HE },
         { "90%",  OPTION_VAL_90_HE },
         { "100%", OPTION_VAL_100_O47_HE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HE },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HE },
         { "2x", OPTION_VAL_2_O26_HE },
         { "3x", OPTION_VAL_3X_HE },
         { "4x", OPTION_VAL_4_HE },
         { "5x", OPTION_VAL_5X_HE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HE },
         { "20%",  OPTION_VAL_20_HE },
         { "30%",  OPTION_VAL_30_HE },
         { "40%",  OPTION_VAL_40_HE },
         { "50%",  OPTION_VAL_50_HE },
         { "60%",  OPTION_VAL_60_HE },
         { "70%",  OPTION_VAL_70_HE },
         { "80%",  OPTION_VAL_80_HE },
         { "90%",  OPTION_VAL_90_HE },
         { "100%", OPTION_VAL_100_O47_HE },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HE },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HE },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HE },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HE },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HE },
         { "2x", OPTION_VAL_2_O26_HE },
         { "3x", OPTION_VAL_3X_HE },
         { "4x", OPTION_VAL_4_HE },
         { "5x", OPTION_VAL_5X_HE },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HE },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HE },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HE },
         { "BLUE 03",           OPTION_VAL_BLUE_HE },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HE },
         { "GREEN 05",          OPTION_VAL_GREEN_HE },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HE },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HE },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HE },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HE },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HE },
         { "RED 11",            OPTION_VAL_RED_HE },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HE },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HE },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HE },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HE },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HE },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HE },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HE },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HE },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HE },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HE },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HE },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HE },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HE },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HE },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HE },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HE },
         { "WHITE 28",          OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_HE,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HE },
         { "20%",  OPTION_VAL_20_HE },
         { "30%",  OPTION_VAL_30_HE },
         { "40%",  OPTION_VAL_40_HE },
         { "50%",  OPTION_VAL_50_HE },
         { "60%",  OPTION_VAL_60_HE },
         { "70%",  OPTION_VAL_70_HE },
         { "80%",  OPTION_VAL_80_HE },
         { "90%",  OPTION_VAL_90_HE },
         { "100%", OPTION_VAL_100_O47_HE },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_he = {
   option_cats_he,
   option_defs_he
};

/* RETRO_LANGUAGE_HR */

#define CATEGORY_SYSTEM_LABEL_HR "Sustav"
#define CATEGORY_SYSTEM_INFO_0_HR NULL
#define CATEGORY_VIDEO_LABEL_HR NULL
#define CATEGORY_VIDEO_INFO_0_HR NULL
#define CATEGORY_PERFORMANCE_LABEL_HR "Performanse"
#define CATEGORY_PERFORMANCE_INFO_0_HR NULL
#define CATEGORY_HACKS_LABEL_HR "Hakovanje emulacije"
#define CATEGORY_HACKS_INFO_0_HR NULL
#define CATEGORY_INPUT_LABEL_HR NULL
#define CATEGORY_INPUT_INFO_0_HR NULL
#define CATEGORY_EXPANSIONS_LABEL_HR NULL
#define CATEGORY_EXPANSIONS_INFO_0_HR NULL
#define CATEGORY_VMU_LABEL_HR NULL
#define CATEGORY_VMU_INFO_0_HR NULL
#define CORE_OPTION_NAME_REGION_LABEL_HR "Regija"
#define OPTION_VAL_JAPAN_HR NULL
#define OPTION_VAL_USA_HR "SAD"
#define OPTION_VAL_EUROPE_HR "Europa"
#define OPTION_VAL_DEFAULT_HR NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_HR "Jezik"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_HR NULL
#define OPTION_VAL_JAPANESE_HR "Japanski"
#define OPTION_VAL_ENGLISH_HR "Engleski"
#define OPTION_VAL_GERMAN_HR "Njemački"
#define OPTION_VAL_FRENCH_HR "Francuski"
#define OPTION_VAL_SPANISH_HR "Španjolski"
#define OPTION_VAL_ITALIAN_HR "Talijanski"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_HR NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_HR "Prisilno korištenje BIOS-a visoke razine emulacije."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_HR "Omogući DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_HR "Omogućuje emulaciju Dreamcastovog audio DSP-a (digitalnog procesora signala). Poboljšava točnost generiranog zvuka, ali povećava zahtjeve za performansama."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_HR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_HR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_HR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_HR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_HR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_HR NULL
#define CORE_OPTION_NAME_UPNP_LABEL_HR NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_HR NULL
#define CORE_OPTION_NAME_DCNET_LABEL_HR NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_HR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_HR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_HR "Promijenite rezoluciju renderiranja."
#define OPTION_VAL_320X240_HR NULL
#define OPTION_VAL_640X480_HR NULL
#define OPTION_VAL_800X600_HR NULL
#define OPTION_VAL_960X720_HR NULL
#define OPTION_VAL_1024X768_HR NULL
#define OPTION_VAL_1280X960_HR NULL
#define OPTION_VAL_1440X1080_HR NULL
#define OPTION_VAL_1600X1200_HR NULL
#define OPTION_VAL_1920X1440_HR NULL
#define OPTION_VAL_2560X1920_HR NULL
#define OPTION_VAL_2880X2160_HR NULL
#define OPTION_VAL_3200X2400_HR NULL
#define OPTION_VAL_3840X2880_HR NULL
#define OPTION_VAL_4480X3360_HR NULL
#define OPTION_VAL_5120X3840_HR NULL
#define OPTION_VAL_5760X4320_HR NULL
#define OPTION_VAL_6400X4800_HR NULL
#define OPTION_VAL_7040X5280_HR NULL
#define OPTION_VAL_7680X5760_HR NULL
#define OPTION_VAL_8320X6240_HR NULL
#define OPTION_VAL_8960X6720_HR NULL
#define OPTION_VAL_9600X7200_HR NULL
#define OPTION_VAL_10240X7680_HR NULL
#define OPTION_VAL_10880X8160_HR NULL
#define OPTION_VAL_11520X8640_HR NULL
#define OPTION_VAL_12160X9120_HR NULL
#define OPTION_VAL_12800X9600_HR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_HR "Vrsta kabela"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_HR NULL
#define OPTION_VAL_VGA_HR NULL
#define OPTION_VAL_TV_RGB_HR NULL
#define OPTION_VAL_TV_COMPOSITE_HR "TV (Kompozitni)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_HR "Televizijski standard"
#define OPTION_VAL_NTSC_HR NULL
#define OPTION_VAL_PAL_HR "PAL (Svijet)"
#define OPTION_VAL_PAL_N_HR "PAL-N (Argentina, Paragvaj, Urugvaj)"
#define OPTION_VAL_PAL_M_HR NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_HR NULL
#define OPTION_VAL_HORIZONTAL_HR "Horizontalno"
#define OPTION_VAL_VERTICAL_HR "Vertikalno"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_HR "Abecedno razvrstavanje"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_HR NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_HR "Po traci (brzo, najmanje precizno)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_HR "Po trokutu (normalno)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_HR "Po pikselu (precizno, ali najsporije)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_HR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_HR "Veće vrijednosti mogu biti potrebne za ispravan prikaz pri višim rezolucijama."
#define OPTION_VAL_512MB_HR NULL
#define OPTION_VAL_1GB_HR NULL
#define OPTION_VAL_2GB_HR NULL
#define OPTION_VAL_4GB_HR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_HR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_HR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_HR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_HR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_HR "Omogući RTT (Render To Texture) međuspremnik"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_HR NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_HR "Mipmapiranje"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_HR NULL
#define CORE_OPTION_NAME_FOG_LABEL_HR "Efekti magle"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_HR "Modifikator glasnoće"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_HR "Značajka Dreamcast GPU-a koju igre obično koriste za iscrtavanje sjena objekata. Ovo bi inače trebalo biti omogućeno - utjecaj na performanse obično je minimalan do zanemariv."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_HR "Anizotropno filtriranje"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_HR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_HR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_HR NULL
#define OPTION_VAL_1_HR NULL
#define OPTION_VAL_2_HR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_HR "Odgodi zamjenu kadrova"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_HR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_HR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_HR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_HR "PowerVR2 naknadni filtar"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_HR "Naknadno obradite renderiranu sliku kako biste simulirali efekte specifične za PowerVR2 GPU i analogne video signale."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_HR "Povećanje razlučivosti tekstura (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_HR "Poboljšajte rukom crtanu 2D pikseliziranu grafiku. Trebalo bi se koristiti samo s 2D pikseliziranim igrama."
#define OPTION_VAL_2_O26_HR NULL
#define OPTION_VAL_4_HR NULL
#define OPTION_VAL_6_HR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_HR "Maksimalna filtrirana veličina za povećanje razlučivosti tekstura"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_HR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_HR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_HR NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_HR NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_HR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_HR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_HR "Pokreće GPU i CPU na različitim dretvama. Strogo preporučeno."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_HR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_HR NULL
#define OPTION_VAL_SOME_HR NULL
#define OPTION_VAL_MORE_HR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_HR "Preskakanje kadrova"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_HR "Postavlja broj kadrova koje treba preskočiti između svakog prikazanog kadra."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_HR "Šifre za široki zaslon (potrebno ponovno pokretanje)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_HR "Aktivira šifre koje određenim igrama omogućuju prikaz u širokom zaslonu."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_HR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_HR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_HR "Brzo učitavanje GD-ROM-a (netočno)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_HR "Ubrzava učitavanje GD-ROM-a."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_HR NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_HR NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_HR NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_HR NULL
#define OPTION_VAL_100_HR NULL
#define OPTION_VAL_110_HR NULL
#define OPTION_VAL_120_HR NULL
#define OPTION_VAL_130_HR NULL
#define OPTION_VAL_140_HR NULL
#define OPTION_VAL_150_HR NULL
#define OPTION_VAL_160_HR NULL
#define OPTION_VAL_170_HR NULL
#define OPTION_VAL_180_HR NULL
#define OPTION_VAL_190_HR NULL
#define OPTION_VAL_200_HR NULL
#define OPTION_VAL_210_HR NULL
#define OPTION_VAL_220_HR NULL
#define OPTION_VAL_230_HR NULL
#define OPTION_VAL_240_HR NULL
#define OPTION_VAL_250_HR NULL
#define OPTION_VAL_260_HR NULL
#define OPTION_VAL_270_HR NULL
#define OPTION_VAL_280_HR NULL
#define OPTION_VAL_290_HR NULL
#define OPTION_VAL_300_HR NULL
#define OPTION_VAL_310_HR NULL
#define OPTION_VAL_320_HR NULL
#define OPTION_VAL_330_HR NULL
#define OPTION_VAL_340_HR NULL
#define OPTION_VAL_350_HR NULL
#define OPTION_VAL_360_HR NULL
#define OPTION_VAL_370_HR NULL
#define OPTION_VAL_380_HR NULL
#define OPTION_VAL_390_HR NULL
#define OPTION_VAL_400_HR NULL
#define OPTION_VAL_410_HR NULL
#define OPTION_VAL_420_HR NULL
#define OPTION_VAL_430_HR NULL
#define OPTION_VAL_440_HR NULL
#define OPTION_VAL_450_HR NULL
#define OPTION_VAL_460_HR NULL
#define OPTION_VAL_470_HR NULL
#define OPTION_VAL_480_HR NULL
#define OPTION_VAL_490_HR NULL
#define OPTION_VAL_500_HR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_HR "Učitaj prilagođene teksture"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_HR NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_HR NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_HR NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_HR NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_HR NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_HR NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_HR NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_HR "Mrtva zona analogne palice"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_HR NULL
#define OPTION_VAL_0_HR NULL
#define OPTION_VAL_5_HR NULL
#define OPTION_VAL_10_HR NULL
#define OPTION_VAL_15_HR NULL
#define OPTION_VAL_20_HR NULL
#define OPTION_VAL_25_HR NULL
#define OPTION_VAL_30_HR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_HR "Mrtva zona okidača"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_HR NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_HR "Digitalni okidači"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_HR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_HR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_HR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_HR "Prikaz postavki svjetlosnog pištolja"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_HR NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_HR NULL
#define OPTION_VAL_50_HR NULL
#define OPTION_VAL_60_HR NULL
#define OPTION_VAL_70_HR NULL
#define OPTION_VAL_80_HR NULL
#define OPTION_VAL_90_HR NULL
#define OPTION_VAL_100_O47_HR NULL
#define OPTION_VAL_110_O47_HR NULL
#define OPTION_VAL_120_O47_HR NULL
#define OPTION_VAL_130_O47_HR NULL
#define OPTION_VAL_140_O47_HR NULL
#define OPTION_VAL_150_O47_HR NULL
#define OPTION_VAL_160_O47_HR NULL
#define OPTION_VAL_170_O47_HR NULL
#define OPTION_VAL_180_O47_HR NULL
#define OPTION_VAL_190_O47_HR NULL
#define OPTION_VAL_200_O47_HR NULL
#define OPTION_VAL_210_O47_HR NULL
#define OPTION_VAL_220_O47_HR NULL
#define OPTION_VAL_230_O47_HR NULL
#define OPTION_VAL_240_O47_HR NULL
#define OPTION_VAL_250_O47_HR NULL
#define OPTION_VAL_260_O47_HR NULL
#define OPTION_VAL_270_O47_HR NULL
#define OPTION_VAL_280_O47_HR NULL
#define OPTION_VAL_290_O47_HR NULL
#define OPTION_VAL_300_O47_HR NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_HR "Prikaz nišana pištolja 1"
#define OPTION_VAL_WHITE_HR NULL
#define OPTION_VAL_RED_HR NULL
#define OPTION_VAL_GREEN_HR NULL
#define OPTION_VAL_BLUE_HR NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_HR "Prikaz nišana pištolja 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_HR "Prikaz nišana pištolja 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_HR "Prikaz nišana pištolja 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_HR NULL
#define OPTION_VAL_VMU_HR NULL
#define OPTION_VAL_PURUPURU_HR NULL
#define OPTION_VAL_DREAMPOTATO_HR NULL
#define OPTION_VAL_NONE_HR "Ništa"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_HR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_HR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_HR "Memorijske jedinice (VMU) / Sustavi po igri"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_HR "VMU-ovi po igri"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_HR NULL
#define OPTION_VAL_VMU_A1_HR NULL
#define OPTION_VAL_ALL_VMUS_HR "Svi VMU-ovi"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_HR NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_HR NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_HR NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_HR NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_HR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_HR "Prikaži postavke zaslona memorijske jedinice (VMU) / sustava"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_HR "Prikaži postavke VMU zaslona"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_HR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_HR "VMU zaslon 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_HR "Pozicija ekrana VMU-a 1"
#define OPTION_VAL_UPPER_LEFT_HR "Gore lijevo"
#define OPTION_VAL_UPPER_RIGHT_HR "Gore desno"
#define OPTION_VAL_LOWER_LEFT_HR "Dolje lijevo"
#define OPTION_VAL_LOWER_RIGHT_HR "Dolje desno"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_HR "Veličina ekrana VMU-a 1"
#define OPTION_VAL_1X_HR NULL
#define OPTION_VAL_3X_HR NULL
#define OPTION_VAL_5X_HR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_HR "Boja uključenog piksela na ekranu VMU-a 1"
#define OPTION_VAL_DEFAULT_ON_00_HR "Zadano UKLJUČENO"
#define OPTION_VAL_DEFAULT_OFF_01_HR "Zadano ISKLJUČENO"
#define OPTION_VAL_BLACK_02_HR NULL
#define OPTION_VAL_LIGHT_BLUE_04_HR "Svijetlo plava"
#define OPTION_VAL_CYAN_06_HR NULL
#define OPTION_VAL_CYAN_BLUE_07_HR "Cijan plava"
#define OPTION_VAL_LIGHT_GREEN_08_HR "Svijetlo zelena"
#define OPTION_VAL_CYAN_GREEN_09_HR "Cijan zelena"
#define OPTION_VAL_LIGHT_CYAN_10_HR "Svijetlo cijan"
#define OPTION_VAL_PURPLE_12_HR NULL
#define OPTION_VAL_LIGHT_PURPLE_13_HR "Svijetlo ljubičasta"
#define OPTION_VAL_YELLOW_14_HR NULL
#define OPTION_VAL_GRAY_15_HR "Siva"
#define OPTION_VAL_LIGHT_PURPLE_2_16_HR "Svijetlo ljubičasta (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_HR "Svijetlo zelena (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_HR "Svijetlo zelena (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_HR "Svijetlo cijan (2)"
#define OPTION_VAL_LIGHT_RED_2_20_HR "Svijetlo crvena (2)"
#define OPTION_VAL_MAGENTA_21_HR NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_HR "Svijetlo ljubičasta (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_HR "Svijetlo narančasta"
#define OPTION_VAL_ORANGE_24_HR NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_HR NULL
#define OPTION_VAL_LIGHT_YELLOW_26_HR "Svijetlo žuta"
#define OPTION_VAL_LIGHT_YELLOW_2_27_HR "Svijetlo žuta (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_HR "Boja isključenih piksela VMU zaslona 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_HR "Prozirnost zaslona 1 VMU-a"
#define OPTION_VAL_40_HR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_HR "Prikaz zaslona 2 VMU-a"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_HR "Položaj zaslona 2 VMU-a"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_HR "Veličina zaslona 2 VMU-a"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_HR "Boja uključenog piksela na zaslonu 2 VMU-a"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_HR "Boja isključenog piksela na zaslonu 2 VMU-a"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_HR "Prozirnost VMU zaslona 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_HR "Prikaz VMU zaslona 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_HR "Položaj VMU zaslona 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_HR "Veličina VMU zaslona 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_HR "Boja uključenih piksela VMU zaslona 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_HR "Boja isključenih piksela VMU zaslona 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_HR "Prozirnost VMU zaslona 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_HR "Prikaz VMU zaslona 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_HR "Položaj VMU zaslona 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_HR "Veličina VMU zaslona 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_HR "Boja uključenih piksela VMU zaslona 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_HR "Boja isključenih piksela VMU zaslona 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_HR "Prozirnost VMU zaslona 4"

struct retro_core_option_v2_category option_cats_hr[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_HR,
      CATEGORY_SYSTEM_INFO_0_HR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_HR,
      CATEGORY_VIDEO_INFO_0_HR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_HR,
      CATEGORY_PERFORMANCE_INFO_0_HR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_HR,
      CATEGORY_HACKS_INFO_0_HR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_HR,
      CATEGORY_INPUT_INFO_0_HR
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_HR,
      CATEGORY_EXPANSIONS_INFO_0_HR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_HR,
      CATEGORY_VMU_INFO_0_HR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_hr[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_HR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_HR },
         { "USA",     OPTION_VAL_USA_HR },
         { "Europe",  OPTION_VAL_EUROPE_HR },
         { "Default", OPTION_VAL_DEFAULT_HR },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_HR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_HR },
         { "English",  OPTION_VAL_ENGLISH_HR },
         { "German",   OPTION_VAL_GERMAN_HR },
         { "French",   OPTION_VAL_FRENCH_HR },
         { "Spanish",  OPTION_VAL_SPANISH_HR },
         { "Italian",  OPTION_VAL_ITALIAN_HR },
         { "Default",  OPTION_VAL_DEFAULT_HR },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_HR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_HR },
         { "640x480",    OPTION_VAL_640X480_HR },
         { "800x600",    OPTION_VAL_800X600_HR },
         { "960x720",    OPTION_VAL_960X720_HR },
         { "1024x768",   OPTION_VAL_1024X768_HR },
         { "1280x960",   OPTION_VAL_1280X960_HR },
         { "1440x1080",  OPTION_VAL_1440X1080_HR },
         { "1600x1200",  OPTION_VAL_1600X1200_HR },
         { "1920x1440",  OPTION_VAL_1920X1440_HR },
         { "2560x1920",  OPTION_VAL_2560X1920_HR },
         { "2880x2160",  OPTION_VAL_2880X2160_HR },
         { "3200x2400",  OPTION_VAL_3200X2400_HR },
         { "3840x2880",  OPTION_VAL_3840X2880_HR },
         { "4480x3360",  OPTION_VAL_4480X3360_HR },
         { "5120x3840",  OPTION_VAL_5120X3840_HR },
         { "5760x4320",  OPTION_VAL_5760X4320_HR },
         { "6400x4800",  OPTION_VAL_6400X4800_HR },
         { "7040x5280",  OPTION_VAL_7040X5280_HR },
         { "7680x5760",  OPTION_VAL_7680X5760_HR },
         { "8320x6240",  OPTION_VAL_8320X6240_HR },
         { "8960x6720",  OPTION_VAL_8960X6720_HR },
         { "9600x7200",  OPTION_VAL_9600X7200_HR },
         { "10240x7680", OPTION_VAL_10240X7680_HR },
         { "10880x8160", OPTION_VAL_10880X8160_HR },
         { "11520x8640", OPTION_VAL_11520X8640_HR },
         { "12160x9120", OPTION_VAL_12160X9120_HR },
         { "12800x9600", OPTION_VAL_12800X9600_HR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_HR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_HR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_HR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_HR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_HR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_HR },
         { "PAL",     OPTION_VAL_PAL_HR },
         { "PAL_N",   OPTION_VAL_PAL_N_HR },
         { "PAL_M",   OPTION_VAL_PAL_M_HR },
         { "Default", OPTION_VAL_DEFAULT_HR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_HR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_HR },
         { "vertical",   OPTION_VAL_VERTICAL_HR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_HR,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_HR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_HR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_HR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_HR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_HR },
         { "1GB",   OPTION_VAL_1GB_HR },
         { "2GB",   OPTION_VAL_2GB_HR },
         { "4GB",   OPTION_VAL_4GB_HR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_HR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_HR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_HR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_HR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_HR },
         { "1",  OPTION_VAL_1_HR },
         { "2",  OPTION_VAL_2_HR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_HR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_HR },
         { "4",  OPTION_VAL_4_HR },
         { "6",  OPTION_VAL_6_HR },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_HR,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_HR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_HR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_HR,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_HR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_HR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_HR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_HR },
         { "more", OPTION_VAL_MORE_HR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_HR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_HR },
         { "110", OPTION_VAL_110_HR },
         { "120", OPTION_VAL_120_HR },
         { "130", OPTION_VAL_130_HR },
         { "140", OPTION_VAL_140_HR },
         { "150", OPTION_VAL_150_HR },
         { "160", OPTION_VAL_160_HR },
         { "170", OPTION_VAL_170_HR },
         { "180", OPTION_VAL_180_HR },
         { "190", OPTION_VAL_190_HR },
         { "200", OPTION_VAL_200_HR },
         { "210", OPTION_VAL_210_HR },
         { "220", OPTION_VAL_220_HR },
         { "230", OPTION_VAL_230_HR },
         { "240", OPTION_VAL_240_HR },
         { "250", OPTION_VAL_250_HR },
         { "260", OPTION_VAL_260_HR },
         { "270", OPTION_VAL_270_HR },
         { "280", OPTION_VAL_280_HR },
         { "290", OPTION_VAL_290_HR },
         { "300", OPTION_VAL_300_HR },
         { "310", OPTION_VAL_310_HR },
         { "320", OPTION_VAL_320_HR },
         { "330", OPTION_VAL_330_HR },
         { "340", OPTION_VAL_340_HR },
         { "350", OPTION_VAL_350_HR },
         { "360", OPTION_VAL_360_HR },
         { "370", OPTION_VAL_370_HR },
         { "380", OPTION_VAL_380_HR },
         { "390", OPTION_VAL_390_HR },
         { "400", OPTION_VAL_400_HR },
         { "410", OPTION_VAL_410_HR },
         { "420", OPTION_VAL_420_HR },
         { "430", OPTION_VAL_430_HR },
         { "440", OPTION_VAL_440_HR },
         { "450", OPTION_VAL_450_HR },
         { "460", OPTION_VAL_460_HR },
         { "470", OPTION_VAL_470_HR },
         { "480", OPTION_VAL_480_HR },
         { "490", OPTION_VAL_490_HR },
         { "500", OPTION_VAL_500_HR },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_HR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_HR },
         { "5%",  OPTION_VAL_5_HR },
         { "10%", OPTION_VAL_10_HR },
         { "15%", OPTION_VAL_15_HR },
         { "20%", OPTION_VAL_20_HR },
         { "25%", OPTION_VAL_25_HR },
         { "30%", OPTION_VAL_30_HR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_HR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_HR },
         { "5%",  OPTION_VAL_5_HR },
         { "10%", OPTION_VAL_10_HR },
         { "15%", OPTION_VAL_15_HR },
         { "20%", OPTION_VAL_20_HR },
         { "25%", OPTION_VAL_25_HR },
         { "30%", OPTION_VAL_30_HR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_HR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_HR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_HR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_HR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_HR },
         { "60%",  OPTION_VAL_60_HR },
         { "70%",  OPTION_VAL_70_HR },
         { "80%",  OPTION_VAL_80_HR },
         { "90%",  OPTION_VAL_90_HR },
         { "100%", OPTION_VAL_100_O47_HR },
         { "110%", OPTION_VAL_110_O47_HR },
         { "120%", OPTION_VAL_120_O47_HR },
         { "130%", OPTION_VAL_130_O47_HR },
         { "140%", OPTION_VAL_140_O47_HR },
         { "150%", OPTION_VAL_150_O47_HR },
         { "160%", OPTION_VAL_160_O47_HR },
         { "170%", OPTION_VAL_170_O47_HR },
         { "180%", OPTION_VAL_180_O47_HR },
         { "190%", OPTION_VAL_190_O47_HR },
         { "200%", OPTION_VAL_200_O47_HR },
         { "210%", OPTION_VAL_210_O47_HR },
         { "220%", OPTION_VAL_220_O47_HR },
         { "230%", OPTION_VAL_230_O47_HR },
         { "240%", OPTION_VAL_240_O47_HR },
         { "250%", OPTION_VAL_250_O47_HR },
         { "260%", OPTION_VAL_260_O47_HR },
         { "270%", OPTION_VAL_270_O47_HR },
         { "280%", OPTION_VAL_280_O47_HR },
         { "290%", OPTION_VAL_290_O47_HR },
         { "300%", OPTION_VAL_300_O47_HR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_HR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HR },
         { "Red",      OPTION_VAL_RED_HR },
         { "Green",    OPTION_VAL_GREEN_HR },
         { "Blue",     OPTION_VAL_BLUE_HR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_HR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HR },
         { "Red",      OPTION_VAL_RED_HR },
         { "Green",    OPTION_VAL_GREEN_HR },
         { "Blue",     OPTION_VAL_BLUE_HR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_HR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HR },
         { "Red",      OPTION_VAL_RED_HR },
         { "Green",    OPTION_VAL_GREEN_HR },
         { "Blue",     OPTION_VAL_BLUE_HR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_HR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HR },
         { "Red",      OPTION_VAL_RED_HR },
         { "Green",    OPTION_VAL_GREEN_HR },
         { "Blue",     OPTION_VAL_BLUE_HR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_HR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HR },
         { "Purupuru", OPTION_VAL_PURUPURU_HR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_HR },
         { "None",     OPTION_VAL_NONE_HR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_HR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HR },
         { "Purupuru", OPTION_VAL_PURUPURU_HR },
         { "None",     OPTION_VAL_NONE_HR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_HR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HR },
         { "Purupuru", OPTION_VAL_PURUPURU_HR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_HR },
         { "None",     OPTION_VAL_NONE_HR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_HR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HR },
         { "Purupuru", OPTION_VAL_PURUPURU_HR },
         { "None",     OPTION_VAL_NONE_HR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_HR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HR },
         { "Purupuru", OPTION_VAL_PURUPURU_HR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_HR },
         { "None",     OPTION_VAL_NONE_HR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_HR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HR },
         { "Purupuru", OPTION_VAL_PURUPURU_HR },
         { "None",     OPTION_VAL_NONE_HR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_HR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HR },
         { "Purupuru", OPTION_VAL_PURUPURU_HR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_HR },
         { "None",     OPTION_VAL_NONE_HR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_HR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HR },
         { "Purupuru", OPTION_VAL_PURUPURU_HR },
         { "None",     OPTION_VAL_NONE_HR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_HR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_HR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_HR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_HR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_HR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_HR,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_HR,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_HR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_HR,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_HR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_HR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_HR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_HR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HR },
         { "2x", OPTION_VAL_2_O26_HR },
         { "3x", OPTION_VAL_3X_HR },
         { "4x", OPTION_VAL_4_HR },
         { "5x", OPTION_VAL_5X_HR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HR },
         { "20%",  OPTION_VAL_20_HR },
         { "30%",  OPTION_VAL_30_HR },
         { "40%",  OPTION_VAL_40_HR },
         { "50%",  OPTION_VAL_50_HR },
         { "60%",  OPTION_VAL_60_HR },
         { "70%",  OPTION_VAL_70_HR },
         { "80%",  OPTION_VAL_80_HR },
         { "90%",  OPTION_VAL_90_HR },
         { "100%", OPTION_VAL_100_O47_HR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HR },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HR },
         { "2x", OPTION_VAL_2_O26_HR },
         { "3x", OPTION_VAL_3X_HR },
         { "4x", OPTION_VAL_4_HR },
         { "5x", OPTION_VAL_5X_HR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HR },
         { "20%",  OPTION_VAL_20_HR },
         { "30%",  OPTION_VAL_30_HR },
         { "40%",  OPTION_VAL_40_HR },
         { "50%",  OPTION_VAL_50_HR },
         { "60%",  OPTION_VAL_60_HR },
         { "70%",  OPTION_VAL_70_HR },
         { "80%",  OPTION_VAL_80_HR },
         { "90%",  OPTION_VAL_90_HR },
         { "100%", OPTION_VAL_100_O47_HR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HR },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HR },
         { "2x", OPTION_VAL_2_O26_HR },
         { "3x", OPTION_VAL_3X_HR },
         { "4x", OPTION_VAL_4_HR },
         { "5x", OPTION_VAL_5X_HR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HR },
         { "20%",  OPTION_VAL_20_HR },
         { "30%",  OPTION_VAL_30_HR },
         { "40%",  OPTION_VAL_40_HR },
         { "50%",  OPTION_VAL_50_HR },
         { "60%",  OPTION_VAL_60_HR },
         { "70%",  OPTION_VAL_70_HR },
         { "80%",  OPTION_VAL_80_HR },
         { "90%",  OPTION_VAL_90_HR },
         { "100%", OPTION_VAL_100_O47_HR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HR },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HR },
         { "2x", OPTION_VAL_2_O26_HR },
         { "3x", OPTION_VAL_3X_HR },
         { "4x", OPTION_VAL_4_HR },
         { "5x", OPTION_VAL_5X_HR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HR },
         { "BLUE 03",           OPTION_VAL_BLUE_HR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HR },
         { "GREEN 05",          OPTION_VAL_GREEN_HR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HR },
         { "RED 11",            OPTION_VAL_RED_HR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HR },
         { "WHITE 28",          OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_HR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HR },
         { "20%",  OPTION_VAL_20_HR },
         { "30%",  OPTION_VAL_30_HR },
         { "40%",  OPTION_VAL_40_HR },
         { "50%",  OPTION_VAL_50_HR },
         { "60%",  OPTION_VAL_60_HR },
         { "70%",  OPTION_VAL_70_HR },
         { "80%",  OPTION_VAL_80_HR },
         { "90%",  OPTION_VAL_90_HR },
         { "100%", OPTION_VAL_100_O47_HR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_hr = {
   option_cats_hr,
   option_defs_hr
};

/* RETRO_LANGUAGE_HU */

#define CATEGORY_SYSTEM_LABEL_HU "Rendszer"
#define CATEGORY_SYSTEM_INFO_0_HU "A régió, nyelv, BIOS, és a hardveralap beállításai."
#define CATEGORY_VIDEO_LABEL_HU "Kép"
#define CATEGORY_VIDEO_INFO_0_HU "A felbontás, sorrendfüggetlen átlátszóság, és vizuális effektek beállításai."
#define CATEGORY_PERFORMANCE_LABEL_HU "Teljesítmény"
#define CATEGORY_PERFORMANCE_INFO_0_HU "A többszálas megjelenítés és képkockakihagyás beállításai."
#define CATEGORY_HACKS_LABEL_HU "Emulációs trükkök"
#define CATEGORY_HACKS_INFO_0_HU "A szélesvásznú módosítások, GD-ROM betöltés sebessége, és a textúracserék beállításai."
#define CATEGORY_INPUT_LABEL_HU "Bemenet"
#define CATEGORY_INPUT_INFO_0_HU "A kontroller és fénypisztoly beállításai."
#define CATEGORY_EXPANSIONS_LABEL_HU "Kontroller bővítőhelyek"
#define CATEGORY_EXPANSIONS_INFO_0_HU "Az egyes kontroller bővítőhelyekbe csatlakoztatott eszköz (VMU, rezgő eszköz)."
#define CATEGORY_VMU_LABEL_HU NULL
#define CATEGORY_VMU_INFO_0_HU "A játékonkénti VMU mentések és a képernyőn látható VMU beállításai."
#define CORE_OPTION_NAME_REGION_LABEL_HU "Régió"
#define OPTION_VAL_JAPAN_HU "Japán"
#define OPTION_VAL_USA_HU NULL
#define OPTION_VAL_EUROPE_HU "Európa"
#define OPTION_VAL_DEFAULT_HU "Alapértelmezett"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_HU "Nyelv"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_HU "A nyelv, amelyet a BIOS és a több nyelvet tartalmazó játékok használnak."
#define OPTION_VAL_JAPANESE_HU "japán"
#define OPTION_VAL_ENGLISH_HU "angol"
#define OPTION_VAL_GERMAN_HU "német"
#define OPTION_VAL_FRENCH_HU "francia"
#define OPTION_VAL_SPANISH_HU "spanyol"
#define OPTION_VAL_ITALIAN_HU "olasz"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_HU "HLE BIOS (újraindítás szükséges)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_HU "Mindenképp a magas szinten emulált BIOS használata."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_HU "DSP engedélyezése"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_HU "A Dreamcast DSP (digitális jelfeldolgozó) emulálásának engedélyezése. Javítja a generált hang pontosságát, de megnöveli a teljesítményigényt."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_HU "Játéktermi szervizgombok engedélyezése"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_HU "A játéktermi SERVICE gomb engedélyezése, a játékgép beállításaihoz."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_HU "Ingyen játék a NAOMI játékokhoz"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_HU "Az érmebeállítások ingyen játékra állítása."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_HU "Szélessávú adapter emulálása"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_HU "Modem helyett a szélessávú Ethernet adapter emulálása. (Újraindítás szükséges.)"
#define CORE_OPTION_NAME_UPNP_LABEL_HU "UPnP engedélyezése"
#define CORE_OPTION_NAME_UPNP_INFO_0_HU "Az internetre csatlakozó router automatikus konfigurálása az online játékokhoz UPnP-n keresztül."
#define CORE_OPTION_NAME_DCNET_LABEL_HU "DCNet használata"
#define CORE_OPTION_NAME_DCNET_INFO_0_HU "A DCNet felhőszolgáltatás használata a Dreamcast internet hozzáféréséhez."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_HU "Belső felbontás"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_HU "A képelőállítás felbontása."
#define OPTION_VAL_320X240_HU "320x240 (Felezett)"
#define OPTION_VAL_640X480_HU "640x480 (Eredeti)"
#define OPTION_VAL_800X600_HU NULL
#define OPTION_VAL_960X720_HU NULL
#define OPTION_VAL_1024X768_HU NULL
#define OPTION_VAL_1280X960_HU NULL
#define OPTION_VAL_1440X1080_HU NULL
#define OPTION_VAL_1600X1200_HU NULL
#define OPTION_VAL_1920X1440_HU NULL
#define OPTION_VAL_2560X1920_HU NULL
#define OPTION_VAL_2880X2160_HU NULL
#define OPTION_VAL_3200X2400_HU NULL
#define OPTION_VAL_3840X2880_HU NULL
#define OPTION_VAL_4480X3360_HU NULL
#define OPTION_VAL_5120X3840_HU NULL
#define OPTION_VAL_5760X4320_HU NULL
#define OPTION_VAL_6400X4800_HU NULL
#define OPTION_VAL_7040X5280_HU NULL
#define OPTION_VAL_7680X5760_HU NULL
#define OPTION_VAL_8320X6240_HU NULL
#define OPTION_VAL_8960X6720_HU NULL
#define OPTION_VAL_9600X7200_HU NULL
#define OPTION_VAL_10240X7680_HU NULL
#define OPTION_VAL_10880X8160_HU NULL
#define OPTION_VAL_11520X8640_HU NULL
#define OPTION_VAL_12160X9120_HU NULL
#define OPTION_VAL_12800X9600_HU NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_HU "Kábel típusa"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_HU "A kimenő jel típusa. A \"TV (kompozit)\" támogatása a legelterjedtebb."
#define OPTION_VAL_VGA_HU NULL
#define OPTION_VAL_TV_RGB_HU NULL
#define OPTION_VAL_TV_COMPOSITE_HU "TV (kompozit)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_HU "Műsorszórási szabvány"
#define OPTION_VAL_NTSC_HU NULL
#define OPTION_VAL_PAL_HU "PAL (Világszerte)"
#define OPTION_VAL_PAL_N_HU NULL
#define OPTION_VAL_PAL_M_HU NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_HU "Képernyő tájolása"
#define OPTION_VAL_HORIZONTAL_HU "Vízszintes"
#define OPTION_VAL_VERTICAL_HU "Függőleges"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_HU "Alfa rendezés"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_HU "Az átlátszó poligonok rendezése."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_HU "Per-strip (gyors, legkevésbé pontos)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_HU "Per-triangle (normál)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_HU "Per-pixel (pontos, de a leglassabb)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_HU "Pixel gyűjtőbuffer mérete"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_HU "Magasabb felbontások helyes visszaadásához nagyobb értékek kellhetnek."
#define OPTION_VAL_512MB_HU NULL
#define OPTION_VAL_1GB_HU NULL
#define OPTION_VAL_2GB_HU NULL
#define OPTION_VAL_4GB_HU NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_HU "Átlátszó rétegek maximális száma"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_HU "Összetett jelenetekhez nagyobb értékekre lehet szükség."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_HU "Teljes framebuffer emuláció"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_HU "A VRAM framebuffer teljes emulálásának engedélyezése. Olyan játékokhoz hasznos, amelyek közvetlenül a VRAM-ban írják vagy olvassák a framebuffert. Bekapcsolva a belső felbontás mindenképpen 640x480 lesz és a teljesítmény erősen visszaeshet."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_HU "RTT (Render to texture) puffer engedélyezése"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_HU "Az előállított textúrák visszamásolása a GPU-ból a VRAM-ba. Ez a lehetőség rendszerint be van kapcsolva azoknál a játékoknál, amelyek igénylik. Bekapcsolt állapotban a textúra előállítás felméretezése kikapcsol, és hatással lehet a teljesítményre."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_HU NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_HU "Bekapcsolva a textúrák saját maguk kisebb változatát használják, amikor távolabb tűnnek fel. Javítja a teljesítményt és csökkenti a remegést."
#define CORE_OPTION_NAME_FOG_LABEL_HU "Köd effekt"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_HU "Térfogatmódosító"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_HU "A Dreamcast GPU-jának egy szolgáltatása, amit a játékok rendszerint tárgyak árnyékának megrajzolásához használnak. Normál esetben legyen bekapcsolva, hatása a teljesítményre rendszerint minimális vagy elhanyagolható."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_HU "Anizotróp szûrés"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_HU "Minőségjavítás azokon a felületi textúrákon, amik a kamerához képest rézsútosan helyezkednek el. Nagyobb érték jobban terheli a GPU-t. A változtatás csak újraindítás után lép életbe."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_HU "Textúraszűrés"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_HU "A textúraszűrés módja. Ezzel lehet egy bizonyos textúraszűrési módot rákényszeríteni minden textúrára, hogy az alapértelmezésnél élesebb (vagy simább) megjelenésük legyen. Az alapértelmezettől eltérő értékek különféle megjelenítési hibákat okozhatnak. A változások csak újraindítás után lépnek életbe."
#define OPTION_VAL_1_HU "Mindenképp Nearest-Neighbor"
#define OPTION_VAL_2_HU "Mindenképp lineáris"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_HU "Képkockacsere késleltetése"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_HU "A villódzó képernyőt vagy hibás mozgóképet segít elkerülni. Lassú platformokon nem ajánlott."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_HU "Képkockasebesség változásainak érzékelése"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_HU "A frontend értesítése, amikor a belső képsebesség megváltozik (pl. 60 fps-ről 30 fps-re). Javítja a megjelenítés ütemét olyan játékoknál, amelyek fix 20 vagy 30 fps-sel futnak, de legyen kikapcsolva nem rögzített (instabil) képsebességű játékoknál (pl. Ecco the Dolphin, Unreal Tournament). Figyelem: nem elérhető, ha az automatikus képkocka kihagyás be van kapcsolva."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_HU "PowerVR2 utófeldolgozó szűrő"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_HU "Az előállított kép utófeldolgozása a PowerVR2 GPU és az analóg videojelek egyedi effektjeinek szimulálásához."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_HU "Textúra felméretezés (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_HU "A kézzel rajzolt 2D pixel art grafikák feljavítása. Csak a 2D pixelgrafikus játékokhoz ajánlott."
#define OPTION_VAL_2_O26_HU NULL
#define OPTION_VAL_4_HU NULL
#define OPTION_VAL_6_HU NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_HU "Textúra felméretezés maximális szűrt mérete"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_HU "A textúra legnagyobb mérete felméretezéshez. Ha a textúra mérete nagyobb, mint az itt kiválasztott, nem lesz felméretezve."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_HU "Natív mélységi interpoláció"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_HU "Segíthet a textúrahibákon és mélységi problémákon AMD GPU-k, vagy néhány esetben Intel GPU-k használatakor is."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_HU "Felméretezés széleinek javítása"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_HU "Segít a textúrák túlcsordulásán felméretezéskor. Kikapcsolva segíthet a 2D játékok (MVC2, CVS, KOF stb.) felméretezésekor jelentkező pixeltorzuláson."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_HU "Többszálas képelőállítás"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_HU "A CPU és GPU külön szálon futnak. Erősen ajánlott."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_HU "Automatikus képkocka kihagyás"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_HU "Képkockák automatikus kihagyása, ha az emulátor lassan fut. Figyelem: csak akkor érvényes, ha a többszálas képelőállítás engedélyezett."
#define OPTION_VAL_SOME_HU "Normál"
#define OPTION_VAL_MORE_HU "Maximális"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_HU "Képkocka kihagyás"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_HU "Két megjelenített képkocka közt ennyi képkocka eldobása."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_HU "Szélesvásznú csalások (újraindítás szükséges)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_HU "Csalások alkalmazása, amelyekkel bizonyos játékok szélesvásznú formátumban jelennek meg."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_HU "Szélesvásznú trükk"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_HU "A normál 4:3-as képarányon kívüli geometria felrajzolása. Grafikai hibákat tartalmazhat az így megjelenített területeken."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_HU "GD-ROM gyorstöltés (pontatlan)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_HU "A GD-ROM betöltésének felgyorsítása."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_HU NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_HU "A Dreamcast 32 MB-os RAM módosítása. Hatással lehet a kompatibilitásra"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_HU "SH4 CPU alul/túlhajtás"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_HU "Az SH4 fő CPU órajelének átállítása az alapértelmezett 200 Mhz-ről. Az alulhajtás segíthet lassú platformokon, a túlhajtás növelheti a képsebességet néhány játéknál. Óvatos használatra."
#define OPTION_VAL_100_HU NULL
#define OPTION_VAL_110_HU NULL
#define OPTION_VAL_120_HU NULL
#define OPTION_VAL_130_HU NULL
#define OPTION_VAL_140_HU NULL
#define OPTION_VAL_150_HU NULL
#define OPTION_VAL_160_HU NULL
#define OPTION_VAL_170_HU NULL
#define OPTION_VAL_180_HU NULL
#define OPTION_VAL_190_HU NULL
#define OPTION_VAL_200_HU NULL
#define OPTION_VAL_210_HU NULL
#define OPTION_VAL_220_HU NULL
#define OPTION_VAL_230_HU NULL
#define OPTION_VAL_240_HU NULL
#define OPTION_VAL_250_HU NULL
#define OPTION_VAL_260_HU NULL
#define OPTION_VAL_270_HU NULL
#define OPTION_VAL_280_HU NULL
#define OPTION_VAL_290_HU NULL
#define OPTION_VAL_300_HU NULL
#define OPTION_VAL_310_HU NULL
#define OPTION_VAL_320_HU NULL
#define OPTION_VAL_330_HU NULL
#define OPTION_VAL_340_HU NULL
#define OPTION_VAL_350_HU NULL
#define OPTION_VAL_360_HU NULL
#define OPTION_VAL_370_HU NULL
#define OPTION_VAL_380_HU NULL
#define OPTION_VAL_390_HU NULL
#define OPTION_VAL_400_HU NULL
#define OPTION_VAL_410_HU NULL
#define OPTION_VAL_420_HU NULL
#define OPTION_VAL_430_HU NULL
#define OPTION_VAL_440_HU NULL
#define OPTION_VAL_450_HU NULL
#define OPTION_VAL_460_HU NULL
#define OPTION_VAL_470_HU NULL
#define OPTION_VAL_480_HU NULL
#define OPTION_VAL_490_HU NULL
#define OPTION_VAL_500_HU NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_HU "Egyedi textúrák betöltése"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_HU "Egyedi textúrák betöltése a 'system/dc/textures/<game-id>/' mappából."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_HU "Egyéni textúrák előzetes betöltése"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_HU "A játék indításakor előre betölti az egyéni textúrákat. Javíthatja a teljesítményt, de növeli a memóriahasználatot."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_HU "Textúrák lementése"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_HU "A textúrák lementése .png fájlként a 'system/dc/texdump/<game-id>/' mappába, amint a játék új textúrát használ."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_HU "Lecserélt textúrák kimentése"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_HU "Az egyéni textúrák által lecserélt textúrák kiírása."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_HU "Analóg kar holtsávja"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_HU "Az analóg kart ilyen mértékben kell elmozdítani ahhoz, hogy hatása legyen."
#define OPTION_VAL_0_HU NULL
#define OPTION_VAL_5_HU NULL
#define OPTION_VAL_10_HU NULL
#define OPTION_VAL_15_HU NULL
#define OPTION_VAL_20_HU NULL
#define OPTION_VAL_25_HU NULL
#define OPTION_VAL_30_HU NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_HU "Ravasz holtsávja"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_HU "A ravaszt ilyen mértékben kell elmozdítani ahhoz, hogy hatása legyen."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_HU "Digitális ravasz"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_HU "Bekapcsolva a ravaszok normális gombokként működnek, teljesen vagy egyáltalán nem lenyomott állapotot visszaadva, köztes érték nélkül."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_HU "Digitális kimenetek sugárzása"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_HU "A digitális kimenetek és az erővisszacsatolás állapotának közzététele a 8000-es TCP porton. Kompatibilis a MAME \"-output network\" opciójával."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_HU "Fénypisztoly beállítások megjelenítése"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_HU "A fénypisztoly célkereszt megjelenítés beállításai. Figyelem: a változtatás életbelépéséhez újra be kell lépni a gyorsmenübe."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_HU "A fénypisztoly célkeresztjének méretezése"
#define OPTION_VAL_50_HU NULL
#define OPTION_VAL_60_HU NULL
#define OPTION_VAL_70_HU NULL
#define OPTION_VAL_80_HU NULL
#define OPTION_VAL_90_HU NULL
#define OPTION_VAL_100_O47_HU NULL
#define OPTION_VAL_110_O47_HU NULL
#define OPTION_VAL_120_O47_HU NULL
#define OPTION_VAL_130_O47_HU NULL
#define OPTION_VAL_140_O47_HU NULL
#define OPTION_VAL_150_O47_HU NULL
#define OPTION_VAL_160_O47_HU NULL
#define OPTION_VAL_170_O47_HU NULL
#define OPTION_VAL_180_O47_HU NULL
#define OPTION_VAL_190_O47_HU NULL
#define OPTION_VAL_200_O47_HU NULL
#define OPTION_VAL_210_O47_HU NULL
#define OPTION_VAL_220_O47_HU NULL
#define OPTION_VAL_230_O47_HU NULL
#define OPTION_VAL_240_O47_HU NULL
#define OPTION_VAL_250_O47_HU NULL
#define OPTION_VAL_260_O47_HU NULL
#define OPTION_VAL_270_O47_HU NULL
#define OPTION_VAL_280_O47_HU NULL
#define OPTION_VAL_290_O47_HU NULL
#define OPTION_VAL_300_O47_HU NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_HU "1. fénypisztoly célkeresztjének megjelenítése"
#define OPTION_VAL_WHITE_HU "Fehér"
#define OPTION_VAL_RED_HU "Vörös"
#define OPTION_VAL_GREEN_HU "Zöld"
#define OPTION_VAL_BLUE_HU "Kék"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_HU "2. fénypisztoly célkeresztjének megjelenítése"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_HU "3. fénypisztoly célkeresztjének megjelenítése"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_HU "4. fénypisztoly célkeresztjének megjelenítése"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_HU "Az A1 bővítőhelyen található eszköz"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_HU "Az A1 bővítőhelybe csatlakoztatott eszköz (A csatlakozó, 1. hely)."
#define OPTION_VAL_VMU_HU NULL
#define OPTION_VAL_PURUPURU_HU NULL
#define OPTION_VAL_DREAMPOTATO_HU NULL
#define OPTION_VAL_NONE_HU "Nincs"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_HU "Az A2 bővítőhelyen található eszköz"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_HU "Az A2 bővítőhelybe csatlakoztatott eszköz (A csatlakozó, 2. hely)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_HU "A B1 bővítőhelyen található eszköz"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_HU "A B1 bővítőhelybe csatlakoztatott eszköz (B csatlakozó, 1. hely)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_HU "A B2 bővítőhelyen található eszköz"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_HU "A B2 bővítőhelybe csatlakoztatott eszköz (B csatlakozó, 2. hely)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_HU "A C1 bővítőhelyen található eszköz"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_HU "A C1 bővítőhelybe csatlakoztatott eszköz (C csatlakozó, 1. hely)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_HU "A C2 bővítőhelyen található eszköz"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_HU "A C2 bővítőhelybe csatlakoztatott eszköz (C csatlakozó, 2. hely)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_HU "A D1 bővítőhelyen található eszköz"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_HU "A D1 bővítőhelybe csatlakoztatott eszköz (D csatlakozó, 1. hely)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_HU "A D2 bővítőhelyen található eszköz"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_HU "A D2 bővítőhelybe csatlakoztatott eszköz (D csatlakozó, 2. hely)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_HU "Játékonként külön Visual Memory Unit/System (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_HU "Játékonként külön VMU"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_HU "Kikapcsolva minden játék ugyanazt a legfeljebb 8 VMU mentés fájlt (A1/A2/B1/B2/C1/C2/D1/D2) használja a RetroArch rendszerkönyvtárban.\n\"VMU A1\" beállításnál minden elindított játéknak külön VMU \"A1\" fájlt hoz létre.\n\"Minden VMU\" beállításnál minden elindított játéknak legfeljebb 8 külön VMU fájlt (A1/A2/B1/B2/C1/C2/D1/D2) hoz létre."
#define OPTION_VAL_VMU_A1_HU NULL
#define OPTION_VAL_ALL_VMUS_HU "Minden VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_HU "A Visual Memory Unit / System (VMU) hangjai"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_HU "VMU hangok"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_HU "Bekapcsolva a VMU csipogása is hallható."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_HU "DreamPotato VMU tárhely"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_HU "A VMU tárolása a csatlakoztatott DreamPotato emulátoron. Bekapcsolt állapotban a helyi VMU mentés fájl nincs használatban."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_HU "Visual Memory Unit/System (VMU) megjelenítési beállítások jelenjenek meg"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_HU "VMU megjelenítési beállítások jelenjenek meg"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_HU "Az emulált VMU LCD képernyő láthatósági, méret, helyzet, és szín beállításainak engedélyezése. Figyelem: a gyorsmenüt ki-be kell kapcsolni, hogy ez a beállítás érvényre jusson."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_HU "VMU 1 képernyő megjelenítése"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_HU "VMU 1 képernyő helyzete"
#define OPTION_VAL_UPPER_LEFT_HU "Bal felső"
#define OPTION_VAL_UPPER_RIGHT_HU "Jobb felső"
#define OPTION_VAL_LOWER_LEFT_HU "Bal alsó"
#define OPTION_VAL_LOWER_RIGHT_HU "Jobb alsó"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_HU "VMU 1 képernyő mérete"
#define OPTION_VAL_1X_HU NULL
#define OPTION_VAL_3X_HU NULL
#define OPTION_VAL_5X_HU NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_HU "VMU 1 képernyő bekapcsolt pixel színe"
#define OPTION_VAL_DEFAULT_ON_00_HU "Alapértelmezésben bekapcsolva"
#define OPTION_VAL_DEFAULT_OFF_01_HU "Alapértelmezésben kikapcsolva"
#define OPTION_VAL_BLACK_02_HU "Fekete"
#define OPTION_VAL_LIGHT_BLUE_04_HU "Világoskék"
#define OPTION_VAL_CYAN_06_HU "Ciánkék"
#define OPTION_VAL_CYAN_BLUE_07_HU "Ciánkék"
#define OPTION_VAL_LIGHT_GREEN_08_HU "Világoszöld"
#define OPTION_VAL_CYAN_GREEN_09_HU "Ciánzöld"
#define OPTION_VAL_LIGHT_CYAN_10_HU "Világos ciánkék"
#define OPTION_VAL_PURPLE_12_HU "Lila"
#define OPTION_VAL_LIGHT_PURPLE_13_HU "Világoslila"
#define OPTION_VAL_YELLOW_14_HU "Sárga"
#define OPTION_VAL_GRAY_15_HU "Szürke"
#define OPTION_VAL_LIGHT_PURPLE_2_16_HU "Világoslila (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_HU "Világoszöld (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_HU "Világoszöld (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_HU "Világos ciánkék (2)"
#define OPTION_VAL_LIGHT_RED_2_20_HU "Világospiros (2)"
#define OPTION_VAL_MAGENTA_21_HU "Bíbor"
#define OPTION_VAL_LIGHT_PURPLE_3_22_HU "Világoslila (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_HU "Világos narancs"
#define OPTION_VAL_ORANGE_24_HU "Narancssárga"
#define OPTION_VAL_LIGHT_PURPLE_4_25_HU "Világoslila (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_HU "Halványsárga"
#define OPTION_VAL_LIGHT_YELLOW_2_27_HU "Halványsárga (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_HU "VMU 1 képernyő kikapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_HU "VMU 1 képernyő átlátszatlansága"
#define OPTION_VAL_40_HU NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_HU "VMU 2 képernyő megjelenítése"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_HU "VMU 2 képernyő helyzete"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_HU "VMU 2 képernyő mérete"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_HU "VMU 2 képernyő bekapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_HU "VMU 2 képernyő kikapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_HU "VMU 2 képernyő átlátszatlansága"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_HU "VMU 3 képernyő megjelenítése"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_HU "VMU 3 képernyő helyzete"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_HU "VMU 3 képernyő mérete"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_HU "VMU 3 képernyő bekapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_HU "VMU 3 képernyő kikapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_HU "VMU 3 képernyő átlátszatlansága"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_HU "VMU 4 képernyő megjelenítése"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_HU "VMU 4 képernyő helyzete"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_HU "VMU 4 képernyő mérete"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_HU "VMU 4 képernyő bekapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_HU "VMU 4 képernyő kikapcsolt pixel színe"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_HU "VMU 4 képernyő átlátszatlansága"

struct retro_core_option_v2_category option_cats_hu[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_HU,
      CATEGORY_SYSTEM_INFO_0_HU
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_HU,
      CATEGORY_VIDEO_INFO_0_HU
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_HU,
      CATEGORY_PERFORMANCE_INFO_0_HU
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_HU,
      CATEGORY_HACKS_INFO_0_HU
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_HU,
      CATEGORY_INPUT_INFO_0_HU
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_HU,
      CATEGORY_EXPANSIONS_INFO_0_HU
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_HU,
      CATEGORY_VMU_INFO_0_HU
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_hu[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_HU,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_HU },
         { "USA",     OPTION_VAL_USA_HU },
         { "Europe",  OPTION_VAL_EUROPE_HU },
         { "Default", OPTION_VAL_DEFAULT_HU },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_HU,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_HU },
         { "English",  OPTION_VAL_ENGLISH_HU },
         { "German",   OPTION_VAL_GERMAN_HU },
         { "French",   OPTION_VAL_FRENCH_HU },
         { "Spanish",  OPTION_VAL_SPANISH_HU },
         { "Italian",  OPTION_VAL_ITALIAN_HU },
         { "Default",  OPTION_VAL_DEFAULT_HU },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_HU,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_HU },
         { "640x480",    OPTION_VAL_640X480_HU },
         { "800x600",    OPTION_VAL_800X600_HU },
         { "960x720",    OPTION_VAL_960X720_HU },
         { "1024x768",   OPTION_VAL_1024X768_HU },
         { "1280x960",   OPTION_VAL_1280X960_HU },
         { "1440x1080",  OPTION_VAL_1440X1080_HU },
         { "1600x1200",  OPTION_VAL_1600X1200_HU },
         { "1920x1440",  OPTION_VAL_1920X1440_HU },
         { "2560x1920",  OPTION_VAL_2560X1920_HU },
         { "2880x2160",  OPTION_VAL_2880X2160_HU },
         { "3200x2400",  OPTION_VAL_3200X2400_HU },
         { "3840x2880",  OPTION_VAL_3840X2880_HU },
         { "4480x3360",  OPTION_VAL_4480X3360_HU },
         { "5120x3840",  OPTION_VAL_5120X3840_HU },
         { "5760x4320",  OPTION_VAL_5760X4320_HU },
         { "6400x4800",  OPTION_VAL_6400X4800_HU },
         { "7040x5280",  OPTION_VAL_7040X5280_HU },
         { "7680x5760",  OPTION_VAL_7680X5760_HU },
         { "8320x6240",  OPTION_VAL_8320X6240_HU },
         { "8960x6720",  OPTION_VAL_8960X6720_HU },
         { "9600x7200",  OPTION_VAL_9600X7200_HU },
         { "10240x7680", OPTION_VAL_10240X7680_HU },
         { "10880x8160", OPTION_VAL_10880X8160_HU },
         { "11520x8640", OPTION_VAL_11520X8640_HU },
         { "12160x9120", OPTION_VAL_12160X9120_HU },
         { "12800x9600", OPTION_VAL_12800X9600_HU },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_HU,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_HU },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_HU },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_HU },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_HU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_HU },
         { "PAL",     OPTION_VAL_PAL_HU },
         { "PAL_N",   OPTION_VAL_PAL_N_HU },
         { "PAL_M",   OPTION_VAL_PAL_M_HU },
         { "Default", OPTION_VAL_DEFAULT_HU },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_HU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_HU },
         { "vertical",   OPTION_VAL_VERTICAL_HU },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_HU,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_HU },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_HU },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_HU },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_HU,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_HU },
         { "1GB",   OPTION_VAL_1GB_HU },
         { "2GB",   OPTION_VAL_2GB_HU },
         { "4GB",   OPTION_VAL_4GB_HU },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_HU,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_HU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_HU,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_HU,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_HU },
         { "1",  OPTION_VAL_1_HU },
         { "2",  OPTION_VAL_2_HU },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_HU,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_HU },
         { "4",  OPTION_VAL_4_HU },
         { "6",  OPTION_VAL_6_HU },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_HU,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_HU,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_HU,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_HU,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_HU,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_HU,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_HU,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_HU },
         { "more", OPTION_VAL_MORE_HU },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_HU,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_HU },
         { "110", OPTION_VAL_110_HU },
         { "120", OPTION_VAL_120_HU },
         { "130", OPTION_VAL_130_HU },
         { "140", OPTION_VAL_140_HU },
         { "150", OPTION_VAL_150_HU },
         { "160", OPTION_VAL_160_HU },
         { "170", OPTION_VAL_170_HU },
         { "180", OPTION_VAL_180_HU },
         { "190", OPTION_VAL_190_HU },
         { "200", OPTION_VAL_200_HU },
         { "210", OPTION_VAL_210_HU },
         { "220", OPTION_VAL_220_HU },
         { "230", OPTION_VAL_230_HU },
         { "240", OPTION_VAL_240_HU },
         { "250", OPTION_VAL_250_HU },
         { "260", OPTION_VAL_260_HU },
         { "270", OPTION_VAL_270_HU },
         { "280", OPTION_VAL_280_HU },
         { "290", OPTION_VAL_290_HU },
         { "300", OPTION_VAL_300_HU },
         { "310", OPTION_VAL_310_HU },
         { "320", OPTION_VAL_320_HU },
         { "330", OPTION_VAL_330_HU },
         { "340", OPTION_VAL_340_HU },
         { "350", OPTION_VAL_350_HU },
         { "360", OPTION_VAL_360_HU },
         { "370", OPTION_VAL_370_HU },
         { "380", OPTION_VAL_380_HU },
         { "390", OPTION_VAL_390_HU },
         { "400", OPTION_VAL_400_HU },
         { "410", OPTION_VAL_410_HU },
         { "420", OPTION_VAL_420_HU },
         { "430", OPTION_VAL_430_HU },
         { "440", OPTION_VAL_440_HU },
         { "450", OPTION_VAL_450_HU },
         { "460", OPTION_VAL_460_HU },
         { "470", OPTION_VAL_470_HU },
         { "480", OPTION_VAL_480_HU },
         { "490", OPTION_VAL_490_HU },
         { "500", OPTION_VAL_500_HU },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_HU,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_HU },
         { "5%",  OPTION_VAL_5_HU },
         { "10%", OPTION_VAL_10_HU },
         { "15%", OPTION_VAL_15_HU },
         { "20%", OPTION_VAL_20_HU },
         { "25%", OPTION_VAL_25_HU },
         { "30%", OPTION_VAL_30_HU },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_HU,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_HU },
         { "5%",  OPTION_VAL_5_HU },
         { "10%", OPTION_VAL_10_HU },
         { "15%", OPTION_VAL_15_HU },
         { "20%", OPTION_VAL_20_HU },
         { "25%", OPTION_VAL_25_HU },
         { "30%", OPTION_VAL_30_HU },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_HU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_HU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_HU,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_HU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_HU },
         { "60%",  OPTION_VAL_60_HU },
         { "70%",  OPTION_VAL_70_HU },
         { "80%",  OPTION_VAL_80_HU },
         { "90%",  OPTION_VAL_90_HU },
         { "100%", OPTION_VAL_100_O47_HU },
         { "110%", OPTION_VAL_110_O47_HU },
         { "120%", OPTION_VAL_120_O47_HU },
         { "130%", OPTION_VAL_130_O47_HU },
         { "140%", OPTION_VAL_140_O47_HU },
         { "150%", OPTION_VAL_150_O47_HU },
         { "160%", OPTION_VAL_160_O47_HU },
         { "170%", OPTION_VAL_170_O47_HU },
         { "180%", OPTION_VAL_180_O47_HU },
         { "190%", OPTION_VAL_190_O47_HU },
         { "200%", OPTION_VAL_200_O47_HU },
         { "210%", OPTION_VAL_210_O47_HU },
         { "220%", OPTION_VAL_220_O47_HU },
         { "230%", OPTION_VAL_230_O47_HU },
         { "240%", OPTION_VAL_240_O47_HU },
         { "250%", OPTION_VAL_250_O47_HU },
         { "260%", OPTION_VAL_260_O47_HU },
         { "270%", OPTION_VAL_270_O47_HU },
         { "280%", OPTION_VAL_280_O47_HU },
         { "290%", OPTION_VAL_290_O47_HU },
         { "300%", OPTION_VAL_300_O47_HU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_HU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HU },
         { "Red",      OPTION_VAL_RED_HU },
         { "Green",    OPTION_VAL_GREEN_HU },
         { "Blue",     OPTION_VAL_BLUE_HU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_HU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HU },
         { "Red",      OPTION_VAL_RED_HU },
         { "Green",    OPTION_VAL_GREEN_HU },
         { "Blue",     OPTION_VAL_BLUE_HU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_HU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HU },
         { "Red",      OPTION_VAL_RED_HU },
         { "Green",    OPTION_VAL_GREEN_HU },
         { "Blue",     OPTION_VAL_BLUE_HU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_HU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_HU },
         { "Red",      OPTION_VAL_RED_HU },
         { "Green",    OPTION_VAL_GREEN_HU },
         { "Blue",     OPTION_VAL_BLUE_HU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_HU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HU },
         { "Purupuru", OPTION_VAL_PURUPURU_HU },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_HU },
         { "None",     OPTION_VAL_NONE_HU },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_HU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HU },
         { "Purupuru", OPTION_VAL_PURUPURU_HU },
         { "None",     OPTION_VAL_NONE_HU },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_HU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HU },
         { "Purupuru", OPTION_VAL_PURUPURU_HU },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_HU },
         { "None",     OPTION_VAL_NONE_HU },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_HU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HU },
         { "Purupuru", OPTION_VAL_PURUPURU_HU },
         { "None",     OPTION_VAL_NONE_HU },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_HU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HU },
         { "Purupuru", OPTION_VAL_PURUPURU_HU },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_HU },
         { "None",     OPTION_VAL_NONE_HU },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_HU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HU },
         { "Purupuru", OPTION_VAL_PURUPURU_HU },
         { "None",     OPTION_VAL_NONE_HU },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_HU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HU },
         { "Purupuru", OPTION_VAL_PURUPURU_HU },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_HU },
         { "None",     OPTION_VAL_NONE_HU },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_HU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_HU },
         { "Purupuru", OPTION_VAL_PURUPURU_HU },
         { "None",     OPTION_VAL_NONE_HU },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_HU,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_HU,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_HU,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_HU },
         { "All VMUs", OPTION_VAL_ALL_VMUS_HU },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_HU,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_HU,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_HU,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_HU,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_HU,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_HU,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_HU,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_HU,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HU },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HU },
         { "2x", OPTION_VAL_2_O26_HU },
         { "3x", OPTION_VAL_3X_HU },
         { "4x", OPTION_VAL_4_HU },
         { "5x", OPTION_VAL_5X_HU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HU },
         { "20%",  OPTION_VAL_20_HU },
         { "30%",  OPTION_VAL_30_HU },
         { "40%",  OPTION_VAL_40_HU },
         { "50%",  OPTION_VAL_50_HU },
         { "60%",  OPTION_VAL_60_HU },
         { "70%",  OPTION_VAL_70_HU },
         { "80%",  OPTION_VAL_80_HU },
         { "90%",  OPTION_VAL_90_HU },
         { "100%", OPTION_VAL_100_O47_HU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HU },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HU },
         { "2x", OPTION_VAL_2_O26_HU },
         { "3x", OPTION_VAL_3X_HU },
         { "4x", OPTION_VAL_4_HU },
         { "5x", OPTION_VAL_5X_HU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HU },
         { "20%",  OPTION_VAL_20_HU },
         { "30%",  OPTION_VAL_30_HU },
         { "40%",  OPTION_VAL_40_HU },
         { "50%",  OPTION_VAL_50_HU },
         { "60%",  OPTION_VAL_60_HU },
         { "70%",  OPTION_VAL_70_HU },
         { "80%",  OPTION_VAL_80_HU },
         { "90%",  OPTION_VAL_90_HU },
         { "100%", OPTION_VAL_100_O47_HU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HU },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HU },
         { "2x", OPTION_VAL_2_O26_HU },
         { "3x", OPTION_VAL_3X_HU },
         { "4x", OPTION_VAL_4_HU },
         { "5x", OPTION_VAL_5X_HU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HU },
         { "20%",  OPTION_VAL_20_HU },
         { "30%",  OPTION_VAL_30_HU },
         { "40%",  OPTION_VAL_40_HU },
         { "50%",  OPTION_VAL_50_HU },
         { "60%",  OPTION_VAL_60_HU },
         { "70%",  OPTION_VAL_70_HU },
         { "80%",  OPTION_VAL_80_HU },
         { "90%",  OPTION_VAL_90_HU },
         { "100%", OPTION_VAL_100_O47_HU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_HU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_HU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_HU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_HU },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_HU },
         { "2x", OPTION_VAL_2_O26_HU },
         { "3x", OPTION_VAL_3X_HU },
         { "4x", OPTION_VAL_4_HU },
         { "5x", OPTION_VAL_5X_HU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_HU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_HU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_HU },
         { "BLUE 03",           OPTION_VAL_BLUE_HU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_HU },
         { "GREEN 05",          OPTION_VAL_GREEN_HU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_HU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_HU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_HU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_HU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_HU },
         { "RED 11",            OPTION_VAL_RED_HU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_HU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_HU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_HU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_HU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_HU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_HU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_HU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_HU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_HU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_HU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_HU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_HU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_HU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_HU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_HU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_HU },
         { "WHITE 28",          OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_HU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_HU },
         { "20%",  OPTION_VAL_20_HU },
         { "30%",  OPTION_VAL_30_HU },
         { "40%",  OPTION_VAL_40_HU },
         { "50%",  OPTION_VAL_50_HU },
         { "60%",  OPTION_VAL_60_HU },
         { "70%",  OPTION_VAL_70_HU },
         { "80%",  OPTION_VAL_80_HU },
         { "90%",  OPTION_VAL_90_HU },
         { "100%", OPTION_VAL_100_O47_HU },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_hu = {
   option_cats_hu,
   option_defs_hu
};

/* RETRO_LANGUAGE_ID */

#define CATEGORY_SYSTEM_LABEL_ID "Sistem"
#define CATEGORY_SYSTEM_INFO_0_ID "Menyetel daerah, bahasa, BIOS, dan pengaturan peranti keras dasar."
#define CATEGORY_VIDEO_LABEL_ID NULL
#define CATEGORY_VIDEO_INFO_0_ID NULL
#define CATEGORY_PERFORMANCE_LABEL_ID "Kinerja"
#define CATEGORY_PERFORMANCE_INFO_0_ID NULL
#define CATEGORY_HACKS_LABEL_ID "Peretasan Emulasi"
#define CATEGORY_HACKS_INFO_0_ID NULL
#define CATEGORY_INPUT_LABEL_ID "Masukan"
#define CATEGORY_INPUT_INFO_0_ID NULL
#define CATEGORY_EXPANSIONS_LABEL_ID NULL
#define CATEGORY_EXPANSIONS_INFO_0_ID NULL
#define CATEGORY_VMU_LABEL_ID NULL
#define CATEGORY_VMU_INFO_0_ID NULL
#define CORE_OPTION_NAME_REGION_LABEL_ID "Daerah"
#define OPTION_VAL_JAPAN_ID "Jepang"
#define OPTION_VAL_USA_ID NULL
#define OPTION_VAL_EUROPE_ID "Eropa"
#define OPTION_VAL_DEFAULT_ID NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_ID "Bahasa"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_ID NULL
#define OPTION_VAL_JAPANESE_ID "Bahasa Jepang"
#define OPTION_VAL_ENGLISH_ID "Bahasa Inggris"
#define OPTION_VAL_GERMAN_ID "Bahasa Jerman"
#define OPTION_VAL_FRENCH_ID "Bahasa Perancis"
#define OPTION_VAL_SPANISH_ID "Bahasa Spanyol"
#define OPTION_VAL_ITALIAN_ID "Bahasa Itali"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_ID "BIOS HLE (Perlu Mulai Ulang)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_ID "Paksa penggunaan BIOS emulasi tingkat tinggi."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_ID "Aktifkan DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_ID "Aktifkan emulasi DSP audio Dreamcast (pengolah sinyal digital). Perbaiki akurasi suara yang dihasilakan, tetapi tingkatkan persyaratan kinerja."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_ID NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_ID NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_ID NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_ID NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_ID NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_ID NULL
#define CORE_OPTION_NAME_UPNP_LABEL_ID NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_ID NULL
#define CORE_OPTION_NAME_DCNET_LABEL_ID NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_ID NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_ID "Resolusi Internal"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_ID "Modifikasi resolusi render."
#define OPTION_VAL_320X240_ID NULL
#define OPTION_VAL_640X480_ID "640x480 (Lokal)"
#define OPTION_VAL_800X600_ID NULL
#define OPTION_VAL_960X720_ID NULL
#define OPTION_VAL_1024X768_ID NULL
#define OPTION_VAL_1280X960_ID NULL
#define OPTION_VAL_1440X1080_ID NULL
#define OPTION_VAL_1600X1200_ID NULL
#define OPTION_VAL_1920X1440_ID NULL
#define OPTION_VAL_2560X1920_ID NULL
#define OPTION_VAL_2880X2160_ID NULL
#define OPTION_VAL_3200X2400_ID NULL
#define OPTION_VAL_3840X2880_ID NULL
#define OPTION_VAL_4480X3360_ID NULL
#define OPTION_VAL_5120X3840_ID NULL
#define OPTION_VAL_5760X4320_ID NULL
#define OPTION_VAL_6400X4800_ID NULL
#define OPTION_VAL_7040X5280_ID NULL
#define OPTION_VAL_7680X5760_ID NULL
#define OPTION_VAL_8320X6240_ID NULL
#define OPTION_VAL_8960X6720_ID NULL
#define OPTION_VAL_9600X7200_ID NULL
#define OPTION_VAL_10240X7680_ID NULL
#define OPTION_VAL_10880X8160_ID NULL
#define OPTION_VAL_11520X8640_ID NULL
#define OPTION_VAL_12160X9120_ID NULL
#define OPTION_VAL_12800X9600_ID NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_ID NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_ID NULL
#define OPTION_VAL_VGA_ID NULL
#define OPTION_VAL_TV_RGB_ID NULL
#define OPTION_VAL_TV_COMPOSITE_ID NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_ID NULL
#define OPTION_VAL_NTSC_ID NULL
#define OPTION_VAL_PAL_ID NULL
#define OPTION_VAL_PAL_N_ID NULL
#define OPTION_VAL_PAL_M_ID NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_ID "Orientasi Layar"
#define OPTION_VAL_HORIZONTAL_ID NULL
#define OPTION_VAL_VERTICAL_ID "Vertikal"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_ID NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_ID NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_ID NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_ID NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_ID NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_ID NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_ID "Mungkin perlu nilai yang lebih besar agar resolusi yang lebih besar terlihat benar."
#define OPTION_VAL_512MB_ID NULL
#define OPTION_VAL_1GB_ID NULL
#define OPTION_VAL_2GB_ID NULL
#define OPTION_VAL_4GB_ID NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_ID NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_ID NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_ID NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_ID NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_ID "Mengaktifkan Dapar RTT (Render Ke Tekstur)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_ID NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_ID NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_ID NULL
#define CORE_OPTION_NAME_FOG_LABEL_ID NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_ID NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_ID NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_ID "Penyaringan Anisotropic"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_ID "Meningkatkan mutu tekstur permukaan yang bersudut pandang serong di mata kamera. Opsi nilai tinggi lebih menguras GPU. Perubahan ini hanya berlaku setelah dimulai ulang."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_ID "Penyaringan Tekstur"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_ID "Pilihan mode saring tekstur. Ini dapat memaksa mode saring tekstur tertentu pada semua tekstur agar berpenampilan lebih tajam (atau mulus) dibandingkan Bawaan. Opsi nilai selain Bawaan mungkin akan muncul masalah visual. Perubahan ini hanya berlaku setelah dimulai ulang."
#define OPTION_VAL_1_ID NULL
#define OPTION_VAL_2_ID NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_ID NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_ID NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_ID NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_ID NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_ID "Saring Pasca-pengolahan PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_ID "Memasca-olah gambar yang di-render untuk meniru efek seperti di GPU PowerVR2 dan sinyal video analog."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_ID "Penskalaan Naik Tekstur (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_ID "Meningkatkan grafis seni piksel 2D karya tangan. Hanya bisa digunakan di pemainan piksel 2D."
#define OPTION_VAL_2_O26_ID NULL
#define OPTION_VAL_4_ID NULL
#define OPTION_VAL_6_ID NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_ID "Ukuran Saring Maks. Tekstur Naik Skala"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_ID NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_ID NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_ID NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_ID NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_ID NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_ID "Perenderan Berutas"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_ID "Jalankan GPU dan CPU pada Thread berbeda. Sangat direkomendasikan."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_ID NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_ID NULL
#define OPTION_VAL_SOME_ID NULL
#define OPTION_VAL_MORE_ID NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_ID NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_ID NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_ID "Cheat Layar Lebar (Perlu Mulai Ulang)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_ID NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_ID NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_ID NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_ID NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_ID NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_ID NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_ID NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_ID NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_ID NULL
#define OPTION_VAL_100_ID NULL
#define OPTION_VAL_110_ID NULL
#define OPTION_VAL_120_ID NULL
#define OPTION_VAL_130_ID NULL
#define OPTION_VAL_140_ID NULL
#define OPTION_VAL_150_ID NULL
#define OPTION_VAL_160_ID NULL
#define OPTION_VAL_170_ID NULL
#define OPTION_VAL_180_ID NULL
#define OPTION_VAL_190_ID NULL
#define OPTION_VAL_200_ID NULL
#define OPTION_VAL_210_ID NULL
#define OPTION_VAL_220_ID NULL
#define OPTION_VAL_230_ID NULL
#define OPTION_VAL_240_ID NULL
#define OPTION_VAL_250_ID NULL
#define OPTION_VAL_260_ID NULL
#define OPTION_VAL_270_ID NULL
#define OPTION_VAL_280_ID NULL
#define OPTION_VAL_290_ID NULL
#define OPTION_VAL_300_ID NULL
#define OPTION_VAL_310_ID NULL
#define OPTION_VAL_320_ID NULL
#define OPTION_VAL_330_ID NULL
#define OPTION_VAL_340_ID NULL
#define OPTION_VAL_350_ID NULL
#define OPTION_VAL_360_ID NULL
#define OPTION_VAL_370_ID NULL
#define OPTION_VAL_380_ID NULL
#define OPTION_VAL_390_ID NULL
#define OPTION_VAL_400_ID NULL
#define OPTION_VAL_410_ID NULL
#define OPTION_VAL_420_ID NULL
#define OPTION_VAL_430_ID NULL
#define OPTION_VAL_440_ID NULL
#define OPTION_VAL_450_ID NULL
#define OPTION_VAL_460_ID NULL
#define OPTION_VAL_470_ID NULL
#define OPTION_VAL_480_ID NULL
#define OPTION_VAL_490_ID NULL
#define OPTION_VAL_500_ID NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_ID NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_ID NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_ID NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_ID NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_ID "Dump Tekstur"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_ID NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_ID NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_ID NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_ID NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_ID NULL
#define OPTION_VAL_0_ID NULL
#define OPTION_VAL_5_ID NULL
#define OPTION_VAL_10_ID NULL
#define OPTION_VAL_15_ID NULL
#define OPTION_VAL_20_ID NULL
#define OPTION_VAL_25_ID NULL
#define OPTION_VAL_30_ID NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_ID "Pemicu Deadzone"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_ID NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_ID "Pemicu Digital"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_ID NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_ID NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_ID NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_ID NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_ID NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_ID NULL
#define OPTION_VAL_50_ID NULL
#define OPTION_VAL_60_ID NULL
#define OPTION_VAL_70_ID NULL
#define OPTION_VAL_80_ID NULL
#define OPTION_VAL_90_ID NULL
#define OPTION_VAL_100_O47_ID NULL
#define OPTION_VAL_110_O47_ID NULL
#define OPTION_VAL_120_O47_ID NULL
#define OPTION_VAL_130_O47_ID NULL
#define OPTION_VAL_140_O47_ID NULL
#define OPTION_VAL_150_O47_ID NULL
#define OPTION_VAL_160_O47_ID NULL
#define OPTION_VAL_170_O47_ID NULL
#define OPTION_VAL_180_O47_ID NULL
#define OPTION_VAL_190_O47_ID NULL
#define OPTION_VAL_200_O47_ID NULL
#define OPTION_VAL_210_O47_ID NULL
#define OPTION_VAL_220_O47_ID NULL
#define OPTION_VAL_230_O47_ID NULL
#define OPTION_VAL_240_O47_ID NULL
#define OPTION_VAL_250_O47_ID NULL
#define OPTION_VAL_260_O47_ID NULL
#define OPTION_VAL_270_O47_ID NULL
#define OPTION_VAL_280_O47_ID NULL
#define OPTION_VAL_290_O47_ID NULL
#define OPTION_VAL_300_O47_ID NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_ID NULL
#define OPTION_VAL_WHITE_ID "Putih"
#define OPTION_VAL_RED_ID NULL
#define OPTION_VAL_GREEN_ID NULL
#define OPTION_VAL_BLUE_ID NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_ID NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_ID NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_ID NULL
#define OPTION_VAL_VMU_ID NULL
#define OPTION_VAL_PURUPURU_ID NULL
#define OPTION_VAL_DREAMPOTATO_ID NULL
#define OPTION_VAL_NONE_ID "Tak ada"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_ID NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_ID NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_ID NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_ID NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_ID NULL
#define OPTION_VAL_VMU_A1_ID NULL
#define OPTION_VAL_ALL_VMUS_ID NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_ID NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_ID NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_ID NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_ID NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_ID NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_ID NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_ID NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_ID NULL
#define OPTION_VAL_UPPER_LEFT_ID "Kiri Atas"
#define OPTION_VAL_UPPER_RIGHT_ID "Kanan Atas"
#define OPTION_VAL_LOWER_LEFT_ID "Kiri Bawah"
#define OPTION_VAL_LOWER_RIGHT_ID "Kanan Bawah"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_ID NULL
#define OPTION_VAL_1X_ID NULL
#define OPTION_VAL_3X_ID NULL
#define OPTION_VAL_5X_ID NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_ID NULL
#define OPTION_VAL_DEFAULT_ON_00_ID NULL
#define OPTION_VAL_DEFAULT_OFF_01_ID NULL
#define OPTION_VAL_BLACK_02_ID "Hitam"
#define OPTION_VAL_LIGHT_BLUE_04_ID NULL
#define OPTION_VAL_CYAN_06_ID NULL
#define OPTION_VAL_CYAN_BLUE_07_ID NULL
#define OPTION_VAL_LIGHT_GREEN_08_ID NULL
#define OPTION_VAL_CYAN_GREEN_09_ID NULL
#define OPTION_VAL_LIGHT_CYAN_10_ID NULL
#define OPTION_VAL_PURPLE_12_ID "Ungu"
#define OPTION_VAL_LIGHT_PURPLE_13_ID NULL
#define OPTION_VAL_YELLOW_14_ID NULL
#define OPTION_VAL_GRAY_15_ID NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_ID NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_ID NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_ID NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_ID NULL
#define OPTION_VAL_LIGHT_RED_2_20_ID NULL
#define OPTION_VAL_MAGENTA_21_ID NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_ID NULL
#define OPTION_VAL_LIGHT_ORANGE_23_ID NULL
#define OPTION_VAL_ORANGE_24_ID "Oranye"
#define OPTION_VAL_LIGHT_PURPLE_4_25_ID NULL
#define OPTION_VAL_LIGHT_YELLOW_26_ID NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_ID NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_ID NULL
#define OPTION_VAL_40_ID NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_ID NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_ID NULL

struct retro_core_option_v2_category option_cats_id[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_ID,
      CATEGORY_SYSTEM_INFO_0_ID
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_ID,
      CATEGORY_VIDEO_INFO_0_ID
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_ID,
      CATEGORY_PERFORMANCE_INFO_0_ID
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_ID,
      CATEGORY_HACKS_INFO_0_ID
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_ID,
      CATEGORY_INPUT_INFO_0_ID
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_ID,
      CATEGORY_EXPANSIONS_INFO_0_ID
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_ID,
      CATEGORY_VMU_INFO_0_ID
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_id[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_ID,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_ID },
         { "USA",     OPTION_VAL_USA_ID },
         { "Europe",  OPTION_VAL_EUROPE_ID },
         { "Default", OPTION_VAL_DEFAULT_ID },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_ID,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_ID },
         { "English",  OPTION_VAL_ENGLISH_ID },
         { "German",   OPTION_VAL_GERMAN_ID },
         { "French",   OPTION_VAL_FRENCH_ID },
         { "Spanish",  OPTION_VAL_SPANISH_ID },
         { "Italian",  OPTION_VAL_ITALIAN_ID },
         { "Default",  OPTION_VAL_DEFAULT_ID },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_ID,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_ID },
         { "640x480",    OPTION_VAL_640X480_ID },
         { "800x600",    OPTION_VAL_800X600_ID },
         { "960x720",    OPTION_VAL_960X720_ID },
         { "1024x768",   OPTION_VAL_1024X768_ID },
         { "1280x960",   OPTION_VAL_1280X960_ID },
         { "1440x1080",  OPTION_VAL_1440X1080_ID },
         { "1600x1200",  OPTION_VAL_1600X1200_ID },
         { "1920x1440",  OPTION_VAL_1920X1440_ID },
         { "2560x1920",  OPTION_VAL_2560X1920_ID },
         { "2880x2160",  OPTION_VAL_2880X2160_ID },
         { "3200x2400",  OPTION_VAL_3200X2400_ID },
         { "3840x2880",  OPTION_VAL_3840X2880_ID },
         { "4480x3360",  OPTION_VAL_4480X3360_ID },
         { "5120x3840",  OPTION_VAL_5120X3840_ID },
         { "5760x4320",  OPTION_VAL_5760X4320_ID },
         { "6400x4800",  OPTION_VAL_6400X4800_ID },
         { "7040x5280",  OPTION_VAL_7040X5280_ID },
         { "7680x5760",  OPTION_VAL_7680X5760_ID },
         { "8320x6240",  OPTION_VAL_8320X6240_ID },
         { "8960x6720",  OPTION_VAL_8960X6720_ID },
         { "9600x7200",  OPTION_VAL_9600X7200_ID },
         { "10240x7680", OPTION_VAL_10240X7680_ID },
         { "10880x8160", OPTION_VAL_10880X8160_ID },
         { "11520x8640", OPTION_VAL_11520X8640_ID },
         { "12160x9120", OPTION_VAL_12160X9120_ID },
         { "12800x9600", OPTION_VAL_12800X9600_ID },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_ID,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_ID },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_ID },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_ID },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_ID,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_ID },
         { "PAL",     OPTION_VAL_PAL_ID },
         { "PAL_N",   OPTION_VAL_PAL_N_ID },
         { "PAL_M",   OPTION_VAL_PAL_M_ID },
         { "Default", OPTION_VAL_DEFAULT_ID },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_ID,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_ID },
         { "vertical",   OPTION_VAL_VERTICAL_ID },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_ID,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_ID },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_ID },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_ID },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_ID,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_ID },
         { "1GB",   OPTION_VAL_1GB_ID },
         { "2GB",   OPTION_VAL_2GB_ID },
         { "4GB",   OPTION_VAL_4GB_ID },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_ID,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_ID,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_ID,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_ID,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_ID },
         { "1",  OPTION_VAL_1_ID },
         { "2",  OPTION_VAL_2_ID },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_ID,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_ID },
         { "4",  OPTION_VAL_4_ID },
         { "6",  OPTION_VAL_6_ID },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_ID,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_ID,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_ID,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_ID,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_ID,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_ID,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_ID,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_ID },
         { "more", OPTION_VAL_MORE_ID },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_ID,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_ID },
         { "110", OPTION_VAL_110_ID },
         { "120", OPTION_VAL_120_ID },
         { "130", OPTION_VAL_130_ID },
         { "140", OPTION_VAL_140_ID },
         { "150", OPTION_VAL_150_ID },
         { "160", OPTION_VAL_160_ID },
         { "170", OPTION_VAL_170_ID },
         { "180", OPTION_VAL_180_ID },
         { "190", OPTION_VAL_190_ID },
         { "200", OPTION_VAL_200_ID },
         { "210", OPTION_VAL_210_ID },
         { "220", OPTION_VAL_220_ID },
         { "230", OPTION_VAL_230_ID },
         { "240", OPTION_VAL_240_ID },
         { "250", OPTION_VAL_250_ID },
         { "260", OPTION_VAL_260_ID },
         { "270", OPTION_VAL_270_ID },
         { "280", OPTION_VAL_280_ID },
         { "290", OPTION_VAL_290_ID },
         { "300", OPTION_VAL_300_ID },
         { "310", OPTION_VAL_310_ID },
         { "320", OPTION_VAL_320_ID },
         { "330", OPTION_VAL_330_ID },
         { "340", OPTION_VAL_340_ID },
         { "350", OPTION_VAL_350_ID },
         { "360", OPTION_VAL_360_ID },
         { "370", OPTION_VAL_370_ID },
         { "380", OPTION_VAL_380_ID },
         { "390", OPTION_VAL_390_ID },
         { "400", OPTION_VAL_400_ID },
         { "410", OPTION_VAL_410_ID },
         { "420", OPTION_VAL_420_ID },
         { "430", OPTION_VAL_430_ID },
         { "440", OPTION_VAL_440_ID },
         { "450", OPTION_VAL_450_ID },
         { "460", OPTION_VAL_460_ID },
         { "470", OPTION_VAL_470_ID },
         { "480", OPTION_VAL_480_ID },
         { "490", OPTION_VAL_490_ID },
         { "500", OPTION_VAL_500_ID },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_ID,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_ID },
         { "5%",  OPTION_VAL_5_ID },
         { "10%", OPTION_VAL_10_ID },
         { "15%", OPTION_VAL_15_ID },
         { "20%", OPTION_VAL_20_ID },
         { "25%", OPTION_VAL_25_ID },
         { "30%", OPTION_VAL_30_ID },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_ID,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_ID },
         { "5%",  OPTION_VAL_5_ID },
         { "10%", OPTION_VAL_10_ID },
         { "15%", OPTION_VAL_15_ID },
         { "20%", OPTION_VAL_20_ID },
         { "25%", OPTION_VAL_25_ID },
         { "30%", OPTION_VAL_30_ID },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_ID,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_ID,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_ID,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_ID,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_ID },
         { "60%",  OPTION_VAL_60_ID },
         { "70%",  OPTION_VAL_70_ID },
         { "80%",  OPTION_VAL_80_ID },
         { "90%",  OPTION_VAL_90_ID },
         { "100%", OPTION_VAL_100_O47_ID },
         { "110%", OPTION_VAL_110_O47_ID },
         { "120%", OPTION_VAL_120_O47_ID },
         { "130%", OPTION_VAL_130_O47_ID },
         { "140%", OPTION_VAL_140_O47_ID },
         { "150%", OPTION_VAL_150_O47_ID },
         { "160%", OPTION_VAL_160_O47_ID },
         { "170%", OPTION_VAL_170_O47_ID },
         { "180%", OPTION_VAL_180_O47_ID },
         { "190%", OPTION_VAL_190_O47_ID },
         { "200%", OPTION_VAL_200_O47_ID },
         { "210%", OPTION_VAL_210_O47_ID },
         { "220%", OPTION_VAL_220_O47_ID },
         { "230%", OPTION_VAL_230_O47_ID },
         { "240%", OPTION_VAL_240_O47_ID },
         { "250%", OPTION_VAL_250_O47_ID },
         { "260%", OPTION_VAL_260_O47_ID },
         { "270%", OPTION_VAL_270_O47_ID },
         { "280%", OPTION_VAL_280_O47_ID },
         { "290%", OPTION_VAL_290_O47_ID },
         { "300%", OPTION_VAL_300_O47_ID },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_ID,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ID },
         { "Red",      OPTION_VAL_RED_ID },
         { "Green",    OPTION_VAL_GREEN_ID },
         { "Blue",     OPTION_VAL_BLUE_ID },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_ID,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ID },
         { "Red",      OPTION_VAL_RED_ID },
         { "Green",    OPTION_VAL_GREEN_ID },
         { "Blue",     OPTION_VAL_BLUE_ID },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_ID,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ID },
         { "Red",      OPTION_VAL_RED_ID },
         { "Green",    OPTION_VAL_GREEN_ID },
         { "Blue",     OPTION_VAL_BLUE_ID },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_ID,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_ID },
         { "Red",      OPTION_VAL_RED_ID },
         { "Green",    OPTION_VAL_GREEN_ID },
         { "Blue",     OPTION_VAL_BLUE_ID },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_ID,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ID },
         { "Purupuru", OPTION_VAL_PURUPURU_ID },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_ID },
         { "None",     OPTION_VAL_NONE_ID },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_ID,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ID },
         { "Purupuru", OPTION_VAL_PURUPURU_ID },
         { "None",     OPTION_VAL_NONE_ID },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_ID,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ID },
         { "Purupuru", OPTION_VAL_PURUPURU_ID },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_ID },
         { "None",     OPTION_VAL_NONE_ID },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_ID,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ID },
         { "Purupuru", OPTION_VAL_PURUPURU_ID },
         { "None",     OPTION_VAL_NONE_ID },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_ID,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ID },
         { "Purupuru", OPTION_VAL_PURUPURU_ID },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_ID },
         { "None",     OPTION_VAL_NONE_ID },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_ID,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ID },
         { "Purupuru", OPTION_VAL_PURUPURU_ID },
         { "None",     OPTION_VAL_NONE_ID },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_ID,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ID },
         { "Purupuru", OPTION_VAL_PURUPURU_ID },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_ID },
         { "None",     OPTION_VAL_NONE_ID },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_ID,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_ID },
         { "Purupuru", OPTION_VAL_PURUPURU_ID },
         { "None",     OPTION_VAL_NONE_ID },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_ID,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_ID,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_ID,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_ID },
         { "All VMUs", OPTION_VAL_ALL_VMUS_ID },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_ID,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_ID,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_ID,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_ID,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_ID,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_ID,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_ID,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_ID,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ID },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ID },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ID },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ID },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ID },
         { "2x", OPTION_VAL_2_O26_ID },
         { "3x", OPTION_VAL_3X_ID },
         { "4x", OPTION_VAL_4_ID },
         { "5x", OPTION_VAL_5X_ID },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ID },
         { "20%",  OPTION_VAL_20_ID },
         { "30%",  OPTION_VAL_30_ID },
         { "40%",  OPTION_VAL_40_ID },
         { "50%",  OPTION_VAL_50_ID },
         { "60%",  OPTION_VAL_60_ID },
         { "70%",  OPTION_VAL_70_ID },
         { "80%",  OPTION_VAL_80_ID },
         { "90%",  OPTION_VAL_90_ID },
         { "100%", OPTION_VAL_100_O47_ID },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ID },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ID },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ID },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ID },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ID },
         { "2x", OPTION_VAL_2_O26_ID },
         { "3x", OPTION_VAL_3X_ID },
         { "4x", OPTION_VAL_4_ID },
         { "5x", OPTION_VAL_5X_ID },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ID },
         { "20%",  OPTION_VAL_20_ID },
         { "30%",  OPTION_VAL_30_ID },
         { "40%",  OPTION_VAL_40_ID },
         { "50%",  OPTION_VAL_50_ID },
         { "60%",  OPTION_VAL_60_ID },
         { "70%",  OPTION_VAL_70_ID },
         { "80%",  OPTION_VAL_80_ID },
         { "90%",  OPTION_VAL_90_ID },
         { "100%", OPTION_VAL_100_O47_ID },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ID },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ID },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ID },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ID },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ID },
         { "2x", OPTION_VAL_2_O26_ID },
         { "3x", OPTION_VAL_3X_ID },
         { "4x", OPTION_VAL_4_ID },
         { "5x", OPTION_VAL_5X_ID },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ID },
         { "20%",  OPTION_VAL_20_ID },
         { "30%",  OPTION_VAL_30_ID },
         { "40%",  OPTION_VAL_40_ID },
         { "50%",  OPTION_VAL_50_ID },
         { "60%",  OPTION_VAL_60_ID },
         { "70%",  OPTION_VAL_70_ID },
         { "80%",  OPTION_VAL_80_ID },
         { "90%",  OPTION_VAL_90_ID },
         { "100%", OPTION_VAL_100_O47_ID },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_ID },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_ID },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_ID },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_ID },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_ID },
         { "2x", OPTION_VAL_2_O26_ID },
         { "3x", OPTION_VAL_3X_ID },
         { "4x", OPTION_VAL_4_ID },
         { "5x", OPTION_VAL_5X_ID },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_ID },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_ID },
         { "BLACK 02",          OPTION_VAL_BLACK_02_ID },
         { "BLUE 03",           OPTION_VAL_BLUE_ID },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_ID },
         { "GREEN 05",          OPTION_VAL_GREEN_ID },
         { "CYAN 06",           OPTION_VAL_CYAN_06_ID },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_ID },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_ID },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_ID },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_ID },
         { "RED 11",            OPTION_VAL_RED_ID },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_ID },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_ID },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_ID },
         { "GRAY 15",           OPTION_VAL_GRAY_15_ID },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_ID },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_ID },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_ID },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_ID },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_ID },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_ID },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_ID },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_ID },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_ID },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_ID },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_ID },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_ID },
         { "WHITE 28",          OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_ID,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_ID },
         { "20%",  OPTION_VAL_20_ID },
         { "30%",  OPTION_VAL_30_ID },
         { "40%",  OPTION_VAL_40_ID },
         { "50%",  OPTION_VAL_50_ID },
         { "60%",  OPTION_VAL_60_ID },
         { "70%",  OPTION_VAL_70_ID },
         { "80%",  OPTION_VAL_80_ID },
         { "90%",  OPTION_VAL_90_ID },
         { "100%", OPTION_VAL_100_O47_ID },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_id = {
   option_cats_id,
   option_defs_id
};

/* RETRO_LANGUAGE_IT */

#define CATEGORY_SYSTEM_LABEL_IT "Sistema"
#define CATEGORY_SYSTEM_INFO_0_IT "Configurare la regione, la lingua, il BIOS e le impostazioni hardware di base."
#define CATEGORY_VIDEO_LABEL_IT NULL
#define CATEGORY_VIDEO_INFO_0_IT "Configurare la risoluzione, la trasparenza indipendente dall'ordine e le impostazioni degli effetti visivi."
#define CATEGORY_PERFORMANCE_LABEL_IT "Prestazioni"
#define CATEGORY_PERFORMANCE_INFO_0_IT "Configurare il rendering filettato e le impostazioni del salto di fotogramma."
#define CATEGORY_HACKS_LABEL_IT "Hack di Emulazione"
#define CATEGORY_HACKS_INFO_0_IT "Configura override widescreen, velocità di caricamento GD-ROM e impostazioni di sostituzione texture."
#define CATEGORY_INPUT_LABEL_IT NULL
#define CATEGORY_INPUT_INFO_0_IT "Configura impostazioni gamepad e pistola ottica."
#define CATEGORY_EXPANSIONS_LABEL_IT "Slot Di Espansione Controller"
#define CATEGORY_EXPANSIONS_INFO_0_IT "Selezionare il dispositivo (VMU, dispositivo rumble) collegato in ogni slot di espansione del regolatore."
#define CATEGORY_VMU_LABEL_IT "Unità Memoria Visiva"
#define CATEGORY_VMU_INFO_0_IT "Configura i file VMU per gioco e le impostazioni di visibilità VMU sullo schermo."
#define CORE_OPTION_NAME_REGION_LABEL_IT "Regione"
#define OPTION_VAL_JAPAN_IT "Giappone"
#define OPTION_VAL_USA_IT NULL
#define OPTION_VAL_EUROPE_IT "Europa"
#define OPTION_VAL_DEFAULT_IT "Predefinito"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_IT "Lingua"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_IT "Cambia la lingua usata dal BIOS e da qualsiasi gioco che contenga più lingue."
#define OPTION_VAL_JAPANESE_IT "Giapponese"
#define OPTION_VAL_ENGLISH_IT "Inglese"
#define OPTION_VAL_GERMAN_IT "Tedesco"
#define OPTION_VAL_FRENCH_IT "Francese"
#define OPTION_VAL_SPANISH_IT "Spagnolo"
#define OPTION_VAL_ITALIAN_IT "Italiano"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_IT "HLE BIOS (Riavvio Richiesto)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_IT "Forza l'uso di BIOS di emulazione di alto livello."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_IT "Abilita DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_IT "Abilita l'emulazione del DSP audio di Dreamcast (processore di segnale digitale). Migliora l'accuratezza del suono generato, ma aumenta i requisiti di prestazioni."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_IT "Consenti I Pulsanti del Servizio Arcade"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_IT "Abilita il pulsante SERVICE per i giochi arcade, per inserire le impostazioni del cabinato."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_IT "Imposta giochi NAOMI a Free Play"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_IT "Modificare le impostazioni del gioco per free play."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_IT "Emulazione Adattatore A banda larga"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_IT "Emulare l'adattatore a banda larga Ethernet invece del modem. (Riavvio Richiesto)"
#define CORE_OPTION_NAME_UPNP_LABEL_IT "Abilita UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_IT "Usa UPnP per configurare automaticamente il router Internet per i giochi online."
#define CORE_OPTION_NAME_DCNET_LABEL_IT "Usa DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_IT "Utilizzare il servizio cloud DCNet per l'accesso a Internet Dreamcast."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_IT "Risoluzione Interna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_IT "Modificare la risoluzione di rendering."
#define OPTION_VAL_320X240_IT "320x240 (Metà)"
#define OPTION_VAL_640X480_IT "640x480 (Nativo)"
#define OPTION_VAL_800X600_IT NULL
#define OPTION_VAL_960X720_IT NULL
#define OPTION_VAL_1024X768_IT NULL
#define OPTION_VAL_1280X960_IT NULL
#define OPTION_VAL_1440X1080_IT "1440x1080 (x2,25)"
#define OPTION_VAL_1600X1200_IT "1600x1200 (x2,5)"
#define OPTION_VAL_1920X1440_IT NULL
#define OPTION_VAL_2560X1920_IT NULL
#define OPTION_VAL_2880X2160_IT "2880x2160 (x4,5)"
#define OPTION_VAL_3200X2400_IT NULL
#define OPTION_VAL_3840X2880_IT NULL
#define OPTION_VAL_4480X3360_IT NULL
#define OPTION_VAL_5120X3840_IT NULL
#define OPTION_VAL_5760X4320_IT NULL
#define OPTION_VAL_6400X4800_IT NULL
#define OPTION_VAL_7040X5280_IT NULL
#define OPTION_VAL_7680X5760_IT NULL
#define OPTION_VAL_8320X6240_IT NULL
#define OPTION_VAL_8960X6720_IT NULL
#define OPTION_VAL_9600X7200_IT NULL
#define OPTION_VAL_10240X7680_IT NULL
#define OPTION_VAL_10880X8160_IT NULL
#define OPTION_VAL_11520X8640_IT NULL
#define OPTION_VAL_12160X9120_IT NULL
#define OPTION_VAL_12800X9600_IT NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_IT "Tipo Di Cavo"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_IT "Il tipo di segnale di uscita. 'TV (Composito)' è il più ampiamente supportato."
#define OPTION_VAL_VGA_IT NULL
#define OPTION_VAL_TV_RGB_IT NULL
#define OPTION_VAL_TV_COMPOSITE_IT "TV (Composto)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_IT "Standard di trasmissione"
#define OPTION_VAL_NTSC_IT NULL
#define OPTION_VAL_PAL_IT "PAL (Mondo)"
#define OPTION_VAL_PAL_N_IT NULL
#define OPTION_VAL_PAL_M_IT "PAL-M (Brasile)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_IT "Orientamento Schermo"
#define OPTION_VAL_HORIZONTAL_IT "Orizzontale"
#define OPTION_VAL_VERTICAL_IT "Verticale"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_IT "Ordinamento Alfa"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_IT "Seleziona come vengono ordinati i poligoni trasparenti."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_IT "Per-Strip (veloce, meno preciso)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_IT "Per-Triangolo (normale)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_IT "Per-Pixel (preciso, ma più lento)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_IT "Dimensione di accumulo del Buffer Pixel"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_IT "Potrebbero essere necessari valori più alti per ottenere risoluzioni più elevate in uscita correttamente."
#define OPTION_VAL_512MB_IT NULL
#define OPTION_VAL_1GB_IT NULL
#define OPTION_VAL_2GB_IT NULL
#define OPTION_VAL_4GB_IT NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_IT "Strati trasparenti massimi"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_IT "Potrebbero essere necessari valori più elevati per scene complesse."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_IT "Emulazione framebuffer completa"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_IT "Abilita l'emulazione completa del framebuffer in VRAM. Ciò è utile per i giochi che leggono o scrivono direttamente il framebuffer in VRAM. Se abilitata, la risoluzione interna è costretta a 640x480 e le prestazioni possono essere gravemente colpite."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_IT "Abilita Buffer RTT (Render To Texture)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_IT "Copia le texture renderizzate dalla GPU a VRAM. Questa opzione è normalmente abilitata per i giochi che lo richiedono. Se abilitata, il rendering delle texture è disabilitato e le prestazioni potrebbero essere influenzate."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_IT NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_IT "Quando abilitato le texture useranno la versione più piccola di se stesse quando appaiono più lontane, può aumentare le prestazioni e ridurre lo shimmering."
#define CORE_OPTION_NAME_FOG_LABEL_IT "Effetti Nebbia"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_IT "Modificatore del Volume"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_IT "Una funzionalità della GPU Dreamcast che viene tipicamente utilizzata dai giochi per disegnare ombre degli oggetti. Questo dovrebbe normalmente essere abilitato - l'impatto delle prestazioni di solito è minimo o trascurabile."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_IT "Filtro Anisotropico"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_IT "Migliorare la qualità delle texture su superfici che sono ad angoli obliqui rispetto alla fotocamera. Valori più alti sono più impegnativi per la GPU. Le modifiche a questa impostazione si applicano solo dopo il riavvio."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_IT "Filtro Texture"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_IT "La modalità di filtraggio delle texture da usare. Questo può essere usato per forzare una certa modalità di filtraggio delle texture su tutte le texture per ottenere un aspetto più nitido (o più liscio) di quello predefinito. Valori diversi da quelli predefiniti possono causare vari problemi di rendering. Le modifiche a questa impostazione si applicano solo dopo il riavvio."
#define OPTION_VAL_1_IT "Forza il Vicino più Vicino"
#define OPTION_VAL_2_IT "Forza Lineare"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_IT "Ritardo Scambio fotogrammi"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_IT "Utile per evitare di lampeggiare schermi o video glitchati. Sconsigliato su piattaforme lente."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_IT "Rileva Modifiche Frame Rate"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_IT "Notificare il frontend quando il frame rate interno cambia (ad esempio da 60 fps a 30 fps). Migliora il ritmo dei fotogrammi nei giochi che girano a un bloccato 30 fps o 20 fps, ma dovrebbe essere disabilitato per i giochi con frame rate sbloccati (instabili) (e.. Ecco il delfino, torneo irreale). Nota: non disponibile quando 'Auto Salta Frame' è abilitato."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_IT "Filtro Di Post-Elaborazione PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_IT "Post-elaborare l'immagine renderizzata per simulare effetti specifici alla GPU PowerVR2 e segnali video analogici."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_IT "Miglioramento Della Texture (xbrz)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_IT "Migliora la grafica grafica grafica a pixel 2D disegnata a mano. Dovrebbe essere usata solo con giochi in pixel 2D."
#define OPTION_VAL_2_O26_IT NULL
#define OPTION_VAL_4_IT NULL
#define OPTION_VAL_6_IT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_IT "Miglioramento Texture Dimensione Massima Filtrata"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_IT "Seleziona un valore di dimensione massima per una texture da aggiornare, se la dimensione della texture è superiore al valore selezionato, allora non verrà modificata."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_IT "Interpolazione Di Profondità Nativa"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_IT "Aiuta con la corruzione della texture e problemi di profondità sulle GPU AMD. Può anche aiutare le GPU Intel in alcuni casi."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_IT "Correzione Upscale Bleeding Edge"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_IT "Aiuta con il caso di sanguinamento della texture durante l'upscaling. Disabilitarlo può aiutare se i pixel sono deformazioni durante l'upscaling in giochi 2D (MVC2, CVS, KOF, ecc.)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_IT "Rendering Filettato"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_IT "Esegue la GPU e la CPU su diversi thread. Altamente raccomandato."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_IT "Auto Salto Fotogrammi"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_IT "Salta automaticamente i fotogrammi quando l'emulatore è lento. Nota: questa impostazione si applica solo quando il Rendering Threaded è abilitato."
#define OPTION_VAL_SOME_IT "Normale"
#define OPTION_VAL_MORE_IT "Massimo"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_IT "Salto fotogrammi"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_IT "Imposta il numero di fotogrammi da saltare tra ogni fotogramma visualizzato."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_IT "Trucchi Widescreen (Riavvio Richiesto)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_IT "Attiva i trucchi che permettono ad alcuni giochi di visualizzare in formato widescreen."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_IT "Hack Widescreen"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_IT "Disegnare la geometria al di fuori del normale rapporto di aspetto 4:3. Può produrre difetti grafici nelle aree rivelate."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_IT "Caricamento Veloce GD-ROM (inaccurato)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_IT "Accelera il caricamento di GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_IT NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_IT "Abilita 32MB RAM Mod per Dreamcast. Può influire sulla compatibilità"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_IT "CPU SH4 sotto/overclock"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_IT "Cambia l'orologio principale della CPU SH4 da 200 MHz predefinito. Underclocking può aiutare le piattaforme lente. L'overclocking può aumentare il frame rate per alcuni giochi. Usare con cautela."
#define OPTION_VAL_100_IT NULL
#define OPTION_VAL_110_IT NULL
#define OPTION_VAL_120_IT NULL
#define OPTION_VAL_130_IT NULL
#define OPTION_VAL_140_IT NULL
#define OPTION_VAL_150_IT NULL
#define OPTION_VAL_160_IT NULL
#define OPTION_VAL_170_IT NULL
#define OPTION_VAL_180_IT NULL
#define OPTION_VAL_190_IT NULL
#define OPTION_VAL_200_IT NULL
#define OPTION_VAL_210_IT NULL
#define OPTION_VAL_220_IT NULL
#define OPTION_VAL_230_IT NULL
#define OPTION_VAL_240_IT NULL
#define OPTION_VAL_250_IT NULL
#define OPTION_VAL_260_IT NULL
#define OPTION_VAL_270_IT NULL
#define OPTION_VAL_280_IT NULL
#define OPTION_VAL_290_IT NULL
#define OPTION_VAL_300_IT NULL
#define OPTION_VAL_310_IT NULL
#define OPTION_VAL_320_IT NULL
#define OPTION_VAL_330_IT NULL
#define OPTION_VAL_340_IT NULL
#define OPTION_VAL_350_IT NULL
#define OPTION_VAL_360_IT NULL
#define OPTION_VAL_370_IT NULL
#define OPTION_VAL_380_IT NULL
#define OPTION_VAL_390_IT NULL
#define OPTION_VAL_400_IT NULL
#define OPTION_VAL_410_IT NULL
#define OPTION_VAL_420_IT NULL
#define OPTION_VAL_430_IT NULL
#define OPTION_VAL_440_IT NULL
#define OPTION_VAL_450_IT NULL
#define OPTION_VAL_460_IT NULL
#define OPTION_VAL_470_IT NULL
#define OPTION_VAL_480_IT NULL
#define OPTION_VAL_490_IT NULL
#define OPTION_VAL_500_IT NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_IT "Carica Texture Personalizzate"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_IT "Carica texture personalizzate nella cartella 'system/dc/textures/<game-id>/'."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_IT "Precarica texture personalizzate"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_IT "Precarica texture personalizzate all'avvio del gioco. Può migliorare le prestazioni, ma aumenta l'utilizzo della memoria."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_IT "Scarica Texture"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_IT "Ogni volta che un nuovo texture viene utilizzato dal gioco, verrà salvato come file .png nella cartella 'system/dc/texdump/<game-id>/'."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_IT "Scarica Texture sostituite"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_IT "Scarica sempre le texture che sono già sostituite da texture personalizzate."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_IT "Zona morta dello sticl analogico"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_IT "Seleziona fino a che punto devi spingere la levetta analogica prima che inizi ad essere elaborata."
#define OPTION_VAL_0_IT NULL
#define OPTION_VAL_5_IT NULL
#define OPTION_VAL_10_IT NULL
#define OPTION_VAL_15_IT NULL
#define OPTION_VAL_20_IT NULL
#define OPTION_VAL_25_IT NULL
#define OPTION_VAL_30_IT NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_IT "Zona morta dei trigger"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_IT "Selezionare quanto è necessario premere il tasto prima di iniziare l'elaborazione."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_IT "Trigger Digitali"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_IT "Quando abilitato, i tasti agiranno come pulsanti regolari, il che significa che saranno elaborati come o completamente premuto o non premuto affatto, senza alcun in-tra."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_IT "Uscite di trasmissione digitale"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_IT "Trasmissione di uscite digitali e stato di forza-feedback sulla porta TCP 8000. Compatibile con l'opzione MAME \"-output network\"."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_IT "Mostra Impostazioni Pistola Ottica"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_IT "Abilita la configurazione delle opzioni di visualizzazione di una pistola leggera. NOTA: potrebbe essere necessario attivare il menu rapido perché questa impostazione abbia effetto."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_IT "Scala Dimensione del Mirino della Lightgun"
#define OPTION_VAL_50_IT NULL
#define OPTION_VAL_60_IT NULL
#define OPTION_VAL_70_IT NULL
#define OPTION_VAL_80_IT NULL
#define OPTION_VAL_90_IT NULL
#define OPTION_VAL_100_O47_IT NULL
#define OPTION_VAL_110_O47_IT NULL
#define OPTION_VAL_120_O47_IT NULL
#define OPTION_VAL_130_O47_IT NULL
#define OPTION_VAL_140_O47_IT NULL
#define OPTION_VAL_150_O47_IT NULL
#define OPTION_VAL_160_O47_IT NULL
#define OPTION_VAL_170_O47_IT NULL
#define OPTION_VAL_180_O47_IT NULL
#define OPTION_VAL_190_O47_IT NULL
#define OPTION_VAL_200_O47_IT NULL
#define OPTION_VAL_210_O47_IT NULL
#define OPTION_VAL_220_O47_IT NULL
#define OPTION_VAL_230_O47_IT NULL
#define OPTION_VAL_240_O47_IT NULL
#define OPTION_VAL_250_O47_IT NULL
#define OPTION_VAL_260_O47_IT NULL
#define OPTION_VAL_270_O47_IT NULL
#define OPTION_VAL_280_O47_IT NULL
#define OPTION_VAL_290_O47_IT NULL
#define OPTION_VAL_300_O47_IT NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_IT "Mostra mirino della Pistola 1"
#define OPTION_VAL_WHITE_IT "Bianco"
#define OPTION_VAL_RED_IT "Rosso"
#define OPTION_VAL_GREEN_IT "Verde"
#define OPTION_VAL_BLUE_IT "Blu"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_IT "Mostra mirino della Pistola 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_IT "Mostra mirino della Pistola 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_IT "Mostra mirino della Pistola 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_IT "Dispositivo nello slot di espansione A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_IT "Selezionare il dispositivo collegato allo slot di espansione A1 (porta A slot 1)."
#define OPTION_VAL_VMU_IT NULL
#define OPTION_VAL_PURUPURU_IT NULL
#define OPTION_VAL_DREAMPOTATO_IT NULL
#define OPTION_VAL_NONE_IT "Nessuno"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_IT "Dispositivo nello slot di espansione A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_IT "Selezionare il dispositivo collegato allo slot di espansione A2 (porta A slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_IT "Dispositivo nello slot di espansione B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_IT "Selezionare il dispositivo collegato allo slot di espansione B1 (porta B slot 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_IT "Dispositivo nello slot di espansione B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_IT "Selezionare il dispositivo collegato allo slot di espansione B2 (porta B slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_IT "Dispositivo nello slot di espansione C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_IT "Selezionare il dispositivo collegato allo slot di espansione C1 (porta C slot 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_IT "Dispositivo nello slot di espansione C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_IT "Selezionare il dispositivo collegato allo slot di espansione C2 (porta C slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_IT "Dispositivo nello slot di espansione D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_IT "Selezionare il dispositivo collegato allo slot di espansione D1 (porta D slot 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_IT "Dispositivo nello slot di espansione D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_IT "Selezionare il dispositivo collegato allo slot di espansione D2 (porta D slot 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_IT "Visual Memory Units/Sistemi per Gioco (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_IT "VMU per gioco"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_IT "Se disabilitato, tutti i giochi condividono fino a 8 file VMU di salvataggio (A1/A2/B1/B2/C1/C2/D1/D2) situati nella cartella di sistema di RetroArch.\nL'impostazione 'VMU A1' crea un file VMU 'A1' unico nella cartella di salvataggio di RetroArch, per ogni gioco che viene lanciato.\nL'impostazione 'All VMUs' crea fino a 8 file VMU unici (A1/A2/B1/B2/C1/C2/D1/D2) per ogni gioco che viene lanciato."
#define OPTION_VAL_VMU_A1_IT NULL
#define OPTION_VAL_ALL_VMUS_IT "Tutte Le VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_IT "Suoni Delle Unità Di Memoria Visiva/Sistemi (VMU)"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_IT "Suoni VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_IT "Se abilitato, vengono riprodotti i segnali acustici VMU."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_IT "Capienza DreamPotato VMU"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_IT "Abilita l'archiviazione dei dati VMU sull'emulatore DreamPotato connesso. Quando abilitato, il file di salvataggio VMU locale non viene utilizzato."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_IT "Mostra Le Impostazioni Di Visualizzazione della Visual Memory Unit/System (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_IT "Mostra Impostazioni Di Visualizzazione VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_IT "Abilita la configurazione dello schermo LCD VMU emulato, la dimensione, la posizione e il colore. NOTA: potrebbe essere necessario attivare il menu rapido affinché questa impostazione abbia effetto."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_IT "Visualizzazione Schermo VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_IT "Posizione VMU Schermo 1"
#define OPTION_VAL_UPPER_LEFT_IT "In Alto a Sinistra"
#define OPTION_VAL_UPPER_RIGHT_IT "In Alto a Destra"
#define OPTION_VAL_LOWER_LEFT_IT "In Basso a Sinistra"
#define OPTION_VAL_LOWER_RIGHT_IT "In Basso a Destra"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_IT "Dimensione VMU Schermo 1"
#define OPTION_VAL_1X_IT NULL
#define OPTION_VAL_3X_IT NULL
#define OPTION_VAL_5X_IT NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_IT "Schermo VMU 1 Pixel Su Colore"
#define OPTION_VAL_DEFAULT_ON_00_IT "Predefinito ON"
#define OPTION_VAL_DEFAULT_OFF_01_IT "Predefinito OFF"
#define OPTION_VAL_BLACK_02_IT "Nero"
#define OPTION_VAL_LIGHT_BLUE_04_IT "Azzurro"
#define OPTION_VAL_CYAN_06_IT "Ciano"
#define OPTION_VAL_CYAN_BLUE_07_IT "Blu Ciano"
#define OPTION_VAL_LIGHT_GREEN_08_IT "Verde Chiaro"
#define OPTION_VAL_CYAN_GREEN_09_IT "Verde Ciano"
#define OPTION_VAL_LIGHT_CYAN_10_IT "Ciano Chiaro"
#define OPTION_VAL_PURPLE_12_IT "Viola"
#define OPTION_VAL_LIGHT_PURPLE_13_IT "Viola Chiaro"
#define OPTION_VAL_YELLOW_14_IT "Giallo"
#define OPTION_VAL_GRAY_15_IT "Grigio"
#define OPTION_VAL_LIGHT_PURPLE_2_16_IT "Viola Chiaro (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_IT "Verde Chiaro (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_IT "Verde Chiaro (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_IT "Ciano Chiaro (2)"
#define OPTION_VAL_LIGHT_RED_2_20_IT "Rosso Chiaro (2)"
#define OPTION_VAL_MAGENTA_21_IT NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_IT "Viola Chiaro (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_IT "Arancione Chiaro"
#define OPTION_VAL_ORANGE_24_IT "Arancione"
#define OPTION_VAL_LIGHT_PURPLE_4_25_IT "Viola Chiaro (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_IT "Giallo Chiaro"
#define OPTION_VAL_LIGHT_YELLOW_2_27_IT "Giallo Chiaro (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_IT "Colore Vmu Schermo 1 Pixel Off"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_IT "Opacità VMU Screen 1"
#define OPTION_VAL_40_IT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_IT "Visualizzazione Schermo VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_IT "Posizione VMU Schermo 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_IT "Dimensione VMU Schermo 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_IT "Colore VMU Schermo 2 Pixel On"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_IT "Colore VMU Schermo 2 Pixel Off"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_IT "Opacità VMU Schermo 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_IT "Visualizzazione Schermo 3 VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_IT "Posizione Schermo 3 VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_IT "Dimensione Schermo 3 VMU"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_IT "Colore su Pixel Schermo VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_IT "Colore Pixel Off Schermo VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_IT "Opacità Schermo VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_IT "Visualizzazione Schermo VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_IT "Posizione Schermo VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_IT "Dimensione Schermo VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_IT "Colore Pixel On Schermo VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_IT "Colore Pixel Off Schermo VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_IT "Opacità Schermo VMU 4"

struct retro_core_option_v2_category option_cats_it[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_IT,
      CATEGORY_SYSTEM_INFO_0_IT
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_IT,
      CATEGORY_VIDEO_INFO_0_IT
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_IT,
      CATEGORY_PERFORMANCE_INFO_0_IT
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_IT,
      CATEGORY_HACKS_INFO_0_IT
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_IT,
      CATEGORY_INPUT_INFO_0_IT
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_IT,
      CATEGORY_EXPANSIONS_INFO_0_IT
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_IT,
      CATEGORY_VMU_INFO_0_IT
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_it[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_IT,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_IT },
         { "USA",     OPTION_VAL_USA_IT },
         { "Europe",  OPTION_VAL_EUROPE_IT },
         { "Default", OPTION_VAL_DEFAULT_IT },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_IT,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_IT },
         { "English",  OPTION_VAL_ENGLISH_IT },
         { "German",   OPTION_VAL_GERMAN_IT },
         { "French",   OPTION_VAL_FRENCH_IT },
         { "Spanish",  OPTION_VAL_SPANISH_IT },
         { "Italian",  OPTION_VAL_ITALIAN_IT },
         { "Default",  OPTION_VAL_DEFAULT_IT },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_IT,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_IT },
         { "640x480",    OPTION_VAL_640X480_IT },
         { "800x600",    OPTION_VAL_800X600_IT },
         { "960x720",    OPTION_VAL_960X720_IT },
         { "1024x768",   OPTION_VAL_1024X768_IT },
         { "1280x960",   OPTION_VAL_1280X960_IT },
         { "1440x1080",  OPTION_VAL_1440X1080_IT },
         { "1600x1200",  OPTION_VAL_1600X1200_IT },
         { "1920x1440",  OPTION_VAL_1920X1440_IT },
         { "2560x1920",  OPTION_VAL_2560X1920_IT },
         { "2880x2160",  OPTION_VAL_2880X2160_IT },
         { "3200x2400",  OPTION_VAL_3200X2400_IT },
         { "3840x2880",  OPTION_VAL_3840X2880_IT },
         { "4480x3360",  OPTION_VAL_4480X3360_IT },
         { "5120x3840",  OPTION_VAL_5120X3840_IT },
         { "5760x4320",  OPTION_VAL_5760X4320_IT },
         { "6400x4800",  OPTION_VAL_6400X4800_IT },
         { "7040x5280",  OPTION_VAL_7040X5280_IT },
         { "7680x5760",  OPTION_VAL_7680X5760_IT },
         { "8320x6240",  OPTION_VAL_8320X6240_IT },
         { "8960x6720",  OPTION_VAL_8960X6720_IT },
         { "9600x7200",  OPTION_VAL_9600X7200_IT },
         { "10240x7680", OPTION_VAL_10240X7680_IT },
         { "10880x8160", OPTION_VAL_10880X8160_IT },
         { "11520x8640", OPTION_VAL_11520X8640_IT },
         { "12160x9120", OPTION_VAL_12160X9120_IT },
         { "12800x9600", OPTION_VAL_12800X9600_IT },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_IT,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_IT },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_IT },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_IT },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_IT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_IT },
         { "PAL",     OPTION_VAL_PAL_IT },
         { "PAL_N",   OPTION_VAL_PAL_N_IT },
         { "PAL_M",   OPTION_VAL_PAL_M_IT },
         { "Default", OPTION_VAL_DEFAULT_IT },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_IT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_IT },
         { "vertical",   OPTION_VAL_VERTICAL_IT },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_IT,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_IT },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_IT },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_IT },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_IT,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_IT },
         { "1GB",   OPTION_VAL_1GB_IT },
         { "2GB",   OPTION_VAL_2GB_IT },
         { "4GB",   OPTION_VAL_4GB_IT },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_IT,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_IT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_IT,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_IT,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_IT },
         { "1",  OPTION_VAL_1_IT },
         { "2",  OPTION_VAL_2_IT },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_IT,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_IT },
         { "4",  OPTION_VAL_4_IT },
         { "6",  OPTION_VAL_6_IT },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_IT,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_IT,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_IT,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_IT,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_IT,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_IT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_IT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_IT },
         { "more", OPTION_VAL_MORE_IT },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_IT,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_IT },
         { "110", OPTION_VAL_110_IT },
         { "120", OPTION_VAL_120_IT },
         { "130", OPTION_VAL_130_IT },
         { "140", OPTION_VAL_140_IT },
         { "150", OPTION_VAL_150_IT },
         { "160", OPTION_VAL_160_IT },
         { "170", OPTION_VAL_170_IT },
         { "180", OPTION_VAL_180_IT },
         { "190", OPTION_VAL_190_IT },
         { "200", OPTION_VAL_200_IT },
         { "210", OPTION_VAL_210_IT },
         { "220", OPTION_VAL_220_IT },
         { "230", OPTION_VAL_230_IT },
         { "240", OPTION_VAL_240_IT },
         { "250", OPTION_VAL_250_IT },
         { "260", OPTION_VAL_260_IT },
         { "270", OPTION_VAL_270_IT },
         { "280", OPTION_VAL_280_IT },
         { "290", OPTION_VAL_290_IT },
         { "300", OPTION_VAL_300_IT },
         { "310", OPTION_VAL_310_IT },
         { "320", OPTION_VAL_320_IT },
         { "330", OPTION_VAL_330_IT },
         { "340", OPTION_VAL_340_IT },
         { "350", OPTION_VAL_350_IT },
         { "360", OPTION_VAL_360_IT },
         { "370", OPTION_VAL_370_IT },
         { "380", OPTION_VAL_380_IT },
         { "390", OPTION_VAL_390_IT },
         { "400", OPTION_VAL_400_IT },
         { "410", OPTION_VAL_410_IT },
         { "420", OPTION_VAL_420_IT },
         { "430", OPTION_VAL_430_IT },
         { "440", OPTION_VAL_440_IT },
         { "450", OPTION_VAL_450_IT },
         { "460", OPTION_VAL_460_IT },
         { "470", OPTION_VAL_470_IT },
         { "480", OPTION_VAL_480_IT },
         { "490", OPTION_VAL_490_IT },
         { "500", OPTION_VAL_500_IT },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_IT,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_IT },
         { "5%",  OPTION_VAL_5_IT },
         { "10%", OPTION_VAL_10_IT },
         { "15%", OPTION_VAL_15_IT },
         { "20%", OPTION_VAL_20_IT },
         { "25%", OPTION_VAL_25_IT },
         { "30%", OPTION_VAL_30_IT },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_IT,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_IT },
         { "5%",  OPTION_VAL_5_IT },
         { "10%", OPTION_VAL_10_IT },
         { "15%", OPTION_VAL_15_IT },
         { "20%", OPTION_VAL_20_IT },
         { "25%", OPTION_VAL_25_IT },
         { "30%", OPTION_VAL_30_IT },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_IT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_IT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_IT,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_IT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_IT },
         { "60%",  OPTION_VAL_60_IT },
         { "70%",  OPTION_VAL_70_IT },
         { "80%",  OPTION_VAL_80_IT },
         { "90%",  OPTION_VAL_90_IT },
         { "100%", OPTION_VAL_100_O47_IT },
         { "110%", OPTION_VAL_110_O47_IT },
         { "120%", OPTION_VAL_120_O47_IT },
         { "130%", OPTION_VAL_130_O47_IT },
         { "140%", OPTION_VAL_140_O47_IT },
         { "150%", OPTION_VAL_150_O47_IT },
         { "160%", OPTION_VAL_160_O47_IT },
         { "170%", OPTION_VAL_170_O47_IT },
         { "180%", OPTION_VAL_180_O47_IT },
         { "190%", OPTION_VAL_190_O47_IT },
         { "200%", OPTION_VAL_200_O47_IT },
         { "210%", OPTION_VAL_210_O47_IT },
         { "220%", OPTION_VAL_220_O47_IT },
         { "230%", OPTION_VAL_230_O47_IT },
         { "240%", OPTION_VAL_240_O47_IT },
         { "250%", OPTION_VAL_250_O47_IT },
         { "260%", OPTION_VAL_260_O47_IT },
         { "270%", OPTION_VAL_270_O47_IT },
         { "280%", OPTION_VAL_280_O47_IT },
         { "290%", OPTION_VAL_290_O47_IT },
         { "300%", OPTION_VAL_300_O47_IT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_IT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_IT },
         { "Red",      OPTION_VAL_RED_IT },
         { "Green",    OPTION_VAL_GREEN_IT },
         { "Blue",     OPTION_VAL_BLUE_IT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_IT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_IT },
         { "Red",      OPTION_VAL_RED_IT },
         { "Green",    OPTION_VAL_GREEN_IT },
         { "Blue",     OPTION_VAL_BLUE_IT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_IT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_IT },
         { "Red",      OPTION_VAL_RED_IT },
         { "Green",    OPTION_VAL_GREEN_IT },
         { "Blue",     OPTION_VAL_BLUE_IT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_IT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_IT },
         { "Red",      OPTION_VAL_RED_IT },
         { "Green",    OPTION_VAL_GREEN_IT },
         { "Blue",     OPTION_VAL_BLUE_IT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_IT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_IT },
         { "Purupuru", OPTION_VAL_PURUPURU_IT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_IT },
         { "None",     OPTION_VAL_NONE_IT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_IT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_IT },
         { "Purupuru", OPTION_VAL_PURUPURU_IT },
         { "None",     OPTION_VAL_NONE_IT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_IT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_IT },
         { "Purupuru", OPTION_VAL_PURUPURU_IT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_IT },
         { "None",     OPTION_VAL_NONE_IT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_IT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_IT },
         { "Purupuru", OPTION_VAL_PURUPURU_IT },
         { "None",     OPTION_VAL_NONE_IT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_IT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_IT },
         { "Purupuru", OPTION_VAL_PURUPURU_IT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_IT },
         { "None",     OPTION_VAL_NONE_IT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_IT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_IT },
         { "Purupuru", OPTION_VAL_PURUPURU_IT },
         { "None",     OPTION_VAL_NONE_IT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_IT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_IT },
         { "Purupuru", OPTION_VAL_PURUPURU_IT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_IT },
         { "None",     OPTION_VAL_NONE_IT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_IT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_IT },
         { "Purupuru", OPTION_VAL_PURUPURU_IT },
         { "None",     OPTION_VAL_NONE_IT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_IT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_IT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_IT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_IT },
         { "All VMUs", OPTION_VAL_ALL_VMUS_IT },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_IT,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_IT,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_IT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_IT,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_IT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_IT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_IT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_IT,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_IT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_IT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_IT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_IT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_IT },
         { "2x", OPTION_VAL_2_O26_IT },
         { "3x", OPTION_VAL_3X_IT },
         { "4x", OPTION_VAL_4_IT },
         { "5x", OPTION_VAL_5X_IT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_IT },
         { "20%",  OPTION_VAL_20_IT },
         { "30%",  OPTION_VAL_30_IT },
         { "40%",  OPTION_VAL_40_IT },
         { "50%",  OPTION_VAL_50_IT },
         { "60%",  OPTION_VAL_60_IT },
         { "70%",  OPTION_VAL_70_IT },
         { "80%",  OPTION_VAL_80_IT },
         { "90%",  OPTION_VAL_90_IT },
         { "100%", OPTION_VAL_100_O47_IT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_IT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_IT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_IT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_IT },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_IT },
         { "2x", OPTION_VAL_2_O26_IT },
         { "3x", OPTION_VAL_3X_IT },
         { "4x", OPTION_VAL_4_IT },
         { "5x", OPTION_VAL_5X_IT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_IT },
         { "20%",  OPTION_VAL_20_IT },
         { "30%",  OPTION_VAL_30_IT },
         { "40%",  OPTION_VAL_40_IT },
         { "50%",  OPTION_VAL_50_IT },
         { "60%",  OPTION_VAL_60_IT },
         { "70%",  OPTION_VAL_70_IT },
         { "80%",  OPTION_VAL_80_IT },
         { "90%",  OPTION_VAL_90_IT },
         { "100%", OPTION_VAL_100_O47_IT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_IT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_IT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_IT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_IT },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_IT },
         { "2x", OPTION_VAL_2_O26_IT },
         { "3x", OPTION_VAL_3X_IT },
         { "4x", OPTION_VAL_4_IT },
         { "5x", OPTION_VAL_5X_IT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_IT },
         { "20%",  OPTION_VAL_20_IT },
         { "30%",  OPTION_VAL_30_IT },
         { "40%",  OPTION_VAL_40_IT },
         { "50%",  OPTION_VAL_50_IT },
         { "60%",  OPTION_VAL_60_IT },
         { "70%",  OPTION_VAL_70_IT },
         { "80%",  OPTION_VAL_80_IT },
         { "90%",  OPTION_VAL_90_IT },
         { "100%", OPTION_VAL_100_O47_IT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_IT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_IT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_IT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_IT },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_IT },
         { "2x", OPTION_VAL_2_O26_IT },
         { "3x", OPTION_VAL_3X_IT },
         { "4x", OPTION_VAL_4_IT },
         { "5x", OPTION_VAL_5X_IT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_IT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_IT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_IT },
         { "BLUE 03",           OPTION_VAL_BLUE_IT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_IT },
         { "GREEN 05",          OPTION_VAL_GREEN_IT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_IT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_IT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_IT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_IT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_IT },
         { "RED 11",            OPTION_VAL_RED_IT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_IT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_IT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_IT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_IT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_IT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_IT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_IT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_IT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_IT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_IT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_IT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_IT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_IT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_IT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_IT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_IT },
         { "WHITE 28",          OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_IT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_IT },
         { "20%",  OPTION_VAL_20_IT },
         { "30%",  OPTION_VAL_30_IT },
         { "40%",  OPTION_VAL_40_IT },
         { "50%",  OPTION_VAL_50_IT },
         { "60%",  OPTION_VAL_60_IT },
         { "70%",  OPTION_VAL_70_IT },
         { "80%",  OPTION_VAL_80_IT },
         { "90%",  OPTION_VAL_90_IT },
         { "100%", OPTION_VAL_100_O47_IT },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_it = {
   option_cats_it,
   option_defs_it
};

/* RETRO_LANGUAGE_JA */

#define CATEGORY_SYSTEM_LABEL_JA "システム"
#define CATEGORY_SYSTEM_INFO_0_JA "地域、言語、BIOS および基本的なハードウェアの設定を変更します。"
#define CATEGORY_VIDEO_LABEL_JA "ビデオ"
#define CATEGORY_VIDEO_INFO_0_JA "解像度、描画順に依存しない透明度および視覚効果の設定を変更します。"
#define CATEGORY_PERFORMANCE_LABEL_JA "パフォーマンス"
#define CATEGORY_PERFORMANCE_INFO_0_JA "スレッド化したレンダリングとフレームスキップの設定を変更します。"
#define CATEGORY_HACKS_LABEL_JA "エミュレーションハック"
#define CATEGORY_HACKS_INFO_0_JA "ワイドスクリーンハック、GD-ROM 読み込み速度およびテクスチャの置換設定を変更します。"
#define CATEGORY_INPUT_LABEL_JA "入力"
#define CATEGORY_INPUT_INFO_0_JA "ゲームパッドとライトガンの設定を変更します。"
#define CATEGORY_EXPANSIONS_LABEL_JA "コントローラー拡張スロット"
#define CATEGORY_EXPANSIONS_INFO_0_JA "各コントローラー拡張スロットに接続されるデバイス (VMU、振動デバイス) を選択します。"
#define CATEGORY_VMU_LABEL_JA "ビジュアルメモリユニット"
#define CATEGORY_VMU_INFO_0_JA "ゲームごとの VMU セーブファイルと OSD VMU の表示設定を変更します。"
#define CORE_OPTION_NAME_REGION_LABEL_JA "地域"
#define OPTION_VAL_JAPAN_JA "日本"
#define OPTION_VAL_USA_JA "アメリカ合衆国"
#define OPTION_VAL_EUROPE_JA "ヨーロッパ"
#define OPTION_VAL_DEFAULT_JA "デフォルト"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_JA "言語"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_JA "BIOS や複数の言語を含むゲームで使用される言語を変更します。"
#define OPTION_VAL_JAPANESE_JA "日本語"
#define OPTION_VAL_ENGLISH_JA "英語"
#define OPTION_VAL_GERMAN_JA "ドイツ語"
#define OPTION_VAL_FRENCH_JA "フランス語"
#define OPTION_VAL_SPANISH_JA "スペイン語"
#define OPTION_VAL_ITALIAN_JA "イタリア語"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_JA "HLE BIOS (再起動が必要)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_JA "高レベルエミュレーション BIOS の使用を強制します。"
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_JA "DSP を有効にする"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_JA "ドリームキャストのオーディオ DSP のエミュレーションを有効にします。生成されるサウンドの精度を向上させますが、パフォーマンス要件が増加します。"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_JA "アーケードサービスボタンを許可"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_JA "アーケードゲームでサービスボタンを有効にし、キャビネット設定を行います。"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_JA "NAOMI ゲームを無料プレイにする"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_JA "無料プレイを行うためにゲームのコイン設定を変更します。"
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_JA "ブロードバンドアダプタエミュレーション"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_JA NULL
#define CORE_OPTION_NAME_UPNP_LABEL_JA "UPnP を有効にする"
#define CORE_OPTION_NAME_UPNP_INFO_0_JA "オンラインゲーム用にルーターの UPnP を自動的に設定し使用します。"
#define CORE_OPTION_NAME_DCNET_LABEL_JA NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_JA NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_JA "内部解像度"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_JA "描画解像度を変更します。"
#define OPTION_VAL_320X240_JA "320x240 (半分)"
#define OPTION_VAL_640X480_JA "640x480 (ネイティブ)"
#define OPTION_VAL_800X600_JA NULL
#define OPTION_VAL_960X720_JA NULL
#define OPTION_VAL_1024X768_JA NULL
#define OPTION_VAL_1280X960_JA NULL
#define OPTION_VAL_1440X1080_JA NULL
#define OPTION_VAL_1600X1200_JA NULL
#define OPTION_VAL_1920X1440_JA NULL
#define OPTION_VAL_2560X1920_JA NULL
#define OPTION_VAL_2880X2160_JA NULL
#define OPTION_VAL_3200X2400_JA NULL
#define OPTION_VAL_3840X2880_JA NULL
#define OPTION_VAL_4480X3360_JA NULL
#define OPTION_VAL_5120X3840_JA NULL
#define OPTION_VAL_5760X4320_JA NULL
#define OPTION_VAL_6400X4800_JA NULL
#define OPTION_VAL_7040X5280_JA "7040x580 (x11)"
#define OPTION_VAL_7680X5760_JA NULL
#define OPTION_VAL_8320X6240_JA NULL
#define OPTION_VAL_8960X6720_JA NULL
#define OPTION_VAL_9600X7200_JA NULL
#define OPTION_VAL_10240X7680_JA NULL
#define OPTION_VAL_10880X8160_JA NULL
#define OPTION_VAL_11520X8640_JA NULL
#define OPTION_VAL_12160X9120_JA NULL
#define OPTION_VAL_12800X9600_JA NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_JA "ケーブルの種類"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_JA "出力信号の種類を選択します。[TV (コンポジット)] が最も広くサポートされています。"
#define OPTION_VAL_VGA_JA NULL
#define OPTION_VAL_TV_RGB_JA NULL
#define OPTION_VAL_TV_COMPOSITE_JA "TV (コンポジット)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_JA "放送方式"
#define OPTION_VAL_NTSC_JA NULL
#define OPTION_VAL_PAL_JA "PAL (世界)"
#define OPTION_VAL_PAL_N_JA "PAL-N (アルゼンチン, パラグアイ, ウルグアイ)"
#define OPTION_VAL_PAL_M_JA "PAL-M (ブラジル)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_JA "画面の向き"
#define OPTION_VAL_HORIZONTAL_JA "水平"
#define OPTION_VAL_VERTICAL_JA "垂直"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_JA "半透明ソート"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_JA "半透明ポリゴンがどのように並べ替えられるかを選択します。"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_JA "ストリップ単位 (低精度, 高速)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_JA "トライアングル単位 (通常)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_JA "ピクセル単位 (高精度, 低速)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_JA "累積ピクセルバッファサイズ"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_JA "より高い解像度で正しく出力するには、より高い値が必要になる可能性があります。"
#define OPTION_VAL_512MB_JA NULL
#define OPTION_VAL_1GB_JA NULL
#define OPTION_VAL_2GB_JA NULL
#define OPTION_VAL_4GB_JA NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_JA "最大透明レイヤー数"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_JA "複雑なシーンでは、より高い値が必要になる可能性があります。"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_JA "完全なフレームバッファエミュレーション"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_JA "VRAM で完全なフレームバッファエミュレーションを有効にします。フレームバッファを VRAM で直接読み書きするゲームに便利です。有効にすると、内部解像度は 640x480 に固定され、実行速度に深刻な影響を与える可能性があります。"
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_JA "RTT (テクスチャへのレンダリング) バッファを有効にする"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_JA "レンダリングされたテクスチャを GPU から VRAM にコピーします。このオプションは通常、これを必要とするゲームで有効になります。有効にすると、テクスチャアップスケーリングが無効になり、実行速度に影響を与える可能性があります。"
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_JA "ミップマッピング"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_JA "有効にすると、テクスチャが遠くに表示される際に小さいバージョンが使用されます。パフォーマンスが向上し、チラつきが減少します。"
#define CORE_OPTION_NAME_FOG_LABEL_JA "フォグエフェクト"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_JA "ボリュームモディファイア"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_JA "オブジェクトの影を描画するためにゲームで一般的に使用されるドリームキャスト GPU の機能です。通常は有効にする必要があります。パフォーマンスへの影響は最小限か無視できる程度です。"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_JA "異方性フィルタリング"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_JA "カメラに対して斜めの視野角にあるテクスチャ表面の品質を向上させます。値を高くするほど GPU への負荷が大きくなります。この設定を反映するには再起動が必要です。"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_JA "テクスチャフィルタリング"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_JA "使用するテクスチャフィルタリングモードを選択します。すべてのテクスチャに特定のフィルタリングモードを強制することで、デフォルトよりも鮮明 (または滑らかな) 外観を得るために使用できます。デフォルト以外の値はレンダリングに問題が発生する可能性があります。この設定を反映するには再起動が必要です。"
#define OPTION_VAL_1_JA "強制ニアレストネイバー"
#define OPTION_VAL_2_JA "強制リニア"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_JA "フレーム入れ替え遅延"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_JA "画面の点滅やビデオの描画バグを回避するために便利です。遅いプラットフォームではお勧めしません。"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_JA "フレームレートの変更を検出"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_JA "内部フレームレートが変更された際にフロントエンドに通知します (例: 60fps から 30fps)。30fps または 20fps の固定フレームレートで実行されるゲームでフレームペーシングを改善します。ゲームが固定フレームレートでない場合は無効にする必要があります。注意: 自動フレームスキップが有効の場合は使用できません。"
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_JA "PowerVR2 ポストプロセスフィルター"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_JA "レンダリングされた画像にポストプロセスを適用し、PowerVR2 GPU およびアナログビデオ信号に特定の効果をシミュレートします。"
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_JA "テクスチャアップスケーリング (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_JA "手書きの 2D ピクセルアートグラフィックを強化します。2D ピクセルアートのゲームでのみ使用することをお勧めします。"
#define OPTION_VAL_2_O26_JA NULL
#define OPTION_VAL_4_JA NULL
#define OPTION_VAL_6_JA NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_JA "テクスチャアップスケーリング最大フィルターサイズ"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_JA "アップスケーリングされるテクスチャの最大サイズ値を選択します。テクスチャサイズが選択された値より大きい場合はアップスケーリングされなくなります。"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_JA "ネイティブ深度補完"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_JA "AMD GPU でテクスチャの破損や深度の問題を改善します。場合によっては インテル GPU でを改善することもできます。"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_JA "アップスケーリング滲み修正"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_JA "アップスケーリング時のテクスチャの滲みを改善します。2D ゲームのアップスケーリング時にピクセルが歪む場合、無効にすることで解決できます。"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_JA "スレッド化されたレンダリング"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_JA "GPU と CPU を別のスレッドで実行します。強く推奨されます。"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_JA "自動フレームスキップ"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_JA "エミュレータの動作が遅い場合、自動的にフレームをスキップします。注意: この設定は [レンダリングのスレッド化] が有効な場合にのみ適用されます。"
#define OPTION_VAL_SOME_JA "通常"
#define OPTION_VAL_MORE_JA "最大"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_JA "フレームスキップ"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_JA "表示される各フレーム間でスキップするフレーム数を設定します。"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_JA "ワイドスクリーンチート (再起動が必要)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_JA "特定のゲームをワイドスクリーン形式で表示できるようにするチートを有効にします。"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_JA "ワイドスクリーンハック"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_JA "通常の 4:3 アスペクト比の外側にジオメトリを描画します。新たに表示された領域でグラフィックに不具合が発生する可能性があります。"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_JA "GD-ROM 高速読み込み (不正確)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_JA "GD-ROM の読み込みを高速化します。"
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_JA "ドリームキャスト 32MB RAM MOD"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_JA "ドリームキャストの 32MB RAM MOD を有効にします。互換性に影響を与える可能性があります。"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_JA "SH4 CPU アンダー/オーバークロック"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_JA "SH4 メイン CPU クロックをデフォルトの 200 MHz から変更します。アンダークロックは遅いプラットフォームで動作を改善する可能性があります。オーバークロックは一部のゲームのフレームレートを上昇させる可能性があります。注意して使用してください。"
#define OPTION_VAL_100_JA NULL
#define OPTION_VAL_110_JA NULL
#define OPTION_VAL_120_JA NULL
#define OPTION_VAL_130_JA NULL
#define OPTION_VAL_140_JA NULL
#define OPTION_VAL_150_JA NULL
#define OPTION_VAL_160_JA NULL
#define OPTION_VAL_170_JA NULL
#define OPTION_VAL_180_JA NULL
#define OPTION_VAL_190_JA NULL
#define OPTION_VAL_200_JA NULL
#define OPTION_VAL_210_JA NULL
#define OPTION_VAL_220_JA NULL
#define OPTION_VAL_230_JA NULL
#define OPTION_VAL_240_JA NULL
#define OPTION_VAL_250_JA NULL
#define OPTION_VAL_260_JA NULL
#define OPTION_VAL_270_JA NULL
#define OPTION_VAL_280_JA NULL
#define OPTION_VAL_290_JA NULL
#define OPTION_VAL_300_JA NULL
#define OPTION_VAL_310_JA NULL
#define OPTION_VAL_320_JA NULL
#define OPTION_VAL_330_JA NULL
#define OPTION_VAL_340_JA NULL
#define OPTION_VAL_350_JA NULL
#define OPTION_VAL_360_JA NULL
#define OPTION_VAL_370_JA NULL
#define OPTION_VAL_380_JA NULL
#define OPTION_VAL_390_JA NULL
#define OPTION_VAL_400_JA NULL
#define OPTION_VAL_410_JA NULL
#define OPTION_VAL_420_JA NULL
#define OPTION_VAL_430_JA NULL
#define OPTION_VAL_440_JA NULL
#define OPTION_VAL_450_JA NULL
#define OPTION_VAL_460_JA NULL
#define OPTION_VAL_470_JA NULL
#define OPTION_VAL_480_JA NULL
#define OPTION_VAL_490_JA NULL
#define OPTION_VAL_500_JA NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_JA "カスタムテクスチャをロード"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_JA "[system/dc/textures/<game-id>/] フォルダに置かれたカスタムテクスチャを読み込みます。"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_JA NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_JA NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_JA "テクスチャをダンプ"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_JA "ゲームで新しいテクスチャが使用されるたび、テクスチャを .png ファイルとして [system/dc/texdump/<game-id>] フォルダに保存します。"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_JA NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_JA NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_JA "アナログスティックデッドゾーン"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_JA "アナログスティックが処理され始めるまでどれだけ傾ける必要があるかを選択します。"
#define OPTION_VAL_0_JA NULL
#define OPTION_VAL_5_JA NULL
#define OPTION_VAL_10_JA NULL
#define OPTION_VAL_15_JA NULL
#define OPTION_VAL_20_JA NULL
#define OPTION_VAL_25_JA NULL
#define OPTION_VAL_30_JA NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_JA "トリガーデッドゾーン"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_JA "トリガーが処理されるまでどれだけ押し込む必要があるかを選択します。"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_JA "デジタルトリガー"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_JA "有効にすると、トリガーは通常のボタンのように処理されます。つまり、押し込み量は考慮されず、完全に押されているか全く押されていないかのように処理されます。"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_JA "デジタル出力配信"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_JA "TCP ポート 8000 でデジタル出力とフォースフィードバックの状態を配信します。MAME の [-output network] オプションと互換性があります。"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_JA "ライトガン設定を表示"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_JA "ライトガンのクロスヘア表示に関する設定を有効にします。注意: この設定を反映するにはクイックメニューを開き直す必要があります。"
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_JA "ガンクロスヘアサイズ表示倍率"
#define OPTION_VAL_50_JA NULL
#define OPTION_VAL_60_JA NULL
#define OPTION_VAL_70_JA NULL
#define OPTION_VAL_80_JA NULL
#define OPTION_VAL_90_JA NULL
#define OPTION_VAL_100_O47_JA NULL
#define OPTION_VAL_110_O47_JA NULL
#define OPTION_VAL_120_O47_JA NULL
#define OPTION_VAL_130_O47_JA NULL
#define OPTION_VAL_140_O47_JA NULL
#define OPTION_VAL_150_O47_JA NULL
#define OPTION_VAL_160_O47_JA NULL
#define OPTION_VAL_170_O47_JA NULL
#define OPTION_VAL_180_O47_JA NULL
#define OPTION_VAL_190_O47_JA NULL
#define OPTION_VAL_200_O47_JA NULL
#define OPTION_VAL_210_O47_JA NULL
#define OPTION_VAL_220_O47_JA NULL
#define OPTION_VAL_230_O47_JA NULL
#define OPTION_VAL_240_O47_JA NULL
#define OPTION_VAL_250_O47_JA NULL
#define OPTION_VAL_260_O47_JA NULL
#define OPTION_VAL_270_O47_JA NULL
#define OPTION_VAL_280_O47_JA NULL
#define OPTION_VAL_290_O47_JA NULL
#define OPTION_VAL_300_O47_JA NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_JA "ガンクロスヘア 1 表示"
#define OPTION_VAL_WHITE_JA "白"
#define OPTION_VAL_RED_JA "レッド"
#define OPTION_VAL_GREEN_JA "グリーン"
#define OPTION_VAL_BLUE_JA "ブルー"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_JA "ガンクロスヘア 2 表示"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_JA "ガンクロスヘア 3 表示"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_JA "ガンクロスヘア 4 表示"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_JA "拡張スロット A1 デバイス"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_JA "拡張スロット A1 (ポート A スロット 1) に接続するデバイスを選択します。"
#define OPTION_VAL_VMU_JA NULL
#define OPTION_VAL_PURUPURU_JA "ぷるぷるぱっく"
#define OPTION_VAL_DREAMPOTATO_JA NULL
#define OPTION_VAL_NONE_JA "なし"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_JA "拡張スロット A2 デバイス"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_JA "拡張スロット A2 (ポート A スロット 2) に接続するデバイスを選択します。"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_JA "拡張スロット B1 デバイス"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_JA "拡張スロット B1 (ポート B スロット 1) に接続するデバイスを選択します。"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_JA "拡張スロット B2 デバイス"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_JA "拡張スロット B2 (ポート B スロット 2) に接続するデバイスを選択します。"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_JA "拡張スロット C1 デバイス"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_JA "拡張スロット C1 (ポート C スロット 1) に接続するデバイスを選択します。"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_JA "拡張スロット C2 デバイス"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_JA "拡張スロット C2 (ポート C スロット 2) に接続するデバイスを選択します。"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_JA "拡張スロット D1 デバイス"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_JA "拡張スロット D1 (ポート D スロット 1) に接続するデバイスを選択します。"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_JA "拡張スロット D2 デバイス"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_JA "拡張スロット D2 (ポート D スロット 2) に接続するデバイスを選択します。"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_JA "ゲームごとのビジュアルメモリユニット/システム (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_JA "ゲームごとの VMU"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_JA NULL
#define OPTION_VAL_VMU_A1_JA NULL
#define OPTION_VAL_ALL_VMUS_JA "すべての VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_JA "ビジュアルメモリユニット/システム (VMU) サウンド"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_JA "VMU サウンド"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_JA "有効にすると、VMU ビープ音が再生されます。"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_JA NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_JA NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_JA "ビジュアルメモリユニット/システム (VMU) 表示設定を表示"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_JA "VMU 表示設定を表示"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_JA "エミュレートされた VMU 液晶ディスプレイの表示/非表示、サイズ、位置および色の設定を有効にします。注意: この設定を反映するにはクイックメニューを開き直す必要があります。"
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_JA "VMU 画面 1 表示"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_JA "VMU 画面 1 位置"
#define OPTION_VAL_UPPER_LEFT_JA "左上"
#define OPTION_VAL_UPPER_RIGHT_JA "右上"
#define OPTION_VAL_LOWER_LEFT_JA "左下"
#define OPTION_VAL_LOWER_RIGHT_JA "右下"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_JA "VMU 画面 1 サイズ"
#define OPTION_VAL_1X_JA NULL
#define OPTION_VAL_3X_JA NULL
#define OPTION_VAL_5X_JA NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_JA "VMU 画面 1ピクセルオンカラー"
#define OPTION_VAL_DEFAULT_ON_00_JA "デフォルトでオン"
#define OPTION_VAL_DEFAULT_OFF_01_JA "デフォルトでオフ"
#define OPTION_VAL_BLACK_02_JA "黒"
#define OPTION_VAL_LIGHT_BLUE_04_JA "ライトブルー"
#define OPTION_VAL_CYAN_06_JA "シアン"
#define OPTION_VAL_CYAN_BLUE_07_JA "シアンブルー"
#define OPTION_VAL_LIGHT_GREEN_08_JA "ライトグリーン"
#define OPTION_VAL_CYAN_GREEN_09_JA "シアングリーン"
#define OPTION_VAL_LIGHT_CYAN_10_JA "ライトシアン"
#define OPTION_VAL_PURPLE_12_JA "紫"
#define OPTION_VAL_LIGHT_PURPLE_13_JA "ライトパープル"
#define OPTION_VAL_YELLOW_14_JA "イエロー"
#define OPTION_VAL_GRAY_15_JA "グレー"
#define OPTION_VAL_LIGHT_PURPLE_2_16_JA "ライトパープル (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_JA "ライトグリーン (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_JA "ライトグリーン (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_JA "ライトシアン (2)"
#define OPTION_VAL_LIGHT_RED_2_20_JA "ライトレッド (2)"
#define OPTION_VAL_MAGENTA_21_JA "マゼンタ"
#define OPTION_VAL_LIGHT_PURPLE_3_22_JA "ライトパープル (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_JA "ライトオレンジ"
#define OPTION_VAL_ORANGE_24_JA "オレンジ"
#define OPTION_VAL_LIGHT_PURPLE_4_25_JA "ライトパープル (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_JA "ライトイエロー"
#define OPTION_VAL_LIGHT_YELLOW_2_27_JA "ライトイエロー (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_JA "VMU 画面 1ピクセルオフカラー"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_JA "VMU 画面 1 不透明度"
#define OPTION_VAL_40_JA NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_JA "VMU 画面 2 表示"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_JA "VMU 画面 2 位置"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_JA "VMU 画面 2 サイズ"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_JA "VMU 画面 2 ピクセルオンカラー"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_JA "VMU 画面 2 ピクセルオフカラー"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_JA "VMU 画面 2 不透明度"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_JA "VMU 画面 3 表示"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_JA "VMU 画面 3 位置"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_JA "VMU 画面 3 サイズ"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_JA "VMU 画面 3 ピクセルオンカラー"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_JA "VMU 画面 3 ピクセルオフカラー"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_JA "VMU 画面 3 不透明度"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_JA "VMU 画面 4 表示"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_JA "VMU 画面 4 位置"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_JA "VMU 画面 4 サイズ"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_JA "VMU 画面 4 ピクセルオンカラー"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_JA "VMU 画面 4 ピクセルオフカラー"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_JA "VMU 画面 4 不透明度"

struct retro_core_option_v2_category option_cats_ja[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_JA,
      CATEGORY_SYSTEM_INFO_0_JA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_JA,
      CATEGORY_VIDEO_INFO_0_JA
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_JA,
      CATEGORY_PERFORMANCE_INFO_0_JA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_JA,
      CATEGORY_HACKS_INFO_0_JA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_JA,
      CATEGORY_INPUT_INFO_0_JA
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_JA,
      CATEGORY_EXPANSIONS_INFO_0_JA
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_JA,
      CATEGORY_VMU_INFO_0_JA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ja[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_JA,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_JA },
         { "USA",     OPTION_VAL_USA_JA },
         { "Europe",  OPTION_VAL_EUROPE_JA },
         { "Default", OPTION_VAL_DEFAULT_JA },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_JA,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_JA },
         { "English",  OPTION_VAL_ENGLISH_JA },
         { "German",   OPTION_VAL_GERMAN_JA },
         { "French",   OPTION_VAL_FRENCH_JA },
         { "Spanish",  OPTION_VAL_SPANISH_JA },
         { "Italian",  OPTION_VAL_ITALIAN_JA },
         { "Default",  OPTION_VAL_DEFAULT_JA },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_JA,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_JA },
         { "640x480",    OPTION_VAL_640X480_JA },
         { "800x600",    OPTION_VAL_800X600_JA },
         { "960x720",    OPTION_VAL_960X720_JA },
         { "1024x768",   OPTION_VAL_1024X768_JA },
         { "1280x960",   OPTION_VAL_1280X960_JA },
         { "1440x1080",  OPTION_VAL_1440X1080_JA },
         { "1600x1200",  OPTION_VAL_1600X1200_JA },
         { "1920x1440",  OPTION_VAL_1920X1440_JA },
         { "2560x1920",  OPTION_VAL_2560X1920_JA },
         { "2880x2160",  OPTION_VAL_2880X2160_JA },
         { "3200x2400",  OPTION_VAL_3200X2400_JA },
         { "3840x2880",  OPTION_VAL_3840X2880_JA },
         { "4480x3360",  OPTION_VAL_4480X3360_JA },
         { "5120x3840",  OPTION_VAL_5120X3840_JA },
         { "5760x4320",  OPTION_VAL_5760X4320_JA },
         { "6400x4800",  OPTION_VAL_6400X4800_JA },
         { "7040x5280",  OPTION_VAL_7040X5280_JA },
         { "7680x5760",  OPTION_VAL_7680X5760_JA },
         { "8320x6240",  OPTION_VAL_8320X6240_JA },
         { "8960x6720",  OPTION_VAL_8960X6720_JA },
         { "9600x7200",  OPTION_VAL_9600X7200_JA },
         { "10240x7680", OPTION_VAL_10240X7680_JA },
         { "10880x8160", OPTION_VAL_10880X8160_JA },
         { "11520x8640", OPTION_VAL_11520X8640_JA },
         { "12160x9120", OPTION_VAL_12160X9120_JA },
         { "12800x9600", OPTION_VAL_12800X9600_JA },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_JA,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_JA },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_JA },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_JA },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_JA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_JA },
         { "PAL",     OPTION_VAL_PAL_JA },
         { "PAL_N",   OPTION_VAL_PAL_N_JA },
         { "PAL_M",   OPTION_VAL_PAL_M_JA },
         { "Default", OPTION_VAL_DEFAULT_JA },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_JA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_JA },
         { "vertical",   OPTION_VAL_VERTICAL_JA },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_JA,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_JA },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_JA },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_JA },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_JA,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_JA },
         { "1GB",   OPTION_VAL_1GB_JA },
         { "2GB",   OPTION_VAL_2GB_JA },
         { "4GB",   OPTION_VAL_4GB_JA },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_JA,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_JA,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_JA,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_JA,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_JA },
         { "1",  OPTION_VAL_1_JA },
         { "2",  OPTION_VAL_2_JA },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_JA,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_JA },
         { "4",  OPTION_VAL_4_JA },
         { "6",  OPTION_VAL_6_JA },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_JA,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_JA,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_JA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_JA,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_JA,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_JA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_JA,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_JA },
         { "more", OPTION_VAL_MORE_JA },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_JA,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_JA },
         { "110", OPTION_VAL_110_JA },
         { "120", OPTION_VAL_120_JA },
         { "130", OPTION_VAL_130_JA },
         { "140", OPTION_VAL_140_JA },
         { "150", OPTION_VAL_150_JA },
         { "160", OPTION_VAL_160_JA },
         { "170", OPTION_VAL_170_JA },
         { "180", OPTION_VAL_180_JA },
         { "190", OPTION_VAL_190_JA },
         { "200", OPTION_VAL_200_JA },
         { "210", OPTION_VAL_210_JA },
         { "220", OPTION_VAL_220_JA },
         { "230", OPTION_VAL_230_JA },
         { "240", OPTION_VAL_240_JA },
         { "250", OPTION_VAL_250_JA },
         { "260", OPTION_VAL_260_JA },
         { "270", OPTION_VAL_270_JA },
         { "280", OPTION_VAL_280_JA },
         { "290", OPTION_VAL_290_JA },
         { "300", OPTION_VAL_300_JA },
         { "310", OPTION_VAL_310_JA },
         { "320", OPTION_VAL_320_JA },
         { "330", OPTION_VAL_330_JA },
         { "340", OPTION_VAL_340_JA },
         { "350", OPTION_VAL_350_JA },
         { "360", OPTION_VAL_360_JA },
         { "370", OPTION_VAL_370_JA },
         { "380", OPTION_VAL_380_JA },
         { "390", OPTION_VAL_390_JA },
         { "400", OPTION_VAL_400_JA },
         { "410", OPTION_VAL_410_JA },
         { "420", OPTION_VAL_420_JA },
         { "430", OPTION_VAL_430_JA },
         { "440", OPTION_VAL_440_JA },
         { "450", OPTION_VAL_450_JA },
         { "460", OPTION_VAL_460_JA },
         { "470", OPTION_VAL_470_JA },
         { "480", OPTION_VAL_480_JA },
         { "490", OPTION_VAL_490_JA },
         { "500", OPTION_VAL_500_JA },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_JA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_JA },
         { "5%",  OPTION_VAL_5_JA },
         { "10%", OPTION_VAL_10_JA },
         { "15%", OPTION_VAL_15_JA },
         { "20%", OPTION_VAL_20_JA },
         { "25%", OPTION_VAL_25_JA },
         { "30%", OPTION_VAL_30_JA },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_JA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_JA },
         { "5%",  OPTION_VAL_5_JA },
         { "10%", OPTION_VAL_10_JA },
         { "15%", OPTION_VAL_15_JA },
         { "20%", OPTION_VAL_20_JA },
         { "25%", OPTION_VAL_25_JA },
         { "30%", OPTION_VAL_30_JA },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_JA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_JA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_JA,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_JA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_JA },
         { "60%",  OPTION_VAL_60_JA },
         { "70%",  OPTION_VAL_70_JA },
         { "80%",  OPTION_VAL_80_JA },
         { "90%",  OPTION_VAL_90_JA },
         { "100%", OPTION_VAL_100_O47_JA },
         { "110%", OPTION_VAL_110_O47_JA },
         { "120%", OPTION_VAL_120_O47_JA },
         { "130%", OPTION_VAL_130_O47_JA },
         { "140%", OPTION_VAL_140_O47_JA },
         { "150%", OPTION_VAL_150_O47_JA },
         { "160%", OPTION_VAL_160_O47_JA },
         { "170%", OPTION_VAL_170_O47_JA },
         { "180%", OPTION_VAL_180_O47_JA },
         { "190%", OPTION_VAL_190_O47_JA },
         { "200%", OPTION_VAL_200_O47_JA },
         { "210%", OPTION_VAL_210_O47_JA },
         { "220%", OPTION_VAL_220_O47_JA },
         { "230%", OPTION_VAL_230_O47_JA },
         { "240%", OPTION_VAL_240_O47_JA },
         { "250%", OPTION_VAL_250_O47_JA },
         { "260%", OPTION_VAL_260_O47_JA },
         { "270%", OPTION_VAL_270_O47_JA },
         { "280%", OPTION_VAL_280_O47_JA },
         { "290%", OPTION_VAL_290_O47_JA },
         { "300%", OPTION_VAL_300_O47_JA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_JA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_JA },
         { "Red",      OPTION_VAL_RED_JA },
         { "Green",    OPTION_VAL_GREEN_JA },
         { "Blue",     OPTION_VAL_BLUE_JA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_JA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_JA },
         { "Red",      OPTION_VAL_RED_JA },
         { "Green",    OPTION_VAL_GREEN_JA },
         { "Blue",     OPTION_VAL_BLUE_JA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_JA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_JA },
         { "Red",      OPTION_VAL_RED_JA },
         { "Green",    OPTION_VAL_GREEN_JA },
         { "Blue",     OPTION_VAL_BLUE_JA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_JA,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_JA },
         { "Red",      OPTION_VAL_RED_JA },
         { "Green",    OPTION_VAL_GREEN_JA },
         { "Blue",     OPTION_VAL_BLUE_JA },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_JA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_JA },
         { "Purupuru", OPTION_VAL_PURUPURU_JA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_JA },
         { "None",     OPTION_VAL_NONE_JA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_JA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_JA },
         { "Purupuru", OPTION_VAL_PURUPURU_JA },
         { "None",     OPTION_VAL_NONE_JA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_JA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_JA },
         { "Purupuru", OPTION_VAL_PURUPURU_JA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_JA },
         { "None",     OPTION_VAL_NONE_JA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_JA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_JA },
         { "Purupuru", OPTION_VAL_PURUPURU_JA },
         { "None",     OPTION_VAL_NONE_JA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_JA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_JA },
         { "Purupuru", OPTION_VAL_PURUPURU_JA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_JA },
         { "None",     OPTION_VAL_NONE_JA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_JA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_JA },
         { "Purupuru", OPTION_VAL_PURUPURU_JA },
         { "None",     OPTION_VAL_NONE_JA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_JA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_JA },
         { "Purupuru", OPTION_VAL_PURUPURU_JA },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_JA },
         { "None",     OPTION_VAL_NONE_JA },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_JA,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_JA },
         { "Purupuru", OPTION_VAL_PURUPURU_JA },
         { "None",     OPTION_VAL_NONE_JA },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_JA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_JA,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_JA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_JA },
         { "All VMUs", OPTION_VAL_ALL_VMUS_JA },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_JA,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_JA,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_JA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_JA,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_JA,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_JA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_JA,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_JA,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_JA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_JA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_JA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_JA },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_JA },
         { "2x", OPTION_VAL_2_O26_JA },
         { "3x", OPTION_VAL_3X_JA },
         { "4x", OPTION_VAL_4_JA },
         { "5x", OPTION_VAL_5X_JA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_JA },
         { "20%",  OPTION_VAL_20_JA },
         { "30%",  OPTION_VAL_30_JA },
         { "40%",  OPTION_VAL_40_JA },
         { "50%",  OPTION_VAL_50_JA },
         { "60%",  OPTION_VAL_60_JA },
         { "70%",  OPTION_VAL_70_JA },
         { "80%",  OPTION_VAL_80_JA },
         { "90%",  OPTION_VAL_90_JA },
         { "100%", OPTION_VAL_100_O47_JA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_JA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_JA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_JA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_JA },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_JA },
         { "2x", OPTION_VAL_2_O26_JA },
         { "3x", OPTION_VAL_3X_JA },
         { "4x", OPTION_VAL_4_JA },
         { "5x", OPTION_VAL_5X_JA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_JA },
         { "20%",  OPTION_VAL_20_JA },
         { "30%",  OPTION_VAL_30_JA },
         { "40%",  OPTION_VAL_40_JA },
         { "50%",  OPTION_VAL_50_JA },
         { "60%",  OPTION_VAL_60_JA },
         { "70%",  OPTION_VAL_70_JA },
         { "80%",  OPTION_VAL_80_JA },
         { "90%",  OPTION_VAL_90_JA },
         { "100%", OPTION_VAL_100_O47_JA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_JA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_JA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_JA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_JA },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_JA },
         { "2x", OPTION_VAL_2_O26_JA },
         { "3x", OPTION_VAL_3X_JA },
         { "4x", OPTION_VAL_4_JA },
         { "5x", OPTION_VAL_5X_JA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_JA },
         { "20%",  OPTION_VAL_20_JA },
         { "30%",  OPTION_VAL_30_JA },
         { "40%",  OPTION_VAL_40_JA },
         { "50%",  OPTION_VAL_50_JA },
         { "60%",  OPTION_VAL_60_JA },
         { "70%",  OPTION_VAL_70_JA },
         { "80%",  OPTION_VAL_80_JA },
         { "90%",  OPTION_VAL_90_JA },
         { "100%", OPTION_VAL_100_O47_JA },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_JA },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_JA },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_JA },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_JA },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_JA },
         { "2x", OPTION_VAL_2_O26_JA },
         { "3x", OPTION_VAL_3X_JA },
         { "4x", OPTION_VAL_4_JA },
         { "5x", OPTION_VAL_5X_JA },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_JA },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_JA },
         { "BLACK 02",          OPTION_VAL_BLACK_02_JA },
         { "BLUE 03",           OPTION_VAL_BLUE_JA },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_JA },
         { "GREEN 05",          OPTION_VAL_GREEN_JA },
         { "CYAN 06",           OPTION_VAL_CYAN_06_JA },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_JA },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_JA },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_JA },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_JA },
         { "RED 11",            OPTION_VAL_RED_JA },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_JA },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_JA },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_JA },
         { "GRAY 15",           OPTION_VAL_GRAY_15_JA },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_JA },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_JA },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_JA },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_JA },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_JA },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_JA },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_JA },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_JA },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_JA },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_JA },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_JA },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_JA },
         { "WHITE 28",          OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_JA,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_JA },
         { "20%",  OPTION_VAL_20_JA },
         { "30%",  OPTION_VAL_30_JA },
         { "40%",  OPTION_VAL_40_JA },
         { "50%",  OPTION_VAL_50_JA },
         { "60%",  OPTION_VAL_60_JA },
         { "70%",  OPTION_VAL_70_JA },
         { "80%",  OPTION_VAL_80_JA },
         { "90%",  OPTION_VAL_90_JA },
         { "100%", OPTION_VAL_100_O47_JA },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ja = {
   option_cats_ja,
   option_defs_ja
};

/* RETRO_LANGUAGE_KO */

#define CATEGORY_SYSTEM_LABEL_KO "시스템"
#define CATEGORY_SYSTEM_INFO_0_KO "지역, 언어, BIOS 및 하드웨어 설정을 구성합니다."
#define CATEGORY_VIDEO_LABEL_KO "비디오"
#define CATEGORY_VIDEO_INFO_0_KO "해상도, 투명도 및 화면 효과 설정을 구성합니다."
#define CATEGORY_PERFORMANCE_LABEL_KO "성능"
#define CATEGORY_PERFORMANCE_INFO_0_KO "스레드 렌더링 및 프레임 스킵 설정을 구성합니다."
#define CATEGORY_HACKS_LABEL_KO "에뮬레이션 핵"
#define CATEGORY_HACKS_INFO_0_KO "와이드스크린 핵, GD롬 로딩 속도 조절 및 텍스쳐 변경 설정등을 구성합니다."
#define CATEGORY_INPUT_LABEL_KO "입력"
#define CATEGORY_INPUT_INFO_0_KO "게임패드 및 라이트 건 설정을 구성합니다."
#define CATEGORY_EXPANSIONS_LABEL_KO "컨트롤러 확장 슬롯"
#define CATEGORY_EXPANSIONS_INFO_0_KO "컨트롤러의 각 확장 슬롯에 연결할 장치(VMU, 진동 장치)를 선택합니다."
#define CATEGORY_VMU_LABEL_KO "비쥬얼 메모리 유닛"
#define CATEGORY_VMU_INFO_0_KO "게임 별 VMU 저장 파일 및 VMU OSD 설정을 구성합니다."
#define CORE_OPTION_NAME_REGION_LABEL_KO "지역"
#define OPTION_VAL_JAPAN_KO "일본"
#define OPTION_VAL_USA_KO "미국"
#define OPTION_VAL_EUROPE_KO "유럽"
#define OPTION_VAL_DEFAULT_KO "기본"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_KO "언어"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_KO "BIOS 및 다국어 게임에서 사용되는 언어 설정을 변경합니다"
#define OPTION_VAL_JAPANESE_KO "일본어"
#define OPTION_VAL_ENGLISH_KO "영어"
#define OPTION_VAL_GERMAN_KO "독일어"
#define OPTION_VAL_FRENCH_KO "프랑스어"
#define OPTION_VAL_SPANISH_KO "스페인어"
#define OPTION_VAL_ITALIAN_KO "이탈리아어"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_KO "HLE BIOS (재시작 필요)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_KO "HLE BIOS 사용을 강제합니다."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_KO "HDR 활성화"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_KO "Dreamcast의 오디오 DSP(디지털 신호 프로세서)의 에뮬레이션을 활성화합니다. 사운드의 정확도는 향상되지만 성능 요구 사항이 증가합니다."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_KO "아케이드 서비스 버튼 허용"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_KO "아케이드 게임에서 캐비넷 설정에 들어가기 위한 서비스 버튼을 활성화합니다."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_KO "NAOMI 게임을 무료 플레이로 설정"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_KO "무료 플레이를 위해 게임의 코인 설정을 수정합니다."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_KO "브로드밴드 어댑터 에뮬레이션"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_KO "모뎀 대신 이더넷 브로드밴드 어댑터를 에뮬레이트합니다. (재시작 필요)"
#define CORE_OPTION_NAME_UPNP_LABEL_KO "UPnP 활성화"
#define CORE_OPTION_NAME_UPNP_INFO_0_KO "온라인 게임에서 UPnP를 사용해 자동으로 인터넷 라우터를 설정합니다."
#define CORE_OPTION_NAME_DCNET_LABEL_KO "DCNet 사용"
#define CORE_OPTION_NAME_DCNET_INFO_0_KO "DCNet 클라우드 서비스를 통해 Dreamcast를 인터넷에 연결합니다."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_KO "내부 해상도"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_KO "렌더링 해상도를 변경합니다."
#define OPTION_VAL_320X240_KO "320x240 (절반)"
#define OPTION_VAL_640X480_KO "640x480 (네이티브)"
#define OPTION_VAL_800X600_KO NULL
#define OPTION_VAL_960X720_KO NULL
#define OPTION_VAL_1024X768_KO NULL
#define OPTION_VAL_1280X960_KO NULL
#define OPTION_VAL_1440X1080_KO NULL
#define OPTION_VAL_1600X1200_KO NULL
#define OPTION_VAL_1920X1440_KO NULL
#define OPTION_VAL_2560X1920_KO NULL
#define OPTION_VAL_2880X2160_KO NULL
#define OPTION_VAL_3200X2400_KO NULL
#define OPTION_VAL_3840X2880_KO NULL
#define OPTION_VAL_4480X3360_KO NULL
#define OPTION_VAL_5120X3840_KO NULL
#define OPTION_VAL_5760X4320_KO NULL
#define OPTION_VAL_6400X4800_KO NULL
#define OPTION_VAL_7040X5280_KO NULL
#define OPTION_VAL_7680X5760_KO NULL
#define OPTION_VAL_8320X6240_KO NULL
#define OPTION_VAL_8960X6720_KO NULL
#define OPTION_VAL_9600X7200_KO NULL
#define OPTION_VAL_10240X7680_KO NULL
#define OPTION_VAL_10880X8160_KO NULL
#define OPTION_VAL_11520X8640_KO NULL
#define OPTION_VAL_12160X9120_KO NULL
#define OPTION_VAL_12800X9600_KO NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_KO "케이블 타입"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_KO "외부 출력 형식. 'TV (컴포지트)'가 가장 많이 지원됨."
#define OPTION_VAL_VGA_KO NULL
#define OPTION_VAL_TV_RGB_KO NULL
#define OPTION_VAL_TV_COMPOSITE_KO "TV (컴포지트)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_KO "방송 표준"
#define OPTION_VAL_NTSC_KO NULL
#define OPTION_VAL_PAL_KO "PAL (국제)"
#define OPTION_VAL_PAL_N_KO "PAL-N (아르헨티나, 파라과이, 우르과이)"
#define OPTION_VAL_PAL_M_KO "PAL-M (브라질)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_KO "화면 방향"
#define OPTION_VAL_HORIZONTAL_KO "가로"
#define OPTION_VAL_VERTICAL_KO "세로"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_KO "알파 정렬"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_KO "투명 폴리곤을 정렬할 방식을 선택합니다."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_KO "줄당 (빠름, 가장 부정확)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_KO "삼각형당 (중간)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_KO "픽셀당 (정확, 가장 느림)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_KO "누산 픽셀 버퍼 크기"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_KO "높은 해상도 출력을 올바르게 표시하려면 이 값을 높은 값으로 설정해야 할 수도 있습니다."
#define OPTION_VAL_512MB_KO NULL
#define OPTION_VAL_1GB_KO NULL
#define OPTION_VAL_2GB_KO NULL
#define OPTION_VAL_4GB_KO NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_KO "최대 투명 레이어"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_KO "복잡한 장면을 그릴 때 높은 값이 필요할 수 있습니다."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_KO "전체 프레임버퍼 에뮬레이션"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_KO "VRAM 상 전체 프레임버퍼 에뮬레이션을 활성화합니다. VRAM 상의 프레임버퍼에 직접 읽기 또는 쓰기를 수행하는 게임을 실행할 때 유용합니다. 활성화할 경우, 내부 해상도가 640x480으로 고정되며 성능에 큰 영향을 끼칠 수 있습니다."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_KO "RTT (Render To Texture) 버퍼 사용"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_KO "렌더된 텍스쳐를 GPU에서 VRAM으로 복사합니다. 이 옵션은 이 옵션을 필요로 하는 게임에서는 기본적으로 활성화됩니다. 활성화할 경우, 텍스쳐 업스케일링이 비활성화되며 성능에 큰 영향을 끼칠 수 있습니다."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_KO "밉맵"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_KO "활성화할 경우 멀리 떨어진 곳에 텍스처를 그릴 때 작은 버전의 텍스처를 사용하여 성능을 개선하고 번쩍이는 효과를 억제합니다."
#define CORE_OPTION_NAME_FOG_LABEL_KO "안개 효과"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_KO "볼륨 모디파이어"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_KO "드림캐스트 GPU의 기능으로 보통 개체의 그림자를 그리는 데 사용됩니다. 이 기능은 보통 활성화해두어야 하며 성능에 미치는 영향은 아주 적습니다."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_KO "비등방성 필터링"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_KO "카메라와 비스듬하게 배치된 표면의 텍스쳐 품질을 향상합니다. 높은 값으로 설정하면 GPU 부하가 높아집니다. 변경 사항은 코어를 다시 시작해야 적용됩니다."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_KO "텍스쳐 필터링"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_KO "사용할 텍스쳐 필터링 모드를 선택합니다. 모든 텍스쳐에 강제로 특정 필터링 모드를 적용하여 더 날카로운(또는 부드러운) 비주얼을 얻을 수 있습니다. 기본 외의 다른 값으로 설정 시 다양한 렌더링 문제가 발생할 수 있습니다. 변경 사항은 코어를 다시 시작해야 적용됩니다."
#define OPTION_VAL_1_KO "강제 최근접이웃"
#define OPTION_VAL_2_KO "강제 선형"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_KO "프레임 스왑 지연"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_KO "화면이 번쩍이거나 깨지는 것을 방지할 수 있습니다. 느린 플랫폼에서는 사용을 권장하지 않습니다."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_KO "프레임 레이트 변화 감지"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_KO "내부 프레임 레이트가 변경될 때(예: 60FPS에서 30FPS로) 프론트엔드에게 알림을 보냅니다. 30FPS 또는 20FPS 고정으로 실행되는 게임에서는 프레임 페이스를 개선하지만, 무제한(불안정) 프레임 레이트로 실행되는 게임(예: Ecco the Dolphin, Unreal Tournament)을 실행할 때는 비활성화되어야 합니다. 주의: '자동 프레임 건너뛰기'가 활성화되어있으면 작동하지 않습니다."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_KO "PowerVR2 포스트프로세싱 필터"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_KO "렌더된 이미지에 PowerVR2 GPU와 아날로그 비디오 신호에서 나타나는 효과를 후처리로 구현합니다."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_KO "텍스쳐 업스케일링 (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_KO "손으로 그려진 2D 픽셀아트 그래픽을 개선합니다. 2D 픽셀아트 게임에서만 사용되어야 합니다."
#define OPTION_VAL_2_O26_KO NULL
#define OPTION_VAL_4_KO NULL
#define OPTION_VAL_6_KO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_KO "텍스쳐 업스케일링 필터 최대 크기"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_KO "텍스처를 업스케일할 최대 크기를 선택합니다. 텍스처의 크기가 설정한 크기보다 클 경우에는 텍스처를 업스케일하지 않습니다."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_KO "네이티브 깊이 보간"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_KO "AMD GPU에서 깊이 문제와 텍스쳐 망가짐 증상을 개선하는 데 도움이 됩니다. 일부 상황에서는 인텔 GPU에서도 도움이 될 수 있습니다."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_KO "업스케일 시 테두리 블리딩 억제"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_KO "업스케일 시 텍스처 블리딩 문제를 완화합니다. 비활성화할 경우 일부 2D 게임(MVC2, CVS, KOF 등)에서 픽셀이 뒤틀리는 현상이 해결될 수 있습니다."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_KO "스레드 렌더링"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_KO "GPU와 CPU를 별도의 스레드에서 실행합니다. 사용하는 것이 매우 권장됩니다."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_KO "자동 프레임 건너뛰기"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_KO "에뮬레이터 실행이 느려질 경우 자동으로 프레임을 건너뜁니다. 주의: 이 설정은 '스레드 렌더링'이 활성화되어 있어야 적용됩니다."
#define OPTION_VAL_SOME_KO "보통"
#define OPTION_VAL_MORE_KO "최대"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_KO "프레임 건너뛰기"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_KO "표시되는 각 프레임 사이에서 건너뛸 프레임 수를 설정합니다."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_KO "와이드스크린 핵 (재시작 필요)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_KO "일부 게임이 와이드스크린 형식으로 출력되게 하는 치트를 활성화합니다."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_KO "와이드스크린 핵"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_KO "일반적인 4:3 화면비 밖에도 개체를 그립니다. 드러난 부분에서 그래픽 오류가 발생할 수도 있습니다."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_KO "GD-ROM 빠른 로딩 (정확하지 않음)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_KO "GD-ROM 로딩 속도를 높입니다."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_KO "드림캐스트 32MB RAM 모드"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_KO "드림캐스트용 32MB RAM 모드를 활성화합니다. 호환성에 영향이 있을 수 있습니다"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_KO "SH4 CPU 언더/오버클럭"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_KO "SH4 메인 CPU 클럭을 기본 200MHz에서 변경합니다. 언더클럭하면 플랫폼 속도가 느려질 수 있습니다. 오버클럭하면 일부 게임에서 프레임 속도가 증가할 수 있습니다. 주의해서 사용하세요."
#define OPTION_VAL_100_KO NULL
#define OPTION_VAL_110_KO NULL
#define OPTION_VAL_120_KO NULL
#define OPTION_VAL_130_KO NULL
#define OPTION_VAL_140_KO NULL
#define OPTION_VAL_150_KO NULL
#define OPTION_VAL_160_KO NULL
#define OPTION_VAL_170_KO NULL
#define OPTION_VAL_180_KO NULL
#define OPTION_VAL_190_KO NULL
#define OPTION_VAL_200_KO NULL
#define OPTION_VAL_210_KO NULL
#define OPTION_VAL_220_KO NULL
#define OPTION_VAL_230_KO NULL
#define OPTION_VAL_240_KO NULL
#define OPTION_VAL_250_KO NULL
#define OPTION_VAL_260_KO NULL
#define OPTION_VAL_270_KO NULL
#define OPTION_VAL_280_KO NULL
#define OPTION_VAL_290_KO NULL
#define OPTION_VAL_300_KO NULL
#define OPTION_VAL_310_KO NULL
#define OPTION_VAL_320_KO NULL
#define OPTION_VAL_330_KO NULL
#define OPTION_VAL_340_KO NULL
#define OPTION_VAL_350_KO NULL
#define OPTION_VAL_360_KO NULL
#define OPTION_VAL_370_KO NULL
#define OPTION_VAL_380_KO NULL
#define OPTION_VAL_390_KO NULL
#define OPTION_VAL_400_KO NULL
#define OPTION_VAL_410_KO NULL
#define OPTION_VAL_420_KO NULL
#define OPTION_VAL_430_KO NULL
#define OPTION_VAL_440_KO NULL
#define OPTION_VAL_450_KO NULL
#define OPTION_VAL_460_KO NULL
#define OPTION_VAL_470_KO NULL
#define OPTION_VAL_480_KO NULL
#define OPTION_VAL_490_KO NULL
#define OPTION_VAL_500_KO NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_KO "커스텀 텍스처 불러오기"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_KO "'system/dc/textures/<game-id>/' 폴더에서 커스텀 텍스처를 불러옵니다."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_KO "커스텀 텍스처 미리 불러오기"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_KO "커스텀 텍스처를 게임 시작 시에 미리 불러옵니다. 성능이 향상될 수 있지만 메모리 사용량이 늘어납니다."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_KO "텍스쳐 덤프하기"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_KO "게임이 새로운 텍스처를 사용할 때마다, 'system/dc/texdump/<game-id>/' 폴더 안에 .png 파일로 저장되게 합니다."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_KO "교체된 텍스처 덤프하기"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_KO "커스텀 텍스처로 교체된 텍스처를 항상 덤프합니다."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_KO "아날로그 스틱 데드존"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_KO "아날로그 스틱을 얼마나 기울여야 입력으로 인식할지를 선택합니다."
#define OPTION_VAL_0_KO NULL
#define OPTION_VAL_5_KO NULL
#define OPTION_VAL_10_KO NULL
#define OPTION_VAL_15_KO NULL
#define OPTION_VAL_20_KO NULL
#define OPTION_VAL_25_KO NULL
#define OPTION_VAL_30_KO NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_KO "트리거 데드존"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_KO "트리거를 얼마나 눌러야 입력으로 인식할지를 선택합니다."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_KO "디지털 트리거"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_KO "활성화할 경우 트리거가 평범한 버튼으로 취급되며, 중간 상태 없이 완전히 눌리거나 눌리지 않은 상태로만 처리됩니다."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_KO "디지털 출력 송출하기"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_KO "디지털 출력과 포스 피드백 상태를 TCP 포트 8000번으로 송출합니다. MAME의 \"-output network\" 옵션과 호환됩니다."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_KO "라이트 건 설정 표시"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_KO "라이트건 십자 조준선 옵션 구성을 활성화합니다. 참고: 이 설정을 적용하려면 빠른 메뉴를 다시 표시해야 할 수 있습니다."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_KO "총 십자선 크기 조정"
#define OPTION_VAL_50_KO NULL
#define OPTION_VAL_60_KO NULL
#define OPTION_VAL_70_KO NULL
#define OPTION_VAL_80_KO NULL
#define OPTION_VAL_90_KO NULL
#define OPTION_VAL_100_O47_KO NULL
#define OPTION_VAL_110_O47_KO NULL
#define OPTION_VAL_120_O47_KO NULL
#define OPTION_VAL_130_O47_KO NULL
#define OPTION_VAL_140_O47_KO NULL
#define OPTION_VAL_150_O47_KO NULL
#define OPTION_VAL_160_O47_KO NULL
#define OPTION_VAL_170_O47_KO NULL
#define OPTION_VAL_180_O47_KO NULL
#define OPTION_VAL_190_O47_KO NULL
#define OPTION_VAL_200_O47_KO NULL
#define OPTION_VAL_210_O47_KO NULL
#define OPTION_VAL_220_O47_KO NULL
#define OPTION_VAL_230_O47_KO NULL
#define OPTION_VAL_240_O47_KO NULL
#define OPTION_VAL_250_O47_KO NULL
#define OPTION_VAL_260_O47_KO NULL
#define OPTION_VAL_270_O47_KO NULL
#define OPTION_VAL_280_O47_KO NULL
#define OPTION_VAL_290_O47_KO NULL
#define OPTION_VAL_300_O47_KO NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_KO "라이트건 조준선 1 표시"
#define OPTION_VAL_WHITE_KO "흰색"
#define OPTION_VAL_RED_KO "빨강"
#define OPTION_VAL_GREEN_KO "초록"
#define OPTION_VAL_BLUE_KO "파랑"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_KO "라이트건 조준선 2 표시"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_KO "라이트건 조준선 3 표시"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_KO "라이트건 조준선 4 표시"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_KO "확장 슬롯 A1의 장치"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_KO "확장 슬롯 A1(포트 A 슬롯 1)에 연결할 장치를 선택합니다."
#define OPTION_VAL_VMU_KO NULL
#define OPTION_VAL_PURUPURU_KO "진동 팩"
#define OPTION_VAL_DREAMPOTATO_KO NULL
#define OPTION_VAL_NONE_KO "없음"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_KO "확장 슬롯 A2의 장치"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_KO "확장 슬롯 A2(포트 A 슬롯 2)에 연결할 장치를 선택합니다."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_KO "확장 슬롯 B1의 장치"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_KO "확장 슬롯 B1(포트 B 슬롯 1)에 연결할 장치를 선택합니다."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_KO "확장 슬롯 B2의 장치"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_KO "확장 슬롯 B2(포트 B 슬롯 2)에 연결할 장치를 선택합니다."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_KO "확장 슬롯 C1의 장치"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_KO "확장 슬롯 C1(포트 C 슬롯 1)에 연결할 장치를 선택합니다."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_KO "확장 슬롯 C2의 장치"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_KO "확장 슬롯 C2(포트 C 슬롯 2)에 연결할 장치를 선택합니다."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_KO "확장 슬롯 D1의 장치"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_KO "확장 슬롯 D1(포트 D 슬롯 1)에 연결할 장치를 선택합니다."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_KO "확장 슬롯 D2의 장치"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_KO "확장 슬롯 D2(포트 D 슬롯 2)에 연결할 장치를 선택합니다."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_KO "게임당 별도 Visual Memory Units/Systems(VMU) 사용"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_KO "게임당 별도 VMU 사용"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_KO "비활성화할 경우, 모든 게임이 RetroArch의 시스템 폴더에 저장된 최대 8개의 VMU 저장 파일(A1/A2/B1/B2/C1/C2/D1/D2)을 공유합니다.\n'VMU A1' 옵션은 RetroArch의 저장 폴더에 게임 하나 당 하나의 VMU 'A1' 파일을 생성합니다.\n'모든 VMU'는 게임 하나 당 최대 8개의 VMU 파일(A1/A2/B1/B2/C1/C2/D1/D2)을 생성합니다."
#define OPTION_VAL_VMU_A1_KO NULL
#define OPTION_VAL_ALL_VMUS_KO "전체 VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_KO "비주얼 메모리 유닛/시스템 (VMU) 사운드"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_KO "VMU 사운드"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_KO "활성화하면 VMU 신호음이 재생됩니다."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_KO "DreamPotato VMU 저장소 사용"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_KO "VMU 데이터를 연결된 DreamPotato 에뮬레이터에 저장합니다. 활성화할 경우, 로컬 VMU 세이브 파일은 사용되지 않습니다."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_KO "Visual Memory Unit/System(VMU) 디스플레이 설정 표시"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_KO "VMU 디스플레이 설정 표시"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_KO "에뮬레이트되는 VMU LCD 화면의 표시 여부, 크기, 위치, 색상 설정을 표시합니다. 주의: 변경 사항이 적용되려면 빠른 메뉴를 껐다 켜야 할 수도 있습니다."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_KO "VMU 화면 1 표시"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_KO "VMU 화면 1 위치"
#define OPTION_VAL_UPPER_LEFT_KO "왼쪽 위"
#define OPTION_VAL_UPPER_RIGHT_KO "오른쪽 위"
#define OPTION_VAL_LOWER_LEFT_KO "왼쪽 아래"
#define OPTION_VAL_LOWER_RIGHT_KO "오른쪽 아래"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_KO "VMU 화면 1 크기"
#define OPTION_VAL_1X_KO NULL
#define OPTION_VAL_3X_KO NULL
#define OPTION_VAL_5X_KO NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_KO "VMU 화면 1 켜진 픽셀 색상"
#define OPTION_VAL_DEFAULT_ON_00_KO "기본설정 켜짐"
#define OPTION_VAL_DEFAULT_OFF_01_KO "기본설정 꺼짐"
#define OPTION_VAL_BLACK_02_KO "검정"
#define OPTION_VAL_LIGHT_BLUE_04_KO "밝은 파랑"
#define OPTION_VAL_CYAN_06_KO "청록"
#define OPTION_VAL_CYAN_BLUE_07_KO "청록 파랑"
#define OPTION_VAL_LIGHT_GREEN_08_KO "밝은 녹색"
#define OPTION_VAL_CYAN_GREEN_09_KO "청록 초록"
#define OPTION_VAL_LIGHT_CYAN_10_KO "밝은 청록"
#define OPTION_VAL_PURPLE_12_KO "보라"
#define OPTION_VAL_LIGHT_PURPLE_13_KO "밝은 보라"
#define OPTION_VAL_YELLOW_14_KO "노란색"
#define OPTION_VAL_GRAY_15_KO "회색"
#define OPTION_VAL_LIGHT_PURPLE_2_16_KO "연보라색 (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_KO "밝은 녹색 (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_KO "밝은 녹색 (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_KO "밝은 청록 (2)"
#define OPTION_VAL_LIGHT_RED_2_20_KO "밝은 빨강 (2)"
#define OPTION_VAL_MAGENTA_21_KO "자홍색"
#define OPTION_VAL_LIGHT_PURPLE_3_22_KO "연보라색 (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_KO "밝은 주황색"
#define OPTION_VAL_ORANGE_24_KO "오렌지"
#define OPTION_VAL_LIGHT_PURPLE_4_25_KO "연보라색 (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_KO "연한 노란색"
#define OPTION_VAL_LIGHT_YELLOW_2_27_KO "연한 노란색 (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_KO "VMU 화면 1 꺼진 픽셀 색상"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_KO "VMU 화면 1 투명도"
#define OPTION_VAL_40_KO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_KO "VMU 화면 2 표시"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_KO "VMU 화면 2 위치"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_KO "VMU 화면 2 크기"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_KO "VMU 화면 2 켜진 픽셀 색상"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_KO "VMU 화면 2 꺼진 픽셀 색상"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_KO "VMU 화면 2 투명도"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_KO "VMU 화면 3 표시"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_KO "VMU 화면 3 위치"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_KO "VMU 화면 3 크기"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_KO "VMU 화면 3 켜진 픽셀 색상"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_KO "VMU 화면 3 꺼진 픽셀 색상"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_KO "VMU 화면 3 투명도"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_KO "VMU 화면 4 표시"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_KO "VMU 화면 4 위치"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_KO "VMU 화면 4 크기"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_KO "VMU 화면 4 켜진 픽셀 색상"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_KO "VMU 화면 4 꺼진 픽셀 색상"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_KO "VMU 화면 4 투명도"

struct retro_core_option_v2_category option_cats_ko[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_KO,
      CATEGORY_SYSTEM_INFO_0_KO
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_KO,
      CATEGORY_VIDEO_INFO_0_KO
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_KO,
      CATEGORY_PERFORMANCE_INFO_0_KO
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_KO,
      CATEGORY_HACKS_INFO_0_KO
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_KO,
      CATEGORY_INPUT_INFO_0_KO
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_KO,
      CATEGORY_EXPANSIONS_INFO_0_KO
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_KO,
      CATEGORY_VMU_INFO_0_KO
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ko[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_KO,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_KO },
         { "USA",     OPTION_VAL_USA_KO },
         { "Europe",  OPTION_VAL_EUROPE_KO },
         { "Default", OPTION_VAL_DEFAULT_KO },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_KO,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_KO },
         { "English",  OPTION_VAL_ENGLISH_KO },
         { "German",   OPTION_VAL_GERMAN_KO },
         { "French",   OPTION_VAL_FRENCH_KO },
         { "Spanish",  OPTION_VAL_SPANISH_KO },
         { "Italian",  OPTION_VAL_ITALIAN_KO },
         { "Default",  OPTION_VAL_DEFAULT_KO },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_KO,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_KO },
         { "640x480",    OPTION_VAL_640X480_KO },
         { "800x600",    OPTION_VAL_800X600_KO },
         { "960x720",    OPTION_VAL_960X720_KO },
         { "1024x768",   OPTION_VAL_1024X768_KO },
         { "1280x960",   OPTION_VAL_1280X960_KO },
         { "1440x1080",  OPTION_VAL_1440X1080_KO },
         { "1600x1200",  OPTION_VAL_1600X1200_KO },
         { "1920x1440",  OPTION_VAL_1920X1440_KO },
         { "2560x1920",  OPTION_VAL_2560X1920_KO },
         { "2880x2160",  OPTION_VAL_2880X2160_KO },
         { "3200x2400",  OPTION_VAL_3200X2400_KO },
         { "3840x2880",  OPTION_VAL_3840X2880_KO },
         { "4480x3360",  OPTION_VAL_4480X3360_KO },
         { "5120x3840",  OPTION_VAL_5120X3840_KO },
         { "5760x4320",  OPTION_VAL_5760X4320_KO },
         { "6400x4800",  OPTION_VAL_6400X4800_KO },
         { "7040x5280",  OPTION_VAL_7040X5280_KO },
         { "7680x5760",  OPTION_VAL_7680X5760_KO },
         { "8320x6240",  OPTION_VAL_8320X6240_KO },
         { "8960x6720",  OPTION_VAL_8960X6720_KO },
         { "9600x7200",  OPTION_VAL_9600X7200_KO },
         { "10240x7680", OPTION_VAL_10240X7680_KO },
         { "10880x8160", OPTION_VAL_10880X8160_KO },
         { "11520x8640", OPTION_VAL_11520X8640_KO },
         { "12160x9120", OPTION_VAL_12160X9120_KO },
         { "12800x9600", OPTION_VAL_12800X9600_KO },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_KO,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_KO },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_KO },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_KO },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_KO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_KO },
         { "PAL",     OPTION_VAL_PAL_KO },
         { "PAL_N",   OPTION_VAL_PAL_N_KO },
         { "PAL_M",   OPTION_VAL_PAL_M_KO },
         { "Default", OPTION_VAL_DEFAULT_KO },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_KO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_KO },
         { "vertical",   OPTION_VAL_VERTICAL_KO },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_KO,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_KO },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_KO },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_KO },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_KO,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_KO },
         { "1GB",   OPTION_VAL_1GB_KO },
         { "2GB",   OPTION_VAL_2GB_KO },
         { "4GB",   OPTION_VAL_4GB_KO },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_KO,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_KO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_KO,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_KO,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_KO },
         { "1",  OPTION_VAL_1_KO },
         { "2",  OPTION_VAL_2_KO },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_KO,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_KO },
         { "4",  OPTION_VAL_4_KO },
         { "6",  OPTION_VAL_6_KO },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_KO,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_KO,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_KO,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_KO,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_KO,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_KO,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_KO,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_KO },
         { "more", OPTION_VAL_MORE_KO },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_KO,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_KO },
         { "110", OPTION_VAL_110_KO },
         { "120", OPTION_VAL_120_KO },
         { "130", OPTION_VAL_130_KO },
         { "140", OPTION_VAL_140_KO },
         { "150", OPTION_VAL_150_KO },
         { "160", OPTION_VAL_160_KO },
         { "170", OPTION_VAL_170_KO },
         { "180", OPTION_VAL_180_KO },
         { "190", OPTION_VAL_190_KO },
         { "200", OPTION_VAL_200_KO },
         { "210", OPTION_VAL_210_KO },
         { "220", OPTION_VAL_220_KO },
         { "230", OPTION_VAL_230_KO },
         { "240", OPTION_VAL_240_KO },
         { "250", OPTION_VAL_250_KO },
         { "260", OPTION_VAL_260_KO },
         { "270", OPTION_VAL_270_KO },
         { "280", OPTION_VAL_280_KO },
         { "290", OPTION_VAL_290_KO },
         { "300", OPTION_VAL_300_KO },
         { "310", OPTION_VAL_310_KO },
         { "320", OPTION_VAL_320_KO },
         { "330", OPTION_VAL_330_KO },
         { "340", OPTION_VAL_340_KO },
         { "350", OPTION_VAL_350_KO },
         { "360", OPTION_VAL_360_KO },
         { "370", OPTION_VAL_370_KO },
         { "380", OPTION_VAL_380_KO },
         { "390", OPTION_VAL_390_KO },
         { "400", OPTION_VAL_400_KO },
         { "410", OPTION_VAL_410_KO },
         { "420", OPTION_VAL_420_KO },
         { "430", OPTION_VAL_430_KO },
         { "440", OPTION_VAL_440_KO },
         { "450", OPTION_VAL_450_KO },
         { "460", OPTION_VAL_460_KO },
         { "470", OPTION_VAL_470_KO },
         { "480", OPTION_VAL_480_KO },
         { "490", OPTION_VAL_490_KO },
         { "500", OPTION_VAL_500_KO },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_KO,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_KO },
         { "5%",  OPTION_VAL_5_KO },
         { "10%", OPTION_VAL_10_KO },
         { "15%", OPTION_VAL_15_KO },
         { "20%", OPTION_VAL_20_KO },
         { "25%", OPTION_VAL_25_KO },
         { "30%", OPTION_VAL_30_KO },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_KO,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_KO },
         { "5%",  OPTION_VAL_5_KO },
         { "10%", OPTION_VAL_10_KO },
         { "15%", OPTION_VAL_15_KO },
         { "20%", OPTION_VAL_20_KO },
         { "25%", OPTION_VAL_25_KO },
         { "30%", OPTION_VAL_30_KO },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_KO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_KO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_KO,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_KO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_KO },
         { "60%",  OPTION_VAL_60_KO },
         { "70%",  OPTION_VAL_70_KO },
         { "80%",  OPTION_VAL_80_KO },
         { "90%",  OPTION_VAL_90_KO },
         { "100%", OPTION_VAL_100_O47_KO },
         { "110%", OPTION_VAL_110_O47_KO },
         { "120%", OPTION_VAL_120_O47_KO },
         { "130%", OPTION_VAL_130_O47_KO },
         { "140%", OPTION_VAL_140_O47_KO },
         { "150%", OPTION_VAL_150_O47_KO },
         { "160%", OPTION_VAL_160_O47_KO },
         { "170%", OPTION_VAL_170_O47_KO },
         { "180%", OPTION_VAL_180_O47_KO },
         { "190%", OPTION_VAL_190_O47_KO },
         { "200%", OPTION_VAL_200_O47_KO },
         { "210%", OPTION_VAL_210_O47_KO },
         { "220%", OPTION_VAL_220_O47_KO },
         { "230%", OPTION_VAL_230_O47_KO },
         { "240%", OPTION_VAL_240_O47_KO },
         { "250%", OPTION_VAL_250_O47_KO },
         { "260%", OPTION_VAL_260_O47_KO },
         { "270%", OPTION_VAL_270_O47_KO },
         { "280%", OPTION_VAL_280_O47_KO },
         { "290%", OPTION_VAL_290_O47_KO },
         { "300%", OPTION_VAL_300_O47_KO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_KO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_KO },
         { "Red",      OPTION_VAL_RED_KO },
         { "Green",    OPTION_VAL_GREEN_KO },
         { "Blue",     OPTION_VAL_BLUE_KO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_KO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_KO },
         { "Red",      OPTION_VAL_RED_KO },
         { "Green",    OPTION_VAL_GREEN_KO },
         { "Blue",     OPTION_VAL_BLUE_KO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_KO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_KO },
         { "Red",      OPTION_VAL_RED_KO },
         { "Green",    OPTION_VAL_GREEN_KO },
         { "Blue",     OPTION_VAL_BLUE_KO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_KO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_KO },
         { "Red",      OPTION_VAL_RED_KO },
         { "Green",    OPTION_VAL_GREEN_KO },
         { "Blue",     OPTION_VAL_BLUE_KO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_KO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_KO },
         { "Purupuru", OPTION_VAL_PURUPURU_KO },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_KO },
         { "None",     OPTION_VAL_NONE_KO },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_KO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_KO },
         { "Purupuru", OPTION_VAL_PURUPURU_KO },
         { "None",     OPTION_VAL_NONE_KO },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_KO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_KO },
         { "Purupuru", OPTION_VAL_PURUPURU_KO },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_KO },
         { "None",     OPTION_VAL_NONE_KO },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_KO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_KO },
         { "Purupuru", OPTION_VAL_PURUPURU_KO },
         { "None",     OPTION_VAL_NONE_KO },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_KO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_KO },
         { "Purupuru", OPTION_VAL_PURUPURU_KO },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_KO },
         { "None",     OPTION_VAL_NONE_KO },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_KO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_KO },
         { "Purupuru", OPTION_VAL_PURUPURU_KO },
         { "None",     OPTION_VAL_NONE_KO },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_KO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_KO },
         { "Purupuru", OPTION_VAL_PURUPURU_KO },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_KO },
         { "None",     OPTION_VAL_NONE_KO },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_KO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_KO },
         { "Purupuru", OPTION_VAL_PURUPURU_KO },
         { "None",     OPTION_VAL_NONE_KO },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_KO,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_KO,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_KO,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_KO },
         { "All VMUs", OPTION_VAL_ALL_VMUS_KO },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_KO,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_KO,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_KO,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_KO,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_KO,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_KO,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_KO,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_KO,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_KO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_KO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_KO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_KO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_KO },
         { "2x", OPTION_VAL_2_O26_KO },
         { "3x", OPTION_VAL_3X_KO },
         { "4x", OPTION_VAL_4_KO },
         { "5x", OPTION_VAL_5X_KO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_KO },
         { "20%",  OPTION_VAL_20_KO },
         { "30%",  OPTION_VAL_30_KO },
         { "40%",  OPTION_VAL_40_KO },
         { "50%",  OPTION_VAL_50_KO },
         { "60%",  OPTION_VAL_60_KO },
         { "70%",  OPTION_VAL_70_KO },
         { "80%",  OPTION_VAL_80_KO },
         { "90%",  OPTION_VAL_90_KO },
         { "100%", OPTION_VAL_100_O47_KO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_KO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_KO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_KO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_KO },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_KO },
         { "2x", OPTION_VAL_2_O26_KO },
         { "3x", OPTION_VAL_3X_KO },
         { "4x", OPTION_VAL_4_KO },
         { "5x", OPTION_VAL_5X_KO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_KO },
         { "20%",  OPTION_VAL_20_KO },
         { "30%",  OPTION_VAL_30_KO },
         { "40%",  OPTION_VAL_40_KO },
         { "50%",  OPTION_VAL_50_KO },
         { "60%",  OPTION_VAL_60_KO },
         { "70%",  OPTION_VAL_70_KO },
         { "80%",  OPTION_VAL_80_KO },
         { "90%",  OPTION_VAL_90_KO },
         { "100%", OPTION_VAL_100_O47_KO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_KO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_KO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_KO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_KO },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_KO },
         { "2x", OPTION_VAL_2_O26_KO },
         { "3x", OPTION_VAL_3X_KO },
         { "4x", OPTION_VAL_4_KO },
         { "5x", OPTION_VAL_5X_KO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_KO },
         { "20%",  OPTION_VAL_20_KO },
         { "30%",  OPTION_VAL_30_KO },
         { "40%",  OPTION_VAL_40_KO },
         { "50%",  OPTION_VAL_50_KO },
         { "60%",  OPTION_VAL_60_KO },
         { "70%",  OPTION_VAL_70_KO },
         { "80%",  OPTION_VAL_80_KO },
         { "90%",  OPTION_VAL_90_KO },
         { "100%", OPTION_VAL_100_O47_KO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_KO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_KO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_KO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_KO },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_KO },
         { "2x", OPTION_VAL_2_O26_KO },
         { "3x", OPTION_VAL_3X_KO },
         { "4x", OPTION_VAL_4_KO },
         { "5x", OPTION_VAL_5X_KO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_KO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_KO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_KO },
         { "BLUE 03",           OPTION_VAL_BLUE_KO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_KO },
         { "GREEN 05",          OPTION_VAL_GREEN_KO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_KO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_KO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_KO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_KO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_KO },
         { "RED 11",            OPTION_VAL_RED_KO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_KO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_KO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_KO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_KO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_KO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_KO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_KO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_KO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_KO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_KO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_KO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_KO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_KO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_KO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_KO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_KO },
         { "WHITE 28",          OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_KO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_KO },
         { "20%",  OPTION_VAL_20_KO },
         { "30%",  OPTION_VAL_30_KO },
         { "40%",  OPTION_VAL_40_KO },
         { "50%",  OPTION_VAL_50_KO },
         { "60%",  OPTION_VAL_60_KO },
         { "70%",  OPTION_VAL_70_KO },
         { "80%",  OPTION_VAL_80_KO },
         { "90%",  OPTION_VAL_90_KO },
         { "100%", OPTION_VAL_100_O47_KO },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ko = {
   option_cats_ko,
   option_defs_ko
};

/* RETRO_LANGUAGE_NL */

#define CATEGORY_SYSTEM_LABEL_NL "Systeem"
#define CATEGORY_SYSTEM_INFO_0_NL NULL
#define CATEGORY_VIDEO_LABEL_NL NULL
#define CATEGORY_VIDEO_INFO_0_NL NULL
#define CATEGORY_PERFORMANCE_LABEL_NL NULL
#define CATEGORY_PERFORMANCE_INFO_0_NL NULL
#define CATEGORY_HACKS_LABEL_NL NULL
#define CATEGORY_HACKS_INFO_0_NL NULL
#define CATEGORY_INPUT_LABEL_NL "Invoer"
#define CATEGORY_INPUT_INFO_0_NL NULL
#define CATEGORY_EXPANSIONS_LABEL_NL NULL
#define CATEGORY_EXPANSIONS_INFO_0_NL NULL
#define CATEGORY_VMU_LABEL_NL NULL
#define CATEGORY_VMU_INFO_0_NL NULL
#define CORE_OPTION_NAME_REGION_LABEL_NL "Regio"
#define OPTION_VAL_JAPAN_NL NULL
#define OPTION_VAL_USA_NL NULL
#define OPTION_VAL_EUROPE_NL NULL
#define OPTION_VAL_DEFAULT_NL "Standaard"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_NL "Taal"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_NL NULL
#define OPTION_VAL_JAPANESE_NL "Japans"
#define OPTION_VAL_ENGLISH_NL "Engels"
#define OPTION_VAL_GERMAN_NL "Duits"
#define OPTION_VAL_FRENCH_NL "Frans"
#define OPTION_VAL_SPANISH_NL "Spaans"
#define OPTION_VAL_ITALIAN_NL "Italiaans"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_NL NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_NL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_NL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_NL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_NL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_NL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_NL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_NL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_NL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_NL NULL
#define CORE_OPTION_NAME_UPNP_LABEL_NL NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_NL NULL
#define CORE_OPTION_NAME_DCNET_LABEL_NL NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_NL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_NL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_NL NULL
#define OPTION_VAL_320X240_NL NULL
#define OPTION_VAL_640X480_NL NULL
#define OPTION_VAL_800X600_NL NULL
#define OPTION_VAL_960X720_NL NULL
#define OPTION_VAL_1024X768_NL NULL
#define OPTION_VAL_1280X960_NL NULL
#define OPTION_VAL_1440X1080_NL NULL
#define OPTION_VAL_1600X1200_NL NULL
#define OPTION_VAL_1920X1440_NL NULL
#define OPTION_VAL_2560X1920_NL NULL
#define OPTION_VAL_2880X2160_NL NULL
#define OPTION_VAL_3200X2400_NL NULL
#define OPTION_VAL_3840X2880_NL NULL
#define OPTION_VAL_4480X3360_NL NULL
#define OPTION_VAL_5120X3840_NL NULL
#define OPTION_VAL_5760X4320_NL NULL
#define OPTION_VAL_6400X4800_NL NULL
#define OPTION_VAL_7040X5280_NL NULL
#define OPTION_VAL_7680X5760_NL NULL
#define OPTION_VAL_8320X6240_NL NULL
#define OPTION_VAL_8960X6720_NL NULL
#define OPTION_VAL_9600X7200_NL NULL
#define OPTION_VAL_10240X7680_NL NULL
#define OPTION_VAL_10880X8160_NL NULL
#define OPTION_VAL_11520X8640_NL NULL
#define OPTION_VAL_12160X9120_NL NULL
#define OPTION_VAL_12800X9600_NL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_NL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_NL NULL
#define OPTION_VAL_VGA_NL NULL
#define OPTION_VAL_TV_RGB_NL NULL
#define OPTION_VAL_TV_COMPOSITE_NL NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_NL NULL
#define OPTION_VAL_NTSC_NL NULL
#define OPTION_VAL_PAL_NL NULL
#define OPTION_VAL_PAL_N_NL NULL
#define OPTION_VAL_PAL_M_NL NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_NL "Schermoriëntatie"
#define OPTION_VAL_HORIZONTAL_NL NULL
#define OPTION_VAL_VERTICAL_NL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_NL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_NL NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_NL NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_NL NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_NL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_NL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_NL NULL
#define OPTION_VAL_512MB_NL NULL
#define OPTION_VAL_1GB_NL NULL
#define OPTION_VAL_2GB_NL NULL
#define OPTION_VAL_4GB_NL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_NL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_NL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_NL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_NL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_NL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_NL NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_NL NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_NL NULL
#define CORE_OPTION_NAME_FOG_LABEL_NL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_NL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_NL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_NL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_NL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_NL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_NL NULL
#define OPTION_VAL_1_NL NULL
#define OPTION_VAL_2_NL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_NL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_NL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_NL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_NL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_NL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_NL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_NL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_NL NULL
#define OPTION_VAL_2_O26_NL NULL
#define OPTION_VAL_4_NL NULL
#define OPTION_VAL_6_NL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_NL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_NL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_NL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_NL NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_NL NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_NL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_NL "Gedraaide Rendering"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_NL "Voert de GPU en CPU uit op verschillende threads. Zeer aanbevolen."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_NL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_NL "Automatisch frames overslaan wanneer de emulator traag draait. Deze instelling is alleen van toepassing wanneer 'Gedraaide Rendering' is ingeschakeld."
#define OPTION_VAL_SOME_NL "Normaal"
#define OPTION_VAL_MORE_NL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_NL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_NL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_NL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_NL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_NL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_NL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_NL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_NL NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_NL NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_NL NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_NL NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_NL NULL
#define OPTION_VAL_100_NL NULL
#define OPTION_VAL_110_NL NULL
#define OPTION_VAL_120_NL NULL
#define OPTION_VAL_130_NL NULL
#define OPTION_VAL_140_NL NULL
#define OPTION_VAL_150_NL NULL
#define OPTION_VAL_160_NL NULL
#define OPTION_VAL_170_NL NULL
#define OPTION_VAL_180_NL NULL
#define OPTION_VAL_190_NL NULL
#define OPTION_VAL_200_NL NULL
#define OPTION_VAL_210_NL NULL
#define OPTION_VAL_220_NL NULL
#define OPTION_VAL_230_NL NULL
#define OPTION_VAL_240_NL NULL
#define OPTION_VAL_250_NL NULL
#define OPTION_VAL_260_NL NULL
#define OPTION_VAL_270_NL NULL
#define OPTION_VAL_280_NL NULL
#define OPTION_VAL_290_NL NULL
#define OPTION_VAL_300_NL NULL
#define OPTION_VAL_310_NL NULL
#define OPTION_VAL_320_NL NULL
#define OPTION_VAL_330_NL NULL
#define OPTION_VAL_340_NL NULL
#define OPTION_VAL_350_NL NULL
#define OPTION_VAL_360_NL NULL
#define OPTION_VAL_370_NL NULL
#define OPTION_VAL_380_NL NULL
#define OPTION_VAL_390_NL NULL
#define OPTION_VAL_400_NL NULL
#define OPTION_VAL_410_NL NULL
#define OPTION_VAL_420_NL NULL
#define OPTION_VAL_430_NL NULL
#define OPTION_VAL_440_NL NULL
#define OPTION_VAL_450_NL NULL
#define OPTION_VAL_460_NL NULL
#define OPTION_VAL_470_NL NULL
#define OPTION_VAL_480_NL NULL
#define OPTION_VAL_490_NL NULL
#define OPTION_VAL_500_NL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_NL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_NL NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_NL NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_NL NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_NL NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_NL NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_NL NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_NL NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_NL NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_NL NULL
#define OPTION_VAL_0_NL NULL
#define OPTION_VAL_5_NL NULL
#define OPTION_VAL_10_NL NULL
#define OPTION_VAL_15_NL NULL
#define OPTION_VAL_20_NL NULL
#define OPTION_VAL_25_NL NULL
#define OPTION_VAL_30_NL NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_NL NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_NL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_NL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_NL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_NL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_NL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_NL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_NL NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_NL NULL
#define OPTION_VAL_50_NL NULL
#define OPTION_VAL_60_NL NULL
#define OPTION_VAL_70_NL NULL
#define OPTION_VAL_80_NL NULL
#define OPTION_VAL_90_NL NULL
#define OPTION_VAL_100_O47_NL NULL
#define OPTION_VAL_110_O47_NL NULL
#define OPTION_VAL_120_O47_NL NULL
#define OPTION_VAL_130_O47_NL NULL
#define OPTION_VAL_140_O47_NL NULL
#define OPTION_VAL_150_O47_NL NULL
#define OPTION_VAL_160_O47_NL NULL
#define OPTION_VAL_170_O47_NL NULL
#define OPTION_VAL_180_O47_NL NULL
#define OPTION_VAL_190_O47_NL NULL
#define OPTION_VAL_200_O47_NL NULL
#define OPTION_VAL_210_O47_NL NULL
#define OPTION_VAL_220_O47_NL NULL
#define OPTION_VAL_230_O47_NL NULL
#define OPTION_VAL_240_O47_NL NULL
#define OPTION_VAL_250_O47_NL NULL
#define OPTION_VAL_260_O47_NL NULL
#define OPTION_VAL_270_O47_NL NULL
#define OPTION_VAL_280_O47_NL NULL
#define OPTION_VAL_290_O47_NL NULL
#define OPTION_VAL_300_O47_NL NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_NL NULL
#define OPTION_VAL_WHITE_NL NULL
#define OPTION_VAL_RED_NL "Rood"
#define OPTION_VAL_GREEN_NL "Groen"
#define OPTION_VAL_BLUE_NL "Blauw"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_NL NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_NL NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_NL NULL
#define OPTION_VAL_VMU_NL NULL
#define OPTION_VAL_PURUPURU_NL NULL
#define OPTION_VAL_DREAMPOTATO_NL NULL
#define OPTION_VAL_NONE_NL "Geen"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_NL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_NL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_NL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_NL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_NL NULL
#define OPTION_VAL_VMU_A1_NL NULL
#define OPTION_VAL_ALL_VMUS_NL NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_NL NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_NL NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_NL NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_NL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_NL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_NL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_NL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_NL NULL
#define OPTION_VAL_UPPER_LEFT_NL NULL
#define OPTION_VAL_UPPER_RIGHT_NL NULL
#define OPTION_VAL_LOWER_LEFT_NL NULL
#define OPTION_VAL_LOWER_RIGHT_NL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_NL NULL
#define OPTION_VAL_1X_NL NULL
#define OPTION_VAL_3X_NL NULL
#define OPTION_VAL_5X_NL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_NL NULL
#define OPTION_VAL_DEFAULT_ON_00_NL NULL
#define OPTION_VAL_DEFAULT_OFF_01_NL NULL
#define OPTION_VAL_BLACK_02_NL NULL
#define OPTION_VAL_LIGHT_BLUE_04_NL NULL
#define OPTION_VAL_CYAN_06_NL NULL
#define OPTION_VAL_CYAN_BLUE_07_NL NULL
#define OPTION_VAL_LIGHT_GREEN_08_NL NULL
#define OPTION_VAL_CYAN_GREEN_09_NL NULL
#define OPTION_VAL_LIGHT_CYAN_10_NL NULL
#define OPTION_VAL_PURPLE_12_NL NULL
#define OPTION_VAL_LIGHT_PURPLE_13_NL NULL
#define OPTION_VAL_YELLOW_14_NL "Geel"
#define OPTION_VAL_GRAY_15_NL NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_NL NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_NL NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_NL NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_NL NULL
#define OPTION_VAL_LIGHT_RED_2_20_NL NULL
#define OPTION_VAL_MAGENTA_21_NL NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_NL NULL
#define OPTION_VAL_LIGHT_ORANGE_23_NL NULL
#define OPTION_VAL_ORANGE_24_NL NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_NL NULL
#define OPTION_VAL_LIGHT_YELLOW_26_NL NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_NL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_NL NULL
#define OPTION_VAL_40_NL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_NL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_NL NULL

struct retro_core_option_v2_category option_cats_nl[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_NL,
      CATEGORY_SYSTEM_INFO_0_NL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_NL,
      CATEGORY_VIDEO_INFO_0_NL
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_NL,
      CATEGORY_PERFORMANCE_INFO_0_NL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_NL,
      CATEGORY_HACKS_INFO_0_NL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_NL,
      CATEGORY_INPUT_INFO_0_NL
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_NL,
      CATEGORY_EXPANSIONS_INFO_0_NL
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_NL,
      CATEGORY_VMU_INFO_0_NL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_nl[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_NL,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_NL },
         { "USA",     OPTION_VAL_USA_NL },
         { "Europe",  OPTION_VAL_EUROPE_NL },
         { "Default", OPTION_VAL_DEFAULT_NL },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_NL,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_NL },
         { "English",  OPTION_VAL_ENGLISH_NL },
         { "German",   OPTION_VAL_GERMAN_NL },
         { "French",   OPTION_VAL_FRENCH_NL },
         { "Spanish",  OPTION_VAL_SPANISH_NL },
         { "Italian",  OPTION_VAL_ITALIAN_NL },
         { "Default",  OPTION_VAL_DEFAULT_NL },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_NL,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_NL },
         { "640x480",    OPTION_VAL_640X480_NL },
         { "800x600",    OPTION_VAL_800X600_NL },
         { "960x720",    OPTION_VAL_960X720_NL },
         { "1024x768",   OPTION_VAL_1024X768_NL },
         { "1280x960",   OPTION_VAL_1280X960_NL },
         { "1440x1080",  OPTION_VAL_1440X1080_NL },
         { "1600x1200",  OPTION_VAL_1600X1200_NL },
         { "1920x1440",  OPTION_VAL_1920X1440_NL },
         { "2560x1920",  OPTION_VAL_2560X1920_NL },
         { "2880x2160",  OPTION_VAL_2880X2160_NL },
         { "3200x2400",  OPTION_VAL_3200X2400_NL },
         { "3840x2880",  OPTION_VAL_3840X2880_NL },
         { "4480x3360",  OPTION_VAL_4480X3360_NL },
         { "5120x3840",  OPTION_VAL_5120X3840_NL },
         { "5760x4320",  OPTION_VAL_5760X4320_NL },
         { "6400x4800",  OPTION_VAL_6400X4800_NL },
         { "7040x5280",  OPTION_VAL_7040X5280_NL },
         { "7680x5760",  OPTION_VAL_7680X5760_NL },
         { "8320x6240",  OPTION_VAL_8320X6240_NL },
         { "8960x6720",  OPTION_VAL_8960X6720_NL },
         { "9600x7200",  OPTION_VAL_9600X7200_NL },
         { "10240x7680", OPTION_VAL_10240X7680_NL },
         { "10880x8160", OPTION_VAL_10880X8160_NL },
         { "11520x8640", OPTION_VAL_11520X8640_NL },
         { "12160x9120", OPTION_VAL_12160X9120_NL },
         { "12800x9600", OPTION_VAL_12800X9600_NL },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_NL,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_NL },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_NL },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_NL },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_NL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_NL },
         { "PAL",     OPTION_VAL_PAL_NL },
         { "PAL_N",   OPTION_VAL_PAL_N_NL },
         { "PAL_M",   OPTION_VAL_PAL_M_NL },
         { "Default", OPTION_VAL_DEFAULT_NL },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_NL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_NL },
         { "vertical",   OPTION_VAL_VERTICAL_NL },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_NL,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_NL },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_NL },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_NL },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_NL,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_NL },
         { "1GB",   OPTION_VAL_1GB_NL },
         { "2GB",   OPTION_VAL_2GB_NL },
         { "4GB",   OPTION_VAL_4GB_NL },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_NL,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_NL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_NL,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_NL,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_NL },
         { "1",  OPTION_VAL_1_NL },
         { "2",  OPTION_VAL_2_NL },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_NL,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_NL },
         { "4",  OPTION_VAL_4_NL },
         { "6",  OPTION_VAL_6_NL },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_NL,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_NL,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_NL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_NL,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_NL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_NL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_NL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_NL },
         { "more", OPTION_VAL_MORE_NL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_NL,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_NL },
         { "110", OPTION_VAL_110_NL },
         { "120", OPTION_VAL_120_NL },
         { "130", OPTION_VAL_130_NL },
         { "140", OPTION_VAL_140_NL },
         { "150", OPTION_VAL_150_NL },
         { "160", OPTION_VAL_160_NL },
         { "170", OPTION_VAL_170_NL },
         { "180", OPTION_VAL_180_NL },
         { "190", OPTION_VAL_190_NL },
         { "200", OPTION_VAL_200_NL },
         { "210", OPTION_VAL_210_NL },
         { "220", OPTION_VAL_220_NL },
         { "230", OPTION_VAL_230_NL },
         { "240", OPTION_VAL_240_NL },
         { "250", OPTION_VAL_250_NL },
         { "260", OPTION_VAL_260_NL },
         { "270", OPTION_VAL_270_NL },
         { "280", OPTION_VAL_280_NL },
         { "290", OPTION_VAL_290_NL },
         { "300", OPTION_VAL_300_NL },
         { "310", OPTION_VAL_310_NL },
         { "320", OPTION_VAL_320_NL },
         { "330", OPTION_VAL_330_NL },
         { "340", OPTION_VAL_340_NL },
         { "350", OPTION_VAL_350_NL },
         { "360", OPTION_VAL_360_NL },
         { "370", OPTION_VAL_370_NL },
         { "380", OPTION_VAL_380_NL },
         { "390", OPTION_VAL_390_NL },
         { "400", OPTION_VAL_400_NL },
         { "410", OPTION_VAL_410_NL },
         { "420", OPTION_VAL_420_NL },
         { "430", OPTION_VAL_430_NL },
         { "440", OPTION_VAL_440_NL },
         { "450", OPTION_VAL_450_NL },
         { "460", OPTION_VAL_460_NL },
         { "470", OPTION_VAL_470_NL },
         { "480", OPTION_VAL_480_NL },
         { "490", OPTION_VAL_490_NL },
         { "500", OPTION_VAL_500_NL },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_NL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_NL },
         { "5%",  OPTION_VAL_5_NL },
         { "10%", OPTION_VAL_10_NL },
         { "15%", OPTION_VAL_15_NL },
         { "20%", OPTION_VAL_20_NL },
         { "25%", OPTION_VAL_25_NL },
         { "30%", OPTION_VAL_30_NL },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_NL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_NL },
         { "5%",  OPTION_VAL_5_NL },
         { "10%", OPTION_VAL_10_NL },
         { "15%", OPTION_VAL_15_NL },
         { "20%", OPTION_VAL_20_NL },
         { "25%", OPTION_VAL_25_NL },
         { "30%", OPTION_VAL_30_NL },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_NL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_NL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_NL,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_NL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_NL },
         { "60%",  OPTION_VAL_60_NL },
         { "70%",  OPTION_VAL_70_NL },
         { "80%",  OPTION_VAL_80_NL },
         { "90%",  OPTION_VAL_90_NL },
         { "100%", OPTION_VAL_100_O47_NL },
         { "110%", OPTION_VAL_110_O47_NL },
         { "120%", OPTION_VAL_120_O47_NL },
         { "130%", OPTION_VAL_130_O47_NL },
         { "140%", OPTION_VAL_140_O47_NL },
         { "150%", OPTION_VAL_150_O47_NL },
         { "160%", OPTION_VAL_160_O47_NL },
         { "170%", OPTION_VAL_170_O47_NL },
         { "180%", OPTION_VAL_180_O47_NL },
         { "190%", OPTION_VAL_190_O47_NL },
         { "200%", OPTION_VAL_200_O47_NL },
         { "210%", OPTION_VAL_210_O47_NL },
         { "220%", OPTION_VAL_220_O47_NL },
         { "230%", OPTION_VAL_230_O47_NL },
         { "240%", OPTION_VAL_240_O47_NL },
         { "250%", OPTION_VAL_250_O47_NL },
         { "260%", OPTION_VAL_260_O47_NL },
         { "270%", OPTION_VAL_270_O47_NL },
         { "280%", OPTION_VAL_280_O47_NL },
         { "290%", OPTION_VAL_290_O47_NL },
         { "300%", OPTION_VAL_300_O47_NL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_NL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NL },
         { "Red",      OPTION_VAL_RED_NL },
         { "Green",    OPTION_VAL_GREEN_NL },
         { "Blue",     OPTION_VAL_BLUE_NL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_NL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NL },
         { "Red",      OPTION_VAL_RED_NL },
         { "Green",    OPTION_VAL_GREEN_NL },
         { "Blue",     OPTION_VAL_BLUE_NL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_NL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NL },
         { "Red",      OPTION_VAL_RED_NL },
         { "Green",    OPTION_VAL_GREEN_NL },
         { "Blue",     OPTION_VAL_BLUE_NL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_NL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NL },
         { "Red",      OPTION_VAL_RED_NL },
         { "Green",    OPTION_VAL_GREEN_NL },
         { "Blue",     OPTION_VAL_BLUE_NL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_NL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NL },
         { "Purupuru", OPTION_VAL_PURUPURU_NL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_NL },
         { "None",     OPTION_VAL_NONE_NL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_NL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NL },
         { "Purupuru", OPTION_VAL_PURUPURU_NL },
         { "None",     OPTION_VAL_NONE_NL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_NL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NL },
         { "Purupuru", OPTION_VAL_PURUPURU_NL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_NL },
         { "None",     OPTION_VAL_NONE_NL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_NL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NL },
         { "Purupuru", OPTION_VAL_PURUPURU_NL },
         { "None",     OPTION_VAL_NONE_NL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_NL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NL },
         { "Purupuru", OPTION_VAL_PURUPURU_NL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_NL },
         { "None",     OPTION_VAL_NONE_NL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_NL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NL },
         { "Purupuru", OPTION_VAL_PURUPURU_NL },
         { "None",     OPTION_VAL_NONE_NL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_NL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NL },
         { "Purupuru", OPTION_VAL_PURUPURU_NL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_NL },
         { "None",     OPTION_VAL_NONE_NL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_NL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NL },
         { "Purupuru", OPTION_VAL_PURUPURU_NL },
         { "None",     OPTION_VAL_NONE_NL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_NL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_NL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_NL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_NL },
         { "All VMUs", OPTION_VAL_ALL_VMUS_NL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_NL,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_NL,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_NL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_NL,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_NL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_NL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_NL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_NL,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NL },
         { "2x", OPTION_VAL_2_O26_NL },
         { "3x", OPTION_VAL_3X_NL },
         { "4x", OPTION_VAL_4_NL },
         { "5x", OPTION_VAL_5X_NL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NL },
         { "20%",  OPTION_VAL_20_NL },
         { "30%",  OPTION_VAL_30_NL },
         { "40%",  OPTION_VAL_40_NL },
         { "50%",  OPTION_VAL_50_NL },
         { "60%",  OPTION_VAL_60_NL },
         { "70%",  OPTION_VAL_70_NL },
         { "80%",  OPTION_VAL_80_NL },
         { "90%",  OPTION_VAL_90_NL },
         { "100%", OPTION_VAL_100_O47_NL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NL },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NL },
         { "2x", OPTION_VAL_2_O26_NL },
         { "3x", OPTION_VAL_3X_NL },
         { "4x", OPTION_VAL_4_NL },
         { "5x", OPTION_VAL_5X_NL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NL },
         { "20%",  OPTION_VAL_20_NL },
         { "30%",  OPTION_VAL_30_NL },
         { "40%",  OPTION_VAL_40_NL },
         { "50%",  OPTION_VAL_50_NL },
         { "60%",  OPTION_VAL_60_NL },
         { "70%",  OPTION_VAL_70_NL },
         { "80%",  OPTION_VAL_80_NL },
         { "90%",  OPTION_VAL_90_NL },
         { "100%", OPTION_VAL_100_O47_NL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NL },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NL },
         { "2x", OPTION_VAL_2_O26_NL },
         { "3x", OPTION_VAL_3X_NL },
         { "4x", OPTION_VAL_4_NL },
         { "5x", OPTION_VAL_5X_NL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NL },
         { "20%",  OPTION_VAL_20_NL },
         { "30%",  OPTION_VAL_30_NL },
         { "40%",  OPTION_VAL_40_NL },
         { "50%",  OPTION_VAL_50_NL },
         { "60%",  OPTION_VAL_60_NL },
         { "70%",  OPTION_VAL_70_NL },
         { "80%",  OPTION_VAL_80_NL },
         { "90%",  OPTION_VAL_90_NL },
         { "100%", OPTION_VAL_100_O47_NL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NL },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NL },
         { "2x", OPTION_VAL_2_O26_NL },
         { "3x", OPTION_VAL_3X_NL },
         { "4x", OPTION_VAL_4_NL },
         { "5x", OPTION_VAL_5X_NL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NL },
         { "BLUE 03",           OPTION_VAL_BLUE_NL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NL },
         { "GREEN 05",          OPTION_VAL_GREEN_NL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NL },
         { "RED 11",            OPTION_VAL_RED_NL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NL },
         { "WHITE 28",          OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_NL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NL },
         { "20%",  OPTION_VAL_20_NL },
         { "30%",  OPTION_VAL_30_NL },
         { "40%",  OPTION_VAL_40_NL },
         { "50%",  OPTION_VAL_50_NL },
         { "60%",  OPTION_VAL_60_NL },
         { "70%",  OPTION_VAL_70_NL },
         { "80%",  OPTION_VAL_80_NL },
         { "90%",  OPTION_VAL_90_NL },
         { "100%", OPTION_VAL_100_O47_NL },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_nl = {
   option_cats_nl,
   option_defs_nl
};

/* RETRO_LANGUAGE_NO */

#define CATEGORY_SYSTEM_LABEL_NO NULL
#define CATEGORY_SYSTEM_INFO_0_NO NULL
#define CATEGORY_VIDEO_LABEL_NO NULL
#define CATEGORY_VIDEO_INFO_0_NO NULL
#define CATEGORY_PERFORMANCE_LABEL_NO NULL
#define CATEGORY_PERFORMANCE_INFO_0_NO NULL
#define CATEGORY_HACKS_LABEL_NO NULL
#define CATEGORY_HACKS_INFO_0_NO NULL
#define CATEGORY_INPUT_LABEL_NO "Inndata"
#define CATEGORY_INPUT_INFO_0_NO NULL
#define CATEGORY_EXPANSIONS_LABEL_NO NULL
#define CATEGORY_EXPANSIONS_INFO_0_NO NULL
#define CATEGORY_VMU_LABEL_NO NULL
#define CATEGORY_VMU_INFO_0_NO NULL
#define CORE_OPTION_NAME_REGION_LABEL_NO NULL
#define OPTION_VAL_JAPAN_NO NULL
#define OPTION_VAL_USA_NO NULL
#define OPTION_VAL_EUROPE_NO NULL
#define OPTION_VAL_DEFAULT_NO NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_NO "Språk"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_NO NULL
#define OPTION_VAL_JAPANESE_NO NULL
#define OPTION_VAL_ENGLISH_NO NULL
#define OPTION_VAL_GERMAN_NO NULL
#define OPTION_VAL_FRENCH_NO NULL
#define OPTION_VAL_SPANISH_NO NULL
#define OPTION_VAL_ITALIAN_NO NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_NO NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_NO NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_NO NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_NO NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_NO NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_NO NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_NO NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_NO NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_NO NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_NO NULL
#define CORE_OPTION_NAME_UPNP_LABEL_NO NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_NO NULL
#define CORE_OPTION_NAME_DCNET_LABEL_NO NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_NO NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_NO "Intern oppløsning"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_NO NULL
#define OPTION_VAL_320X240_NO NULL
#define OPTION_VAL_640X480_NO NULL
#define OPTION_VAL_800X600_NO NULL
#define OPTION_VAL_960X720_NO NULL
#define OPTION_VAL_1024X768_NO NULL
#define OPTION_VAL_1280X960_NO NULL
#define OPTION_VAL_1440X1080_NO NULL
#define OPTION_VAL_1600X1200_NO NULL
#define OPTION_VAL_1920X1440_NO NULL
#define OPTION_VAL_2560X1920_NO NULL
#define OPTION_VAL_2880X2160_NO NULL
#define OPTION_VAL_3200X2400_NO NULL
#define OPTION_VAL_3840X2880_NO NULL
#define OPTION_VAL_4480X3360_NO NULL
#define OPTION_VAL_5120X3840_NO NULL
#define OPTION_VAL_5760X4320_NO NULL
#define OPTION_VAL_6400X4800_NO NULL
#define OPTION_VAL_7040X5280_NO NULL
#define OPTION_VAL_7680X5760_NO NULL
#define OPTION_VAL_8320X6240_NO NULL
#define OPTION_VAL_8960X6720_NO NULL
#define OPTION_VAL_9600X7200_NO NULL
#define OPTION_VAL_10240X7680_NO NULL
#define OPTION_VAL_10880X8160_NO NULL
#define OPTION_VAL_11520X8640_NO NULL
#define OPTION_VAL_12160X9120_NO NULL
#define OPTION_VAL_12800X9600_NO NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_NO NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_NO NULL
#define OPTION_VAL_VGA_NO NULL
#define OPTION_VAL_TV_RGB_NO NULL
#define OPTION_VAL_TV_COMPOSITE_NO NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_NO NULL
#define OPTION_VAL_NTSC_NO NULL
#define OPTION_VAL_PAL_NO NULL
#define OPTION_VAL_PAL_N_NO NULL
#define OPTION_VAL_PAL_M_NO NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_NO "Skjermorientering"
#define OPTION_VAL_HORIZONTAL_NO NULL
#define OPTION_VAL_VERTICAL_NO NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_NO NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_NO NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_NO NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_NO NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_NO NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_NO NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_NO NULL
#define OPTION_VAL_512MB_NO NULL
#define OPTION_VAL_1GB_NO NULL
#define OPTION_VAL_2GB_NO NULL
#define OPTION_VAL_4GB_NO NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_NO NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_NO NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_NO NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_NO NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_NO NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_NO NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_NO NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_NO NULL
#define CORE_OPTION_NAME_FOG_LABEL_NO NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_NO NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_NO NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_NO NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_NO NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_NO NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_NO NULL
#define OPTION_VAL_1_NO NULL
#define OPTION_VAL_2_NO NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_NO NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_NO NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_NO NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_NO NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_NO NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_NO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_NO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_NO NULL
#define OPTION_VAL_2_O26_NO NULL
#define OPTION_VAL_4_NO NULL
#define OPTION_VAL_6_NO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_NO NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_NO NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_NO NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_NO NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_NO NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_NO NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_NO NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_NO NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_NO NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_NO NULL
#define OPTION_VAL_SOME_NO NULL
#define OPTION_VAL_MORE_NO NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_NO NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_NO NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_NO NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_NO NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_NO NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_NO NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_NO NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_NO NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_NO NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_NO NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_NO NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_NO NULL
#define OPTION_VAL_100_NO NULL
#define OPTION_VAL_110_NO NULL
#define OPTION_VAL_120_NO NULL
#define OPTION_VAL_130_NO NULL
#define OPTION_VAL_140_NO NULL
#define OPTION_VAL_150_NO NULL
#define OPTION_VAL_160_NO NULL
#define OPTION_VAL_170_NO NULL
#define OPTION_VAL_180_NO NULL
#define OPTION_VAL_190_NO NULL
#define OPTION_VAL_200_NO NULL
#define OPTION_VAL_210_NO NULL
#define OPTION_VAL_220_NO NULL
#define OPTION_VAL_230_NO NULL
#define OPTION_VAL_240_NO NULL
#define OPTION_VAL_250_NO NULL
#define OPTION_VAL_260_NO NULL
#define OPTION_VAL_270_NO NULL
#define OPTION_VAL_280_NO NULL
#define OPTION_VAL_290_NO NULL
#define OPTION_VAL_300_NO NULL
#define OPTION_VAL_310_NO NULL
#define OPTION_VAL_320_NO NULL
#define OPTION_VAL_330_NO NULL
#define OPTION_VAL_340_NO NULL
#define OPTION_VAL_350_NO NULL
#define OPTION_VAL_360_NO NULL
#define OPTION_VAL_370_NO NULL
#define OPTION_VAL_380_NO NULL
#define OPTION_VAL_390_NO NULL
#define OPTION_VAL_400_NO NULL
#define OPTION_VAL_410_NO NULL
#define OPTION_VAL_420_NO NULL
#define OPTION_VAL_430_NO NULL
#define OPTION_VAL_440_NO NULL
#define OPTION_VAL_450_NO NULL
#define OPTION_VAL_460_NO NULL
#define OPTION_VAL_470_NO NULL
#define OPTION_VAL_480_NO NULL
#define OPTION_VAL_490_NO NULL
#define OPTION_VAL_500_NO NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_NO NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_NO NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_NO NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_NO NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_NO NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_NO NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_NO NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_NO NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_NO NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_NO NULL
#define OPTION_VAL_0_NO NULL
#define OPTION_VAL_5_NO NULL
#define OPTION_VAL_10_NO NULL
#define OPTION_VAL_15_NO NULL
#define OPTION_VAL_20_NO NULL
#define OPTION_VAL_25_NO NULL
#define OPTION_VAL_30_NO NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_NO NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_NO NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_NO NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_NO NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_NO NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_NO NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_NO NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_NO NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_NO NULL
#define OPTION_VAL_50_NO NULL
#define OPTION_VAL_60_NO NULL
#define OPTION_VAL_70_NO NULL
#define OPTION_VAL_80_NO NULL
#define OPTION_VAL_90_NO NULL
#define OPTION_VAL_100_O47_NO NULL
#define OPTION_VAL_110_O47_NO NULL
#define OPTION_VAL_120_O47_NO NULL
#define OPTION_VAL_130_O47_NO NULL
#define OPTION_VAL_140_O47_NO NULL
#define OPTION_VAL_150_O47_NO NULL
#define OPTION_VAL_160_O47_NO NULL
#define OPTION_VAL_170_O47_NO NULL
#define OPTION_VAL_180_O47_NO NULL
#define OPTION_VAL_190_O47_NO NULL
#define OPTION_VAL_200_O47_NO NULL
#define OPTION_VAL_210_O47_NO NULL
#define OPTION_VAL_220_O47_NO NULL
#define OPTION_VAL_230_O47_NO NULL
#define OPTION_VAL_240_O47_NO NULL
#define OPTION_VAL_250_O47_NO NULL
#define OPTION_VAL_260_O47_NO NULL
#define OPTION_VAL_270_O47_NO NULL
#define OPTION_VAL_280_O47_NO NULL
#define OPTION_VAL_290_O47_NO NULL
#define OPTION_VAL_300_O47_NO NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_NO NULL
#define OPTION_VAL_WHITE_NO NULL
#define OPTION_VAL_RED_NO "Rød"
#define OPTION_VAL_GREEN_NO "Grønn"
#define OPTION_VAL_BLUE_NO "Blå"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_NO NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_NO NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_NO NULL
#define OPTION_VAL_VMU_NO NULL
#define OPTION_VAL_PURUPURU_NO NULL
#define OPTION_VAL_DREAMPOTATO_NO NULL
#define OPTION_VAL_NONE_NO "Ingen"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_NO NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_NO NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_NO NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_NO NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_NO NULL
#define OPTION_VAL_VMU_A1_NO NULL
#define OPTION_VAL_ALL_VMUS_NO NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_NO NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_NO NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_NO NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_NO NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_NO NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_NO NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_NO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_NO NULL
#define OPTION_VAL_UPPER_LEFT_NO NULL
#define OPTION_VAL_UPPER_RIGHT_NO NULL
#define OPTION_VAL_LOWER_LEFT_NO NULL
#define OPTION_VAL_LOWER_RIGHT_NO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_NO NULL
#define OPTION_VAL_1X_NO NULL
#define OPTION_VAL_3X_NO NULL
#define OPTION_VAL_5X_NO NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_NO NULL
#define OPTION_VAL_DEFAULT_ON_00_NO NULL
#define OPTION_VAL_DEFAULT_OFF_01_NO NULL
#define OPTION_VAL_BLACK_02_NO NULL
#define OPTION_VAL_LIGHT_BLUE_04_NO NULL
#define OPTION_VAL_CYAN_06_NO NULL
#define OPTION_VAL_CYAN_BLUE_07_NO NULL
#define OPTION_VAL_LIGHT_GREEN_08_NO NULL
#define OPTION_VAL_CYAN_GREEN_09_NO NULL
#define OPTION_VAL_LIGHT_CYAN_10_NO NULL
#define OPTION_VAL_PURPLE_12_NO NULL
#define OPTION_VAL_LIGHT_PURPLE_13_NO NULL
#define OPTION_VAL_YELLOW_14_NO "Gul"
#define OPTION_VAL_GRAY_15_NO NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_NO NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_NO NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_NO NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_NO NULL
#define OPTION_VAL_LIGHT_RED_2_20_NO NULL
#define OPTION_VAL_MAGENTA_21_NO NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_NO NULL
#define OPTION_VAL_LIGHT_ORANGE_23_NO NULL
#define OPTION_VAL_ORANGE_24_NO NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_NO NULL
#define OPTION_VAL_LIGHT_YELLOW_26_NO NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_NO NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_NO NULL
#define OPTION_VAL_40_NO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_NO NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_NO NULL

struct retro_core_option_v2_category option_cats_no[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_NO,
      CATEGORY_SYSTEM_INFO_0_NO
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_NO,
      CATEGORY_VIDEO_INFO_0_NO
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_NO,
      CATEGORY_PERFORMANCE_INFO_0_NO
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_NO,
      CATEGORY_HACKS_INFO_0_NO
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_NO,
      CATEGORY_INPUT_INFO_0_NO
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_NO,
      CATEGORY_EXPANSIONS_INFO_0_NO
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_NO,
      CATEGORY_VMU_INFO_0_NO
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_no[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_NO,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_NO },
         { "USA",     OPTION_VAL_USA_NO },
         { "Europe",  OPTION_VAL_EUROPE_NO },
         { "Default", OPTION_VAL_DEFAULT_NO },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_NO,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_NO },
         { "English",  OPTION_VAL_ENGLISH_NO },
         { "German",   OPTION_VAL_GERMAN_NO },
         { "French",   OPTION_VAL_FRENCH_NO },
         { "Spanish",  OPTION_VAL_SPANISH_NO },
         { "Italian",  OPTION_VAL_ITALIAN_NO },
         { "Default",  OPTION_VAL_DEFAULT_NO },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_NO,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_NO },
         { "640x480",    OPTION_VAL_640X480_NO },
         { "800x600",    OPTION_VAL_800X600_NO },
         { "960x720",    OPTION_VAL_960X720_NO },
         { "1024x768",   OPTION_VAL_1024X768_NO },
         { "1280x960",   OPTION_VAL_1280X960_NO },
         { "1440x1080",  OPTION_VAL_1440X1080_NO },
         { "1600x1200",  OPTION_VAL_1600X1200_NO },
         { "1920x1440",  OPTION_VAL_1920X1440_NO },
         { "2560x1920",  OPTION_VAL_2560X1920_NO },
         { "2880x2160",  OPTION_VAL_2880X2160_NO },
         { "3200x2400",  OPTION_VAL_3200X2400_NO },
         { "3840x2880",  OPTION_VAL_3840X2880_NO },
         { "4480x3360",  OPTION_VAL_4480X3360_NO },
         { "5120x3840",  OPTION_VAL_5120X3840_NO },
         { "5760x4320",  OPTION_VAL_5760X4320_NO },
         { "6400x4800",  OPTION_VAL_6400X4800_NO },
         { "7040x5280",  OPTION_VAL_7040X5280_NO },
         { "7680x5760",  OPTION_VAL_7680X5760_NO },
         { "8320x6240",  OPTION_VAL_8320X6240_NO },
         { "8960x6720",  OPTION_VAL_8960X6720_NO },
         { "9600x7200",  OPTION_VAL_9600X7200_NO },
         { "10240x7680", OPTION_VAL_10240X7680_NO },
         { "10880x8160", OPTION_VAL_10880X8160_NO },
         { "11520x8640", OPTION_VAL_11520X8640_NO },
         { "12160x9120", OPTION_VAL_12160X9120_NO },
         { "12800x9600", OPTION_VAL_12800X9600_NO },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_NO,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_NO },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_NO },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_NO },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_NO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_NO },
         { "PAL",     OPTION_VAL_PAL_NO },
         { "PAL_N",   OPTION_VAL_PAL_N_NO },
         { "PAL_M",   OPTION_VAL_PAL_M_NO },
         { "Default", OPTION_VAL_DEFAULT_NO },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_NO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_NO },
         { "vertical",   OPTION_VAL_VERTICAL_NO },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_NO,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_NO },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_NO },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_NO },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_NO,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_NO },
         { "1GB",   OPTION_VAL_1GB_NO },
         { "2GB",   OPTION_VAL_2GB_NO },
         { "4GB",   OPTION_VAL_4GB_NO },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_NO,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_NO,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_NO,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_NO,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_NO },
         { "1",  OPTION_VAL_1_NO },
         { "2",  OPTION_VAL_2_NO },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_NO,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_NO },
         { "4",  OPTION_VAL_4_NO },
         { "6",  OPTION_VAL_6_NO },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_NO,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_NO,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_NO,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_NO,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_NO,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_NO,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_NO,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_NO },
         { "more", OPTION_VAL_MORE_NO },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_NO,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_NO },
         { "110", OPTION_VAL_110_NO },
         { "120", OPTION_VAL_120_NO },
         { "130", OPTION_VAL_130_NO },
         { "140", OPTION_VAL_140_NO },
         { "150", OPTION_VAL_150_NO },
         { "160", OPTION_VAL_160_NO },
         { "170", OPTION_VAL_170_NO },
         { "180", OPTION_VAL_180_NO },
         { "190", OPTION_VAL_190_NO },
         { "200", OPTION_VAL_200_NO },
         { "210", OPTION_VAL_210_NO },
         { "220", OPTION_VAL_220_NO },
         { "230", OPTION_VAL_230_NO },
         { "240", OPTION_VAL_240_NO },
         { "250", OPTION_VAL_250_NO },
         { "260", OPTION_VAL_260_NO },
         { "270", OPTION_VAL_270_NO },
         { "280", OPTION_VAL_280_NO },
         { "290", OPTION_VAL_290_NO },
         { "300", OPTION_VAL_300_NO },
         { "310", OPTION_VAL_310_NO },
         { "320", OPTION_VAL_320_NO },
         { "330", OPTION_VAL_330_NO },
         { "340", OPTION_VAL_340_NO },
         { "350", OPTION_VAL_350_NO },
         { "360", OPTION_VAL_360_NO },
         { "370", OPTION_VAL_370_NO },
         { "380", OPTION_VAL_380_NO },
         { "390", OPTION_VAL_390_NO },
         { "400", OPTION_VAL_400_NO },
         { "410", OPTION_VAL_410_NO },
         { "420", OPTION_VAL_420_NO },
         { "430", OPTION_VAL_430_NO },
         { "440", OPTION_VAL_440_NO },
         { "450", OPTION_VAL_450_NO },
         { "460", OPTION_VAL_460_NO },
         { "470", OPTION_VAL_470_NO },
         { "480", OPTION_VAL_480_NO },
         { "490", OPTION_VAL_490_NO },
         { "500", OPTION_VAL_500_NO },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_NO,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_NO },
         { "5%",  OPTION_VAL_5_NO },
         { "10%", OPTION_VAL_10_NO },
         { "15%", OPTION_VAL_15_NO },
         { "20%", OPTION_VAL_20_NO },
         { "25%", OPTION_VAL_25_NO },
         { "30%", OPTION_VAL_30_NO },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_NO,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_NO },
         { "5%",  OPTION_VAL_5_NO },
         { "10%", OPTION_VAL_10_NO },
         { "15%", OPTION_VAL_15_NO },
         { "20%", OPTION_VAL_20_NO },
         { "25%", OPTION_VAL_25_NO },
         { "30%", OPTION_VAL_30_NO },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_NO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_NO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_NO,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_NO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_NO },
         { "60%",  OPTION_VAL_60_NO },
         { "70%",  OPTION_VAL_70_NO },
         { "80%",  OPTION_VAL_80_NO },
         { "90%",  OPTION_VAL_90_NO },
         { "100%", OPTION_VAL_100_O47_NO },
         { "110%", OPTION_VAL_110_O47_NO },
         { "120%", OPTION_VAL_120_O47_NO },
         { "130%", OPTION_VAL_130_O47_NO },
         { "140%", OPTION_VAL_140_O47_NO },
         { "150%", OPTION_VAL_150_O47_NO },
         { "160%", OPTION_VAL_160_O47_NO },
         { "170%", OPTION_VAL_170_O47_NO },
         { "180%", OPTION_VAL_180_O47_NO },
         { "190%", OPTION_VAL_190_O47_NO },
         { "200%", OPTION_VAL_200_O47_NO },
         { "210%", OPTION_VAL_210_O47_NO },
         { "220%", OPTION_VAL_220_O47_NO },
         { "230%", OPTION_VAL_230_O47_NO },
         { "240%", OPTION_VAL_240_O47_NO },
         { "250%", OPTION_VAL_250_O47_NO },
         { "260%", OPTION_VAL_260_O47_NO },
         { "270%", OPTION_VAL_270_O47_NO },
         { "280%", OPTION_VAL_280_O47_NO },
         { "290%", OPTION_VAL_290_O47_NO },
         { "300%", OPTION_VAL_300_O47_NO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_NO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NO },
         { "Red",      OPTION_VAL_RED_NO },
         { "Green",    OPTION_VAL_GREEN_NO },
         { "Blue",     OPTION_VAL_BLUE_NO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_NO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NO },
         { "Red",      OPTION_VAL_RED_NO },
         { "Green",    OPTION_VAL_GREEN_NO },
         { "Blue",     OPTION_VAL_BLUE_NO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_NO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NO },
         { "Red",      OPTION_VAL_RED_NO },
         { "Green",    OPTION_VAL_GREEN_NO },
         { "Blue",     OPTION_VAL_BLUE_NO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_NO,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_NO },
         { "Red",      OPTION_VAL_RED_NO },
         { "Green",    OPTION_VAL_GREEN_NO },
         { "Blue",     OPTION_VAL_BLUE_NO },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_NO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NO },
         { "Purupuru", OPTION_VAL_PURUPURU_NO },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_NO },
         { "None",     OPTION_VAL_NONE_NO },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_NO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NO },
         { "Purupuru", OPTION_VAL_PURUPURU_NO },
         { "None",     OPTION_VAL_NONE_NO },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_NO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NO },
         { "Purupuru", OPTION_VAL_PURUPURU_NO },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_NO },
         { "None",     OPTION_VAL_NONE_NO },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_NO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NO },
         { "Purupuru", OPTION_VAL_PURUPURU_NO },
         { "None",     OPTION_VAL_NONE_NO },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_NO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NO },
         { "Purupuru", OPTION_VAL_PURUPURU_NO },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_NO },
         { "None",     OPTION_VAL_NONE_NO },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_NO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NO },
         { "Purupuru", OPTION_VAL_PURUPURU_NO },
         { "None",     OPTION_VAL_NONE_NO },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_NO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NO },
         { "Purupuru", OPTION_VAL_PURUPURU_NO },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_NO },
         { "None",     OPTION_VAL_NONE_NO },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_NO,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_NO },
         { "Purupuru", OPTION_VAL_PURUPURU_NO },
         { "None",     OPTION_VAL_NONE_NO },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_NO,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_NO,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_NO,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_NO },
         { "All VMUs", OPTION_VAL_ALL_VMUS_NO },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_NO,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_NO,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_NO,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_NO,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_NO,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_NO,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_NO,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_NO,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NO },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NO },
         { "2x", OPTION_VAL_2_O26_NO },
         { "3x", OPTION_VAL_3X_NO },
         { "4x", OPTION_VAL_4_NO },
         { "5x", OPTION_VAL_5X_NO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NO },
         { "20%",  OPTION_VAL_20_NO },
         { "30%",  OPTION_VAL_30_NO },
         { "40%",  OPTION_VAL_40_NO },
         { "50%",  OPTION_VAL_50_NO },
         { "60%",  OPTION_VAL_60_NO },
         { "70%",  OPTION_VAL_70_NO },
         { "80%",  OPTION_VAL_80_NO },
         { "90%",  OPTION_VAL_90_NO },
         { "100%", OPTION_VAL_100_O47_NO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NO },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NO },
         { "2x", OPTION_VAL_2_O26_NO },
         { "3x", OPTION_VAL_3X_NO },
         { "4x", OPTION_VAL_4_NO },
         { "5x", OPTION_VAL_5X_NO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NO },
         { "20%",  OPTION_VAL_20_NO },
         { "30%",  OPTION_VAL_30_NO },
         { "40%",  OPTION_VAL_40_NO },
         { "50%",  OPTION_VAL_50_NO },
         { "60%",  OPTION_VAL_60_NO },
         { "70%",  OPTION_VAL_70_NO },
         { "80%",  OPTION_VAL_80_NO },
         { "90%",  OPTION_VAL_90_NO },
         { "100%", OPTION_VAL_100_O47_NO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NO },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NO },
         { "2x", OPTION_VAL_2_O26_NO },
         { "3x", OPTION_VAL_3X_NO },
         { "4x", OPTION_VAL_4_NO },
         { "5x", OPTION_VAL_5X_NO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NO },
         { "20%",  OPTION_VAL_20_NO },
         { "30%",  OPTION_VAL_30_NO },
         { "40%",  OPTION_VAL_40_NO },
         { "50%",  OPTION_VAL_50_NO },
         { "60%",  OPTION_VAL_60_NO },
         { "70%",  OPTION_VAL_70_NO },
         { "80%",  OPTION_VAL_80_NO },
         { "90%",  OPTION_VAL_90_NO },
         { "100%", OPTION_VAL_100_O47_NO },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_NO },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_NO },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_NO },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_NO },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_NO },
         { "2x", OPTION_VAL_2_O26_NO },
         { "3x", OPTION_VAL_3X_NO },
         { "4x", OPTION_VAL_4_NO },
         { "5x", OPTION_VAL_5X_NO },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_NO },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_NO },
         { "BLACK 02",          OPTION_VAL_BLACK_02_NO },
         { "BLUE 03",           OPTION_VAL_BLUE_NO },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_NO },
         { "GREEN 05",          OPTION_VAL_GREEN_NO },
         { "CYAN 06",           OPTION_VAL_CYAN_06_NO },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_NO },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_NO },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_NO },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_NO },
         { "RED 11",            OPTION_VAL_RED_NO },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_NO },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_NO },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_NO },
         { "GRAY 15",           OPTION_VAL_GRAY_15_NO },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_NO },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_NO },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_NO },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_NO },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_NO },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_NO },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_NO },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_NO },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_NO },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_NO },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_NO },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_NO },
         { "WHITE 28",          OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_NO,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_NO },
         { "20%",  OPTION_VAL_20_NO },
         { "30%",  OPTION_VAL_30_NO },
         { "40%",  OPTION_VAL_40_NO },
         { "50%",  OPTION_VAL_50_NO },
         { "60%",  OPTION_VAL_60_NO },
         { "70%",  OPTION_VAL_70_NO },
         { "80%",  OPTION_VAL_80_NO },
         { "90%",  OPTION_VAL_90_NO },
         { "100%", OPTION_VAL_100_O47_NO },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_no = {
   option_cats_no,
   option_defs_no
};

/* RETRO_LANGUAGE_OR */

#define CATEGORY_SYSTEM_LABEL_OR NULL
#define CATEGORY_SYSTEM_INFO_0_OR NULL
#define CATEGORY_VIDEO_LABEL_OR "ଵିଡ଼ିଓ"
#define CATEGORY_VIDEO_INFO_0_OR NULL
#define CATEGORY_PERFORMANCE_LABEL_OR NULL
#define CATEGORY_PERFORMANCE_INFO_0_OR NULL
#define CATEGORY_HACKS_LABEL_OR NULL
#define CATEGORY_HACKS_INFO_0_OR NULL
#define CATEGORY_INPUT_LABEL_OR NULL
#define CATEGORY_INPUT_INFO_0_OR NULL
#define CATEGORY_EXPANSIONS_LABEL_OR NULL
#define CATEGORY_EXPANSIONS_INFO_0_OR NULL
#define CATEGORY_VMU_LABEL_OR NULL
#define CATEGORY_VMU_INFO_0_OR NULL
#define CORE_OPTION_NAME_REGION_LABEL_OR "ଅଞ୍ଚଳ"
#define OPTION_VAL_JAPAN_OR NULL
#define OPTION_VAL_USA_OR NULL
#define OPTION_VAL_EUROPE_OR NULL
#define OPTION_VAL_DEFAULT_OR NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_OR "ଭାଷା"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_OR NULL
#define OPTION_VAL_JAPANESE_OR NULL
#define OPTION_VAL_ENGLISH_OR NULL
#define OPTION_VAL_GERMAN_OR NULL
#define OPTION_VAL_FRENCH_OR NULL
#define OPTION_VAL_SPANISH_OR NULL
#define OPTION_VAL_ITALIAN_OR NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_OR NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_OR NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_OR NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_OR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_OR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_OR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_OR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_OR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_OR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_OR NULL
#define CORE_OPTION_NAME_UPNP_LABEL_OR NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_OR NULL
#define CORE_OPTION_NAME_DCNET_LABEL_OR NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_OR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_OR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_OR NULL
#define OPTION_VAL_320X240_OR NULL
#define OPTION_VAL_640X480_OR NULL
#define OPTION_VAL_800X600_OR NULL
#define OPTION_VAL_960X720_OR NULL
#define OPTION_VAL_1024X768_OR NULL
#define OPTION_VAL_1280X960_OR NULL
#define OPTION_VAL_1440X1080_OR NULL
#define OPTION_VAL_1600X1200_OR NULL
#define OPTION_VAL_1920X1440_OR NULL
#define OPTION_VAL_2560X1920_OR NULL
#define OPTION_VAL_2880X2160_OR NULL
#define OPTION_VAL_3200X2400_OR NULL
#define OPTION_VAL_3840X2880_OR NULL
#define OPTION_VAL_4480X3360_OR NULL
#define OPTION_VAL_5120X3840_OR NULL
#define OPTION_VAL_5760X4320_OR NULL
#define OPTION_VAL_6400X4800_OR NULL
#define OPTION_VAL_7040X5280_OR NULL
#define OPTION_VAL_7680X5760_OR NULL
#define OPTION_VAL_8320X6240_OR NULL
#define OPTION_VAL_8960X6720_OR NULL
#define OPTION_VAL_9600X7200_OR NULL
#define OPTION_VAL_10240X7680_OR NULL
#define OPTION_VAL_10880X8160_OR NULL
#define OPTION_VAL_11520X8640_OR NULL
#define OPTION_VAL_12160X9120_OR NULL
#define OPTION_VAL_12800X9600_OR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_OR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_OR NULL
#define OPTION_VAL_VGA_OR NULL
#define OPTION_VAL_TV_RGB_OR NULL
#define OPTION_VAL_TV_COMPOSITE_OR NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_OR NULL
#define OPTION_VAL_NTSC_OR NULL
#define OPTION_VAL_PAL_OR NULL
#define OPTION_VAL_PAL_N_OR NULL
#define OPTION_VAL_PAL_M_OR NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_OR NULL
#define OPTION_VAL_HORIZONTAL_OR NULL
#define OPTION_VAL_VERTICAL_OR NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_OR NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_OR NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_OR NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_OR NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_OR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_OR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_OR NULL
#define OPTION_VAL_512MB_OR NULL
#define OPTION_VAL_1GB_OR NULL
#define OPTION_VAL_2GB_OR NULL
#define OPTION_VAL_4GB_OR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_OR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_OR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_OR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_OR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_OR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_OR NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_OR NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_OR NULL
#define CORE_OPTION_NAME_FOG_LABEL_OR NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_OR NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_OR NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_OR NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_OR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_OR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_OR NULL
#define OPTION_VAL_1_OR NULL
#define OPTION_VAL_2_OR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_OR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_OR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_OR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_OR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_OR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_OR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_OR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_OR NULL
#define OPTION_VAL_2_O26_OR NULL
#define OPTION_VAL_4_OR NULL
#define OPTION_VAL_6_OR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_OR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_OR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_OR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_OR NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_OR NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_OR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_OR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_OR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_OR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_OR NULL
#define OPTION_VAL_SOME_OR NULL
#define OPTION_VAL_MORE_OR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_OR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_OR NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_OR NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_OR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_OR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_OR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_OR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_OR NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_OR NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_OR NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_OR NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_OR NULL
#define OPTION_VAL_100_OR NULL
#define OPTION_VAL_110_OR NULL
#define OPTION_VAL_120_OR NULL
#define OPTION_VAL_130_OR NULL
#define OPTION_VAL_140_OR NULL
#define OPTION_VAL_150_OR NULL
#define OPTION_VAL_160_OR NULL
#define OPTION_VAL_170_OR NULL
#define OPTION_VAL_180_OR NULL
#define OPTION_VAL_190_OR NULL
#define OPTION_VAL_200_OR NULL
#define OPTION_VAL_210_OR NULL
#define OPTION_VAL_220_OR NULL
#define OPTION_VAL_230_OR NULL
#define OPTION_VAL_240_OR NULL
#define OPTION_VAL_250_OR NULL
#define OPTION_VAL_260_OR NULL
#define OPTION_VAL_270_OR NULL
#define OPTION_VAL_280_OR NULL
#define OPTION_VAL_290_OR NULL
#define OPTION_VAL_300_OR NULL
#define OPTION_VAL_310_OR NULL
#define OPTION_VAL_320_OR NULL
#define OPTION_VAL_330_OR NULL
#define OPTION_VAL_340_OR NULL
#define OPTION_VAL_350_OR NULL
#define OPTION_VAL_360_OR NULL
#define OPTION_VAL_370_OR NULL
#define OPTION_VAL_380_OR NULL
#define OPTION_VAL_390_OR NULL
#define OPTION_VAL_400_OR NULL
#define OPTION_VAL_410_OR NULL
#define OPTION_VAL_420_OR NULL
#define OPTION_VAL_430_OR NULL
#define OPTION_VAL_440_OR NULL
#define OPTION_VAL_450_OR NULL
#define OPTION_VAL_460_OR NULL
#define OPTION_VAL_470_OR NULL
#define OPTION_VAL_480_OR NULL
#define OPTION_VAL_490_OR NULL
#define OPTION_VAL_500_OR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_OR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_OR NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_OR NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_OR NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_OR NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_OR NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_OR NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_OR NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_OR NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_OR NULL
#define OPTION_VAL_0_OR NULL
#define OPTION_VAL_5_OR NULL
#define OPTION_VAL_10_OR NULL
#define OPTION_VAL_15_OR NULL
#define OPTION_VAL_20_OR NULL
#define OPTION_VAL_25_OR NULL
#define OPTION_VAL_30_OR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_OR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_OR NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_OR NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_OR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_OR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_OR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_OR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_OR NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_OR NULL
#define OPTION_VAL_50_OR NULL
#define OPTION_VAL_60_OR NULL
#define OPTION_VAL_70_OR NULL
#define OPTION_VAL_80_OR NULL
#define OPTION_VAL_90_OR NULL
#define OPTION_VAL_100_O47_OR NULL
#define OPTION_VAL_110_O47_OR NULL
#define OPTION_VAL_120_O47_OR NULL
#define OPTION_VAL_130_O47_OR NULL
#define OPTION_VAL_140_O47_OR NULL
#define OPTION_VAL_150_O47_OR NULL
#define OPTION_VAL_160_O47_OR NULL
#define OPTION_VAL_170_O47_OR NULL
#define OPTION_VAL_180_O47_OR NULL
#define OPTION_VAL_190_O47_OR NULL
#define OPTION_VAL_200_O47_OR NULL
#define OPTION_VAL_210_O47_OR NULL
#define OPTION_VAL_220_O47_OR NULL
#define OPTION_VAL_230_O47_OR NULL
#define OPTION_VAL_240_O47_OR NULL
#define OPTION_VAL_250_O47_OR NULL
#define OPTION_VAL_260_O47_OR NULL
#define OPTION_VAL_270_O47_OR NULL
#define OPTION_VAL_280_O47_OR NULL
#define OPTION_VAL_290_O47_OR NULL
#define OPTION_VAL_300_O47_OR NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_OR NULL
#define OPTION_VAL_WHITE_OR NULL
#define OPTION_VAL_RED_OR NULL
#define OPTION_VAL_GREEN_OR NULL
#define OPTION_VAL_BLUE_OR NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_OR NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_OR NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_OR NULL
#define OPTION_VAL_VMU_OR NULL
#define OPTION_VAL_PURUPURU_OR NULL
#define OPTION_VAL_DREAMPOTATO_OR NULL
#define OPTION_VAL_NONE_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_OR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_OR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_OR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_OR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_OR NULL
#define OPTION_VAL_VMU_A1_OR NULL
#define OPTION_VAL_ALL_VMUS_OR NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_OR NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_OR NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_OR NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_OR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_OR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_OR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_OR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_OR NULL
#define OPTION_VAL_UPPER_LEFT_OR NULL
#define OPTION_VAL_UPPER_RIGHT_OR NULL
#define OPTION_VAL_LOWER_LEFT_OR NULL
#define OPTION_VAL_LOWER_RIGHT_OR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_OR NULL
#define OPTION_VAL_1X_OR NULL
#define OPTION_VAL_3X_OR NULL
#define OPTION_VAL_5X_OR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_OR NULL
#define OPTION_VAL_DEFAULT_ON_00_OR NULL
#define OPTION_VAL_DEFAULT_OFF_01_OR NULL
#define OPTION_VAL_BLACK_02_OR NULL
#define OPTION_VAL_LIGHT_BLUE_04_OR NULL
#define OPTION_VAL_CYAN_06_OR NULL
#define OPTION_VAL_CYAN_BLUE_07_OR NULL
#define OPTION_VAL_LIGHT_GREEN_08_OR NULL
#define OPTION_VAL_CYAN_GREEN_09_OR NULL
#define OPTION_VAL_LIGHT_CYAN_10_OR NULL
#define OPTION_VAL_PURPLE_12_OR NULL
#define OPTION_VAL_LIGHT_PURPLE_13_OR NULL
#define OPTION_VAL_YELLOW_14_OR NULL
#define OPTION_VAL_GRAY_15_OR NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_OR NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_OR NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_OR NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_OR NULL
#define OPTION_VAL_LIGHT_RED_2_20_OR NULL
#define OPTION_VAL_MAGENTA_21_OR NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_OR NULL
#define OPTION_VAL_LIGHT_ORANGE_23_OR NULL
#define OPTION_VAL_ORANGE_24_OR NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_OR NULL
#define OPTION_VAL_LIGHT_YELLOW_26_OR NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_OR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_OR NULL
#define OPTION_VAL_40_OR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_OR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_OR NULL

struct retro_core_option_v2_category option_cats_or[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_OR,
      CATEGORY_SYSTEM_INFO_0_OR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_OR,
      CATEGORY_VIDEO_INFO_0_OR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_OR,
      CATEGORY_PERFORMANCE_INFO_0_OR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_OR,
      CATEGORY_HACKS_INFO_0_OR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_OR,
      CATEGORY_INPUT_INFO_0_OR
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_OR,
      CATEGORY_EXPANSIONS_INFO_0_OR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_OR,
      CATEGORY_VMU_INFO_0_OR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_or[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_OR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_OR },
         { "USA",     OPTION_VAL_USA_OR },
         { "Europe",  OPTION_VAL_EUROPE_OR },
         { "Default", OPTION_VAL_DEFAULT_OR },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_OR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_OR },
         { "English",  OPTION_VAL_ENGLISH_OR },
         { "German",   OPTION_VAL_GERMAN_OR },
         { "French",   OPTION_VAL_FRENCH_OR },
         { "Spanish",  OPTION_VAL_SPANISH_OR },
         { "Italian",  OPTION_VAL_ITALIAN_OR },
         { "Default",  OPTION_VAL_DEFAULT_OR },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_OR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_OR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_OR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_OR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_OR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_OR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_OR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_OR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_OR },
         { "640x480",    OPTION_VAL_640X480_OR },
         { "800x600",    OPTION_VAL_800X600_OR },
         { "960x720",    OPTION_VAL_960X720_OR },
         { "1024x768",   OPTION_VAL_1024X768_OR },
         { "1280x960",   OPTION_VAL_1280X960_OR },
         { "1440x1080",  OPTION_VAL_1440X1080_OR },
         { "1600x1200",  OPTION_VAL_1600X1200_OR },
         { "1920x1440",  OPTION_VAL_1920X1440_OR },
         { "2560x1920",  OPTION_VAL_2560X1920_OR },
         { "2880x2160",  OPTION_VAL_2880X2160_OR },
         { "3200x2400",  OPTION_VAL_3200X2400_OR },
         { "3840x2880",  OPTION_VAL_3840X2880_OR },
         { "4480x3360",  OPTION_VAL_4480X3360_OR },
         { "5120x3840",  OPTION_VAL_5120X3840_OR },
         { "5760x4320",  OPTION_VAL_5760X4320_OR },
         { "6400x4800",  OPTION_VAL_6400X4800_OR },
         { "7040x5280",  OPTION_VAL_7040X5280_OR },
         { "7680x5760",  OPTION_VAL_7680X5760_OR },
         { "8320x6240",  OPTION_VAL_8320X6240_OR },
         { "8960x6720",  OPTION_VAL_8960X6720_OR },
         { "9600x7200",  OPTION_VAL_9600X7200_OR },
         { "10240x7680", OPTION_VAL_10240X7680_OR },
         { "10880x8160", OPTION_VAL_10880X8160_OR },
         { "11520x8640", OPTION_VAL_11520X8640_OR },
         { "12160x9120", OPTION_VAL_12160X9120_OR },
         { "12800x9600", OPTION_VAL_12800X9600_OR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_OR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_OR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_OR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_OR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_OR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_OR },
         { "PAL",     OPTION_VAL_PAL_OR },
         { "PAL_N",   OPTION_VAL_PAL_N_OR },
         { "PAL_M",   OPTION_VAL_PAL_M_OR },
         { "Default", OPTION_VAL_DEFAULT_OR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_OR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_OR },
         { "vertical",   OPTION_VAL_VERTICAL_OR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_OR,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_OR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_OR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_OR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_OR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_OR },
         { "1GB",   OPTION_VAL_1GB_OR },
         { "2GB",   OPTION_VAL_2GB_OR },
         { "4GB",   OPTION_VAL_4GB_OR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_OR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_OR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_OR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_OR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_OR },
         { "1",  OPTION_VAL_1_OR },
         { "2",  OPTION_VAL_2_OR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_OR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_OR },
         { "4",  OPTION_VAL_4_OR },
         { "6",  OPTION_VAL_6_OR },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_OR,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_OR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_OR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_OR,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_OR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_OR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_OR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_OR },
         { "more", OPTION_VAL_MORE_OR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_OR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_OR },
         { "110", OPTION_VAL_110_OR },
         { "120", OPTION_VAL_120_OR },
         { "130", OPTION_VAL_130_OR },
         { "140", OPTION_VAL_140_OR },
         { "150", OPTION_VAL_150_OR },
         { "160", OPTION_VAL_160_OR },
         { "170", OPTION_VAL_170_OR },
         { "180", OPTION_VAL_180_OR },
         { "190", OPTION_VAL_190_OR },
         { "200", OPTION_VAL_200_OR },
         { "210", OPTION_VAL_210_OR },
         { "220", OPTION_VAL_220_OR },
         { "230", OPTION_VAL_230_OR },
         { "240", OPTION_VAL_240_OR },
         { "250", OPTION_VAL_250_OR },
         { "260", OPTION_VAL_260_OR },
         { "270", OPTION_VAL_270_OR },
         { "280", OPTION_VAL_280_OR },
         { "290", OPTION_VAL_290_OR },
         { "300", OPTION_VAL_300_OR },
         { "310", OPTION_VAL_310_OR },
         { "320", OPTION_VAL_320_OR },
         { "330", OPTION_VAL_330_OR },
         { "340", OPTION_VAL_340_OR },
         { "350", OPTION_VAL_350_OR },
         { "360", OPTION_VAL_360_OR },
         { "370", OPTION_VAL_370_OR },
         { "380", OPTION_VAL_380_OR },
         { "390", OPTION_VAL_390_OR },
         { "400", OPTION_VAL_400_OR },
         { "410", OPTION_VAL_410_OR },
         { "420", OPTION_VAL_420_OR },
         { "430", OPTION_VAL_430_OR },
         { "440", OPTION_VAL_440_OR },
         { "450", OPTION_VAL_450_OR },
         { "460", OPTION_VAL_460_OR },
         { "470", OPTION_VAL_470_OR },
         { "480", OPTION_VAL_480_OR },
         { "490", OPTION_VAL_490_OR },
         { "500", OPTION_VAL_500_OR },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_OR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_OR },
         { "5%",  OPTION_VAL_5_OR },
         { "10%", OPTION_VAL_10_OR },
         { "15%", OPTION_VAL_15_OR },
         { "20%", OPTION_VAL_20_OR },
         { "25%", OPTION_VAL_25_OR },
         { "30%", OPTION_VAL_30_OR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_OR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_OR },
         { "5%",  OPTION_VAL_5_OR },
         { "10%", OPTION_VAL_10_OR },
         { "15%", OPTION_VAL_15_OR },
         { "20%", OPTION_VAL_20_OR },
         { "25%", OPTION_VAL_25_OR },
         { "30%", OPTION_VAL_30_OR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_OR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_OR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_OR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_OR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_OR },
         { "60%",  OPTION_VAL_60_OR },
         { "70%",  OPTION_VAL_70_OR },
         { "80%",  OPTION_VAL_80_OR },
         { "90%",  OPTION_VAL_90_OR },
         { "100%", OPTION_VAL_100_O47_OR },
         { "110%", OPTION_VAL_110_O47_OR },
         { "120%", OPTION_VAL_120_O47_OR },
         { "130%", OPTION_VAL_130_O47_OR },
         { "140%", OPTION_VAL_140_O47_OR },
         { "150%", OPTION_VAL_150_O47_OR },
         { "160%", OPTION_VAL_160_O47_OR },
         { "170%", OPTION_VAL_170_O47_OR },
         { "180%", OPTION_VAL_180_O47_OR },
         { "190%", OPTION_VAL_190_O47_OR },
         { "200%", OPTION_VAL_200_O47_OR },
         { "210%", OPTION_VAL_210_O47_OR },
         { "220%", OPTION_VAL_220_O47_OR },
         { "230%", OPTION_VAL_230_O47_OR },
         { "240%", OPTION_VAL_240_O47_OR },
         { "250%", OPTION_VAL_250_O47_OR },
         { "260%", OPTION_VAL_260_O47_OR },
         { "270%", OPTION_VAL_270_O47_OR },
         { "280%", OPTION_VAL_280_O47_OR },
         { "290%", OPTION_VAL_290_O47_OR },
         { "300%", OPTION_VAL_300_O47_OR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_OR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_OR },
         { "Red",      OPTION_VAL_RED_OR },
         { "Green",    OPTION_VAL_GREEN_OR },
         { "Blue",     OPTION_VAL_BLUE_OR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_OR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_OR },
         { "Red",      OPTION_VAL_RED_OR },
         { "Green",    OPTION_VAL_GREEN_OR },
         { "Blue",     OPTION_VAL_BLUE_OR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_OR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_OR },
         { "Red",      OPTION_VAL_RED_OR },
         { "Green",    OPTION_VAL_GREEN_OR },
         { "Blue",     OPTION_VAL_BLUE_OR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_OR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_OR },
         { "Red",      OPTION_VAL_RED_OR },
         { "Green",    OPTION_VAL_GREEN_OR },
         { "Blue",     OPTION_VAL_BLUE_OR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_OR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_OR },
         { "Purupuru", OPTION_VAL_PURUPURU_OR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_OR },
         { "None",     OPTION_VAL_NONE_OR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_OR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_OR },
         { "Purupuru", OPTION_VAL_PURUPURU_OR },
         { "None",     OPTION_VAL_NONE_OR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_OR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_OR },
         { "Purupuru", OPTION_VAL_PURUPURU_OR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_OR },
         { "None",     OPTION_VAL_NONE_OR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_OR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_OR },
         { "Purupuru", OPTION_VAL_PURUPURU_OR },
         { "None",     OPTION_VAL_NONE_OR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_OR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_OR },
         { "Purupuru", OPTION_VAL_PURUPURU_OR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_OR },
         { "None",     OPTION_VAL_NONE_OR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_OR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_OR },
         { "Purupuru", OPTION_VAL_PURUPURU_OR },
         { "None",     OPTION_VAL_NONE_OR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_OR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_OR },
         { "Purupuru", OPTION_VAL_PURUPURU_OR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_OR },
         { "None",     OPTION_VAL_NONE_OR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_OR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_OR },
         { "Purupuru", OPTION_VAL_PURUPURU_OR },
         { "None",     OPTION_VAL_NONE_OR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_OR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_OR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_OR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_OR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_OR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_OR,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_OR,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_OR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_OR,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_OR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_OR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_OR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_OR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_OR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_OR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_OR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_OR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_OR },
         { "2x", OPTION_VAL_2_O26_OR },
         { "3x", OPTION_VAL_3X_OR },
         { "4x", OPTION_VAL_4_OR },
         { "5x", OPTION_VAL_5X_OR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_OR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_OR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_OR },
         { "BLUE 03",           OPTION_VAL_BLUE_OR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_OR },
         { "GREEN 05",          OPTION_VAL_GREEN_OR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_OR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_OR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_OR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_OR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_OR },
         { "RED 11",            OPTION_VAL_RED_OR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_OR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_OR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_OR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_OR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_OR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_OR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_OR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_OR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_OR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_OR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_OR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_OR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_OR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_OR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_OR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_OR },
         { "WHITE 28",          OPTION_VAL_WHITE_OR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_OR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_OR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_OR },
         { "BLUE 03",           OPTION_VAL_BLUE_OR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_OR },
         { "GREEN 05",          OPTION_VAL_GREEN_OR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_OR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_OR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_OR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_OR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_OR },
         { "RED 11",            OPTION_VAL_RED_OR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_OR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_OR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_OR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_OR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_OR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_OR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_OR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_OR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_OR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_OR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_OR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_OR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_OR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_OR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_OR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_OR },
         { "WHITE 28",          OPTION_VAL_WHITE_OR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_OR },
         { "20%",  OPTION_VAL_20_OR },
         { "30%",  OPTION_VAL_30_OR },
         { "40%",  OPTION_VAL_40_OR },
         { "50%",  OPTION_VAL_50_OR },
         { "60%",  OPTION_VAL_60_OR },
         { "70%",  OPTION_VAL_70_OR },
         { "80%",  OPTION_VAL_80_OR },
         { "90%",  OPTION_VAL_90_OR },
         { "100%", OPTION_VAL_100_O47_OR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_OR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_OR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_OR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_OR },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_OR },
         { "2x", OPTION_VAL_2_O26_OR },
         { "3x", OPTION_VAL_3X_OR },
         { "4x", OPTION_VAL_4_OR },
         { "5x", OPTION_VAL_5X_OR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_OR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_OR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_OR },
         { "BLUE 03",           OPTION_VAL_BLUE_OR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_OR },
         { "GREEN 05",          OPTION_VAL_GREEN_OR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_OR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_OR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_OR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_OR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_OR },
         { "RED 11",            OPTION_VAL_RED_OR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_OR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_OR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_OR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_OR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_OR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_OR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_OR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_OR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_OR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_OR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_OR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_OR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_OR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_OR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_OR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_OR },
         { "WHITE 28",          OPTION_VAL_WHITE_OR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_OR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_OR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_OR },
         { "BLUE 03",           OPTION_VAL_BLUE_OR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_OR },
         { "GREEN 05",          OPTION_VAL_GREEN_OR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_OR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_OR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_OR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_OR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_OR },
         { "RED 11",            OPTION_VAL_RED_OR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_OR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_OR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_OR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_OR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_OR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_OR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_OR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_OR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_OR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_OR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_OR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_OR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_OR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_OR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_OR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_OR },
         { "WHITE 28",          OPTION_VAL_WHITE_OR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_OR },
         { "20%",  OPTION_VAL_20_OR },
         { "30%",  OPTION_VAL_30_OR },
         { "40%",  OPTION_VAL_40_OR },
         { "50%",  OPTION_VAL_50_OR },
         { "60%",  OPTION_VAL_60_OR },
         { "70%",  OPTION_VAL_70_OR },
         { "80%",  OPTION_VAL_80_OR },
         { "90%",  OPTION_VAL_90_OR },
         { "100%", OPTION_VAL_100_O47_OR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_OR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_OR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_OR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_OR },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_OR },
         { "2x", OPTION_VAL_2_O26_OR },
         { "3x", OPTION_VAL_3X_OR },
         { "4x", OPTION_VAL_4_OR },
         { "5x", OPTION_VAL_5X_OR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_OR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_OR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_OR },
         { "BLUE 03",           OPTION_VAL_BLUE_OR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_OR },
         { "GREEN 05",          OPTION_VAL_GREEN_OR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_OR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_OR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_OR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_OR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_OR },
         { "RED 11",            OPTION_VAL_RED_OR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_OR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_OR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_OR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_OR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_OR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_OR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_OR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_OR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_OR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_OR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_OR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_OR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_OR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_OR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_OR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_OR },
         { "WHITE 28",          OPTION_VAL_WHITE_OR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_OR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_OR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_OR },
         { "BLUE 03",           OPTION_VAL_BLUE_OR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_OR },
         { "GREEN 05",          OPTION_VAL_GREEN_OR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_OR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_OR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_OR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_OR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_OR },
         { "RED 11",            OPTION_VAL_RED_OR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_OR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_OR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_OR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_OR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_OR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_OR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_OR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_OR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_OR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_OR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_OR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_OR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_OR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_OR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_OR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_OR },
         { "WHITE 28",          OPTION_VAL_WHITE_OR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_OR },
         { "20%",  OPTION_VAL_20_OR },
         { "30%",  OPTION_VAL_30_OR },
         { "40%",  OPTION_VAL_40_OR },
         { "50%",  OPTION_VAL_50_OR },
         { "60%",  OPTION_VAL_60_OR },
         { "70%",  OPTION_VAL_70_OR },
         { "80%",  OPTION_VAL_80_OR },
         { "90%",  OPTION_VAL_90_OR },
         { "100%", OPTION_VAL_100_O47_OR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_OR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_OR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_OR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_OR },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_OR },
         { "2x", OPTION_VAL_2_O26_OR },
         { "3x", OPTION_VAL_3X_OR },
         { "4x", OPTION_VAL_4_OR },
         { "5x", OPTION_VAL_5X_OR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_OR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_OR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_OR },
         { "BLUE 03",           OPTION_VAL_BLUE_OR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_OR },
         { "GREEN 05",          OPTION_VAL_GREEN_OR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_OR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_OR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_OR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_OR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_OR },
         { "RED 11",            OPTION_VAL_RED_OR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_OR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_OR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_OR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_OR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_OR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_OR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_OR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_OR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_OR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_OR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_OR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_OR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_OR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_OR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_OR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_OR },
         { "WHITE 28",          OPTION_VAL_WHITE_OR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_OR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_OR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_OR },
         { "BLUE 03",           OPTION_VAL_BLUE_OR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_OR },
         { "GREEN 05",          OPTION_VAL_GREEN_OR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_OR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_OR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_OR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_OR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_OR },
         { "RED 11",            OPTION_VAL_RED_OR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_OR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_OR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_OR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_OR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_OR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_OR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_OR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_OR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_OR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_OR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_OR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_OR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_OR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_OR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_OR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_OR },
         { "WHITE 28",          OPTION_VAL_WHITE_OR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_OR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_OR },
         { "20%",  OPTION_VAL_20_OR },
         { "30%",  OPTION_VAL_30_OR },
         { "40%",  OPTION_VAL_40_OR },
         { "50%",  OPTION_VAL_50_OR },
         { "60%",  OPTION_VAL_60_OR },
         { "70%",  OPTION_VAL_70_OR },
         { "80%",  OPTION_VAL_80_OR },
         { "90%",  OPTION_VAL_90_OR },
         { "100%", OPTION_VAL_100_O47_OR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_or = {
   option_cats_or,
   option_defs_or
};

/* RETRO_LANGUAGE_PL */

#define CATEGORY_SYSTEM_LABEL_PL NULL
#define CATEGORY_SYSTEM_INFO_0_PL "Skonfiguruj ustawienia regionu, języka, BIOS i podstawowego sprzętu."
#define CATEGORY_VIDEO_LABEL_PL "Wideo"
#define CATEGORY_VIDEO_INFO_0_PL NULL
#define CATEGORY_PERFORMANCE_LABEL_PL "Wydajność"
#define CATEGORY_PERFORMANCE_INFO_0_PL NULL
#define CATEGORY_HACKS_LABEL_PL "Hacki emulacyjne"
#define CATEGORY_HACKS_INFO_0_PL NULL
#define CATEGORY_INPUT_LABEL_PL "Wejście"
#define CATEGORY_INPUT_INFO_0_PL NULL
#define CATEGORY_EXPANSIONS_LABEL_PL "Sloty rozszerzenia kontrolera"
#define CATEGORY_EXPANSIONS_INFO_0_PL "Wybierz urządzenie (VMU, urządzenie z plotką) podłączone do każdego pola rozszerzenia sterownika."
#define CATEGORY_VMU_LABEL_PL "Jednostka pamięci wizualnej"
#define CATEGORY_VMU_INFO_0_PL NULL
#define CORE_OPTION_NAME_REGION_LABEL_PL NULL
#define OPTION_VAL_JAPAN_PL NULL
#define OPTION_VAL_USA_PL NULL
#define OPTION_VAL_EUROPE_PL NULL
#define OPTION_VAL_DEFAULT_PL "Domyślny"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_PL "Język"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_PL "Zmienia język używany przez BIOS i przez gry zawierające wiele języków."
#define OPTION_VAL_JAPANESE_PL "japoński"
#define OPTION_VAL_ENGLISH_PL "angielski"
#define OPTION_VAL_GERMAN_PL "niemiecki"
#define OPTION_VAL_FRENCH_PL "francuski"
#define OPTION_VAL_SPANISH_PL "hiszpański"
#define OPTION_VAL_ITALIAN_PL "włoski"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_PL "HLE BIOS (wymagany restart)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_PL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_PL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_PL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_PL "Zezwalaj na przyciski serwisowe Arcade"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_PL "Włącza przycisk SERVICE dla gier arkadowych, aby wprowadzić ustawienia szafy."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_PL "Ustaw gry NAOMI jako Free Play"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_PL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_PL "Emulacja adaptera szerokopasmowego"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_PL NULL
#define CORE_OPTION_NAME_UPNP_LABEL_PL "Włącz UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_PL "Użyj UPnP, aby automatycznie skonfigurować router internetowy dla gier online."
#define CORE_OPTION_NAME_DCNET_LABEL_PL NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_PL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_PL "Rozdzielczość wewnętrzna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_PL NULL
#define OPTION_VAL_320X240_PL "320x240 (Połowa)"
#define OPTION_VAL_640X480_PL NULL
#define OPTION_VAL_800X600_PL NULL
#define OPTION_VAL_960X720_PL NULL
#define OPTION_VAL_1024X768_PL NULL
#define OPTION_VAL_1280X960_PL NULL
#define OPTION_VAL_1440X1080_PL NULL
#define OPTION_VAL_1600X1200_PL NULL
#define OPTION_VAL_1920X1440_PL NULL
#define OPTION_VAL_2560X1920_PL NULL
#define OPTION_VAL_2880X2160_PL NULL
#define OPTION_VAL_3200X2400_PL NULL
#define OPTION_VAL_3840X2880_PL NULL
#define OPTION_VAL_4480X3360_PL NULL
#define OPTION_VAL_5120X3840_PL NULL
#define OPTION_VAL_5760X4320_PL NULL
#define OPTION_VAL_6400X4800_PL NULL
#define OPTION_VAL_7040X5280_PL NULL
#define OPTION_VAL_7680X5760_PL NULL
#define OPTION_VAL_8320X6240_PL NULL
#define OPTION_VAL_8960X6720_PL NULL
#define OPTION_VAL_9600X7200_PL NULL
#define OPTION_VAL_10240X7680_PL NULL
#define OPTION_VAL_10880X8160_PL NULL
#define OPTION_VAL_11520X8640_PL NULL
#define OPTION_VAL_12160X9120_PL NULL
#define OPTION_VAL_12800X9600_PL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_PL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_PL NULL
#define OPTION_VAL_VGA_PL NULL
#define OPTION_VAL_TV_RGB_PL NULL
#define OPTION_VAL_TV_COMPOSITE_PL NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_PL NULL
#define OPTION_VAL_NTSC_PL NULL
#define OPTION_VAL_PAL_PL NULL
#define OPTION_VAL_PAL_N_PL NULL
#define OPTION_VAL_PAL_M_PL NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_PL "Orientacja ekranu"
#define OPTION_VAL_HORIZONTAL_PL NULL
#define OPTION_VAL_VERTICAL_PL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_PL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_PL NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_PL NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_PL NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_PL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_PL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_PL NULL
#define OPTION_VAL_512MB_PL NULL
#define OPTION_VAL_1GB_PL NULL
#define OPTION_VAL_2GB_PL NULL
#define OPTION_VAL_4GB_PL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_PL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_PL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_PL "Emulacja pełnego bufora ramki"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_PL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_PL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_PL NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_PL NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_PL NULL
#define CORE_OPTION_NAME_FOG_LABEL_PL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_PL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_PL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_PL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_PL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_PL "Filtrowanie tekstur"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_PL NULL
#define OPTION_VAL_1_PL NULL
#define OPTION_VAL_2_PL "Wymuś liniowy"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_PL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_PL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_PL "Wykrywanie zmian częstotliwości klatek"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_PL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_PL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_PL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_PL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_PL NULL
#define OPTION_VAL_2_O26_PL NULL
#define OPTION_VAL_4_PL NULL
#define OPTION_VAL_6_PL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_PL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_PL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_PL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_PL NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_PL NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_PL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_PL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_PL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_PL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_PL NULL
#define OPTION_VAL_SOME_PL "Normalny"
#define OPTION_VAL_MORE_PL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_PL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_PL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_PL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_PL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_PL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_PL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_PL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_PL NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_PL NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_PL "Włącza modyfikację pamięci RAM 32MB dla Dreamcast. Może wpływać na kompatybilność"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_PL NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_PL "Zmień główny zegar procesora SH4 z domyślnego 200 MHz. Odzwierciedlanie może pomóc spowolnić platformy. Przełączanie może zwiększyć szybkość klatek dla niektórych gier. Używaj ostrożnie."
#define OPTION_VAL_100_PL NULL
#define OPTION_VAL_110_PL NULL
#define OPTION_VAL_120_PL NULL
#define OPTION_VAL_130_PL NULL
#define OPTION_VAL_140_PL NULL
#define OPTION_VAL_150_PL NULL
#define OPTION_VAL_160_PL NULL
#define OPTION_VAL_170_PL NULL
#define OPTION_VAL_180_PL NULL
#define OPTION_VAL_190_PL NULL
#define OPTION_VAL_200_PL NULL
#define OPTION_VAL_210_PL NULL
#define OPTION_VAL_220_PL NULL
#define OPTION_VAL_230_PL NULL
#define OPTION_VAL_240_PL NULL
#define OPTION_VAL_250_PL NULL
#define OPTION_VAL_260_PL NULL
#define OPTION_VAL_270_PL NULL
#define OPTION_VAL_280_PL NULL
#define OPTION_VAL_290_PL NULL
#define OPTION_VAL_300_PL NULL
#define OPTION_VAL_310_PL NULL
#define OPTION_VAL_320_PL NULL
#define OPTION_VAL_330_PL NULL
#define OPTION_VAL_340_PL NULL
#define OPTION_VAL_350_PL NULL
#define OPTION_VAL_360_PL NULL
#define OPTION_VAL_370_PL NULL
#define OPTION_VAL_380_PL NULL
#define OPTION_VAL_390_PL NULL
#define OPTION_VAL_400_PL NULL
#define OPTION_VAL_410_PL NULL
#define OPTION_VAL_420_PL NULL
#define OPTION_VAL_430_PL NULL
#define OPTION_VAL_440_PL NULL
#define OPTION_VAL_450_PL NULL
#define OPTION_VAL_460_PL NULL
#define OPTION_VAL_470_PL NULL
#define OPTION_VAL_480_PL NULL
#define OPTION_VAL_490_PL NULL
#define OPTION_VAL_500_PL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_PL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_PL NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_PL NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_PL NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_PL "Zgraj tekstury"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_PL NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_PL NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_PL NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_PL NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_PL NULL
#define OPTION_VAL_0_PL NULL
#define OPTION_VAL_5_PL NULL
#define OPTION_VAL_10_PL NULL
#define OPTION_VAL_15_PL NULL
#define OPTION_VAL_20_PL NULL
#define OPTION_VAL_25_PL NULL
#define OPTION_VAL_30_PL "30 %"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_PL NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_PL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_PL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_PL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_PL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_PL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_PL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_PL NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_PL NULL
#define OPTION_VAL_50_PL "50 %"
#define OPTION_VAL_60_PL "60 %"
#define OPTION_VAL_70_PL NULL
#define OPTION_VAL_80_PL "80 %"
#define OPTION_VAL_90_PL NULL
#define OPTION_VAL_100_O47_PL NULL
#define OPTION_VAL_110_O47_PL NULL
#define OPTION_VAL_120_O47_PL NULL
#define OPTION_VAL_130_O47_PL NULL
#define OPTION_VAL_140_O47_PL NULL
#define OPTION_VAL_150_O47_PL NULL
#define OPTION_VAL_160_O47_PL NULL
#define OPTION_VAL_170_O47_PL NULL
#define OPTION_VAL_180_O47_PL NULL
#define OPTION_VAL_190_O47_PL NULL
#define OPTION_VAL_200_O47_PL NULL
#define OPTION_VAL_210_O47_PL NULL
#define OPTION_VAL_220_O47_PL NULL
#define OPTION_VAL_230_O47_PL NULL
#define OPTION_VAL_240_O47_PL NULL
#define OPTION_VAL_250_O47_PL NULL
#define OPTION_VAL_260_O47_PL NULL
#define OPTION_VAL_270_O47_PL NULL
#define OPTION_VAL_280_O47_PL NULL
#define OPTION_VAL_290_O47_PL NULL
#define OPTION_VAL_300_O47_PL NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_PL NULL
#define OPTION_VAL_WHITE_PL NULL
#define OPTION_VAL_RED_PL "Czerwony"
#define OPTION_VAL_GREEN_PL "Zielony"
#define OPTION_VAL_BLUE_PL "Niebieski"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_PL NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_PL NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_PL NULL
#define OPTION_VAL_VMU_PL NULL
#define OPTION_VAL_PURUPURU_PL NULL
#define OPTION_VAL_DREAMPOTATO_PL NULL
#define OPTION_VAL_NONE_PL "Brak"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_PL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_PL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_PL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_PL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_PL "Gdy wyłączone, wszystkie gry udostępniają do 8 plików zapisów VMU (A1/A2/B1/B2/C1/C2/D1/D2) znajdujących się w folderze systemowym RetroArch.\nUstawienie 'VMU A1' tworzy unikalny plik VMU 'A1' w folderze zapisów RetroArcha dla każdej uruchomionej gry.\nUstawienie 'Wszystkie VMU' tworzy do 8 unikalnych plików VMU (A1/A2/B1/B2/C1/C2/D1/D2/D2) dla każdej uruchomionej gry."
#define OPTION_VAL_VMU_A1_PL NULL
#define OPTION_VAL_ALL_VMUS_PL NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_PL "Dźwięki VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_PL "Po włączeniu dźwięk VMU jest odtwarzany."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_PL NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_PL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_PL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_PL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_PL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_PL NULL
#define OPTION_VAL_UPPER_LEFT_PL NULL
#define OPTION_VAL_UPPER_RIGHT_PL NULL
#define OPTION_VAL_LOWER_LEFT_PL NULL
#define OPTION_VAL_LOWER_RIGHT_PL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_PL NULL
#define OPTION_VAL_1X_PL NULL
#define OPTION_VAL_3X_PL NULL
#define OPTION_VAL_5X_PL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_PL NULL
#define OPTION_VAL_DEFAULT_ON_00_PL NULL
#define OPTION_VAL_DEFAULT_OFF_01_PL NULL
#define OPTION_VAL_BLACK_02_PL NULL
#define OPTION_VAL_LIGHT_BLUE_04_PL NULL
#define OPTION_VAL_CYAN_06_PL NULL
#define OPTION_VAL_CYAN_BLUE_07_PL NULL
#define OPTION_VAL_LIGHT_GREEN_08_PL NULL
#define OPTION_VAL_CYAN_GREEN_09_PL NULL
#define OPTION_VAL_LIGHT_CYAN_10_PL NULL
#define OPTION_VAL_PURPLE_12_PL NULL
#define OPTION_VAL_LIGHT_PURPLE_13_PL NULL
#define OPTION_VAL_YELLOW_14_PL "Żółty"
#define OPTION_VAL_GRAY_15_PL NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_PL NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_PL NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_PL NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_PL NULL
#define OPTION_VAL_LIGHT_RED_2_20_PL NULL
#define OPTION_VAL_MAGENTA_21_PL NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_PL NULL
#define OPTION_VAL_LIGHT_ORANGE_23_PL NULL
#define OPTION_VAL_ORANGE_24_PL NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_PL NULL
#define OPTION_VAL_LIGHT_YELLOW_26_PL NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_PL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_PL NULL
#define OPTION_VAL_40_PL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_PL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_PL NULL

struct retro_core_option_v2_category option_cats_pl[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_PL,
      CATEGORY_SYSTEM_INFO_0_PL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_PL,
      CATEGORY_VIDEO_INFO_0_PL
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_PL,
      CATEGORY_PERFORMANCE_INFO_0_PL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_PL,
      CATEGORY_HACKS_INFO_0_PL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_PL,
      CATEGORY_INPUT_INFO_0_PL
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_PL,
      CATEGORY_EXPANSIONS_INFO_0_PL
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_PL,
      CATEGORY_VMU_INFO_0_PL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_pl[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_PL,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_PL },
         { "USA",     OPTION_VAL_USA_PL },
         { "Europe",  OPTION_VAL_EUROPE_PL },
         { "Default", OPTION_VAL_DEFAULT_PL },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_PL,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_PL },
         { "English",  OPTION_VAL_ENGLISH_PL },
         { "German",   OPTION_VAL_GERMAN_PL },
         { "French",   OPTION_VAL_FRENCH_PL },
         { "Spanish",  OPTION_VAL_SPANISH_PL },
         { "Italian",  OPTION_VAL_ITALIAN_PL },
         { "Default",  OPTION_VAL_DEFAULT_PL },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_PL,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_PL },
         { "640x480",    OPTION_VAL_640X480_PL },
         { "800x600",    OPTION_VAL_800X600_PL },
         { "960x720",    OPTION_VAL_960X720_PL },
         { "1024x768",   OPTION_VAL_1024X768_PL },
         { "1280x960",   OPTION_VAL_1280X960_PL },
         { "1440x1080",  OPTION_VAL_1440X1080_PL },
         { "1600x1200",  OPTION_VAL_1600X1200_PL },
         { "1920x1440",  OPTION_VAL_1920X1440_PL },
         { "2560x1920",  OPTION_VAL_2560X1920_PL },
         { "2880x2160",  OPTION_VAL_2880X2160_PL },
         { "3200x2400",  OPTION_VAL_3200X2400_PL },
         { "3840x2880",  OPTION_VAL_3840X2880_PL },
         { "4480x3360",  OPTION_VAL_4480X3360_PL },
         { "5120x3840",  OPTION_VAL_5120X3840_PL },
         { "5760x4320",  OPTION_VAL_5760X4320_PL },
         { "6400x4800",  OPTION_VAL_6400X4800_PL },
         { "7040x5280",  OPTION_VAL_7040X5280_PL },
         { "7680x5760",  OPTION_VAL_7680X5760_PL },
         { "8320x6240",  OPTION_VAL_8320X6240_PL },
         { "8960x6720",  OPTION_VAL_8960X6720_PL },
         { "9600x7200",  OPTION_VAL_9600X7200_PL },
         { "10240x7680", OPTION_VAL_10240X7680_PL },
         { "10880x8160", OPTION_VAL_10880X8160_PL },
         { "11520x8640", OPTION_VAL_11520X8640_PL },
         { "12160x9120", OPTION_VAL_12160X9120_PL },
         { "12800x9600", OPTION_VAL_12800X9600_PL },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_PL,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_PL },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_PL },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_PL },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_PL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_PL },
         { "PAL",     OPTION_VAL_PAL_PL },
         { "PAL_N",   OPTION_VAL_PAL_N_PL },
         { "PAL_M",   OPTION_VAL_PAL_M_PL },
         { "Default", OPTION_VAL_DEFAULT_PL },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_PL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_PL },
         { "vertical",   OPTION_VAL_VERTICAL_PL },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_PL,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_PL },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_PL },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_PL },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_PL,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_PL },
         { "1GB",   OPTION_VAL_1GB_PL },
         { "2GB",   OPTION_VAL_2GB_PL },
         { "4GB",   OPTION_VAL_4GB_PL },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_PL,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_PL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_PL,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_PL,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_PL },
         { "1",  OPTION_VAL_1_PL },
         { "2",  OPTION_VAL_2_PL },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_PL,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_PL },
         { "4",  OPTION_VAL_4_PL },
         { "6",  OPTION_VAL_6_PL },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_PL,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_PL,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_PL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_PL,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_PL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_PL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_PL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_PL },
         { "more", OPTION_VAL_MORE_PL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_PL,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_PL },
         { "110", OPTION_VAL_110_PL },
         { "120", OPTION_VAL_120_PL },
         { "130", OPTION_VAL_130_PL },
         { "140", OPTION_VAL_140_PL },
         { "150", OPTION_VAL_150_PL },
         { "160", OPTION_VAL_160_PL },
         { "170", OPTION_VAL_170_PL },
         { "180", OPTION_VAL_180_PL },
         { "190", OPTION_VAL_190_PL },
         { "200", OPTION_VAL_200_PL },
         { "210", OPTION_VAL_210_PL },
         { "220", OPTION_VAL_220_PL },
         { "230", OPTION_VAL_230_PL },
         { "240", OPTION_VAL_240_PL },
         { "250", OPTION_VAL_250_PL },
         { "260", OPTION_VAL_260_PL },
         { "270", OPTION_VAL_270_PL },
         { "280", OPTION_VAL_280_PL },
         { "290", OPTION_VAL_290_PL },
         { "300", OPTION_VAL_300_PL },
         { "310", OPTION_VAL_310_PL },
         { "320", OPTION_VAL_320_PL },
         { "330", OPTION_VAL_330_PL },
         { "340", OPTION_VAL_340_PL },
         { "350", OPTION_VAL_350_PL },
         { "360", OPTION_VAL_360_PL },
         { "370", OPTION_VAL_370_PL },
         { "380", OPTION_VAL_380_PL },
         { "390", OPTION_VAL_390_PL },
         { "400", OPTION_VAL_400_PL },
         { "410", OPTION_VAL_410_PL },
         { "420", OPTION_VAL_420_PL },
         { "430", OPTION_VAL_430_PL },
         { "440", OPTION_VAL_440_PL },
         { "450", OPTION_VAL_450_PL },
         { "460", OPTION_VAL_460_PL },
         { "470", OPTION_VAL_470_PL },
         { "480", OPTION_VAL_480_PL },
         { "490", OPTION_VAL_490_PL },
         { "500", OPTION_VAL_500_PL },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_PL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_PL },
         { "5%",  OPTION_VAL_5_PL },
         { "10%", OPTION_VAL_10_PL },
         { "15%", OPTION_VAL_15_PL },
         { "20%", OPTION_VAL_20_PL },
         { "25%", OPTION_VAL_25_PL },
         { "30%", OPTION_VAL_30_PL },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_PL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_PL },
         { "5%",  OPTION_VAL_5_PL },
         { "10%", OPTION_VAL_10_PL },
         { "15%", OPTION_VAL_15_PL },
         { "20%", OPTION_VAL_20_PL },
         { "25%", OPTION_VAL_25_PL },
         { "30%", OPTION_VAL_30_PL },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_PL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_PL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_PL,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_PL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_PL },
         { "60%",  OPTION_VAL_60_PL },
         { "70%",  OPTION_VAL_70_PL },
         { "80%",  OPTION_VAL_80_PL },
         { "90%",  OPTION_VAL_90_PL },
         { "100%", OPTION_VAL_100_O47_PL },
         { "110%", OPTION_VAL_110_O47_PL },
         { "120%", OPTION_VAL_120_O47_PL },
         { "130%", OPTION_VAL_130_O47_PL },
         { "140%", OPTION_VAL_140_O47_PL },
         { "150%", OPTION_VAL_150_O47_PL },
         { "160%", OPTION_VAL_160_O47_PL },
         { "170%", OPTION_VAL_170_O47_PL },
         { "180%", OPTION_VAL_180_O47_PL },
         { "190%", OPTION_VAL_190_O47_PL },
         { "200%", OPTION_VAL_200_O47_PL },
         { "210%", OPTION_VAL_210_O47_PL },
         { "220%", OPTION_VAL_220_O47_PL },
         { "230%", OPTION_VAL_230_O47_PL },
         { "240%", OPTION_VAL_240_O47_PL },
         { "250%", OPTION_VAL_250_O47_PL },
         { "260%", OPTION_VAL_260_O47_PL },
         { "270%", OPTION_VAL_270_O47_PL },
         { "280%", OPTION_VAL_280_O47_PL },
         { "290%", OPTION_VAL_290_O47_PL },
         { "300%", OPTION_VAL_300_O47_PL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_PL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PL },
         { "Red",      OPTION_VAL_RED_PL },
         { "Green",    OPTION_VAL_GREEN_PL },
         { "Blue",     OPTION_VAL_BLUE_PL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_PL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PL },
         { "Red",      OPTION_VAL_RED_PL },
         { "Green",    OPTION_VAL_GREEN_PL },
         { "Blue",     OPTION_VAL_BLUE_PL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_PL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PL },
         { "Red",      OPTION_VAL_RED_PL },
         { "Green",    OPTION_VAL_GREEN_PL },
         { "Blue",     OPTION_VAL_BLUE_PL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_PL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PL },
         { "Red",      OPTION_VAL_RED_PL },
         { "Green",    OPTION_VAL_GREEN_PL },
         { "Blue",     OPTION_VAL_BLUE_PL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_PL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PL },
         { "Purupuru", OPTION_VAL_PURUPURU_PL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_PL },
         { "None",     OPTION_VAL_NONE_PL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_PL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PL },
         { "Purupuru", OPTION_VAL_PURUPURU_PL },
         { "None",     OPTION_VAL_NONE_PL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_PL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PL },
         { "Purupuru", OPTION_VAL_PURUPURU_PL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_PL },
         { "None",     OPTION_VAL_NONE_PL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_PL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PL },
         { "Purupuru", OPTION_VAL_PURUPURU_PL },
         { "None",     OPTION_VAL_NONE_PL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_PL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PL },
         { "Purupuru", OPTION_VAL_PURUPURU_PL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_PL },
         { "None",     OPTION_VAL_NONE_PL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_PL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PL },
         { "Purupuru", OPTION_VAL_PURUPURU_PL },
         { "None",     OPTION_VAL_NONE_PL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_PL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PL },
         { "Purupuru", OPTION_VAL_PURUPURU_PL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_PL },
         { "None",     OPTION_VAL_NONE_PL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_PL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PL },
         { "Purupuru", OPTION_VAL_PURUPURU_PL },
         { "None",     OPTION_VAL_NONE_PL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_PL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_PL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_PL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_PL },
         { "All VMUs", OPTION_VAL_ALL_VMUS_PL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_PL,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_PL,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_PL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_PL,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_PL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_PL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_PL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_PL,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PL },
         { "2x", OPTION_VAL_2_O26_PL },
         { "3x", OPTION_VAL_3X_PL },
         { "4x", OPTION_VAL_4_PL },
         { "5x", OPTION_VAL_5X_PL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PL },
         { "20%",  OPTION_VAL_20_PL },
         { "30%",  OPTION_VAL_30_PL },
         { "40%",  OPTION_VAL_40_PL },
         { "50%",  OPTION_VAL_50_PL },
         { "60%",  OPTION_VAL_60_PL },
         { "70%",  OPTION_VAL_70_PL },
         { "80%",  OPTION_VAL_80_PL },
         { "90%",  OPTION_VAL_90_PL },
         { "100%", OPTION_VAL_100_O47_PL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PL },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PL },
         { "2x", OPTION_VAL_2_O26_PL },
         { "3x", OPTION_VAL_3X_PL },
         { "4x", OPTION_VAL_4_PL },
         { "5x", OPTION_VAL_5X_PL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PL },
         { "20%",  OPTION_VAL_20_PL },
         { "30%",  OPTION_VAL_30_PL },
         { "40%",  OPTION_VAL_40_PL },
         { "50%",  OPTION_VAL_50_PL },
         { "60%",  OPTION_VAL_60_PL },
         { "70%",  OPTION_VAL_70_PL },
         { "80%",  OPTION_VAL_80_PL },
         { "90%",  OPTION_VAL_90_PL },
         { "100%", OPTION_VAL_100_O47_PL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PL },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PL },
         { "2x", OPTION_VAL_2_O26_PL },
         { "3x", OPTION_VAL_3X_PL },
         { "4x", OPTION_VAL_4_PL },
         { "5x", OPTION_VAL_5X_PL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PL },
         { "20%",  OPTION_VAL_20_PL },
         { "30%",  OPTION_VAL_30_PL },
         { "40%",  OPTION_VAL_40_PL },
         { "50%",  OPTION_VAL_50_PL },
         { "60%",  OPTION_VAL_60_PL },
         { "70%",  OPTION_VAL_70_PL },
         { "80%",  OPTION_VAL_80_PL },
         { "90%",  OPTION_VAL_90_PL },
         { "100%", OPTION_VAL_100_O47_PL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PL },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PL },
         { "2x", OPTION_VAL_2_O26_PL },
         { "3x", OPTION_VAL_3X_PL },
         { "4x", OPTION_VAL_4_PL },
         { "5x", OPTION_VAL_5X_PL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PL },
         { "BLUE 03",           OPTION_VAL_BLUE_PL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PL },
         { "GREEN 05",          OPTION_VAL_GREEN_PL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PL },
         { "RED 11",            OPTION_VAL_RED_PL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PL },
         { "WHITE 28",          OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_PL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PL },
         { "20%",  OPTION_VAL_20_PL },
         { "30%",  OPTION_VAL_30_PL },
         { "40%",  OPTION_VAL_40_PL },
         { "50%",  OPTION_VAL_50_PL },
         { "60%",  OPTION_VAL_60_PL },
         { "70%",  OPTION_VAL_70_PL },
         { "80%",  OPTION_VAL_80_PL },
         { "90%",  OPTION_VAL_90_PL },
         { "100%", OPTION_VAL_100_O47_PL },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_pl = {
   option_cats_pl,
   option_defs_pl
};

/* RETRO_LANGUAGE_PT_BR */

#define CATEGORY_SYSTEM_LABEL_PT_BR "Sistema"
#define CATEGORY_SYSTEM_INFO_0_PT_BR "Altera as configurações de hardware básicas, região, idioma e BIOS."
#define CATEGORY_VIDEO_LABEL_PT_BR "Vídeo"
#define CATEGORY_VIDEO_INFO_0_PT_BR "Altera as configurações de resolução, transparência e efeitos visuais."
#define CATEGORY_PERFORMANCE_LABEL_PT_BR "Desempenho"
#define CATEGORY_PERFORMANCE_INFO_0_PT_BR "Altera as configurações da renderização multithread e pulo de quadros."
#define CATEGORY_HACKS_LABEL_PT_BR "Hacks de emulação"
#define CATEGORY_HACKS_INFO_0_PT_BR "Altera as configurações de tela panorâmica, velocidade de carregamento do GD-ROM e substituição de textura."
#define CATEGORY_INPUT_LABEL_PT_BR "Entrada"
#define CATEGORY_INPUT_INFO_0_PT_BR "Altera as configurações do controle e pistola de luz."
#define CATEGORY_EXPANSIONS_LABEL_PT_BR "Encaixe de expansão dos controles"
#define CATEGORY_EXPANSIONS_INFO_0_PT_BR "Define o dispositivo (VMU ou Aparelho de Vibração) conectado em cada encaixe de expansão do controle."
#define CATEGORY_VMU_LABEL_PT_BR "Unidade de Memória Visual (VMU)"
#define CATEGORY_VMU_INFO_0_PT_BR "Altera as configurações de salvamento do VMU e aparência do VMU na tela."
#define CORE_OPTION_NAME_REGION_LABEL_PT_BR "Região"
#define OPTION_VAL_JAPAN_PT_BR "Japão"
#define OPTION_VAL_USA_PT_BR "EUA"
#define OPTION_VAL_EUROPE_PT_BR "Europa"
#define OPTION_VAL_DEFAULT_PT_BR "Padrão"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_PT_BR "Idioma"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_PT_BR "Altera o idioma usado pelo BIOS e por quaisquer jogos com múltiplos idiomas."
#define OPTION_VAL_JAPANESE_PT_BR "Japonês"
#define OPTION_VAL_ENGLISH_PT_BR "Inglês"
#define OPTION_VAL_GERMAN_PT_BR "Alemão"
#define OPTION_VAL_FRENCH_PT_BR "Francês"
#define OPTION_VAL_SPANISH_PT_BR "Espanhol"
#define OPTION_VAL_ITALIAN_PT_BR "Italiano"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_PT_BR "BIOS HLE (requer reinício)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_PT_BR "Força o uso de um BIOS de emulação de alto nível (HLE)."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_PT_BR "Ativar DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_PT_BR "Ativa a emulação de áudio do DSP (processador de sinal digital) do Dreamcast. Otimiza a precisão de emulação do som, mas aumenta os requisitos mínimos de desempenho."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_PT_BR "Permitir botão Assistência do fliperama"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_PT_BR "Habilita o botão ASSISTÊNCIA em jogos de fliperama, para entrar nas configurações."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_PT_BR "Definir jogos NAOMI como gratuitos"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_PT_BR "Define as configurações de moeda do jogo como gratuitas."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_PT_BR "Emulação do Broadband Adapter"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_PT_BR "Emula o adaptador de banda larga por Ethernet em vez do modem. (requer reinício)"
#define CORE_OPTION_NAME_UPNP_LABEL_PT_BR "Ativar UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_PT_BR "Usa o UPnP para configurar automaticamente seu roteador de Internet para jogos on-line."
#define CORE_OPTION_NAME_DCNET_LABEL_PT_BR "Ativar DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_PT_BR "Usa o serviço de nuvem do DCNet para acessar a Internet do Dreamcast."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_PT_BR "Resolução interna"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_PT_BR "Define a resolução de renderização."
#define OPTION_VAL_320X240_PT_BR "320x240 (50%)"
#define OPTION_VAL_640X480_PT_BR "640x480 (nativa)"
#define OPTION_VAL_800X600_PT_BR "800x600 (125%)"
#define OPTION_VAL_960X720_PT_BR "960x720 (150%)"
#define OPTION_VAL_1024X768_PT_BR "1024x768 (160%)"
#define OPTION_VAL_1280X960_PT_BR "1280x960 (200%)"
#define OPTION_VAL_1440X1080_PT_BR "1440x1080 (225%)"
#define OPTION_VAL_1600X1200_PT_BR "1600x1200 (250%)"
#define OPTION_VAL_1920X1440_PT_BR "1920x1440 (300%)"
#define OPTION_VAL_2560X1920_PT_BR "2560x1920 (400%)"
#define OPTION_VAL_2880X2160_PT_BR "2880x2160 (450%)"
#define OPTION_VAL_3200X2400_PT_BR "3200x2400 (500%)"
#define OPTION_VAL_3840X2880_PT_BR "3840x2880 (600%)"
#define OPTION_VAL_4480X3360_PT_BR "4480x3360 (700%)"
#define OPTION_VAL_5120X3840_PT_BR "5120x3840 (800%)"
#define OPTION_VAL_5760X4320_PT_BR "5760x4320 (900%)"
#define OPTION_VAL_6400X4800_PT_BR "6400x4800 (1000%)"
#define OPTION_VAL_7040X5280_PT_BR "7040x5280 (1100%)"
#define OPTION_VAL_7680X5760_PT_BR "7680x5760 (1200%)"
#define OPTION_VAL_8320X6240_PT_BR "8320x6240 (1300%)"
#define OPTION_VAL_8960X6720_PT_BR "8960x6720 (1400%)"
#define OPTION_VAL_9600X7200_PT_BR "9600x7200 (1500%)"
#define OPTION_VAL_10240X7680_PT_BR "10240x7680 (1600%)"
#define OPTION_VAL_10880X8160_PT_BR "10880x8160 (1700%)"
#define OPTION_VAL_11520X8640_PT_BR "11520x8640 (1800%)"
#define OPTION_VAL_12160X9120_PT_BR "12160x9120 (1900%)"
#define OPTION_VAL_12800X9600_PT_BR "12800x9600 (2000%)"
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_PT_BR "Cabo"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_PT_BR "Define o sinal de saída.\n\"TV (composto)\": é o mais recomendando por ter mais suporte."
#define OPTION_VAL_VGA_PT_BR NULL
#define OPTION_VAL_TV_RGB_PT_BR NULL
#define OPTION_VAL_TV_COMPOSITE_PT_BR "TV (composto)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_PT_BR "Sistema de transmissão"
#define OPTION_VAL_NTSC_PT_BR NULL
#define OPTION_VAL_PAL_PT_BR "PAL"
#define OPTION_VAL_PAL_N_PT_BR "PAL-N (Argentina, Paraguai e Uruguai)"
#define OPTION_VAL_PAL_M_PT_BR "PAL-M (Brasil)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_PT_BR "Orientação de tela"
#define OPTION_VAL_HORIZONTAL_PT_BR NULL
#define OPTION_VAL_VERTICAL_PT_BR NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_PT_BR "Separador de Alfa"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_PT_BR "Define como os polígonos de transparência são separados."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_PT_BR "Faixa (rapidez)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_PT_BR "Triângulo (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_PT_BR "Píxel (precisão)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_PT_BR "Buffer da acumulação de píxel"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_PT_BR "Valores maiores podem ser necessários em resoluções maiores."
#define OPTION_VAL_512MB_PT_BR "512 MB"
#define OPTION_VAL_1GB_PT_BR "1 GB"
#define OPTION_VAL_2GB_PT_BR "2 GB"
#define OPTION_VAL_4GB_PT_BR "4 GB"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_PT_BR "Máximo de camadas transparentes"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_PT_BR "Valores maiores podem se necessários em cenários complexos."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_PT_BR "Emulação completa de framebuffer"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_PT_BR "Ativa a emulação completa de framebuffer na VRAM. É útil em jogos que leem ou gravam o framebuffer na VRAM.\nOBSERVAÇÃO: a resolução interna é forçada à 640x480 e o desempenho pode ser gravemente afetado."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_PT_BR "Ativar buffer RTT (renderização para texturas)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_PT_BR "Copia as texturas renderizadas da GPU de volta para a VRAM. É normalmente ativada em jogos que necessitam dessa opção.\nOBSERVAÇÃO: a mudança de escala (upscaling) da renderização de textura é desativada e o desempenho pode ser afetado."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_PT_BR "Piramidar (mipmapping)"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_PT_BR "As texturas usam uma versão menor delas mesmas, ao aparecer mais de longe, pode aumentar o desempenho e reduzir a tremeluzência."
#define CORE_OPTION_NAME_FOG_LABEL_PT_BR "Efeito de nevoeiro"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_PT_BR "Modificador de volume"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_PT_BR "Um recurso da GPU do Dreamcast que é normalmente usado para sombrar objetos e tem pouco efeito sobre o desempenho."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_PT_BR "Filtro anisotrópico"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_PT_BR "Otimiza a qualidade de texturas em superfícies, que estão em ângulos de visão oblíquos em relação à câmera. Valores maiores exigem mais da GPU.\nOBSERVAÇÃO: requer reinício."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_PT_BR "Filtragem de textura"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_PT_BR "Define a filtragem de textura. Pode ser usada para forçar uma filtragem de textura em todas as texturas, para obter uma aparência mais nítida.\nOBSERVAÇÃO: requer reinício e pode causar vários erros de renderização."
#define OPTION_VAL_1_PT_BR "Forçar adjacentes"
#define OPTION_VAL_2_PT_BR "Forçar linear"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_PT_BR "Atrasar troca de quadro"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_PT_BR "Útil para evitar telas intermitentes ou defeitos de vídeo.\nOBSERVAÇÃO: não é recomendado em plataformas lentas."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_PT_BR "Detectar alterações da taxa de quadros"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_PT_BR "Notifica quando a taxa de quadros interna é alterada (por ex.: de 60 qps para 30 qps). Otimiza a frequência de quadros em jogos travados a 30 qps ou 20 qps, mas não deve ser habilitado em jogos com taxa de quadros instável (por ex.: Ecco the Dolphin, Unreal Tournament etc.). OBSERVAÇÃO: requer \"Pulo de quadros automático\" desativado."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_PT_BR "Filtro de pós-processamento (PowerVR2)"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_PT_BR "Processa a imagem renderizada para simular efeitos específicos da GPU PowerVR2 e sinais de vídeo analógicos."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_PT_BR "Mudar escala de textura (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_PT_BR "Realça os gráficos 2D de arte píxel feitos à mão.\nOBSERVAÇÃO: usar apenas em jogos 2D pixelado."
#define OPTION_VAL_2_O26_PT_BR "200%"
#define OPTION_VAL_4_PT_BR "400%"
#define OPTION_VAL_6_PT_BR "600%"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_PT_BR "Tamanho máximo das texturas"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_PT_BR "Define o tamanho máximo que uma textura ser aumentada, se o tamanho da textura for maior que o valor definido, enquanto não será aumentada."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_PT_BR "Interpolação nativa de profundidade"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_PT_BR "Ajuda com corrupção de textura e problemas de profundidade em GPU da AMD. Pode ajudar em GPU da Intel em alguns casos."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_PT_BR "Corrigir vazamento de arestas"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_PT_BR "Ajuda com vazamento de textura ao mudar escala. Desativar pode ajudar se os píxeis são deformados pela mudança de escala em jogos 2D (MVC2, CVS, KOF etc.)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_PT_BR "Renderização multithread"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_PT_BR "Executa a GPU e a CPU em diferentes threads. Altamente recomendado."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_PT_BR "Pulo de quadros automático"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_PT_BR "Pula quadros automaticamente quando o emulador está lento.\nOBSERVAÇÃO: requer \"Renderização multithread\" ativada."
#define OPTION_VAL_SOME_PT_BR NULL
#define OPTION_VAL_MORE_PT_BR "Máximo"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_PT_BR "Pulo de quadros"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_PT_BR "Define o número de quadros a serem pulados entre cada quadro exibido."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_PT_BR "Trapaça panorâmica (requer reinício)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_PT_BR "Permite uma projeção panorâmica (widescreen) em alguns jogos."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_PT_BR "Hack panorâmico"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_PT_BR "Desenha a geometria além da proporção de tela normal de 4:3.\nOBSERVAÇÃO: pode causar falhas gráficas."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_PT_BR "Carregamento rápido do GD-ROM (impreciso)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_PT_BR "Acelera o carregamento do GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_PT_BR "Modificação de 32 MB de RAM"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_PT_BR "Ativa uma modificação de 32 MB de RAM para o Dreamcast.\nOBSERVAÇÃO: pode afetar compatibilidade."
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_PT_BR "Velocidade da CPU SH-4"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_PT_BR "Altera a velocidade padrão (200 MHz) da CPU SH-4. Reduzir a velocidade pode ajudar em plataformas lentas. Aumentar a velocidade pode aumentar a taxa de quadros em alguns jogos.\nOBSERVAÇÃO: cuidado."
#define OPTION_VAL_100_PT_BR NULL
#define OPTION_VAL_110_PT_BR NULL
#define OPTION_VAL_120_PT_BR NULL
#define OPTION_VAL_130_PT_BR NULL
#define OPTION_VAL_140_PT_BR NULL
#define OPTION_VAL_150_PT_BR NULL
#define OPTION_VAL_160_PT_BR NULL
#define OPTION_VAL_170_PT_BR NULL
#define OPTION_VAL_180_PT_BR NULL
#define OPTION_VAL_190_PT_BR NULL
#define OPTION_VAL_200_PT_BR NULL
#define OPTION_VAL_210_PT_BR NULL
#define OPTION_VAL_220_PT_BR NULL
#define OPTION_VAL_230_PT_BR NULL
#define OPTION_VAL_240_PT_BR NULL
#define OPTION_VAL_250_PT_BR NULL
#define OPTION_VAL_260_PT_BR NULL
#define OPTION_VAL_270_PT_BR NULL
#define OPTION_VAL_280_PT_BR NULL
#define OPTION_VAL_290_PT_BR NULL
#define OPTION_VAL_300_PT_BR NULL
#define OPTION_VAL_310_PT_BR NULL
#define OPTION_VAL_320_PT_BR NULL
#define OPTION_VAL_330_PT_BR NULL
#define OPTION_VAL_340_PT_BR NULL
#define OPTION_VAL_350_PT_BR NULL
#define OPTION_VAL_360_PT_BR NULL
#define OPTION_VAL_370_PT_BR NULL
#define OPTION_VAL_380_PT_BR NULL
#define OPTION_VAL_390_PT_BR NULL
#define OPTION_VAL_400_PT_BR NULL
#define OPTION_VAL_410_PT_BR NULL
#define OPTION_VAL_420_PT_BR NULL
#define OPTION_VAL_430_PT_BR NULL
#define OPTION_VAL_440_PT_BR NULL
#define OPTION_VAL_450_PT_BR NULL
#define OPTION_VAL_460_PT_BR NULL
#define OPTION_VAL_470_PT_BR NULL
#define OPTION_VAL_480_PT_BR NULL
#define OPTION_VAL_490_PT_BR NULL
#define OPTION_VAL_500_PT_BR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_PT_BR "Carregar texturas personalizadas"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_PT_BR "Carrega as texturas personalizadas.\nEndereço da pasta: \"system/dc/textures/<game-id>/\"."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_PT_BR "Pré-carregar texturas personalizadas"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_PT_BR "Pré-carrega as texturas personalizadas ao iniciar um jogo. Otimiza o desempenho, mas aumenta o uso de memória."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_PT_BR "Descarregar texturas"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_PT_BR "Toda vez que uma nova textura for usada pelo jogo, a textura será salva como um .PNG na pasta \"system/dc/texdump/<game-id>/\"."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_PT_BR "Descarregar texturas substituídas"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_PT_BR "Descarrega as texturas que já foram substituídas pelas texturas personalizadas."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_PT_BR "Zona morta do direcional analógico"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_PT_BR "Define a distância mínima do direcional analógico para ser ativado."
#define OPTION_VAL_0_PT_BR NULL
#define OPTION_VAL_5_PT_BR NULL
#define OPTION_VAL_10_PT_BR NULL
#define OPTION_VAL_15_PT_BR NULL
#define OPTION_VAL_20_PT_BR NULL
#define OPTION_VAL_25_PT_BR NULL
#define OPTION_VAL_30_PT_BR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_PT_BR "Zona morta dos gatilhos"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_PT_BR "Define a distância mínima do gatilho para ser ativado."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_PT_BR "Gatilhos digitais"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_PT_BR "Os gatilhos digitais agem igual os botões comuns, então serão processados como totalmente pressionados ou não pressionados, sem meio-termo."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_PT_BR "Transmitir saídas digitais"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_PT_BR "Transmite saídas digitais e estado da resposta de força (force-feedback) na porta TCP 8000. Compatível com a configuração MAME \"-output network\"."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_PT_BR "Mostrar configurações da pistola de luz"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_PT_BR "Ativa as configurações de mira na tela.\nOBSERVAÇÃO: requer reinício do menu rápido."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_PT_BR "Tamanho da mira"
#define OPTION_VAL_50_PT_BR NULL
#define OPTION_VAL_60_PT_BR NULL
#define OPTION_VAL_70_PT_BR NULL
#define OPTION_VAL_80_PT_BR NULL
#define OPTION_VAL_90_PT_BR NULL
#define OPTION_VAL_100_O47_PT_BR NULL
#define OPTION_VAL_110_O47_PT_BR NULL
#define OPTION_VAL_120_O47_PT_BR NULL
#define OPTION_VAL_130_O47_PT_BR NULL
#define OPTION_VAL_140_O47_PT_BR NULL
#define OPTION_VAL_150_O47_PT_BR NULL
#define OPTION_VAL_160_O47_PT_BR NULL
#define OPTION_VAL_170_O47_PT_BR NULL
#define OPTION_VAL_180_O47_PT_BR NULL
#define OPTION_VAL_190_O47_PT_BR NULL
#define OPTION_VAL_200_O47_PT_BR NULL
#define OPTION_VAL_210_O47_PT_BR NULL
#define OPTION_VAL_220_O47_PT_BR NULL
#define OPTION_VAL_230_O47_PT_BR NULL
#define OPTION_VAL_240_O47_PT_BR NULL
#define OPTION_VAL_250_O47_PT_BR NULL
#define OPTION_VAL_260_O47_PT_BR NULL
#define OPTION_VAL_270_O47_PT_BR NULL
#define OPTION_VAL_280_O47_PT_BR NULL
#define OPTION_VAL_290_O47_PT_BR NULL
#define OPTION_VAL_300_O47_PT_BR NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_PT_BR "Mira da arma 1"
#define OPTION_VAL_WHITE_PT_BR "Branco"
#define OPTION_VAL_RED_PT_BR "Vermelho"
#define OPTION_VAL_GREEN_PT_BR "Verde"
#define OPTION_VAL_BLUE_PT_BR "Azul"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_PT_BR "Mira da arma 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_PT_BR "Mira da arma 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_PT_BR "Mira da arma 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_PT_BR "Encaixe A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_PT_BR "Define o dispositivo conectado no encaixe de expansão A1 (porta A, compartimento 1)."
#define OPTION_VAL_VMU_PT_BR NULL
#define OPTION_VAL_PURUPURU_PT_BR "Aparelho de Vibração"
#define OPTION_VAL_DREAMPOTATO_PT_BR NULL
#define OPTION_VAL_NONE_PT_BR "Nenhum"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_PT_BR "Encaixe A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_PT_BR "Define o dispositivo conectado no encaixe de expansão A2 (porta A, compartimento 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_PT_BR "Encaixe B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_PT_BR "Define o dispositivo conectado no encaixe de expansão B1 (porta B, compartimento 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_PT_BR "Encaixe B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_PT_BR "Define o dispositivo conectado no encaixe de expansão B2 (porta B, compartimento 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_PT_BR "Encaixe C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_PT_BR "Define o dispositivo conectado no encaixe de expansão C1 (porta C, compartimento 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_PT_BR "Encaixe C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_PT_BR "Define o dispositivo conectado no encaixe de expansão C2 (porta C, compartimento 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_PT_BR "Encaixe D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_PT_BR "Define o dispositivo conectado no encaixe de expansão D1 (porta D, compartimento 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_PT_BR "Encaixe D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_PT_BR "Define o dispositivo conectado no encaixe de expansão D2 (porta D, compartimento 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_PT_BR "VMU por jogo"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_PT_BR "VMU por jogo"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_PT_BR "\"DESLIGADO\": todos os jogos compartilham até 8 arquivos de salvamento (A1/A2/B1/B2/C1/C2/D1/D2), localizados na pasta de sistema do RetroArch.\n\"VMU A1\": cria um único arquivo para todos os jogos.\n\"Todos os VMU\": cria até 8 arquivos (A1/A2/B1/B2/C1/C2/D1/D2) para cada jogo."
#define OPTION_VAL_VMU_A1_PT_BR NULL
#define OPTION_VAL_ALL_VMUS_PT_BR "Todos os VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_PT_BR "Som do VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_PT_BR "Som do VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_PT_BR "Permite reproduzir os bipes do VMU."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_PT_BR "Armazenamento de VMU do DreamPotato"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_PT_BR "Ativa o armazenamento de dados de VMU no emulador DreamPotato conectado. Quando ativado, o arquivo de save local da VMU não é utilizado."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_PT_BR "Mostrar configurações de tela do VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_PT_BR "Mostrar configurações de tela do VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_PT_BR "Permite a configuração de visibilidade, tamanho, posição e cor das telas LCD do VMU emulado.\nOBSERVAÇÃO: requer reinício do menu rápido."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_PT_BR "VMU 1 — Tela"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_PT_BR "VMU 1 — Posição da tela"
#define OPTION_VAL_UPPER_LEFT_PT_BR "Canto superior esquerdo"
#define OPTION_VAL_UPPER_RIGHT_PT_BR "Canto superior direito"
#define OPTION_VAL_LOWER_LEFT_PT_BR "Canto inferior esquerdo"
#define OPTION_VAL_LOWER_RIGHT_PT_BR "Canto inferior direito"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_PT_BR "VMU 1 — Tamanho da tela"
#define OPTION_VAL_1X_PT_BR "Padrão (100%)"
#define OPTION_VAL_3X_PT_BR "300%"
#define OPTION_VAL_5X_PT_BR "500%"
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_PT_BR "VMU 1 — Cor dos píxeis acesos"
#define OPTION_VAL_DEFAULT_ON_00_PT_BR "Padrão (aceso)"
#define OPTION_VAL_DEFAULT_OFF_01_PT_BR "Padrão (apagado)"
#define OPTION_VAL_BLACK_02_PT_BR "Preto"
#define OPTION_VAL_LIGHT_BLUE_04_PT_BR "Azul claro"
#define OPTION_VAL_CYAN_06_PT_BR "Ciano"
#define OPTION_VAL_CYAN_BLUE_07_PT_BR "Azul ciano"
#define OPTION_VAL_LIGHT_GREEN_08_PT_BR "Verde claro"
#define OPTION_VAL_CYAN_GREEN_09_PT_BR "Verde ciano"
#define OPTION_VAL_LIGHT_CYAN_10_PT_BR "Ciano claro"
#define OPTION_VAL_PURPLE_12_PT_BR "Roxo"
#define OPTION_VAL_LIGHT_PURPLE_13_PT_BR "Roxo claro"
#define OPTION_VAL_YELLOW_14_PT_BR "Amarelo"
#define OPTION_VAL_GRAY_15_PT_BR "Cinza"
#define OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR "Roxo claro (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_PT_BR "Verde claro (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_PT_BR "Verde claro (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_PT_BR "Ciano claro (2)"
#define OPTION_VAL_LIGHT_RED_2_20_PT_BR "Vermelho claro (2)"
#define OPTION_VAL_MAGENTA_21_PT_BR NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR "Roxo claro (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_PT_BR "Laranja claro"
#define OPTION_VAL_ORANGE_24_PT_BR "Laranja"
#define OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR "Roxo claro (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_PT_BR "Amarelo claro"
#define OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR "Amarelo claro (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_PT_BR "VMU 1 — Cor dos píxeis apagados"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_PT_BR "VMU 1 — Opacidade da tela"
#define OPTION_VAL_40_PT_BR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_PT_BR "VMU 2 — Tela"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_PT_BR "VMU 2 — Posição da tela"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_PT_BR "VMU 2 — Tamanho da tela"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_PT_BR "VMU 2 — Cor dos píxeis acesos"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_PT_BR "VMU 2 — Cor dos píxeis apagados"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_PT_BR "VMU 2 — Opacidade da tela"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_PT_BR "VMU 3 — Tela"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_PT_BR "VMU 3 — Posição da tela"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_PT_BR "VMU 3 — Tamanho da tela"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_PT_BR "VMU 3 — Cor dos píxeis acesos"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_PT_BR "VMU 3 — Cor dos píxeis apagados"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_PT_BR "VMU 3 — Opacidade da tela"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_PT_BR "VMU 4 — Tela"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_PT_BR "VMU 4 — Posição da tela"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_PT_BR "VMU 4 — Tamanho da tela"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_PT_BR "VMU 4 — Cor dos píxeis acesos"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_PT_BR "VMU 4 — Cor dos píxeis apagados"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_PT_BR "VMU 4 — Opacidade da tela"

struct retro_core_option_v2_category option_cats_pt_br[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_PT_BR,
      CATEGORY_SYSTEM_INFO_0_PT_BR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_PT_BR,
      CATEGORY_VIDEO_INFO_0_PT_BR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_PT_BR,
      CATEGORY_PERFORMANCE_INFO_0_PT_BR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_PT_BR,
      CATEGORY_HACKS_INFO_0_PT_BR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_PT_BR,
      CATEGORY_INPUT_INFO_0_PT_BR
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_PT_BR,
      CATEGORY_EXPANSIONS_INFO_0_PT_BR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_PT_BR,
      CATEGORY_VMU_INFO_0_PT_BR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_pt_br[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_PT_BR },
         { "USA",     OPTION_VAL_USA_PT_BR },
         { "Europe",  OPTION_VAL_EUROPE_PT_BR },
         { "Default", OPTION_VAL_DEFAULT_PT_BR },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_PT_BR },
         { "English",  OPTION_VAL_ENGLISH_PT_BR },
         { "German",   OPTION_VAL_GERMAN_PT_BR },
         { "French",   OPTION_VAL_FRENCH_PT_BR },
         { "Spanish",  OPTION_VAL_SPANISH_PT_BR },
         { "Italian",  OPTION_VAL_ITALIAN_PT_BR },
         { "Default",  OPTION_VAL_DEFAULT_PT_BR },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_PT_BR },
         { "640x480",    OPTION_VAL_640X480_PT_BR },
         { "800x600",    OPTION_VAL_800X600_PT_BR },
         { "960x720",    OPTION_VAL_960X720_PT_BR },
         { "1024x768",   OPTION_VAL_1024X768_PT_BR },
         { "1280x960",   OPTION_VAL_1280X960_PT_BR },
         { "1440x1080",  OPTION_VAL_1440X1080_PT_BR },
         { "1600x1200",  OPTION_VAL_1600X1200_PT_BR },
         { "1920x1440",  OPTION_VAL_1920X1440_PT_BR },
         { "2560x1920",  OPTION_VAL_2560X1920_PT_BR },
         { "2880x2160",  OPTION_VAL_2880X2160_PT_BR },
         { "3200x2400",  OPTION_VAL_3200X2400_PT_BR },
         { "3840x2880",  OPTION_VAL_3840X2880_PT_BR },
         { "4480x3360",  OPTION_VAL_4480X3360_PT_BR },
         { "5120x3840",  OPTION_VAL_5120X3840_PT_BR },
         { "5760x4320",  OPTION_VAL_5760X4320_PT_BR },
         { "6400x4800",  OPTION_VAL_6400X4800_PT_BR },
         { "7040x5280",  OPTION_VAL_7040X5280_PT_BR },
         { "7680x5760",  OPTION_VAL_7680X5760_PT_BR },
         { "8320x6240",  OPTION_VAL_8320X6240_PT_BR },
         { "8960x6720",  OPTION_VAL_8960X6720_PT_BR },
         { "9600x7200",  OPTION_VAL_9600X7200_PT_BR },
         { "10240x7680", OPTION_VAL_10240X7680_PT_BR },
         { "10880x8160", OPTION_VAL_10880X8160_PT_BR },
         { "11520x8640", OPTION_VAL_11520X8640_PT_BR },
         { "12160x9120", OPTION_VAL_12160X9120_PT_BR },
         { "12800x9600", OPTION_VAL_12800X9600_PT_BR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_PT_BR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_PT_BR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_PT_BR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_PT_BR },
         { "PAL",     OPTION_VAL_PAL_PT_BR },
         { "PAL_N",   OPTION_VAL_PAL_N_PT_BR },
         { "PAL_M",   OPTION_VAL_PAL_M_PT_BR },
         { "Default", OPTION_VAL_DEFAULT_PT_BR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_PT_BR },
         { "vertical",   OPTION_VAL_VERTICAL_PT_BR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_PT_BR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_PT_BR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_PT_BR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_PT_BR },
         { "1GB",   OPTION_VAL_1GB_PT_BR },
         { "2GB",   OPTION_VAL_2GB_PT_BR },
         { "4GB",   OPTION_VAL_4GB_PT_BR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_PT_BR },
         { "1",  OPTION_VAL_1_PT_BR },
         { "2",  OPTION_VAL_2_PT_BR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_PT_BR },
         { "4",  OPTION_VAL_4_PT_BR },
         { "6",  OPTION_VAL_6_PT_BR },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_PT_BR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_PT_BR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_PT_BR,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_PT_BR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_PT_BR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_PT_BR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_PT_BR },
         { "more", OPTION_VAL_MORE_PT_BR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_PT_BR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_PT_BR },
         { "110", OPTION_VAL_110_PT_BR },
         { "120", OPTION_VAL_120_PT_BR },
         { "130", OPTION_VAL_130_PT_BR },
         { "140", OPTION_VAL_140_PT_BR },
         { "150", OPTION_VAL_150_PT_BR },
         { "160", OPTION_VAL_160_PT_BR },
         { "170", OPTION_VAL_170_PT_BR },
         { "180", OPTION_VAL_180_PT_BR },
         { "190", OPTION_VAL_190_PT_BR },
         { "200", OPTION_VAL_200_PT_BR },
         { "210", OPTION_VAL_210_PT_BR },
         { "220", OPTION_VAL_220_PT_BR },
         { "230", OPTION_VAL_230_PT_BR },
         { "240", OPTION_VAL_240_PT_BR },
         { "250", OPTION_VAL_250_PT_BR },
         { "260", OPTION_VAL_260_PT_BR },
         { "270", OPTION_VAL_270_PT_BR },
         { "280", OPTION_VAL_280_PT_BR },
         { "290", OPTION_VAL_290_PT_BR },
         { "300", OPTION_VAL_300_PT_BR },
         { "310", OPTION_VAL_310_PT_BR },
         { "320", OPTION_VAL_320_PT_BR },
         { "330", OPTION_VAL_330_PT_BR },
         { "340", OPTION_VAL_340_PT_BR },
         { "350", OPTION_VAL_350_PT_BR },
         { "360", OPTION_VAL_360_PT_BR },
         { "370", OPTION_VAL_370_PT_BR },
         { "380", OPTION_VAL_380_PT_BR },
         { "390", OPTION_VAL_390_PT_BR },
         { "400", OPTION_VAL_400_PT_BR },
         { "410", OPTION_VAL_410_PT_BR },
         { "420", OPTION_VAL_420_PT_BR },
         { "430", OPTION_VAL_430_PT_BR },
         { "440", OPTION_VAL_440_PT_BR },
         { "450", OPTION_VAL_450_PT_BR },
         { "460", OPTION_VAL_460_PT_BR },
         { "470", OPTION_VAL_470_PT_BR },
         { "480", OPTION_VAL_480_PT_BR },
         { "490", OPTION_VAL_490_PT_BR },
         { "500", OPTION_VAL_500_PT_BR },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_PT_BR },
         { "5%",  OPTION_VAL_5_PT_BR },
         { "10%", OPTION_VAL_10_PT_BR },
         { "15%", OPTION_VAL_15_PT_BR },
         { "20%", OPTION_VAL_20_PT_BR },
         { "25%", OPTION_VAL_25_PT_BR },
         { "30%", OPTION_VAL_30_PT_BR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_PT_BR },
         { "5%",  OPTION_VAL_5_PT_BR },
         { "10%", OPTION_VAL_10_PT_BR },
         { "15%", OPTION_VAL_15_PT_BR },
         { "20%", OPTION_VAL_20_PT_BR },
         { "25%", OPTION_VAL_25_PT_BR },
         { "30%", OPTION_VAL_30_PT_BR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_PT_BR },
         { "60%",  OPTION_VAL_60_PT_BR },
         { "70%",  OPTION_VAL_70_PT_BR },
         { "80%",  OPTION_VAL_80_PT_BR },
         { "90%",  OPTION_VAL_90_PT_BR },
         { "100%", OPTION_VAL_100_O47_PT_BR },
         { "110%", OPTION_VAL_110_O47_PT_BR },
         { "120%", OPTION_VAL_120_O47_PT_BR },
         { "130%", OPTION_VAL_130_O47_PT_BR },
         { "140%", OPTION_VAL_140_O47_PT_BR },
         { "150%", OPTION_VAL_150_O47_PT_BR },
         { "160%", OPTION_VAL_160_O47_PT_BR },
         { "170%", OPTION_VAL_170_O47_PT_BR },
         { "180%", OPTION_VAL_180_O47_PT_BR },
         { "190%", OPTION_VAL_190_O47_PT_BR },
         { "200%", OPTION_VAL_200_O47_PT_BR },
         { "210%", OPTION_VAL_210_O47_PT_BR },
         { "220%", OPTION_VAL_220_O47_PT_BR },
         { "230%", OPTION_VAL_230_O47_PT_BR },
         { "240%", OPTION_VAL_240_O47_PT_BR },
         { "250%", OPTION_VAL_250_O47_PT_BR },
         { "260%", OPTION_VAL_260_O47_PT_BR },
         { "270%", OPTION_VAL_270_O47_PT_BR },
         { "280%", OPTION_VAL_280_O47_PT_BR },
         { "290%", OPTION_VAL_290_O47_PT_BR },
         { "300%", OPTION_VAL_300_O47_PT_BR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_BR },
         { "Red",      OPTION_VAL_RED_PT_BR },
         { "Green",    OPTION_VAL_GREEN_PT_BR },
         { "Blue",     OPTION_VAL_BLUE_PT_BR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_BR },
         { "Red",      OPTION_VAL_RED_PT_BR },
         { "Green",    OPTION_VAL_GREEN_PT_BR },
         { "Blue",     OPTION_VAL_BLUE_PT_BR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_BR },
         { "Red",      OPTION_VAL_RED_PT_BR },
         { "Green",    OPTION_VAL_GREEN_PT_BR },
         { "Blue",     OPTION_VAL_BLUE_PT_BR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_BR },
         { "Red",      OPTION_VAL_RED_PT_BR },
         { "Green",    OPTION_VAL_GREEN_PT_BR },
         { "Blue",     OPTION_VAL_BLUE_PT_BR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_PT_BR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_BR },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_BR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_PT_BR },
         { "None",     OPTION_VAL_NONE_PT_BR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_PT_BR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_BR },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_BR },
         { "None",     OPTION_VAL_NONE_PT_BR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_PT_BR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_BR },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_BR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_PT_BR },
         { "None",     OPTION_VAL_NONE_PT_BR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_PT_BR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_BR },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_BR },
         { "None",     OPTION_VAL_NONE_PT_BR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_PT_BR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_BR },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_BR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_PT_BR },
         { "None",     OPTION_VAL_NONE_PT_BR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_PT_BR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_BR },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_BR },
         { "None",     OPTION_VAL_NONE_PT_BR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_PT_BR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_BR },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_BR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_PT_BR },
         { "None",     OPTION_VAL_NONE_PT_BR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_PT_BR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_BR },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_BR },
         { "None",     OPTION_VAL_NONE_PT_BR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_PT_BR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_PT_BR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_PT_BR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_PT_BR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_PT_BR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_PT_BR,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_PT_BR,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_PT_BR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_PT_BR,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_PT_BR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_PT_BR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_PT_BR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_PT_BR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_BR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_BR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_BR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_BR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_BR },
         { "2x", OPTION_VAL_2_O26_PT_BR },
         { "3x", OPTION_VAL_3X_PT_BR },
         { "4x", OPTION_VAL_4_PT_BR },
         { "5x", OPTION_VAL_5X_PT_BR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_BR },
         { "20%",  OPTION_VAL_20_PT_BR },
         { "30%",  OPTION_VAL_30_PT_BR },
         { "40%",  OPTION_VAL_40_PT_BR },
         { "50%",  OPTION_VAL_50_PT_BR },
         { "60%",  OPTION_VAL_60_PT_BR },
         { "70%",  OPTION_VAL_70_PT_BR },
         { "80%",  OPTION_VAL_80_PT_BR },
         { "90%",  OPTION_VAL_90_PT_BR },
         { "100%", OPTION_VAL_100_O47_PT_BR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_BR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_BR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_BR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_BR },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_BR },
         { "2x", OPTION_VAL_2_O26_PT_BR },
         { "3x", OPTION_VAL_3X_PT_BR },
         { "4x", OPTION_VAL_4_PT_BR },
         { "5x", OPTION_VAL_5X_PT_BR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_BR },
         { "20%",  OPTION_VAL_20_PT_BR },
         { "30%",  OPTION_VAL_30_PT_BR },
         { "40%",  OPTION_VAL_40_PT_BR },
         { "50%",  OPTION_VAL_50_PT_BR },
         { "60%",  OPTION_VAL_60_PT_BR },
         { "70%",  OPTION_VAL_70_PT_BR },
         { "80%",  OPTION_VAL_80_PT_BR },
         { "90%",  OPTION_VAL_90_PT_BR },
         { "100%", OPTION_VAL_100_O47_PT_BR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_BR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_BR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_BR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_BR },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_BR },
         { "2x", OPTION_VAL_2_O26_PT_BR },
         { "3x", OPTION_VAL_3X_PT_BR },
         { "4x", OPTION_VAL_4_PT_BR },
         { "5x", OPTION_VAL_5X_PT_BR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_BR },
         { "20%",  OPTION_VAL_20_PT_BR },
         { "30%",  OPTION_VAL_30_PT_BR },
         { "40%",  OPTION_VAL_40_PT_BR },
         { "50%",  OPTION_VAL_50_PT_BR },
         { "60%",  OPTION_VAL_60_PT_BR },
         { "70%",  OPTION_VAL_70_PT_BR },
         { "80%",  OPTION_VAL_80_PT_BR },
         { "90%",  OPTION_VAL_90_PT_BR },
         { "100%", OPTION_VAL_100_O47_PT_BR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_BR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_BR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_BR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_BR },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_BR },
         { "2x", OPTION_VAL_2_O26_PT_BR },
         { "3x", OPTION_VAL_3X_PT_BR },
         { "4x", OPTION_VAL_4_PT_BR },
         { "5x", OPTION_VAL_5X_PT_BR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_BR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_BR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_BR },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_BR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_BR },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_BR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_BR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_BR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_BR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_BR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_BR },
         { "RED 11",            OPTION_VAL_RED_PT_BR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_BR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_BR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_BR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_BR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_BR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_BR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_BR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_BR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_BR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_BR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_BR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_BR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_BR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_BR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_BR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_BR },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_PT_BR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_BR },
         { "20%",  OPTION_VAL_20_PT_BR },
         { "30%",  OPTION_VAL_30_PT_BR },
         { "40%",  OPTION_VAL_40_PT_BR },
         { "50%",  OPTION_VAL_50_PT_BR },
         { "60%",  OPTION_VAL_60_PT_BR },
         { "70%",  OPTION_VAL_70_PT_BR },
         { "80%",  OPTION_VAL_80_PT_BR },
         { "90%",  OPTION_VAL_90_PT_BR },
         { "100%", OPTION_VAL_100_O47_PT_BR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_pt_br = {
   option_cats_pt_br,
   option_defs_pt_br
};

/* RETRO_LANGUAGE_PT_PT */

#define CATEGORY_SYSTEM_LABEL_PT_PT NULL
#define CATEGORY_SYSTEM_INFO_0_PT_PT NULL
#define CATEGORY_VIDEO_LABEL_PT_PT "Vídeo"
#define CATEGORY_VIDEO_INFO_0_PT_PT NULL
#define CATEGORY_PERFORMANCE_LABEL_PT_PT NULL
#define CATEGORY_PERFORMANCE_INFO_0_PT_PT "Configurar a renderização multithread e as definições de salto de fotogramas."
#define CATEGORY_HACKS_LABEL_PT_PT NULL
#define CATEGORY_HACKS_INFO_0_PT_PT NULL
#define CATEGORY_INPUT_LABEL_PT_PT "Entrada"
#define CATEGORY_INPUT_INFO_0_PT_PT NULL
#define CATEGORY_EXPANSIONS_LABEL_PT_PT NULL
#define CATEGORY_EXPANSIONS_INFO_0_PT_PT NULL
#define CATEGORY_VMU_LABEL_PT_PT NULL
#define CATEGORY_VMU_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_REGION_LABEL_PT_PT NULL
#define OPTION_VAL_JAPAN_PT_PT NULL
#define OPTION_VAL_USA_PT_PT NULL
#define OPTION_VAL_EUROPE_PT_PT NULL
#define OPTION_VAL_DEFAULT_PT_PT NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_PT_PT "Idioma"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_PT_PT NULL
#define OPTION_VAL_JAPANESE_PT_PT "Japonês"
#define OPTION_VAL_ENGLISH_PT_PT "Inglês"
#define OPTION_VAL_GERMAN_PT_PT "Alemão"
#define OPTION_VAL_FRENCH_PT_PT "Francês"
#define OPTION_VAL_SPANISH_PT_PT "Espanhol"
#define OPTION_VAL_ITALIAN_PT_PT "Italiano"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_UPNP_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DCNET_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_PT_PT NULL
#define OPTION_VAL_320X240_PT_PT NULL
#define OPTION_VAL_640X480_PT_PT NULL
#define OPTION_VAL_800X600_PT_PT NULL
#define OPTION_VAL_960X720_PT_PT NULL
#define OPTION_VAL_1024X768_PT_PT NULL
#define OPTION_VAL_1280X960_PT_PT NULL
#define OPTION_VAL_1440X1080_PT_PT NULL
#define OPTION_VAL_1600X1200_PT_PT NULL
#define OPTION_VAL_1920X1440_PT_PT NULL
#define OPTION_VAL_2560X1920_PT_PT NULL
#define OPTION_VAL_2880X2160_PT_PT NULL
#define OPTION_VAL_3200X2400_PT_PT NULL
#define OPTION_VAL_3840X2880_PT_PT NULL
#define OPTION_VAL_4480X3360_PT_PT NULL
#define OPTION_VAL_5120X3840_PT_PT NULL
#define OPTION_VAL_5760X4320_PT_PT NULL
#define OPTION_VAL_6400X4800_PT_PT NULL
#define OPTION_VAL_7040X5280_PT_PT NULL
#define OPTION_VAL_7680X5760_PT_PT NULL
#define OPTION_VAL_8320X6240_PT_PT NULL
#define OPTION_VAL_8960X6720_PT_PT NULL
#define OPTION_VAL_9600X7200_PT_PT NULL
#define OPTION_VAL_10240X7680_PT_PT NULL
#define OPTION_VAL_10880X8160_PT_PT NULL
#define OPTION_VAL_11520X8640_PT_PT NULL
#define OPTION_VAL_12160X9120_PT_PT NULL
#define OPTION_VAL_12800X9600_PT_PT NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_PT_PT NULL
#define OPTION_VAL_VGA_PT_PT NULL
#define OPTION_VAL_TV_RGB_PT_PT NULL
#define OPTION_VAL_TV_COMPOSITE_PT_PT NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_PT_PT NULL
#define OPTION_VAL_NTSC_PT_PT NULL
#define OPTION_VAL_PAL_PT_PT NULL
#define OPTION_VAL_PAL_N_PT_PT NULL
#define OPTION_VAL_PAL_M_PT_PT NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_PT_PT "Orientação do ecrã"
#define OPTION_VAL_HORIZONTAL_PT_PT NULL
#define OPTION_VAL_VERTICAL_PT_PT NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_PT_PT NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_PT_PT NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_PT_PT NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_PT_PT NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_PT_PT NULL
#define OPTION_VAL_512MB_PT_PT NULL
#define OPTION_VAL_1GB_PT_PT NULL
#define OPTION_VAL_2GB_PT_PT NULL
#define OPTION_VAL_4GB_PT_PT NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_FOG_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_PT_PT NULL
#define OPTION_VAL_1_PT_PT NULL
#define OPTION_VAL_2_PT_PT NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_PT_PT "Detetar alterações na taxa de fotogramas"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_PT_PT "Notifica a interface quando a taxa de fotogramas interna muda (por exemplo, de 60 fps para 30 fps). Melhora o ritmo dos fotogramas em jogos que funcionam a 30 fps ou 20 fps bloqueados, mas deve ser desativado para jogos com taxas de fotogramas desbloqueadas (instáveis) (por exemplo, Ecco the Dolphin, Unreal Tournament). Nota: Não está disponível quando a opção \"Saltar fotogramas automaticamente\" está ativada."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_PT_PT NULL
#define OPTION_VAL_2_O26_PT_PT NULL
#define OPTION_VAL_4_PT_PT NULL
#define OPTION_VAL_6_PT_PT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_PT_PT "Renderização multithread"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_PT_PT "Executa a GPU e a CPU em threads diferentes. Altamente recomendado."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_PT_PT "Saltar fotogramas automaticamente"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_PT_PT "Saltar automaticamente os fotogramas quando o emulador está a funcionar lentamente. Nota: Esta definição só se aplica quando a opção \"Renderização multithread\" está ativada."
#define OPTION_VAL_SOME_PT_PT NULL
#define OPTION_VAL_MORE_PT_PT NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_PT_PT NULL
#define OPTION_VAL_100_PT_PT NULL
#define OPTION_VAL_110_PT_PT NULL
#define OPTION_VAL_120_PT_PT NULL
#define OPTION_VAL_130_PT_PT NULL
#define OPTION_VAL_140_PT_PT NULL
#define OPTION_VAL_150_PT_PT NULL
#define OPTION_VAL_160_PT_PT NULL
#define OPTION_VAL_170_PT_PT NULL
#define OPTION_VAL_180_PT_PT NULL
#define OPTION_VAL_190_PT_PT NULL
#define OPTION_VAL_200_PT_PT NULL
#define OPTION_VAL_210_PT_PT NULL
#define OPTION_VAL_220_PT_PT NULL
#define OPTION_VAL_230_PT_PT NULL
#define OPTION_VAL_240_PT_PT NULL
#define OPTION_VAL_250_PT_PT NULL
#define OPTION_VAL_260_PT_PT NULL
#define OPTION_VAL_270_PT_PT NULL
#define OPTION_VAL_280_PT_PT NULL
#define OPTION_VAL_290_PT_PT NULL
#define OPTION_VAL_300_PT_PT NULL
#define OPTION_VAL_310_PT_PT NULL
#define OPTION_VAL_320_PT_PT NULL
#define OPTION_VAL_330_PT_PT NULL
#define OPTION_VAL_340_PT_PT NULL
#define OPTION_VAL_350_PT_PT NULL
#define OPTION_VAL_360_PT_PT NULL
#define OPTION_VAL_370_PT_PT NULL
#define OPTION_VAL_380_PT_PT NULL
#define OPTION_VAL_390_PT_PT NULL
#define OPTION_VAL_400_PT_PT NULL
#define OPTION_VAL_410_PT_PT NULL
#define OPTION_VAL_420_PT_PT NULL
#define OPTION_VAL_430_PT_PT NULL
#define OPTION_VAL_440_PT_PT NULL
#define OPTION_VAL_450_PT_PT NULL
#define OPTION_VAL_460_PT_PT NULL
#define OPTION_VAL_470_PT_PT NULL
#define OPTION_VAL_480_PT_PT NULL
#define OPTION_VAL_490_PT_PT NULL
#define OPTION_VAL_500_PT_PT NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_PT_PT NULL
#define OPTION_VAL_0_PT_PT NULL
#define OPTION_VAL_5_PT_PT NULL
#define OPTION_VAL_10_PT_PT NULL
#define OPTION_VAL_15_PT_PT NULL
#define OPTION_VAL_20_PT_PT NULL
#define OPTION_VAL_25_PT_PT NULL
#define OPTION_VAL_30_PT_PT NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_PT_PT NULL
#define OPTION_VAL_50_PT_PT NULL
#define OPTION_VAL_60_PT_PT NULL
#define OPTION_VAL_70_PT_PT NULL
#define OPTION_VAL_80_PT_PT NULL
#define OPTION_VAL_90_PT_PT NULL
#define OPTION_VAL_100_O47_PT_PT NULL
#define OPTION_VAL_110_O47_PT_PT NULL
#define OPTION_VAL_120_O47_PT_PT NULL
#define OPTION_VAL_130_O47_PT_PT NULL
#define OPTION_VAL_140_O47_PT_PT NULL
#define OPTION_VAL_150_O47_PT_PT NULL
#define OPTION_VAL_160_O47_PT_PT NULL
#define OPTION_VAL_170_O47_PT_PT NULL
#define OPTION_VAL_180_O47_PT_PT NULL
#define OPTION_VAL_190_O47_PT_PT NULL
#define OPTION_VAL_200_O47_PT_PT NULL
#define OPTION_VAL_210_O47_PT_PT NULL
#define OPTION_VAL_220_O47_PT_PT NULL
#define OPTION_VAL_230_O47_PT_PT NULL
#define OPTION_VAL_240_O47_PT_PT NULL
#define OPTION_VAL_250_O47_PT_PT NULL
#define OPTION_VAL_260_O47_PT_PT NULL
#define OPTION_VAL_270_O47_PT_PT NULL
#define OPTION_VAL_280_O47_PT_PT NULL
#define OPTION_VAL_290_O47_PT_PT NULL
#define OPTION_VAL_300_O47_PT_PT NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_PT_PT NULL
#define OPTION_VAL_WHITE_PT_PT NULL
#define OPTION_VAL_RED_PT_PT "Vermelho"
#define OPTION_VAL_GREEN_PT_PT "Verde"
#define OPTION_VAL_BLUE_PT_PT "Azul"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_PT_PT NULL
#define OPTION_VAL_VMU_PT_PT NULL
#define OPTION_VAL_PURUPURU_PT_PT NULL
#define OPTION_VAL_DREAMPOTATO_PT_PT NULL
#define OPTION_VAL_NONE_PT_PT "Não partilhar"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_PT_PT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_PT_PT NULL
#define OPTION_VAL_VMU_A1_PT_PT NULL
#define OPTION_VAL_ALL_VMUS_PT_PT NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_PT_PT NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_PT_PT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_PT_PT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_PT_PT NULL
#define OPTION_VAL_UPPER_LEFT_PT_PT NULL
#define OPTION_VAL_UPPER_RIGHT_PT_PT NULL
#define OPTION_VAL_LOWER_LEFT_PT_PT NULL
#define OPTION_VAL_LOWER_RIGHT_PT_PT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_PT_PT NULL
#define OPTION_VAL_1X_PT_PT NULL
#define OPTION_VAL_3X_PT_PT NULL
#define OPTION_VAL_5X_PT_PT NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_PT_PT NULL
#define OPTION_VAL_DEFAULT_ON_00_PT_PT NULL
#define OPTION_VAL_DEFAULT_OFF_01_PT_PT NULL
#define OPTION_VAL_BLACK_02_PT_PT NULL
#define OPTION_VAL_LIGHT_BLUE_04_PT_PT NULL
#define OPTION_VAL_CYAN_06_PT_PT NULL
#define OPTION_VAL_CYAN_BLUE_07_PT_PT NULL
#define OPTION_VAL_LIGHT_GREEN_08_PT_PT NULL
#define OPTION_VAL_CYAN_GREEN_09_PT_PT NULL
#define OPTION_VAL_LIGHT_CYAN_10_PT_PT NULL
#define OPTION_VAL_PURPLE_12_PT_PT NULL
#define OPTION_VAL_LIGHT_PURPLE_13_PT_PT NULL
#define OPTION_VAL_YELLOW_14_PT_PT "Amarelo"
#define OPTION_VAL_GRAY_15_PT_PT NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_PT_PT NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_PT_PT NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_PT_PT NULL
#define OPTION_VAL_LIGHT_RED_2_20_PT_PT NULL
#define OPTION_VAL_MAGENTA_21_PT_PT NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT NULL
#define OPTION_VAL_LIGHT_ORANGE_23_PT_PT NULL
#define OPTION_VAL_ORANGE_24_PT_PT NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT NULL
#define OPTION_VAL_LIGHT_YELLOW_26_PT_PT NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_PT_PT NULL
#define OPTION_VAL_40_PT_PT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_PT_PT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_PT_PT NULL

struct retro_core_option_v2_category option_cats_pt_pt[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_PT_PT,
      CATEGORY_SYSTEM_INFO_0_PT_PT
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_PT_PT,
      CATEGORY_VIDEO_INFO_0_PT_PT
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_PT_PT,
      CATEGORY_PERFORMANCE_INFO_0_PT_PT
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_PT_PT,
      CATEGORY_HACKS_INFO_0_PT_PT
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_PT_PT,
      CATEGORY_INPUT_INFO_0_PT_PT
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_PT_PT,
      CATEGORY_EXPANSIONS_INFO_0_PT_PT
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_PT_PT,
      CATEGORY_VMU_INFO_0_PT_PT
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_pt_pt[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_PT_PT },
         { "USA",     OPTION_VAL_USA_PT_PT },
         { "Europe",  OPTION_VAL_EUROPE_PT_PT },
         { "Default", OPTION_VAL_DEFAULT_PT_PT },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_PT_PT },
         { "English",  OPTION_VAL_ENGLISH_PT_PT },
         { "German",   OPTION_VAL_GERMAN_PT_PT },
         { "French",   OPTION_VAL_FRENCH_PT_PT },
         { "Spanish",  OPTION_VAL_SPANISH_PT_PT },
         { "Italian",  OPTION_VAL_ITALIAN_PT_PT },
         { "Default",  OPTION_VAL_DEFAULT_PT_PT },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_PT_PT },
         { "640x480",    OPTION_VAL_640X480_PT_PT },
         { "800x600",    OPTION_VAL_800X600_PT_PT },
         { "960x720",    OPTION_VAL_960X720_PT_PT },
         { "1024x768",   OPTION_VAL_1024X768_PT_PT },
         { "1280x960",   OPTION_VAL_1280X960_PT_PT },
         { "1440x1080",  OPTION_VAL_1440X1080_PT_PT },
         { "1600x1200",  OPTION_VAL_1600X1200_PT_PT },
         { "1920x1440",  OPTION_VAL_1920X1440_PT_PT },
         { "2560x1920",  OPTION_VAL_2560X1920_PT_PT },
         { "2880x2160",  OPTION_VAL_2880X2160_PT_PT },
         { "3200x2400",  OPTION_VAL_3200X2400_PT_PT },
         { "3840x2880",  OPTION_VAL_3840X2880_PT_PT },
         { "4480x3360",  OPTION_VAL_4480X3360_PT_PT },
         { "5120x3840",  OPTION_VAL_5120X3840_PT_PT },
         { "5760x4320",  OPTION_VAL_5760X4320_PT_PT },
         { "6400x4800",  OPTION_VAL_6400X4800_PT_PT },
         { "7040x5280",  OPTION_VAL_7040X5280_PT_PT },
         { "7680x5760",  OPTION_VAL_7680X5760_PT_PT },
         { "8320x6240",  OPTION_VAL_8320X6240_PT_PT },
         { "8960x6720",  OPTION_VAL_8960X6720_PT_PT },
         { "9600x7200",  OPTION_VAL_9600X7200_PT_PT },
         { "10240x7680", OPTION_VAL_10240X7680_PT_PT },
         { "10880x8160", OPTION_VAL_10880X8160_PT_PT },
         { "11520x8640", OPTION_VAL_11520X8640_PT_PT },
         { "12160x9120", OPTION_VAL_12160X9120_PT_PT },
         { "12800x9600", OPTION_VAL_12800X9600_PT_PT },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_PT_PT },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_PT_PT },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_PT_PT },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_PT_PT },
         { "PAL",     OPTION_VAL_PAL_PT_PT },
         { "PAL_N",   OPTION_VAL_PAL_N_PT_PT },
         { "PAL_M",   OPTION_VAL_PAL_M_PT_PT },
         { "Default", OPTION_VAL_DEFAULT_PT_PT },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_PT_PT },
         { "vertical",   OPTION_VAL_VERTICAL_PT_PT },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_PT_PT },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_PT_PT },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_PT_PT },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_PT_PT },
         { "1GB",   OPTION_VAL_1GB_PT_PT },
         { "2GB",   OPTION_VAL_2GB_PT_PT },
         { "4GB",   OPTION_VAL_4GB_PT_PT },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_PT_PT },
         { "1",  OPTION_VAL_1_PT_PT },
         { "2",  OPTION_VAL_2_PT_PT },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_PT_PT },
         { "4",  OPTION_VAL_4_PT_PT },
         { "6",  OPTION_VAL_6_PT_PT },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_PT_PT,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_PT_PT,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_PT_PT,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_PT_PT,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_PT_PT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_PT_PT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_PT_PT },
         { "more", OPTION_VAL_MORE_PT_PT },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_PT_PT,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_PT_PT },
         { "110", OPTION_VAL_110_PT_PT },
         { "120", OPTION_VAL_120_PT_PT },
         { "130", OPTION_VAL_130_PT_PT },
         { "140", OPTION_VAL_140_PT_PT },
         { "150", OPTION_VAL_150_PT_PT },
         { "160", OPTION_VAL_160_PT_PT },
         { "170", OPTION_VAL_170_PT_PT },
         { "180", OPTION_VAL_180_PT_PT },
         { "190", OPTION_VAL_190_PT_PT },
         { "200", OPTION_VAL_200_PT_PT },
         { "210", OPTION_VAL_210_PT_PT },
         { "220", OPTION_VAL_220_PT_PT },
         { "230", OPTION_VAL_230_PT_PT },
         { "240", OPTION_VAL_240_PT_PT },
         { "250", OPTION_VAL_250_PT_PT },
         { "260", OPTION_VAL_260_PT_PT },
         { "270", OPTION_VAL_270_PT_PT },
         { "280", OPTION_VAL_280_PT_PT },
         { "290", OPTION_VAL_290_PT_PT },
         { "300", OPTION_VAL_300_PT_PT },
         { "310", OPTION_VAL_310_PT_PT },
         { "320", OPTION_VAL_320_PT_PT },
         { "330", OPTION_VAL_330_PT_PT },
         { "340", OPTION_VAL_340_PT_PT },
         { "350", OPTION_VAL_350_PT_PT },
         { "360", OPTION_VAL_360_PT_PT },
         { "370", OPTION_VAL_370_PT_PT },
         { "380", OPTION_VAL_380_PT_PT },
         { "390", OPTION_VAL_390_PT_PT },
         { "400", OPTION_VAL_400_PT_PT },
         { "410", OPTION_VAL_410_PT_PT },
         { "420", OPTION_VAL_420_PT_PT },
         { "430", OPTION_VAL_430_PT_PT },
         { "440", OPTION_VAL_440_PT_PT },
         { "450", OPTION_VAL_450_PT_PT },
         { "460", OPTION_VAL_460_PT_PT },
         { "470", OPTION_VAL_470_PT_PT },
         { "480", OPTION_VAL_480_PT_PT },
         { "490", OPTION_VAL_490_PT_PT },
         { "500", OPTION_VAL_500_PT_PT },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_PT_PT },
         { "5%",  OPTION_VAL_5_PT_PT },
         { "10%", OPTION_VAL_10_PT_PT },
         { "15%", OPTION_VAL_15_PT_PT },
         { "20%", OPTION_VAL_20_PT_PT },
         { "25%", OPTION_VAL_25_PT_PT },
         { "30%", OPTION_VAL_30_PT_PT },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_PT_PT },
         { "5%",  OPTION_VAL_5_PT_PT },
         { "10%", OPTION_VAL_10_PT_PT },
         { "15%", OPTION_VAL_15_PT_PT },
         { "20%", OPTION_VAL_20_PT_PT },
         { "25%", OPTION_VAL_25_PT_PT },
         { "30%", OPTION_VAL_30_PT_PT },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_PT_PT },
         { "60%",  OPTION_VAL_60_PT_PT },
         { "70%",  OPTION_VAL_70_PT_PT },
         { "80%",  OPTION_VAL_80_PT_PT },
         { "90%",  OPTION_VAL_90_PT_PT },
         { "100%", OPTION_VAL_100_O47_PT_PT },
         { "110%", OPTION_VAL_110_O47_PT_PT },
         { "120%", OPTION_VAL_120_O47_PT_PT },
         { "130%", OPTION_VAL_130_O47_PT_PT },
         { "140%", OPTION_VAL_140_O47_PT_PT },
         { "150%", OPTION_VAL_150_O47_PT_PT },
         { "160%", OPTION_VAL_160_O47_PT_PT },
         { "170%", OPTION_VAL_170_O47_PT_PT },
         { "180%", OPTION_VAL_180_O47_PT_PT },
         { "190%", OPTION_VAL_190_O47_PT_PT },
         { "200%", OPTION_VAL_200_O47_PT_PT },
         { "210%", OPTION_VAL_210_O47_PT_PT },
         { "220%", OPTION_VAL_220_O47_PT_PT },
         { "230%", OPTION_VAL_230_O47_PT_PT },
         { "240%", OPTION_VAL_240_O47_PT_PT },
         { "250%", OPTION_VAL_250_O47_PT_PT },
         { "260%", OPTION_VAL_260_O47_PT_PT },
         { "270%", OPTION_VAL_270_O47_PT_PT },
         { "280%", OPTION_VAL_280_O47_PT_PT },
         { "290%", OPTION_VAL_290_O47_PT_PT },
         { "300%", OPTION_VAL_300_O47_PT_PT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_PT },
         { "Red",      OPTION_VAL_RED_PT_PT },
         { "Green",    OPTION_VAL_GREEN_PT_PT },
         { "Blue",     OPTION_VAL_BLUE_PT_PT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_PT },
         { "Red",      OPTION_VAL_RED_PT_PT },
         { "Green",    OPTION_VAL_GREEN_PT_PT },
         { "Blue",     OPTION_VAL_BLUE_PT_PT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_PT },
         { "Red",      OPTION_VAL_RED_PT_PT },
         { "Green",    OPTION_VAL_GREEN_PT_PT },
         { "Blue",     OPTION_VAL_BLUE_PT_PT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_PT_PT },
         { "Red",      OPTION_VAL_RED_PT_PT },
         { "Green",    OPTION_VAL_GREEN_PT_PT },
         { "Blue",     OPTION_VAL_BLUE_PT_PT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_PT_PT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_PT },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_PT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_PT_PT },
         { "None",     OPTION_VAL_NONE_PT_PT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_PT_PT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_PT },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_PT },
         { "None",     OPTION_VAL_NONE_PT_PT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_PT_PT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_PT },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_PT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_PT_PT },
         { "None",     OPTION_VAL_NONE_PT_PT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_PT_PT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_PT },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_PT },
         { "None",     OPTION_VAL_NONE_PT_PT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_PT_PT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_PT },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_PT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_PT_PT },
         { "None",     OPTION_VAL_NONE_PT_PT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_PT_PT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_PT },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_PT },
         { "None",     OPTION_VAL_NONE_PT_PT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_PT_PT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_PT },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_PT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_PT_PT },
         { "None",     OPTION_VAL_NONE_PT_PT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_PT_PT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_PT_PT },
         { "Purupuru", OPTION_VAL_PURUPURU_PT_PT },
         { "None",     OPTION_VAL_NONE_PT_PT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_PT_PT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_PT_PT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_PT_PT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_PT_PT },
         { "All VMUs", OPTION_VAL_ALL_VMUS_PT_PT },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_PT_PT,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_PT_PT,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_PT_PT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_PT_PT,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_PT_PT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_PT_PT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_PT_PT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_PT_PT,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_PT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_PT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_PT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_PT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_PT },
         { "2x", OPTION_VAL_2_O26_PT_PT },
         { "3x", OPTION_VAL_3X_PT_PT },
         { "4x", OPTION_VAL_4_PT_PT },
         { "5x", OPTION_VAL_5X_PT_PT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_PT },
         { "20%",  OPTION_VAL_20_PT_PT },
         { "30%",  OPTION_VAL_30_PT_PT },
         { "40%",  OPTION_VAL_40_PT_PT },
         { "50%",  OPTION_VAL_50_PT_PT },
         { "60%",  OPTION_VAL_60_PT_PT },
         { "70%",  OPTION_VAL_70_PT_PT },
         { "80%",  OPTION_VAL_80_PT_PT },
         { "90%",  OPTION_VAL_90_PT_PT },
         { "100%", OPTION_VAL_100_O47_PT_PT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_PT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_PT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_PT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_PT },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_PT },
         { "2x", OPTION_VAL_2_O26_PT_PT },
         { "3x", OPTION_VAL_3X_PT_PT },
         { "4x", OPTION_VAL_4_PT_PT },
         { "5x", OPTION_VAL_5X_PT_PT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_PT },
         { "20%",  OPTION_VAL_20_PT_PT },
         { "30%",  OPTION_VAL_30_PT_PT },
         { "40%",  OPTION_VAL_40_PT_PT },
         { "50%",  OPTION_VAL_50_PT_PT },
         { "60%",  OPTION_VAL_60_PT_PT },
         { "70%",  OPTION_VAL_70_PT_PT },
         { "80%",  OPTION_VAL_80_PT_PT },
         { "90%",  OPTION_VAL_90_PT_PT },
         { "100%", OPTION_VAL_100_O47_PT_PT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_PT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_PT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_PT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_PT },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_PT },
         { "2x", OPTION_VAL_2_O26_PT_PT },
         { "3x", OPTION_VAL_3X_PT_PT },
         { "4x", OPTION_VAL_4_PT_PT },
         { "5x", OPTION_VAL_5X_PT_PT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_PT },
         { "20%",  OPTION_VAL_20_PT_PT },
         { "30%",  OPTION_VAL_30_PT_PT },
         { "40%",  OPTION_VAL_40_PT_PT },
         { "50%",  OPTION_VAL_50_PT_PT },
         { "60%",  OPTION_VAL_60_PT_PT },
         { "70%",  OPTION_VAL_70_PT_PT },
         { "80%",  OPTION_VAL_80_PT_PT },
         { "90%",  OPTION_VAL_90_PT_PT },
         { "100%", OPTION_VAL_100_O47_PT_PT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_PT_PT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_PT_PT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_PT_PT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_PT_PT },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_PT_PT },
         { "2x", OPTION_VAL_2_O26_PT_PT },
         { "3x", OPTION_VAL_3X_PT_PT },
         { "4x", OPTION_VAL_4_PT_PT },
         { "5x", OPTION_VAL_5X_PT_PT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_PT_PT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_PT_PT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_PT_PT },
         { "BLUE 03",           OPTION_VAL_BLUE_PT_PT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_PT_PT },
         { "GREEN 05",          OPTION_VAL_GREEN_PT_PT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_PT_PT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_PT_PT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_PT_PT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_PT_PT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_PT_PT },
         { "RED 11",            OPTION_VAL_RED_PT_PT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_PT_PT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_PT_PT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_PT_PT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_PT_PT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_PT_PT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_PT_PT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_PT_PT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_PT_PT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_PT_PT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_PT_PT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_PT_PT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_PT_PT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_PT_PT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_PT_PT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_PT_PT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_PT_PT },
         { "WHITE 28",          OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_PT_PT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_PT_PT },
         { "20%",  OPTION_VAL_20_PT_PT },
         { "30%",  OPTION_VAL_30_PT_PT },
         { "40%",  OPTION_VAL_40_PT_PT },
         { "50%",  OPTION_VAL_50_PT_PT },
         { "60%",  OPTION_VAL_60_PT_PT },
         { "70%",  OPTION_VAL_70_PT_PT },
         { "80%",  OPTION_VAL_80_PT_PT },
         { "90%",  OPTION_VAL_90_PT_PT },
         { "100%", OPTION_VAL_100_O47_PT_PT },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_pt_pt = {
   option_cats_pt_pt,
   option_defs_pt_pt
};

/* RETRO_LANGUAGE_RU */

#define CATEGORY_SYSTEM_LABEL_RU "Система"
#define CATEGORY_SYSTEM_INFO_0_RU "Настройка региона, языка, BIOS и основных параметров оборудования."
#define CATEGORY_VIDEO_LABEL_RU "Видео"
#define CATEGORY_VIDEO_INFO_0_RU "Настройки разрешения, порядко-независимой прозрачности и визуальных эффектов."
#define CATEGORY_PERFORMANCE_LABEL_RU "Производительность"
#define CATEGORY_PERFORMANCE_INFO_0_RU "Настройки многопоточного рендеринга и пропуска кадров."
#define CATEGORY_HACKS_LABEL_RU "Хаки эмуляции"
#define CATEGORY_HACKS_INFO_0_RU "Настройки широкоформатных чит-кодов, скорости загрузки GD-ROM и замены текстур."
#define CATEGORY_INPUT_LABEL_RU "Ввод"
#define CATEGORY_INPUT_INFO_0_RU "Настройки геймпада и светового пистолета."
#define CATEGORY_EXPANSIONS_LABEL_RU "Модули расширения контроллеров"
#define CATEGORY_EXPANSIONS_INFO_0_RU "Выбор устройства (VMU, вибропак), подключенного к слотам расширения контроллеров."
#define CATEGORY_VMU_LABEL_RU "Визуальные карты памяти"
#define CATEGORY_VMU_INFO_0_RU "Настройки файлов сохранений VMU для каждой игры и отображения VMU на экране."
#define CORE_OPTION_NAME_REGION_LABEL_RU "Регион"
#define OPTION_VAL_JAPAN_RU "Япония"
#define OPTION_VAL_USA_RU "США"
#define OPTION_VAL_EUROPE_RU "Европа"
#define OPTION_VAL_DEFAULT_RU "По умолчанию"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_RU "Язык"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_RU "Выбор языка, используемого в BIOS, а также в любых играх с поддержкой нескольких языков."
#define OPTION_VAL_JAPANESE_RU "Японский"
#define OPTION_VAL_ENGLISH_RU "Английский"
#define OPTION_VAL_GERMAN_RU "Немецкий"
#define OPTION_VAL_FRENCH_RU "Французский"
#define OPTION_VAL_SPANISH_RU "Испанский"
#define OPTION_VAL_ITALIAN_RU "Итальянский"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_RU "Эмуляция BIOS (требуется перезапуск)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_RU "Принудительно использовать высокоуровневую эмуляцию BIOS."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_RU "Включить DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_RU "Включить эмуляцию звукового DSP (цифрового сигнального процессора) Dreamcast. Улучшает точность генерируемого звука, но повышает системные требования."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_RU "Разрешать сервисные кнопки аркадных игр"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_RU "Включает кнопку SERVICE для входа в настройки автоматов аркадных игр."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_RU "Режим Free Play для игр NAOMI"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_RU "Изменять настройки монетизации для включения бесплатной игры."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_RU "Эмуляция Broadband Adapter"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_RU "Эмулировать вместо модема широкополосный Ethernet-адаптер (требуется перезапуск)"
#define CORE_OPTION_NAME_UPNP_LABEL_RU "Включить UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_RU "Использовать UPnP для автоматической настройки вашего Интернет-роутера для онлайн-игр."
#define CORE_OPTION_NAME_DCNET_LABEL_RU "Использовать DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_RU "Используйте облачный сервис DCNet для доступа в Интернет Dreamcast."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_RU "Внутреннее разрешение"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_RU "Настройка разрешения рендеринга."
#define OPTION_VAL_320X240_RU "320x240 (половина)"
#define OPTION_VAL_640X480_RU "640x480 (исходное)"
#define OPTION_VAL_800X600_RU NULL
#define OPTION_VAL_960X720_RU NULL
#define OPTION_VAL_1024X768_RU NULL
#define OPTION_VAL_1280X960_RU NULL
#define OPTION_VAL_1440X1080_RU NULL
#define OPTION_VAL_1600X1200_RU NULL
#define OPTION_VAL_1920X1440_RU NULL
#define OPTION_VAL_2560X1920_RU NULL
#define OPTION_VAL_2880X2160_RU NULL
#define OPTION_VAL_3200X2400_RU NULL
#define OPTION_VAL_3840X2880_RU NULL
#define OPTION_VAL_4480X3360_RU NULL
#define OPTION_VAL_5120X3840_RU NULL
#define OPTION_VAL_5760X4320_RU NULL
#define OPTION_VAL_6400X4800_RU NULL
#define OPTION_VAL_7040X5280_RU NULL
#define OPTION_VAL_7680X5760_RU NULL
#define OPTION_VAL_8320X6240_RU NULL
#define OPTION_VAL_8960X6720_RU NULL
#define OPTION_VAL_9600X7200_RU NULL
#define OPTION_VAL_10240X7680_RU NULL
#define OPTION_VAL_10880X8160_RU NULL
#define OPTION_VAL_11520X8640_RU NULL
#define OPTION_VAL_12160X9120_RU NULL
#define OPTION_VAL_12800X9600_RU NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_RU "Тип кабеля"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_RU "Тип выходного сигнала. 'ТВ (композит)' является наиболее совместимым."
#define OPTION_VAL_VGA_RU NULL
#define OPTION_VAL_TV_RGB_RU "ТВ (RGB)"
#define OPTION_VAL_TV_COMPOSITE_RU "ТВ (композит)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_RU "Стандарт вещания"
#define OPTION_VAL_NTSC_RU NULL
#define OPTION_VAL_PAL_RU "PAL (Мир)"
#define OPTION_VAL_PAL_N_RU "PAL-N (Аргентина, Парагвай, Уругвай)"
#define OPTION_VAL_PAL_M_RU "PAL-M (Бразилия)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_RU "Ориентация экрана"
#define OPTION_VAL_HORIZONTAL_RU "Горизонтальная"
#define OPTION_VAL_VERTICAL_RU "Вертикальная"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_RU "Альфа-сортировка"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_RU "Выбор способа сортировки прозрачных полигонов."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_RU "По слоям (самая быстрая, наименее точная)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_RU "По треугольникам (стандартная)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_RU "По пикселям (точная, самая медленная)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_RU "Размер накопительного буфера пикселей"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_RU "Большие значения могут требоваться для правильного отображения высоких разрешений."
#define OPTION_VAL_512MB_RU "512 МБ"
#define OPTION_VAL_1GB_RU "1 ГБ"
#define OPTION_VAL_2GB_RU "2 ГБ"
#define OPTION_VAL_4GB_RU "4 ГБ"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_RU "Максимум прозрачных слоёв"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_RU "Для сложных сцен могут потребоваться более высокие значения."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_RU "Полная эмуляция фреймбуфера"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_RU "Включает полную эмуляцию кадрового буфера в VRAM. Полезно для игр, которые напрямую производят чтение или запись фреймбуфера в VRAM. При включении внутреннее разрешение будет принудительно установлено в 640x480 и возможно сильное падение скорости."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_RU "Включить буфер рендера в текстуру"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_RU "Копирует отрисованные текстуры из GPU обратно в VRAM. Данная настройка обычно включена для игр, которые этого требуют. При включении не используется масштабирование рендеринга текстур и возможно падение производительности."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_RU "MIP-текстурирование"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_RU "Может повысить производительность и снизить мерцание при отображении на расстоянии уменьшенных копий текстур."
#define CORE_OPTION_NAME_FOG_LABEL_RU "Эффекты тумана"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_RU "Модификатор объема"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_RU "Функция графического процессора Dreamcast, используемая в играх для отрисовки теней объектов. По умолчанию должно быть включено - влияние на производительность обычно минимально или незначительно."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_RU "Анизотропная фильтрация"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_RU "Улучшает качество текстур на поверхностях под непрямыми углами обзора по отношению к камере. Более высокие значения повышают требования к GPU. Изменения настройки вступают в силу при перезапуске."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_RU "Сглаживание текстур"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_RU "Выбор метода сглаживания текстур. Используется для принудительной фильтрации всех текстур, чтобы получить более чёткое (или сглаженное) изображение. Любое значение, отличное от стандартного может приводить к ошибкам рендеринга. Для применения настройки требуется перезапуск."
#define OPTION_VAL_1_RU "Принудительно методом ближ. соседа"
#define OPTION_VAL_2_RU "Принудительно линейное"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_RU "Задержка замены кадров"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_RU "Помогает избежать мерцания экрана или искажения заставок. Не рекомендуется на медленных платформах."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_RU "Определять изменение частоты кадров"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_RU "Уведомлять фронтенд при изменениях внутренней частоты кадров (напр., с 60 до 30 к/с). Улучшает смену кадров в играх с постоянной частотой в 30 или 20 к/с. В играх с динамической частотой кадров (напр., Ecco the Dolphin, Unreal Tournament) должно быть выключено. Недоступно при включении опции 'Автопропуск кадров'."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_RU "Фильтр постобработки PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_RU "Включает постобработку изображения для имитации эффектов графического процессора PowerVR2 и аналоговых видеосигналов."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_RU "Апскейлинг текстур (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_RU "Улучшает 2D пиксель-арт с ручной отрисовкой. Рекомендуется только для 2D-игр с пиксельной графикой."
#define OPTION_VAL_2_O26_RU NULL
#define OPTION_VAL_4_RU NULL
#define OPTION_VAL_6_RU NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_RU "Макс. размер сглаженных текстур при апскейлинге"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_RU "Установка максимального размера текстур при масштабировании. Для текстур с размером больше заданного значения масштабирование не применяется."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_RU "Нативная глубина интерполяции"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_RU "Помогает при проблемах с глубиной и искажением текстур на видеокартах AMD. В ряде случаев также устраняет проблемы на видеокартах Intel."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_RU "Исправление просачивания при апскейлинге"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_RU "Помогает в случае просачивания текстур при масштабировании. Отключение помогает исправить искажения пикселей при апскейлинге 2D-игр (MVC2, CVS, KOF и др.)."
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_RU "Рендеринг в отдельном потоке"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_RU "Запускает графический и центральный процессоры в разных потоках. Настоятельно рекомендуется."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_RU "Автопропуск кадров"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_RU "Автоматически пропускать кадры при замедлении эмулятора. Применяется только с вкл. опцией 'Рендеринг в отдельном потоке'."
#define OPTION_VAL_SOME_RU "Нормально"
#define OPTION_VAL_MORE_RU "Максимум"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_RU "Пропуск кадров"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_RU "Устанавливает количество кадров пропуска между отображаемыми кадрами."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_RU "Широкоэкранные чит-коды (требуется перезапуск)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_RU "Включает чит-коды для отображения некоторых игр в широкоэкранном формате."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_RU "Широкоэкранный хак"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_RU "Отрисовывать геометрию за пределами стандартного соотношения 4:3. Может вызвать искажения графики в расширенной области."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_RU "Быстрая загрузка GD-ROM (неточно)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_RU "Ускоряет загрузку GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_RU "Мод Dreamcast на 32 Мбайт ОЗУ"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_RU "Вкл. мод для Dreamcast на 32 мегабайта ОЗУ. Может влиять на совместимость."
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_RU "Разгон/замедление SH4"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_RU "Изменение стандартной частоты процессора SH4 в 200 МГц. Снижение частоты помогает для слабых платформ. Разгон может повышать частоту кадров в некоторых играх. Используйте с осторожностью."
#define OPTION_VAL_100_RU "100 МГц"
#define OPTION_VAL_110_RU "110 МГц"
#define OPTION_VAL_120_RU "120 МГц"
#define OPTION_VAL_130_RU "130 МГц"
#define OPTION_VAL_140_RU "140 МГц"
#define OPTION_VAL_150_RU "150 МГц"
#define OPTION_VAL_160_RU "160 МГц"
#define OPTION_VAL_170_RU "170 МГц"
#define OPTION_VAL_180_RU "180 МГц"
#define OPTION_VAL_190_RU "190 МГц"
#define OPTION_VAL_200_RU "200 МГц"
#define OPTION_VAL_210_RU "210 МГц"
#define OPTION_VAL_220_RU "220 МГц"
#define OPTION_VAL_230_RU "230 МГц"
#define OPTION_VAL_240_RU "240 МГц"
#define OPTION_VAL_250_RU "250 МГц"
#define OPTION_VAL_260_RU "260 МГц"
#define OPTION_VAL_270_RU "270 МГц"
#define OPTION_VAL_280_RU "280 МГц"
#define OPTION_VAL_290_RU "290 МГц"
#define OPTION_VAL_300_RU "300 МГц"
#define OPTION_VAL_310_RU "310 МГц"
#define OPTION_VAL_320_RU "320 МГц"
#define OPTION_VAL_330_RU "330 МГц"
#define OPTION_VAL_340_RU "340 МГц"
#define OPTION_VAL_350_RU "350 МГц"
#define OPTION_VAL_360_RU "360 МГц"
#define OPTION_VAL_370_RU "370 МГц"
#define OPTION_VAL_380_RU "380 МГц"
#define OPTION_VAL_390_RU "390 МГц"
#define OPTION_VAL_400_RU "400 МГц"
#define OPTION_VAL_410_RU "410 МГц"
#define OPTION_VAL_420_RU "420 МГц"
#define OPTION_VAL_430_RU "430 МГц"
#define OPTION_VAL_440_RU "440 МГц"
#define OPTION_VAL_450_RU "450 МГц"
#define OPTION_VAL_460_RU "460 МГц"
#define OPTION_VAL_470_RU "470 МГц"
#define OPTION_VAL_480_RU "480 МГц"
#define OPTION_VAL_490_RU "490 МГц"
#define OPTION_VAL_500_RU "500 МГц"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_RU "Загружать внешние текстуры"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_RU "Загружать пользовательские текстуры из папки 'system/dc/textures/<game-id>/'."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_RU "Предзагрузка кастомных текстур"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_RU "Использовать предзагрузку кастомных текстур при запуске игры. Может улучшить производительность, но увеличивает расход памяти."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_RU "Дамп текстур"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_RU "При каждом использовании игрой новой текстуры она будет сохраняться в файл .png в папке 'system/dc/texdump/<game-id>/'."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_RU "Выгружать заменяемые текстуры"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_RU "Всегда создавать дамп текстур, для которых загружены кастомные текстуры."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_RU "Мёртвая зона аналогового стика"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_RU "Настройка величины отклонения аналогового джойстика для срабатывания."
#define OPTION_VAL_0_RU NULL
#define OPTION_VAL_5_RU NULL
#define OPTION_VAL_10_RU NULL
#define OPTION_VAL_15_RU NULL
#define OPTION_VAL_20_RU NULL
#define OPTION_VAL_25_RU NULL
#define OPTION_VAL_30_RU NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_RU "Мёртвая зона триггеров"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_RU "Настройка степени нажатия триггеров для срабатывания."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_RU "Цифровые триггеры"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_RU "При включении триггеры будут обрабатываться как стандартные кнопки, полностью нажатые либо не нажатые, без промежуточных положений."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_RU "Транслировать цифровые сигналы"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_RU "Транслировать события цифрового ввода и состояние вибрации в TCP-порт 8000. Совместимо с параметром MAME \"-output network\"."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_RU "Показывать настройки свет. пистолета"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_RU "Включает конфигурацию параметров прицела светового пистолета. Для применения настройки требуется переключить быстрое меню."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_RU "Масштабирование прицела"
#define OPTION_VAL_50_RU NULL
#define OPTION_VAL_60_RU NULL
#define OPTION_VAL_70_RU NULL
#define OPTION_VAL_80_RU NULL
#define OPTION_VAL_90_RU NULL
#define OPTION_VAL_100_O47_RU NULL
#define OPTION_VAL_110_O47_RU NULL
#define OPTION_VAL_120_O47_RU NULL
#define OPTION_VAL_130_O47_RU NULL
#define OPTION_VAL_140_O47_RU NULL
#define OPTION_VAL_150_O47_RU NULL
#define OPTION_VAL_160_O47_RU NULL
#define OPTION_VAL_170_O47_RU NULL
#define OPTION_VAL_180_O47_RU NULL
#define OPTION_VAL_190_O47_RU NULL
#define OPTION_VAL_200_O47_RU NULL
#define OPTION_VAL_210_O47_RU NULL
#define OPTION_VAL_220_O47_RU NULL
#define OPTION_VAL_230_O47_RU NULL
#define OPTION_VAL_240_O47_RU NULL
#define OPTION_VAL_250_O47_RU NULL
#define OPTION_VAL_260_O47_RU NULL
#define OPTION_VAL_270_O47_RU NULL
#define OPTION_VAL_280_O47_RU NULL
#define OPTION_VAL_290_O47_RU NULL
#define OPTION_VAL_300_O47_RU NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_RU "Отображение прицела пистолета 1"
#define OPTION_VAL_WHITE_RU "Белый"
#define OPTION_VAL_RED_RU "Красный"
#define OPTION_VAL_GREEN_RU "Зелёный"
#define OPTION_VAL_BLUE_RU "Голубой"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_RU "Отображение прицела пистолета 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_RU "Отображение прицела пистолета 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_RU "Отображение прицела пистолета 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_RU "Устройство в слоте расширения A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_RU "Выбор устройства, включенного в слот расширения A1 (порт А слот 1)."
#define OPTION_VAL_VMU_RU NULL
#define OPTION_VAL_PURUPURU_RU "Вибропак"
#define OPTION_VAL_DREAMPOTATO_RU NULL
#define OPTION_VAL_NONE_RU "Нет"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_RU "Устройство в слоте расширения A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_RU "Выбор устройства, включенного в слот расширения A2 (порт А слот 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_RU "Устройство в слоте расширения B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_RU "Выбор устройства, включенного в слот расширения B1 (порт B слот 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_RU "Устройство в слоте расширения B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_RU "Выбор устройства, включенного в слот расширения B2 (порт B слот 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_RU "Устройство в слоте расширения C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_RU "Выбор устройства, включенного в слот расширения C1 (порт C слот 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_RU "Устройство в слоте расширения C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_RU "Выбор устройства, включенного в слот расширения C2 (порт C слот 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_RU "Устройство в слоте расширения D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_RU "Выбор устройства, включенного в слот расширения D1 (порт D слот 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_RU "Устройство в слоте расширения D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_RU "Выбор устройства, включенного в слот расширения D2 (порт D слот 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_RU "Отдельные визуальные карты памяти (VMU) для игр"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_RU "VMU для каждой игры"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_RU "Если выкл., все игры используют до 8 файлов сохранений VMU (A1/A2/B1/B2/C1/C2/D1/D2), расположенных в системном каталоге RetroArch.\nОпция 'VMU A1' создаёт в каталоге сохранений RetroArch уникальный файл VMU 'A1' для каждой запускаемой игры.\nОпция 'Все VMU' создаёт до 8 уникальных файлов VMU (A1/A2/B1/B2/C1/C2/D1/D2) для каждой запускаемой игры."
#define OPTION_VAL_VMU_A1_RU NULL
#define OPTION_VAL_ALL_VMUS_RU "Все VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_RU "Звуковые сигналы визуальных карт памяти (VMU)"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_RU "Сигналы VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_RU "Включает воспроизведение сигналов VMU."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_RU "Хранилище VMU DreamPotato"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_RU "Хранить данные VMU на подключенном эмуляторе DreamPotato. При включении опции локальный файл сохранения VMU не используется."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_RU "Показывать настройки отображения визуальных карт памяти (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_RU "Показывать настройки отображения VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_RU "Включает конфигурацию размера, положения, цвета и видимости эмулируемого ЖК-экрана VMU. Для применения настройки требуется переключить быстрое меню."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_RU "Показывать экран VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_RU "Положение экрана VMU 1"
#define OPTION_VAL_UPPER_LEFT_RU "Вверху слева"
#define OPTION_VAL_UPPER_RIGHT_RU "Вверху справа"
#define OPTION_VAL_LOWER_LEFT_RU "Внизу слева"
#define OPTION_VAL_LOWER_RIGHT_RU "Внизу справа"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_RU "Размер экрана VMU 1"
#define OPTION_VAL_1X_RU NULL
#define OPTION_VAL_3X_RU NULL
#define OPTION_VAL_5X_RU NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_RU "Цвет вкл. пикселей VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_RU "По умолчанию ВКЛ"
#define OPTION_VAL_DEFAULT_OFF_01_RU "По умолчанию ВЫКЛ"
#define OPTION_VAL_BLACK_02_RU "Чёрный"
#define OPTION_VAL_LIGHT_BLUE_04_RU "Светло-синий"
#define OPTION_VAL_CYAN_06_RU "Сине-зелёный"
#define OPTION_VAL_CYAN_BLUE_07_RU "Голубовато-синий"
#define OPTION_VAL_LIGHT_GREEN_08_RU "Светло-зелёный"
#define OPTION_VAL_CYAN_GREEN_09_RU "Голубовато-зелёный"
#define OPTION_VAL_LIGHT_CYAN_10_RU "Светло-голубой"
#define OPTION_VAL_PURPLE_12_RU "Фиолетовый"
#define OPTION_VAL_LIGHT_PURPLE_13_RU "Светло-фиолетовый"
#define OPTION_VAL_YELLOW_14_RU "Жёлтая"
#define OPTION_VAL_GRAY_15_RU "Серый"
#define OPTION_VAL_LIGHT_PURPLE_2_16_RU "Светло-фиолетовый (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_RU "Светло-зелёный (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_RU "Светло-зелёный (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_RU "Светло-циановый (2)"
#define OPTION_VAL_LIGHT_RED_2_20_RU "Светло-красный (2)"
#define OPTION_VAL_MAGENTA_21_RU "Пурпурный"
#define OPTION_VAL_LIGHT_PURPLE_3_22_RU "Светло-фиолетовый (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_RU "Светло-оранжевый"
#define OPTION_VAL_ORANGE_24_RU "Оранжевый"
#define OPTION_VAL_LIGHT_PURPLE_4_25_RU "Светло-фиолетовый (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_RU "Светло-жёлтый"
#define OPTION_VAL_LIGHT_YELLOW_2_27_RU "Светло-жёлтый (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_RU "Цвет выкл. пикселей VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_RU "Видимость экрана VMU 1"
#define OPTION_VAL_40_RU NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_RU "Показывать экран VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_RU "Положение экрана VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_RU "Размер экрана VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_RU "Цвет вкл. пикселей VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_RU "Цвет выкл. пикселей VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_RU "Видимость экрана VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_RU "Показывать экран VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_RU "Положение экрана VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_RU "Размер экрана VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_RU "Цвет вкл. пикселей VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_RU "Цвет выкл. пикселей VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_RU "Видимость экрана VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_RU "Показывать экран VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_RU "Положение экрана VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_RU "Размер экрана VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_RU "Цвет вкл. пикселей VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_RU "Цвет выкл. пикселей VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_RU "Видимость экрана VMU 4"

struct retro_core_option_v2_category option_cats_ru[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_RU,
      CATEGORY_SYSTEM_INFO_0_RU
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_RU,
      CATEGORY_VIDEO_INFO_0_RU
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_RU,
      CATEGORY_PERFORMANCE_INFO_0_RU
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_RU,
      CATEGORY_HACKS_INFO_0_RU
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_RU,
      CATEGORY_INPUT_INFO_0_RU
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_RU,
      CATEGORY_EXPANSIONS_INFO_0_RU
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_RU,
      CATEGORY_VMU_INFO_0_RU
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ru[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_RU,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_RU },
         { "USA",     OPTION_VAL_USA_RU },
         { "Europe",  OPTION_VAL_EUROPE_RU },
         { "Default", OPTION_VAL_DEFAULT_RU },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_RU,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_RU },
         { "English",  OPTION_VAL_ENGLISH_RU },
         { "German",   OPTION_VAL_GERMAN_RU },
         { "French",   OPTION_VAL_FRENCH_RU },
         { "Spanish",  OPTION_VAL_SPANISH_RU },
         { "Italian",  OPTION_VAL_ITALIAN_RU },
         { "Default",  OPTION_VAL_DEFAULT_RU },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_RU,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_RU },
         { "640x480",    OPTION_VAL_640X480_RU },
         { "800x600",    OPTION_VAL_800X600_RU },
         { "960x720",    OPTION_VAL_960X720_RU },
         { "1024x768",   OPTION_VAL_1024X768_RU },
         { "1280x960",   OPTION_VAL_1280X960_RU },
         { "1440x1080",  OPTION_VAL_1440X1080_RU },
         { "1600x1200",  OPTION_VAL_1600X1200_RU },
         { "1920x1440",  OPTION_VAL_1920X1440_RU },
         { "2560x1920",  OPTION_VAL_2560X1920_RU },
         { "2880x2160",  OPTION_VAL_2880X2160_RU },
         { "3200x2400",  OPTION_VAL_3200X2400_RU },
         { "3840x2880",  OPTION_VAL_3840X2880_RU },
         { "4480x3360",  OPTION_VAL_4480X3360_RU },
         { "5120x3840",  OPTION_VAL_5120X3840_RU },
         { "5760x4320",  OPTION_VAL_5760X4320_RU },
         { "6400x4800",  OPTION_VAL_6400X4800_RU },
         { "7040x5280",  OPTION_VAL_7040X5280_RU },
         { "7680x5760",  OPTION_VAL_7680X5760_RU },
         { "8320x6240",  OPTION_VAL_8320X6240_RU },
         { "8960x6720",  OPTION_VAL_8960X6720_RU },
         { "9600x7200",  OPTION_VAL_9600X7200_RU },
         { "10240x7680", OPTION_VAL_10240X7680_RU },
         { "10880x8160", OPTION_VAL_10880X8160_RU },
         { "11520x8640", OPTION_VAL_11520X8640_RU },
         { "12160x9120", OPTION_VAL_12160X9120_RU },
         { "12800x9600", OPTION_VAL_12800X9600_RU },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_RU,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_RU },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_RU },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_RU },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_RU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_RU },
         { "PAL",     OPTION_VAL_PAL_RU },
         { "PAL_N",   OPTION_VAL_PAL_N_RU },
         { "PAL_M",   OPTION_VAL_PAL_M_RU },
         { "Default", OPTION_VAL_DEFAULT_RU },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_RU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_RU },
         { "vertical",   OPTION_VAL_VERTICAL_RU },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_RU,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_RU },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_RU },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_RU },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_RU,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_RU },
         { "1GB",   OPTION_VAL_1GB_RU },
         { "2GB",   OPTION_VAL_2GB_RU },
         { "4GB",   OPTION_VAL_4GB_RU },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_RU,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_RU,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_RU,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_RU,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_RU },
         { "1",  OPTION_VAL_1_RU },
         { "2",  OPTION_VAL_2_RU },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_RU,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_RU },
         { "4",  OPTION_VAL_4_RU },
         { "6",  OPTION_VAL_6_RU },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_RU,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_RU,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_RU,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_RU,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_RU,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_RU,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_RU,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_RU },
         { "more", OPTION_VAL_MORE_RU },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_RU,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_RU },
         { "110", OPTION_VAL_110_RU },
         { "120", OPTION_VAL_120_RU },
         { "130", OPTION_VAL_130_RU },
         { "140", OPTION_VAL_140_RU },
         { "150", OPTION_VAL_150_RU },
         { "160", OPTION_VAL_160_RU },
         { "170", OPTION_VAL_170_RU },
         { "180", OPTION_VAL_180_RU },
         { "190", OPTION_VAL_190_RU },
         { "200", OPTION_VAL_200_RU },
         { "210", OPTION_VAL_210_RU },
         { "220", OPTION_VAL_220_RU },
         { "230", OPTION_VAL_230_RU },
         { "240", OPTION_VAL_240_RU },
         { "250", OPTION_VAL_250_RU },
         { "260", OPTION_VAL_260_RU },
         { "270", OPTION_VAL_270_RU },
         { "280", OPTION_VAL_280_RU },
         { "290", OPTION_VAL_290_RU },
         { "300", OPTION_VAL_300_RU },
         { "310", OPTION_VAL_310_RU },
         { "320", OPTION_VAL_320_RU },
         { "330", OPTION_VAL_330_RU },
         { "340", OPTION_VAL_340_RU },
         { "350", OPTION_VAL_350_RU },
         { "360", OPTION_VAL_360_RU },
         { "370", OPTION_VAL_370_RU },
         { "380", OPTION_VAL_380_RU },
         { "390", OPTION_VAL_390_RU },
         { "400", OPTION_VAL_400_RU },
         { "410", OPTION_VAL_410_RU },
         { "420", OPTION_VAL_420_RU },
         { "430", OPTION_VAL_430_RU },
         { "440", OPTION_VAL_440_RU },
         { "450", OPTION_VAL_450_RU },
         { "460", OPTION_VAL_460_RU },
         { "470", OPTION_VAL_470_RU },
         { "480", OPTION_VAL_480_RU },
         { "490", OPTION_VAL_490_RU },
         { "500", OPTION_VAL_500_RU },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_RU,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_RU },
         { "5%",  OPTION_VAL_5_RU },
         { "10%", OPTION_VAL_10_RU },
         { "15%", OPTION_VAL_15_RU },
         { "20%", OPTION_VAL_20_RU },
         { "25%", OPTION_VAL_25_RU },
         { "30%", OPTION_VAL_30_RU },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_RU,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_RU },
         { "5%",  OPTION_VAL_5_RU },
         { "10%", OPTION_VAL_10_RU },
         { "15%", OPTION_VAL_15_RU },
         { "20%", OPTION_VAL_20_RU },
         { "25%", OPTION_VAL_25_RU },
         { "30%", OPTION_VAL_30_RU },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_RU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_RU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_RU,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_RU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_RU },
         { "60%",  OPTION_VAL_60_RU },
         { "70%",  OPTION_VAL_70_RU },
         { "80%",  OPTION_VAL_80_RU },
         { "90%",  OPTION_VAL_90_RU },
         { "100%", OPTION_VAL_100_O47_RU },
         { "110%", OPTION_VAL_110_O47_RU },
         { "120%", OPTION_VAL_120_O47_RU },
         { "130%", OPTION_VAL_130_O47_RU },
         { "140%", OPTION_VAL_140_O47_RU },
         { "150%", OPTION_VAL_150_O47_RU },
         { "160%", OPTION_VAL_160_O47_RU },
         { "170%", OPTION_VAL_170_O47_RU },
         { "180%", OPTION_VAL_180_O47_RU },
         { "190%", OPTION_VAL_190_O47_RU },
         { "200%", OPTION_VAL_200_O47_RU },
         { "210%", OPTION_VAL_210_O47_RU },
         { "220%", OPTION_VAL_220_O47_RU },
         { "230%", OPTION_VAL_230_O47_RU },
         { "240%", OPTION_VAL_240_O47_RU },
         { "250%", OPTION_VAL_250_O47_RU },
         { "260%", OPTION_VAL_260_O47_RU },
         { "270%", OPTION_VAL_270_O47_RU },
         { "280%", OPTION_VAL_280_O47_RU },
         { "290%", OPTION_VAL_290_O47_RU },
         { "300%", OPTION_VAL_300_O47_RU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_RU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_RU },
         { "Red",      OPTION_VAL_RED_RU },
         { "Green",    OPTION_VAL_GREEN_RU },
         { "Blue",     OPTION_VAL_BLUE_RU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_RU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_RU },
         { "Red",      OPTION_VAL_RED_RU },
         { "Green",    OPTION_VAL_GREEN_RU },
         { "Blue",     OPTION_VAL_BLUE_RU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_RU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_RU },
         { "Red",      OPTION_VAL_RED_RU },
         { "Green",    OPTION_VAL_GREEN_RU },
         { "Blue",     OPTION_VAL_BLUE_RU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_RU,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_RU },
         { "Red",      OPTION_VAL_RED_RU },
         { "Green",    OPTION_VAL_GREEN_RU },
         { "Blue",     OPTION_VAL_BLUE_RU },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_RU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_RU },
         { "Purupuru", OPTION_VAL_PURUPURU_RU },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_RU },
         { "None",     OPTION_VAL_NONE_RU },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_RU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_RU },
         { "Purupuru", OPTION_VAL_PURUPURU_RU },
         { "None",     OPTION_VAL_NONE_RU },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_RU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_RU },
         { "Purupuru", OPTION_VAL_PURUPURU_RU },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_RU },
         { "None",     OPTION_VAL_NONE_RU },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_RU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_RU },
         { "Purupuru", OPTION_VAL_PURUPURU_RU },
         { "None",     OPTION_VAL_NONE_RU },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_RU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_RU },
         { "Purupuru", OPTION_VAL_PURUPURU_RU },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_RU },
         { "None",     OPTION_VAL_NONE_RU },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_RU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_RU },
         { "Purupuru", OPTION_VAL_PURUPURU_RU },
         { "None",     OPTION_VAL_NONE_RU },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_RU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_RU },
         { "Purupuru", OPTION_VAL_PURUPURU_RU },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_RU },
         { "None",     OPTION_VAL_NONE_RU },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_RU,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_RU },
         { "Purupuru", OPTION_VAL_PURUPURU_RU },
         { "None",     OPTION_VAL_NONE_RU },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_RU,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_RU,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_RU,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_RU },
         { "All VMUs", OPTION_VAL_ALL_VMUS_RU },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_RU,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_RU,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_RU,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_RU,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_RU,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_RU,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_RU,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_RU,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_RU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_RU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_RU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_RU },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_RU },
         { "2x", OPTION_VAL_2_O26_RU },
         { "3x", OPTION_VAL_3X_RU },
         { "4x", OPTION_VAL_4_RU },
         { "5x", OPTION_VAL_5X_RU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_RU },
         { "20%",  OPTION_VAL_20_RU },
         { "30%",  OPTION_VAL_30_RU },
         { "40%",  OPTION_VAL_40_RU },
         { "50%",  OPTION_VAL_50_RU },
         { "60%",  OPTION_VAL_60_RU },
         { "70%",  OPTION_VAL_70_RU },
         { "80%",  OPTION_VAL_80_RU },
         { "90%",  OPTION_VAL_90_RU },
         { "100%", OPTION_VAL_100_O47_RU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_RU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_RU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_RU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_RU },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_RU },
         { "2x", OPTION_VAL_2_O26_RU },
         { "3x", OPTION_VAL_3X_RU },
         { "4x", OPTION_VAL_4_RU },
         { "5x", OPTION_VAL_5X_RU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_RU },
         { "20%",  OPTION_VAL_20_RU },
         { "30%",  OPTION_VAL_30_RU },
         { "40%",  OPTION_VAL_40_RU },
         { "50%",  OPTION_VAL_50_RU },
         { "60%",  OPTION_VAL_60_RU },
         { "70%",  OPTION_VAL_70_RU },
         { "80%",  OPTION_VAL_80_RU },
         { "90%",  OPTION_VAL_90_RU },
         { "100%", OPTION_VAL_100_O47_RU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_RU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_RU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_RU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_RU },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_RU },
         { "2x", OPTION_VAL_2_O26_RU },
         { "3x", OPTION_VAL_3X_RU },
         { "4x", OPTION_VAL_4_RU },
         { "5x", OPTION_VAL_5X_RU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_RU },
         { "20%",  OPTION_VAL_20_RU },
         { "30%",  OPTION_VAL_30_RU },
         { "40%",  OPTION_VAL_40_RU },
         { "50%",  OPTION_VAL_50_RU },
         { "60%",  OPTION_VAL_60_RU },
         { "70%",  OPTION_VAL_70_RU },
         { "80%",  OPTION_VAL_80_RU },
         { "90%",  OPTION_VAL_90_RU },
         { "100%", OPTION_VAL_100_O47_RU },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_RU },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_RU },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_RU },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_RU },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_RU },
         { "2x", OPTION_VAL_2_O26_RU },
         { "3x", OPTION_VAL_3X_RU },
         { "4x", OPTION_VAL_4_RU },
         { "5x", OPTION_VAL_5X_RU },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_RU },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_RU },
         { "BLACK 02",          OPTION_VAL_BLACK_02_RU },
         { "BLUE 03",           OPTION_VAL_BLUE_RU },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_RU },
         { "GREEN 05",          OPTION_VAL_GREEN_RU },
         { "CYAN 06",           OPTION_VAL_CYAN_06_RU },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_RU },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_RU },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_RU },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_RU },
         { "RED 11",            OPTION_VAL_RED_RU },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_RU },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_RU },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_RU },
         { "GRAY 15",           OPTION_VAL_GRAY_15_RU },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_RU },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_RU },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_RU },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_RU },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_RU },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_RU },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_RU },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_RU },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_RU },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_RU },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_RU },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_RU },
         { "WHITE 28",          OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_RU,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_RU },
         { "20%",  OPTION_VAL_20_RU },
         { "30%",  OPTION_VAL_30_RU },
         { "40%",  OPTION_VAL_40_RU },
         { "50%",  OPTION_VAL_50_RU },
         { "60%",  OPTION_VAL_60_RU },
         { "70%",  OPTION_VAL_70_RU },
         { "80%",  OPTION_VAL_80_RU },
         { "90%",  OPTION_VAL_90_RU },
         { "100%", OPTION_VAL_100_O47_RU },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ru = {
   option_cats_ru,
   option_defs_ru
};

/* RETRO_LANGUAGE_SK */

#define CATEGORY_SYSTEM_LABEL_SK "Systém"
#define CATEGORY_SYSTEM_INFO_0_SK "Nakonfigurujte nastavenia oblasti, jazyka, BIOS-u a základného hardvéru."
#define CATEGORY_VIDEO_LABEL_SK NULL
#define CATEGORY_VIDEO_INFO_0_SK "Nakonfigurujte nastavenia rozlíšenia, na poradí nezávislej priehľadnosti a vizuálnych efektov."
#define CATEGORY_PERFORMANCE_LABEL_SK "Výkon"
#define CATEGORY_PERFORMANCE_INFO_0_SK "Nakonfigurujte nastavenia viacvláknového vykresľovania a preskakovania snímok."
#define CATEGORY_HACKS_LABEL_SK "Emulačné hacky"
#define CATEGORY_HACKS_INFO_0_SK "Nakonfigurujte nastavenia prepísaní širokouhlého režimu, rýchlosti načítavania GD-ROM a nahrádzania textúr."
#define CATEGORY_INPUT_LABEL_SK "Vstup"
#define CATEGORY_INPUT_INFO_0_SK "Nakonfigurujte nastavenia gamepadu a svetelnej pištole."
#define CATEGORY_EXPANSIONS_LABEL_SK "Rozširujúce sloty ovládača"
#define CATEGORY_EXPANSIONS_INFO_0_SK "Vyberte zariadenie (VMU, vibračné zariadenie) pripojené v každom rozširujúcom slote ovládača."
#define CATEGORY_VMU_LABEL_SK NULL
#define CATEGORY_VMU_INFO_0_SK "Nakonfigurujte nastavenia uložení VMU pre každú hru a viditeľnosti VMU na obrazovke."
#define CORE_OPTION_NAME_REGION_LABEL_SK "Región"
#define OPTION_VAL_JAPAN_SK "Japonsko"
#define OPTION_VAL_USA_SK NULL
#define OPTION_VAL_EUROPE_SK "Európa"
#define OPTION_VAL_DEFAULT_SK "Predvolené"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_SK "Jazyk"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_SK "Zmení jazyk používaný BIOS-om a hrami, ktoré obsahujú viacero jazykov."
#define OPTION_VAL_JAPANESE_SK "Japončina"
#define OPTION_VAL_ENGLISH_SK "Angličtina"
#define OPTION_VAL_GERMAN_SK "Nemčina"
#define OPTION_VAL_FRENCH_SK "Francúzština"
#define OPTION_VAL_SPANISH_SK "Španielčina"
#define OPTION_VAL_ITALIAN_SK "Taliančina"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_SK "HLE BIOS (vyžaduje reštart)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_SK "Vynútiť použitie BIOS-u s vysokoúrovňovou emuláciou."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_SK "Povoliť DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_SK "Zapne emuláciu audio DSP Dreamcastu (digital signal processor). Zlepšuje presnosť generovaného zvuku, ale zvyšuje nároky na výkon."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_SK "Povoliť servisné tlačidlá arkády"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_SK "Povolí tlačidlo SERVICE pre arkádové hry na vstup do nastavení skrinky."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_SK "Nastaviť hry NAOMI na voľné hranie"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_SK "Upraví nastavenia mincí hry na voľné hranie."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_SK "Emulácia širokopásmového adaptéra"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_SK "Emulovať širokopásmový Ethernet adaptér namiesto modemu. (vyžaduje reštart)"
#define CORE_OPTION_NAME_UPNP_LABEL_SK "Povoliť UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_SK "Použiť UPnP na automatické nakonfigurovanie internetového routeru pre online hry."
#define CORE_OPTION_NAME_DCNET_LABEL_SK "Použiť DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_SK "Použiť cloudovú službu DCNet pre prístup Dreamcastu na Internet."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_SK "Vnútorné rozlíšenie"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_SK "Upraviť rozlíšenie renderingu."
#define OPTION_VAL_320X240_SK "320x240 (polovičné)"
#define OPTION_VAL_640X480_SK "640x480 (natívne)"
#define OPTION_VAL_800X600_SK NULL
#define OPTION_VAL_960X720_SK NULL
#define OPTION_VAL_1024X768_SK NULL
#define OPTION_VAL_1280X960_SK NULL
#define OPTION_VAL_1440X1080_SK NULL
#define OPTION_VAL_1600X1200_SK NULL
#define OPTION_VAL_1920X1440_SK NULL
#define OPTION_VAL_2560X1920_SK NULL
#define OPTION_VAL_2880X2160_SK NULL
#define OPTION_VAL_3200X2400_SK NULL
#define OPTION_VAL_3840X2880_SK NULL
#define OPTION_VAL_4480X3360_SK NULL
#define OPTION_VAL_5120X3840_SK NULL
#define OPTION_VAL_5760X4320_SK NULL
#define OPTION_VAL_6400X4800_SK NULL
#define OPTION_VAL_7040X5280_SK NULL
#define OPTION_VAL_7680X5760_SK NULL
#define OPTION_VAL_8320X6240_SK NULL
#define OPTION_VAL_8960X6720_SK NULL
#define OPTION_VAL_9600X7200_SK NULL
#define OPTION_VAL_10240X7680_SK NULL
#define OPTION_VAL_10880X8160_SK NULL
#define OPTION_VAL_11520X8640_SK NULL
#define OPTION_VAL_12160X9120_SK NULL
#define OPTION_VAL_12800X9600_SK NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_SK "Typ kábla"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_SK "Typ výstupného signálu. 'TV (Composite)' je najviac podporovaný."
#define OPTION_VAL_VGA_SK NULL
#define OPTION_VAL_TV_RGB_SK NULL
#define OPTION_VAL_TV_COMPOSITE_SK "TV (composite)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_SK "Vysielací štandard"
#define OPTION_VAL_NTSC_SK NULL
#define OPTION_VAL_PAL_SK "PAL (svet)"
#define OPTION_VAL_PAL_N_SK "PAL-N (Argentína, Paraguaj, Uruguaj)"
#define OPTION_VAL_PAL_M_SK "PAL-M (Brazília)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_SK "Orientácia Obrazovky"
#define OPTION_VAL_HORIZONTAL_SK "Vodorovne"
#define OPTION_VAL_VERTICAL_SK "Zvislo"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_SK "Alfa triedenie"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_SK "Vyberte, ako sa zoraďujú priehľadné polygóny."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_SK "Per-Strip (rýchle, najmenej presné)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_SK "Per-Triangle (normálne)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_SK "Per-Pixel (presné, ale najpomalšie)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_SK "Veľkosť akumulačného pixelového bufferu"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_SK "Vyššie hodnoty môžu byť potrebné pre vyššie rozlíšenia, aby sa správne zobrazovali."
#define OPTION_VAL_512MB_SK "512 MB"
#define OPTION_VAL_1GB_SK NULL
#define OPTION_VAL_2GB_SK NULL
#define OPTION_VAL_4GB_SK NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_SK "Maximálny počet priehľadných vrstiev"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_SK "Pri zložitých scénach môžu byť potrebné vyššie hodnoty."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_SK "Plná emulácia framebufferu"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_SK "Povolí plnú emuláciu framebufferu vo VRAM. Užitočné pre hry, ktoré priamo čítajú alebo zapisujú framebuffer vo VRAM. Po povolení sa interné rozlíšenie vynúti na 640x480 a výkon môže byť výrazne ovplyvnený."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_SK "Zapnúť buffer RTT (Render To Texture)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_SK "Kopíruje vykreslené textúry späť z GPU do VRAM. Táto voľba je zvyčajne povolená pre hry, ktoré ju vyžadujú. Po povolení sa zvýšenie rozlíšenia textúr deaktivuje a výkon môže byť ovplyvnený."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_SK NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_SK "Po povolení textúry použijú menšiu verziu seba samých, keď sú ďalej. Môže to zvýšiť výkon a obmedziť mihanie."
#define CORE_OPTION_NAME_FOG_LABEL_SK "Efekty hmly"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_SK "Modifikátor objemu"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_SK "Funkcia Dreamcast GPU, ktorá sa typicky používa v hrách na vykreslenie tieňov objektov. Mala by byť normálne zapnutá – dopad na výkon je zvyčajne minimálny až zanedbateľný."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_SK "Anizotropné filtrovanie"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_SK "Vylepší kvalitu textúr na povrchoch, ktoré sú voči kamere v šikmom uhle pohľadu. Vyššie hodnoty viac zaťažujú GPU. Zmeny tohto nastavenia sa prejavia až po reštarte."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_SK "Filtrovanie textúr"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_SK "Režim filtrovania textúr, ktorý sa má použiť. Možno použiť na vynútenie určitého režimu filtrovania textúr na všetky textúry, aby sa dosiahol ostrejší (alebo hladší) vzhľad ako Predvolený. Iné hodnoty než Predvolené môžu spôsobiť rôzne problémy s vykresľovaním. Zmeny tohto nastavenia sa prejavia až po reštarte."
#define OPTION_VAL_1_SK "Vynútiť Nearest-Neighbor"
#define OPTION_VAL_2_SK "Vynútiť lineárne"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_SK "Oneskoriť výmenu snímok"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_SK "Užitočné na zabránenie blikajúcim obrazovkám alebo chybnému videu. Neodporúča sa na pomalých platformách."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_SK "Detekovať zmeny snímkovej frekvencie"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_SK "Upozorní frontend, keď sa interná snímková frekvencia zmení (napr. z 60 fps na 30 fps). Zlepší rovnomernosť snímok v hrách, ktoré bežia s pevnou rýchlosťou 30 fps alebo 20 fps, ale malo by byť zakázané pre hry s neuzamknutou (nestabilnou) snímkovou frekvenciou (napr. Ecco the Dolphin, Unreal Tournament). Poznámka: Nedostupné, keď je povolená možnosť 'Auto Skip Frame'."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_SK "Post-procesný filter PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_SK "Post-procesovanie vyrenderovaného obrazu na simuláciu efektov špecifických pre PowerVR2 GPU a analógové video signály."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_SK "Zväčšenie textúr (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_SK "Zlepšiť ručne kreslenú 2D pixel art grafiku. Malo by sa používať len s 2D pixelovanými hrami."
#define OPTION_VAL_2_O26_SK NULL
#define OPTION_VAL_4_SK NULL
#define OPTION_VAL_6_SK NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_SK "Max. filtrovaná veľkosť pri zväčšení textúr"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_SK "Vyberte maximálnu veľkosť textúry, ktorá sa má zväčšiť. Ak je veľkosť textúry vyššia než vybraná hodnota, nebude zväčšená."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_SK "Natívna interpolácia hĺbky"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_SK "Pomáha s poškodením textúr a problémami hĺbky na AMD GPU. Môže pomôcť aj na Intel GPU v niektorých prípadoch."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_SK "Oprava presakujúceho okraja pri zväčšení"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_SK "Pomáha s presakovaním textúr pri zväčšení. Zakázanie môže pomôcť, ak sa pixely deformujú pri zväčšení v 2D hrách (MVC2, CVS, KOF atď.)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_SK "Viacvláknové vykresľovanie"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_SK "Spúšťa GPU a CPU v rôznych vláknach. Veľmi odporúčané."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_SK "Automatické preskakovanie snímok"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_SK "Automaticky preskakuje snímky, keď emulátor beží pomaly. Poznámka: Toto nastavenie sa použije, len keď je povolené 'Viacvláknové vykresľovanie'."
#define OPTION_VAL_SOME_SK "Normálne"
#define OPTION_VAL_MORE_SK "Maximálne"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_SK "Preskakovanie snímkov"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_SK "Nastaví počet snímok, ktoré sa preskočia medzi každou zobrazenou snímkou."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_SK "Cheaty pre širokouhlé zobrazenie (vyžaduje reštart)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_SK "Aktivuje cheaty, ktoré umožňujú zobrazenie niektorých hier vo formáte širokouhlého obrazu."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_SK "Hack širokouhlého režimu"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_SK "Vykreslí geometriu mimo bežného pomeru strán 4:3. Môže spôsobiť grafické chyby v odhalených oblastiach."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_SK "Rýchle načítanie GD-ROM (nepresné)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_SK "Zrýchli načítanie GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_SK "Mod 32MB RAM pre Dreamcast"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_SK "Povolí mod 32MB RAM pre Dreamcast. Môže ovplyvniť kompatibilitu"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_SK "SH4 CPU pretaktovanie/podtaktovanie"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_SK "Zmení frekvenciu hlavného CPU SH4 z predvolených 200 MHz. Podtaktovanie môže pomôcť pomalým platformám. Pretaktovanie môže zvýšiť snímkovú frekvenciu v niektorých hrách. Používajte s opatrnosťou."
#define OPTION_VAL_100_SK NULL
#define OPTION_VAL_110_SK NULL
#define OPTION_VAL_120_SK NULL
#define OPTION_VAL_130_SK NULL
#define OPTION_VAL_140_SK NULL
#define OPTION_VAL_150_SK NULL
#define OPTION_VAL_160_SK NULL
#define OPTION_VAL_170_SK NULL
#define OPTION_VAL_180_SK NULL
#define OPTION_VAL_190_SK NULL
#define OPTION_VAL_200_SK "200 Hz"
#define OPTION_VAL_210_SK "210 Hz"
#define OPTION_VAL_220_SK "220 Hz"
#define OPTION_VAL_230_SK "230 Hz"
#define OPTION_VAL_240_SK "240 Hz"
#define OPTION_VAL_250_SK "250 Hz"
#define OPTION_VAL_260_SK "260 Hz"
#define OPTION_VAL_270_SK "270 Hz"
#define OPTION_VAL_280_SK "280 Hz"
#define OPTION_VAL_290_SK "290 Hz"
#define OPTION_VAL_300_SK "300 Hz"
#define OPTION_VAL_310_SK "310 Hz"
#define OPTION_VAL_320_SK "320 Hz"
#define OPTION_VAL_330_SK NULL
#define OPTION_VAL_340_SK "340 Hz"
#define OPTION_VAL_350_SK "350 Hz"
#define OPTION_VAL_360_SK "360 Hz"
#define OPTION_VAL_370_SK "370 Hz"
#define OPTION_VAL_380_SK "380 Hz"
#define OPTION_VAL_390_SK "390 Hz"
#define OPTION_VAL_400_SK "400 Hz"
#define OPTION_VAL_410_SK "410 Hz"
#define OPTION_VAL_420_SK "420 Hz"
#define OPTION_VAL_430_SK "430 Hz"
#define OPTION_VAL_440_SK "440 Hz"
#define OPTION_VAL_450_SK "450 Hz"
#define OPTION_VAL_460_SK "460 Hz"
#define OPTION_VAL_470_SK "470 Hz"
#define OPTION_VAL_480_SK "480 Hz"
#define OPTION_VAL_490_SK "490 Hz"
#define OPTION_VAL_500_SK NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_SK "Načítať vlastné textúry"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_SK "Načítať vlastné textúry umiestnené v priečinku 'system/dc/textures/<game-id>/'."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_SK "Predčítať vlastné textúry"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_SK "Predčíta vlastné textúry pri spustení hry. Môže zlepšiť výkon, ale zvyšuje využitie pamäte."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_SK "Dumpovať textúry"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_SK "Vždy, keď hra použije novú textúru, sa uloží ako súbor .png v priečinku 'system/dc/texdump/<game-id>/'."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_SK "Dumpovať nahradené textúry"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_SK "Vždy dumpovať textúry, ktoré sú už nahradené vlastnými textúrami."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_SK "Mŕtva zóna analógovej páčky"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_SK "Vyberte, ako ďaleko musíte zatlačiť analógovú páčku predtým, ako sa začne spracúvať."
#define OPTION_VAL_0_SK NULL
#define OPTION_VAL_5_SK NULL
#define OPTION_VAL_10_SK NULL
#define OPTION_VAL_15_SK NULL
#define OPTION_VAL_20_SK NULL
#define OPTION_VAL_25_SK NULL
#define OPTION_VAL_30_SK NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_SK "Mŕtva zóna triggera"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_SK "Vyberte, ako veľmi musíte stlačiť trigger predtým, ako sa začne spracúvať."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_SK "Digitálne triggery"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_SK "Po povolení budú triggery fungovať ako bežné tlačidlá, čo znamená, že sa budú spracúvať buď ako úplne stlačené alebo vôbec nestlačené, bez medzistavu."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_SK "Vysielať digitálne výstupy"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_SK "Vysielať digitálne výstupy a stav force-feedbacku na TCP porte 8000. Kompatibilné s voľbou MAME \"-output network\"."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_SK "Zobraziť nastavenia svetelnej pištole"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_SK "Povolí konfiguráciu zobrazovacích možností zameriavača svetelnej pištole. POZNÁMKA: Možno bude potrebné prepnúť Quick Menu, aby sa nastavenie prejavilo."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_SK "Škálovanie veľkosti zameriavača pištole"
#define OPTION_VAL_50_SK NULL
#define OPTION_VAL_60_SK NULL
#define OPTION_VAL_70_SK NULL
#define OPTION_VAL_80_SK NULL
#define OPTION_VAL_90_SK NULL
#define OPTION_VAL_100_O47_SK NULL
#define OPTION_VAL_110_O47_SK NULL
#define OPTION_VAL_120_O47_SK NULL
#define OPTION_VAL_130_O47_SK NULL
#define OPTION_VAL_140_O47_SK NULL
#define OPTION_VAL_150_O47_SK NULL
#define OPTION_VAL_160_O47_SK NULL
#define OPTION_VAL_170_O47_SK NULL
#define OPTION_VAL_180_O47_SK NULL
#define OPTION_VAL_190_O47_SK NULL
#define OPTION_VAL_200_O47_SK NULL
#define OPTION_VAL_210_O47_SK NULL
#define OPTION_VAL_220_O47_SK NULL
#define OPTION_VAL_230_O47_SK NULL
#define OPTION_VAL_240_O47_SK NULL
#define OPTION_VAL_250_O47_SK NULL
#define OPTION_VAL_260_O47_SK NULL
#define OPTION_VAL_270_O47_SK NULL
#define OPTION_VAL_280_O47_SK NULL
#define OPTION_VAL_290_O47_SK NULL
#define OPTION_VAL_300_O47_SK NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_SK "Displej zameriavača pištole 1"
#define OPTION_VAL_WHITE_SK "Biela"
#define OPTION_VAL_RED_SK "Červená"
#define OPTION_VAL_GREEN_SK "Zelená"
#define OPTION_VAL_BLUE_SK "Modrá"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_SK "Displej zameriavača pištole 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_SK "Displej zameriavača pištole 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_SK "Displej zameriavača pištole 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_SK "Zariadenie v rozširujúcom slote A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_SK "Vyberte zariadenie pripojené v rozširujúcom slote A1 (port A slot 1)."
#define OPTION_VAL_VMU_SK NULL
#define OPTION_VAL_PURUPURU_SK "Vibračný balík"
#define OPTION_VAL_DREAMPOTATO_SK NULL
#define OPTION_VAL_NONE_SK "Žiadne"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_SK "Zariadenie v rozširujúcom slote A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_SK "Vyberte zariadenie pripojené v rozširujúcom slote A2 (port A slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_SK "Zariadenie v rozširujúcom slote B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_SK "Vyberte zariadenie pripojené v rozširujúcom slote B1 (port B slot 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_SK "Zariadenie v rozširujúcom slote B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_SK "Vyberte zariadenie pripojené v rozširujúcom slote B2 (port B slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_SK "Zariadenie v rozširujúcom slote C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_SK "Vyberte zariadenie pripojené v rozširujúcom slote C1 (port C slot 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_SK "Zariadenie v rozširujúcom slote C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_SK "Vyberte zariadenie pripojené v rozširujúcom slote C2 (port C slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_SK "Zariadenie v rozširujúcom slote D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_SK "Vyberte zariadenie pripojené v rozširujúcom slote D1 (port D slot 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_SK "Zariadenie v rozširujúcom slote D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_SK "Vyberte zariadenie pripojené v rozširujúcom slote D2 (port D slot 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_SK "Visual Memory Units/Systems (VMU) pre každú hru"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_SK "VMU pre každú hru"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_SK "Po zakázaní zdieľajú všetky hry až 8 uložení VMU (A1/A2/B1/B2/C1/C2/D1/D2) umiestnených v systémovom priečinku RetroArchu.\nNastavenie 'VMU A1' vytvorí unikátny VMU súbor 'A1' v priečinku uložení RetroArchu pre každú spustenú hru.\nNastavenie 'All VMUs' vytvorí až 8 unikátnych VMU súborov (A1/A2/B1/B2/C1/C2/D1/D2) pre každú spustenú hru."
#define OPTION_VAL_VMU_A1_SK NULL
#define OPTION_VAL_ALL_VMUS_SK "Všetky VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_SK "Zvuky Visual Memory Units/Systems (VMU)"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_SK "Zvuky VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_SK "Po povolení sa prehráva pípanie VMU."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_SK "Úložisko VMU DreamPotato"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_SK "Povolí ukladanie údajov VMU na pripojenom emulátore DreamPotato. Po povolení sa lokálny súbor uložení VMU nepoužije."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_SK "Zobraziť nastavenia displeja Visual Memory Unit/System (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_SK "Zobraziť nastavenia displeja VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_SK "Povolí konfiguráciu viditeľnosti, veľkosti, polohy a farby emulovanej obrazovky LCD VMU. POZNÁMKA: Možno bude potrebné prepnúť Quick Menu, aby sa nastavenie prejavilo."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_SK "Displej obrazovky VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_SK "Poloha obrazovky VMU 1"
#define OPTION_VAL_UPPER_LEFT_SK "Hore vľavo"
#define OPTION_VAL_UPPER_RIGHT_SK "Hore vpravo"
#define OPTION_VAL_LOWER_LEFT_SK "Vľavo dolu"
#define OPTION_VAL_LOWER_RIGHT_SK "Vpravo dolu"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_SK "Veľkosť obrazovky VMU 1"
#define OPTION_VAL_1X_SK NULL
#define OPTION_VAL_3X_SK NULL
#define OPTION_VAL_5X_SK NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_SK "Farba zapnutého pixelu obrazovky VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_SK "Predvolene zapnuté"
#define OPTION_VAL_DEFAULT_OFF_01_SK "Predvolene vypnuté"
#define OPTION_VAL_BLACK_02_SK "Čierna"
#define OPTION_VAL_LIGHT_BLUE_04_SK "Svetlomodrá"
#define OPTION_VAL_CYAN_06_SK "Azúrová"
#define OPTION_VAL_CYAN_BLUE_07_SK "Tyrkysová modrá"
#define OPTION_VAL_LIGHT_GREEN_08_SK "Svetlozelená"
#define OPTION_VAL_CYAN_GREEN_09_SK "Tyrkysová zelená"
#define OPTION_VAL_LIGHT_CYAN_10_SK "Svetloazúrová"
#define OPTION_VAL_PURPLE_12_SK "Fialová"
#define OPTION_VAL_LIGHT_PURPLE_13_SK "Svetlofialová"
#define OPTION_VAL_YELLOW_14_SK "Žltá"
#define OPTION_VAL_GRAY_15_SK "Šedá"
#define OPTION_VAL_LIGHT_PURPLE_2_16_SK "Svetlofialová (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_SK "Svetlozelená (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_SK "Svetlozelená (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_SK "Svetloazúrová (2)"
#define OPTION_VAL_LIGHT_RED_2_20_SK "Svetločervená (2)"
#define OPTION_VAL_MAGENTA_21_SK "Purpurová"
#define OPTION_VAL_LIGHT_PURPLE_3_22_SK "Svetlofialová (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_SK "Svetlooranžová"
#define OPTION_VAL_ORANGE_24_SK "Oranžová"
#define OPTION_VAL_LIGHT_PURPLE_4_25_SK "Svetlofialová (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_SK "Svetložltá"
#define OPTION_VAL_LIGHT_YELLOW_2_27_SK "Svetložltá (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_SK "Farba vypnutého pixelu obrazovky VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_SK "Priehľadnosť obrazovky VMU 1"
#define OPTION_VAL_40_SK NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_SK "Displej obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_SK "Poloha obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_SK "Veľkosť obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_SK "Farba zapnutého pixelu obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_SK "Farba vypnutého pixelu obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_SK "Priehľadnosť obrazovky VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_SK "Displej obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_SK "Poloha obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_SK "Veľkosť obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_SK "Farba zapnutého pixelu obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_SK "Farba vypnutého pixelu obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_SK "Priehľadnosť obrazovky VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_SK "Displej obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_SK "Poloha obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_SK "Veľkosť obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_SK "Farba zapnutého pixelu obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_SK "Farba vypnutého pixelu obrazovky VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_SK "Priehľadnosť obrazovky VMU 4"

struct retro_core_option_v2_category option_cats_sk[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_SK,
      CATEGORY_SYSTEM_INFO_0_SK
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_SK,
      CATEGORY_VIDEO_INFO_0_SK
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_SK,
      CATEGORY_PERFORMANCE_INFO_0_SK
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_SK,
      CATEGORY_HACKS_INFO_0_SK
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_SK,
      CATEGORY_INPUT_INFO_0_SK
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_SK,
      CATEGORY_EXPANSIONS_INFO_0_SK
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_SK,
      CATEGORY_VMU_INFO_0_SK
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_sk[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_SK,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_SK },
         { "USA",     OPTION_VAL_USA_SK },
         { "Europe",  OPTION_VAL_EUROPE_SK },
         { "Default", OPTION_VAL_DEFAULT_SK },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_SK,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_SK },
         { "English",  OPTION_VAL_ENGLISH_SK },
         { "German",   OPTION_VAL_GERMAN_SK },
         { "French",   OPTION_VAL_FRENCH_SK },
         { "Spanish",  OPTION_VAL_SPANISH_SK },
         { "Italian",  OPTION_VAL_ITALIAN_SK },
         { "Default",  OPTION_VAL_DEFAULT_SK },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_SK,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_SK },
         { "640x480",    OPTION_VAL_640X480_SK },
         { "800x600",    OPTION_VAL_800X600_SK },
         { "960x720",    OPTION_VAL_960X720_SK },
         { "1024x768",   OPTION_VAL_1024X768_SK },
         { "1280x960",   OPTION_VAL_1280X960_SK },
         { "1440x1080",  OPTION_VAL_1440X1080_SK },
         { "1600x1200",  OPTION_VAL_1600X1200_SK },
         { "1920x1440",  OPTION_VAL_1920X1440_SK },
         { "2560x1920",  OPTION_VAL_2560X1920_SK },
         { "2880x2160",  OPTION_VAL_2880X2160_SK },
         { "3200x2400",  OPTION_VAL_3200X2400_SK },
         { "3840x2880",  OPTION_VAL_3840X2880_SK },
         { "4480x3360",  OPTION_VAL_4480X3360_SK },
         { "5120x3840",  OPTION_VAL_5120X3840_SK },
         { "5760x4320",  OPTION_VAL_5760X4320_SK },
         { "6400x4800",  OPTION_VAL_6400X4800_SK },
         { "7040x5280",  OPTION_VAL_7040X5280_SK },
         { "7680x5760",  OPTION_VAL_7680X5760_SK },
         { "8320x6240",  OPTION_VAL_8320X6240_SK },
         { "8960x6720",  OPTION_VAL_8960X6720_SK },
         { "9600x7200",  OPTION_VAL_9600X7200_SK },
         { "10240x7680", OPTION_VAL_10240X7680_SK },
         { "10880x8160", OPTION_VAL_10880X8160_SK },
         { "11520x8640", OPTION_VAL_11520X8640_SK },
         { "12160x9120", OPTION_VAL_12160X9120_SK },
         { "12800x9600", OPTION_VAL_12800X9600_SK },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_SK,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_SK },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_SK },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_SK },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_SK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_SK },
         { "PAL",     OPTION_VAL_PAL_SK },
         { "PAL_N",   OPTION_VAL_PAL_N_SK },
         { "PAL_M",   OPTION_VAL_PAL_M_SK },
         { "Default", OPTION_VAL_DEFAULT_SK },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_SK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_SK },
         { "vertical",   OPTION_VAL_VERTICAL_SK },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_SK,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_SK },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_SK },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_SK },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_SK,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_SK },
         { "1GB",   OPTION_VAL_1GB_SK },
         { "2GB",   OPTION_VAL_2GB_SK },
         { "4GB",   OPTION_VAL_4GB_SK },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_SK,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_SK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_SK,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_SK,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_SK },
         { "1",  OPTION_VAL_1_SK },
         { "2",  OPTION_VAL_2_SK },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_SK,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_SK },
         { "4",  OPTION_VAL_4_SK },
         { "6",  OPTION_VAL_6_SK },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_SK,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_SK,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_SK,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_SK,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_SK,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_SK,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_SK,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_SK },
         { "more", OPTION_VAL_MORE_SK },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_SK,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_SK },
         { "110", OPTION_VAL_110_SK },
         { "120", OPTION_VAL_120_SK },
         { "130", OPTION_VAL_130_SK },
         { "140", OPTION_VAL_140_SK },
         { "150", OPTION_VAL_150_SK },
         { "160", OPTION_VAL_160_SK },
         { "170", OPTION_VAL_170_SK },
         { "180", OPTION_VAL_180_SK },
         { "190", OPTION_VAL_190_SK },
         { "200", OPTION_VAL_200_SK },
         { "210", OPTION_VAL_210_SK },
         { "220", OPTION_VAL_220_SK },
         { "230", OPTION_VAL_230_SK },
         { "240", OPTION_VAL_240_SK },
         { "250", OPTION_VAL_250_SK },
         { "260", OPTION_VAL_260_SK },
         { "270", OPTION_VAL_270_SK },
         { "280", OPTION_VAL_280_SK },
         { "290", OPTION_VAL_290_SK },
         { "300", OPTION_VAL_300_SK },
         { "310", OPTION_VAL_310_SK },
         { "320", OPTION_VAL_320_SK },
         { "330", OPTION_VAL_330_SK },
         { "340", OPTION_VAL_340_SK },
         { "350", OPTION_VAL_350_SK },
         { "360", OPTION_VAL_360_SK },
         { "370", OPTION_VAL_370_SK },
         { "380", OPTION_VAL_380_SK },
         { "390", OPTION_VAL_390_SK },
         { "400", OPTION_VAL_400_SK },
         { "410", OPTION_VAL_410_SK },
         { "420", OPTION_VAL_420_SK },
         { "430", OPTION_VAL_430_SK },
         { "440", OPTION_VAL_440_SK },
         { "450", OPTION_VAL_450_SK },
         { "460", OPTION_VAL_460_SK },
         { "470", OPTION_VAL_470_SK },
         { "480", OPTION_VAL_480_SK },
         { "490", OPTION_VAL_490_SK },
         { "500", OPTION_VAL_500_SK },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_SK,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_SK },
         { "5%",  OPTION_VAL_5_SK },
         { "10%", OPTION_VAL_10_SK },
         { "15%", OPTION_VAL_15_SK },
         { "20%", OPTION_VAL_20_SK },
         { "25%", OPTION_VAL_25_SK },
         { "30%", OPTION_VAL_30_SK },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_SK,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_SK },
         { "5%",  OPTION_VAL_5_SK },
         { "10%", OPTION_VAL_10_SK },
         { "15%", OPTION_VAL_15_SK },
         { "20%", OPTION_VAL_20_SK },
         { "25%", OPTION_VAL_25_SK },
         { "30%", OPTION_VAL_30_SK },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_SK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_SK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_SK,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_SK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_SK },
         { "60%",  OPTION_VAL_60_SK },
         { "70%",  OPTION_VAL_70_SK },
         { "80%",  OPTION_VAL_80_SK },
         { "90%",  OPTION_VAL_90_SK },
         { "100%", OPTION_VAL_100_O47_SK },
         { "110%", OPTION_VAL_110_O47_SK },
         { "120%", OPTION_VAL_120_O47_SK },
         { "130%", OPTION_VAL_130_O47_SK },
         { "140%", OPTION_VAL_140_O47_SK },
         { "150%", OPTION_VAL_150_O47_SK },
         { "160%", OPTION_VAL_160_O47_SK },
         { "170%", OPTION_VAL_170_O47_SK },
         { "180%", OPTION_VAL_180_O47_SK },
         { "190%", OPTION_VAL_190_O47_SK },
         { "200%", OPTION_VAL_200_O47_SK },
         { "210%", OPTION_VAL_210_O47_SK },
         { "220%", OPTION_VAL_220_O47_SK },
         { "230%", OPTION_VAL_230_O47_SK },
         { "240%", OPTION_VAL_240_O47_SK },
         { "250%", OPTION_VAL_250_O47_SK },
         { "260%", OPTION_VAL_260_O47_SK },
         { "270%", OPTION_VAL_270_O47_SK },
         { "280%", OPTION_VAL_280_O47_SK },
         { "290%", OPTION_VAL_290_O47_SK },
         { "300%", OPTION_VAL_300_O47_SK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_SK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SK },
         { "Red",      OPTION_VAL_RED_SK },
         { "Green",    OPTION_VAL_GREEN_SK },
         { "Blue",     OPTION_VAL_BLUE_SK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_SK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SK },
         { "Red",      OPTION_VAL_RED_SK },
         { "Green",    OPTION_VAL_GREEN_SK },
         { "Blue",     OPTION_VAL_BLUE_SK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_SK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SK },
         { "Red",      OPTION_VAL_RED_SK },
         { "Green",    OPTION_VAL_GREEN_SK },
         { "Blue",     OPTION_VAL_BLUE_SK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_SK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SK },
         { "Red",      OPTION_VAL_RED_SK },
         { "Green",    OPTION_VAL_GREEN_SK },
         { "Blue",     OPTION_VAL_BLUE_SK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_SK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SK },
         { "Purupuru", OPTION_VAL_PURUPURU_SK },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_SK },
         { "None",     OPTION_VAL_NONE_SK },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_SK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SK },
         { "Purupuru", OPTION_VAL_PURUPURU_SK },
         { "None",     OPTION_VAL_NONE_SK },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_SK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SK },
         { "Purupuru", OPTION_VAL_PURUPURU_SK },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_SK },
         { "None",     OPTION_VAL_NONE_SK },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_SK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SK },
         { "Purupuru", OPTION_VAL_PURUPURU_SK },
         { "None",     OPTION_VAL_NONE_SK },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_SK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SK },
         { "Purupuru", OPTION_VAL_PURUPURU_SK },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_SK },
         { "None",     OPTION_VAL_NONE_SK },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_SK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SK },
         { "Purupuru", OPTION_VAL_PURUPURU_SK },
         { "None",     OPTION_VAL_NONE_SK },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_SK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SK },
         { "Purupuru", OPTION_VAL_PURUPURU_SK },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_SK },
         { "None",     OPTION_VAL_NONE_SK },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_SK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SK },
         { "Purupuru", OPTION_VAL_PURUPURU_SK },
         { "None",     OPTION_VAL_NONE_SK },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_SK,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_SK,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_SK,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_SK },
         { "All VMUs", OPTION_VAL_ALL_VMUS_SK },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_SK,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_SK,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_SK,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_SK,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_SK,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_SK,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_SK,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_SK,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SK },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SK },
         { "2x", OPTION_VAL_2_O26_SK },
         { "3x", OPTION_VAL_3X_SK },
         { "4x", OPTION_VAL_4_SK },
         { "5x", OPTION_VAL_5X_SK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SK },
         { "20%",  OPTION_VAL_20_SK },
         { "30%",  OPTION_VAL_30_SK },
         { "40%",  OPTION_VAL_40_SK },
         { "50%",  OPTION_VAL_50_SK },
         { "60%",  OPTION_VAL_60_SK },
         { "70%",  OPTION_VAL_70_SK },
         { "80%",  OPTION_VAL_80_SK },
         { "90%",  OPTION_VAL_90_SK },
         { "100%", OPTION_VAL_100_O47_SK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SK },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SK },
         { "2x", OPTION_VAL_2_O26_SK },
         { "3x", OPTION_VAL_3X_SK },
         { "4x", OPTION_VAL_4_SK },
         { "5x", OPTION_VAL_5X_SK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SK },
         { "20%",  OPTION_VAL_20_SK },
         { "30%",  OPTION_VAL_30_SK },
         { "40%",  OPTION_VAL_40_SK },
         { "50%",  OPTION_VAL_50_SK },
         { "60%",  OPTION_VAL_60_SK },
         { "70%",  OPTION_VAL_70_SK },
         { "80%",  OPTION_VAL_80_SK },
         { "90%",  OPTION_VAL_90_SK },
         { "100%", OPTION_VAL_100_O47_SK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SK },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SK },
         { "2x", OPTION_VAL_2_O26_SK },
         { "3x", OPTION_VAL_3X_SK },
         { "4x", OPTION_VAL_4_SK },
         { "5x", OPTION_VAL_5X_SK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SK },
         { "20%",  OPTION_VAL_20_SK },
         { "30%",  OPTION_VAL_30_SK },
         { "40%",  OPTION_VAL_40_SK },
         { "50%",  OPTION_VAL_50_SK },
         { "60%",  OPTION_VAL_60_SK },
         { "70%",  OPTION_VAL_70_SK },
         { "80%",  OPTION_VAL_80_SK },
         { "90%",  OPTION_VAL_90_SK },
         { "100%", OPTION_VAL_100_O47_SK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SK },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SK },
         { "2x", OPTION_VAL_2_O26_SK },
         { "3x", OPTION_VAL_3X_SK },
         { "4x", OPTION_VAL_4_SK },
         { "5x", OPTION_VAL_5X_SK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SK },
         { "BLUE 03",           OPTION_VAL_BLUE_SK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SK },
         { "GREEN 05",          OPTION_VAL_GREEN_SK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SK },
         { "RED 11",            OPTION_VAL_RED_SK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SK },
         { "WHITE 28",          OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_SK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SK },
         { "20%",  OPTION_VAL_20_SK },
         { "30%",  OPTION_VAL_30_SK },
         { "40%",  OPTION_VAL_40_SK },
         { "50%",  OPTION_VAL_50_SK },
         { "60%",  OPTION_VAL_60_SK },
         { "70%",  OPTION_VAL_70_SK },
         { "80%",  OPTION_VAL_80_SK },
         { "90%",  OPTION_VAL_90_SK },
         { "100%", OPTION_VAL_100_O47_SK },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_sk = {
   option_cats_sk,
   option_defs_sk
};

/* RETRO_LANGUAGE_SR */

#define CATEGORY_SYSTEM_LABEL_SR NULL
#define CATEGORY_SYSTEM_INFO_0_SR NULL
#define CATEGORY_VIDEO_LABEL_SR NULL
#define CATEGORY_VIDEO_INFO_0_SR NULL
#define CATEGORY_PERFORMANCE_LABEL_SR NULL
#define CATEGORY_PERFORMANCE_INFO_0_SR NULL
#define CATEGORY_HACKS_LABEL_SR NULL
#define CATEGORY_HACKS_INFO_0_SR NULL
#define CATEGORY_INPUT_LABEL_SR "Ulaz"
#define CATEGORY_INPUT_INFO_0_SR NULL
#define CATEGORY_EXPANSIONS_LABEL_SR NULL
#define CATEGORY_EXPANSIONS_INFO_0_SR NULL
#define CATEGORY_VMU_LABEL_SR NULL
#define CATEGORY_VMU_INFO_0_SR NULL
#define CORE_OPTION_NAME_REGION_LABEL_SR NULL
#define OPTION_VAL_JAPAN_SR NULL
#define OPTION_VAL_USA_SR NULL
#define OPTION_VAL_EUROPE_SR NULL
#define OPTION_VAL_DEFAULT_SR NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_SR NULL
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_SR NULL
#define OPTION_VAL_JAPANESE_SR NULL
#define OPTION_VAL_ENGLISH_SR NULL
#define OPTION_VAL_GERMAN_SR NULL
#define OPTION_VAL_FRENCH_SR NULL
#define OPTION_VAL_SPANISH_SR NULL
#define OPTION_VAL_ITALIAN_SR NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_SR NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_SR NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_SR NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_SR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_SR NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_SR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_SR NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_SR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_SR NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_SR NULL
#define CORE_OPTION_NAME_UPNP_LABEL_SR NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_SR NULL
#define CORE_OPTION_NAME_DCNET_LABEL_SR NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_SR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_SR NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_SR NULL
#define OPTION_VAL_320X240_SR NULL
#define OPTION_VAL_640X480_SR NULL
#define OPTION_VAL_800X600_SR NULL
#define OPTION_VAL_960X720_SR NULL
#define OPTION_VAL_1024X768_SR NULL
#define OPTION_VAL_1280X960_SR NULL
#define OPTION_VAL_1440X1080_SR NULL
#define OPTION_VAL_1600X1200_SR NULL
#define OPTION_VAL_1920X1440_SR NULL
#define OPTION_VAL_2560X1920_SR NULL
#define OPTION_VAL_2880X2160_SR NULL
#define OPTION_VAL_3200X2400_SR NULL
#define OPTION_VAL_3840X2880_SR NULL
#define OPTION_VAL_4480X3360_SR NULL
#define OPTION_VAL_5120X3840_SR NULL
#define OPTION_VAL_5760X4320_SR NULL
#define OPTION_VAL_6400X4800_SR NULL
#define OPTION_VAL_7040X5280_SR NULL
#define OPTION_VAL_7680X5760_SR NULL
#define OPTION_VAL_8320X6240_SR NULL
#define OPTION_VAL_8960X6720_SR NULL
#define OPTION_VAL_9600X7200_SR NULL
#define OPTION_VAL_10240X7680_SR NULL
#define OPTION_VAL_10880X8160_SR NULL
#define OPTION_VAL_11520X8640_SR NULL
#define OPTION_VAL_12160X9120_SR NULL
#define OPTION_VAL_12800X9600_SR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_SR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_SR NULL
#define OPTION_VAL_VGA_SR NULL
#define OPTION_VAL_TV_RGB_SR NULL
#define OPTION_VAL_TV_COMPOSITE_SR NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_SR NULL
#define OPTION_VAL_NTSC_SR NULL
#define OPTION_VAL_PAL_SR NULL
#define OPTION_VAL_PAL_N_SR NULL
#define OPTION_VAL_PAL_M_SR NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_SR "Orijentacija ekrana"
#define OPTION_VAL_HORIZONTAL_SR NULL
#define OPTION_VAL_VERTICAL_SR NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_SR NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_SR NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_SR NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_SR NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_SR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_SR NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_SR NULL
#define OPTION_VAL_512MB_SR NULL
#define OPTION_VAL_1GB_SR NULL
#define OPTION_VAL_2GB_SR NULL
#define OPTION_VAL_4GB_SR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_SR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_SR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_SR NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_SR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_SR NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_SR NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_SR NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_SR NULL
#define CORE_OPTION_NAME_FOG_LABEL_SR NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_SR NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_SR NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_SR NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_SR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_SR NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_SR NULL
#define OPTION_VAL_1_SR NULL
#define OPTION_VAL_2_SR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_SR NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_SR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_SR NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_SR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_SR NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_SR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_SR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_SR NULL
#define OPTION_VAL_2_O26_SR NULL
#define OPTION_VAL_4_SR NULL
#define OPTION_VAL_6_SR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_SR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_SR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_SR NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_SR NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_SR NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_SR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_SR NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_SR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_SR NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_SR NULL
#define OPTION_VAL_SOME_SR NULL
#define OPTION_VAL_MORE_SR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_SR NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_SR NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_SR NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_SR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_SR NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_SR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_SR NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_SR NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_SR NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_SR NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_SR NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_SR NULL
#define OPTION_VAL_100_SR NULL
#define OPTION_VAL_110_SR NULL
#define OPTION_VAL_120_SR NULL
#define OPTION_VAL_130_SR NULL
#define OPTION_VAL_140_SR NULL
#define OPTION_VAL_150_SR NULL
#define OPTION_VAL_160_SR NULL
#define OPTION_VAL_170_SR NULL
#define OPTION_VAL_180_SR NULL
#define OPTION_VAL_190_SR NULL
#define OPTION_VAL_200_SR NULL
#define OPTION_VAL_210_SR NULL
#define OPTION_VAL_220_SR NULL
#define OPTION_VAL_230_SR NULL
#define OPTION_VAL_240_SR NULL
#define OPTION_VAL_250_SR NULL
#define OPTION_VAL_260_SR NULL
#define OPTION_VAL_270_SR NULL
#define OPTION_VAL_280_SR NULL
#define OPTION_VAL_290_SR NULL
#define OPTION_VAL_300_SR NULL
#define OPTION_VAL_310_SR NULL
#define OPTION_VAL_320_SR NULL
#define OPTION_VAL_330_SR NULL
#define OPTION_VAL_340_SR NULL
#define OPTION_VAL_350_SR NULL
#define OPTION_VAL_360_SR NULL
#define OPTION_VAL_370_SR NULL
#define OPTION_VAL_380_SR NULL
#define OPTION_VAL_390_SR NULL
#define OPTION_VAL_400_SR NULL
#define OPTION_VAL_410_SR NULL
#define OPTION_VAL_420_SR NULL
#define OPTION_VAL_430_SR NULL
#define OPTION_VAL_440_SR NULL
#define OPTION_VAL_450_SR NULL
#define OPTION_VAL_460_SR NULL
#define OPTION_VAL_470_SR NULL
#define OPTION_VAL_480_SR NULL
#define OPTION_VAL_490_SR NULL
#define OPTION_VAL_500_SR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_SR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_SR NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_SR NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_SR NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_SR NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_SR NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_SR NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_SR NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_SR NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_SR NULL
#define OPTION_VAL_0_SR NULL
#define OPTION_VAL_5_SR NULL
#define OPTION_VAL_10_SR NULL
#define OPTION_VAL_15_SR NULL
#define OPTION_VAL_20_SR NULL
#define OPTION_VAL_25_SR NULL
#define OPTION_VAL_30_SR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_SR NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_SR NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_SR NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_SR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_SR NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_SR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_SR NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_SR NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_SR NULL
#define OPTION_VAL_50_SR NULL
#define OPTION_VAL_60_SR NULL
#define OPTION_VAL_70_SR NULL
#define OPTION_VAL_80_SR NULL
#define OPTION_VAL_90_SR NULL
#define OPTION_VAL_100_O47_SR NULL
#define OPTION_VAL_110_O47_SR NULL
#define OPTION_VAL_120_O47_SR NULL
#define OPTION_VAL_130_O47_SR NULL
#define OPTION_VAL_140_O47_SR NULL
#define OPTION_VAL_150_O47_SR NULL
#define OPTION_VAL_160_O47_SR NULL
#define OPTION_VAL_170_O47_SR NULL
#define OPTION_VAL_180_O47_SR NULL
#define OPTION_VAL_190_O47_SR NULL
#define OPTION_VAL_200_O47_SR NULL
#define OPTION_VAL_210_O47_SR NULL
#define OPTION_VAL_220_O47_SR NULL
#define OPTION_VAL_230_O47_SR NULL
#define OPTION_VAL_240_O47_SR NULL
#define OPTION_VAL_250_O47_SR NULL
#define OPTION_VAL_260_O47_SR NULL
#define OPTION_VAL_270_O47_SR NULL
#define OPTION_VAL_280_O47_SR NULL
#define OPTION_VAL_290_O47_SR NULL
#define OPTION_VAL_300_O47_SR NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_SR NULL
#define OPTION_VAL_WHITE_SR NULL
#define OPTION_VAL_RED_SR NULL
#define OPTION_VAL_GREEN_SR NULL
#define OPTION_VAL_BLUE_SR NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_SR NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_SR NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_SR NULL
#define OPTION_VAL_VMU_SR NULL
#define OPTION_VAL_PURUPURU_SR NULL
#define OPTION_VAL_DREAMPOTATO_SR NULL
#define OPTION_VAL_NONE_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_SR NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_SR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_SR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_SR NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_SR NULL
#define OPTION_VAL_VMU_A1_SR NULL
#define OPTION_VAL_ALL_VMUS_SR NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_SR NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_SR NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_SR NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_SR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_SR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_SR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_SR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_SR NULL
#define OPTION_VAL_UPPER_LEFT_SR NULL
#define OPTION_VAL_UPPER_RIGHT_SR NULL
#define OPTION_VAL_LOWER_LEFT_SR NULL
#define OPTION_VAL_LOWER_RIGHT_SR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_SR NULL
#define OPTION_VAL_1X_SR NULL
#define OPTION_VAL_3X_SR NULL
#define OPTION_VAL_5X_SR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_SR NULL
#define OPTION_VAL_DEFAULT_ON_00_SR NULL
#define OPTION_VAL_DEFAULT_OFF_01_SR NULL
#define OPTION_VAL_BLACK_02_SR NULL
#define OPTION_VAL_LIGHT_BLUE_04_SR NULL
#define OPTION_VAL_CYAN_06_SR NULL
#define OPTION_VAL_CYAN_BLUE_07_SR NULL
#define OPTION_VAL_LIGHT_GREEN_08_SR NULL
#define OPTION_VAL_CYAN_GREEN_09_SR NULL
#define OPTION_VAL_LIGHT_CYAN_10_SR NULL
#define OPTION_VAL_PURPLE_12_SR NULL
#define OPTION_VAL_LIGHT_PURPLE_13_SR NULL
#define OPTION_VAL_YELLOW_14_SR NULL
#define OPTION_VAL_GRAY_15_SR NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_SR NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_SR NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_SR NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_SR NULL
#define OPTION_VAL_LIGHT_RED_2_20_SR NULL
#define OPTION_VAL_MAGENTA_21_SR NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_SR NULL
#define OPTION_VAL_LIGHT_ORANGE_23_SR NULL
#define OPTION_VAL_ORANGE_24_SR NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_SR NULL
#define OPTION_VAL_LIGHT_YELLOW_26_SR NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_SR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_SR NULL
#define OPTION_VAL_40_SR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_SR NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_SR NULL

struct retro_core_option_v2_category option_cats_sr[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_SR,
      CATEGORY_SYSTEM_INFO_0_SR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_SR,
      CATEGORY_VIDEO_INFO_0_SR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_SR,
      CATEGORY_PERFORMANCE_INFO_0_SR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_SR,
      CATEGORY_HACKS_INFO_0_SR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_SR,
      CATEGORY_INPUT_INFO_0_SR
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_SR,
      CATEGORY_EXPANSIONS_INFO_0_SR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_SR,
      CATEGORY_VMU_INFO_0_SR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_sr[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_SR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_SR },
         { "USA",     OPTION_VAL_USA_SR },
         { "Europe",  OPTION_VAL_EUROPE_SR },
         { "Default", OPTION_VAL_DEFAULT_SR },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_SR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_SR },
         { "English",  OPTION_VAL_ENGLISH_SR },
         { "German",   OPTION_VAL_GERMAN_SR },
         { "French",   OPTION_VAL_FRENCH_SR },
         { "Spanish",  OPTION_VAL_SPANISH_SR },
         { "Italian",  OPTION_VAL_ITALIAN_SR },
         { "Default",  OPTION_VAL_DEFAULT_SR },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_SR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_SR },
         { "640x480",    OPTION_VAL_640X480_SR },
         { "800x600",    OPTION_VAL_800X600_SR },
         { "960x720",    OPTION_VAL_960X720_SR },
         { "1024x768",   OPTION_VAL_1024X768_SR },
         { "1280x960",   OPTION_VAL_1280X960_SR },
         { "1440x1080",  OPTION_VAL_1440X1080_SR },
         { "1600x1200",  OPTION_VAL_1600X1200_SR },
         { "1920x1440",  OPTION_VAL_1920X1440_SR },
         { "2560x1920",  OPTION_VAL_2560X1920_SR },
         { "2880x2160",  OPTION_VAL_2880X2160_SR },
         { "3200x2400",  OPTION_VAL_3200X2400_SR },
         { "3840x2880",  OPTION_VAL_3840X2880_SR },
         { "4480x3360",  OPTION_VAL_4480X3360_SR },
         { "5120x3840",  OPTION_VAL_5120X3840_SR },
         { "5760x4320",  OPTION_VAL_5760X4320_SR },
         { "6400x4800",  OPTION_VAL_6400X4800_SR },
         { "7040x5280",  OPTION_VAL_7040X5280_SR },
         { "7680x5760",  OPTION_VAL_7680X5760_SR },
         { "8320x6240",  OPTION_VAL_8320X6240_SR },
         { "8960x6720",  OPTION_VAL_8960X6720_SR },
         { "9600x7200",  OPTION_VAL_9600X7200_SR },
         { "10240x7680", OPTION_VAL_10240X7680_SR },
         { "10880x8160", OPTION_VAL_10880X8160_SR },
         { "11520x8640", OPTION_VAL_11520X8640_SR },
         { "12160x9120", OPTION_VAL_12160X9120_SR },
         { "12800x9600", OPTION_VAL_12800X9600_SR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_SR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_SR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_SR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_SR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_SR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_SR },
         { "PAL",     OPTION_VAL_PAL_SR },
         { "PAL_N",   OPTION_VAL_PAL_N_SR },
         { "PAL_M",   OPTION_VAL_PAL_M_SR },
         { "Default", OPTION_VAL_DEFAULT_SR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_SR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_SR },
         { "vertical",   OPTION_VAL_VERTICAL_SR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_SR,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_SR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_SR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_SR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_SR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_SR },
         { "1GB",   OPTION_VAL_1GB_SR },
         { "2GB",   OPTION_VAL_2GB_SR },
         { "4GB",   OPTION_VAL_4GB_SR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_SR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_SR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_SR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_SR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_SR },
         { "1",  OPTION_VAL_1_SR },
         { "2",  OPTION_VAL_2_SR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_SR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_SR },
         { "4",  OPTION_VAL_4_SR },
         { "6",  OPTION_VAL_6_SR },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_SR,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_SR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_SR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_SR,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_SR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_SR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_SR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_SR },
         { "more", OPTION_VAL_MORE_SR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_SR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_SR },
         { "110", OPTION_VAL_110_SR },
         { "120", OPTION_VAL_120_SR },
         { "130", OPTION_VAL_130_SR },
         { "140", OPTION_VAL_140_SR },
         { "150", OPTION_VAL_150_SR },
         { "160", OPTION_VAL_160_SR },
         { "170", OPTION_VAL_170_SR },
         { "180", OPTION_VAL_180_SR },
         { "190", OPTION_VAL_190_SR },
         { "200", OPTION_VAL_200_SR },
         { "210", OPTION_VAL_210_SR },
         { "220", OPTION_VAL_220_SR },
         { "230", OPTION_VAL_230_SR },
         { "240", OPTION_VAL_240_SR },
         { "250", OPTION_VAL_250_SR },
         { "260", OPTION_VAL_260_SR },
         { "270", OPTION_VAL_270_SR },
         { "280", OPTION_VAL_280_SR },
         { "290", OPTION_VAL_290_SR },
         { "300", OPTION_VAL_300_SR },
         { "310", OPTION_VAL_310_SR },
         { "320", OPTION_VAL_320_SR },
         { "330", OPTION_VAL_330_SR },
         { "340", OPTION_VAL_340_SR },
         { "350", OPTION_VAL_350_SR },
         { "360", OPTION_VAL_360_SR },
         { "370", OPTION_VAL_370_SR },
         { "380", OPTION_VAL_380_SR },
         { "390", OPTION_VAL_390_SR },
         { "400", OPTION_VAL_400_SR },
         { "410", OPTION_VAL_410_SR },
         { "420", OPTION_VAL_420_SR },
         { "430", OPTION_VAL_430_SR },
         { "440", OPTION_VAL_440_SR },
         { "450", OPTION_VAL_450_SR },
         { "460", OPTION_VAL_460_SR },
         { "470", OPTION_VAL_470_SR },
         { "480", OPTION_VAL_480_SR },
         { "490", OPTION_VAL_490_SR },
         { "500", OPTION_VAL_500_SR },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_SR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_SR },
         { "5%",  OPTION_VAL_5_SR },
         { "10%", OPTION_VAL_10_SR },
         { "15%", OPTION_VAL_15_SR },
         { "20%", OPTION_VAL_20_SR },
         { "25%", OPTION_VAL_25_SR },
         { "30%", OPTION_VAL_30_SR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_SR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_SR },
         { "5%",  OPTION_VAL_5_SR },
         { "10%", OPTION_VAL_10_SR },
         { "15%", OPTION_VAL_15_SR },
         { "20%", OPTION_VAL_20_SR },
         { "25%", OPTION_VAL_25_SR },
         { "30%", OPTION_VAL_30_SR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_SR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_SR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_SR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_SR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_SR },
         { "60%",  OPTION_VAL_60_SR },
         { "70%",  OPTION_VAL_70_SR },
         { "80%",  OPTION_VAL_80_SR },
         { "90%",  OPTION_VAL_90_SR },
         { "100%", OPTION_VAL_100_O47_SR },
         { "110%", OPTION_VAL_110_O47_SR },
         { "120%", OPTION_VAL_120_O47_SR },
         { "130%", OPTION_VAL_130_O47_SR },
         { "140%", OPTION_VAL_140_O47_SR },
         { "150%", OPTION_VAL_150_O47_SR },
         { "160%", OPTION_VAL_160_O47_SR },
         { "170%", OPTION_VAL_170_O47_SR },
         { "180%", OPTION_VAL_180_O47_SR },
         { "190%", OPTION_VAL_190_O47_SR },
         { "200%", OPTION_VAL_200_O47_SR },
         { "210%", OPTION_VAL_210_O47_SR },
         { "220%", OPTION_VAL_220_O47_SR },
         { "230%", OPTION_VAL_230_O47_SR },
         { "240%", OPTION_VAL_240_O47_SR },
         { "250%", OPTION_VAL_250_O47_SR },
         { "260%", OPTION_VAL_260_O47_SR },
         { "270%", OPTION_VAL_270_O47_SR },
         { "280%", OPTION_VAL_280_O47_SR },
         { "290%", OPTION_VAL_290_O47_SR },
         { "300%", OPTION_VAL_300_O47_SR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_SR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SR },
         { "Red",      OPTION_VAL_RED_SR },
         { "Green",    OPTION_VAL_GREEN_SR },
         { "Blue",     OPTION_VAL_BLUE_SR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_SR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SR },
         { "Red",      OPTION_VAL_RED_SR },
         { "Green",    OPTION_VAL_GREEN_SR },
         { "Blue",     OPTION_VAL_BLUE_SR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_SR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SR },
         { "Red",      OPTION_VAL_RED_SR },
         { "Green",    OPTION_VAL_GREEN_SR },
         { "Blue",     OPTION_VAL_BLUE_SR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_SR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SR },
         { "Red",      OPTION_VAL_RED_SR },
         { "Green",    OPTION_VAL_GREEN_SR },
         { "Blue",     OPTION_VAL_BLUE_SR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_SR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SR },
         { "Purupuru", OPTION_VAL_PURUPURU_SR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_SR },
         { "None",     OPTION_VAL_NONE_SR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_SR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SR },
         { "Purupuru", OPTION_VAL_PURUPURU_SR },
         { "None",     OPTION_VAL_NONE_SR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_SR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SR },
         { "Purupuru", OPTION_VAL_PURUPURU_SR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_SR },
         { "None",     OPTION_VAL_NONE_SR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_SR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SR },
         { "Purupuru", OPTION_VAL_PURUPURU_SR },
         { "None",     OPTION_VAL_NONE_SR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_SR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SR },
         { "Purupuru", OPTION_VAL_PURUPURU_SR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_SR },
         { "None",     OPTION_VAL_NONE_SR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_SR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SR },
         { "Purupuru", OPTION_VAL_PURUPURU_SR },
         { "None",     OPTION_VAL_NONE_SR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_SR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SR },
         { "Purupuru", OPTION_VAL_PURUPURU_SR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_SR },
         { "None",     OPTION_VAL_NONE_SR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_SR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SR },
         { "Purupuru", OPTION_VAL_PURUPURU_SR },
         { "None",     OPTION_VAL_NONE_SR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_SR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_SR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_SR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_SR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_SR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_SR,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_SR,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_SR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_SR,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_SR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_SR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_SR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_SR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SR },
         { "2x", OPTION_VAL_2_O26_SR },
         { "3x", OPTION_VAL_3X_SR },
         { "4x", OPTION_VAL_4_SR },
         { "5x", OPTION_VAL_5X_SR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SR },
         { "20%",  OPTION_VAL_20_SR },
         { "30%",  OPTION_VAL_30_SR },
         { "40%",  OPTION_VAL_40_SR },
         { "50%",  OPTION_VAL_50_SR },
         { "60%",  OPTION_VAL_60_SR },
         { "70%",  OPTION_VAL_70_SR },
         { "80%",  OPTION_VAL_80_SR },
         { "90%",  OPTION_VAL_90_SR },
         { "100%", OPTION_VAL_100_O47_SR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SR },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SR },
         { "2x", OPTION_VAL_2_O26_SR },
         { "3x", OPTION_VAL_3X_SR },
         { "4x", OPTION_VAL_4_SR },
         { "5x", OPTION_VAL_5X_SR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SR },
         { "20%",  OPTION_VAL_20_SR },
         { "30%",  OPTION_VAL_30_SR },
         { "40%",  OPTION_VAL_40_SR },
         { "50%",  OPTION_VAL_50_SR },
         { "60%",  OPTION_VAL_60_SR },
         { "70%",  OPTION_VAL_70_SR },
         { "80%",  OPTION_VAL_80_SR },
         { "90%",  OPTION_VAL_90_SR },
         { "100%", OPTION_VAL_100_O47_SR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SR },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SR },
         { "2x", OPTION_VAL_2_O26_SR },
         { "3x", OPTION_VAL_3X_SR },
         { "4x", OPTION_VAL_4_SR },
         { "5x", OPTION_VAL_5X_SR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SR },
         { "20%",  OPTION_VAL_20_SR },
         { "30%",  OPTION_VAL_30_SR },
         { "40%",  OPTION_VAL_40_SR },
         { "50%",  OPTION_VAL_50_SR },
         { "60%",  OPTION_VAL_60_SR },
         { "70%",  OPTION_VAL_70_SR },
         { "80%",  OPTION_VAL_80_SR },
         { "90%",  OPTION_VAL_90_SR },
         { "100%", OPTION_VAL_100_O47_SR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SR },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SR },
         { "2x", OPTION_VAL_2_O26_SR },
         { "3x", OPTION_VAL_3X_SR },
         { "4x", OPTION_VAL_4_SR },
         { "5x", OPTION_VAL_5X_SR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SR },
         { "BLUE 03",           OPTION_VAL_BLUE_SR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SR },
         { "GREEN 05",          OPTION_VAL_GREEN_SR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SR },
         { "RED 11",            OPTION_VAL_RED_SR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SR },
         { "WHITE 28",          OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_SR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SR },
         { "20%",  OPTION_VAL_20_SR },
         { "30%",  OPTION_VAL_30_SR },
         { "40%",  OPTION_VAL_40_SR },
         { "50%",  OPTION_VAL_50_SR },
         { "60%",  OPTION_VAL_60_SR },
         { "70%",  OPTION_VAL_70_SR },
         { "80%",  OPTION_VAL_80_SR },
         { "90%",  OPTION_VAL_90_SR },
         { "100%", OPTION_VAL_100_O47_SR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_sr = {
   option_cats_sr,
   option_defs_sr
};

/* RETRO_LANGUAGE_SV */

#define CATEGORY_SYSTEM_LABEL_SV NULL
#define CATEGORY_SYSTEM_INFO_0_SV "Konfigurera region, språk, BIOS och grundläggande hårdvaruinställningar."
#define CATEGORY_VIDEO_LABEL_SV NULL
#define CATEGORY_VIDEO_INFO_0_SV "Konfigurera upplösning, oberoende transparens i ordning och visuella effektinställningar."
#define CATEGORY_PERFORMANCE_LABEL_SV "Prestanda"
#define CATEGORY_PERFORMANCE_INFO_0_SV "Konfigurera inställningar för trådad rendering och bildruteskippning."
#define CATEGORY_HACKS_LABEL_SV "Emulatorhack"
#define CATEGORY_HACKS_INFO_0_SV "Konfigurera inställningar för widescreen-överskridanden, GD-ROM-laddningshastighet och texturrelaterade inställningar."
#define CATEGORY_INPUT_LABEL_SV "Inmatning"
#define CATEGORY_INPUT_INFO_0_SV "Konfigurera inställningar för spelkontrollern och ljuspistolen."
#define CATEGORY_EXPANSIONS_LABEL_SV "Expansionsplatser för kontroller"
#define CATEGORY_EXPANSIONS_INFO_0_SV "Välj enheten (VMU, rumble-enhet) ansluten i varje kontrollers expansionsplats."
#define CATEGORY_VMU_LABEL_SV "Visuell minnesenhet"
#define CATEGORY_VMU_INFO_0_SV "Konfigurera per-spel VMU-sparfiler och inställningar för synlighet av VMU på skärmen."
#define CORE_OPTION_NAME_REGION_LABEL_SV NULL
#define OPTION_VAL_JAPAN_SV NULL
#define OPTION_VAL_USA_SV NULL
#define OPTION_VAL_EUROPE_SV "Europa"
#define OPTION_VAL_DEFAULT_SV "Standard"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_SV "Språk"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_SV "Ändrar språket som används av BIOS och av eventuella spel som innehåller flera språk."
#define OPTION_VAL_JAPANESE_SV "Japanska"
#define OPTION_VAL_ENGLISH_SV "Engelska"
#define OPTION_VAL_GERMAN_SV "Tyska"
#define OPTION_VAL_FRENCH_SV "Franska"
#define OPTION_VAL_SPANISH_SV "Spanska"
#define OPTION_VAL_ITALIAN_SV "Italienska"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_SV "HLE BIOS (omstart krävs)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_SV "Tvinga användning av högnivåemulering av BIOS."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_SV "Aktivera DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_SV "Aktivera emulering av Dreamcasts ljud-DSP (digital signal processor). Förbättrar noggrannheten i genererat ljud, men ökar prestandakraven."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_SV "Tillåt knappar för arkadservice"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_SV "Aktiverar SERVICE-knapp för arkadspel för att ange kabinettinställningar."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_SV "Ställ in NAOMI-spel till Free Play"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_SV "Ändra spelets myntinställningar till free play."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_SV "Emulering av bredbandsadapter"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_SV "Emulera Ethernet-bredbandsadaptern istället för modemet. (Omstart krävs)"
#define CORE_OPTION_NAME_UPNP_LABEL_SV "Aktivera UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_SV "Använd UPnP för att automatiskt konfigurera din internetrouter för onlinespel."
#define CORE_OPTION_NAME_DCNET_LABEL_SV "Använd DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_SV "Använd DCNet-molntjänsten för Dreamcasts internetåtkomst."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_SV "Intern upplösning"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_SV "Ändra renderingsupplösning."
#define OPTION_VAL_320X240_SV "320x240 (Halv)"
#define OPTION_VAL_640X480_SV "640x480 (Inbyggd)"
#define OPTION_VAL_800X600_SV NULL
#define OPTION_VAL_960X720_SV NULL
#define OPTION_VAL_1024X768_SV NULL
#define OPTION_VAL_1280X960_SV NULL
#define OPTION_VAL_1440X1080_SV NULL
#define OPTION_VAL_1600X1200_SV NULL
#define OPTION_VAL_1920X1440_SV NULL
#define OPTION_VAL_2560X1920_SV NULL
#define OPTION_VAL_2880X2160_SV NULL
#define OPTION_VAL_3200X2400_SV NULL
#define OPTION_VAL_3840X2880_SV NULL
#define OPTION_VAL_4480X3360_SV NULL
#define OPTION_VAL_5120X3840_SV NULL
#define OPTION_VAL_5760X4320_SV NULL
#define OPTION_VAL_6400X4800_SV NULL
#define OPTION_VAL_7040X5280_SV NULL
#define OPTION_VAL_7680X5760_SV NULL
#define OPTION_VAL_8320X6240_SV NULL
#define OPTION_VAL_8960X6720_SV NULL
#define OPTION_VAL_9600X7200_SV NULL
#define OPTION_VAL_10240X7680_SV NULL
#define OPTION_VAL_10880X8160_SV NULL
#define OPTION_VAL_11520X8640_SV NULL
#define OPTION_VAL_12160X9120_SV NULL
#define OPTION_VAL_12800X9600_SV NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_SV "Kabeltyp"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_SV "Utgångssignalstyp. \"TV (Komposit)\" är den mest välkomna."
#define OPTION_VAL_VGA_SV NULL
#define OPTION_VAL_TV_RGB_SV NULL
#define OPTION_VAL_TV_COMPOSITE_SV "TV (Komposit)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_SV "Sändningsstandard"
#define OPTION_VAL_NTSC_SV NULL
#define OPTION_VAL_PAL_SV "PAL (Världen)"
#define OPTION_VAL_PAL_N_SV NULL
#define OPTION_VAL_PAL_M_SV "PAL-M (Brasilien)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_SV "Skärmorientering"
#define OPTION_VAL_HORIZONTAL_SV "Horisontell"
#define OPTION_VAL_VERTICAL_SV "Vertikal"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_SV "Alfasortering"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_SV "Välj hur transparenta polygoner sorteras."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_SV "Per-Strip (snabbast, minst noggrann)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_SV "Per-Triangel (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_SV "Per-Pixel (noggrann, men långsammast)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_SV "Storlek på ackumulerande pixelbuffert"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_SV "Högre värden kan vara nödvändiga för att korrekt visa utdata i högre upplösningar."
#define OPTION_VAL_512MB_SV NULL
#define OPTION_VAL_1GB_SV NULL
#define OPTION_VAL_2GB_SV NULL
#define OPTION_VAL_4GB_SV NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_SV "Maximalt antal transparenta lager"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_SV "Högre värden kan krävas för komplexa scener."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_SV "Fullständig framebuffer-emulering"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_SV "Aktivera fullständig framebuffer-emulering i VRAM. Detta är användbart för spel som läser eller skriver direkt till framebuffer i VRAM. När det är aktiverat tvingas intern upplösning till 640x480 och prestanda kan påverkas allvarligt."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_SV "Aktivera RTT (Render To Texture) buffert"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_SV "Kopiera renderade texturer tillbaka från GPU till VRAM. Denna inställning är normalt aktiverad för spel som kräver det. När den är aktiverad avaktiveras uppskalning av texturrendering och prestandan kan påverkas."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_SV "Mipmappning"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_SV "När aktiverade texturer använder mindre versioner av sig själva när de visas längre bort kan det öka prestandan och minska skimret."
#define CORE_OPTION_NAME_FOG_LABEL_SV "Dimeffekter"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_SV "Volymjusterare"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_SV "En Dreamcast GPU-funktion som vanligtvis används av spel för att rita objektsskuggor. Detta bör normalt vara aktiverat - prestandapåverkan är vanligtvis minimal till försumbar."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_SV "Anisotropisk Filtering"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_SV "Förbättrar kvaliteten på texturer på ytor som är i sneda betraktningsvinklar i förhållande till kameran. Högre värden kräver mer av GPU:n. Ändringar av denna inställning gäller endast efter omstart."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_SV "Texturfiltrering"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_SV "Texturfiltreringsläget som ska användas. Detta kan användas för att tvinga ett visst texturfiltreringsläge på alla texturer för att få en skarpare (eller slätare) utseende än standard. Värden som skiljer sig från standard kan orsaka olika renderingsproblem. Ändringar av denna inställning gäller endast efter omstart."
#define OPTION_VAL_1_SV "Tvinga Närmsta granne"
#define OPTION_VAL_2_SV "Tvinga Linjär"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_SV "Fördröj bildrutebyte"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_SV "Användbart för att undvika blinkande skärmar eller glitchiga videor. Rekommenderas inte på långsamma plattformar."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_SV "Upptäck ändringar i bildhastighet"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_SV "Meddela gränssnittet när den interna bildrutefrekvensen ändras (till exempel från 60 fps till 30 fps). Förbättrar taktningen av bildfrekvensen i spel som körs med en låst hastighet på 30 fps eller 20 fps, men bör vara inaktiverad för spel med upplåsta (ostadiga) bildrutefrekvenser (till exempel Ecco the Dolphin, Unreal Tournament). Observera!: Ej tillgänglig när \"Auto Skip Frame\" är aktiverad."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_SV "PowerVR2 Efterbehandlingsfilter"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_SV "Efterbehandla den renderade bilden för att simulera effekter som är specifika för PowerVR2 GPU:n och analoga videosignaler."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_SV "Texturuppskalning (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_SV "Förbättrar handritade 2D-pixelgrafik. Ska endast användas med 2D-pixeliserade spel."
#define OPTION_VAL_2_O26_SV NULL
#define OPTION_VAL_4_SV NULL
#define OPTION_VAL_6_SV NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_SV "Maximalt filtrerad storlek för texturuppskalning"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_SV "Välj ett maximalt storleksvärde för en textur som ska skalas upp, om texturstorleken är högre än det valda värdet kommer den inte att skalas upp."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_SV "Ursprunglig djupinterpolation"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_SV "Hjälper till med texturkorruption och djupproblem på AMD-GPU:n. Kan även hjälpa Intel-GPU:n i vissa fall."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_SV "Rätta till blödande kanter vid uppskalning"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_SV "Hjälper till med texturblödning vid uppskalning. Om du inaktiverar det kan det hjälpa om bildpunkter förvrängs vid uppskalning i 2D-spel (MVC2, CVS, KOF, etc.)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_SV "Trådad rendering"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_SV "Kör GPU och CPU på olika trådar. Starkt rekommenderat."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_SV "Hoppa över bildrutor automatiskt"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_SV "Hoppa automatiskt över bildrutor när emuleringen går långsamt. Observera!: Denna inställning gäller endast när \"Trådad Rendering\" är aktiverad."
#define OPTION_VAL_SOME_SV NULL
#define OPTION_VAL_MORE_SV "Maximalt"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_SV "Bildruteskippning"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_SV "Sätter antalet bildrutor att hoppa över mellan varje visad bildruta."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_SV "Bredbildsfusk (omstart krävs)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_SV "Aktiverar fusk som gör det möjligt för vissa spel att visas i bredbildsformat."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_SV "Bredbildshackning"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_SV "Ritar geometrin utanför den normala 4:3 bildformatet. Kan leda till grafiska fel i de avslöjade områdena."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_SV "GD-ROM Snabbladdning (felaktig)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_SV "Ökar hastigheten på GD-ROM laddningen."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_SV "\"Dreamcast 32MB RAM-modifikation\""
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_SV "Aktiverar 32MB RAM-modifikation för Dreamcast. Kan påverka kompatibiliteten."
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_SV "SH4 CPU under/överklockning"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_SV "Ändra SH4 huvud-CPU-klockan från standarden 200 MHz. Att sänka klockfrekvensen kan hjälpa på långsamma plattformar. Att öka klockfrekvensen kan öka bildhastigheten i vissa spel. Använd på egen risk."
#define OPTION_VAL_100_SV NULL
#define OPTION_VAL_110_SV NULL
#define OPTION_VAL_120_SV NULL
#define OPTION_VAL_130_SV NULL
#define OPTION_VAL_140_SV NULL
#define OPTION_VAL_150_SV NULL
#define OPTION_VAL_160_SV NULL
#define OPTION_VAL_170_SV NULL
#define OPTION_VAL_180_SV NULL
#define OPTION_VAL_190_SV NULL
#define OPTION_VAL_200_SV NULL
#define OPTION_VAL_210_SV NULL
#define OPTION_VAL_220_SV NULL
#define OPTION_VAL_230_SV NULL
#define OPTION_VAL_240_SV NULL
#define OPTION_VAL_250_SV NULL
#define OPTION_VAL_260_SV NULL
#define OPTION_VAL_270_SV NULL
#define OPTION_VAL_280_SV NULL
#define OPTION_VAL_290_SV NULL
#define OPTION_VAL_300_SV NULL
#define OPTION_VAL_310_SV NULL
#define OPTION_VAL_320_SV NULL
#define OPTION_VAL_330_SV NULL
#define OPTION_VAL_340_SV NULL
#define OPTION_VAL_350_SV NULL
#define OPTION_VAL_360_SV NULL
#define OPTION_VAL_370_SV NULL
#define OPTION_VAL_380_SV NULL
#define OPTION_VAL_390_SV NULL
#define OPTION_VAL_400_SV NULL
#define OPTION_VAL_410_SV NULL
#define OPTION_VAL_420_SV NULL
#define OPTION_VAL_430_SV NULL
#define OPTION_VAL_440_SV NULL
#define OPTION_VAL_450_SV NULL
#define OPTION_VAL_460_SV NULL
#define OPTION_VAL_470_SV NULL
#define OPTION_VAL_480_SV NULL
#define OPTION_VAL_490_SV NULL
#define OPTION_VAL_500_SV NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_SV "Läs in anpassade texturer"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_SV "Läs in anpassade texturer som finns i mappen 'system/dc/textures/<game-id>/'."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_SV "Förinläs anpassade texturer"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_SV "Förinläs anpassade texturer vid spelstart. Kan förbättra prestanda men ökar minnesanvändningen."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_SV "Dumpa texturer"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_SV "Varje gång en ny textur används av spelet sparas den som en .png-fil i mappen 'system/dc/texdump/<game-id>/'."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_SV "Dumpa ersatta texturer"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_SV "Dumpa alltid texturer som redan ersatts av anpassade texturer."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_SV "Analoga spakens dödzon"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_SV "Välj hur långt du måste flytta på den analoga spaken innan den börjar bearbetas."
#define OPTION_VAL_0_SV NULL
#define OPTION_VAL_5_SV NULL
#define OPTION_VAL_10_SV NULL
#define OPTION_VAL_15_SV NULL
#define OPTION_VAL_20_SV NULL
#define OPTION_VAL_25_SV NULL
#define OPTION_VAL_30_SV NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_SV "Dödläge för avtryckare"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_SV "Välj hur mycket du måste trycka på avtryckaren innan den börjar bearbetas."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_SV "Digitala avtryckare"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_SV "När avtryckare är aktiverade fungerar de som vanliga knappar, vilket innebär att de behandlas som antingen helt intryckta eller inte intryckta alls, utan något mellanting."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_SV "Sänd ut digitala utgångar"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_SV "Sänd digitala utgångar och force-feedbackstatus på TCP-port 8000. Kompatibelt med MAME-alternativet \"-output network\"."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_SV "Visa ljuspistolinställningar"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_SV "Aktivera konfiguration av alternativ för ljuspistolens hårkors. OBS!: Snabbmenyn kan behöva aktiveras och avaktiveras för att denna inställning ska träda i kraft."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_SV "Storleksskalning för pistolhårkors"
#define OPTION_VAL_50_SV NULL
#define OPTION_VAL_60_SV NULL
#define OPTION_VAL_70_SV NULL
#define OPTION_VAL_80_SV NULL
#define OPTION_VAL_90_SV NULL
#define OPTION_VAL_100_O47_SV NULL
#define OPTION_VAL_110_O47_SV NULL
#define OPTION_VAL_120_O47_SV NULL
#define OPTION_VAL_130_O47_SV NULL
#define OPTION_VAL_140_O47_SV NULL
#define OPTION_VAL_150_O47_SV NULL
#define OPTION_VAL_160_O47_SV NULL
#define OPTION_VAL_170_O47_SV NULL
#define OPTION_VAL_180_O47_SV NULL
#define OPTION_VAL_190_O47_SV NULL
#define OPTION_VAL_200_O47_SV NULL
#define OPTION_VAL_210_O47_SV NULL
#define OPTION_VAL_220_O47_SV NULL
#define OPTION_VAL_230_O47_SV NULL
#define OPTION_VAL_240_O47_SV NULL
#define OPTION_VAL_250_O47_SV NULL
#define OPTION_VAL_260_O47_SV NULL
#define OPTION_VAL_270_O47_SV NULL
#define OPTION_VAL_280_O47_SV NULL
#define OPTION_VAL_290_O47_SV NULL
#define OPTION_VAL_300_O47_SV NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_SV "Visa hårkors för pistol 1"
#define OPTION_VAL_WHITE_SV "Vit"
#define OPTION_VAL_RED_SV "Röd"
#define OPTION_VAL_GREEN_SV "Grön"
#define OPTION_VAL_BLUE_SV "Blå"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_SV "Visa hårkors för pistol 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_SV "Visa hårkors för pistol 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_SV "Visa hårkors för pistol 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_SV "Enhet i expansionsfack A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_SV "Välj den enhet som är ansluten till expansionsfack A1 (port A plats 1)."
#define OPTION_VAL_VMU_SV NULL
#define OPTION_VAL_PURUPURU_SV "Vibrationspaket"
#define OPTION_VAL_DREAMPOTATO_SV NULL
#define OPTION_VAL_NONE_SV "Ingen"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_SV "Enhet i expansionsfack A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_SV "Välj den enhet som är ansluten till expansionsfack A2 (port A plats 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_SV "Enhet i expansionsfack B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_SV "Välj den enhet som är ansluten till expansionsfack B1 (port B plats 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_SV "Enhet i expansionsfack B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_SV "Välj den enhet som är ansluten till expansionsfack B2 (port B plats 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_SV "Enhet i expansionsfack C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_SV "Välj den enhet som är ansluten till expansionsfack C1 (port C plats 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_SV "Enhet i expansionsfack C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_SV "Välj den enhet som är ansluten till expansionsfack C2 (port C plats 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_SV "Enhet i expansionsfack D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_SV "Välj den enhet som är ansluten till expansionsfack D1 (port D plats 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_SV "Enhet i expansionsfack D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_SV "Välj den enhet som är ansluten till expansionsfack D2 (port D plats 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_SV "Per-spel visningsminnesenheter/systemer (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_SV "Per-spel VMU:er"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_SV "När den är inaktiverad delar alla spel upp till 8 VMU-sparfiler (A1/A2/B1/B2/C1/C2/D1/D2) som finns i RetroArchs systemmapp.\nInställningen ”VMU A1” skapar en unik VMU-fil ”A1” i RetroArchs sparmapp för varje spel som startas.\nInställningen ”Alla VMU: er” skapar upp till 8 unika VMU-filer (A1/A2/B1/B2/C1/C2/D1/D2) för varje spel som startas."
#define OPTION_VAL_VMU_A1_SV NULL
#define OPTION_VAL_ALL_VMUS_SV "Alla VMU:er"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_SV "Ljud för visuella minnesenheter/System (VMU)"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_SV "VMU-ljud"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_SV "När detta är aktiverad, spelas VMU-pipen."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_SV "DreamPotato VMU-lagring"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_SV "Aktivera lagring av VMU-data på den anslutna DreamPotato-emulatorn. När denna funktion är aktiverad används inte den lokala VMU-sparfil."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_SV "Visa inställningar för Visual Memory Unit/System (VMU)-skärmen"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_SV "Visa VMU-skärminställningar"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_SV "Möjliggör konfiguration av emulerad VMU LCD-skärmens synlighet, storlek, position och färg. OBS!: Snabbmenyn kan behöva aktiveras/dras fram för att den här inställningen ska träda i kraft."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_SV "Visa VMU-skärm 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_SV "Position för VMU-skärm 1"
#define OPTION_VAL_UPPER_LEFT_SV "Övre vänstra"
#define OPTION_VAL_UPPER_RIGHT_SV "Övre högra"
#define OPTION_VAL_LOWER_LEFT_SV "Nedre vänstra"
#define OPTION_VAL_LOWER_RIGHT_SV "Nedre högra"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_SV "Storlek för VMU-skärm 1"
#define OPTION_VAL_1X_SV NULL
#define OPTION_VAL_3X_SV NULL
#define OPTION_VAL_5X_SV NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_SV "VMU Skärm 1 Pixel PÅ Färg"
#define OPTION_VAL_DEFAULT_ON_00_SV "Standard PÅ"
#define OPTION_VAL_DEFAULT_OFF_01_SV "Standard AV"
#define OPTION_VAL_BLACK_02_SV "Svart"
#define OPTION_VAL_LIGHT_BLUE_04_SV "Ljusblå"
#define OPTION_VAL_CYAN_06_SV NULL
#define OPTION_VAL_CYAN_BLUE_07_SV "Turkosblå"
#define OPTION_VAL_LIGHT_GREEN_08_SV "Ljusgrön"
#define OPTION_VAL_CYAN_GREEN_09_SV "Turkosgrön"
#define OPTION_VAL_LIGHT_CYAN_10_SV "Ljusturkos"
#define OPTION_VAL_PURPLE_12_SV "Lila"
#define OPTION_VAL_LIGHT_PURPLE_13_SV "Ljuslila"
#define OPTION_VAL_YELLOW_14_SV "Gul"
#define OPTION_VAL_GRAY_15_SV "Grå"
#define OPTION_VAL_LIGHT_PURPLE_2_16_SV "Ljuslila (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_SV "Ljusgrön (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_SV "Ljusgrön (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_SV "Ljusturkos (2)"
#define OPTION_VAL_LIGHT_RED_2_20_SV "Ljusröd (2)"
#define OPTION_VAL_MAGENTA_21_SV NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_SV "Ljuslila (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_SV "Ljusorange"
#define OPTION_VAL_ORANGE_24_SV NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_SV "Ljuslila (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_SV "Ljusgul"
#define OPTION_VAL_LIGHT_YELLOW_2_27_SV "Ljusgul (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_SV "VMU Skärm 1 Pixel AV Färg"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_SV "VMU Skärm 1 Opacitet"
#define OPTION_VAL_40_SV NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_SV "VMU Skärm 2 Visa"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_SV "VMU Skärm 2 Position"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_SV "VMU Skärm 2 Storlek"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_SV "VMU Skärm 2 Pixel PÅ Färg"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_SV "VMU Skärm 2 Pixel AV Färg"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_SV "VMU Skärm 2 Opacitet"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_SV "Visa VMU skärm 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_SV "Positionen för VMU skärm 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_SV "VMU skärm 3 Storlek"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_SV "Aktivera pixelfärg på VMU skärm 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_SV "Avaktivera pixelfärg på VMU skärm 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_SV "Opaciteten för VMU Skärm 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_SV "Visa VMU skärm 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_SV "Positionen för VMU skärm 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_SV "VMU skärm 4 Storlek"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_SV "VMU Skärm 4 Pixel PÅ Färg"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_SV "VMU Skärm 4 Pixel AV Färg"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_SV "VMU Skärm 4 Opacitet"

struct retro_core_option_v2_category option_cats_sv[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_SV,
      CATEGORY_SYSTEM_INFO_0_SV
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_SV,
      CATEGORY_VIDEO_INFO_0_SV
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_SV,
      CATEGORY_PERFORMANCE_INFO_0_SV
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_SV,
      CATEGORY_HACKS_INFO_0_SV
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_SV,
      CATEGORY_INPUT_INFO_0_SV
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_SV,
      CATEGORY_EXPANSIONS_INFO_0_SV
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_SV,
      CATEGORY_VMU_INFO_0_SV
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_sv[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_SV,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_SV },
         { "USA",     OPTION_VAL_USA_SV },
         { "Europe",  OPTION_VAL_EUROPE_SV },
         { "Default", OPTION_VAL_DEFAULT_SV },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_SV,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_SV },
         { "English",  OPTION_VAL_ENGLISH_SV },
         { "German",   OPTION_VAL_GERMAN_SV },
         { "French",   OPTION_VAL_FRENCH_SV },
         { "Spanish",  OPTION_VAL_SPANISH_SV },
         { "Italian",  OPTION_VAL_ITALIAN_SV },
         { "Default",  OPTION_VAL_DEFAULT_SV },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_SV,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_SV },
         { "640x480",    OPTION_VAL_640X480_SV },
         { "800x600",    OPTION_VAL_800X600_SV },
         { "960x720",    OPTION_VAL_960X720_SV },
         { "1024x768",   OPTION_VAL_1024X768_SV },
         { "1280x960",   OPTION_VAL_1280X960_SV },
         { "1440x1080",  OPTION_VAL_1440X1080_SV },
         { "1600x1200",  OPTION_VAL_1600X1200_SV },
         { "1920x1440",  OPTION_VAL_1920X1440_SV },
         { "2560x1920",  OPTION_VAL_2560X1920_SV },
         { "2880x2160",  OPTION_VAL_2880X2160_SV },
         { "3200x2400",  OPTION_VAL_3200X2400_SV },
         { "3840x2880",  OPTION_VAL_3840X2880_SV },
         { "4480x3360",  OPTION_VAL_4480X3360_SV },
         { "5120x3840",  OPTION_VAL_5120X3840_SV },
         { "5760x4320",  OPTION_VAL_5760X4320_SV },
         { "6400x4800",  OPTION_VAL_6400X4800_SV },
         { "7040x5280",  OPTION_VAL_7040X5280_SV },
         { "7680x5760",  OPTION_VAL_7680X5760_SV },
         { "8320x6240",  OPTION_VAL_8320X6240_SV },
         { "8960x6720",  OPTION_VAL_8960X6720_SV },
         { "9600x7200",  OPTION_VAL_9600X7200_SV },
         { "10240x7680", OPTION_VAL_10240X7680_SV },
         { "10880x8160", OPTION_VAL_10880X8160_SV },
         { "11520x8640", OPTION_VAL_11520X8640_SV },
         { "12160x9120", OPTION_VAL_12160X9120_SV },
         { "12800x9600", OPTION_VAL_12800X9600_SV },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_SV,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_SV },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_SV },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_SV },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_SV,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_SV },
         { "PAL",     OPTION_VAL_PAL_SV },
         { "PAL_N",   OPTION_VAL_PAL_N_SV },
         { "PAL_M",   OPTION_VAL_PAL_M_SV },
         { "Default", OPTION_VAL_DEFAULT_SV },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_SV,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_SV },
         { "vertical",   OPTION_VAL_VERTICAL_SV },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_SV,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_SV },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_SV },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_SV },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_SV,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_SV },
         { "1GB",   OPTION_VAL_1GB_SV },
         { "2GB",   OPTION_VAL_2GB_SV },
         { "4GB",   OPTION_VAL_4GB_SV },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_SV,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_SV,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_SV,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_SV,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_SV },
         { "1",  OPTION_VAL_1_SV },
         { "2",  OPTION_VAL_2_SV },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_SV,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_SV },
         { "4",  OPTION_VAL_4_SV },
         { "6",  OPTION_VAL_6_SV },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_SV,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_SV,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_SV,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_SV,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_SV,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_SV,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_SV,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_SV },
         { "more", OPTION_VAL_MORE_SV },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_SV,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_SV },
         { "110", OPTION_VAL_110_SV },
         { "120", OPTION_VAL_120_SV },
         { "130", OPTION_VAL_130_SV },
         { "140", OPTION_VAL_140_SV },
         { "150", OPTION_VAL_150_SV },
         { "160", OPTION_VAL_160_SV },
         { "170", OPTION_VAL_170_SV },
         { "180", OPTION_VAL_180_SV },
         { "190", OPTION_VAL_190_SV },
         { "200", OPTION_VAL_200_SV },
         { "210", OPTION_VAL_210_SV },
         { "220", OPTION_VAL_220_SV },
         { "230", OPTION_VAL_230_SV },
         { "240", OPTION_VAL_240_SV },
         { "250", OPTION_VAL_250_SV },
         { "260", OPTION_VAL_260_SV },
         { "270", OPTION_VAL_270_SV },
         { "280", OPTION_VAL_280_SV },
         { "290", OPTION_VAL_290_SV },
         { "300", OPTION_VAL_300_SV },
         { "310", OPTION_VAL_310_SV },
         { "320", OPTION_VAL_320_SV },
         { "330", OPTION_VAL_330_SV },
         { "340", OPTION_VAL_340_SV },
         { "350", OPTION_VAL_350_SV },
         { "360", OPTION_VAL_360_SV },
         { "370", OPTION_VAL_370_SV },
         { "380", OPTION_VAL_380_SV },
         { "390", OPTION_VAL_390_SV },
         { "400", OPTION_VAL_400_SV },
         { "410", OPTION_VAL_410_SV },
         { "420", OPTION_VAL_420_SV },
         { "430", OPTION_VAL_430_SV },
         { "440", OPTION_VAL_440_SV },
         { "450", OPTION_VAL_450_SV },
         { "460", OPTION_VAL_460_SV },
         { "470", OPTION_VAL_470_SV },
         { "480", OPTION_VAL_480_SV },
         { "490", OPTION_VAL_490_SV },
         { "500", OPTION_VAL_500_SV },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_SV,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_SV },
         { "5%",  OPTION_VAL_5_SV },
         { "10%", OPTION_VAL_10_SV },
         { "15%", OPTION_VAL_15_SV },
         { "20%", OPTION_VAL_20_SV },
         { "25%", OPTION_VAL_25_SV },
         { "30%", OPTION_VAL_30_SV },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_SV,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_SV },
         { "5%",  OPTION_VAL_5_SV },
         { "10%", OPTION_VAL_10_SV },
         { "15%", OPTION_VAL_15_SV },
         { "20%", OPTION_VAL_20_SV },
         { "25%", OPTION_VAL_25_SV },
         { "30%", OPTION_VAL_30_SV },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_SV,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_SV,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_SV,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_SV,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_SV },
         { "60%",  OPTION_VAL_60_SV },
         { "70%",  OPTION_VAL_70_SV },
         { "80%",  OPTION_VAL_80_SV },
         { "90%",  OPTION_VAL_90_SV },
         { "100%", OPTION_VAL_100_O47_SV },
         { "110%", OPTION_VAL_110_O47_SV },
         { "120%", OPTION_VAL_120_O47_SV },
         { "130%", OPTION_VAL_130_O47_SV },
         { "140%", OPTION_VAL_140_O47_SV },
         { "150%", OPTION_VAL_150_O47_SV },
         { "160%", OPTION_VAL_160_O47_SV },
         { "170%", OPTION_VAL_170_O47_SV },
         { "180%", OPTION_VAL_180_O47_SV },
         { "190%", OPTION_VAL_190_O47_SV },
         { "200%", OPTION_VAL_200_O47_SV },
         { "210%", OPTION_VAL_210_O47_SV },
         { "220%", OPTION_VAL_220_O47_SV },
         { "230%", OPTION_VAL_230_O47_SV },
         { "240%", OPTION_VAL_240_O47_SV },
         { "250%", OPTION_VAL_250_O47_SV },
         { "260%", OPTION_VAL_260_O47_SV },
         { "270%", OPTION_VAL_270_O47_SV },
         { "280%", OPTION_VAL_280_O47_SV },
         { "290%", OPTION_VAL_290_O47_SV },
         { "300%", OPTION_VAL_300_O47_SV },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_SV,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SV },
         { "Red",      OPTION_VAL_RED_SV },
         { "Green",    OPTION_VAL_GREEN_SV },
         { "Blue",     OPTION_VAL_BLUE_SV },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_SV,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SV },
         { "Red",      OPTION_VAL_RED_SV },
         { "Green",    OPTION_VAL_GREEN_SV },
         { "Blue",     OPTION_VAL_BLUE_SV },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_SV,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SV },
         { "Red",      OPTION_VAL_RED_SV },
         { "Green",    OPTION_VAL_GREEN_SV },
         { "Blue",     OPTION_VAL_BLUE_SV },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_SV,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_SV },
         { "Red",      OPTION_VAL_RED_SV },
         { "Green",    OPTION_VAL_GREEN_SV },
         { "Blue",     OPTION_VAL_BLUE_SV },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_SV,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SV },
         { "Purupuru", OPTION_VAL_PURUPURU_SV },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_SV },
         { "None",     OPTION_VAL_NONE_SV },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_SV,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SV },
         { "Purupuru", OPTION_VAL_PURUPURU_SV },
         { "None",     OPTION_VAL_NONE_SV },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_SV,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SV },
         { "Purupuru", OPTION_VAL_PURUPURU_SV },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_SV },
         { "None",     OPTION_VAL_NONE_SV },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_SV,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SV },
         { "Purupuru", OPTION_VAL_PURUPURU_SV },
         { "None",     OPTION_VAL_NONE_SV },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_SV,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SV },
         { "Purupuru", OPTION_VAL_PURUPURU_SV },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_SV },
         { "None",     OPTION_VAL_NONE_SV },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_SV,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SV },
         { "Purupuru", OPTION_VAL_PURUPURU_SV },
         { "None",     OPTION_VAL_NONE_SV },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_SV,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SV },
         { "Purupuru", OPTION_VAL_PURUPURU_SV },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_SV },
         { "None",     OPTION_VAL_NONE_SV },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_SV,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_SV },
         { "Purupuru", OPTION_VAL_PURUPURU_SV },
         { "None",     OPTION_VAL_NONE_SV },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_SV,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_SV,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_SV,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_SV },
         { "All VMUs", OPTION_VAL_ALL_VMUS_SV },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_SV,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_SV,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_SV,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_SV,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_SV,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_SV,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_SV,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_SV,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SV },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SV },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SV },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SV },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SV },
         { "2x", OPTION_VAL_2_O26_SV },
         { "3x", OPTION_VAL_3X_SV },
         { "4x", OPTION_VAL_4_SV },
         { "5x", OPTION_VAL_5X_SV },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SV },
         { "20%",  OPTION_VAL_20_SV },
         { "30%",  OPTION_VAL_30_SV },
         { "40%",  OPTION_VAL_40_SV },
         { "50%",  OPTION_VAL_50_SV },
         { "60%",  OPTION_VAL_60_SV },
         { "70%",  OPTION_VAL_70_SV },
         { "80%",  OPTION_VAL_80_SV },
         { "90%",  OPTION_VAL_90_SV },
         { "100%", OPTION_VAL_100_O47_SV },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SV },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SV },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SV },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SV },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SV },
         { "2x", OPTION_VAL_2_O26_SV },
         { "3x", OPTION_VAL_3X_SV },
         { "4x", OPTION_VAL_4_SV },
         { "5x", OPTION_VAL_5X_SV },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SV },
         { "20%",  OPTION_VAL_20_SV },
         { "30%",  OPTION_VAL_30_SV },
         { "40%",  OPTION_VAL_40_SV },
         { "50%",  OPTION_VAL_50_SV },
         { "60%",  OPTION_VAL_60_SV },
         { "70%",  OPTION_VAL_70_SV },
         { "80%",  OPTION_VAL_80_SV },
         { "90%",  OPTION_VAL_90_SV },
         { "100%", OPTION_VAL_100_O47_SV },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SV },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SV },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SV },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SV },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SV },
         { "2x", OPTION_VAL_2_O26_SV },
         { "3x", OPTION_VAL_3X_SV },
         { "4x", OPTION_VAL_4_SV },
         { "5x", OPTION_VAL_5X_SV },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SV },
         { "20%",  OPTION_VAL_20_SV },
         { "30%",  OPTION_VAL_30_SV },
         { "40%",  OPTION_VAL_40_SV },
         { "50%",  OPTION_VAL_50_SV },
         { "60%",  OPTION_VAL_60_SV },
         { "70%",  OPTION_VAL_70_SV },
         { "80%",  OPTION_VAL_80_SV },
         { "90%",  OPTION_VAL_90_SV },
         { "100%", OPTION_VAL_100_O47_SV },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_SV },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_SV },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_SV },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_SV },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_SV },
         { "2x", OPTION_VAL_2_O26_SV },
         { "3x", OPTION_VAL_3X_SV },
         { "4x", OPTION_VAL_4_SV },
         { "5x", OPTION_VAL_5X_SV },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_SV },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_SV },
         { "BLACK 02",          OPTION_VAL_BLACK_02_SV },
         { "BLUE 03",           OPTION_VAL_BLUE_SV },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_SV },
         { "GREEN 05",          OPTION_VAL_GREEN_SV },
         { "CYAN 06",           OPTION_VAL_CYAN_06_SV },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_SV },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_SV },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_SV },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_SV },
         { "RED 11",            OPTION_VAL_RED_SV },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_SV },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_SV },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_SV },
         { "GRAY 15",           OPTION_VAL_GRAY_15_SV },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_SV },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_SV },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_SV },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_SV },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_SV },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_SV },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_SV },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_SV },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_SV },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_SV },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_SV },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_SV },
         { "WHITE 28",          OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_SV,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_SV },
         { "20%",  OPTION_VAL_20_SV },
         { "30%",  OPTION_VAL_30_SV },
         { "40%",  OPTION_VAL_40_SV },
         { "50%",  OPTION_VAL_50_SV },
         { "60%",  OPTION_VAL_60_SV },
         { "70%",  OPTION_VAL_70_SV },
         { "80%",  OPTION_VAL_80_SV },
         { "90%",  OPTION_VAL_90_SV },
         { "100%", OPTION_VAL_100_O47_SV },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_sv = {
   option_cats_sv,
   option_defs_sv
};

/* RETRO_LANGUAGE_TH */

#define CATEGORY_SYSTEM_LABEL_TH "ระบบ"
#define CATEGORY_SYSTEM_INFO_0_TH "ตั้งค่าภูมิภาค, ภาษา, BIOS และการตั้งค่าฮาร์ดแวร์พื้นฐาน"
#define CATEGORY_VIDEO_LABEL_TH "วิดีโอ"
#define CATEGORY_VIDEO_INFO_0_TH "ตั้งค่าความละเอียด, การประมวลผลความโปร่งใส และการแสดงผลเอฟเฟกต์ภาพต่างๆ"
#define CATEGORY_PERFORMANCE_LABEL_TH "ประสิทธิภาพ"
#define CATEGORY_PERFORMANCE_INFO_0_TH "ตั้งค่าการประมวลผลแบบเธรด และการข้ามเฟรม"
#define CATEGORY_HACKS_LABEL_TH NULL
#define CATEGORY_HACKS_INFO_0_TH "ตั้งค่าการเขียนทับหน้าจอกว้าง, ความเร็วในการโหลด GD-ROM และการแทนที่พื้นผิว"
#define CATEGORY_INPUT_LABEL_TH "การควบคุม"
#define CATEGORY_INPUT_INFO_0_TH "ตั้งค่าจอยเกมและปืนแสง"
#define CATEGORY_EXPANSIONS_LABEL_TH "ช่องเสียบส่วนขยายของคอนโทรลเลอร์"
#define CATEGORY_EXPANSIONS_INFO_0_TH "เลือกอุปกรณ์ (VMU, อุปกรณ์สั่น) ที่เสียบอยู่ในแต่ละช่องเสียบส่วนขยายของคอนโทรลเลอร์"
#define CATEGORY_VMU_LABEL_TH "หน่วยความจำจำลอง (VMU)"
#define CATEGORY_VMU_INFO_0_TH "ตั้งค่าการแยกไฟล์เซฟ VMU รายเกม และการแสดงผลหน้าจอ VMU บนหน้าจอหลัก"
#define CORE_OPTION_NAME_REGION_LABEL_TH "ภูมิภาค"
#define OPTION_VAL_JAPAN_TH "ญี่ปุ่น"
#define OPTION_VAL_USA_TH NULL
#define OPTION_VAL_EUROPE_TH "ยุโรป"
#define OPTION_VAL_DEFAULT_TH "ค่าเริ่มต้น"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_TH "ภาษา"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_TH "เปลี่ยนภาษาที่ใช้โดย BIOS และเกมที่มีหลายภาษาในตัวเลือก"
#define OPTION_VAL_JAPANESE_TH "ญี่ปุ่น"
#define OPTION_VAL_ENGLISH_TH "อังกฤษ"
#define OPTION_VAL_GERMAN_TH "เยอรมัน"
#define OPTION_VAL_FRENCH_TH "ฝรั่งเศส"
#define OPTION_VAL_SPANISH_TH "สเปน"
#define OPTION_VAL_ITALIAN_TH "อิตาลี"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_TH "จำลอง BIOS (HLE) (ต้องรีสตาร์ท)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_TH "ใช้งาน BIOS จำลอง (HLE) (ต้องรีสตาร์ท)"
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_TH "เปิดใช้ DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_TH "เปิดใช้งานการจำลองระบบประมวลผลสัญญาณเสียงดิจิทัล (DSP) ของ Dreamcast ช่วยให้เสียงที่ออกมามีความแม่นยำสูงขึ้น แต่จะใช้ทรัพยากรเครื่องมากขึ้นตามไปด้วย"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_TH "อนุญาตให้ใช้งานปุ่ม Arcade Service"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_TH "เปิดใช้งานปุ่ม SERVICE สำหรับเกมตู้ Arcade เพื่อเข้าสู่การตั้งค่าภายในของตัวเครื่อง"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_TH "ตั้งค่าเกม NAOMI ให้เป็น Free Play"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_TH "ปรับเปลี่ยนการตั้งค่าการหยอดเหรียญของเกมให้เป็นแบบเล่นฟรี"
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_TH "การจำลองอแดปเตอร์บรอดแบนด์"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_TH "จำลองอแดปเตอร์บรอดแบนด์อีเธอร์เน็ตแทนโมเด็ม (ต้องรีสตาร์ท)"
#define CORE_OPTION_NAME_UPNP_LABEL_TH "เปิดการใช้ UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_TH "ใช้ UPnP เพื่อกำหนดค่าเราเตอร์อินเทอร์เน็ตโดยอัตโนมัติสำหรับการเล่นเกมออนไลน์"
#define CORE_OPTION_NAME_DCNET_LABEL_TH "ใช้ DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_TH "ใช้บริการคลาวด์ DCNet เพื่อเข้าถึงอินเทอร์เน็ตของ Dreamcast"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_TH "ความละเอียดภายใน"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_TH "ปรับเปลี่ยนความละเอียดในการแสดงผล"
#define OPTION_VAL_320X240_TH NULL
#define OPTION_VAL_640X480_TH NULL
#define OPTION_VAL_800X600_TH NULL
#define OPTION_VAL_960X720_TH NULL
#define OPTION_VAL_1024X768_TH NULL
#define OPTION_VAL_1280X960_TH NULL
#define OPTION_VAL_1440X1080_TH NULL
#define OPTION_VAL_1600X1200_TH NULL
#define OPTION_VAL_1920X1440_TH NULL
#define OPTION_VAL_2560X1920_TH NULL
#define OPTION_VAL_2880X2160_TH NULL
#define OPTION_VAL_3200X2400_TH NULL
#define OPTION_VAL_3840X2880_TH NULL
#define OPTION_VAL_4480X3360_TH NULL
#define OPTION_VAL_5120X3840_TH NULL
#define OPTION_VAL_5760X4320_TH NULL
#define OPTION_VAL_6400X4800_TH NULL
#define OPTION_VAL_7040X5280_TH NULL
#define OPTION_VAL_7680X5760_TH NULL
#define OPTION_VAL_8320X6240_TH NULL
#define OPTION_VAL_8960X6720_TH NULL
#define OPTION_VAL_9600X7200_TH NULL
#define OPTION_VAL_10240X7680_TH NULL
#define OPTION_VAL_10880X8160_TH NULL
#define OPTION_VAL_11520X8640_TH NULL
#define OPTION_VAL_12160X9120_TH NULL
#define OPTION_VAL_12800X9600_TH NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_TH "ประเภทสายสัญญาณ"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_TH "ประเภทสัญญาณเอาต์พุต 'TV (คอมโพสิต)' ได้รับการรองรับกว้างขวางที่สุด"
#define OPTION_VAL_VGA_TH NULL
#define OPTION_VAL_TV_RGB_TH NULL
#define OPTION_VAL_TV_COMPOSITE_TH NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_TH "มาตรฐานการแพร่ภาพ"
#define OPTION_VAL_NTSC_TH NULL
#define OPTION_VAL_PAL_TH "PAL (ทั่วโลก)"
#define OPTION_VAL_PAL_N_TH NULL
#define OPTION_VAL_PAL_M_TH NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_TH "การจัดวางหน้าจอ"
#define OPTION_VAL_HORIZONTAL_TH "แนวนอน"
#define OPTION_VAL_VERTICAL_TH "แนวตั้ง"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_TH "การเรียงลำดับอัลฟา"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_TH "เลือกวิธีเรียงลำดับรูปหลายเหลี่ยมโปร่งใส"
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_TH "แบบต่อแถบ (เร็ว, แม่นยำน้อยที่สุด)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_TH "แบบต่อรูปสามเหลี่ยม (ปกติ)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_TH "แบบต่อพิกเซล (แม่นยำ แต่ช้าที่สุด)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_TH "ขนาดบัฟเฟอร์สะสมพิกเซล"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_TH "ค่าที่สูงกว่าอาจจำเป็นสำหรับความละเอียดที่สูงขึ้นเพื่อให้แสดงผลได้อย่างถูกต้อง"
#define OPTION_VAL_512MB_TH NULL
#define OPTION_VAL_1GB_TH NULL
#define OPTION_VAL_2GB_TH NULL
#define OPTION_VAL_4GB_TH NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_TH "จำนวนชั้นความโปร่งใสสูงสุด"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_TH "ค่าที่สูงกว่าอาจจำเป็นสำหรับฉากที่มีความซับซ้อนสูง"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_TH "การจำลองเฟรมบัฟเฟอร์แบบเต็มรูปแบบ"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_TH "เปิดใช้งานการจำลองเฟรมบัฟเฟอร์ แบบเต็มรูปแบบใน VRAM ซึ่งมีประโยชน์สำหรับเกมที่มีการอ่านหรือเขียนข้อมูลเฟรมบัฟเฟอร์ใน VRAM โดยตรง เมื่อเปิดใช้งาน ความละเอียดภายในจะถูกบังคับไว้ที่ 640x480 และอาจส่งผลกระทบต่อประสิทธิภาพการทำงานอย่างรุนแรง"
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_TH "เปิดใช้งานบัฟเฟอร์ RTT (Render To Texture)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_TH "คัดลอกพื้นผิวที่เรนเดอร์แล้วจาก GPU กลับไปยัง VRAM โดยปกติจะเปิดใช้งานตัวเลือกนี้สำหรับเกมที่จำเป็นต้องใช้ เมื่อเปิดใช้งาน จะไม่สามารถขยายความละเอียดของพื้นผิวได้ และอาจส่งผลกระทบต่อประสิทธิภาพการทำงาน"
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_TH NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_TH "เมื่อเปิดใช้งาน พื้นผิว จะใช้รูปภาพขนาดที่เล็กลงเมื่อวัตถุนั้นปรากฏอยู่ไกลออกไป ซึ่งจะช่วยเพิ่มประสิทธิภาพการทำงานและลดอาการภาพระยิบระยับลงได้"
#define CORE_OPTION_NAME_FOG_LABEL_TH "เอฟเฟกต์หมอก"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_TH "ตัวปรับแต่งระดับเสียง"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_TH "คุณลักษณะของ GPU ในเครื่อง Dreamcast ที่โดยปกติแล้วเกมต่างๆ จะใช้เพื่อ วาดเงาของวัตถุ แนะนำให้เปิดใช้งานไว้เป็นปกติ เนื่องจากส่งผลกระทบต่อประสิทธิภาพการทำงานน้อยมากจนแทบไม่รู้สึก"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_TH "ตัวกรอง Anisotropic"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_TH "ปรับปรุงคุณภาพของพื้นผิวบนวัตถุ ที่ทำมุมเอียงหรืออยู่ในระนาบแนวลึกเมื่อมองจากมุมกล้อง ค่าที่สูงขึ้นจะใช้ทรัพยากร GPU มากขึ้น และการเปลี่ยนแปลงค่านี้จะมีผลหลังจากรีสตาร์ทแอปพลิเคชันหรือเกมใหม่เท่านั้น"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_TH "การกรองพื้นผิว"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_TH "โหมดการกรองพื้นผิว ที่ต้องการใช้งาน ตัวเลือกนี้สามารถใช้เพื่อ บังคับ รูปแบบการกรองพื้นผิวในทุกๆ ส่วนของเกม เพื่อให้ได้ภาพที่ดูคมชัดขึ้น หรือดูเนียนตาขึ้นกว่าค่าเริ่มต้น"
#define OPTION_VAL_1_TH "บังคับใช้ Nearest-Neighbor"
#define OPTION_VAL_2_TH "บังคับใช้ Linear"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_TH "หน่วงเวลาการสลับเฟรม"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_TH "ช่วยลดอาการหน้าจอกระพริบหรือวิดีโอที่แสดงผลผิดปกติ แต่ไม่แนะนำให้ใช้กับเครื่องที่มีสเปกต่ำหรือทำงานช้า"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_TH "ตรวจหาการเปลี่ยนแปลงของเฟรมเรต"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_TH "แจ้งเตือนไปยัง Frontend เมื่อเฟรมเรตภายในมีการเปลี่ยนแปลง (เช่น จาก 60 fps เป็น 30 fps) ช่วยให้การจัดจังหวะเฟรม (Frame Pacing) ดีขึ้นในเกมที่ล็อกเฟรมเรตไว้ที่ 30 fps หรือ 20 fps แต่ควรปิดการใช้งานสำหรับเกมที่มีเฟรมเรตแบบปลดล็อก (ไม่คงที่) (เช่น Ecco the Dolphin, Unreal Tournament)"
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_TH "ตัวกรองประมวลผลภายหลังของ PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_TH "ประมวลผลภาพที่เรนเดอร์แล้ว เพื่อจำลองเอฟเฟกต์เฉพาะของ GPU PowerVR2 และสัญญาณวิดีโอแบบอนาล็อก"
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_TH NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_TH "ปรับปรุงกราฟิกภาพพิกเซล แบบ 2D ที่วาดด้วยมือให้มีความคมชัดยิ่งขึ้น ควรใช้เฉพาะกับเกมที่เป็นภาพพิกเซล 2D เท่านั้น"
#define OPTION_VAL_2_O26_TH NULL
#define OPTION_VAL_4_TH NULL
#define OPTION_VAL_6_TH NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_TH NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_TH "เลือกค่าขนาดสูงสุดของพื้นผิวที่จะทำการขยาย (Upscale) หากพื้นผิวมีขนาดใหญ่กว่าค่าที่เลือกไว้ พื้นผิวนั้นจะไม่ถูกนำไปขยายขนาด"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_TH NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_TH "ช่วยแก้ไขปัญหาพื้นผิวผิดเพี้ยน และปัญหาด้านมิติความลึก สำหรับการ์ดจอ AMD และแก้ปัญหาในบางกรณีสำหรับการ์ดจอ Intel ได้ด้วย"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_TH "แก้ไขขอบภาพเบลอจากการขยายขนาด"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_TH "ช่วยแก้ปัญหาขอบสีพื้นผิวเหลือบหรือกลืนกัน (Bleeding) เมื่อมีการขยายขนาด (Upscaling) การปิดฟังก์ชันนี้อาจช่วยได้หากพบอาการพิกเซลบิดเบี้ยวในขณะขยายขนาดภาพในเกม 2D (เช่น MVC2, CVS, KOF ฯลฯ)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_TH "การเรนเดอร์แบบแยกเธรด"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_TH "ประมวลผล GPU และ CPU แยกเธรดกัน แนะนำให้เปิดใช้งานเป็นอย่างยิ่ง"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_TH "ข้ามเฟรมอัตโนมัติ"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_TH "ข้ามเฟรมโดยอัตโนมัติเมื่ออีมูเลเตอร์ทำงานช้าลง หมายเหตุ: การตั้งค่านี้จะมีผลเมื่อเปิดใช้งาน 'การเรนเดอร์แบบแยกเธรด' เท่านั้น"
#define OPTION_VAL_SOME_TH "ปกติ"
#define OPTION_VAL_MORE_TH "สูงสุด"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_TH "การข้ามเฟรม"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_TH "กำหนดจำนวนเฟรมที่จะข้ามในระหว่างแต่ละเฟรมที่แสดงผล"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_TH "สูตรจอกว้าง (จำเป็นต้องเริ่มใหม่)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_TH "เปิดใช้งานสูตรที่ช่วยให้บางเกมแสดงผลในรูปแบบจอกว้าง"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_TH "การแฮ็กโหมดจอกว้าง"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_TH "วาดเรขาคณิตนอกอัตราส่วนภาพ 4:3 ปกติ อาจทำให้เกิดกราฟิกผิดเพี้ยนในพื้นที่ที่แสดงเพิ่มขึ้น"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_TH "GD-ROM โหลดเร็ว (ไม่แม่นยำ)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_TH "เร่งความเร็วการโหลด GD-ROM"
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_TH NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_TH "เปิดใช้งานมอด RAM 32MB สำหรับ Dreamcast อาจส่งผลต่อความเข้ากันได้"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_TH "SH4 CPU ลด/เพิ่มความเร็วสัญญาณนาฬิกา"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_TH "เปลี่ยนความเร็วสัญญาณนาฬิกาของ CPU หลัก SH4 จากค่าเริ่มต้น 200 MHz การลดความเร็วอาจช่วยแพลตฟอร์มที่ช้า การเพิ่มความเร็วอาจช่วยเพิ่มอัตราเฟรมสำหรับบางเกม ควรใช้ด้วยความระมัดระวัง"
#define OPTION_VAL_100_TH NULL
#define OPTION_VAL_110_TH NULL
#define OPTION_VAL_120_TH NULL
#define OPTION_VAL_130_TH NULL
#define OPTION_VAL_140_TH NULL
#define OPTION_VAL_150_TH NULL
#define OPTION_VAL_160_TH NULL
#define OPTION_VAL_170_TH NULL
#define OPTION_VAL_180_TH NULL
#define OPTION_VAL_190_TH NULL
#define OPTION_VAL_200_TH NULL
#define OPTION_VAL_210_TH NULL
#define OPTION_VAL_220_TH NULL
#define OPTION_VAL_230_TH NULL
#define OPTION_VAL_240_TH NULL
#define OPTION_VAL_250_TH NULL
#define OPTION_VAL_260_TH NULL
#define OPTION_VAL_270_TH NULL
#define OPTION_VAL_280_TH NULL
#define OPTION_VAL_290_TH NULL
#define OPTION_VAL_300_TH NULL
#define OPTION_VAL_310_TH NULL
#define OPTION_VAL_320_TH NULL
#define OPTION_VAL_330_TH NULL
#define OPTION_VAL_340_TH NULL
#define OPTION_VAL_350_TH NULL
#define OPTION_VAL_360_TH NULL
#define OPTION_VAL_370_TH NULL
#define OPTION_VAL_380_TH NULL
#define OPTION_VAL_390_TH NULL
#define OPTION_VAL_400_TH NULL
#define OPTION_VAL_410_TH NULL
#define OPTION_VAL_420_TH NULL
#define OPTION_VAL_430_TH NULL
#define OPTION_VAL_440_TH NULL
#define OPTION_VAL_450_TH NULL
#define OPTION_VAL_460_TH NULL
#define OPTION_VAL_470_TH NULL
#define OPTION_VAL_480_TH NULL
#define OPTION_VAL_490_TH NULL
#define OPTION_VAL_500_TH NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_TH "โหลดพื้นผิวแบบกำหนดเอง"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_TH "โหลดพื้นผิวแบบกำหนดเองที่อยู่ในโฟลเดอร์ 'system/dc/textures/<game-id>/'"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_TH "โหลดพื้นผิวแบบกำหนดเองล่วงหน้า"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_TH "โหลดพื้นผิวแบบกำหนดเองล่วงหน้า เมื่อเริ่มเกม อาจช่วยปรับปรุงประสิทธิภาพให้ดีขึ้น แต่จะใช้ RAM เพิ่มมากขึ้น"
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_TH NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_TH "ทุกครั้งที่มีการใช้งานพื้นผิวใหม่ในเกม ไฟล์นั้นจะถูกบันทึกเป็นไฟล์ .png ไว้ในโฟลเดอร์ 'system/dc/texdump/<game-id>/'"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_TH "ดัมป์พื้นผิวที่ถูกแทนที่"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_TH "ดัมป์พื้นผิวที่ถูกแทนที่ด้วยพื้นผิวแบบกำหนดเองเสมอ"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_TH "ระยะหยุดการทำงานของก้านอะนาล็อก"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_TH "เลือกตั้งค่าระยะที่คุณต้องขยับก้านอะนาล็อก ก่อนที่ระบบจะเริ่มประมวลผลการเคลื่อนที่ (Deadzone)"
#define OPTION_VAL_0_TH NULL
#define OPTION_VAL_5_TH NULL
#define OPTION_VAL_10_TH NULL
#define OPTION_VAL_15_TH NULL
#define OPTION_VAL_20_TH NULL
#define OPTION_VAL_25_TH NULL
#define OPTION_VAL_30_TH NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_TH NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_TH "เลือกตั้งค่าระยะที่คุณต้องกดปุ่มทริกเกอร์ ก่อนที่ระบบจะเริ่มประมวลผล"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_TH NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_TH "เมื่อเปิดใช้งาน ปุ่มทริกเกอร์จะทำงานเหมือนปุ่มกดทั่วไป ซึ่งจะถูกประมวลผลเป็นแบบกดจนสุดหรือไม่ได้กดเลย โดยจะไม่มีค่าความหนักเบาระหว่างกลาง (Analog)"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_TH NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_TH "กระจายสัญญาณเอาต์พุตดิจิทัลและสถานะการตอบสนองแบบสั่น (Force-feedback) ผ่านพอร์ต TCP 8000 ซึ่งสามารถใช้งานร่วมกับตัวเลือก \"-output network\" ของ MAME ได้"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_TH "แสดงการตั้งปืนแสง"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_TH "เปิดใช้งานการกำหนดค่าตัวเลือกการแสดงผลเป้าเล็งของปืนแสง หมายเหตุ: อาจจำเป็นต้องเปิด-ปิดเมนูทางลัดใหม่ เพื่อให้การตั้งค่านี้มีผลในทันที"
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_TH "การปรับขนาดเป้าเล็งของปืน"
#define OPTION_VAL_50_TH NULL
#define OPTION_VAL_60_TH NULL
#define OPTION_VAL_70_TH NULL
#define OPTION_VAL_80_TH NULL
#define OPTION_VAL_90_TH NULL
#define OPTION_VAL_100_O47_TH NULL
#define OPTION_VAL_110_O47_TH NULL
#define OPTION_VAL_120_O47_TH NULL
#define OPTION_VAL_130_O47_TH NULL
#define OPTION_VAL_140_O47_TH NULL
#define OPTION_VAL_150_O47_TH NULL
#define OPTION_VAL_160_O47_TH NULL
#define OPTION_VAL_170_O47_TH NULL
#define OPTION_VAL_180_O47_TH NULL
#define OPTION_VAL_190_O47_TH NULL
#define OPTION_VAL_200_O47_TH NULL
#define OPTION_VAL_210_O47_TH NULL
#define OPTION_VAL_220_O47_TH NULL
#define OPTION_VAL_230_O47_TH NULL
#define OPTION_VAL_240_O47_TH NULL
#define OPTION_VAL_250_O47_TH NULL
#define OPTION_VAL_260_O47_TH NULL
#define OPTION_VAL_270_O47_TH NULL
#define OPTION_VAL_280_O47_TH NULL
#define OPTION_VAL_290_O47_TH NULL
#define OPTION_VAL_300_O47_TH NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_TH "การแสดงผลเป้าเล็งของปืน 1"
#define OPTION_VAL_WHITE_TH "ขาว"
#define OPTION_VAL_RED_TH NULL
#define OPTION_VAL_GREEN_TH "เขียว"
#define OPTION_VAL_BLUE_TH "น้ำเงิน"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_TH "การแสดงผลเป้าเล็งของปืน 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_TH "การแสดงผลเป้าเล็งของปืน 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_TH "การแสดงผลเป้าเล็งของปืน 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_TH "อุปกรณ์ในช่องเสียบขยาย A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_TH "เลือกอุปกรณ์ที่เสียบอยู่ในช่องเสียบขยาย A1 (จอยพอร์ต A ช่องที่ 1)"
#define OPTION_VAL_VMU_TH NULL
#define OPTION_VAL_PURUPURU_TH "ชุดสั่น"
#define OPTION_VAL_DREAMPOTATO_TH NULL
#define OPTION_VAL_NONE_TH "ไม่แชร์อินพุต"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_TH "อุปกรณ์ในช่องเสียบขยาย A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_TH "เลือกอุปกรณ์ที่เสียบอยู่ในช่องเสียบขยาย A2 (จอยพอร์ต A ช่องที่ 2)"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_TH "อุปกรณ์ในช่องเสียบขยาย B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_TH "เลือกอุปกรณ์ที่เสียบอยู่ในช่องเสียบขยาย B1 (พอร์ต B ช่องที่ 1)"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_TH "อุปกรณ์ในช่องเสียบขยาย B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_TH "เลือกอุปกรณ์ที่เสียบอยู่ในช่องเสียบขยาย B2 (พอร์ต B ช่องที่ 2)"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_TH "อุปกรณ์ในช่องเสียบขยาย C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_TH "เลือกอุปกรณ์ที่เสียบอยู่ในช่องเสียบขยาย C1 (พอร์ต C ช่องที่ 1)"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_TH "อุปกรณ์ในช่องเสียบขยาย C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_TH "เลือกอุปกรณ์ที่เสียบอยู่ในช่องเสียบขยาย C2 (พอร์ต C ช่องที่ 2)"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_TH "อุปกรณ์ในช่องเสียบขยาย D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_TH "เลือกอุปกรณ์ที่เสียบอยู่ในช่องเสียบขยาย D1 (พอร์ต D ช่องที่ 1)"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_TH "อุปกรณ์ในช่องเสียบขยาย D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_TH "เลือกอุปกรณ์ที่เสียบอยู่ในช่องเสียบขยาย D2 (พอร์ต D ช่องที่ 2)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_TH "หน่วยความจำจำลอง (VMU) แยกตามรายเกม"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_TH "VMU แยกตามเกม"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_TH "เมื่อปิดใช้งาน ทุกเกมจะใช้ไฟล์เซฟ VMUร่วมกันสูงสุด 8 ไฟล์ (A1/A2/B1/B2/C1/C2/D1/D2) ซึ่งจัดเก็บอยู่ในโฟลเดอร์ ระบบ ของ RetroArch\nการตั้งค่า 'VMU A1' จะสร้างไฟล์ VMU 'A1' เฉพาะตัวในโฟลเดอร์ บันทึกสถานะ ของ RetroArch สำหรับแต่ละเนื้อหาที่เริ่มใช้งาน\nการตั้งค่า 'All VMUs' จะสร้างไฟล์ VMU เฉพาะตัวสูงสุด 8 ไฟล์ (A1/A2/B1/B2/C1/C2/D1/D2) สำหรับแต่ละเนื้อหาที่เริ่มใช้งาน"
#define OPTION_VAL_VMU_A1_TH NULL
#define OPTION_VAL_ALL_VMUS_TH "VMU ทั้งหมด"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_TH "เสียงของระบบหน่วยความจำจำลอง (VMU)"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_TH "เสียงของ VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_TH "เมื่อเปิดใช้งาน เสียงบี๊บของ VMU จะถูกเล่น"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_TH "ที่เก็บข้อมูล VMU ของ DreamPotato"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_TH "เปิดใช้งานการจัดเก็บข้อมูล VMU บนตัวจำลอง DreamPotato ที่เชื่อมต่ออยู่ เมื่อเปิดใช้งาน จะไม่มีการใช้ไฟล์บันทึกสถานะ VMU ในเครื่อง"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_TH "แสดงการตั้งค่างานภาพหน้าจอหน่วยความจำจำลอง (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_TH "แสดงการตั้งค่าหน้าจอ VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_TH "เปิดใช้งานการกำหนดค่าการแสดงผล ขนาด ตำแหน่ง และสีของหน้าจอ LCD ของ VMU ที่จำลองขึ้น หมายเหตุ: อาจต้อง เปิด-ปิด เมนูทางลัดเพื่อให้การตั้งค่านี้มีผล"
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_TH "การแสดงผลหน้าจอ VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_TH "ตำแหน่งหน้าจอ VMU 1"
#define OPTION_VAL_UPPER_LEFT_TH "บนซ้าย"
#define OPTION_VAL_UPPER_RIGHT_TH "บนขวา"
#define OPTION_VAL_LOWER_LEFT_TH "ล่างซ้าย"
#define OPTION_VAL_LOWER_RIGHT_TH "ล่างขวา"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_TH "ขนาดหน้าจอ VMU 1"
#define OPTION_VAL_1X_TH NULL
#define OPTION_VAL_3X_TH NULL
#define OPTION_VAL_5X_TH NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_TH "สีพิกเซลที่แสดงผลบนหน้าจอ VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_TH "เปิดเป็นค่าเริ่มต้น"
#define OPTION_VAL_DEFAULT_OFF_01_TH "ปิดเป็นค่าเริ่มต้น"
#define OPTION_VAL_BLACK_02_TH "ดำ"
#define OPTION_VAL_LIGHT_BLUE_04_TH "ฟ้าอ่อน"
#define OPTION_VAL_CYAN_06_TH "ฟ้าอมเขียว"
#define OPTION_VAL_CYAN_BLUE_07_TH "สีฟ้า"
#define OPTION_VAL_LIGHT_GREEN_08_TH "เขียวอ่อน"
#define OPTION_VAL_CYAN_GREEN_09_TH "เขียวอมฟ้า"
#define OPTION_VAL_LIGHT_CYAN_10_TH "ฟ้าสว่าง"
#define OPTION_VAL_PURPLE_12_TH "ม่วง"
#define OPTION_VAL_LIGHT_PURPLE_13_TH "ม่วงอ่อน"
#define OPTION_VAL_YELLOW_14_TH NULL
#define OPTION_VAL_GRAY_15_TH "เทา"
#define OPTION_VAL_LIGHT_PURPLE_2_16_TH "ม่วงอ่อน (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_TH "เขียวอ่อน (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_TH "เขียวอ่อน (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_TH "ฟ้าสว่าง (2)"
#define OPTION_VAL_LIGHT_RED_2_20_TH "แดงสว่าง (2)"
#define OPTION_VAL_MAGENTA_21_TH "ม่วงอมแดง"
#define OPTION_VAL_LIGHT_PURPLE_3_22_TH "ม่วงสว่าง (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_TH "ส้มสว่าง"
#define OPTION_VAL_ORANGE_24_TH "ส้ม"
#define OPTION_VAL_LIGHT_PURPLE_4_25_TH "ม่วงสว่าง (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_TH "เหลืองสว่าง"
#define OPTION_VAL_LIGHT_YELLOW_2_27_TH "เหลืองสว่าง (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_TH "สีพิกเซลขณะปิดของหน้าจอ VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_TH "ความโปร่งใสของหน้าจอ VMU 1"
#define OPTION_VAL_40_TH NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_TH "การแสดงผลหน้าจอ VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_TH "ตำแหน่งหน้าจอ VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_TH "ขนาดหน้าจอ VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_TH "สีพิกเซลเปิดของหน้าจอ VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_TH "สีพิกเซลขณะปิดของหน้าจอ VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_TH "ความโปร่งใสหน้าจอ VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_TH "การแสดงผลหน้าจอ VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_TH "ตำแหน่งหน้าจอ VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_TH "ขนาดหน้าจอ VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_TH "สีพิกเซลเปิดของหน้าจอ VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_TH "สีพิกเซลขณะปิดของหน้าจอ VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_TH "ความโปร่งใสหน้าจอ VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_TH "การแสดงผลหน้าจอ VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_TH "ตำแหน่งหน้าจอ VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_TH "ขนาดหน้าจอ VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_TH "สีพิกเซลเปิดของหน้าจอ VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_TH "สีพิกเซลขณะปิดของหน้าจอ VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_TH "ความโปร่งใสหน้าจอ VMU 4"

struct retro_core_option_v2_category option_cats_th[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_TH,
      CATEGORY_SYSTEM_INFO_0_TH
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_TH,
      CATEGORY_VIDEO_INFO_0_TH
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_TH,
      CATEGORY_PERFORMANCE_INFO_0_TH
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_TH,
      CATEGORY_HACKS_INFO_0_TH
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_TH,
      CATEGORY_INPUT_INFO_0_TH
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_TH,
      CATEGORY_EXPANSIONS_INFO_0_TH
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_TH,
      CATEGORY_VMU_INFO_0_TH
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_th[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_TH,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_TH },
         { "USA",     OPTION_VAL_USA_TH },
         { "Europe",  OPTION_VAL_EUROPE_TH },
         { "Default", OPTION_VAL_DEFAULT_TH },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_TH,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_TH },
         { "English",  OPTION_VAL_ENGLISH_TH },
         { "German",   OPTION_VAL_GERMAN_TH },
         { "French",   OPTION_VAL_FRENCH_TH },
         { "Spanish",  OPTION_VAL_SPANISH_TH },
         { "Italian",  OPTION_VAL_ITALIAN_TH },
         { "Default",  OPTION_VAL_DEFAULT_TH },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_TH,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_TH,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_TH,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_TH,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_TH,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_TH,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_TH,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_TH,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_TH },
         { "640x480",    OPTION_VAL_640X480_TH },
         { "800x600",    OPTION_VAL_800X600_TH },
         { "960x720",    OPTION_VAL_960X720_TH },
         { "1024x768",   OPTION_VAL_1024X768_TH },
         { "1280x960",   OPTION_VAL_1280X960_TH },
         { "1440x1080",  OPTION_VAL_1440X1080_TH },
         { "1600x1200",  OPTION_VAL_1600X1200_TH },
         { "1920x1440",  OPTION_VAL_1920X1440_TH },
         { "2560x1920",  OPTION_VAL_2560X1920_TH },
         { "2880x2160",  OPTION_VAL_2880X2160_TH },
         { "3200x2400",  OPTION_VAL_3200X2400_TH },
         { "3840x2880",  OPTION_VAL_3840X2880_TH },
         { "4480x3360",  OPTION_VAL_4480X3360_TH },
         { "5120x3840",  OPTION_VAL_5120X3840_TH },
         { "5760x4320",  OPTION_VAL_5760X4320_TH },
         { "6400x4800",  OPTION_VAL_6400X4800_TH },
         { "7040x5280",  OPTION_VAL_7040X5280_TH },
         { "7680x5760",  OPTION_VAL_7680X5760_TH },
         { "8320x6240",  OPTION_VAL_8320X6240_TH },
         { "8960x6720",  OPTION_VAL_8960X6720_TH },
         { "9600x7200",  OPTION_VAL_9600X7200_TH },
         { "10240x7680", OPTION_VAL_10240X7680_TH },
         { "10880x8160", OPTION_VAL_10880X8160_TH },
         { "11520x8640", OPTION_VAL_11520X8640_TH },
         { "12160x9120", OPTION_VAL_12160X9120_TH },
         { "12800x9600", OPTION_VAL_12800X9600_TH },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_TH,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_TH },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_TH },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_TH },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_TH,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_TH },
         { "PAL",     OPTION_VAL_PAL_TH },
         { "PAL_N",   OPTION_VAL_PAL_N_TH },
         { "PAL_M",   OPTION_VAL_PAL_M_TH },
         { "Default", OPTION_VAL_DEFAULT_TH },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_TH,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_TH },
         { "vertical",   OPTION_VAL_VERTICAL_TH },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_TH,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_TH },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_TH },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_TH },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_TH,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_TH },
         { "1GB",   OPTION_VAL_1GB_TH },
         { "2GB",   OPTION_VAL_2GB_TH },
         { "4GB",   OPTION_VAL_4GB_TH },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_TH,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_TH,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_TH,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_TH,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_TH },
         { "1",  OPTION_VAL_1_TH },
         { "2",  OPTION_VAL_2_TH },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_TH,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_TH },
         { "4",  OPTION_VAL_4_TH },
         { "6",  OPTION_VAL_6_TH },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_TH,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_TH,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_TH,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_TH,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_TH,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_TH,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_TH,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_TH },
         { "more", OPTION_VAL_MORE_TH },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_TH,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_TH },
         { "110", OPTION_VAL_110_TH },
         { "120", OPTION_VAL_120_TH },
         { "130", OPTION_VAL_130_TH },
         { "140", OPTION_VAL_140_TH },
         { "150", OPTION_VAL_150_TH },
         { "160", OPTION_VAL_160_TH },
         { "170", OPTION_VAL_170_TH },
         { "180", OPTION_VAL_180_TH },
         { "190", OPTION_VAL_190_TH },
         { "200", OPTION_VAL_200_TH },
         { "210", OPTION_VAL_210_TH },
         { "220", OPTION_VAL_220_TH },
         { "230", OPTION_VAL_230_TH },
         { "240", OPTION_VAL_240_TH },
         { "250", OPTION_VAL_250_TH },
         { "260", OPTION_VAL_260_TH },
         { "270", OPTION_VAL_270_TH },
         { "280", OPTION_VAL_280_TH },
         { "290", OPTION_VAL_290_TH },
         { "300", OPTION_VAL_300_TH },
         { "310", OPTION_VAL_310_TH },
         { "320", OPTION_VAL_320_TH },
         { "330", OPTION_VAL_330_TH },
         { "340", OPTION_VAL_340_TH },
         { "350", OPTION_VAL_350_TH },
         { "360", OPTION_VAL_360_TH },
         { "370", OPTION_VAL_370_TH },
         { "380", OPTION_VAL_380_TH },
         { "390", OPTION_VAL_390_TH },
         { "400", OPTION_VAL_400_TH },
         { "410", OPTION_VAL_410_TH },
         { "420", OPTION_VAL_420_TH },
         { "430", OPTION_VAL_430_TH },
         { "440", OPTION_VAL_440_TH },
         { "450", OPTION_VAL_450_TH },
         { "460", OPTION_VAL_460_TH },
         { "470", OPTION_VAL_470_TH },
         { "480", OPTION_VAL_480_TH },
         { "490", OPTION_VAL_490_TH },
         { "500", OPTION_VAL_500_TH },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_TH,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_TH },
         { "5%",  OPTION_VAL_5_TH },
         { "10%", OPTION_VAL_10_TH },
         { "15%", OPTION_VAL_15_TH },
         { "20%", OPTION_VAL_20_TH },
         { "25%", OPTION_VAL_25_TH },
         { "30%", OPTION_VAL_30_TH },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_TH,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_TH },
         { "5%",  OPTION_VAL_5_TH },
         { "10%", OPTION_VAL_10_TH },
         { "15%", OPTION_VAL_15_TH },
         { "20%", OPTION_VAL_20_TH },
         { "25%", OPTION_VAL_25_TH },
         { "30%", OPTION_VAL_30_TH },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_TH,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_TH,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_TH,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_TH,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_TH },
         { "60%",  OPTION_VAL_60_TH },
         { "70%",  OPTION_VAL_70_TH },
         { "80%",  OPTION_VAL_80_TH },
         { "90%",  OPTION_VAL_90_TH },
         { "100%", OPTION_VAL_100_O47_TH },
         { "110%", OPTION_VAL_110_O47_TH },
         { "120%", OPTION_VAL_120_O47_TH },
         { "130%", OPTION_VAL_130_O47_TH },
         { "140%", OPTION_VAL_140_O47_TH },
         { "150%", OPTION_VAL_150_O47_TH },
         { "160%", OPTION_VAL_160_O47_TH },
         { "170%", OPTION_VAL_170_O47_TH },
         { "180%", OPTION_VAL_180_O47_TH },
         { "190%", OPTION_VAL_190_O47_TH },
         { "200%", OPTION_VAL_200_O47_TH },
         { "210%", OPTION_VAL_210_O47_TH },
         { "220%", OPTION_VAL_220_O47_TH },
         { "230%", OPTION_VAL_230_O47_TH },
         { "240%", OPTION_VAL_240_O47_TH },
         { "250%", OPTION_VAL_250_O47_TH },
         { "260%", OPTION_VAL_260_O47_TH },
         { "270%", OPTION_VAL_270_O47_TH },
         { "280%", OPTION_VAL_280_O47_TH },
         { "290%", OPTION_VAL_290_O47_TH },
         { "300%", OPTION_VAL_300_O47_TH },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_TH,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TH },
         { "Red",      OPTION_VAL_RED_TH },
         { "Green",    OPTION_VAL_GREEN_TH },
         { "Blue",     OPTION_VAL_BLUE_TH },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_TH,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TH },
         { "Red",      OPTION_VAL_RED_TH },
         { "Green",    OPTION_VAL_GREEN_TH },
         { "Blue",     OPTION_VAL_BLUE_TH },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_TH,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TH },
         { "Red",      OPTION_VAL_RED_TH },
         { "Green",    OPTION_VAL_GREEN_TH },
         { "Blue",     OPTION_VAL_BLUE_TH },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_TH,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TH },
         { "Red",      OPTION_VAL_RED_TH },
         { "Green",    OPTION_VAL_GREEN_TH },
         { "Blue",     OPTION_VAL_BLUE_TH },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_TH,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TH },
         { "Purupuru", OPTION_VAL_PURUPURU_TH },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_TH },
         { "None",     OPTION_VAL_NONE_TH },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_TH,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TH },
         { "Purupuru", OPTION_VAL_PURUPURU_TH },
         { "None",     OPTION_VAL_NONE_TH },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_TH,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TH },
         { "Purupuru", OPTION_VAL_PURUPURU_TH },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_TH },
         { "None",     OPTION_VAL_NONE_TH },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_TH,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TH },
         { "Purupuru", OPTION_VAL_PURUPURU_TH },
         { "None",     OPTION_VAL_NONE_TH },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_TH,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TH },
         { "Purupuru", OPTION_VAL_PURUPURU_TH },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_TH },
         { "None",     OPTION_VAL_NONE_TH },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_TH,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TH },
         { "Purupuru", OPTION_VAL_PURUPURU_TH },
         { "None",     OPTION_VAL_NONE_TH },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_TH,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TH },
         { "Purupuru", OPTION_VAL_PURUPURU_TH },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_TH },
         { "None",     OPTION_VAL_NONE_TH },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_TH,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TH },
         { "Purupuru", OPTION_VAL_PURUPURU_TH },
         { "None",     OPTION_VAL_NONE_TH },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_TH,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_TH,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_TH,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_TH },
         { "All VMUs", OPTION_VAL_ALL_VMUS_TH },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_TH,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_TH,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_TH,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_TH,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_TH,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_TH,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_TH,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_TH,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TH },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TH },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TH },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TH },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TH },
         { "2x", OPTION_VAL_2_O26_TH },
         { "3x", OPTION_VAL_3X_TH },
         { "4x", OPTION_VAL_4_TH },
         { "5x", OPTION_VAL_5X_TH },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TH },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TH },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TH },
         { "BLUE 03",           OPTION_VAL_BLUE_TH },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TH },
         { "GREEN 05",          OPTION_VAL_GREEN_TH },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TH },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TH },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TH },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TH },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TH },
         { "RED 11",            OPTION_VAL_RED_TH },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TH },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TH },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TH },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TH },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TH },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TH },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TH },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TH },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TH },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TH },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TH },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TH },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TH },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TH },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TH },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TH },
         { "WHITE 28",          OPTION_VAL_WHITE_TH },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TH },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TH },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TH },
         { "BLUE 03",           OPTION_VAL_BLUE_TH },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TH },
         { "GREEN 05",          OPTION_VAL_GREEN_TH },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TH },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TH },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TH },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TH },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TH },
         { "RED 11",            OPTION_VAL_RED_TH },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TH },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TH },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TH },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TH },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TH },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TH },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TH },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TH },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TH },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TH },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TH },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TH },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TH },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TH },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TH },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TH },
         { "WHITE 28",          OPTION_VAL_WHITE_TH },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TH },
         { "20%",  OPTION_VAL_20_TH },
         { "30%",  OPTION_VAL_30_TH },
         { "40%",  OPTION_VAL_40_TH },
         { "50%",  OPTION_VAL_50_TH },
         { "60%",  OPTION_VAL_60_TH },
         { "70%",  OPTION_VAL_70_TH },
         { "80%",  OPTION_VAL_80_TH },
         { "90%",  OPTION_VAL_90_TH },
         { "100%", OPTION_VAL_100_O47_TH },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TH },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TH },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TH },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TH },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TH },
         { "2x", OPTION_VAL_2_O26_TH },
         { "3x", OPTION_VAL_3X_TH },
         { "4x", OPTION_VAL_4_TH },
         { "5x", OPTION_VAL_5X_TH },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TH },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TH },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TH },
         { "BLUE 03",           OPTION_VAL_BLUE_TH },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TH },
         { "GREEN 05",          OPTION_VAL_GREEN_TH },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TH },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TH },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TH },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TH },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TH },
         { "RED 11",            OPTION_VAL_RED_TH },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TH },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TH },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TH },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TH },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TH },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TH },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TH },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TH },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TH },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TH },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TH },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TH },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TH },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TH },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TH },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TH },
         { "WHITE 28",          OPTION_VAL_WHITE_TH },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TH },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TH },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TH },
         { "BLUE 03",           OPTION_VAL_BLUE_TH },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TH },
         { "GREEN 05",          OPTION_VAL_GREEN_TH },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TH },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TH },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TH },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TH },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TH },
         { "RED 11",            OPTION_VAL_RED_TH },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TH },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TH },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TH },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TH },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TH },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TH },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TH },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TH },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TH },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TH },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TH },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TH },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TH },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TH },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TH },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TH },
         { "WHITE 28",          OPTION_VAL_WHITE_TH },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TH },
         { "20%",  OPTION_VAL_20_TH },
         { "30%",  OPTION_VAL_30_TH },
         { "40%",  OPTION_VAL_40_TH },
         { "50%",  OPTION_VAL_50_TH },
         { "60%",  OPTION_VAL_60_TH },
         { "70%",  OPTION_VAL_70_TH },
         { "80%",  OPTION_VAL_80_TH },
         { "90%",  OPTION_VAL_90_TH },
         { "100%", OPTION_VAL_100_O47_TH },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TH },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TH },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TH },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TH },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TH },
         { "2x", OPTION_VAL_2_O26_TH },
         { "3x", OPTION_VAL_3X_TH },
         { "4x", OPTION_VAL_4_TH },
         { "5x", OPTION_VAL_5X_TH },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TH },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TH },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TH },
         { "BLUE 03",           OPTION_VAL_BLUE_TH },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TH },
         { "GREEN 05",          OPTION_VAL_GREEN_TH },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TH },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TH },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TH },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TH },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TH },
         { "RED 11",            OPTION_VAL_RED_TH },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TH },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TH },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TH },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TH },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TH },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TH },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TH },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TH },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TH },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TH },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TH },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TH },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TH },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TH },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TH },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TH },
         { "WHITE 28",          OPTION_VAL_WHITE_TH },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TH },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TH },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TH },
         { "BLUE 03",           OPTION_VAL_BLUE_TH },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TH },
         { "GREEN 05",          OPTION_VAL_GREEN_TH },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TH },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TH },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TH },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TH },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TH },
         { "RED 11",            OPTION_VAL_RED_TH },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TH },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TH },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TH },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TH },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TH },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TH },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TH },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TH },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TH },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TH },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TH },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TH },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TH },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TH },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TH },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TH },
         { "WHITE 28",          OPTION_VAL_WHITE_TH },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TH },
         { "20%",  OPTION_VAL_20_TH },
         { "30%",  OPTION_VAL_30_TH },
         { "40%",  OPTION_VAL_40_TH },
         { "50%",  OPTION_VAL_50_TH },
         { "60%",  OPTION_VAL_60_TH },
         { "70%",  OPTION_VAL_70_TH },
         { "80%",  OPTION_VAL_80_TH },
         { "90%",  OPTION_VAL_90_TH },
         { "100%", OPTION_VAL_100_O47_TH },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TH },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TH },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TH },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TH },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TH },
         { "2x", OPTION_VAL_2_O26_TH },
         { "3x", OPTION_VAL_3X_TH },
         { "4x", OPTION_VAL_4_TH },
         { "5x", OPTION_VAL_5X_TH },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TH },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TH },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TH },
         { "BLUE 03",           OPTION_VAL_BLUE_TH },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TH },
         { "GREEN 05",          OPTION_VAL_GREEN_TH },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TH },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TH },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TH },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TH },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TH },
         { "RED 11",            OPTION_VAL_RED_TH },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TH },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TH },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TH },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TH },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TH },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TH },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TH },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TH },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TH },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TH },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TH },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TH },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TH },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TH },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TH },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TH },
         { "WHITE 28",          OPTION_VAL_WHITE_TH },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TH },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TH },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TH },
         { "BLUE 03",           OPTION_VAL_BLUE_TH },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TH },
         { "GREEN 05",          OPTION_VAL_GREEN_TH },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TH },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TH },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TH },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TH },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TH },
         { "RED 11",            OPTION_VAL_RED_TH },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TH },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TH },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TH },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TH },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TH },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TH },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TH },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TH },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TH },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TH },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TH },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TH },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TH },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TH },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TH },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TH },
         { "WHITE 28",          OPTION_VAL_WHITE_TH },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_TH,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TH },
         { "20%",  OPTION_VAL_20_TH },
         { "30%",  OPTION_VAL_30_TH },
         { "40%",  OPTION_VAL_40_TH },
         { "50%",  OPTION_VAL_50_TH },
         { "60%",  OPTION_VAL_60_TH },
         { "70%",  OPTION_VAL_70_TH },
         { "80%",  OPTION_VAL_80_TH },
         { "90%",  OPTION_VAL_90_TH },
         { "100%", OPTION_VAL_100_O47_TH },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_th = {
   option_cats_th,
   option_defs_th
};

/* RETRO_LANGUAGE_TR */

#define CATEGORY_SYSTEM_LABEL_TR "Sistem"
#define CATEGORY_SYSTEM_INFO_0_TR "Bölge, dil, BIOS ve temel donanım ayarlarını yapılandırın."
#define CATEGORY_VIDEO_LABEL_TR NULL
#define CATEGORY_VIDEO_INFO_0_TR "Çözünürlüğü, istenenden bağımsız şeffaflığı ve görsel efekt ayarlarını yapılandırın."
#define CATEGORY_PERFORMANCE_LABEL_TR "Performans"
#define CATEGORY_PERFORMANCE_INFO_0_TR "Zorlu işleme ve kare atlama ayarlarını yapılandırın."
#define CATEGORY_HACKS_LABEL_TR "Emülatör Geliştirmeleri"
#define CATEGORY_HACKS_INFO_0_TR "Geniş ekran geçersiz kılmaları, GD-ROM yükleme hızı ve doku değiştirme ayarlarını yapılandırın."
#define CATEGORY_INPUT_LABEL_TR "Girdi"
#define CATEGORY_INPUT_INFO_0_TR "Oyun kolu ve light gun ayarlarını yapılandırın."
#define CATEGORY_EXPANSIONS_LABEL_TR "Kontrolcü Genişletme Yuvaları"
#define CATEGORY_EXPANSIONS_INFO_0_TR "Her denetleyici genişletme yuvasına takılı cihazı (VMU, titreşim cihazı) seçin."
#define CATEGORY_VMU_LABEL_TR "Görsel Bellek Birimi"
#define CATEGORY_VMU_INFO_0_TR "Oyun başına VMU kayıt dosyaları ve ekrandaki VMU görünürlük ayarlarını yapılandırın."
#define CORE_OPTION_NAME_REGION_LABEL_TR "Bölge"
#define OPTION_VAL_JAPAN_TR "Japonya"
#define OPTION_VAL_USA_TR "ABD"
#define OPTION_VAL_EUROPE_TR "Avrupa"
#define OPTION_VAL_DEFAULT_TR "Varsayılan"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_TR "Dil"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_TR "BIOS ve birden çok dil içeren oyunlar tarafından kullanılan dili değiştirir."
#define OPTION_VAL_JAPANESE_TR "Japonca"
#define OPTION_VAL_ENGLISH_TR "İngilizce"
#define OPTION_VAL_GERMAN_TR "Almanca"
#define OPTION_VAL_FRENCH_TR "Fransızca"
#define OPTION_VAL_SPANISH_TR "İspanyolca"
#define OPTION_VAL_ITALIAN_TR "İtalyanca"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_TR "HLE BIOS (Yeniden Başlatılmalı)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_TR "Yüksek seviyeli taklit BIOS kullanımını zorunlu kılın."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_TR "DSP'yi Etkinleştir"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_TR "Dreamcast ses DSP (dijital sinyal işlemcisi) taklidini etkinleştirin. Üretilen sesin doğruluğunu artırır, ancak performans gereksinimlerini artırır."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_TR "Kabin Hizmet Düğmelerine İzin Ver"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_TR "Atari salonu oyunlarının kabin ayarlarına girmek üzere SERVİS düğmesini etkinleştirir."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_TR "NAOMI Oyunlarını Serbest Oynamaya Ayarlayın"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_TR "Serbest oynamak için oyunun jeton ayarlarını değiştirin."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_TR "Geniş Bant Adaptör Taklidi"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_TR "Modem yerine Ethernet geniş bant adaptörünü taklit edin. (Yeniden Başlatılmalı)"
#define CORE_OPTION_NAME_UPNP_LABEL_TR "UPnP Etkinleştir"
#define CORE_OPTION_NAME_UPNP_INFO_0_TR "İnternet yönlendiricinizi çevrimiçi oyunlar için otomatik olarak yapılandırmak üzere UPnP kullanın."
#define CORE_OPTION_NAME_DCNET_LABEL_TR "DCNet Kullan"
#define CORE_OPTION_NAME_DCNET_INFO_0_TR "Dreamcast İnternet erişimi için DCNet bulut hizmetini kullanın."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_TR "Dahili Çözünürlük"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_TR "İşleme çözünürlüğünü değiştirin."
#define OPTION_VAL_320X240_TR "320x240 (Yarım)"
#define OPTION_VAL_640X480_TR "640x480 (Doğal)"
#define OPTION_VAL_800X600_TR NULL
#define OPTION_VAL_960X720_TR NULL
#define OPTION_VAL_1024X768_TR NULL
#define OPTION_VAL_1280X960_TR NULL
#define OPTION_VAL_1440X1080_TR NULL
#define OPTION_VAL_1600X1200_TR NULL
#define OPTION_VAL_1920X1440_TR NULL
#define OPTION_VAL_2560X1920_TR NULL
#define OPTION_VAL_2880X2160_TR NULL
#define OPTION_VAL_3200X2400_TR NULL
#define OPTION_VAL_3840X2880_TR NULL
#define OPTION_VAL_4480X3360_TR NULL
#define OPTION_VAL_5120X3840_TR NULL
#define OPTION_VAL_5760X4320_TR NULL
#define OPTION_VAL_6400X4800_TR NULL
#define OPTION_VAL_7040X5280_TR NULL
#define OPTION_VAL_7680X5760_TR NULL
#define OPTION_VAL_8320X6240_TR NULL
#define OPTION_VAL_8960X6720_TR NULL
#define OPTION_VAL_9600X7200_TR NULL
#define OPTION_VAL_10240X7680_TR NULL
#define OPTION_VAL_10880X8160_TR NULL
#define OPTION_VAL_11520X8640_TR NULL
#define OPTION_VAL_12160X9120_TR NULL
#define OPTION_VAL_12800X9600_TR NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_TR "Kablo Tipi"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_TR "Çıkış sinyali türü. 'TV (Kompozit)' en yaygın olarak desteklenendir."
#define OPTION_VAL_VGA_TR NULL
#define OPTION_VAL_TV_RGB_TR NULL
#define OPTION_VAL_TV_COMPOSITE_TR NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_TR "Standart Yayın"
#define OPTION_VAL_NTSC_TR NULL
#define OPTION_VAL_PAL_TR "PAL (Dünya)"
#define OPTION_VAL_PAL_N_TR "PAL-N (Arjantin, Paraguay, Uruguay)"
#define OPTION_VAL_PAL_M_TR "PAL-M (Brezilya)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_TR "Ekran Yönü"
#define OPTION_VAL_HORIZONTAL_TR "Yatay"
#define OPTION_VAL_VERTICAL_TR "Dikey"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_TR "Alfabetik Sıralama"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_TR "Saydam çokgenlerin nasıl sıralanacağını seçin."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_TR "Şerit Başına (hızlı, en az doğru)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_TR "Üçgen Başına (normal)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_TR "Piksel Başına (doğru, ancak en yavaş)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_TR "Toplama Piksel Tampon Boyutu"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_TR "Daha yüksek çözünürlüklerin doğru çıktı alması için daha yüksek değerler gerekebilir."
#define OPTION_VAL_512MB_TR NULL
#define OPTION_VAL_1GB_TR NULL
#define OPTION_VAL_2GB_TR NULL
#define OPTION_VAL_4GB_TR NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_TR "En Fazla Şeffaf Katman"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_TR "Karmaşık sahneler için daha yüksek değerler gerekebilir."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_TR "Tam kare önbellek taklidi"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_TR "VRAM üstünde tam kare önbellek taklidini etkinleştirin. Kare önbelleğini doğrudan VRAM üstünden okuyan veya yazan oyunlar için kullanışlıdır. Etkinleştirildiğinde, Dahili Çözünürlük 640x480 olarak zorlanır ve performans ciddi şekilde etkilenebilir."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_TR "RTT (Dokuya Dönüştür) Önbelleğini Etkinleştir"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_TR "İşlenen dokuları GPU'dan VRAM'e geri kopyalayın. Bu seçenek normalde bunu gerektiren oyunlar için etkinleştirilir. Etkinleştirildiğinde, doku işleme yükseltme devre dışı bırakılır ve performans etkilenebilir."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_TR "Görüş Açısı"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_TR "Etkinleştirilen dokular daha uzakta göründüklerinde kendilerinin daha küçük benzerlerini kullanacakları zaman performansları artırabilir ve parıltıyı azaltabilirler."
#define CORE_OPTION_NAME_FOG_LABEL_TR "Sis Efektleri"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_TR "Hacim Değiştirici"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_TR "Genellikle oyunlar tarafından nesne gölgeleri çizmek için kullanılan bir Dreamcast GPU özelliği. Bu normalde etkinleştirilmelidir - performansa olumsuz etkisi yok denilebilecek düzeydedir."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_TR "Eşyönsüz Filtreleme"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_TR "Kameraya göre eğik görüş açılarında olan yüzeylerdeki dokuların kalitesini iyileştirin. Daha yüksek değerler GPU'da daha fazla talepkardır. Bu ayarda yapılan değişiklikler yalnızca yeniden başlatmanın ardından geçerlidir."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_TR "Doku Filtreleme"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_TR "Kullanılacak doku filtreleme kipi. Bu, Varsayılan'dan daha net (veya pürüzsüz) bir görünüm elde etmek için tüm dokularda belirli bir doku filtreleme kipini zorlamak için kullanılabilir. Varsayılan dışındaki değerler, çeşitli işleme sorunlarına neden olabilir. Bu ayarda yapılan değişiklikler yalnızca yeniden başlatmanın ardından geçerlidir."
#define OPTION_VAL_1_TR "En Yakına Zorla"
#define OPTION_VAL_2_TR "Doğrusala Zorla"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_TR "Gecikmeli Kare Değişimi"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_TR "Yanıp sönen ekranlardan veya sorunlu videolardan kaçınmak için kullanışlıdır. Yavaş platformlarda önerilmez."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_TR "Kare Hızı Değişikliklerini Algıla"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_TR "Dahili kare hızı değiştiğinde (örneğin, 60 fps'den 30 fps'ye) ön ucu bilgilendirin. Kilitli 30 fps veya 20 fps'de çalışan oyunlarda kare hızını iyileştirir, ancak kilitlenmemiş (dengesiz) kare hızlarına sahip oyunlar için devre dışı bırakılması gerekir (örn. Ecco the Dolphin, Unreal Tournament). Not: 'Otomatik Kare Atla' etkinleştirildiğinde kullanılamaz."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_TR "PowerVR2 Rötuş Filtresi"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_TR "PowerVR2 GPU'ya ve analog video sinyallerine özgü efektleri simüle etmek için işlenen görüntüyü son işlemden geçirin."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_TR "Doku Yükseltme (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_TR "Elle çizilmiş 2B piksel sanatsal grafiklerini geliştirin. Yalnızca 2D piksel oyunlarla kullanılmalıdır."
#define OPTION_VAL_2_O26_TR NULL
#define OPTION_VAL_4_TR NULL
#define OPTION_VAL_6_TR NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_TR "Doku Yükseltme Azami Filtre Boyutu"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_TR "Yükseltilecek doku için azami boyut değerini seçin; eğer doku boyutu seçilen değerden yüksekse büyütülmeyecektir."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_TR "Yerel Derinlik Enterpolasyonu"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_TR "AMD GPU'larda doku bozulması ve derinlik sorunlarına yardımcı olur. Bazı durumlarda Intel GPU'lara da yardımcı olabilir."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_TR "Yükseltirken Doku Kenarlarını Düzelt"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_TR "Yükseltme işleminde doku bozulması durumunda yardımcı olur. Bunu devre dışı bırakmak, 2D oyunlarda (MVC2, CVS, KOF, vb.) yükseltme sırasında piksellerin çarpık olması durumunda yardımcı olabilir"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_TR "Zorlu İşleme"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_TR "GPU ve CPU'yu farklı iş parçacıklarında çalıştırır. Şiddetle tavsiye edilir."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_TR "Kareyi Otomatik Atla"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_TR "Emulatör yavaş çalışırken kareleri otomatik olarak atlayın. Not: Bu ayar yalnızca 'Zorlu İşleme' etkinleştirildiğinde geçerlidir."
#define OPTION_VAL_SOME_TR NULL
#define OPTION_VAL_MORE_TR "Azami"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_TR "Kare Atlama"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_TR "Görüntülenen her kare arasında atlanacak kare sayısını ayarlar."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_TR "Geniş Ekran Hileleri (Yeniden Başlatılmalı)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_TR "Belirli oyunların geniş ekran biçiminde görüntülenmesine izin veren hileleri etkinleştirir."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_TR "Geniş Ekran Geliştirmesi"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_TR "Normal 4:3 en boy oranının dışına geometri çizin. Ortaya çıkan alanlarda grafiksel hatalara neden olabilir."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_TR "GDROM Hızlı Yükleme (kusurlu)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_TR "GD-ROM yüklemesini hızlandırır."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_TR "Dreamcast 32MB RAM Modu"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_TR "Dreamcast için 32 MB RAM Modunu etkinleştirir. Uyumluluğu etkileyebilir"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_TR "SH4 CPU düşük/hız aşırtması"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_TR "SH4 ana CPU saatini varsayılan 200 MHz değerinden değiştirin. Hız aşırtması platformların yavaşlamasına yardımcı olabilir. Hız aşırtma bazı oyunlarda kare hızını artırabilir. Dikkatle kullanın."
#define OPTION_VAL_100_TR NULL
#define OPTION_VAL_110_TR NULL
#define OPTION_VAL_120_TR NULL
#define OPTION_VAL_130_TR NULL
#define OPTION_VAL_140_TR NULL
#define OPTION_VAL_150_TR NULL
#define OPTION_VAL_160_TR NULL
#define OPTION_VAL_170_TR NULL
#define OPTION_VAL_180_TR NULL
#define OPTION_VAL_190_TR NULL
#define OPTION_VAL_200_TR NULL
#define OPTION_VAL_210_TR NULL
#define OPTION_VAL_220_TR NULL
#define OPTION_VAL_230_TR NULL
#define OPTION_VAL_240_TR NULL
#define OPTION_VAL_250_TR NULL
#define OPTION_VAL_260_TR NULL
#define OPTION_VAL_270_TR NULL
#define OPTION_VAL_280_TR NULL
#define OPTION_VAL_290_TR NULL
#define OPTION_VAL_300_TR NULL
#define OPTION_VAL_310_TR NULL
#define OPTION_VAL_320_TR NULL
#define OPTION_VAL_330_TR NULL
#define OPTION_VAL_340_TR NULL
#define OPTION_VAL_350_TR NULL
#define OPTION_VAL_360_TR NULL
#define OPTION_VAL_370_TR NULL
#define OPTION_VAL_380_TR NULL
#define OPTION_VAL_390_TR NULL
#define OPTION_VAL_400_TR NULL
#define OPTION_VAL_410_TR NULL
#define OPTION_VAL_420_TR NULL
#define OPTION_VAL_430_TR NULL
#define OPTION_VAL_440_TR NULL
#define OPTION_VAL_450_TR NULL
#define OPTION_VAL_460_TR NULL
#define OPTION_VAL_470_TR NULL
#define OPTION_VAL_480_TR NULL
#define OPTION_VAL_490_TR NULL
#define OPTION_VAL_500_TR NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_TR "Özel Dokular Yükle"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_TR "'system/dc/textures/<game-id>/' klasöründe bulunan özel dokuları yükleyin."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_TR "Özel Dokuları Önceden Yükle"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_TR "Oyun başlangıcında özel dokuları önceden yükleyin. Bu, performansı artırabilir ancak bellek kullanımını yükseltir."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_TR "Dokuları Sakla"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_TR "Oyun tarafından her yeni doku kullanıldığında, 'system/dc/texdump/<game-id>/' klasörüne .png dosyası olarak kaydedilecektir."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_TR NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_TR NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_TR "Analog Çubuğu Ölü Bölge"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_TR "İşlenmeye başlamadan önce analog çubuğu ne kadar itmeniz gerektiğini seçin."
#define OPTION_VAL_0_TR "%0"
#define OPTION_VAL_5_TR "%5"
#define OPTION_VAL_10_TR "%10"
#define OPTION_VAL_15_TR "%15"
#define OPTION_VAL_20_TR "%20"
#define OPTION_VAL_25_TR "%25"
#define OPTION_VAL_30_TR "%30"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_TR "Tetik Ölü Bölge"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_TR "İşlenmeye başlamadan önce tetiğe ne kadar basmanız gerektiğini seçin."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_TR "Dijital Tetikleyiciler"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_TR "Etkinleştirildiğinde, tetikleyiciler normal düğmeler gibi davranacak, yani tamamen basılmış veya hiç basılmamış olarak, arada bir şey olmadan işlenecekler."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_TR "Yayın Dijital Çıkışları"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_TR "TCP port 8000'de dijital çıkışları ve zorla geri bildirim durumunu yayınlayın. \"-output network\" MAME seçeneğiyle uyumludur."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_TR "Light Gun Ayarlarını Göster"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_TR "Light gun artı gösterge seçeneklerinin yapılandırılmasını etkinleştirin. NOT: Bu ayarın etkili olması için Hızlı Menü'nün değiştirilmesi gerekebilir."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_TR "Silah Nişangah Boyut Ölçeği"
#define OPTION_VAL_50_TR "%50"
#define OPTION_VAL_60_TR "%60"
#define OPTION_VAL_70_TR "%70"
#define OPTION_VAL_80_TR "%80"
#define OPTION_VAL_90_TR "%90"
#define OPTION_VAL_100_O47_TR "%100"
#define OPTION_VAL_110_O47_TR "%110"
#define OPTION_VAL_120_O47_TR "%120"
#define OPTION_VAL_130_O47_TR "%130"
#define OPTION_VAL_140_O47_TR "%140"
#define OPTION_VAL_150_O47_TR "%150"
#define OPTION_VAL_160_O47_TR "%160"
#define OPTION_VAL_170_O47_TR "%170"
#define OPTION_VAL_180_O47_TR "%180"
#define OPTION_VAL_190_O47_TR "%190"
#define OPTION_VAL_200_O47_TR "%200"
#define OPTION_VAL_210_O47_TR "%210"
#define OPTION_VAL_220_O47_TR "%220"
#define OPTION_VAL_230_O47_TR "%230"
#define OPTION_VAL_240_O47_TR "%240"
#define OPTION_VAL_250_O47_TR "%250"
#define OPTION_VAL_260_O47_TR "%260"
#define OPTION_VAL_270_O47_TR "%270"
#define OPTION_VAL_280_O47_TR "%280"
#define OPTION_VAL_290_O47_TR "%290"
#define OPTION_VAL_300_O47_TR "%300"
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_TR "Silah Artı Göstergesi 1 Görüntüle"
#define OPTION_VAL_WHITE_TR "Beyaz"
#define OPTION_VAL_RED_TR "Kırmızı"
#define OPTION_VAL_GREEN_TR "Yeşil"
#define OPTION_VAL_BLUE_TR "Mavi"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_TR "Silah Artı Göstergesi 2 Görüntüle"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_TR "Silah Artı Göstergesi 3 Görüntüle"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_TR "Silah Artı Göstergesi 4 Görüntüle"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_TR "Genişletme Yuvası Cihaz A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_TR "Genişletme yuvası A1 için takılı cihazı seçin (port A yuva 1)."
#define OPTION_VAL_VMU_TR NULL
#define OPTION_VAL_PURUPURU_TR "Titreşim Paketi"
#define OPTION_VAL_DREAMPOTATO_TR NULL
#define OPTION_VAL_NONE_TR "Yok"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_TR "Genişletme Yuvası Cihaz A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_TR "Genişletme yuvası A2 için takılı cihazı seçin (port A yuva 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_TR "Genişletme Yuvası Cihaz B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_TR "Genişletme yuvası B1 için takılı cihazı seçin (port B yuva 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_TR "Genişletme Yuvası Cihaz B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_TR "Genişletme yuvası B2 için takılı cihazı seçin (port B yuva 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_TR "Genişletme Yuvası Cihaz C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_TR "Genişletme yuvası C1 için takılı cihazı seçin (port C yuva 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_TR "Genişletme Yuvası Cihaz C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_TR "Genişletme yuvası C2 için takılı cihazı seçin (port C yuva 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_TR "Genişletme Yuvası Cihaz D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_TR "Genişletme yuvası D1 için takılı cihazı seçin (port D yuva 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_TR "Genişletme Yuvası Cihaz D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_TR "Genişletme yuvası D2 için takılı cihazı seçin (port D yuva 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_TR "Oyun Başına Görsel Bellek Birimleri/Sistemleri (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_TR "Oyun Başına VMU"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_TR "Devre dışı bırakıldığında, tüm oyunlar RetroArch sistem klasöründe bulunan en fazla 8 VMU kayıt dosyasını (A1/A2/B1/B2/C1/C2/D1/D2) paylaşır.\n'VMU A1' ayarı, başlatılan her oyun için RetroArch kayıt klasörüne benzersiz bir VMU 'A1' dosyası oluşturur.\n'Tüm VMU'' ayarları, başlatılan her oyun için en fazla 8 benzersiz VMU dosyası (A1/A2/B1/B2/C1/C2/D1/D2) oluşturur."
#define OPTION_VAL_VMU_A1_TR NULL
#define OPTION_VAL_ALL_VMUS_TR "Tüm VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_TR "Görsel Bellek Birimleri/Sistemleri (VMU) Sesleri"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_TR "VMU Sesleri"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_TR "Etkinleştirildiğinde VMU bip sesleri oynatılır."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_TR NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_TR NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_TR "Görsel Bellek Birimi/Sistem (VMU) Görüntü Ayarlarını Göster"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_TR "VMU Görüntü Ayarlarını Göster"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_TR "Taklit edilmiş VMU LCD ekran görünürlüğünün, boyutunun, konumunun ve renginin yapılandırılmasını etkinleştirin. NOT: Bu ayarın etkili olması için Hızlı Menü'nün değiştirilmesi gerekebilir."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_TR "VMU Ekranı 1 Görüntüle"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_TR "VMU Ekranı 1 Pozisyon"
#define OPTION_VAL_UPPER_LEFT_TR "Sol Üst"
#define OPTION_VAL_UPPER_RIGHT_TR "Sağ Üst"
#define OPTION_VAL_LOWER_LEFT_TR "Sol Alt"
#define OPTION_VAL_LOWER_RIGHT_TR "Sağ Alt"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_TR "VMU Ekranı 1 Boyutu"
#define OPTION_VAL_1X_TR NULL
#define OPTION_VAL_3X_TR NULL
#define OPTION_VAL_5X_TR NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_TR "VMU Ekranı 1 Piksel Açık Renkli"
#define OPTION_VAL_DEFAULT_ON_00_TR "Varsayılan AÇIK"
#define OPTION_VAL_DEFAULT_OFF_01_TR "Varsayılan KAPALI"
#define OPTION_VAL_BLACK_02_TR "Siyah"
#define OPTION_VAL_LIGHT_BLUE_04_TR "Açık Mavi"
#define OPTION_VAL_CYAN_06_TR "Camgöbeği"
#define OPTION_VAL_CYAN_BLUE_07_TR "Camgöbeği Mavi"
#define OPTION_VAL_LIGHT_GREEN_08_TR "Açık Yeşil"
#define OPTION_VAL_CYAN_GREEN_09_TR "Camgöbeği Yeşil"
#define OPTION_VAL_LIGHT_CYAN_10_TR "Açık Camgöbeği"
#define OPTION_VAL_PURPLE_12_TR "Mor"
#define OPTION_VAL_LIGHT_PURPLE_13_TR "Açık Mor"
#define OPTION_VAL_YELLOW_14_TR "Sarı"
#define OPTION_VAL_GRAY_15_TR "Boz"
#define OPTION_VAL_LIGHT_PURPLE_2_16_TR "Açık Mor (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_TR "Açık Yeşil (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_TR "Açık Yeşil (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_TR "Açık Camgöbeği (2)"
#define OPTION_VAL_LIGHT_RED_2_20_TR "Açık Kırmızı (2)"
#define OPTION_VAL_MAGENTA_21_TR "Eflatun"
#define OPTION_VAL_LIGHT_PURPLE_3_22_TR "Açık Mor (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_TR "Açık Turuncu"
#define OPTION_VAL_ORANGE_24_TR "Turuncu"
#define OPTION_VAL_LIGHT_PURPLE_4_25_TR "Açık Mor (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_TR "Açık Sarı"
#define OPTION_VAL_LIGHT_YELLOW_2_27_TR "Açık Sarı (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_TR "VMU Ekranı 1 Piksel Kapalı Renk"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_TR "VMU Ekranı 1 Şeffaflığı"
#define OPTION_VAL_40_TR "%40"
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_TR "VMU Ekranı 2 Görüntüle"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_TR "VMU Ekranı 2 Pozisyon"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_TR "VMU Ekranı 2 Boyut"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_TR "VMU Ekranı 2 Piksel Açık Renkli"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_TR "VMU Ekranı 2 Piksel Kapalı Renk"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_TR "VMU Ekranı 2 Şeffaflığı"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_TR "VMU Ekranı 3 Görüntüle"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_TR "VMU Ekranı 3 Pozisyon"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_TR "VMU Ekranı 3 Boyut"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_TR "VMU Ekranı 3 Piksel Açık Renkli"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_TR "VMU Ekranı 3 Piksel Kapalı Renk"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_TR "VMU Ekranı 3 Şeffaflığı"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_TR "VMU Ekranı 4 Görüntüle"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_TR "VMU Ekranı 4 Pozisyon"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_TR "VMU Ekranı 4 Boyut"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_TR "VMU Ekranı 4 Piksel Açık Renkli"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_TR "VMU Ekranı 4 Piksel Kapalı Renk"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_TR "VMU Ekranı 4 Şeffaflığı"

struct retro_core_option_v2_category option_cats_tr[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_TR,
      CATEGORY_SYSTEM_INFO_0_TR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_TR,
      CATEGORY_VIDEO_INFO_0_TR
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_TR,
      CATEGORY_PERFORMANCE_INFO_0_TR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_TR,
      CATEGORY_HACKS_INFO_0_TR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_TR,
      CATEGORY_INPUT_INFO_0_TR
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_TR,
      CATEGORY_EXPANSIONS_INFO_0_TR
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_TR,
      CATEGORY_VMU_INFO_0_TR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_tr[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_TR,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_TR },
         { "USA",     OPTION_VAL_USA_TR },
         { "Europe",  OPTION_VAL_EUROPE_TR },
         { "Default", OPTION_VAL_DEFAULT_TR },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_TR,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_TR },
         { "English",  OPTION_VAL_ENGLISH_TR },
         { "German",   OPTION_VAL_GERMAN_TR },
         { "French",   OPTION_VAL_FRENCH_TR },
         { "Spanish",  OPTION_VAL_SPANISH_TR },
         { "Italian",  OPTION_VAL_ITALIAN_TR },
         { "Default",  OPTION_VAL_DEFAULT_TR },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_TR,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_TR },
         { "640x480",    OPTION_VAL_640X480_TR },
         { "800x600",    OPTION_VAL_800X600_TR },
         { "960x720",    OPTION_VAL_960X720_TR },
         { "1024x768",   OPTION_VAL_1024X768_TR },
         { "1280x960",   OPTION_VAL_1280X960_TR },
         { "1440x1080",  OPTION_VAL_1440X1080_TR },
         { "1600x1200",  OPTION_VAL_1600X1200_TR },
         { "1920x1440",  OPTION_VAL_1920X1440_TR },
         { "2560x1920",  OPTION_VAL_2560X1920_TR },
         { "2880x2160",  OPTION_VAL_2880X2160_TR },
         { "3200x2400",  OPTION_VAL_3200X2400_TR },
         { "3840x2880",  OPTION_VAL_3840X2880_TR },
         { "4480x3360",  OPTION_VAL_4480X3360_TR },
         { "5120x3840",  OPTION_VAL_5120X3840_TR },
         { "5760x4320",  OPTION_VAL_5760X4320_TR },
         { "6400x4800",  OPTION_VAL_6400X4800_TR },
         { "7040x5280",  OPTION_VAL_7040X5280_TR },
         { "7680x5760",  OPTION_VAL_7680X5760_TR },
         { "8320x6240",  OPTION_VAL_8320X6240_TR },
         { "8960x6720",  OPTION_VAL_8960X6720_TR },
         { "9600x7200",  OPTION_VAL_9600X7200_TR },
         { "10240x7680", OPTION_VAL_10240X7680_TR },
         { "10880x8160", OPTION_VAL_10880X8160_TR },
         { "11520x8640", OPTION_VAL_11520X8640_TR },
         { "12160x9120", OPTION_VAL_12160X9120_TR },
         { "12800x9600", OPTION_VAL_12800X9600_TR },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_TR,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_TR },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_TR },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_TR },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_TR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_TR },
         { "PAL",     OPTION_VAL_PAL_TR },
         { "PAL_N",   OPTION_VAL_PAL_N_TR },
         { "PAL_M",   OPTION_VAL_PAL_M_TR },
         { "Default", OPTION_VAL_DEFAULT_TR },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_TR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_TR },
         { "vertical",   OPTION_VAL_VERTICAL_TR },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_TR,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_TR },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_TR },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_TR },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_TR,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_TR },
         { "1GB",   OPTION_VAL_1GB_TR },
         { "2GB",   OPTION_VAL_2GB_TR },
         { "4GB",   OPTION_VAL_4GB_TR },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_TR,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_TR,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_TR,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_TR,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_TR },
         { "1",  OPTION_VAL_1_TR },
         { "2",  OPTION_VAL_2_TR },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_TR,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_TR },
         { "4",  OPTION_VAL_4_TR },
         { "6",  OPTION_VAL_6_TR },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_TR,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_TR,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_TR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_TR,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_TR,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_TR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_TR,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_TR },
         { "more", OPTION_VAL_MORE_TR },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_TR,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_TR },
         { "110", OPTION_VAL_110_TR },
         { "120", OPTION_VAL_120_TR },
         { "130", OPTION_VAL_130_TR },
         { "140", OPTION_VAL_140_TR },
         { "150", OPTION_VAL_150_TR },
         { "160", OPTION_VAL_160_TR },
         { "170", OPTION_VAL_170_TR },
         { "180", OPTION_VAL_180_TR },
         { "190", OPTION_VAL_190_TR },
         { "200", OPTION_VAL_200_TR },
         { "210", OPTION_VAL_210_TR },
         { "220", OPTION_VAL_220_TR },
         { "230", OPTION_VAL_230_TR },
         { "240", OPTION_VAL_240_TR },
         { "250", OPTION_VAL_250_TR },
         { "260", OPTION_VAL_260_TR },
         { "270", OPTION_VAL_270_TR },
         { "280", OPTION_VAL_280_TR },
         { "290", OPTION_VAL_290_TR },
         { "300", OPTION_VAL_300_TR },
         { "310", OPTION_VAL_310_TR },
         { "320", OPTION_VAL_320_TR },
         { "330", OPTION_VAL_330_TR },
         { "340", OPTION_VAL_340_TR },
         { "350", OPTION_VAL_350_TR },
         { "360", OPTION_VAL_360_TR },
         { "370", OPTION_VAL_370_TR },
         { "380", OPTION_VAL_380_TR },
         { "390", OPTION_VAL_390_TR },
         { "400", OPTION_VAL_400_TR },
         { "410", OPTION_VAL_410_TR },
         { "420", OPTION_VAL_420_TR },
         { "430", OPTION_VAL_430_TR },
         { "440", OPTION_VAL_440_TR },
         { "450", OPTION_VAL_450_TR },
         { "460", OPTION_VAL_460_TR },
         { "470", OPTION_VAL_470_TR },
         { "480", OPTION_VAL_480_TR },
         { "490", OPTION_VAL_490_TR },
         { "500", OPTION_VAL_500_TR },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_TR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_TR },
         { "5%",  OPTION_VAL_5_TR },
         { "10%", OPTION_VAL_10_TR },
         { "15%", OPTION_VAL_15_TR },
         { "20%", OPTION_VAL_20_TR },
         { "25%", OPTION_VAL_25_TR },
         { "30%", OPTION_VAL_30_TR },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_TR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_TR },
         { "5%",  OPTION_VAL_5_TR },
         { "10%", OPTION_VAL_10_TR },
         { "15%", OPTION_VAL_15_TR },
         { "20%", OPTION_VAL_20_TR },
         { "25%", OPTION_VAL_25_TR },
         { "30%", OPTION_VAL_30_TR },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_TR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_TR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_TR,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_TR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_TR },
         { "60%",  OPTION_VAL_60_TR },
         { "70%",  OPTION_VAL_70_TR },
         { "80%",  OPTION_VAL_80_TR },
         { "90%",  OPTION_VAL_90_TR },
         { "100%", OPTION_VAL_100_O47_TR },
         { "110%", OPTION_VAL_110_O47_TR },
         { "120%", OPTION_VAL_120_O47_TR },
         { "130%", OPTION_VAL_130_O47_TR },
         { "140%", OPTION_VAL_140_O47_TR },
         { "150%", OPTION_VAL_150_O47_TR },
         { "160%", OPTION_VAL_160_O47_TR },
         { "170%", OPTION_VAL_170_O47_TR },
         { "180%", OPTION_VAL_180_O47_TR },
         { "190%", OPTION_VAL_190_O47_TR },
         { "200%", OPTION_VAL_200_O47_TR },
         { "210%", OPTION_VAL_210_O47_TR },
         { "220%", OPTION_VAL_220_O47_TR },
         { "230%", OPTION_VAL_230_O47_TR },
         { "240%", OPTION_VAL_240_O47_TR },
         { "250%", OPTION_VAL_250_O47_TR },
         { "260%", OPTION_VAL_260_O47_TR },
         { "270%", OPTION_VAL_270_O47_TR },
         { "280%", OPTION_VAL_280_O47_TR },
         { "290%", OPTION_VAL_290_O47_TR },
         { "300%", OPTION_VAL_300_O47_TR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_TR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TR },
         { "Red",      OPTION_VAL_RED_TR },
         { "Green",    OPTION_VAL_GREEN_TR },
         { "Blue",     OPTION_VAL_BLUE_TR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_TR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TR },
         { "Red",      OPTION_VAL_RED_TR },
         { "Green",    OPTION_VAL_GREEN_TR },
         { "Blue",     OPTION_VAL_BLUE_TR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_TR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TR },
         { "Red",      OPTION_VAL_RED_TR },
         { "Green",    OPTION_VAL_GREEN_TR },
         { "Blue",     OPTION_VAL_BLUE_TR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_TR,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TR },
         { "Red",      OPTION_VAL_RED_TR },
         { "Green",    OPTION_VAL_GREEN_TR },
         { "Blue",     OPTION_VAL_BLUE_TR },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_TR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TR },
         { "Purupuru", OPTION_VAL_PURUPURU_TR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_TR },
         { "None",     OPTION_VAL_NONE_TR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_TR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TR },
         { "Purupuru", OPTION_VAL_PURUPURU_TR },
         { "None",     OPTION_VAL_NONE_TR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_TR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TR },
         { "Purupuru", OPTION_VAL_PURUPURU_TR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_TR },
         { "None",     OPTION_VAL_NONE_TR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_TR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TR },
         { "Purupuru", OPTION_VAL_PURUPURU_TR },
         { "None",     OPTION_VAL_NONE_TR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_TR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TR },
         { "Purupuru", OPTION_VAL_PURUPURU_TR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_TR },
         { "None",     OPTION_VAL_NONE_TR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_TR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TR },
         { "Purupuru", OPTION_VAL_PURUPURU_TR },
         { "None",     OPTION_VAL_NONE_TR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_TR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TR },
         { "Purupuru", OPTION_VAL_PURUPURU_TR },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_TR },
         { "None",     OPTION_VAL_NONE_TR },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_TR,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TR },
         { "Purupuru", OPTION_VAL_PURUPURU_TR },
         { "None",     OPTION_VAL_NONE_TR },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_TR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_TR,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_TR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_TR },
         { "All VMUs", OPTION_VAL_ALL_VMUS_TR },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_TR,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_TR,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_TR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_TR,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_TR,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_TR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_TR,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_TR,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TR },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TR },
         { "2x", OPTION_VAL_2_O26_TR },
         { "3x", OPTION_VAL_3X_TR },
         { "4x", OPTION_VAL_4_TR },
         { "5x", OPTION_VAL_5X_TR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TR },
         { "20%",  OPTION_VAL_20_TR },
         { "30%",  OPTION_VAL_30_TR },
         { "40%",  OPTION_VAL_40_TR },
         { "50%",  OPTION_VAL_50_TR },
         { "60%",  OPTION_VAL_60_TR },
         { "70%",  OPTION_VAL_70_TR },
         { "80%",  OPTION_VAL_80_TR },
         { "90%",  OPTION_VAL_90_TR },
         { "100%", OPTION_VAL_100_O47_TR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TR },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TR },
         { "2x", OPTION_VAL_2_O26_TR },
         { "3x", OPTION_VAL_3X_TR },
         { "4x", OPTION_VAL_4_TR },
         { "5x", OPTION_VAL_5X_TR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TR },
         { "20%",  OPTION_VAL_20_TR },
         { "30%",  OPTION_VAL_30_TR },
         { "40%",  OPTION_VAL_40_TR },
         { "50%",  OPTION_VAL_50_TR },
         { "60%",  OPTION_VAL_60_TR },
         { "70%",  OPTION_VAL_70_TR },
         { "80%",  OPTION_VAL_80_TR },
         { "90%",  OPTION_VAL_90_TR },
         { "100%", OPTION_VAL_100_O47_TR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TR },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TR },
         { "2x", OPTION_VAL_2_O26_TR },
         { "3x", OPTION_VAL_3X_TR },
         { "4x", OPTION_VAL_4_TR },
         { "5x", OPTION_VAL_5X_TR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TR },
         { "20%",  OPTION_VAL_20_TR },
         { "30%",  OPTION_VAL_30_TR },
         { "40%",  OPTION_VAL_40_TR },
         { "50%",  OPTION_VAL_50_TR },
         { "60%",  OPTION_VAL_60_TR },
         { "70%",  OPTION_VAL_70_TR },
         { "80%",  OPTION_VAL_80_TR },
         { "90%",  OPTION_VAL_90_TR },
         { "100%", OPTION_VAL_100_O47_TR },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TR },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TR },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TR },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TR },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TR },
         { "2x", OPTION_VAL_2_O26_TR },
         { "3x", OPTION_VAL_3X_TR },
         { "4x", OPTION_VAL_4_TR },
         { "5x", OPTION_VAL_5X_TR },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TR },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TR },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TR },
         { "BLUE 03",           OPTION_VAL_BLUE_TR },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TR },
         { "GREEN 05",          OPTION_VAL_GREEN_TR },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TR },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TR },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TR },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TR },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TR },
         { "RED 11",            OPTION_VAL_RED_TR },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TR },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TR },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TR },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TR },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TR },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TR },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TR },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TR },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TR },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TR },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TR },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TR },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TR },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TR },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TR },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TR },
         { "WHITE 28",          OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_TR,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TR },
         { "20%",  OPTION_VAL_20_TR },
         { "30%",  OPTION_VAL_30_TR },
         { "40%",  OPTION_VAL_40_TR },
         { "50%",  OPTION_VAL_50_TR },
         { "60%",  OPTION_VAL_60_TR },
         { "70%",  OPTION_VAL_70_TR },
         { "80%",  OPTION_VAL_80_TR },
         { "90%",  OPTION_VAL_90_TR },
         { "100%", OPTION_VAL_100_O47_TR },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_tr = {
   option_cats_tr,
   option_defs_tr
};

/* RETRO_LANGUAGE_TT */

#define CATEGORY_SYSTEM_LABEL_TT NULL
#define CATEGORY_SYSTEM_INFO_0_TT NULL
#define CATEGORY_VIDEO_LABEL_TT "Видео"
#define CATEGORY_VIDEO_INFO_0_TT NULL
#define CATEGORY_PERFORMANCE_LABEL_TT NULL
#define CATEGORY_PERFORMANCE_INFO_0_TT NULL
#define CATEGORY_HACKS_LABEL_TT NULL
#define CATEGORY_HACKS_INFO_0_TT NULL
#define CATEGORY_INPUT_LABEL_TT NULL
#define CATEGORY_INPUT_INFO_0_TT NULL
#define CATEGORY_EXPANSIONS_LABEL_TT NULL
#define CATEGORY_EXPANSIONS_INFO_0_TT NULL
#define CATEGORY_VMU_LABEL_TT NULL
#define CATEGORY_VMU_INFO_0_TT NULL
#define CORE_OPTION_NAME_REGION_LABEL_TT NULL
#define OPTION_VAL_JAPAN_TT NULL
#define OPTION_VAL_USA_TT NULL
#define OPTION_VAL_EUROPE_TT "Аурупа"
#define OPTION_VAL_DEFAULT_TT NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_TT "Тел"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_TT NULL
#define OPTION_VAL_JAPANESE_TT "Япоча"
#define OPTION_VAL_ENGLISH_TT "Инглизчә"
#define OPTION_VAL_GERMAN_TT "Алманча"
#define OPTION_VAL_FRENCH_TT "Французча"
#define OPTION_VAL_SPANISH_TT "Испанча"
#define OPTION_VAL_ITALIAN_TT "Итальянча"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_TT NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_TT NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_TT NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_TT NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_TT NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_TT NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_TT NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_TT NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_TT NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_TT NULL
#define CORE_OPTION_NAME_UPNP_LABEL_TT NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_TT NULL
#define CORE_OPTION_NAME_DCNET_LABEL_TT NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_TT NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_TT NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_TT NULL
#define OPTION_VAL_320X240_TT NULL
#define OPTION_VAL_640X480_TT NULL
#define OPTION_VAL_800X600_TT NULL
#define OPTION_VAL_960X720_TT NULL
#define OPTION_VAL_1024X768_TT NULL
#define OPTION_VAL_1280X960_TT NULL
#define OPTION_VAL_1440X1080_TT NULL
#define OPTION_VAL_1600X1200_TT NULL
#define OPTION_VAL_1920X1440_TT NULL
#define OPTION_VAL_2560X1920_TT NULL
#define OPTION_VAL_2880X2160_TT NULL
#define OPTION_VAL_3200X2400_TT NULL
#define OPTION_VAL_3840X2880_TT NULL
#define OPTION_VAL_4480X3360_TT NULL
#define OPTION_VAL_5120X3840_TT NULL
#define OPTION_VAL_5760X4320_TT NULL
#define OPTION_VAL_6400X4800_TT NULL
#define OPTION_VAL_7040X5280_TT NULL
#define OPTION_VAL_7680X5760_TT NULL
#define OPTION_VAL_8320X6240_TT NULL
#define OPTION_VAL_8960X6720_TT NULL
#define OPTION_VAL_9600X7200_TT NULL
#define OPTION_VAL_10240X7680_TT NULL
#define OPTION_VAL_10880X8160_TT NULL
#define OPTION_VAL_11520X8640_TT NULL
#define OPTION_VAL_12160X9120_TT NULL
#define OPTION_VAL_12800X9600_TT NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_TT NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_TT NULL
#define OPTION_VAL_VGA_TT NULL
#define OPTION_VAL_TV_RGB_TT "ТВ (RGB)"
#define OPTION_VAL_TV_COMPOSITE_TT NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_TT NULL
#define OPTION_VAL_NTSC_TT NULL
#define OPTION_VAL_PAL_TT NULL
#define OPTION_VAL_PAL_N_TT NULL
#define OPTION_VAL_PAL_M_TT NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_TT NULL
#define OPTION_VAL_HORIZONTAL_TT NULL
#define OPTION_VAL_VERTICAL_TT NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_TT NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_TT NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_TT NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_TT NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_TT NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_TT NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_TT NULL
#define OPTION_VAL_512MB_TT NULL
#define OPTION_VAL_1GB_TT "1 Гб"
#define OPTION_VAL_2GB_TT "2 Гб"
#define OPTION_VAL_4GB_TT "4 Гб"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_TT NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_TT NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_TT NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_TT NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_TT NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_TT NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_TT NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_TT NULL
#define CORE_OPTION_NAME_FOG_LABEL_TT NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_TT NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_TT NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_TT NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_TT NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_TT NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_TT NULL
#define OPTION_VAL_1_TT NULL
#define OPTION_VAL_2_TT NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_TT NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_TT NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_TT NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_TT NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_TT NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_TT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_TT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_TT NULL
#define OPTION_VAL_2_O26_TT NULL
#define OPTION_VAL_4_TT NULL
#define OPTION_VAL_6_TT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_TT NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_TT NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_TT NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_TT NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_TT NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_TT NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_TT NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_TT NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_TT NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_TT NULL
#define OPTION_VAL_SOME_TT "Нормаль"
#define OPTION_VAL_MORE_TT NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_TT NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_TT NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_TT NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_TT NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_TT NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_TT NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_TT NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_TT NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_TT NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_TT NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_TT NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_TT NULL
#define OPTION_VAL_100_TT "100 МГц"
#define OPTION_VAL_110_TT "110 МГц"
#define OPTION_VAL_120_TT "120 МГц"
#define OPTION_VAL_130_TT "130 МГц"
#define OPTION_VAL_140_TT "140 МГц"
#define OPTION_VAL_150_TT "150 МГц"
#define OPTION_VAL_160_TT "160 МГц"
#define OPTION_VAL_170_TT "170 МГц"
#define OPTION_VAL_180_TT "180 МГц"
#define OPTION_VAL_190_TT "190 МГц"
#define OPTION_VAL_200_TT "200 МГц"
#define OPTION_VAL_210_TT "210 МГц"
#define OPTION_VAL_220_TT "220 МГц"
#define OPTION_VAL_230_TT "230 МГц"
#define OPTION_VAL_240_TT "240 МГц"
#define OPTION_VAL_250_TT "250 МГц"
#define OPTION_VAL_260_TT "260 МГц"
#define OPTION_VAL_270_TT "270 МГц"
#define OPTION_VAL_280_TT "280 МГц"
#define OPTION_VAL_290_TT "290 МГц"
#define OPTION_VAL_300_TT "300 МГц"
#define OPTION_VAL_310_TT "310 МГц"
#define OPTION_VAL_320_TT "320 МГц"
#define OPTION_VAL_330_TT "330 МГц"
#define OPTION_VAL_340_TT "340 МГц"
#define OPTION_VAL_350_TT "350 МГц"
#define OPTION_VAL_360_TT "360 МГц"
#define OPTION_VAL_370_TT "370 МГц"
#define OPTION_VAL_380_TT "380 МГц"
#define OPTION_VAL_390_TT "390 МГц"
#define OPTION_VAL_400_TT "400 МГц"
#define OPTION_VAL_410_TT "410 МГц"
#define OPTION_VAL_420_TT "420 МГц"
#define OPTION_VAL_430_TT "430 МГц"
#define OPTION_VAL_440_TT "440 МГц"
#define OPTION_VAL_450_TT "450 МГц"
#define OPTION_VAL_460_TT "460 МГц"
#define OPTION_VAL_470_TT "470 МГц"
#define OPTION_VAL_480_TT "480 МГц"
#define OPTION_VAL_490_TT "490 МГц"
#define OPTION_VAL_500_TT "500 МГц"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_TT NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_TT NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_TT NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_TT NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_TT NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_TT NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_TT NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_TT NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_TT NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_TT NULL
#define OPTION_VAL_0_TT NULL
#define OPTION_VAL_5_TT NULL
#define OPTION_VAL_10_TT NULL
#define OPTION_VAL_15_TT NULL
#define OPTION_VAL_20_TT NULL
#define OPTION_VAL_25_TT NULL
#define OPTION_VAL_30_TT NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_TT NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_TT NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_TT NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_TT NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_TT NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_TT NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_TT NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_TT NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_TT NULL
#define OPTION_VAL_50_TT NULL
#define OPTION_VAL_60_TT NULL
#define OPTION_VAL_70_TT NULL
#define OPTION_VAL_80_TT NULL
#define OPTION_VAL_90_TT NULL
#define OPTION_VAL_100_O47_TT NULL
#define OPTION_VAL_110_O47_TT NULL
#define OPTION_VAL_120_O47_TT NULL
#define OPTION_VAL_130_O47_TT NULL
#define OPTION_VAL_140_O47_TT NULL
#define OPTION_VAL_150_O47_TT NULL
#define OPTION_VAL_160_O47_TT NULL
#define OPTION_VAL_170_O47_TT NULL
#define OPTION_VAL_180_O47_TT NULL
#define OPTION_VAL_190_O47_TT NULL
#define OPTION_VAL_200_O47_TT NULL
#define OPTION_VAL_210_O47_TT NULL
#define OPTION_VAL_220_O47_TT NULL
#define OPTION_VAL_230_O47_TT NULL
#define OPTION_VAL_240_O47_TT NULL
#define OPTION_VAL_250_O47_TT NULL
#define OPTION_VAL_260_O47_TT NULL
#define OPTION_VAL_270_O47_TT NULL
#define OPTION_VAL_280_O47_TT NULL
#define OPTION_VAL_290_O47_TT NULL
#define OPTION_VAL_300_O47_TT NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_TT NULL
#define OPTION_VAL_WHITE_TT "Ак"
#define OPTION_VAL_RED_TT "Кызыл"
#define OPTION_VAL_GREEN_TT NULL
#define OPTION_VAL_BLUE_TT NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_TT NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_TT NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_TT NULL
#define OPTION_VAL_VMU_TT NULL
#define OPTION_VAL_PURUPURU_TT NULL
#define OPTION_VAL_DREAMPOTATO_TT NULL
#define OPTION_VAL_NONE_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_TT NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_TT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_TT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_TT NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_TT NULL
#define OPTION_VAL_VMU_A1_TT NULL
#define OPTION_VAL_ALL_VMUS_TT NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_TT NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_TT NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_TT NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_TT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_TT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_TT NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_TT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_TT NULL
#define OPTION_VAL_UPPER_LEFT_TT NULL
#define OPTION_VAL_UPPER_RIGHT_TT NULL
#define OPTION_VAL_LOWER_LEFT_TT NULL
#define OPTION_VAL_LOWER_RIGHT_TT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_TT NULL
#define OPTION_VAL_1X_TT NULL
#define OPTION_VAL_3X_TT NULL
#define OPTION_VAL_5X_TT NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_TT NULL
#define OPTION_VAL_DEFAULT_ON_00_TT NULL
#define OPTION_VAL_DEFAULT_OFF_01_TT NULL
#define OPTION_VAL_BLACK_02_TT "Кара"
#define OPTION_VAL_LIGHT_BLUE_04_TT NULL
#define OPTION_VAL_CYAN_06_TT NULL
#define OPTION_VAL_CYAN_BLUE_07_TT NULL
#define OPTION_VAL_LIGHT_GREEN_08_TT NULL
#define OPTION_VAL_CYAN_GREEN_09_TT NULL
#define OPTION_VAL_LIGHT_CYAN_10_TT NULL
#define OPTION_VAL_PURPLE_12_TT NULL
#define OPTION_VAL_LIGHT_PURPLE_13_TT NULL
#define OPTION_VAL_YELLOW_14_TT NULL
#define OPTION_VAL_GRAY_15_TT NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_TT NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_TT NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_TT NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_TT NULL
#define OPTION_VAL_LIGHT_RED_2_20_TT NULL
#define OPTION_VAL_MAGENTA_21_TT NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_TT NULL
#define OPTION_VAL_LIGHT_ORANGE_23_TT NULL
#define OPTION_VAL_ORANGE_24_TT NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_TT NULL
#define OPTION_VAL_LIGHT_YELLOW_26_TT NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_TT NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_TT NULL
#define OPTION_VAL_40_TT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_TT NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_TT NULL

struct retro_core_option_v2_category option_cats_tt[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_TT,
      CATEGORY_SYSTEM_INFO_0_TT
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_TT,
      CATEGORY_VIDEO_INFO_0_TT
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_TT,
      CATEGORY_PERFORMANCE_INFO_0_TT
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_TT,
      CATEGORY_HACKS_INFO_0_TT
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_TT,
      CATEGORY_INPUT_INFO_0_TT
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_TT,
      CATEGORY_EXPANSIONS_INFO_0_TT
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_TT,
      CATEGORY_VMU_INFO_0_TT
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_tt[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_TT,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_TT },
         { "USA",     OPTION_VAL_USA_TT },
         { "Europe",  OPTION_VAL_EUROPE_TT },
         { "Default", OPTION_VAL_DEFAULT_TT },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_TT,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_TT },
         { "English",  OPTION_VAL_ENGLISH_TT },
         { "German",   OPTION_VAL_GERMAN_TT },
         { "French",   OPTION_VAL_FRENCH_TT },
         { "Spanish",  OPTION_VAL_SPANISH_TT },
         { "Italian",  OPTION_VAL_ITALIAN_TT },
         { "Default",  OPTION_VAL_DEFAULT_TT },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_TT,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_TT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_TT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_TT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_TT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_TT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_TT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_TT,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_TT },
         { "640x480",    OPTION_VAL_640X480_TT },
         { "800x600",    OPTION_VAL_800X600_TT },
         { "960x720",    OPTION_VAL_960X720_TT },
         { "1024x768",   OPTION_VAL_1024X768_TT },
         { "1280x960",   OPTION_VAL_1280X960_TT },
         { "1440x1080",  OPTION_VAL_1440X1080_TT },
         { "1600x1200",  OPTION_VAL_1600X1200_TT },
         { "1920x1440",  OPTION_VAL_1920X1440_TT },
         { "2560x1920",  OPTION_VAL_2560X1920_TT },
         { "2880x2160",  OPTION_VAL_2880X2160_TT },
         { "3200x2400",  OPTION_VAL_3200X2400_TT },
         { "3840x2880",  OPTION_VAL_3840X2880_TT },
         { "4480x3360",  OPTION_VAL_4480X3360_TT },
         { "5120x3840",  OPTION_VAL_5120X3840_TT },
         { "5760x4320",  OPTION_VAL_5760X4320_TT },
         { "6400x4800",  OPTION_VAL_6400X4800_TT },
         { "7040x5280",  OPTION_VAL_7040X5280_TT },
         { "7680x5760",  OPTION_VAL_7680X5760_TT },
         { "8320x6240",  OPTION_VAL_8320X6240_TT },
         { "8960x6720",  OPTION_VAL_8960X6720_TT },
         { "9600x7200",  OPTION_VAL_9600X7200_TT },
         { "10240x7680", OPTION_VAL_10240X7680_TT },
         { "10880x8160", OPTION_VAL_10880X8160_TT },
         { "11520x8640", OPTION_VAL_11520X8640_TT },
         { "12160x9120", OPTION_VAL_12160X9120_TT },
         { "12800x9600", OPTION_VAL_12800X9600_TT },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_TT,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_TT },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_TT },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_TT },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_TT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_TT },
         { "PAL",     OPTION_VAL_PAL_TT },
         { "PAL_N",   OPTION_VAL_PAL_N_TT },
         { "PAL_M",   OPTION_VAL_PAL_M_TT },
         { "Default", OPTION_VAL_DEFAULT_TT },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_TT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_TT },
         { "vertical",   OPTION_VAL_VERTICAL_TT },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_TT,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_TT },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_TT },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_TT },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_TT,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_TT },
         { "1GB",   OPTION_VAL_1GB_TT },
         { "2GB",   OPTION_VAL_2GB_TT },
         { "4GB",   OPTION_VAL_4GB_TT },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_TT,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_TT,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_TT,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_TT,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_TT },
         { "1",  OPTION_VAL_1_TT },
         { "2",  OPTION_VAL_2_TT },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_TT,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_TT },
         { "4",  OPTION_VAL_4_TT },
         { "6",  OPTION_VAL_6_TT },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_TT,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_TT,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_TT,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_TT,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_TT,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_TT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_TT,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_TT },
         { "more", OPTION_VAL_MORE_TT },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_TT,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_TT },
         { "110", OPTION_VAL_110_TT },
         { "120", OPTION_VAL_120_TT },
         { "130", OPTION_VAL_130_TT },
         { "140", OPTION_VAL_140_TT },
         { "150", OPTION_VAL_150_TT },
         { "160", OPTION_VAL_160_TT },
         { "170", OPTION_VAL_170_TT },
         { "180", OPTION_VAL_180_TT },
         { "190", OPTION_VAL_190_TT },
         { "200", OPTION_VAL_200_TT },
         { "210", OPTION_VAL_210_TT },
         { "220", OPTION_VAL_220_TT },
         { "230", OPTION_VAL_230_TT },
         { "240", OPTION_VAL_240_TT },
         { "250", OPTION_VAL_250_TT },
         { "260", OPTION_VAL_260_TT },
         { "270", OPTION_VAL_270_TT },
         { "280", OPTION_VAL_280_TT },
         { "290", OPTION_VAL_290_TT },
         { "300", OPTION_VAL_300_TT },
         { "310", OPTION_VAL_310_TT },
         { "320", OPTION_VAL_320_TT },
         { "330", OPTION_VAL_330_TT },
         { "340", OPTION_VAL_340_TT },
         { "350", OPTION_VAL_350_TT },
         { "360", OPTION_VAL_360_TT },
         { "370", OPTION_VAL_370_TT },
         { "380", OPTION_VAL_380_TT },
         { "390", OPTION_VAL_390_TT },
         { "400", OPTION_VAL_400_TT },
         { "410", OPTION_VAL_410_TT },
         { "420", OPTION_VAL_420_TT },
         { "430", OPTION_VAL_430_TT },
         { "440", OPTION_VAL_440_TT },
         { "450", OPTION_VAL_450_TT },
         { "460", OPTION_VAL_460_TT },
         { "470", OPTION_VAL_470_TT },
         { "480", OPTION_VAL_480_TT },
         { "490", OPTION_VAL_490_TT },
         { "500", OPTION_VAL_500_TT },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_TT,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_TT },
         { "5%",  OPTION_VAL_5_TT },
         { "10%", OPTION_VAL_10_TT },
         { "15%", OPTION_VAL_15_TT },
         { "20%", OPTION_VAL_20_TT },
         { "25%", OPTION_VAL_25_TT },
         { "30%", OPTION_VAL_30_TT },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_TT,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_TT },
         { "5%",  OPTION_VAL_5_TT },
         { "10%", OPTION_VAL_10_TT },
         { "15%", OPTION_VAL_15_TT },
         { "20%", OPTION_VAL_20_TT },
         { "25%", OPTION_VAL_25_TT },
         { "30%", OPTION_VAL_30_TT },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_TT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_TT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_TT,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_TT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_TT },
         { "60%",  OPTION_VAL_60_TT },
         { "70%",  OPTION_VAL_70_TT },
         { "80%",  OPTION_VAL_80_TT },
         { "90%",  OPTION_VAL_90_TT },
         { "100%", OPTION_VAL_100_O47_TT },
         { "110%", OPTION_VAL_110_O47_TT },
         { "120%", OPTION_VAL_120_O47_TT },
         { "130%", OPTION_VAL_130_O47_TT },
         { "140%", OPTION_VAL_140_O47_TT },
         { "150%", OPTION_VAL_150_O47_TT },
         { "160%", OPTION_VAL_160_O47_TT },
         { "170%", OPTION_VAL_170_O47_TT },
         { "180%", OPTION_VAL_180_O47_TT },
         { "190%", OPTION_VAL_190_O47_TT },
         { "200%", OPTION_VAL_200_O47_TT },
         { "210%", OPTION_VAL_210_O47_TT },
         { "220%", OPTION_VAL_220_O47_TT },
         { "230%", OPTION_VAL_230_O47_TT },
         { "240%", OPTION_VAL_240_O47_TT },
         { "250%", OPTION_VAL_250_O47_TT },
         { "260%", OPTION_VAL_260_O47_TT },
         { "270%", OPTION_VAL_270_O47_TT },
         { "280%", OPTION_VAL_280_O47_TT },
         { "290%", OPTION_VAL_290_O47_TT },
         { "300%", OPTION_VAL_300_O47_TT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_TT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TT },
         { "Red",      OPTION_VAL_RED_TT },
         { "Green",    OPTION_VAL_GREEN_TT },
         { "Blue",     OPTION_VAL_BLUE_TT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_TT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TT },
         { "Red",      OPTION_VAL_RED_TT },
         { "Green",    OPTION_VAL_GREEN_TT },
         { "Blue",     OPTION_VAL_BLUE_TT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_TT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TT },
         { "Red",      OPTION_VAL_RED_TT },
         { "Green",    OPTION_VAL_GREEN_TT },
         { "Blue",     OPTION_VAL_BLUE_TT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_TT,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_TT },
         { "Red",      OPTION_VAL_RED_TT },
         { "Green",    OPTION_VAL_GREEN_TT },
         { "Blue",     OPTION_VAL_BLUE_TT },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_TT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TT },
         { "Purupuru", OPTION_VAL_PURUPURU_TT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_TT },
         { "None",     OPTION_VAL_NONE_TT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_TT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TT },
         { "Purupuru", OPTION_VAL_PURUPURU_TT },
         { "None",     OPTION_VAL_NONE_TT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_TT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TT },
         { "Purupuru", OPTION_VAL_PURUPURU_TT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_TT },
         { "None",     OPTION_VAL_NONE_TT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_TT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TT },
         { "Purupuru", OPTION_VAL_PURUPURU_TT },
         { "None",     OPTION_VAL_NONE_TT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_TT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TT },
         { "Purupuru", OPTION_VAL_PURUPURU_TT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_TT },
         { "None",     OPTION_VAL_NONE_TT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_TT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TT },
         { "Purupuru", OPTION_VAL_PURUPURU_TT },
         { "None",     OPTION_VAL_NONE_TT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_TT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TT },
         { "Purupuru", OPTION_VAL_PURUPURU_TT },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_TT },
         { "None",     OPTION_VAL_NONE_TT },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_TT,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_TT },
         { "Purupuru", OPTION_VAL_PURUPURU_TT },
         { "None",     OPTION_VAL_NONE_TT },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_TT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_TT,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_TT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_TT },
         { "All VMUs", OPTION_VAL_ALL_VMUS_TT },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_TT,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_TT,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_TT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_TT,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_TT,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_TT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_TT,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_TT,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TT },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TT },
         { "2x", OPTION_VAL_2_O26_TT },
         { "3x", OPTION_VAL_3X_TT },
         { "4x", OPTION_VAL_4_TT },
         { "5x", OPTION_VAL_5X_TT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TT },
         { "BLUE 03",           OPTION_VAL_BLUE_TT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TT },
         { "GREEN 05",          OPTION_VAL_GREEN_TT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TT },
         { "RED 11",            OPTION_VAL_RED_TT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TT },
         { "WHITE 28",          OPTION_VAL_WHITE_TT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TT },
         { "BLUE 03",           OPTION_VAL_BLUE_TT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TT },
         { "GREEN 05",          OPTION_VAL_GREEN_TT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TT },
         { "RED 11",            OPTION_VAL_RED_TT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TT },
         { "WHITE 28",          OPTION_VAL_WHITE_TT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TT },
         { "20%",  OPTION_VAL_20_TT },
         { "30%",  OPTION_VAL_30_TT },
         { "40%",  OPTION_VAL_40_TT },
         { "50%",  OPTION_VAL_50_TT },
         { "60%",  OPTION_VAL_60_TT },
         { "70%",  OPTION_VAL_70_TT },
         { "80%",  OPTION_VAL_80_TT },
         { "90%",  OPTION_VAL_90_TT },
         { "100%", OPTION_VAL_100_O47_TT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TT },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TT },
         { "2x", OPTION_VAL_2_O26_TT },
         { "3x", OPTION_VAL_3X_TT },
         { "4x", OPTION_VAL_4_TT },
         { "5x", OPTION_VAL_5X_TT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TT },
         { "BLUE 03",           OPTION_VAL_BLUE_TT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TT },
         { "GREEN 05",          OPTION_VAL_GREEN_TT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TT },
         { "RED 11",            OPTION_VAL_RED_TT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TT },
         { "WHITE 28",          OPTION_VAL_WHITE_TT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TT },
         { "BLUE 03",           OPTION_VAL_BLUE_TT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TT },
         { "GREEN 05",          OPTION_VAL_GREEN_TT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TT },
         { "RED 11",            OPTION_VAL_RED_TT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TT },
         { "WHITE 28",          OPTION_VAL_WHITE_TT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TT },
         { "20%",  OPTION_VAL_20_TT },
         { "30%",  OPTION_VAL_30_TT },
         { "40%",  OPTION_VAL_40_TT },
         { "50%",  OPTION_VAL_50_TT },
         { "60%",  OPTION_VAL_60_TT },
         { "70%",  OPTION_VAL_70_TT },
         { "80%",  OPTION_VAL_80_TT },
         { "90%",  OPTION_VAL_90_TT },
         { "100%", OPTION_VAL_100_O47_TT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TT },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TT },
         { "2x", OPTION_VAL_2_O26_TT },
         { "3x", OPTION_VAL_3X_TT },
         { "4x", OPTION_VAL_4_TT },
         { "5x", OPTION_VAL_5X_TT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TT },
         { "BLUE 03",           OPTION_VAL_BLUE_TT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TT },
         { "GREEN 05",          OPTION_VAL_GREEN_TT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TT },
         { "RED 11",            OPTION_VAL_RED_TT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TT },
         { "WHITE 28",          OPTION_VAL_WHITE_TT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TT },
         { "BLUE 03",           OPTION_VAL_BLUE_TT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TT },
         { "GREEN 05",          OPTION_VAL_GREEN_TT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TT },
         { "RED 11",            OPTION_VAL_RED_TT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TT },
         { "WHITE 28",          OPTION_VAL_WHITE_TT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TT },
         { "20%",  OPTION_VAL_20_TT },
         { "30%",  OPTION_VAL_30_TT },
         { "40%",  OPTION_VAL_40_TT },
         { "50%",  OPTION_VAL_50_TT },
         { "60%",  OPTION_VAL_60_TT },
         { "70%",  OPTION_VAL_70_TT },
         { "80%",  OPTION_VAL_80_TT },
         { "90%",  OPTION_VAL_90_TT },
         { "100%", OPTION_VAL_100_O47_TT },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_TT },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_TT },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_TT },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_TT },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_TT },
         { "2x", OPTION_VAL_2_O26_TT },
         { "3x", OPTION_VAL_3X_TT },
         { "4x", OPTION_VAL_4_TT },
         { "5x", OPTION_VAL_5X_TT },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TT },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TT },
         { "BLUE 03",           OPTION_VAL_BLUE_TT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TT },
         { "GREEN 05",          OPTION_VAL_GREEN_TT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TT },
         { "RED 11",            OPTION_VAL_RED_TT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TT },
         { "WHITE 28",          OPTION_VAL_WHITE_TT },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_TT },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_TT },
         { "BLACK 02",          OPTION_VAL_BLACK_02_TT },
         { "BLUE 03",           OPTION_VAL_BLUE_TT },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_TT },
         { "GREEN 05",          OPTION_VAL_GREEN_TT },
         { "CYAN 06",           OPTION_VAL_CYAN_06_TT },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_TT },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_TT },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_TT },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_TT },
         { "RED 11",            OPTION_VAL_RED_TT },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_TT },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_TT },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_TT },
         { "GRAY 15",           OPTION_VAL_GRAY_15_TT },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_TT },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_TT },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_TT },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_TT },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_TT },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_TT },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_TT },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_TT },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_TT },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_TT },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_TT },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_TT },
         { "WHITE 28",          OPTION_VAL_WHITE_TT },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_TT,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_TT },
         { "20%",  OPTION_VAL_20_TT },
         { "30%",  OPTION_VAL_30_TT },
         { "40%",  OPTION_VAL_40_TT },
         { "50%",  OPTION_VAL_50_TT },
         { "60%",  OPTION_VAL_60_TT },
         { "70%",  OPTION_VAL_70_TT },
         { "80%",  OPTION_VAL_80_TT },
         { "90%",  OPTION_VAL_90_TT },
         { "100%", OPTION_VAL_100_O47_TT },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_tt = {
   option_cats_tt,
   option_defs_tt
};

/* RETRO_LANGUAGE_UK */

#define CATEGORY_SYSTEM_LABEL_UK "Система"
#define CATEGORY_SYSTEM_INFO_0_UK "Налаштування регіону, мова, BIOS і базові параметри обладнання."
#define CATEGORY_VIDEO_LABEL_UK "Відео"
#define CATEGORY_VIDEO_INFO_0_UK "Налаштувати роздільну здатність, прозорість і візуальні налаштування ефектів."
#define CATEGORY_PERFORMANCE_LABEL_UK "Продуктивність"
#define CATEGORY_PERFORMANCE_INFO_0_UK "Налаштування потокового рендерингу та пропуску кадрів."
#define CATEGORY_HACKS_LABEL_UK "Гаки емуляції"
#define CATEGORY_HACKS_INFO_0_UK "Налаштування поширених змінень, швидкість завантаження GD-ROM та налаштування заміни текстур."
#define CATEGORY_INPUT_LABEL_UK "Ввід"
#define CATEGORY_INPUT_INFO_0_UK "Налаштуйте налаштування геймпада та Підсвітки пістолетів."
#define CATEGORY_EXPANSIONS_LABEL_UK "Збільшення слотів контролера"
#define CATEGORY_EXPANSIONS_INFO_0_UK "Виберіть пристрій (VMU, чудово підключений до кожного слоту розширення контролерів."
#define CATEGORY_VMU_LABEL_UK "Одиниця візуальної пам'яті"
#define CATEGORY_VMU_INFO_0_UK "Налаштування видимості окремої VMU окремо для кожного робочого столу."
#define CORE_OPTION_NAME_REGION_LABEL_UK "Регіон"
#define OPTION_VAL_JAPAN_UK "Японія"
#define OPTION_VAL_USA_UK "США"
#define OPTION_VAL_EUROPE_UK "Європа"
#define OPTION_VAL_DEFAULT_UK "За замовчуванням"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_UK "Мова"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_UK "Змінює мову для BIOS та будь-які ігри, які містять різноманітні мови."
#define OPTION_VAL_JAPANESE_UK "Японська"
#define OPTION_VAL_ENGLISH_UK "Англійська"
#define OPTION_VAL_GERMAN_UK "Німецька"
#define OPTION_VAL_FRENCH_UK "Французька"
#define OPTION_VAL_SPANISH_UK "Іспанська"
#define OPTION_VAL_ITALIAN_UK "Італійська"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_UK "HLE BIOS (потрібен перезапуск)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_UK "Примусово використовувати BIOS високого рівня."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_UK "Увімкнути DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_UK "Увімкнути емуляцію аудіо DSP(цифровий процесор сигналів). Покращує точність згенерованого звуку, але збільшує вимоги до продуктивності."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_UK "Дозволити кнопки служіння Arcade"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_UK "Активує кнопку SERVICE для аркадних ігор, щоб увійти в налаштування кабінету."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_UK "Встановити NAOMI ігри для безкоштовної гри"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_UK "Змінювати параметри гри за допомогою монет \"Вільна гра\"."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_UK "Широкосмугова емуляція адаптера"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_UK "Емуляція Ethernet широкосмугового адаптера замість модема. (Перезавантажити необхідно)"
#define CORE_OPTION_NAME_UPNP_LABEL_UK "Активувати UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_UK "Використовуйте UPnP для автоматичного налаштування інтернет-маршрутизатора для онлайн-ігор."
#define CORE_OPTION_NAME_DCNET_LABEL_UK "Використовувати DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_UK "Використовувати сервіс у хмарі DCNet для доступу в Інтернет Dreamcast."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_UK "Роздільна здатність"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_UK "Змінити роздільну здатність візуалізації."
#define OPTION_VAL_320X240_UK "320х240 (Half)"
#define OPTION_VAL_640X480_UK NULL
#define OPTION_VAL_800X600_UK NULL
#define OPTION_VAL_960X720_UK "960х720 (x1.5)"
#define OPTION_VAL_1024X768_UK "1024х768 (x1.6)"
#define OPTION_VAL_1280X960_UK "1280х960 (x2)"
#define OPTION_VAL_1440X1080_UK NULL
#define OPTION_VAL_1600X1200_UK "1600х1200 (x2.5)"
#define OPTION_VAL_1920X1440_UK "1920х1440 (x3)"
#define OPTION_VAL_2560X1920_UK "2560х1920 (x4)"
#define OPTION_VAL_2880X2160_UK NULL
#define OPTION_VAL_3200X2400_UK "3200х2400 (x5)"
#define OPTION_VAL_3840X2880_UK NULL
#define OPTION_VAL_4480X3360_UK "4480х3360 (x7)"
#define OPTION_VAL_5120X3840_UK NULL
#define OPTION_VAL_5760X4320_UK "5760х4320 (x9)"
#define OPTION_VAL_6400X4800_UK NULL
#define OPTION_VAL_7040X5280_UK "7040х5280 (x11)"
#define OPTION_VAL_7680X5760_UK NULL
#define OPTION_VAL_8320X6240_UK NULL
#define OPTION_VAL_8960X6720_UK NULL
#define OPTION_VAL_9600X7200_UK NULL
#define OPTION_VAL_10240X7680_UK "10240х7680 (x16)"
#define OPTION_VAL_10880X8160_UK "10880х8160 (x17)"
#define OPTION_VAL_11520X8640_UK NULL
#define OPTION_VAL_12160X9120_UK NULL
#define OPTION_VAL_12800X9600_UK NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_UK "Тип кабель"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_UK "Тип вихідного сигналу. 'TV (Composite)' є найбільш поширеним."
#define OPTION_VAL_VGA_UK NULL
#define OPTION_VAL_TV_RGB_UK NULL
#define OPTION_VAL_TV_COMPOSITE_UK "Телевізор (Composite)"
#define CORE_OPTION_NAME_BROADCAST_LABEL_UK "Транслювати стандарт"
#define OPTION_VAL_NTSC_UK NULL
#define OPTION_VAL_PAL_UK "Пал (Світ)"
#define OPTION_VAL_PAL_N_UK NULL
#define OPTION_VAL_PAL_M_UK "PAL-M (Бразилія)"
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_UK "Орієнтація екрану"
#define OPTION_VAL_HORIZONTAL_UK "Горизонтально"
#define OPTION_VAL_VERTICAL_UK "Вертикально"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_UK "Альфа-Сортування"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_UK "Виберіть, як сортуються прозорі багатокутники."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_UK "Per-Strip (швидкий, найточніший)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_UK "Відсоток трикутник (звичайний)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_UK "Per-Pixel (точний, але повільний)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_UK "Розмір буфера Pixel"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_UK "Для коректного виводу можуть бути потрібні більші значення."
#define OPTION_VAL_512MB_UK "512 МБ"
#define OPTION_VAL_1GB_UK "1 ГБ"
#define OPTION_VAL_2GB_UK "2 ГБ"
#define OPTION_VAL_4GB_UK "4 ГБ"
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_UK "Максимальний прозорий шар"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_UK "Вищі значення можуть бути необхідними для складних сцен."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_UK "Повноцінне графічний інтерфейс емуляції"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_UK "Увімкнути повну фреймбуфер в VRAM. Це корисно для ігор, які безпосередньо читають або записують фреймбуфер у VRAM. Коли ввімкнено, Внутрішня роздільність змушена до 640х480, а продуктивність може сильно вплинути."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_UK "Увімкнути Буфер RTT (Render в Texture)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_UK "Копіювання текстур відтворюється з GPU до відеопам'яті. Цей параметр зазвичай увімкнено для ігор, які потребують ігор. Якщо увімкнено, то масштабування текстур вимкнене та може вплинути на продуктивність."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_UK "Зіставлення"
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_UK "Якщо ввімкнені текстури будуть використовувати меншу версію, коли вони з'являються далі, це може збільшити продуктивність і зменшити мереживання."
#define CORE_OPTION_NAME_FOG_LABEL_UK "Туманні ефекти"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_UK NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_UK "Функція Dreamcast GPU для малювання тіней об'єктів. Зазвичай це слід увімкнути - вплив продуктивності зазвичай є мінімальним до незначним."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_UK "Анізотропна фільтрація"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_UK "Підвищення якості текстур на поверхні, які є на забутті кутів візуального перегляду відповідно до камери. Чим вищі значення, тим більше вимагають зміни GPU. Зміни застосовані лише після перезапуску."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_UK "Фільтрація текстур"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_UK "Режим фільтрації текстур для використання. Це можна використати для примусового режиму фільтрації текстур на всіх текстурах для отримання кризисного (або плавного) зовнішнього вигляду, ніж стандартний режим. Значення за замовчуванням можуть спричинити різні проблеми візуалізації. Зміни застосовані лише після перезапуску."
#define OPTION_VAL_1_UK "Форс-Неарест-Нейбор"
#define OPTION_VAL_2_UK "Примусова лінія"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_UK "Затримка заміни кадрів"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_UK "Корисно, щоб уникнути блимання екрану або відео з відблиманням. Не рекомендовано на повільних платформах."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_UK "Визначити зміни частоти кадрів"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_UK "Повідомляти фронт, коли внутрішня частота кадрів (напр. від 60 до 30 fp). Поліпшує отвір у ігри, що запускається за 30 кадрів в секунду або 20 футів, але слід вимкнути для ігор з розблокованими (нестабільними) темпами (наприклад. Ecco, Dolphin, Unreal Tournament). Примітка: Недоступно при увімкненому «Авто-пропуск кадрів»."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_UK "Фільтр постобробки PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_UK "Постобробка зображення для імітації ефектів відеоданих на PowerVR2 та аналогових відео сигналів."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_UK "Вдосконалення текстур (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_UK "Покращення графіки 2D пікселів. Має використовуватися лише з 2D піксельними іграми."
#define OPTION_VAL_2_O26_UK NULL
#define OPTION_VAL_4_UK NULL
#define OPTION_VAL_6_UK NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_UK "Збільшення масштабу текстур. Відфільтрований розмір"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_UK "Виберіть максимальний розмір для масштабування текстури, якщо розмір текстури більший за вибране значення, тоді значення не буде збільшуватися."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_UK "Інтерполяція Рідної глибини"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_UK "Допомагає в корупції текстур та глибинних питаннях AMD GPU. Може також допомогти Intel GPU в деяких випадках."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_UK "Закруглити відстань по краї"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_UK "Підтримує випадок кровотечі текстур під час масштабування. Відключення може допомогти, якщо пікселі попереджають при збільшенні масштабу в 2D іграх (MVC2, CVS, KOF тощо)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_UK "Візуалізація за потоком"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_UK "Запускає відеопроцесор та процесор на різних потоках. Настійно рекомендується."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_UK NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_UK "Автоматично пропускати кадри, коли емулятор працює повільно. Примітка: Цей параметр застосовується лише тоді, коли увімкнуто 'Threaded Rendering'."
#define OPTION_VAL_SOME_UK "Нормально"
#define OPTION_VAL_MORE_UK "Максимум"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_UK "Пропуск кадрів"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_UK "Встановлює кількість кадрів для переходу між кожними відображеними кадрами."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_UK "Широкі коди (потрібен перезапуск)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_UK "Активує чіти, які дозволяють певним іграм відображати у форматі поширеного режиму."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_UK "Широкоекранний Hack"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_UK "Намалювати геометрію поза нормальним співвідношенням сторін 4:3. Може видавати графічні залози в розкритих районах."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_UK "Швидке завантаження GD-ROM (неточне)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_UK "Прискорює завантаження GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_UK NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_UK "Увімкнути 32МБ Мод ОЗУ для Dreamcast. Може вплинути на сумісність"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_UK "SH4 under/overclock процесора"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_UK "Змініть SH4 головний годинник ЦП за замовчуванням 200 МГц. Разом може допомогти уповільнити платформи. Покриття може збільшити частоту кадрів для деяких ігор. Використовуйте з обережністю."
#define OPTION_VAL_100_UK "100 МГц"
#define OPTION_VAL_110_UK NULL
#define OPTION_VAL_120_UK "120 МГц"
#define OPTION_VAL_130_UK "130 МГц"
#define OPTION_VAL_140_UK "140 МГц"
#define OPTION_VAL_150_UK "150 МГц"
#define OPTION_VAL_160_UK "160 МГц"
#define OPTION_VAL_170_UK "170 МГц"
#define OPTION_VAL_180_UK "180 МГц"
#define OPTION_VAL_190_UK "190 МГц"
#define OPTION_VAL_200_UK "200 МГц"
#define OPTION_VAL_210_UK "210 МГц"
#define OPTION_VAL_220_UK "220 МГц"
#define OPTION_VAL_230_UK "230 МГц"
#define OPTION_VAL_240_UK "240 МГц"
#define OPTION_VAL_250_UK "250 МГц"
#define OPTION_VAL_260_UK "260 МГц"
#define OPTION_VAL_270_UK "270 МГц"
#define OPTION_VAL_280_UK "280 МГц"
#define OPTION_VAL_290_UK "290 МГц"
#define OPTION_VAL_300_UK "300 МГц"
#define OPTION_VAL_310_UK "310 МГц"
#define OPTION_VAL_320_UK "320 МГц"
#define OPTION_VAL_330_UK "330 МГц"
#define OPTION_VAL_340_UK "340 МГц"
#define OPTION_VAL_350_UK "350 МГц"
#define OPTION_VAL_360_UK "360 МГц"
#define OPTION_VAL_370_UK "370 МГц"
#define OPTION_VAL_380_UK "380 МГц"
#define OPTION_VAL_390_UK "390 МГц"
#define OPTION_VAL_400_UK "400 МГц"
#define OPTION_VAL_410_UK "410 МГц"
#define OPTION_VAL_420_UK "420 МГц"
#define OPTION_VAL_430_UK "430 МГц"
#define OPTION_VAL_440_UK "440 МГц"
#define OPTION_VAL_450_UK "450 МГц"
#define OPTION_VAL_460_UK "460 МГц"
#define OPTION_VAL_470_UK "470 МГц"
#define OPTION_VAL_480_UK "480 МГц"
#define OPTION_VAL_490_UK "490 МГц"
#define OPTION_VAL_500_UK "500 МГц"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_UK "Завантажити користувацькі текстури"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_UK "Завантажити спеціальні текстури, які можна знайти в теці 'system/dc/textures/<game-id>/'."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_UK "Попередити завантаження користувацьких текстур"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_UK "При запуску гри попередньо завантажуйте спеціальні текстури. Може покращити продуктивність, але збільшує використання пам'яті."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_UK "Дамп текстур"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_UK "Кожного разу, коли гра використовує нову текстуру, вона буде збережена як файл .png у теці 'system/dc/texdump/<game-id>/'."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_UK "Дамп замінених текстур"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_UK "Завжди вивантажувати текстури, які вже замінені спеціальними текстами."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_UK "Аналоговий стік Deadzone"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_UK "Вкажіть, як далеко потрібно перемістити аналоговий стік перед тим, як він почав оброблятися."
#define OPTION_VAL_0_UK NULL
#define OPTION_VAL_5_UK NULL
#define OPTION_VAL_10_UK NULL
#define OPTION_VAL_15_UK NULL
#define OPTION_VAL_20_UK NULL
#define OPTION_VAL_25_UK NULL
#define OPTION_VAL_30_UK NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_UK "Запустити Deadzone"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_UK "Виберіть скільки доведеться натиснути на пристрій перед тим, як почати обробку."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_UK "Цифрові тригери"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_UK "Якщо увімкнено, тригери будуть діяти як звичайні кнопки, означає, що вони будуть оброблені як повністю або повністю натиснуті або взагалі не натиснуті, без проміжків."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_UK "Транслювати цифрові виходи"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_UK "Транслювати цифрові вихідні дані і стан подачі при TCP порту 8000. Сумісно з параметром \"-вихідний мережу\" MAME."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_UK "Показати параметри освітлення"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_UK "Увімкнути налаштування кросшейдерів дисплея. Примітка: швидке меню може бути перезапущено, щоб цей параметр набув чинності."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_UK "Прицільне масштабування для зброї"
#define OPTION_VAL_50_UK NULL
#define OPTION_VAL_60_UK NULL
#define OPTION_VAL_70_UK NULL
#define OPTION_VAL_80_UK NULL
#define OPTION_VAL_90_UK NULL
#define OPTION_VAL_100_O47_UK NULL
#define OPTION_VAL_110_O47_UK NULL
#define OPTION_VAL_120_O47_UK NULL
#define OPTION_VAL_130_O47_UK NULL
#define OPTION_VAL_140_O47_UK NULL
#define OPTION_VAL_150_O47_UK NULL
#define OPTION_VAL_160_O47_UK NULL
#define OPTION_VAL_170_O47_UK NULL
#define OPTION_VAL_180_O47_UK NULL
#define OPTION_VAL_190_O47_UK NULL
#define OPTION_VAL_200_O47_UK NULL
#define OPTION_VAL_210_O47_UK NULL
#define OPTION_VAL_220_O47_UK NULL
#define OPTION_VAL_230_O47_UK NULL
#define OPTION_VAL_240_O47_UK NULL
#define OPTION_VAL_250_O47_UK NULL
#define OPTION_VAL_260_O47_UK NULL
#define OPTION_VAL_270_O47_UK NULL
#define OPTION_VAL_280_O47_UK NULL
#define OPTION_VAL_290_O47_UK NULL
#define OPTION_VAL_300_O47_UK NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_UK "Збройний приціл 1 дисплей"
#define OPTION_VAL_WHITE_UK "Білі"
#define OPTION_VAL_RED_UK "Червоний"
#define OPTION_VAL_GREEN_UK "Зелений"
#define OPTION_VAL_BLUE_UK "Синій"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_UK "Приціл 2 зброю"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_UK "Приціл 3 Зброї"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_UK "Приціл 4 Зброї"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_UK "Пристрій у слоті розширення A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_UK "Виберіть пристрій підключений до слоту розширення A1 (порт A слот 1)."
#define OPTION_VAL_VMU_UK "ВМУ"
#define OPTION_VAL_PURUPURU_UK "Пакунок Вібрації"
#define OPTION_VAL_DREAMPOTATO_UK NULL
#define OPTION_VAL_NONE_UK "Немає"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_UK "Пристрій у слоті розширення A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_UK "Виберіть пристрій підключений до слоту розширення A2 (порт A slot 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_UK "Пристрій у полі розширення B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_UK "Виберіть пристрій підключений до слоту розширення B1 (порт B слот 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_UK "Пристрій у слоті розширення B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_UK "Виберіть пристрій підключений до слоту розширення B2 (порт B слот 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_UK "Пристрій в розширенні слоту C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_UK "Виберіть пристрій підключений до слоту розширення C1 (порт C слот 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_UK "Пристрій у слоті розширення C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_UK "Виберіть пристрій підключений до слоту розширення C2 (порт C слот 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_UK "Пристрій у полі розширення D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_UK "Виберіть пристрій підключений до слоту розширення D1 (порт D слот 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_UK "Пристрій у слоті розширення D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_UK "Виберіть пристрій підключений до слоту розширення D2 (порт D слот 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_UK "Per-Game Visual Memory Units/Системи (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_UK "Виступи для гри"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_UK "Якщо вимкнено, усі ігри спільно використовують до 8 файлів збереження VMU (A1/A2/B1/B2/C1/C2/D1/D2), розташованих у системній папці RetroArch.\nНалаштування «VMU A1» створює унікальний файл VMU «A1». у папці збереження RetroArch для кожної запущеної гри.\nНалаштування «Усі VMU» створює до 8 унікальних файлів VMU (A1/A2/B1/B2/C1/C2/D1/D2) для кожної запущеної гри."
#define OPTION_VAL_VMU_A1_UK "ВМУ А1"
#define OPTION_VAL_ALL_VMUS_UK "Всі VMUs"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_UK "Звуки візуальної пам'яті/системи (VMU)"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_UK "Звук VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_UK "Коли увімкнено, відео відтворюються VMU біки."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_UK NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_UK NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_UK "Показати параметри показу юніту/системи (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_UK "Показати параметри показу vMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_UK "Увімкніть конфігурацію емуляції видимості екрана VMU LCD, розміру, позиції та кольору. Примітка: Швидке Меню потрібно перемкнути для застосування цього параметру."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_UK "Екран VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_UK "Позиція екрану VMU 1"
#define OPTION_VAL_UPPER_LEFT_UK "Верхній лівий"
#define OPTION_VAL_UPPER_RIGHT_UK "Верхній правий"
#define OPTION_VAL_LOWER_LEFT_UK "Менше ліворуч"
#define OPTION_VAL_LOWER_RIGHT_UK "Нижній правий"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_UK "Розмір екрану VMU 1"
#define OPTION_VAL_1X_UK NULL
#define OPTION_VAL_3X_UK NULL
#define OPTION_VAL_5X_UK "5х"
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_UK "ВМУ 1 піксель за кольором"
#define OPTION_VAL_DEFAULT_ON_00_UK "УВІМК"
#define OPTION_VAL_DEFAULT_OFF_01_UK "Вимкнено за промовчанням"
#define OPTION_VAL_BLACK_02_UK "Чорний колір"
#define OPTION_VAL_LIGHT_BLUE_04_UK "Світло-синій"
#define OPTION_VAL_CYAN_06_UK "Бірюзовий"
#define OPTION_VAL_CYAN_BLUE_07_UK "Блакитно-синій"
#define OPTION_VAL_LIGHT_GREEN_08_UK "Світло-зелений"
#define OPTION_VAL_CYAN_GREEN_09_UK "Бірюзово-зелений"
#define OPTION_VAL_LIGHT_CYAN_10_UK "Світлий блакитний"
#define OPTION_VAL_PURPLE_12_UK "Пурпурний"
#define OPTION_VAL_LIGHT_PURPLE_13_UK "Світло-фіолетовий"
#define OPTION_VAL_YELLOW_14_UK "Жовтий"
#define OPTION_VAL_GRAY_15_UK "Сірий"
#define OPTION_VAL_LIGHT_PURPLE_2_16_UK "Світло-фіолетовий (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_UK "Світло-зелений (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_UK "Світло-зелений (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_UK "Бірюзовий кут (2)"
#define OPTION_VAL_LIGHT_RED_2_20_UK "Світло-червоний (2)"
#define OPTION_VAL_MAGENTA_21_UK "Малиновий"
#define OPTION_VAL_LIGHT_PURPLE_3_22_UK "Світло-бузковий (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_UK "Світло-оранжевий"
#define OPTION_VAL_ORANGE_24_UK "Помаранчевий"
#define OPTION_VAL_LIGHT_PURPLE_4_25_UK "Світло-бузковий (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_UK "Світлий жовтий"
#define OPTION_VAL_LIGHT_YELLOW_2_27_UK "Світлий жовтий (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_UK "Колір дисплея VMU екрану 1 за замовчуванням"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_UK "Прозорість екрану VMU 1"
#define OPTION_VAL_40_UK NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_UK "Екран VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_UK "Позиція екрану VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_UK "Розмір екрану VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_UK "VMU екран 2 пікселя за кольором"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_UK "Колір екрану VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_UK "Прозорість екрану VMU"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_UK "Екран VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_UK "Позиція екрану VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_UK "Розмір екрану VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_UK "Екран 3 пікселя VMU за кольором"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_UK "Колір при вимкненому екрані 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_UK "Прозорість екрану VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_UK "Екран VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_UK "Позиція екрану VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_UK "Розмір екрану VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_UK "VMU екран 4 пікселя за кольором"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_UK "Колір при вимкненому екрані 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_UK "Прозорість екрану VMU 4"

struct retro_core_option_v2_category option_cats_uk[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_UK,
      CATEGORY_SYSTEM_INFO_0_UK
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_UK,
      CATEGORY_VIDEO_INFO_0_UK
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_UK,
      CATEGORY_PERFORMANCE_INFO_0_UK
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_UK,
      CATEGORY_HACKS_INFO_0_UK
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_UK,
      CATEGORY_INPUT_INFO_0_UK
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_UK,
      CATEGORY_EXPANSIONS_INFO_0_UK
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_UK,
      CATEGORY_VMU_INFO_0_UK
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_uk[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_UK,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_UK },
         { "USA",     OPTION_VAL_USA_UK },
         { "Europe",  OPTION_VAL_EUROPE_UK },
         { "Default", OPTION_VAL_DEFAULT_UK },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_UK,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_UK },
         { "English",  OPTION_VAL_ENGLISH_UK },
         { "German",   OPTION_VAL_GERMAN_UK },
         { "French",   OPTION_VAL_FRENCH_UK },
         { "Spanish",  OPTION_VAL_SPANISH_UK },
         { "Italian",  OPTION_VAL_ITALIAN_UK },
         { "Default",  OPTION_VAL_DEFAULT_UK },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_UK,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_UK },
         { "640x480",    OPTION_VAL_640X480_UK },
         { "800x600",    OPTION_VAL_800X600_UK },
         { "960x720",    OPTION_VAL_960X720_UK },
         { "1024x768",   OPTION_VAL_1024X768_UK },
         { "1280x960",   OPTION_VAL_1280X960_UK },
         { "1440x1080",  OPTION_VAL_1440X1080_UK },
         { "1600x1200",  OPTION_VAL_1600X1200_UK },
         { "1920x1440",  OPTION_VAL_1920X1440_UK },
         { "2560x1920",  OPTION_VAL_2560X1920_UK },
         { "2880x2160",  OPTION_VAL_2880X2160_UK },
         { "3200x2400",  OPTION_VAL_3200X2400_UK },
         { "3840x2880",  OPTION_VAL_3840X2880_UK },
         { "4480x3360",  OPTION_VAL_4480X3360_UK },
         { "5120x3840",  OPTION_VAL_5120X3840_UK },
         { "5760x4320",  OPTION_VAL_5760X4320_UK },
         { "6400x4800",  OPTION_VAL_6400X4800_UK },
         { "7040x5280",  OPTION_VAL_7040X5280_UK },
         { "7680x5760",  OPTION_VAL_7680X5760_UK },
         { "8320x6240",  OPTION_VAL_8320X6240_UK },
         { "8960x6720",  OPTION_VAL_8960X6720_UK },
         { "9600x7200",  OPTION_VAL_9600X7200_UK },
         { "10240x7680", OPTION_VAL_10240X7680_UK },
         { "10880x8160", OPTION_VAL_10880X8160_UK },
         { "11520x8640", OPTION_VAL_11520X8640_UK },
         { "12160x9120", OPTION_VAL_12160X9120_UK },
         { "12800x9600", OPTION_VAL_12800X9600_UK },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_UK,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_UK },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_UK },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_UK },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_UK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_UK },
         { "PAL",     OPTION_VAL_PAL_UK },
         { "PAL_N",   OPTION_VAL_PAL_N_UK },
         { "PAL_M",   OPTION_VAL_PAL_M_UK },
         { "Default", OPTION_VAL_DEFAULT_UK },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_UK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_UK },
         { "vertical",   OPTION_VAL_VERTICAL_UK },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_UK,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_UK },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_UK },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_UK },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_UK,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_UK },
         { "1GB",   OPTION_VAL_1GB_UK },
         { "2GB",   OPTION_VAL_2GB_UK },
         { "4GB",   OPTION_VAL_4GB_UK },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_UK,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_UK,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_UK,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_UK,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_UK },
         { "1",  OPTION_VAL_1_UK },
         { "2",  OPTION_VAL_2_UK },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_UK,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_UK },
         { "4",  OPTION_VAL_4_UK },
         { "6",  OPTION_VAL_6_UK },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_UK,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_UK,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_UK,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_UK,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_UK,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_UK,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_UK,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_UK },
         { "more", OPTION_VAL_MORE_UK },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_UK,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_UK },
         { "110", OPTION_VAL_110_UK },
         { "120", OPTION_VAL_120_UK },
         { "130", OPTION_VAL_130_UK },
         { "140", OPTION_VAL_140_UK },
         { "150", OPTION_VAL_150_UK },
         { "160", OPTION_VAL_160_UK },
         { "170", OPTION_VAL_170_UK },
         { "180", OPTION_VAL_180_UK },
         { "190", OPTION_VAL_190_UK },
         { "200", OPTION_VAL_200_UK },
         { "210", OPTION_VAL_210_UK },
         { "220", OPTION_VAL_220_UK },
         { "230", OPTION_VAL_230_UK },
         { "240", OPTION_VAL_240_UK },
         { "250", OPTION_VAL_250_UK },
         { "260", OPTION_VAL_260_UK },
         { "270", OPTION_VAL_270_UK },
         { "280", OPTION_VAL_280_UK },
         { "290", OPTION_VAL_290_UK },
         { "300", OPTION_VAL_300_UK },
         { "310", OPTION_VAL_310_UK },
         { "320", OPTION_VAL_320_UK },
         { "330", OPTION_VAL_330_UK },
         { "340", OPTION_VAL_340_UK },
         { "350", OPTION_VAL_350_UK },
         { "360", OPTION_VAL_360_UK },
         { "370", OPTION_VAL_370_UK },
         { "380", OPTION_VAL_380_UK },
         { "390", OPTION_VAL_390_UK },
         { "400", OPTION_VAL_400_UK },
         { "410", OPTION_VAL_410_UK },
         { "420", OPTION_VAL_420_UK },
         { "430", OPTION_VAL_430_UK },
         { "440", OPTION_VAL_440_UK },
         { "450", OPTION_VAL_450_UK },
         { "460", OPTION_VAL_460_UK },
         { "470", OPTION_VAL_470_UK },
         { "480", OPTION_VAL_480_UK },
         { "490", OPTION_VAL_490_UK },
         { "500", OPTION_VAL_500_UK },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_UK,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_UK },
         { "5%",  OPTION_VAL_5_UK },
         { "10%", OPTION_VAL_10_UK },
         { "15%", OPTION_VAL_15_UK },
         { "20%", OPTION_VAL_20_UK },
         { "25%", OPTION_VAL_25_UK },
         { "30%", OPTION_VAL_30_UK },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_UK,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_UK },
         { "5%",  OPTION_VAL_5_UK },
         { "10%", OPTION_VAL_10_UK },
         { "15%", OPTION_VAL_15_UK },
         { "20%", OPTION_VAL_20_UK },
         { "25%", OPTION_VAL_25_UK },
         { "30%", OPTION_VAL_30_UK },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_UK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_UK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_UK,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_UK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_UK },
         { "60%",  OPTION_VAL_60_UK },
         { "70%",  OPTION_VAL_70_UK },
         { "80%",  OPTION_VAL_80_UK },
         { "90%",  OPTION_VAL_90_UK },
         { "100%", OPTION_VAL_100_O47_UK },
         { "110%", OPTION_VAL_110_O47_UK },
         { "120%", OPTION_VAL_120_O47_UK },
         { "130%", OPTION_VAL_130_O47_UK },
         { "140%", OPTION_VAL_140_O47_UK },
         { "150%", OPTION_VAL_150_O47_UK },
         { "160%", OPTION_VAL_160_O47_UK },
         { "170%", OPTION_VAL_170_O47_UK },
         { "180%", OPTION_VAL_180_O47_UK },
         { "190%", OPTION_VAL_190_O47_UK },
         { "200%", OPTION_VAL_200_O47_UK },
         { "210%", OPTION_VAL_210_O47_UK },
         { "220%", OPTION_VAL_220_O47_UK },
         { "230%", OPTION_VAL_230_O47_UK },
         { "240%", OPTION_VAL_240_O47_UK },
         { "250%", OPTION_VAL_250_O47_UK },
         { "260%", OPTION_VAL_260_O47_UK },
         { "270%", OPTION_VAL_270_O47_UK },
         { "280%", OPTION_VAL_280_O47_UK },
         { "290%", OPTION_VAL_290_O47_UK },
         { "300%", OPTION_VAL_300_O47_UK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_UK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_UK },
         { "Red",      OPTION_VAL_RED_UK },
         { "Green",    OPTION_VAL_GREEN_UK },
         { "Blue",     OPTION_VAL_BLUE_UK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_UK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_UK },
         { "Red",      OPTION_VAL_RED_UK },
         { "Green",    OPTION_VAL_GREEN_UK },
         { "Blue",     OPTION_VAL_BLUE_UK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_UK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_UK },
         { "Red",      OPTION_VAL_RED_UK },
         { "Green",    OPTION_VAL_GREEN_UK },
         { "Blue",     OPTION_VAL_BLUE_UK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_UK,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_UK },
         { "Red",      OPTION_VAL_RED_UK },
         { "Green",    OPTION_VAL_GREEN_UK },
         { "Blue",     OPTION_VAL_BLUE_UK },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_UK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_UK },
         { "Purupuru", OPTION_VAL_PURUPURU_UK },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_UK },
         { "None",     OPTION_VAL_NONE_UK },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_UK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_UK },
         { "Purupuru", OPTION_VAL_PURUPURU_UK },
         { "None",     OPTION_VAL_NONE_UK },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_UK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_UK },
         { "Purupuru", OPTION_VAL_PURUPURU_UK },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_UK },
         { "None",     OPTION_VAL_NONE_UK },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_UK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_UK },
         { "Purupuru", OPTION_VAL_PURUPURU_UK },
         { "None",     OPTION_VAL_NONE_UK },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_UK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_UK },
         { "Purupuru", OPTION_VAL_PURUPURU_UK },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_UK },
         { "None",     OPTION_VAL_NONE_UK },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_UK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_UK },
         { "Purupuru", OPTION_VAL_PURUPURU_UK },
         { "None",     OPTION_VAL_NONE_UK },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_UK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_UK },
         { "Purupuru", OPTION_VAL_PURUPURU_UK },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_UK },
         { "None",     OPTION_VAL_NONE_UK },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_UK,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_UK },
         { "Purupuru", OPTION_VAL_PURUPURU_UK },
         { "None",     OPTION_VAL_NONE_UK },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_UK,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_UK,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_UK,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_UK },
         { "All VMUs", OPTION_VAL_ALL_VMUS_UK },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_UK,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_UK,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_UK,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_UK,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_UK,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_UK,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_UK,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_UK,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_UK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_UK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_UK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_UK },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_UK },
         { "2x", OPTION_VAL_2_O26_UK },
         { "3x", OPTION_VAL_3X_UK },
         { "4x", OPTION_VAL_4_UK },
         { "5x", OPTION_VAL_5X_UK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_UK },
         { "20%",  OPTION_VAL_20_UK },
         { "30%",  OPTION_VAL_30_UK },
         { "40%",  OPTION_VAL_40_UK },
         { "50%",  OPTION_VAL_50_UK },
         { "60%",  OPTION_VAL_60_UK },
         { "70%",  OPTION_VAL_70_UK },
         { "80%",  OPTION_VAL_80_UK },
         { "90%",  OPTION_VAL_90_UK },
         { "100%", OPTION_VAL_100_O47_UK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_UK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_UK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_UK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_UK },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_UK },
         { "2x", OPTION_VAL_2_O26_UK },
         { "3x", OPTION_VAL_3X_UK },
         { "4x", OPTION_VAL_4_UK },
         { "5x", OPTION_VAL_5X_UK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_UK },
         { "20%",  OPTION_VAL_20_UK },
         { "30%",  OPTION_VAL_30_UK },
         { "40%",  OPTION_VAL_40_UK },
         { "50%",  OPTION_VAL_50_UK },
         { "60%",  OPTION_VAL_60_UK },
         { "70%",  OPTION_VAL_70_UK },
         { "80%",  OPTION_VAL_80_UK },
         { "90%",  OPTION_VAL_90_UK },
         { "100%", OPTION_VAL_100_O47_UK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_UK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_UK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_UK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_UK },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_UK },
         { "2x", OPTION_VAL_2_O26_UK },
         { "3x", OPTION_VAL_3X_UK },
         { "4x", OPTION_VAL_4_UK },
         { "5x", OPTION_VAL_5X_UK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_UK },
         { "20%",  OPTION_VAL_20_UK },
         { "30%",  OPTION_VAL_30_UK },
         { "40%",  OPTION_VAL_40_UK },
         { "50%",  OPTION_VAL_50_UK },
         { "60%",  OPTION_VAL_60_UK },
         { "70%",  OPTION_VAL_70_UK },
         { "80%",  OPTION_VAL_80_UK },
         { "90%",  OPTION_VAL_90_UK },
         { "100%", OPTION_VAL_100_O47_UK },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_UK },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_UK },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_UK },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_UK },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_UK },
         { "2x", OPTION_VAL_2_O26_UK },
         { "3x", OPTION_VAL_3X_UK },
         { "4x", OPTION_VAL_4_UK },
         { "5x", OPTION_VAL_5X_UK },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_UK },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_UK },
         { "BLACK 02",          OPTION_VAL_BLACK_02_UK },
         { "BLUE 03",           OPTION_VAL_BLUE_UK },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_UK },
         { "GREEN 05",          OPTION_VAL_GREEN_UK },
         { "CYAN 06",           OPTION_VAL_CYAN_06_UK },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_UK },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_UK },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_UK },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_UK },
         { "RED 11",            OPTION_VAL_RED_UK },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_UK },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_UK },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_UK },
         { "GRAY 15",           OPTION_VAL_GRAY_15_UK },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_UK },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_UK },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_UK },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_UK },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_UK },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_UK },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_UK },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_UK },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_UK },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_UK },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_UK },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_UK },
         { "WHITE 28",          OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_UK,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_UK },
         { "20%",  OPTION_VAL_20_UK },
         { "30%",  OPTION_VAL_30_UK },
         { "40%",  OPTION_VAL_40_UK },
         { "50%",  OPTION_VAL_50_UK },
         { "60%",  OPTION_VAL_60_UK },
         { "70%",  OPTION_VAL_70_UK },
         { "80%",  OPTION_VAL_80_UK },
         { "90%",  OPTION_VAL_90_UK },
         { "100%", OPTION_VAL_100_O47_UK },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_uk = {
   option_cats_uk,
   option_defs_uk
};

/* RETRO_LANGUAGE_VAL */

#define CATEGORY_SYSTEM_LABEL_VAL "Sistema"
#define CATEGORY_SYSTEM_INFO_0_VAL "Configura regió, llengua, BIOS i ajustos de maquinari."
#define CATEGORY_VIDEO_LABEL_VAL "Vídeo"
#define CATEGORY_VIDEO_INFO_0_VAL NULL
#define CATEGORY_PERFORMANCE_LABEL_VAL NULL
#define CATEGORY_PERFORMANCE_INFO_0_VAL NULL
#define CATEGORY_HACKS_LABEL_VAL NULL
#define CATEGORY_HACKS_INFO_0_VAL NULL
#define CATEGORY_INPUT_LABEL_VAL "Entrada"
#define CATEGORY_INPUT_INFO_0_VAL NULL
#define CATEGORY_EXPANSIONS_LABEL_VAL NULL
#define CATEGORY_EXPANSIONS_INFO_0_VAL NULL
#define CATEGORY_VMU_LABEL_VAL NULL
#define CATEGORY_VMU_INFO_0_VAL NULL
#define CORE_OPTION_NAME_REGION_LABEL_VAL "Regió"
#define OPTION_VAL_JAPAN_VAL NULL
#define OPTION_VAL_USA_VAL NULL
#define OPTION_VAL_EUROPE_VAL NULL
#define OPTION_VAL_DEFAULT_VAL NULL
#define CORE_OPTION_NAME_LANGUAGE_LABEL_VAL "Llengua"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_VAL NULL
#define OPTION_VAL_JAPANESE_VAL NULL
#define OPTION_VAL_ENGLISH_VAL NULL
#define OPTION_VAL_GERMAN_VAL NULL
#define OPTION_VAL_FRENCH_VAL NULL
#define OPTION_VAL_SPANISH_VAL NULL
#define OPTION_VAL_ITALIAN_VAL NULL
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_VAL NULL
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_VAL NULL
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_VAL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_VAL NULL
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_VAL NULL
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_VAL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_VAL NULL
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_VAL NULL
#define CORE_OPTION_NAME_UPNP_LABEL_VAL NULL
#define CORE_OPTION_NAME_UPNP_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DCNET_LABEL_VAL NULL
#define CORE_OPTION_NAME_DCNET_INFO_0_VAL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_VAL NULL
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_VAL NULL
#define OPTION_VAL_320X240_VAL NULL
#define OPTION_VAL_640X480_VAL NULL
#define OPTION_VAL_800X600_VAL NULL
#define OPTION_VAL_960X720_VAL NULL
#define OPTION_VAL_1024X768_VAL NULL
#define OPTION_VAL_1280X960_VAL NULL
#define OPTION_VAL_1440X1080_VAL NULL
#define OPTION_VAL_1600X1200_VAL NULL
#define OPTION_VAL_1920X1440_VAL NULL
#define OPTION_VAL_2560X1920_VAL NULL
#define OPTION_VAL_2880X2160_VAL NULL
#define OPTION_VAL_3200X2400_VAL NULL
#define OPTION_VAL_3840X2880_VAL NULL
#define OPTION_VAL_4480X3360_VAL NULL
#define OPTION_VAL_5120X3840_VAL NULL
#define OPTION_VAL_5760X4320_VAL NULL
#define OPTION_VAL_6400X4800_VAL NULL
#define OPTION_VAL_7040X5280_VAL NULL
#define OPTION_VAL_7680X5760_VAL NULL
#define OPTION_VAL_8320X6240_VAL NULL
#define OPTION_VAL_8960X6720_VAL NULL
#define OPTION_VAL_9600X7200_VAL NULL
#define OPTION_VAL_10240X7680_VAL NULL
#define OPTION_VAL_10880X8160_VAL NULL
#define OPTION_VAL_11520X8640_VAL NULL
#define OPTION_VAL_12160X9120_VAL NULL
#define OPTION_VAL_12800X9600_VAL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_VAL NULL
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_VAL NULL
#define OPTION_VAL_VGA_VAL NULL
#define OPTION_VAL_TV_RGB_VAL NULL
#define OPTION_VAL_TV_COMPOSITE_VAL NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_VAL NULL
#define OPTION_VAL_NTSC_VAL NULL
#define OPTION_VAL_PAL_VAL NULL
#define OPTION_VAL_PAL_N_VAL NULL
#define OPTION_VAL_PAL_M_VAL NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_VAL "Orientació de pantalla"
#define OPTION_VAL_HORIZONTAL_VAL NULL
#define OPTION_VAL_VERTICAL_VAL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_VAL NULL
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_VAL NULL
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_VAL NULL
#define OPTION_VAL_PER_TRIANGLE_NORMAL_VAL NULL
#define OPTION_VAL_PER_PIXEL_ACCURATE_VAL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_VAL NULL
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_VAL NULL
#define OPTION_VAL_512MB_VAL NULL
#define OPTION_VAL_1GB_VAL NULL
#define OPTION_VAL_2GB_VAL NULL
#define OPTION_VAL_4GB_VAL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_VAL NULL
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_VAL NULL
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_VAL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_VAL NULL
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_VAL NULL
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_VAL NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_FOG_LABEL_VAL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_VAL NULL
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_VAL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_VAL NULL
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_VAL NULL
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_VAL NULL
#define OPTION_VAL_1_VAL NULL
#define OPTION_VAL_2_VAL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_VAL NULL
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_VAL NULL
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_VAL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_VAL NULL
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_VAL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_VAL NULL
#define OPTION_VAL_2_O26_VAL NULL
#define OPTION_VAL_4_VAL NULL
#define OPTION_VAL_6_VAL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_VAL NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_VAL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_VAL NULL
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_VAL NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_VAL NULL
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_VAL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_VAL NULL
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_VAL NULL
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_VAL NULL
#define OPTION_VAL_SOME_VAL NULL
#define OPTION_VAL_MORE_VAL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_VAL NULL
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_VAL NULL
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_VAL NULL
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_VAL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_VAL NULL
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_VAL NULL
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_VAL NULL
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_VAL NULL
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_VAL NULL
#define OPTION_VAL_100_VAL NULL
#define OPTION_VAL_110_VAL NULL
#define OPTION_VAL_120_VAL NULL
#define OPTION_VAL_130_VAL NULL
#define OPTION_VAL_140_VAL NULL
#define OPTION_VAL_150_VAL NULL
#define OPTION_VAL_160_VAL NULL
#define OPTION_VAL_170_VAL NULL
#define OPTION_VAL_180_VAL NULL
#define OPTION_VAL_190_VAL NULL
#define OPTION_VAL_200_VAL NULL
#define OPTION_VAL_210_VAL NULL
#define OPTION_VAL_220_VAL NULL
#define OPTION_VAL_230_VAL NULL
#define OPTION_VAL_240_VAL NULL
#define OPTION_VAL_250_VAL NULL
#define OPTION_VAL_260_VAL NULL
#define OPTION_VAL_270_VAL NULL
#define OPTION_VAL_280_VAL NULL
#define OPTION_VAL_290_VAL NULL
#define OPTION_VAL_300_VAL NULL
#define OPTION_VAL_310_VAL NULL
#define OPTION_VAL_320_VAL NULL
#define OPTION_VAL_330_VAL NULL
#define OPTION_VAL_340_VAL NULL
#define OPTION_VAL_350_VAL NULL
#define OPTION_VAL_360_VAL NULL
#define OPTION_VAL_370_VAL NULL
#define OPTION_VAL_380_VAL NULL
#define OPTION_VAL_390_VAL NULL
#define OPTION_VAL_400_VAL NULL
#define OPTION_VAL_410_VAL NULL
#define OPTION_VAL_420_VAL NULL
#define OPTION_VAL_430_VAL NULL
#define OPTION_VAL_440_VAL NULL
#define OPTION_VAL_450_VAL NULL
#define OPTION_VAL_460_VAL NULL
#define OPTION_VAL_470_VAL NULL
#define OPTION_VAL_480_VAL NULL
#define OPTION_VAL_490_VAL NULL
#define OPTION_VAL_500_VAL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_VAL NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_VAL NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_VAL NULL
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_VAL NULL
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_VAL NULL
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_VAL NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_VAL NULL
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_VAL NULL
#define OPTION_VAL_0_VAL NULL
#define OPTION_VAL_5_VAL NULL
#define OPTION_VAL_10_VAL NULL
#define OPTION_VAL_15_VAL NULL
#define OPTION_VAL_20_VAL NULL
#define OPTION_VAL_25_VAL NULL
#define OPTION_VAL_30_VAL NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_VAL NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_VAL NULL
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_VAL NULL
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_VAL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_VAL NULL
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_VAL NULL
#define OPTION_VAL_50_VAL NULL
#define OPTION_VAL_60_VAL NULL
#define OPTION_VAL_70_VAL NULL
#define OPTION_VAL_80_VAL NULL
#define OPTION_VAL_90_VAL NULL
#define OPTION_VAL_100_O47_VAL NULL
#define OPTION_VAL_110_O47_VAL NULL
#define OPTION_VAL_120_O47_VAL NULL
#define OPTION_VAL_130_O47_VAL NULL
#define OPTION_VAL_140_O47_VAL NULL
#define OPTION_VAL_150_O47_VAL NULL
#define OPTION_VAL_160_O47_VAL NULL
#define OPTION_VAL_170_O47_VAL NULL
#define OPTION_VAL_180_O47_VAL NULL
#define OPTION_VAL_190_O47_VAL NULL
#define OPTION_VAL_200_O47_VAL NULL
#define OPTION_VAL_210_O47_VAL NULL
#define OPTION_VAL_220_O47_VAL NULL
#define OPTION_VAL_230_O47_VAL NULL
#define OPTION_VAL_240_O47_VAL NULL
#define OPTION_VAL_250_O47_VAL NULL
#define OPTION_VAL_260_O47_VAL NULL
#define OPTION_VAL_270_O47_VAL NULL
#define OPTION_VAL_280_O47_VAL NULL
#define OPTION_VAL_290_O47_VAL NULL
#define OPTION_VAL_300_O47_VAL NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_VAL NULL
#define OPTION_VAL_WHITE_VAL NULL
#define OPTION_VAL_RED_VAL NULL
#define OPTION_VAL_GREEN_VAL NULL
#define OPTION_VAL_BLUE_VAL NULL
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_VAL NULL
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_VAL NULL
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_VAL NULL
#define OPTION_VAL_VMU_VAL NULL
#define OPTION_VAL_PURUPURU_VAL NULL
#define OPTION_VAL_DREAMPOTATO_VAL NULL
#define OPTION_VAL_NONE_VAL "Cap"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_VAL NULL
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_VAL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_VAL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_VAL NULL
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_VAL NULL
#define OPTION_VAL_VMU_A1_VAL NULL
#define OPTION_VAL_ALL_VMUS_VAL NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_VAL NULL
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_VAL NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_VAL NULL
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_VAL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_VAL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_VAL NULL
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_VAL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_VAL NULL
#define OPTION_VAL_UPPER_LEFT_VAL NULL
#define OPTION_VAL_UPPER_RIGHT_VAL NULL
#define OPTION_VAL_LOWER_LEFT_VAL NULL
#define OPTION_VAL_LOWER_RIGHT_VAL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_VAL NULL
#define OPTION_VAL_1X_VAL NULL
#define OPTION_VAL_3X_VAL NULL
#define OPTION_VAL_5X_VAL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_VAL NULL
#define OPTION_VAL_DEFAULT_ON_00_VAL NULL
#define OPTION_VAL_DEFAULT_OFF_01_VAL NULL
#define OPTION_VAL_BLACK_02_VAL NULL
#define OPTION_VAL_LIGHT_BLUE_04_VAL NULL
#define OPTION_VAL_CYAN_06_VAL NULL
#define OPTION_VAL_CYAN_BLUE_07_VAL NULL
#define OPTION_VAL_LIGHT_GREEN_08_VAL NULL
#define OPTION_VAL_CYAN_GREEN_09_VAL NULL
#define OPTION_VAL_LIGHT_CYAN_10_VAL NULL
#define OPTION_VAL_PURPLE_12_VAL NULL
#define OPTION_VAL_LIGHT_PURPLE_13_VAL NULL
#define OPTION_VAL_YELLOW_14_VAL NULL
#define OPTION_VAL_GRAY_15_VAL NULL
#define OPTION_VAL_LIGHT_PURPLE_2_16_VAL NULL
#define OPTION_VAL_LIGHT_GREEN_2_17_VAL NULL
#define OPTION_VAL_LIGHT_GREEN_3_18_VAL NULL
#define OPTION_VAL_LIGHT_CYAN_2_19_VAL NULL
#define OPTION_VAL_LIGHT_RED_2_20_VAL NULL
#define OPTION_VAL_MAGENTA_21_VAL NULL
#define OPTION_VAL_LIGHT_PURPLE_3_22_VAL NULL
#define OPTION_VAL_LIGHT_ORANGE_23_VAL NULL
#define OPTION_VAL_ORANGE_24_VAL NULL
#define OPTION_VAL_LIGHT_PURPLE_4_25_VAL NULL
#define OPTION_VAL_LIGHT_YELLOW_26_VAL NULL
#define OPTION_VAL_LIGHT_YELLOW_2_27_VAL NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_VAL NULL
#define OPTION_VAL_40_VAL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_VAL NULL
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_VAL NULL

struct retro_core_option_v2_category option_cats_val[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_VAL,
      CATEGORY_SYSTEM_INFO_0_VAL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_VAL,
      CATEGORY_VIDEO_INFO_0_VAL
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_VAL,
      CATEGORY_PERFORMANCE_INFO_0_VAL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_VAL,
      CATEGORY_HACKS_INFO_0_VAL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_VAL,
      CATEGORY_INPUT_INFO_0_VAL
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_VAL,
      CATEGORY_EXPANSIONS_INFO_0_VAL
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_VAL,
      CATEGORY_VMU_INFO_0_VAL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_val[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_VAL,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_VAL },
         { "USA",     OPTION_VAL_USA_VAL },
         { "Europe",  OPTION_VAL_EUROPE_VAL },
         { "Default", OPTION_VAL_DEFAULT_VAL },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_VAL,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_VAL },
         { "English",  OPTION_VAL_ENGLISH_VAL },
         { "German",   OPTION_VAL_GERMAN_VAL },
         { "French",   OPTION_VAL_FRENCH_VAL },
         { "Spanish",  OPTION_VAL_SPANISH_VAL },
         { "Italian",  OPTION_VAL_ITALIAN_VAL },
         { "Default",  OPTION_VAL_DEFAULT_VAL },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_VAL,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_VAL },
         { "640x480",    OPTION_VAL_640X480_VAL },
         { "800x600",    OPTION_VAL_800X600_VAL },
         { "960x720",    OPTION_VAL_960X720_VAL },
         { "1024x768",   OPTION_VAL_1024X768_VAL },
         { "1280x960",   OPTION_VAL_1280X960_VAL },
         { "1440x1080",  OPTION_VAL_1440X1080_VAL },
         { "1600x1200",  OPTION_VAL_1600X1200_VAL },
         { "1920x1440",  OPTION_VAL_1920X1440_VAL },
         { "2560x1920",  OPTION_VAL_2560X1920_VAL },
         { "2880x2160",  OPTION_VAL_2880X2160_VAL },
         { "3200x2400",  OPTION_VAL_3200X2400_VAL },
         { "3840x2880",  OPTION_VAL_3840X2880_VAL },
         { "4480x3360",  OPTION_VAL_4480X3360_VAL },
         { "5120x3840",  OPTION_VAL_5120X3840_VAL },
         { "5760x4320",  OPTION_VAL_5760X4320_VAL },
         { "6400x4800",  OPTION_VAL_6400X4800_VAL },
         { "7040x5280",  OPTION_VAL_7040X5280_VAL },
         { "7680x5760",  OPTION_VAL_7680X5760_VAL },
         { "8320x6240",  OPTION_VAL_8320X6240_VAL },
         { "8960x6720",  OPTION_VAL_8960X6720_VAL },
         { "9600x7200",  OPTION_VAL_9600X7200_VAL },
         { "10240x7680", OPTION_VAL_10240X7680_VAL },
         { "10880x8160", OPTION_VAL_10880X8160_VAL },
         { "11520x8640", OPTION_VAL_11520X8640_VAL },
         { "12160x9120", OPTION_VAL_12160X9120_VAL },
         { "12800x9600", OPTION_VAL_12800X9600_VAL },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_VAL },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_VAL },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_VAL },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_VAL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_VAL },
         { "PAL",     OPTION_VAL_PAL_VAL },
         { "PAL_N",   OPTION_VAL_PAL_N_VAL },
         { "PAL_M",   OPTION_VAL_PAL_M_VAL },
         { "Default", OPTION_VAL_DEFAULT_VAL },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_VAL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_VAL },
         { "vertical",   OPTION_VAL_VERTICAL_VAL },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_VAL },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_VAL },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_VAL },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_VAL },
         { "1GB",   OPTION_VAL_1GB_VAL },
         { "2GB",   OPTION_VAL_2GB_VAL },
         { "4GB",   OPTION_VAL_4GB_VAL },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_VAL,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_VAL,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_VAL },
         { "1",  OPTION_VAL_1_VAL },
         { "2",  OPTION_VAL_2_VAL },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_VAL },
         { "4",  OPTION_VAL_4_VAL },
         { "6",  OPTION_VAL_6_VAL },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_VAL,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_VAL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_VAL,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_VAL,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_VAL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_VAL,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_VAL },
         { "more", OPTION_VAL_MORE_VAL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_VAL,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_VAL },
         { "110", OPTION_VAL_110_VAL },
         { "120", OPTION_VAL_120_VAL },
         { "130", OPTION_VAL_130_VAL },
         { "140", OPTION_VAL_140_VAL },
         { "150", OPTION_VAL_150_VAL },
         { "160", OPTION_VAL_160_VAL },
         { "170", OPTION_VAL_170_VAL },
         { "180", OPTION_VAL_180_VAL },
         { "190", OPTION_VAL_190_VAL },
         { "200", OPTION_VAL_200_VAL },
         { "210", OPTION_VAL_210_VAL },
         { "220", OPTION_VAL_220_VAL },
         { "230", OPTION_VAL_230_VAL },
         { "240", OPTION_VAL_240_VAL },
         { "250", OPTION_VAL_250_VAL },
         { "260", OPTION_VAL_260_VAL },
         { "270", OPTION_VAL_270_VAL },
         { "280", OPTION_VAL_280_VAL },
         { "290", OPTION_VAL_290_VAL },
         { "300", OPTION_VAL_300_VAL },
         { "310", OPTION_VAL_310_VAL },
         { "320", OPTION_VAL_320_VAL },
         { "330", OPTION_VAL_330_VAL },
         { "340", OPTION_VAL_340_VAL },
         { "350", OPTION_VAL_350_VAL },
         { "360", OPTION_VAL_360_VAL },
         { "370", OPTION_VAL_370_VAL },
         { "380", OPTION_VAL_380_VAL },
         { "390", OPTION_VAL_390_VAL },
         { "400", OPTION_VAL_400_VAL },
         { "410", OPTION_VAL_410_VAL },
         { "420", OPTION_VAL_420_VAL },
         { "430", OPTION_VAL_430_VAL },
         { "440", OPTION_VAL_440_VAL },
         { "450", OPTION_VAL_450_VAL },
         { "460", OPTION_VAL_460_VAL },
         { "470", OPTION_VAL_470_VAL },
         { "480", OPTION_VAL_480_VAL },
         { "490", OPTION_VAL_490_VAL },
         { "500", OPTION_VAL_500_VAL },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_VAL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_VAL },
         { "5%",  OPTION_VAL_5_VAL },
         { "10%", OPTION_VAL_10_VAL },
         { "15%", OPTION_VAL_15_VAL },
         { "20%", OPTION_VAL_20_VAL },
         { "25%", OPTION_VAL_25_VAL },
         { "30%", OPTION_VAL_30_VAL },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_VAL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_VAL },
         { "5%",  OPTION_VAL_5_VAL },
         { "10%", OPTION_VAL_10_VAL },
         { "15%", OPTION_VAL_15_VAL },
         { "20%", OPTION_VAL_20_VAL },
         { "25%", OPTION_VAL_25_VAL },
         { "30%", OPTION_VAL_30_VAL },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_VAL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_VAL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_VAL,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_VAL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_VAL },
         { "60%",  OPTION_VAL_60_VAL },
         { "70%",  OPTION_VAL_70_VAL },
         { "80%",  OPTION_VAL_80_VAL },
         { "90%",  OPTION_VAL_90_VAL },
         { "100%", OPTION_VAL_100_O47_VAL },
         { "110%", OPTION_VAL_110_O47_VAL },
         { "120%", OPTION_VAL_120_O47_VAL },
         { "130%", OPTION_VAL_130_O47_VAL },
         { "140%", OPTION_VAL_140_O47_VAL },
         { "150%", OPTION_VAL_150_O47_VAL },
         { "160%", OPTION_VAL_160_O47_VAL },
         { "170%", OPTION_VAL_170_O47_VAL },
         { "180%", OPTION_VAL_180_O47_VAL },
         { "190%", OPTION_VAL_190_O47_VAL },
         { "200%", OPTION_VAL_200_O47_VAL },
         { "210%", OPTION_VAL_210_O47_VAL },
         { "220%", OPTION_VAL_220_O47_VAL },
         { "230%", OPTION_VAL_230_O47_VAL },
         { "240%", OPTION_VAL_240_O47_VAL },
         { "250%", OPTION_VAL_250_O47_VAL },
         { "260%", OPTION_VAL_260_O47_VAL },
         { "270%", OPTION_VAL_270_O47_VAL },
         { "280%", OPTION_VAL_280_O47_VAL },
         { "290%", OPTION_VAL_290_O47_VAL },
         { "300%", OPTION_VAL_300_O47_VAL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VAL },
         { "Red",      OPTION_VAL_RED_VAL },
         { "Green",    OPTION_VAL_GREEN_VAL },
         { "Blue",     OPTION_VAL_BLUE_VAL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VAL },
         { "Red",      OPTION_VAL_RED_VAL },
         { "Green",    OPTION_VAL_GREEN_VAL },
         { "Blue",     OPTION_VAL_BLUE_VAL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VAL },
         { "Red",      OPTION_VAL_RED_VAL },
         { "Green",    OPTION_VAL_GREEN_VAL },
         { "Blue",     OPTION_VAL_BLUE_VAL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VAL },
         { "Red",      OPTION_VAL_RED_VAL },
         { "Green",    OPTION_VAL_GREEN_VAL },
         { "Blue",     OPTION_VAL_BLUE_VAL },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_VAL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VAL },
         { "Purupuru", OPTION_VAL_PURUPURU_VAL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_VAL },
         { "None",     OPTION_VAL_NONE_VAL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_VAL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VAL },
         { "Purupuru", OPTION_VAL_PURUPURU_VAL },
         { "None",     OPTION_VAL_NONE_VAL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_VAL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VAL },
         { "Purupuru", OPTION_VAL_PURUPURU_VAL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_VAL },
         { "None",     OPTION_VAL_NONE_VAL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_VAL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VAL },
         { "Purupuru", OPTION_VAL_PURUPURU_VAL },
         { "None",     OPTION_VAL_NONE_VAL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_VAL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VAL },
         { "Purupuru", OPTION_VAL_PURUPURU_VAL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_VAL },
         { "None",     OPTION_VAL_NONE_VAL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_VAL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VAL },
         { "Purupuru", OPTION_VAL_PURUPURU_VAL },
         { "None",     OPTION_VAL_NONE_VAL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_VAL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VAL },
         { "Purupuru", OPTION_VAL_PURUPURU_VAL },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_VAL },
         { "None",     OPTION_VAL_NONE_VAL },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_VAL,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VAL },
         { "Purupuru", OPTION_VAL_PURUPURU_VAL },
         { "None",     OPTION_VAL_NONE_VAL },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_VAL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_VAL,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_VAL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_VAL },
         { "All VMUs", OPTION_VAL_ALL_VMUS_VAL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_VAL,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_VAL,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_VAL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_VAL,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_VAL,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_VAL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_VAL,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_VAL,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VAL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VAL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VAL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VAL },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VAL },
         { "2x", OPTION_VAL_2_O26_VAL },
         { "3x", OPTION_VAL_3X_VAL },
         { "4x", OPTION_VAL_4_VAL },
         { "5x", OPTION_VAL_5X_VAL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VAL },
         { "20%",  OPTION_VAL_20_VAL },
         { "30%",  OPTION_VAL_30_VAL },
         { "40%",  OPTION_VAL_40_VAL },
         { "50%",  OPTION_VAL_50_VAL },
         { "60%",  OPTION_VAL_60_VAL },
         { "70%",  OPTION_VAL_70_VAL },
         { "80%",  OPTION_VAL_80_VAL },
         { "90%",  OPTION_VAL_90_VAL },
         { "100%", OPTION_VAL_100_O47_VAL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VAL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VAL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VAL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VAL },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VAL },
         { "2x", OPTION_VAL_2_O26_VAL },
         { "3x", OPTION_VAL_3X_VAL },
         { "4x", OPTION_VAL_4_VAL },
         { "5x", OPTION_VAL_5X_VAL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VAL },
         { "20%",  OPTION_VAL_20_VAL },
         { "30%",  OPTION_VAL_30_VAL },
         { "40%",  OPTION_VAL_40_VAL },
         { "50%",  OPTION_VAL_50_VAL },
         { "60%",  OPTION_VAL_60_VAL },
         { "70%",  OPTION_VAL_70_VAL },
         { "80%",  OPTION_VAL_80_VAL },
         { "90%",  OPTION_VAL_90_VAL },
         { "100%", OPTION_VAL_100_O47_VAL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VAL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VAL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VAL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VAL },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VAL },
         { "2x", OPTION_VAL_2_O26_VAL },
         { "3x", OPTION_VAL_3X_VAL },
         { "4x", OPTION_VAL_4_VAL },
         { "5x", OPTION_VAL_5X_VAL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VAL },
         { "20%",  OPTION_VAL_20_VAL },
         { "30%",  OPTION_VAL_30_VAL },
         { "40%",  OPTION_VAL_40_VAL },
         { "50%",  OPTION_VAL_50_VAL },
         { "60%",  OPTION_VAL_60_VAL },
         { "70%",  OPTION_VAL_70_VAL },
         { "80%",  OPTION_VAL_80_VAL },
         { "90%",  OPTION_VAL_90_VAL },
         { "100%", OPTION_VAL_100_O47_VAL },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VAL },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VAL },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VAL },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VAL },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VAL },
         { "2x", OPTION_VAL_2_O26_VAL },
         { "3x", OPTION_VAL_3X_VAL },
         { "4x", OPTION_VAL_4_VAL },
         { "5x", OPTION_VAL_5X_VAL },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VAL },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VAL },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VAL },
         { "BLUE 03",           OPTION_VAL_BLUE_VAL },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VAL },
         { "GREEN 05",          OPTION_VAL_GREEN_VAL },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VAL },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VAL },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VAL },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VAL },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VAL },
         { "RED 11",            OPTION_VAL_RED_VAL },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VAL },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VAL },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VAL },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VAL },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VAL },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VAL },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VAL },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VAL },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VAL },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VAL },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VAL },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VAL },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VAL },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VAL },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VAL },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VAL },
         { "WHITE 28",          OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_VAL,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VAL },
         { "20%",  OPTION_VAL_20_VAL },
         { "30%",  OPTION_VAL_30_VAL },
         { "40%",  OPTION_VAL_40_VAL },
         { "50%",  OPTION_VAL_50_VAL },
         { "60%",  OPTION_VAL_60_VAL },
         { "70%",  OPTION_VAL_70_VAL },
         { "80%",  OPTION_VAL_80_VAL },
         { "90%",  OPTION_VAL_90_VAL },
         { "100%", OPTION_VAL_100_O47_VAL },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_val = {
   option_cats_val,
   option_defs_val
};

/* RETRO_LANGUAGE_VN */

#define CATEGORY_SYSTEM_LABEL_VN "Hệ thống"
#define CATEGORY_SYSTEM_INFO_0_VN "Cấu hình vùng, ngôn ngữ, BIOS và phần cứng cơ bản."
#define CATEGORY_VIDEO_LABEL_VN NULL
#define CATEGORY_VIDEO_INFO_0_VN "Cấu hình độ phân giải, trong suốt không phụ thuộc thứ tự và hiệu ứng hình ảnh."
#define CATEGORY_PERFORMANCE_LABEL_VN "Hiệu năng"
#define CATEGORY_PERFORMANCE_INFO_0_VN "Cấu hình kết xuất đa luồng và bỏ khung hình."
#define CATEGORY_HACKS_LABEL_VN "Thủ thuật giả lập"
#define CATEGORY_HACKS_INFO_0_VN "Cấu hình chế độ màn hình rộng, tốc độ tải GD-ROM và thay thế texture."
#define CATEGORY_INPUT_LABEL_VN "Đều khiển"
#define CATEGORY_INPUT_INFO_0_VN "Cấu hình tay cầm và súng ánh sáng."
#define CATEGORY_EXPANSIONS_LABEL_VN "Khe mở rộng tay cầm"
#define CATEGORY_EXPANSIONS_INFO_0_VN "Chọn thiết bị (VMU, thiết bị rung) cắm vào mỗi khe mở rộng tay cầm."
#define CATEGORY_VMU_LABEL_VN "Bộ nhớ trực quan Vmu"
#define CATEGORY_VMU_INFO_0_VN "Cấu hình file lưu VMU theo game và hiển thị VMU trên màn hình."
#define CORE_OPTION_NAME_REGION_LABEL_VN "Khu vực"
#define OPTION_VAL_JAPAN_VN "Nhật Bản"
#define OPTION_VAL_USA_VN "Mỹ"
#define OPTION_VAL_EUROPE_VN "Châu Âu"
#define OPTION_VAL_DEFAULT_VN "Mặc định"
#define CORE_OPTION_NAME_LANGUAGE_LABEL_VN "Ngôn ngữ"
#define CORE_OPTION_NAME_LANGUAGE_INFO_0_VN "Thay đổi ngôn ngữ được BIOS và các game nhiều ngôn ngữ sử dụng."
#define OPTION_VAL_JAPANESE_VN "Tiếng Nhật"
#define OPTION_VAL_ENGLISH_VN "Tiếng Anh"
#define OPTION_VAL_GERMAN_VN "Tiếng Đức"
#define OPTION_VAL_FRENCH_VN "Tiếng Pháp"
#define OPTION_VAL_SPANISH_VN "Tiếng Tây Ban Nha"
#define OPTION_VAL_ITALIAN_VN "Tiếng Ý"
#define CORE_OPTION_NAME_HLE_BIOS_LABEL_VN "HLE BIOS (Cần khởi động lại)"
#define CORE_OPTION_NAME_HLE_BIOS_INFO_0_VN "Ép sử dụng BIOS giả lập cấp cao."
#define CORE_OPTION_NAME_ENABLE_DSP_LABEL_VN "Bật DSP"
#define CORE_OPTION_NAME_ENABLE_DSP_INFO_0_VN "Bật mô phỏng DSP âm thanh (bộ xử lý tín hiệu số) của Dreamcast. Cải thiện độ chính xác của âm thanh phát ra, nhưng tăng yêu cầu về hiệu năng."
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_VN "Cho phép nút Service cho Arcade"
#define CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_VN "Bật nút SERVICE để truy cập cài đặt trong máy arcade."
#define CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_VN "Chế độ Free Play cho game NAOMI"
#define CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_VN "Chỉnh cài đặt đồng xu của trò chơi sang chế độ chơi miễn phí."
#define CORE_OPTION_NAME_EMULATE_BBA_LABEL_VN "Giả lập Bộ chuyển đổi Băng thông Rộng"
#define CORE_OPTION_NAME_EMULATE_BBA_INFO_0_VN "Giả lập bộ chuyển đổi Ethernet băng thông rộng thay vì modem. (Cần khởi động lại)"
#define CORE_OPTION_NAME_UPNP_LABEL_VN "Bật UPnP"
#define CORE_OPTION_NAME_UPNP_INFO_0_VN "Sử dụng UPnP để tự động cấu hình router Internet cho trò chơi trực tuyến."
#define CORE_OPTION_NAME_DCNET_LABEL_VN "Sử dụng DCNet"
#define CORE_OPTION_NAME_DCNET_INFO_0_VN "Sử dụng dịch vụ đám mây DCNet để truy cập Internet trên Dreamcast."
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_VN "Độ phân giải nội bộ"
#define CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_VN "Thay đổi độ phân giải khi render."
#define OPTION_VAL_320X240_VN "320x240 (Một nửa)"
#define OPTION_VAL_640X480_VN "640x480 (Nguyên bản)"
#define OPTION_VAL_800X600_VN NULL
#define OPTION_VAL_960X720_VN NULL
#define OPTION_VAL_1024X768_VN NULL
#define OPTION_VAL_1280X960_VN NULL
#define OPTION_VAL_1440X1080_VN NULL
#define OPTION_VAL_1600X1200_VN NULL
#define OPTION_VAL_1920X1440_VN NULL
#define OPTION_VAL_2560X1920_VN NULL
#define OPTION_VAL_2880X2160_VN NULL
#define OPTION_VAL_3200X2400_VN NULL
#define OPTION_VAL_3840X2880_VN NULL
#define OPTION_VAL_4480X3360_VN NULL
#define OPTION_VAL_5120X3840_VN NULL
#define OPTION_VAL_5760X4320_VN NULL
#define OPTION_VAL_6400X4800_VN NULL
#define OPTION_VAL_7040X5280_VN NULL
#define OPTION_VAL_7680X5760_VN NULL
#define OPTION_VAL_8320X6240_VN NULL
#define OPTION_VAL_8960X6720_VN NULL
#define OPTION_VAL_9600X7200_VN NULL
#define OPTION_VAL_10240X7680_VN NULL
#define OPTION_VAL_10880X8160_VN NULL
#define OPTION_VAL_11520X8640_VN NULL
#define OPTION_VAL_12160X9120_VN NULL
#define OPTION_VAL_12800X9600_VN NULL
#define CORE_OPTION_NAME_CABLE_TYPE_LABEL_VN "Loại cáp"
#define CORE_OPTION_NAME_CABLE_TYPE_INFO_0_VN "Loại tín hiệu xuất. 'TV (Composite)' được hỗ trợ rộng rãi nhất."
#define OPTION_VAL_VGA_VN NULL
#define OPTION_VAL_TV_RGB_VN NULL
#define OPTION_VAL_TV_COMPOSITE_VN NULL
#define CORE_OPTION_NAME_BROADCAST_LABEL_VN "Chuẩn Phát sóng"
#define OPTION_VAL_NTSC_VN NULL
#define OPTION_VAL_PAL_VN "PAL (Toàn cầu)"
#define OPTION_VAL_PAL_N_VN NULL
#define OPTION_VAL_PAL_M_VN NULL
#define CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_VN "Hướng màn hình"
#define OPTION_VAL_HORIZONTAL_VN "Ngang"
#define OPTION_VAL_VERTICAL_VN "Dọc"
#define CORE_OPTION_NAME_ALPHA_SORTING_LABEL_VN "Sắp xếp Alpha"
#define CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_VN "Chọn cách sắp xếp các đa giác trong suốt."
#define OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_VN "Theo Dải (nhanh, chính xác thấp nhất)"
#define OPTION_VAL_PER_TRIANGLE_NORMAL_VN "Theo Tam giác (bình thường)"
#define OPTION_VAL_PER_PIXEL_ACCURATE_VN "Theo từng điểm ảnh (chính xác, nhưng chậm nhất)"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_VN "Kích thước bộ đệm điểm ảnh tích lũy"
#define CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_VN "Các giá trị cao hơn có thể cần thiết để hiển thị đúng ở độ phân giải cao hơn."
#define OPTION_VAL_512MB_VN NULL
#define OPTION_VAL_1GB_VN NULL
#define OPTION_VAL_2GB_VN NULL
#define OPTION_VAL_4GB_VN NULL
#define CORE_OPTION_NAME_OIT_LAYERS_LABEL_VN "Số lớp trong suốt tối đa"
#define CORE_OPTION_NAME_OIT_LAYERS_INFO_0_VN "Giá trị cao hơn có thể cần cho các cảnh phức tạp."
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_VN "Mô phỏng toàn bộ framebuffer"
#define CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_VN "Bật mô phỏng toàn bộ framebuffer trong VRAM. Điều này hữu ích cho các trò chơi đọc hoặc ghi trực tiếp framebuffer trong VRAM. Khi bật, độ phân giải nội bộ bị ép ở 640x480 và hiệu năng có thể bị ảnh hưởng nghiêm trọng."
#define CORE_OPTION_NAME_ENABLE_RTTB_LABEL_VN "Bật bộ đệm RTT (Render To Texture)"
#define CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_VN "Sao chép các texture đã kết xuất từ GPU trở lại VRAM. Tùy chọn này thường được bật cho các trò chơi yêu cầu. Khi bật, việc tăng cường độ phân giải texture bị vô hiệu và hiệu năng có thể bị ảnh hưởng."
#define CORE_OPTION_NAME_MIPMAPPING_LABEL_VN NULL
#define CORE_OPTION_NAME_MIPMAPPING_INFO_0_VN "Khi bật, texture sẽ sử dụng phiên bản nhỏ hơn của chính nó khi xuất hiện ở xa hơn, có thể tăng hiệu năng và giảm hiện tượng nhấp nháy."
#define CORE_OPTION_NAME_FOG_LABEL_VN "Hiệu ứng sương mù"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_VN "Điều chỉnh âm lượng"
#define CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_VN "Một tính năng của GPU Dreamcast, thường được các trò chơi sử dụng để vẽ bóng của các đối tượng. Thông thường nên bật tính năng này — ảnh hưởng đến hiệu năng thường rất ít hoặc hầu như không đáng kể."
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_VN "Lọc dị hướng"
#define CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_VN "Cải thiện chất lượng texture trên các bề mặt ở góc nhìn xiên so với camera. Giá trị cao hơn sẽ đòi hỏi GPU nhiều hơn. Thay đổi thiết lập này chỉ có hiệu lực sau khi khởi động lại."
#define CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_VN "Lọc kết cấu"
#define CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_VN "Chế độ lọc texture cần sử dụng. Có thể ép buộc một chế độ lọc texture nhất định cho tất cả texture để có hình ảnh sắc nét (hoặc mượt mà) hơn so với Mặc định. Các giá trị khác ngoài Mặc định có thể gây ra nhiều vấn đề kết xuất. Thay đổi thiết lập này chỉ có hiệu lực sau khi khởi động lại."
#define OPTION_VAL_1_VN "Ép dùng Nearest-Neighbor"
#define OPTION_VAL_2_VN "Ép dùng Linear"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_VN "Độ trễ đổi khung hình"
#define CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_VN "Hữu ích để tránh màn hình nhấp nháy hoặc video bị giật.Không khuyến nghị trên thiết bị chậm."
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_VN "Phát hiện sự thay đổi tốc độ khung hình"
#define CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_VN "Thông báo cho giao diện khi tốc độ khung hình thay đổi (ví dụ từ 60 fps xuống 30 fps). Cải thiện nhịp khung hình trong game chạy cố định 30 fps hoặc 20 fps, nhưng nên tắt cho các game có tốc độ khung hình không cố định (ví dụ Ecco the Dolphin, Unreal Tournament). Lưu ý: không khả dụng khi 'Bỏ qua khung hình tự động' bật."
#define CORE_OPTION_NAME_PVR2_FILTERING_LABEL_VN "Bộ lọc hậu xử lý PowerVR2"
#define CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_VN "Hậu xử lý hình ảnh đã render để mô phỏng các hiệu ứng đặc trưng của GPU PowerVR2 và tín hiệu video analog."
#define CORE_OPTION_NAME_TEXUPSCALE_LABEL_VN "Phóng to kết cấu (xBRZ)"
#define CORE_OPTION_NAME_TEXUPSCALE_INFO_0_VN "Cải thiện đồ họa 2D vẽ tay dạng pixel. Chỉ nên dùng cho các game 2D pixel."
#define OPTION_VAL_2_O26_VN NULL
#define OPTION_VAL_4_VN NULL
#define OPTION_VAL_6_VN NULL
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_VN "Kích thước tối đa khi phóng to kết cấu đã lọc"
#define CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_VN "Chọn giá trị kích thước tối đa của texture khi tăng độ phân giải.Nếu kích thước texture lớn hơn giá trị chọn, nó sẽ không được tăng."
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_VN "Nội suy độ sâu gốc"
#define CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_VN "Giúp giảm lỗi texture và vấn đề độ sâu trên GPU AMD. Cũng có thể hỗ trợ GPU Intel trong một số trường hợp."
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_VN "Khắc phục tràn màu khi tăng độ phân giải"
#define CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_VN "Giúp xử lý tình trạng tràn màu khi tăng độ phân giải.Tắt tính năng này có thể giúp nếu pixel bị méo khi tăng độ phân giải trong game 2D (MVC2, CVS, KOF…)"
#define CORE_OPTION_NAME_THREADED_RENDERING_LABEL_VN "Kết xuất đa luồng"
#define CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_VN "Chạy GPU và CPU trên các luồng riêng. Rất khuyến nghị."
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_VN "Bỏ qua khung hình tự động"
#define CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_VN "Tự động bỏ qua khung hình khi trình giả lập chạy chậm. Lưu ý: chỉ áp dụng khi 'Kết xuất đa luồng' bật."
#define OPTION_VAL_SOME_VN "Bình thường"
#define OPTION_VAL_MORE_VN "Tối đa"
#define CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_VN "Bỏ qua khung hình"
#define CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_VN "Thiết lập số khung hình bỏ qua giữa các khung hình hiển thị."
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_VN "Mã gian lận màn hình rộng (cần khởi động lại)"
#define CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_VN "Kích hoạt các mã gian lận cho phép một số trò chơi hiển thị ở chế độ màn hình rộng."
#define CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_VN "Mở rộng màn hình"
#define CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_VN "Vẽ hình học ngoài tỉ lệ khung hình 4:3 chuẩn. Có thể gây lỗi đồ họa ở các vùng hiển thị thêm."
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_VN "Tăng tốc tải GD-ROM (không chính xác)"
#define CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_VN "Tăng tốc quá trình tải GD-ROM."
#define CORE_OPTION_NAME_DC_32MB_MOD_LABEL_VN "Chế độ RAM 32MB cho Dreamcast"
#define CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_VN "Bật chế độ RAM 32MB cho Dreamcast. Có thể ảnh hưởng đến khả năng tương thích"
#define CORE_OPTION_NAME_SH4CLOCK_LABEL_VN "Điều chỉnh xung CPU SH4"
#define CORE_OPTION_NAME_SH4CLOCK_INFO_0_VN "Thay đổi xung chính của CPU SH4 từ mức mặc định 200 MHz. Giảm xung có thể giúp các nền tảng chậm. Tăng xung có thể tăng tốc khung hình cho một số trò chơi. Sử dụng cẩn thận."
#define OPTION_VAL_100_VN NULL
#define OPTION_VAL_110_VN NULL
#define OPTION_VAL_120_VN NULL
#define OPTION_VAL_130_VN NULL
#define OPTION_VAL_140_VN NULL
#define OPTION_VAL_150_VN NULL
#define OPTION_VAL_160_VN NULL
#define OPTION_VAL_170_VN NULL
#define OPTION_VAL_180_VN NULL
#define OPTION_VAL_190_VN NULL
#define OPTION_VAL_200_VN NULL
#define OPTION_VAL_210_VN NULL
#define OPTION_VAL_220_VN NULL
#define OPTION_VAL_230_VN NULL
#define OPTION_VAL_240_VN NULL
#define OPTION_VAL_250_VN NULL
#define OPTION_VAL_260_VN NULL
#define OPTION_VAL_270_VN NULL
#define OPTION_VAL_280_VN NULL
#define OPTION_VAL_290_VN NULL
#define OPTION_VAL_300_VN NULL
#define OPTION_VAL_310_VN NULL
#define OPTION_VAL_320_VN NULL
#define OPTION_VAL_330_VN NULL
#define OPTION_VAL_340_VN NULL
#define OPTION_VAL_350_VN NULL
#define OPTION_VAL_360_VN NULL
#define OPTION_VAL_370_VN NULL
#define OPTION_VAL_380_VN NULL
#define OPTION_VAL_390_VN NULL
#define OPTION_VAL_400_VN NULL
#define OPTION_VAL_410_VN NULL
#define OPTION_VAL_420_VN NULL
#define OPTION_VAL_430_VN NULL
#define OPTION_VAL_440_VN NULL
#define OPTION_VAL_450_VN NULL
#define OPTION_VAL_460_VN NULL
#define OPTION_VAL_470_VN NULL
#define OPTION_VAL_480_VN NULL
#define OPTION_VAL_490_VN NULL
#define OPTION_VAL_500_VN NULL
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_VN "Tải Textures Tùy Chỉnh"
#define CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_VN "Tải các texture tùy chỉnh từ thư mục 'system/dc/textures/<game-id>/'."
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_VN "Tải trước Texture Tùy Chỉnh"
#define CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_VN "Tải trước các texture tùy chỉnh khi khởi động trò chơi. Có thể cải thiện hiệu năng nhưng sẽ tăng mức sử dụng bộ nhớ."
#define CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_VN "Xuất kết cấu"
#define CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_VN "Mỗi khi trò chơi sử dụng một kết cấu mới, nó sẽ được lưu dưới dạng tệp .png trong thư mục 'system/dc/texdump/<game-id>/'."
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_VN "Xuất Texture Đã Thay Thế"
#define CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_VN "Luôn xuất các texture đã được thay thế bằng texture tùy chỉnh."
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_VN "Vùng chết của cần analog"
#define CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_VN "Chọn mức độ di chuyển cần thiết của cần analog trước khi nó bắt đầu được xử lý."
#define OPTION_VAL_0_VN NULL
#define OPTION_VAL_5_VN NULL
#define OPTION_VAL_10_VN NULL
#define OPTION_VAL_15_VN NULL
#define OPTION_VAL_20_VN NULL
#define OPTION_VAL_25_VN NULL
#define OPTION_VAL_30_VN NULL
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_VN "Vùng chết của nút kích hoạt - trigger"
#define CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_VN "Chọn lực nhấn cần thiết của nút trigger trước khi nó bắt đầu được xử lý."
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_VN "Nút kích hoạt kỹ thuật số"
#define CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_VN "Khi bật, các trigger sẽ hoạt động như nút thường, nghĩa là chỉ được xử lý ở trạng thái nhấn hoàn toàn hoặc không nhấn, không có mức trung gian."
#define CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_VN "Phát tín hiệu Digital Outputs"
#define CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_VN "Phát tín hiệu digital và trạng thái phản hồi lực qua cổng TCP 8000. Tương thích với tùy chọn \"-output network\" của MAME."
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_VN "Hiển thị Cài đặt Súng Ánh Sáng"
#define CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_VN "Bật cấu hình các tùy chọn hiển thị con trỏ súng. Lưu ý: Có thể cần mở Quick Menu để thiết lập này có hiệu lực."
#define CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_VN "Tỉ lệ kích thước con trỏ súng"
#define OPTION_VAL_50_VN NULL
#define OPTION_VAL_60_VN NULL
#define OPTION_VAL_70_VN NULL
#define OPTION_VAL_80_VN NULL
#define OPTION_VAL_90_VN NULL
#define OPTION_VAL_100_O47_VN NULL
#define OPTION_VAL_110_O47_VN NULL
#define OPTION_VAL_120_O47_VN NULL
#define OPTION_VAL_130_O47_VN NULL
#define OPTION_VAL_140_O47_VN NULL
#define OPTION_VAL_150_O47_VN NULL
#define OPTION_VAL_160_O47_VN NULL
#define OPTION_VAL_170_O47_VN NULL
#define OPTION_VAL_180_O47_VN NULL
#define OPTION_VAL_190_O47_VN NULL
#define OPTION_VAL_200_O47_VN NULL
#define OPTION_VAL_210_O47_VN NULL
#define OPTION_VAL_220_O47_VN NULL
#define OPTION_VAL_230_O47_VN NULL
#define OPTION_VAL_240_O47_VN NULL
#define OPTION_VAL_250_O47_VN NULL
#define OPTION_VAL_260_O47_VN NULL
#define OPTION_VAL_270_O47_VN NULL
#define OPTION_VAL_280_O47_VN NULL
#define OPTION_VAL_290_O47_VN NULL
#define OPTION_VAL_300_O47_VN NULL
#define CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_VN "Hiển thị Tâm ngắm 1"
#define OPTION_VAL_WHITE_VN "Trắng"
#define OPTION_VAL_RED_VN "Đỏ"
#define OPTION_VAL_GREEN_VN "Xanh lá"
#define OPTION_VAL_BLUE_VN "Xanh dương"
#define CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_VN "Hiển thị Tâm ngắm 2"
#define CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_VN "Hiển thị Tâm ngắm 3"
#define CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_VN "Hiển thị Tâm ngắm 4"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_VN "Thiết bị ở khe mở rộng A1"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_VN "Chọn thiết bị cắm ở khe mở rộng A1 (cổng A, khe 1)."
#define OPTION_VAL_VMU_VN NULL
#define OPTION_VAL_PURUPURU_VN "Gói rung tay cầm"
#define OPTION_VAL_DREAMPOTATO_VN NULL
#define OPTION_VAL_NONE_VN "Không"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_VN "Thiết bị trong khe mở rộng A2"
#define CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_VN "Chọn thiết bị cắm vào khe mở rộng A2 (cổng A khe 2)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_VN "Thiết bị trong khe mở rộng B1"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_VN "Chọn thiết bị cắm vào khe mở rộng B1 (cổng B khe 1)."
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_VN "Thiết bị trong khe mở rộng B2"
#define CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_VN "Chọn thiết bị cắm vào khe mở rộng B2 (cổng B khe 2)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_VN "Thiết bị trong khe mở rộng C1"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_VN "Chọn thiết bị cắm vào khe mở rộng C1 (cổng C khe 1)."
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_VN "Thiết bị trong khe mở rộng C2"
#define CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_VN "Chọn thiết bị cắm vào khe mở rộng C2 (cổng C khe 2)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_VN "Thiết bị trong khe mở rộng D1"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_VN "Chọn thiết bị cắm vào khe mở rộng D1 (cổng D khe 1)."
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_VN "Thiết bị trong khe mở rộng D2"
#define CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_VN "Chọn thiết bị cắm vào khe mở rộng D2 (cổng D khe 2)."
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_VN "Đơn vị/ Hệ thống Bộ Nhớ Hình Ảnh theo từng trò chơi (VMU)"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_VN "VMU theo trò chơi"
#define CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_VN "Khi tắt, tất cả trò chơi chia sẻ tối đa 8 file lưu VMU (A1/A2/B1/B2/C1/C2/D1/D2) nằm trong thư mục hệ thống của RetroArch..\nCài đặt 'VMU A1' tạo một file VMU riêng 'A1' trong thư mục lưu của RetroArch cho mỗi trò chơi được khởi chạy.\nCài đặt 'Tất cả VMU' tạo tối đa 8 file VMU riêng (A1/A2/B1/B2/C1/C2/D1/D2) cho mỗi trò chơi được khởi chạy."
#define OPTION_VAL_VMU_A1_VN NULL
#define OPTION_VAL_ALL_VMUS_VN NULL
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_VN "Âm thanh của Bộ nhớ ảo/VMU"
#define CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_VN "Âm thanh VMU"
#define CORE_OPTION_NAME_VMU_SOUND_INFO_0_VN "Khi bật, VMU sẽ phát tiếng bíp."
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_VN "Bộ lưu trữ VMU DreamPotato"
#define CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_VN "Cho phép lưu trữ dữ liệu VMU trên trình giả lập DreamPotato đã kết nối. Khi được bật, tệp lưu VMU cục bộ sẽ không được sử dụng."
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_VN "Hiển thị cài đặt Bộ Nhớ Hình Ảnh/Hệ thống (VMU)"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_VN "Hiển thị Cài đặt VMU"
#define CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_VN "Bật cấu hình hiển thị màn hình LCD VMU giả lập, bao gồm: độ hiển thị, kích thước, vị trí và màu sắc.LƯU Ý: Có thể cần mở lại Menu Nhanh để thiết lập có hiệu lực."
#define CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_VN "Hiển thị Màn hình VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_VN "Vị trí Màn hình VMU 1"
#define OPTION_VAL_UPPER_LEFT_VN "Trên cùng bên trái"
#define OPTION_VAL_UPPER_RIGHT_VN "Trên cùng bên phải"
#define OPTION_VAL_LOWER_LEFT_VN "Dưới cùng bên trái"
#define OPTION_VAL_LOWER_RIGHT_VN "Dưới cùng bên phải"
#define CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_VN "Kích thước Màn hình VMU 1"
#define OPTION_VAL_1X_VN NULL
#define OPTION_VAL_3X_VN NULL
#define OPTION_VAL_5X_VN NULL
#define CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_VN "Màu pixel BẬT của Màn hình VMU 1"
#define OPTION_VAL_DEFAULT_ON_00_VN "Bật Mặc định"
#define OPTION_VAL_DEFAULT_OFF_01_VN "Tắt Mặc định"
#define OPTION_VAL_BLACK_02_VN "Đen"
#define OPTION_VAL_LIGHT_BLUE_04_VN "Xanh dương nhạt"
#define OPTION_VAL_CYAN_06_VN "Lục lam"
#define OPTION_VAL_CYAN_BLUE_07_VN "Xanh lơ"
#define OPTION_VAL_LIGHT_GREEN_08_VN "Xanh lá nhạt"
#define OPTION_VAL_CYAN_GREEN_09_VN "Xanh lơ lá"
#define OPTION_VAL_LIGHT_CYAN_10_VN "Xanh lơ nhạt"
#define OPTION_VAL_PURPLE_12_VN "Tím"
#define OPTION_VAL_LIGHT_PURPLE_13_VN "Tím nhạt"
#define OPTION_VAL_YELLOW_14_VN "Vàng"
#define OPTION_VAL_GRAY_15_VN "Xám"
#define OPTION_VAL_LIGHT_PURPLE_2_16_VN "Tím nhạt (2)"
#define OPTION_VAL_LIGHT_GREEN_2_17_VN "Xanh lá nhạt (2)"
#define OPTION_VAL_LIGHT_GREEN_3_18_VN "Xanh lá nhạt (3)"
#define OPTION_VAL_LIGHT_CYAN_2_19_VN "Xanh lam nhạt (2)"
#define OPTION_VAL_LIGHT_RED_2_20_VN "Đỏ nhạt (2)"
#define OPTION_VAL_MAGENTA_21_VN "Hồng tím"
#define OPTION_VAL_LIGHT_PURPLE_3_22_VN "Tím nhạt (3)"
#define OPTION_VAL_LIGHT_ORANGE_23_VN "Cam nhạt"
#define OPTION_VAL_ORANGE_24_VN "Cam"
#define OPTION_VAL_LIGHT_PURPLE_4_25_VN "Tím nhạt (4)"
#define OPTION_VAL_LIGHT_YELLOW_26_VN "Vàng nhạt"
#define OPTION_VAL_LIGHT_YELLOW_2_27_VN "Vàng nhạt (2)"
#define CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_VN "Màu Pixel Tắt của Màn hình VMU 1"
#define CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_VN "Độ Mờ của Màn hình VMU 1"
#define OPTION_VAL_40_VN NULL
#define CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_VN "Hiển thị Màn hình VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_VN "Vị trí Màn hình VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_VN "Kích thước Màn hình VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_VN "Màu pixel Bật của Màn hình VMU 2"
#define CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_VN "Màu pixel Tắt của Màn hình VMU 2"
#define CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_VN "Độ mờ của Màn hình VMU 2"
#define CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_VN "Hiển thị Màn hình VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_VN "Vị trí Màn hình VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_VN "Kích thước Màn hình VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_VN "Màu pixel Bật của Màn hình VMU 3"
#define CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_VN "Màu pixel Tắt của Màn hình VMU 3"
#define CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_VN "Độ mờ của Màn hình VMU 3"
#define CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_VN "Hiển thị Màn hình VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_VN "Vị trí Màn hình VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_VN "Kích thước Màn hình VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_VN "Màu Pixel Bật của Màn hình VMU 4"
#define CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_VN "Màu Pixel Tắt của Màn hình VMU 4"
#define CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_VN "Độ mờ của Màn hình VMU 4"

struct retro_core_option_v2_category option_cats_vn[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_VN,
      CATEGORY_SYSTEM_INFO_0_VN
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_VN,
      CATEGORY_VIDEO_INFO_0_VN
   },
   {
      "performance",
      CATEGORY_PERFORMANCE_LABEL_VN,
      CATEGORY_PERFORMANCE_INFO_0_VN
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_VN,
      CATEGORY_HACKS_INFO_0_VN
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_VN,
      CATEGORY_INPUT_INFO_0_VN
   },
   {
      "expansions",
      CATEGORY_EXPANSIONS_LABEL_VN,
      CATEGORY_EXPANSIONS_INFO_0_VN
   },
   {
      "vmu",
      CATEGORY_VMU_LABEL_VN,
      CATEGORY_VMU_INFO_0_VN
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_vn[] = {
   {
      CORE_OPTION_NAME "_region",
      CORE_OPTION_NAME_REGION_LABEL_VN,
      NULL,
      "",
      NULL,
      "system",
      {
         { "Japan",   OPTION_VAL_JAPAN_VN },
         { "USA",     OPTION_VAL_USA_VN },
         { "Europe",  OPTION_VAL_EUROPE_VN },
         { "Default", OPTION_VAL_DEFAULT_VN },
         { NULL, NULL },
      },
      "USA",
   },
   {
      CORE_OPTION_NAME "_language",
      CORE_OPTION_NAME_LANGUAGE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_LANGUAGE_INFO_0_VN,
      NULL,
      "system",
      {
         { "Japanese", OPTION_VAL_JAPANESE_VN },
         { "English",  OPTION_VAL_ENGLISH_VN },
         { "German",   OPTION_VAL_GERMAN_VN },
         { "French",   OPTION_VAL_FRENCH_VN },
         { "Spanish",  OPTION_VAL_SPANISH_VN },
         { "Italian",  OPTION_VAL_ITALIAN_VN },
         { "Default",  OPTION_VAL_DEFAULT_VN },
         { NULL, NULL },
      },
      "English",
   },
   {
      CORE_OPTION_NAME "_hle_bios",
      CORE_OPTION_NAME_HLE_BIOS_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_HLE_BIOS_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled",  NULL },
         { "enabled",  NULL },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_enable_dsp",
      CORE_OPTION_NAME_ENABLE_DSP_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_ENABLE_DSP_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "disabled",
#else
      "enabled",
#endif
   },
   {
      CORE_OPTION_NAME "_allow_service_buttons",
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_ALLOW_SERVICE_BUTTONS_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_force_freeplay",
      CORE_OPTION_NAME_FORCE_FREEPLAY_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_FORCE_FREEPLAY_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_emulate_bba",
      CORE_OPTION_NAME_EMULATE_BBA_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_EMULATE_BBA_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_upnp",
      CORE_OPTION_NAME_UPNP_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_UPNP_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_dcnet",
      CORE_OPTION_NAME_DCNET_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DCNET_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },

   {
      CORE_OPTION_NAME "_internal_resolution",
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_INTERNAL_RESOLUTION_INFO_0_VN,
      NULL,
      "video",
      {
         { "320x240",    OPTION_VAL_320X240_VN },
         { "640x480",    OPTION_VAL_640X480_VN },
         { "800x600",    OPTION_VAL_800X600_VN },
         { "960x720",    OPTION_VAL_960X720_VN },
         { "1024x768",   OPTION_VAL_1024X768_VN },
         { "1280x960",   OPTION_VAL_1280X960_VN },
         { "1440x1080",  OPTION_VAL_1440X1080_VN },
         { "1600x1200",  OPTION_VAL_1600X1200_VN },
         { "1920x1440",  OPTION_VAL_1920X1440_VN },
         { "2560x1920",  OPTION_VAL_2560X1920_VN },
         { "2880x2160",  OPTION_VAL_2880X2160_VN },
         { "3200x2400",  OPTION_VAL_3200X2400_VN },
         { "3840x2880",  OPTION_VAL_3840X2880_VN },
         { "4480x3360",  OPTION_VAL_4480X3360_VN },
         { "5120x3840",  OPTION_VAL_5120X3840_VN },
         { "5760x4320",  OPTION_VAL_5760X4320_VN },
         { "6400x4800",  OPTION_VAL_6400X4800_VN },
         { "7040x5280",  OPTION_VAL_7040X5280_VN },
         { "7680x5760",  OPTION_VAL_7680X5760_VN },
         { "8320x6240",  OPTION_VAL_8320X6240_VN },
         { "8960x6720",  OPTION_VAL_8960X6720_VN },
         { "9600x7200",  OPTION_VAL_9600X7200_VN },
         { "10240x7680", OPTION_VAL_10240X7680_VN },
         { "10880x8160", OPTION_VAL_10880X8160_VN },
         { "11520x8640", OPTION_VAL_11520X8640_VN },
         { "12160x9120", OPTION_VAL_12160X9120_VN },
         { "12800x9600", OPTION_VAL_12800X9600_VN },
         { NULL, NULL },
      },
#ifdef LOW_RES
      "320x240",
#else
      "640x480",
#endif
   },
   {
      CORE_OPTION_NAME "_cable_type",
      CORE_OPTION_NAME_CABLE_TYPE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_CABLE_TYPE_INFO_0_VN,
      NULL,
      "video",
      {
         { "VGA",	    	 OPTION_VAL_VGA_VN },
         { "TV (RGB)",       OPTION_VAL_TV_RGB_VN },
         { "TV (Composite)", OPTION_VAL_TV_COMPOSITE_VN },
         { NULL, NULL },
      },
      "TV (Composite)",
   },
   {
      CORE_OPTION_NAME "_broadcast",
      CORE_OPTION_NAME_BROADCAST_LABEL_VN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "NTSC",    OPTION_VAL_NTSC_VN },
         { "PAL",     OPTION_VAL_PAL_VN },
         { "PAL_N",   OPTION_VAL_PAL_N_VN },
         { "PAL_M",   OPTION_VAL_PAL_M_VN },
         { "Default", OPTION_VAL_DEFAULT_VN },
         { NULL, NULL },
      },
      "NTSC",
   },
   {
      CORE_OPTION_NAME "_screen_rotation",
      CORE_OPTION_NAME_SCREEN_ROTATION_LABEL_VN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "horizontal", OPTION_VAL_HORIZONTAL_VN },
         { "vertical",   OPTION_VAL_VERTICAL_VN },
         { NULL, NULL },
      },
      "horizontal",
   },
   {/* TODO: needs better explanation? */
      CORE_OPTION_NAME "_alpha_sorting",
      CORE_OPTION_NAME_ALPHA_SORTING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_ALPHA_SORTING_INFO_0_VN,
      NULL,
      "video",
      {
         { "per-strip (fast, least accurate)", OPTION_VAL_PER_STRIP_FAST_LEAST_ACCURATE_VN },
         { "per-triangle (normal)",            OPTION_VAL_PER_TRIANGLE_NORMAL_VN },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
         { "per-pixel (accurate)",             OPTION_VAL_PER_PIXEL_ACCURATE_VN },
#endif
         { NULL, NULL },
      },
#if defined(LOW_END)
      "per-strip (fast, least accurate)",
#else
      "per-triangle (normal)",
#endif
   },
#if defined(HAVE_OIT) || defined(HAVE_VULKAN) || defined(HAVE_D3D11)
   {
      CORE_OPTION_NAME "_oit_abuffer_size",
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_OIT_ABUFFER_SIZE_INFO_0_VN,
      NULL,
      "video",
      {
         { "512MB", OPTION_VAL_512MB_VN },
         { "1GB",   OPTION_VAL_1GB_VN },
         { "2GB",   OPTION_VAL_2GB_VN },
         { "4GB",   OPTION_VAL_4GB_VN },
         { NULL, NULL },
      },
      "512MB",
   },
   {
      CORE_OPTION_NAME "_oit_layers",
      CORE_OPTION_NAME_OIT_LAYERS_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_OIT_LAYERS_INFO_0_VN,
      NULL,
      "video",
      {
         { "8", NULL },
         { "16",   NULL },
         { "32",   NULL },
         { "64",   NULL },
         { "96",   NULL },
         { "128",   NULL },
         { NULL, NULL },
      },
      "32",
   },
#endif
   {
      CORE_OPTION_NAME "_emulate_framebuffer",
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_EMULATE_FRAMEBUFFER_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {/* TODO: needs explanation */
      CORE_OPTION_NAME "_enable_rttb",
      CORE_OPTION_NAME_ENABLE_RTTB_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_ENABLE_RTTB_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_mipmapping",
      CORE_OPTION_NAME_MIPMAPPING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_MIPMAPPING_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_fog",
      CORE_OPTION_NAME_FOG_LABEL_VN,
      NULL,
      "",
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_volume_modifier_enable",
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_VOLUME_MODIFIER_ENABLE_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_anisotropic_filtering",
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_ANISOTROPIC_FILTERING_INFO_0_VN,
      NULL,
      "video",
      {
         { "off", "disabled" },
         { "2",  NULL },
         { "4",  NULL },
         { "8",  NULL },
         { "16",  NULL },
         { NULL, NULL },
      },
      "4",
   },
   {
      CORE_OPTION_NAME "_texture_filtering",
      CORE_OPTION_NAME_TEXTURE_FILTERING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_TEXTURE_FILTERING_INFO_0_VN,
      NULL,
      "video",
      {
         { "0", OPTION_VAL_DEFAULT_VN },
         { "1",  OPTION_VAL_1_VN },
         { "2",  OPTION_VAL_2_VN },
         { NULL, NULL },
      },
      "0",
   },
   {
      CORE_OPTION_NAME "_delay_frame_swapping",
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DELAY_FRAME_SWAPPING_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_detect_vsync_swap_interval",
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DETECT_VSYNC_SWAP_INTERVAL_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_pvr2_filtering",
      CORE_OPTION_NAME_PVR2_FILTERING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_PVR2_FILTERING_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
#ifdef _OPENMP
   {
      CORE_OPTION_NAME "_texupscale",
      CORE_OPTION_NAME_TEXUPSCALE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_INFO_0_VN,
      NULL,
      "video",
      {
         { "1", "disabled" },
         { "2",  OPTION_VAL_2_O26_VN },
         { "4",  OPTION_VAL_4_VN },
         { "6",  OPTION_VAL_6_VN },
         { NULL, NULL },
      },
      "1",
   },
   {
      CORE_OPTION_NAME "_texupscale_max_filtered_texture_size",
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_TEXUPSCALE_MAX_FILTERED_TEXTURE_SIZE_INFO_0_VN,
      NULL,
      "video",
      {
         { "256",  NULL },
         { "512",  NULL },
         { "1024", NULL },
         { NULL, NULL },
      },
      "256",
   },
#endif
   {
      CORE_OPTION_NAME "_native_depth_interpolation",
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_LABEL_VN,
	  NULL,
	  CORE_OPTION_NAME_NATIVE_DEPTH_INTERPOLATION_INFO_0_VN,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_fix_upscale_bleeding_edge",
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_LABEL_VN,
	  NULL,
	  CORE_OPTION_NAME_FIX_UPSCALE_BLEEDING_EDGE_INFO_0_VN,
	  NULL,
	  "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_threaded_rendering",
      CORE_OPTION_NAME_THREADED_RENDERING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_THREADED_RENDERING_INFO_0_VN,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled",
   },
   {
      CORE_OPTION_NAME "_auto_skip_frame",
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_AUTO_SKIP_FRAME_INFO_0_VN,
      NULL,
      "performance",
      {
         { "disabled", NULL },
         { "some", OPTION_VAL_SOME_VN },
         { "more", OPTION_VAL_MORE_VN },
         { NULL, NULL },
      },
#ifdef LOW_END
      "some",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_frame_skipping",
      CORE_OPTION_NAME_FRAME_SKIPPING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_FRAME_SKIPPING_INFO_0_VN,
      NULL,
      "performance",
      {
         { "disabled",  NULL },
         { "1",         NULL },
         { "2",         NULL },
         { "3",         NULL },
         { "4",         NULL },
         { "5",         NULL },
         { "6",         NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_cheats",
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_CHEATS_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_widescreen_hack",
      CORE_OPTION_NAME_WIDESCREEN_HACK_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_WIDESCREEN_HACK_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_gdrom_fast_loading",
      CORE_OPTION_NAME_GDROM_FAST_LOADING_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_GDROM_FAST_LOADING_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
#ifdef LOW_END
      "enabled",
#else
      "disabled",
#endif
   },
   {
      CORE_OPTION_NAME "_dc_32mb_mod",
      CORE_OPTION_NAME_DC_32MB_MOD_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DC_32MB_MOD_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled", NULL },
         {  NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_sh4clock",
      CORE_OPTION_NAME_SH4CLOCK_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_SH4CLOCK_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "100", OPTION_VAL_100_VN },
         { "110", OPTION_VAL_110_VN },
         { "120", OPTION_VAL_120_VN },
         { "130", OPTION_VAL_130_VN },
         { "140", OPTION_VAL_140_VN },
         { "150", OPTION_VAL_150_VN },
         { "160", OPTION_VAL_160_VN },
         { "170", OPTION_VAL_170_VN },
         { "180", OPTION_VAL_180_VN },
         { "190", OPTION_VAL_190_VN },
         { "200", OPTION_VAL_200_VN },
         { "210", OPTION_VAL_210_VN },
         { "220", OPTION_VAL_220_VN },
         { "230", OPTION_VAL_230_VN },
         { "240", OPTION_VAL_240_VN },
         { "250", OPTION_VAL_250_VN },
         { "260", OPTION_VAL_260_VN },
         { "270", OPTION_VAL_270_VN },
         { "280", OPTION_VAL_280_VN },
         { "290", OPTION_VAL_290_VN },
         { "300", OPTION_VAL_300_VN },
         { "310", OPTION_VAL_310_VN },
         { "320", OPTION_VAL_320_VN },
         { "330", OPTION_VAL_330_VN },
         { "340", OPTION_VAL_340_VN },
         { "350", OPTION_VAL_350_VN },
         { "360", OPTION_VAL_360_VN },
         { "370", OPTION_VAL_370_VN },
         { "380", OPTION_VAL_380_VN },
         { "390", OPTION_VAL_390_VN },
         { "400", OPTION_VAL_400_VN },
         { "410", OPTION_VAL_410_VN },
         { "420", OPTION_VAL_420_VN },
         { "430", OPTION_VAL_430_VN },
         { "440", OPTION_VAL_440_VN },
         { "450", OPTION_VAL_450_VN },
         { "460", OPTION_VAL_460_VN },
         { "470", OPTION_VAL_470_VN },
         { "480", OPTION_VAL_480_VN },
         { "490", OPTION_VAL_490_VN },
         { "500", OPTION_VAL_500_VN },
         { NULL, NULL },
      },
      "200",
   },
   {
      CORE_OPTION_NAME "_custom_textures",
      CORE_OPTION_NAME_CUSTOM_TEXTURES_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_CUSTOM_TEXTURES_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_preload_custom_textures",
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_PRELOAD_CUSTOM_TEXTURES_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_textures",
      CORE_OPTION_NAME_DUMP_TEXTURES_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DUMP_TEXTURES_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_dump_replaced_textures",
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DUMP_REPLACED_TEXTURES_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_analog_stick_deadzone",
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_ANALOG_STICK_DEADZONE_INFO_0_VN,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_VN },
         { "5%",  OPTION_VAL_5_VN },
         { "10%", OPTION_VAL_10_VN },
         { "15%", OPTION_VAL_15_VN },
         { "20%", OPTION_VAL_20_VN },
         { "25%", OPTION_VAL_25_VN },
         { "30%", OPTION_VAL_30_VN },
         { NULL, NULL },
      },
      "15%",
   },
   {
      CORE_OPTION_NAME "_trigger_deadzone",
      CORE_OPTION_NAME_TRIGGER_DEADZONE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_TRIGGER_DEADZONE_INFO_0_VN,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_VN },
         { "5%",  OPTION_VAL_5_VN },
         { "10%", OPTION_VAL_10_VN },
         { "15%", OPTION_VAL_15_VN },
         { "20%", OPTION_VAL_20_VN },
         { "25%", OPTION_VAL_25_VN },
         { "30%", OPTION_VAL_30_VN },
         { NULL, NULL },
      },
      "0%",
   },
   {
      CORE_OPTION_NAME "_digital_triggers",
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DIGITAL_TRIGGERS_INFO_0_VN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
	  CORE_OPTION_NAME "_network_output",
      CORE_OPTION_NAME_NETWORK_OUTPUT_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_NETWORK_OUTPUT_INFO_0_VN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_lightgun_settings",
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_SHOW_LIGHTGUN_SETTINGS_INFO_0_VN,
      NULL,
      "input",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_lightgun_crosshair_size_scaling",
      CORE_OPTION_NAME_LIGHTGUN_CROSSHAIR_SIZE_SCALING_LABEL_VN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "50%",  OPTION_VAL_50_VN },
         { "60%",  OPTION_VAL_60_VN },
         { "70%",  OPTION_VAL_70_VN },
         { "80%",  OPTION_VAL_80_VN },
         { "90%",  OPTION_VAL_90_VN },
         { "100%", OPTION_VAL_100_O47_VN },
         { "110%", OPTION_VAL_110_O47_VN },
         { "120%", OPTION_VAL_120_O47_VN },
         { "130%", OPTION_VAL_130_O47_VN },
         { "140%", OPTION_VAL_140_O47_VN },
         { "150%", OPTION_VAL_150_O47_VN },
         { "160%", OPTION_VAL_160_O47_VN },
         { "170%", OPTION_VAL_170_O47_VN },
         { "180%", OPTION_VAL_180_O47_VN },
         { "190%", OPTION_VAL_190_O47_VN },
         { "200%", OPTION_VAL_200_O47_VN },
         { "210%", OPTION_VAL_210_O47_VN },
         { "220%", OPTION_VAL_220_O47_VN },
         { "230%", OPTION_VAL_230_O47_VN },
         { "240%", OPTION_VAL_240_O47_VN },
         { "250%", OPTION_VAL_250_O47_VN },
         { "260%", OPTION_VAL_260_O47_VN },
         { "270%", OPTION_VAL_270_O47_VN },
         { "280%", OPTION_VAL_280_O47_VN },
         { "290%", OPTION_VAL_290_O47_VN },
         { "300%", OPTION_VAL_300_O47_VN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_lightgun1_crosshair",
      CORE_OPTION_NAME_LIGHTGUN1_CROSSHAIR_LABEL_VN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VN },
         { "Red",      OPTION_VAL_RED_VN },
         { "Green",    OPTION_VAL_GREEN_VN },
         { "Blue",     OPTION_VAL_BLUE_VN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun2_crosshair",
      CORE_OPTION_NAME_LIGHTGUN2_CROSSHAIR_LABEL_VN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VN },
         { "Red",      OPTION_VAL_RED_VN },
         { "Green",    OPTION_VAL_GREEN_VN },
         { "Blue",     OPTION_VAL_BLUE_VN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun3_crosshair",
      CORE_OPTION_NAME_LIGHTGUN3_CROSSHAIR_LABEL_VN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VN },
         { "Red",      OPTION_VAL_RED_VN },
         { "Green",    OPTION_VAL_GREEN_VN },
         { "Blue",     OPTION_VAL_BLUE_VN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_lightgun4_crosshair",
      CORE_OPTION_NAME_LIGHTGUN4_CROSSHAIR_LABEL_VN,
      NULL,
      "",
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "White",    OPTION_VAL_WHITE_VN },
         { "Red",      OPTION_VAL_RED_VN },
         { "Green",    OPTION_VAL_GREEN_VN },
         { "Blue",     OPTION_VAL_BLUE_VN },
         { NULL,       NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot1",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT1_INFO_0_VN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VN },
         { "Purupuru", OPTION_VAL_PURUPURU_VN },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_VN },
         { "None",     OPTION_VAL_NONE_VN },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port1_slot2",
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT1_SLOT2_INFO_0_VN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VN },
         { "Purupuru", OPTION_VAL_PURUPURU_VN },
         { "None",     OPTION_VAL_NONE_VN },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot1",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT1_INFO_0_VN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VN },
         { "Purupuru", OPTION_VAL_PURUPURU_VN },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_VN },
         { "None",     OPTION_VAL_NONE_VN },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port2_slot2",
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT2_SLOT2_INFO_0_VN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VN },
         { "Purupuru", OPTION_VAL_PURUPURU_VN },
         { "None",     OPTION_VAL_NONE_VN },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot1",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT1_INFO_0_VN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VN },
         { "Purupuru", OPTION_VAL_PURUPURU_VN },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_VN },
         { "None",     OPTION_VAL_NONE_VN },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port3_slot2",
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT3_SLOT2_INFO_0_VN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VN },
         { "Purupuru", OPTION_VAL_PURUPURU_VN },
         { "None",     OPTION_VAL_NONE_VN },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot1",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT1_INFO_0_VN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VN },
         { "Purupuru", OPTION_VAL_PURUPURU_VN },
         { "DreamPotato", OPTION_VAL_DREAMPOTATO_VN },
         { "None",     OPTION_VAL_NONE_VN },
         { NULL, NULL },
      },
      "VMU",
   },
   {
      CORE_OPTION_NAME "_device_port4_slot2",
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_DEVICE_PORT4_SLOT2_INFO_0_VN,
      NULL,
      "expansions",
      {
         { "VMU",      OPTION_VAL_VMU_VN },
         { "Purupuru", OPTION_VAL_PURUPURU_VN },
         { "None",     OPTION_VAL_NONE_VN },
         { NULL, NULL },
      },
      "Purupuru",
   },
   {
      CORE_OPTION_NAME "_per_content_vmus",
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_VN,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_LABEL_CAT_VN,
      CORE_OPTION_NAME_PER_CONTENT_VMUS_INFO_0_VN,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "VMU A1",   OPTION_VAL_VMU_A1_VN },
         { "All VMUs", OPTION_VAL_ALL_VMUS_VN },
         { NULL, NULL},
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu_sound",
      CORE_OPTION_NAME_VMU_SOUND_LABEL_VN,
      CORE_OPTION_NAME_VMU_SOUND_LABEL_CAT_VN,
      CORE_OPTION_NAME_VMU_SOUND_INFO_0_VN,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_linked_vmu_storage",
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_LABEL_VN,
      NULL,
      CORE_OPTION_NAME_LINKED_VMU_STORAGE_INFO_0_VN,
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",   NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_show_vmu_screen_settings",
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_VN,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_LABEL_CAT_VN,
      CORE_OPTION_NAME_SHOW_VMU_SCREEN_SETTINGS_INFO_0_VN,
      NULL,
      "vmu",
      {
         { "enabled",  NULL },
         { "disabled", NULL },
         { NULL, NULL},
      },
      "disabled"
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_display",
      CORE_OPTION_NAME_VMU1_SCREEN_DISPLAY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_position",
      CORE_OPTION_NAME_VMU1_SCREEN_POSITION_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VN },
         { NULL, NULL },
      },
      "Upper Left",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_size_mult",
      CORE_OPTION_NAME_VMU1_SCREEN_SIZE_MULT_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VN },
         { "2x", OPTION_VAL_2_O26_VN },
         { "3x", OPTION_VAL_3X_VN },
         { "4x", OPTION_VAL_4_VN },
         { "5x", OPTION_VAL_5X_VN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_on_color",
      CORE_OPTION_NAME_VMU1_PIXEL_ON_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu1_pixel_off_color",
      CORE_OPTION_NAME_VMU1_PIXEL_OFF_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu1_screen_opacity",
      CORE_OPTION_NAME_VMU1_SCREEN_OPACITY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VN },
         { "20%",  OPTION_VAL_20_VN },
         { "30%",  OPTION_VAL_30_VN },
         { "40%",  OPTION_VAL_40_VN },
         { "50%",  OPTION_VAL_50_VN },
         { "60%",  OPTION_VAL_60_VN },
         { "70%",  OPTION_VAL_70_VN },
         { "80%",  OPTION_VAL_80_VN },
         { "90%",  OPTION_VAL_90_VN },
         { "100%", OPTION_VAL_100_O47_VN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_display",
      CORE_OPTION_NAME_VMU2_SCREEN_DISPLAY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_position",
      CORE_OPTION_NAME_VMU2_SCREEN_POSITION_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VN },
         { NULL, NULL },
      },
      "Upper Right",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_size_mult",
      CORE_OPTION_NAME_VMU2_SCREEN_SIZE_MULT_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VN },
         { "2x", OPTION_VAL_2_O26_VN },
         { "3x", OPTION_VAL_3X_VN },
         { "4x", OPTION_VAL_4_VN },
         { "5x", OPTION_VAL_5X_VN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_on_color",
      CORE_OPTION_NAME_VMU2_PIXEL_ON_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu2_pixel_off_color",
      CORE_OPTION_NAME_VMU2_PIXEL_OFF_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu2_screen_opacity",
      CORE_OPTION_NAME_VMU2_SCREEN_OPACITY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VN },
         { "20%",  OPTION_VAL_20_VN },
         { "30%",  OPTION_VAL_30_VN },
         { "40%",  OPTION_VAL_40_VN },
         { "50%",  OPTION_VAL_50_VN },
         { "60%",  OPTION_VAL_60_VN },
         { "70%",  OPTION_VAL_70_VN },
         { "80%",  OPTION_VAL_80_VN },
         { "90%",  OPTION_VAL_90_VN },
         { "100%", OPTION_VAL_100_O47_VN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_display",
      CORE_OPTION_NAME_VMU3_SCREEN_DISPLAY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_position",
      CORE_OPTION_NAME_VMU3_SCREEN_POSITION_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VN },
         { NULL, NULL },
      },
      "Lower Left",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_size_mult",
      CORE_OPTION_NAME_VMU3_SCREEN_SIZE_MULT_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VN },
         { "2x", OPTION_VAL_2_O26_VN },
         { "3x", OPTION_VAL_3X_VN },
         { "4x", OPTION_VAL_4_VN },
         { "5x", OPTION_VAL_5X_VN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_on_color",
      CORE_OPTION_NAME_VMU3_PIXEL_ON_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu3_pixel_off_color",
      CORE_OPTION_NAME_VMU3_PIXEL_OFF_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu3_screen_opacity",
      CORE_OPTION_NAME_VMU3_SCREEN_OPACITY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VN },
         { "20%",  OPTION_VAL_20_VN },
         { "30%",  OPTION_VAL_30_VN },
         { "40%",  OPTION_VAL_40_VN },
         { "50%",  OPTION_VAL_50_VN },
         { "60%",  OPTION_VAL_60_VN },
         { "70%",  OPTION_VAL_70_VN },
         { "80%",  OPTION_VAL_80_VN },
         { "90%",  OPTION_VAL_90_VN },
         { "100%", OPTION_VAL_100_O47_VN },
         { NULL,   NULL },
      },
      "100%",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_display",
      CORE_OPTION_NAME_VMU4_SCREEN_DISPLAY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_position",
      CORE_OPTION_NAME_VMU4_SCREEN_POSITION_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "Upper Left",  OPTION_VAL_UPPER_LEFT_VN },
         { "Upper Right", OPTION_VAL_UPPER_RIGHT_VN },
         { "Lower Left",  OPTION_VAL_LOWER_LEFT_VN },
         { "Lower Right", OPTION_VAL_LOWER_RIGHT_VN },
         { NULL, NULL },
      },
      "Lower Right",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_size_mult",
      CORE_OPTION_NAME_VMU4_SCREEN_SIZE_MULT_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "1x", OPTION_VAL_1X_VN },
         { "2x", OPTION_VAL_2_O26_VN },
         { "3x", OPTION_VAL_3X_VN },
         { "4x", OPTION_VAL_4_VN },
         { "5x", OPTION_VAL_5X_VN },
         { NULL, NULL },
      },
      "1x",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_on_color",
      CORE_OPTION_NAME_VMU4_PIXEL_ON_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_ON 00",
   },
   {
      CORE_OPTION_NAME "_vmu4_pixel_off_color",
      CORE_OPTION_NAME_VMU4_PIXEL_OFF_COLOR_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "DEFAULT_OFF 01", OPTION_VAL_DEFAULT_OFF_01_VN },
         { "DEFAULT_ON 00",  OPTION_VAL_DEFAULT_ON_00_VN },
         { "BLACK 02",          OPTION_VAL_BLACK_02_VN },
         { "BLUE 03",           OPTION_VAL_BLUE_VN },
         { "LIGHT_BLUE 04",     OPTION_VAL_LIGHT_BLUE_04_VN },
         { "GREEN 05",          OPTION_VAL_GREEN_VN },
         { "CYAN 06",           OPTION_VAL_CYAN_06_VN },
         { "CYAN_BLUE 07",      OPTION_VAL_CYAN_BLUE_07_VN },
         { "LIGHT_GREEN 08",    OPTION_VAL_LIGHT_GREEN_08_VN },
         { "CYAN_GREEN 09",     OPTION_VAL_CYAN_GREEN_09_VN },
         { "LIGHT_CYAN 10",     OPTION_VAL_LIGHT_CYAN_10_VN },
         { "RED 11",            OPTION_VAL_RED_VN },
         { "PURPLE 12",         OPTION_VAL_PURPLE_12_VN },
         { "LIGHT_PURPLE 13",   OPTION_VAL_LIGHT_PURPLE_13_VN },
         { "YELLOW 14",         OPTION_VAL_YELLOW_14_VN },
         { "GRAY 15",           OPTION_VAL_GRAY_15_VN },
         { "LIGHT_PURPLE_2 16", OPTION_VAL_LIGHT_PURPLE_2_16_VN },
         { "LIGHT_GREEN_2 17",  OPTION_VAL_LIGHT_GREEN_2_17_VN },
         { "LIGHT_GREEN_3 18",  OPTION_VAL_LIGHT_GREEN_3_18_VN },
         { "LIGHT_CYAN_2 19",   OPTION_VAL_LIGHT_CYAN_2_19_VN },
         { "LIGHT_RED_2 20",    OPTION_VAL_LIGHT_RED_2_20_VN },
         { "MAGENTA 21",        OPTION_VAL_MAGENTA_21_VN },
         { "LIGHT_PURPLE_3 22",   OPTION_VAL_LIGHT_PURPLE_3_22_VN },
         { "LIGHT_ORANGE 23",   OPTION_VAL_LIGHT_ORANGE_23_VN },
         { "ORANGE 24",         OPTION_VAL_ORANGE_24_VN },
         { "LIGHT_PURPLE_4 25", OPTION_VAL_LIGHT_PURPLE_4_25_VN },
         { "LIGHT_YELLOW 26",   OPTION_VAL_LIGHT_YELLOW_26_VN },
         { "LIGHT_YELLOW_2 27", OPTION_VAL_LIGHT_YELLOW_2_27_VN },
         { "WHITE 28",          OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "DEFAULT_OFF 01",
   },
   {
      CORE_OPTION_NAME "_vmu4_screen_opacity",
      CORE_OPTION_NAME_VMU4_SCREEN_OPACITY_LABEL_VN,
      NULL,
      "",
      NULL,
      "vmu",
      {
         { "10%",  OPTION_VAL_10_VN },
         { "20%",  OPTION_VAL_20_VN },
         { "30%",  OPTION_VAL_30_VN },
         { "40%",  OPTION_VAL_40_VN },
         { "50%",  OPTION_VAL_50_VN },
         { "60%",  OPTION_VAL_60_VN },
         { "70%",  OPTION_VAL_70_VN },
         { "80%",  OPTION_VAL_80_VN },
         { "90%",  OPTION_VAL_90_VN },
         { "100%", OPTION_VAL_100_O47_VN },
         { NULL,   NULL },
      },
      "100%",
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_vn = {
   option_cats_vn,
   option_defs_vn
};


#ifdef __cplusplus
}
#endif

#endif